--==============================================================================
--
--     File Name: axi4_256_tb_pkg.vhd
--         Model: Simulation only
--
--  Dependencies: Set simulator resolution to ps for accurate simulation
--
--       Company: Alpha Data
--        Design: Package axi4_256_tb_pkg
--
--   Description: Package of components, procedures and functions for applying
--                testbench stimulus to a target FPGA using ADB3 OCP.
--
--   Limitations:
--
--         Notes:
--
--    Disclaimer: THESE DESIGNS ARE PROVIDED "AS IS" WITH NO WARRANTY
--                WHATSOEVER AND ALPHA DATA SPECIFICALLY DISCLAIMS ANY
--                IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR
--                A PARTICULAR PURPOSE, OR AGAINST INFRINGEMENT.
--
--                Copyright (C) 2015 Alpha Data, All rights reserved
--
--     Revisions: 0.1, KDC, 23 Oct 2015: Initial design
--
--==============================================================================

--
-- Exported data types:
--
--   byte_t
--   byte_enable_t
--   byte_vector_t
--
-- Exported functions (parenthesis indicated number of overloads):
--
--   conv_byte_vector
--   conv_byte_enable
--   conv_vector
--   conv_string_hex
--   conv_string (2)
--
-- Exported procedures
--
--   axi4_256_sim_read_reg32
--   axi4_256_sim_read
--   axi4_256_sim_read_nb
--   axi4_256_sim_read_cmd
--   axi4_256_sim_read_resp
--
--   axi4_256_sim_write_reg32
--   axi4_256_sim_write
--
--   axi4_256_sim_wait_cycles
--

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.numeric_std.all;

library work;
use work.axi4_256.all;

package axi4_256_tb_pkg is

  --------------------------
  -- Testbench data types --
  --------------------------

  -- A byte datatype
  subtype byte_t is std_logic_vector(7 downto 0);

  -- Array of std_logic, intended to hold byte enables in ACTIVE HIGH representation
  type byte_enable_t is array(natural range <>) of std_logic;

  -- Array of bytes, intended to hold data to be written to the
  -- local bus or read from the local bus.
  type byte_vector_t is array(natural range <>) of byte_t;


  ---------------------------
  -- Function declarations --
  ---------------------------

  -- Convert between std_logic_vector, byte_vector and integer datatypes.
  function conv_byte_vector(
    constant x : in std_logic_vector)
  return byte_vector_t;

  function conv_byte_enable(
    constant x : in std_logic_vector)
  return byte_enable_t;

  function conv_vector(
    constant x : in byte_vector_t)
  return std_logic_vector;

  function conv_string_hex(
    constant x : in byte_vector_t)
  return string;

  function conv_string_hex(
    constant x : in std_logic_vector)
  return string;

  function conv_string_bin(
    constant x : in byte_enable_t)
  return string;

  function conv_string_bin(
    constant x : in std_logic_vector)
  return string;


  -- ------------------------------------------------------------
  -- Public routines for performing ADB3 transfers
  -- ------------------------------------------------------------

  -- Perform an AXI 256 interface direct slave read.
  --
  --     address     Byte address (4-byte/16-byte aligned).
  --
  --     tag         Identifier for transfer
  --
  --     data        Data to be read. data'length indicates
  --                 desired number of bytes to read.
  --
  --     nxfered     Returns the actual number of bytes transferred.

  procedure axi4_256_sim_read_reg32(
           address      : in    axi4_256_addr_s;
           tag          : in    natural;
           data         : out   byte_vector_t;
           nxfered      : out   natural;
    signal axi_clk      : in    std_logic;
    signal axi4_256_m2s : out   axi4_256_m2s_t;
    signal axi4_256_s2m : in    axi4_256_s2m_t);

  -- Blocking read procedure
  -- Waits until all data has been read before completing
  procedure axi4_256_sim_read(
           address      : in    axi4_256_addr_s;
           tag          : in    natural;
           data         : out   byte_vector_t;
           nxfered      : out   natural;
    signal axi_clk      : in    std_logic;
    signal axi4_256_m2s : out   axi4_256_m2s_t;
    signal axi4_256_s2m : in    axi4_256_s2m_t);

  -- Non-Blocking read procedure
  -- Issues commands and receives responeses
--  procedure axi4_256_sim_read_nb(
--    constant c_addr       : in    axi4_256_addr_s;
--    constant c_bl         : in    natural;
--    constant c_tag        : in    natural;
--    variable c_done       : out   boolean;
--    variable r_data       : out   byte_vector_t;
--    variable r_tag        : out   natural;
--    variable r_done       : out   boolean;
--    signal   axi_clk      : in    std_logic;
--    signal   axi4_256_m2s : out   axi4_256_m2s_t;
--    signal   axi4_256_s2m : in    axi4_256_s2m_t);

  procedure axi4_256_sim_read_nb(
           address      : in    axi4_256_addr_s;
           tag          : in    natural;
           data         : out   byte_vector_t;
           nxfered      : out   natural;
    signal axi_clk      : in    std_logic;
    signal axi4_256_m2s : out   axi4_256_m2s_t;
    signal axi4_256_s2m : in    axi4_256_s2m_t);

  -- Non-Blocking read command procedure
  -- Produces read commands only
  procedure axi4_256_sim_read_cmd(
           address      : in    axi4_256_addr_s;
           tag          : in    natural;
           data         : out   byte_vector_t;
           nxfered      : out   natural;
    signal axi_clk      : in    std_logic;
    signal axi4_256_m2s : out   axi4_256_m2s_t;
    signal axi4_256_s2m : in    axi4_256_s2m_t);

  -- Non-Blocking read response procedure
  -- waits for read response only
  procedure axi4_256_sim_read_resp(
           data         : out   byte_vector_t;
           tag          : out   natural;
           nxfered      : out   natural;
    signal axi_clk      : in    std_logic;
    signal axi4_256_m2s : out   axi4_256_m2s_t;
    signal axi4_256_s2m : in    axi4_256_s2m_t);


  -- Perform an AXI 256 interface direct slave write.
  --
  --     address     Byte address (4-byte/16-byte aligned).
  --
  --     tag         Identifier for transfer.
  --
  --     be          Byte enables. be'length must be the same as
  --                 data'length. Byte enables used are limited
  --                 to data bus width (16) as they are ignored
  --                 thereafter.
  --
  --     data        Data to be written. data'length indicates
  --                 desired number of bytes to write.
  --
  --     nxfered     Returns the actual number of bytes transferred.

  procedure axi4_256_sim_write_reg32(
           address      : in    axi4_256_addr_s;
           tag          : in    natural;
           be           : in    byte_enable_t;
           data         : in    byte_vector_t;
           nxfered      : out   natural;
    signal axi_clk      : in    std_logic;
    signal axi4_256_m2s : out   axi4_256_m2s_t;
    signal axi4_256_s2m : in    axi4_256_s2m_t);

  procedure axi4_256_sim_write(
           address      : in    axi4_256_addr_s;
           tag          : in    natural;
           be           : in    byte_enable_t;
           data         : in    byte_vector_t;
           nxfered      : out   natural;
    signal axi_clk      : in    std_logic;
    signal axi4_256_m2s : out   axi4_256_m2s_t;
    signal axi4_256_s2m : in    axi4_256_s2m_t);


  -- Delay for a specified number of clock cycles
  --
  --     cycles      Number of clock cycles of delay.
  --     clk         Clock signal to use for delay

  procedure axi4_256_sim_wait_cycles(
           cycles : in    natural;
    signal clk    : in    std_logic);

end;


package body axi4_256_tb_pkg is

  -- ---------------------------------------------------------------
  -- Private
  -- ---------------------------------------------------------------

  function conv_string_hexdigit(
    constant val : in std_logic_vector)
  return character is
    variable tmp : std_logic_vector(3 downto 0);
  begin
    assert not(val'length /= 4)
      report "(conv_string_hex_digit): input has invalid length " & integer'image(val'length) & " /= 4"
      severity failure;

    tmp := val;
    case tmp is
      when "0000" => return '0';
      when "0001" => return '1';
      when "0010" => return '2';
      when "0011" => return '3';
      when "0100" => return '4';
      when "0101" => return '5';
      when "0110" => return '6';
      when "0111" => return '7';
      when "1000" => return '8';
      when "1001" => return '9';
      when "1010" => return 'A';
      when "1011" => return 'B';
      when "1100" => return 'C';
      when "1101" => return 'D';
      when "1110" => return 'E';
      when "1111" => return 'F';
      when others => return 'X';
    end case;
  end;


  function conv_std_logic_char(
    constant val: in std_logic)
  return character is
  begin
    case val is
      when 'U' => return 'U';
      when 'X' => return 'X';
      when '0' => return '0';
      when '1' => return '1';
      when 'Z' => return 'Z';
      when 'W' => return 'W';
      when 'L' => return 'L';
      when 'H' => return 'H';
      when '-' => return '-';
    end case;
  end;


  -- ---------------------------------------------------------------
  -- Private ADB3 routines
  -- ---------------------------------------------------------------

  procedure axi4_256_sim_burst_read_nb(
    constant c_addr       : in    axi4_256_addr_s;
    constant c_bl         : in    natural;  -- Burst length 1..Max burst length. AXI len = BL-1
    constant c_tag        : in    natural;
    variable c_done       : out   boolean;
    variable r_data       : out   byte_vector_t;
    variable r_tag        : out   natural;
    variable r_done       : out   boolean;
    signal   axi_clk      : in    std_logic;
    signal   axi4_256_m2s : out   axi4_256_m2s_t;
    signal   axi4_256_s2m : in    axi4_256_s2m_t)
  is
    variable r_data_offset : natural := r_data'low;
    variable done_c : boolean := false;
    variable done_r : boolean := false;
  begin
    -- Setup valid command if command BL /= 0
    if (c_bl /= 0) then
      assert not(c_addr(3 downto 0) /= "0000")
        report "(axi4_256_sim_burst_read_nb): Address is not 32-byte aligned during AXI 256 read"
        severity error;

      axi4_256_m2s.arvalid  <= '1';
      axi4_256_m2s.araddr   <= std_logic_vector(RESIZE(unsigned(c_addr),AXI4_256_ADDR_WIDTH));
      axi4_256_m2s.arlen    <= std_logic_vector(TO_UNSIGNED(c_bl-1,AXI4_256_LEN_WIDTH));
      axi4_256_m2s.arid     <= std_logic_vector(TO_UNSIGNED(c_tag,AXI4_256_ID_WIDTH));
      axi4_256_m2s.arsize   <= AXI4_256_SIZE;
      axi4_256_m2s.arburst  <= AXI4_256_BURST;
      axi4_256_m2s.arcache  <= AXI4_256_CACHE;
      axi4_256_m2s.arprot   <= AXI4_256_PROT;
      axi4_256_m2s.arlock   <= AXI4_256_LOCK;
      axi4_256_m2s.arqos    <= AXI4_256_QOS;
      axi4_256_m2s.arregion <= AXI4_256_REGION;
    end if;

    -- Accept data until done_rd
    axi4_256_m2s.rready <= '1';

    while not(done_c or done_r) loop
      wait until rising_edge(axi_clk);

      if (c_bl /= 0)and(axi4_256_s2m.arready = '1') then
        done_c := true;
      else
        done_c := false;
      end if;

      if (axi4_256_s2m.rresp = AXI4_256_RESP_OKAY)and(axi4_256_s2m.rvalid = '1') then
        for i in 0 to (AXI4_256_STRB_WIDTH-1) loop
          r_data(r_data_offset) := axi4_256_s2m.rdata(8*i+7 downto 8*i);
          r_data_offset := r_data_offset + 1;
        end loop;
        r_tag  := TO_INTEGER(UNSIGNED(axi4_256_s2m.rid));
        done_r := true;
      end if;
    end loop;

    c_done := done_c;
    r_done := done_r;

    -- Set command and accept to default values
    axi4_256_m2s.arvalid  <= AXI4_256_DEFAULT_M2S.arvalid;
    axi4_256_m2s.araddr   <= AXI4_256_DEFAULT_M2S.araddr;
    axi4_256_m2s.arlen    <= AXI4_256_DEFAULT_M2S.arlen;
    axi4_256_m2s.arid     <= AXI4_256_DEFAULT_M2S.arid;
    axi4_256_m2s.arsize   <= AXI4_256_DEFAULT_M2S.arsize;  
    axi4_256_m2s.arburst  <= AXI4_256_DEFAULT_M2S.arburst;
    axi4_256_m2s.arcache  <= AXI4_256_DEFAULT_M2S.arcache;
    axi4_256_m2s.arprot   <= AXI4_256_DEFAULT_M2S.arprot;
    axi4_256_m2s.arlock   <= AXI4_256_DEFAULT_M2S.arlock;
    axi4_256_m2s.arqos    <= AXI4_256_DEFAULT_M2S.arqos;
    axi4_256_m2s.arregion <= AXI4_256_DEFAULT_M2S.arregion;

    axi4_256_m2s.rready   <= AXI4_256_DEFAULT_M2S.rready;
  end;


  procedure axi4_256_sim_burst_read_cmd(
           address      : in    axi4_256_addr_s;
           tag          : in    natural;
           data         : out   byte_vector_t;
           nxfered      : out   natural;
    signal axi_clk      : in    std_logic;
    signal axi4_256_m2s : out   axi4_256_m2s_t;
    signal axi4_256_s2m : in    axi4_256_s2m_t)
  is
    constant AXI_WIDTH : natural := AXI4_256_STRB_WIDTH;
    constant BST_MAX   : natural := 2**AXI4_256_LEN_WIDTH - 1;

    variable data_offset : natural;
    variable burst_len   : natural range 0 to BST_MAX;

    type burst_state_type is (axi_cmd,axi_acc,axi_finished);
    variable state, n_state : burst_state_type;

    variable n_valid       : std_logic;
    variable n_addr        : axi4_256_addr_s;
    variable n_burstlength : axi4_256_len_s;
    variable n_datavalid   : std_logic;
    variable n_tag         : axi4_256_id_s;
    variable n_size        : axi4_size_s;
    variable n_burst       : axi4_burst_s;
    variable n_cache       : axi4_cache_s;
    variable n_prot        : axi4_prot_s;
    variable n_lock        : axi4_lock_s;
    variable n_qos         : axi4_qos_s;
    variable n_region      : axi4_region_s;

  begin
    assert not(address(3 downto 0) /= "0000")
      report "(axi4_256_sim_burst_read_cmd): Address is not 32-byte aligned during AXI 256 read"
      severity error;

    state       := axi_cmd;
    data_offset := data'low;

    -- remaining =  1..16 => burst_len = 0
    -- remaining = 17..32 => burst_len = 1 etc.
    burst_len := ((data'length + AXI_WIDTH - 1) / AXI_WIDTH) - 1; -- AXI BL = OCP BL -1

    while (state /= axi_finished) loop
      wait until rising_edge(axi_clk);

      case state is
        when axi_cmd =>
          -- Address channel
          n_valid       := '1';
          n_addr        := std_logic_vector(RESIZE(unsigned(address),AXI4_256_ADDR_WIDTH));
          n_burstlength := std_logic_vector(TO_UNSIGNED(burst_len,AXI4_256_LEN_WIDTH));
          n_tag         := std_logic_vector(TO_UNSIGNED(tag,AXI4_256_ID_WIDTH));
          n_size        := AXI4_256_SIZE;
          n_burst       := AXI4_256_BURST;
          n_cache       := AXI4_256_CACHE;
          n_prot        := AXI4_256_PROT;
          n_lock        := AXI4_256_LOCK;
          n_qos         := AXI4_256_QOS;
          n_region      := AXI4_256_REGION;

          n_state       := axi_acc;

        when axi_acc =>
          if (axi4_256_s2m.arready = '1') then
            -- Address channel defaults
            n_valid       := AXI4_256_DEFAULT_M2S.arvalid;
            n_addr        := AXI4_256_DEFAULT_M2S.araddr;
            n_burstlength := AXI4_256_DEFAULT_M2S.arlen;
            n_tag         := AXI4_256_DEFAULT_M2S.arid;
            n_size        := AXI4_256_DEFAULT_M2S.arsize;
            n_burst       := AXI4_256_DEFAULT_M2S.arburst;
            n_cache       := AXI4_256_DEFAULT_M2S.arcache;
            n_prot        := AXI4_256_DEFAULT_M2S.arprot;
            n_lock        := AXI4_256_DEFAULT_M2S.arlock;
            n_qos         := AXI4_256_DEFAULT_M2S.arqos;
            n_region      := AXI4_256_DEFAULT_M2S.arregion;

            n_state       := axi_finished;
          else
            n_state       := axi_acc;
          end if;

        when axi_finished =>
          n_state := axi_finished;
      end case;

      axi4_256_m2s.arvalid  <= n_valid;
      axi4_256_m2s.araddr   <= n_addr;
      axi4_256_m2s.arlen    <= n_burstlength;
      axi4_256_m2s.arid     <= n_tag;
      axi4_256_m2s.arsize   <= n_size;  
      axi4_256_m2s.arburst  <= n_burst;
      axi4_256_m2s.arcache  <= n_cache;
      axi4_256_m2s.arprot   <= n_prot;
      axi4_256_m2s.arlock   <= n_lock;
      axi4_256_m2s.arqos    <= n_qos;
      axi4_256_m2s.arregion <= n_region;

      state := n_state;
    end loop;

    nxfered := data'length;
  end;


  procedure axi4_256_sim_burst_read_resp(
           data         : out   byte_vector_t;
           tag          : out   natural;
           nxfered      : out   natural;
    signal axi_clk      : in    std_logic;
    signal axi4_256_m2s : out   axi4_256_m2s_t;
    signal axi4_256_s2m : in    axi4_256_s2m_t)
  is
    constant AXI_WIDTH    : natural := AXI4_256_STRB_WIDTH;
    constant BST_MAX      : natural := 2**AXI4_256_LEN_WIDTH - 1;

    variable data_offset  : natural;
    variable remaining_d  : natural;
    variable burst_len    : natural range 0 to BST_MAX;
    variable m_d          : natural range 0 to AXI_WIDTH;
  begin
    -- Response always accepted during read response
    axi4_256_m2s.rready <= '1';

    data_offset := data'low;
    remaining_d := data'length;

    -- remaining =  1..16 => burst_len = 0
    -- remaining = 17..32 => burst_len = 1 etc.
    burst_len := ((data'length + AXI_WIDTH - 1) / AXI_WIDTH) - 1; -- AXI BL = OCP BL -1

    while (remaining_d /= 0) loop
      wait until rising_edge(axi_clk);

      if (axi4_256_s2m.rresp = AXI4_256_RESP_OKAY)and(axi4_256_s2m.rvalid = '1') then
        m_d := AXI_WIDTH;
        if (m_d > remaining_d) then
          m_d := remaining_d;
        end if;

        for i in 0 to (AXI_WIDTH - 1) loop
          if (m_d > 0) then
            data(data_offset) := axi4_256_s2m.rdata(8 * (i + 1) - 1 downto 8 * i);
            data_offset := data_offset + 1;
            remaining_d := remaining_d - 1;
            m_d := m_d - 1;
          end if;
        end loop;
        tag := TO_INTEGER(UNSIGNED(axi4_256_s2m.rid));
      end if;
    end loop;

    axi4_256_m2s.rready <= AXI4_256_DEFAULT_M2S.rready;
    nxfered := data'length - remaining_d;
  end;


  procedure axi4_256_sim_burst_write(
           address      : in    axi4_256_addr_s;
           tag          : in    natural;
           be           : in    byte_enable_t;
           data         : in    byte_vector_t;
           nxfered      : out   natural;
    signal axi_clk      : in    std_logic;
    signal axi4_256_m2s : out   axi4_256_m2s_t;
    signal axi4_256_s2m : in    axi4_256_s2m_t)
  is
    constant AXI_WIDTH    : natural := AXI4_256_STRB_WIDTH;
    constant BST_MAX      : natural := 2**AXI4_256_LEN_WIDTH - 1;

    variable i            : natural;
    variable data_offset  : natural;
    variable be_offset    : natural;
    variable remaining_d  : natural;
    variable remaining_r  : natural;
    variable next_be      : boolean;
    variable burst_len    : natural range 0 to BST_MAX;
    variable m            : natural range 0 to AXI_WIDTH;

    type burst_state_type is (axi_cmd,axi_data,axi_finished);
    variable state, n_state : burst_state_type;

    variable cmd_finished  : boolean;
    variable data_finished : boolean;
    variable resp_finished : boolean;

    variable n_valid       : std_logic;
    variable n_addr        : axi4_256_addr_s;
    variable n_burstlength : axi4_256_len_s;
    variable n_tag         : axi4_256_id_s;
    variable n_size        : axi4_size_s;
    variable n_burst       : axi4_burst_s;
    variable n_cache       : axi4_cache_s;
    variable n_prot        : axi4_prot_s;
    variable n_lock        : axi4_lock_s;
    variable n_qos         : axi4_qos_s;
    variable n_region      : axi4_region_s;

    variable n_respaccept  : std_logic;
    variable n_datavalid   : std_logic;
    variable n_data        : axi4_256_data_s;
    variable n_databyteen  : axi4_256_strb_s;
    variable n_datalast    : std_logic;
    variable n_datatag     : axi4_256_id_s;

  begin
    assert not(be'length /= data'length)
      report "(axi4_256_sim_burst_write): byte enable length (" & integer'image(be'length) &
             ") does not match data length (" & integer'image(data'length) & ")"
      severity failure;

    assert not(address(3 downto 0) /= "0000")
      report "(axi4_256_sim_burst_write): Address is not 32-byte aligned during AXI 256 write"
      severity error;

    state       := axi_cmd;
    data_offset := data'low;
    be_offset   := be'low;
    remaining_d := data'length;
    remaining_r := data'length;

    -- remaining =  1..16 => burst_len = 0
    -- remaining = 17..32 => burst_len = 1 etc.
    burst_len := ((data'length + AXI_WIDTH - 1) / AXI_WIDTH) - 1; -- AXI BL = OCP BL -1

    while (state /= axi_finished) loop
      wait until rising_edge(axi_clk);
      next_be := false;

      case state is
        when axi_cmd =>
          -- Address channel
          n_valid       := '1';
          n_addr        := std_logic_vector(RESIZE(unsigned(address),AXI4_256_ADDR_WIDTH));
          n_burstlength := std_logic_vector(TO_UNSIGNED(burst_len,AXI4_256_LEN_WIDTH));
          n_tag         := std_logic_vector(TO_UNSIGNED(tag,AXI4_256_ID_WIDTH));
          n_size        := AXI4_256_SIZE;
          n_burst       := AXI4_256_BURST;
          n_cache       := AXI4_256_CACHE;
          n_prot        := AXI4_256_PROT;
          n_lock        := AXI4_256_LOCK;
          n_qos         := AXI4_256_QOS;
          n_region      := AXI4_256_REGION;

          -- Write data channel
          n_datavalid   := '1';
          n_datalast    := '0';
          n_datatag     := (others => '0');
          next_be       := true;
          n_state       := axi_data;
          n_respaccept  := '1';
          cmd_finished  := false;
          data_finished := false;
          resp_finished := false;

        when axi_data =>
          n_state := axi_data;

          if (axi4_256_s2m.awready = '1') then
            -- Address channel defaults
            n_valid       := AXI4_256_DEFAULT_M2S.awvalid;
            n_addr        := AXI4_256_DEFAULT_M2S.awaddr;
            n_burstlength := AXI4_256_DEFAULT_M2S.awlen;
            n_tag         := AXI4_256_DEFAULT_M2S.awid;
            n_size        := AXI4_256_DEFAULT_M2S.awsize;
            n_burst       := AXI4_256_DEFAULT_M2S.awburst;
            n_cache       := AXI4_256_DEFAULT_M2S.awcache;
            n_prot        := AXI4_256_DEFAULT_M2S.awprot;
            n_lock        := AXI4_256_DEFAULT_M2S.awlock;
            n_qos         := AXI4_256_DEFAULT_M2S.awqos;
            n_region      := AXI4_256_DEFAULT_M2S.awregion;

            cmd_finished  := true;
          end if;

          if (axi4_256_s2m.wready = '1') then
            if (remaining_d /= 0) then
              next_be := true;
            else
              -- Write data channel defaults
              n_data       := AXI4_256_DEFAULT_M2S.wdata;
              n_databyteen := AXI4_256_DEFAULT_M2S.wstrb;
              n_datavalid  := AXI4_256_DEFAULT_M2S.wvalid;
              n_datalast   := AXI4_256_DEFAULT_M2S.wlast;
              n_datatag    := AXI4_256_DEFAULT_M2S.wid;

              data_finished := true;
            end if;
          end if;

          if (axi4_256_s2m.bresp = AXI4_256_RESP_OKAY)and(axi4_256_s2m.bvalid = '1') then
            n_respaccept := AXI4_256_DEFAULT_M2S.bready;
            resp_finished := true;
          end if;

          if (cmd_finished and data_finished and resp_finished) then
            n_state := axi_finished;
          end if;

        when axi_finished =>    -- State is never executed - used in while test only
          null;
      end case;

      state := n_state;

      if next_be then
        if (remaining_d < AXI_WIDTH) then
          m := remaining_d;
        else
          m := AXI_WIDTH;
          if (remaining_d = AXI_WIDTH) then
            n_datalast := '1';
          else
            n_datalast := '0';
          end if;
        end if;

        -- Bytes > m are set to zero
        for i in 0 to (AXI_WIDTH - 1) loop
          if (m = 0) then
            n_databyteen(i) := '0';
            n_data(8 * (i + 1) - 1 downto 8 * i) := (others => '0');
          else
            n_databyteen(i) := be(be_offset);
            be_offset := be_offset + 1;
            n_data(8 * (i + 1) - 1 downto 8 * i) := data(data_offset);
            data_offset := data_offset + 1;
            remaining_d := remaining_d - 1;
            m := m - 1;
          end if;
        end loop;
      end if;

      axi4_256_m2s.awvalid  <= n_valid;
      axi4_256_m2s.awaddr   <= n_addr;
      axi4_256_m2s.awlen    <= n_burstlength;
      axi4_256_m2s.awid     <= n_tag;
      axi4_256_m2s.awsize   <= n_size;  
      axi4_256_m2s.awburst  <= n_burst;
      axi4_256_m2s.awcache  <= n_cache;
      axi4_256_m2s.awprot   <= n_prot;
      axi4_256_m2s.awlock   <= n_lock;
      axi4_256_m2s.awqos    <= n_qos;
      axi4_256_m2s.awregion <= n_region;

      axi4_256_m2s.bready   <= n_respaccept;

      axi4_256_m2s.wvalid   <= n_datavalid;
      axi4_256_m2s.wdata    <= n_data;
      axi4_256_m2s.wstrb    <= n_databyteen;
      axi4_256_m2s.wlast    <= n_datalast;
      axi4_256_m2s.wid      <= n_datatag;
    end loop;

    nxfered := data'length - remaining_d;
  end;


  -- ---------------------------------------------------------------
  -- Public
  -- ---------------------------------------------------------------

  function conv_byte_vector(
    constant x : in std_logic_vector)
  return byte_vector_t is
    variable n : natural := (x'length + 7) / 8;         -- Size of x in bytes (min size of 1)
    variable m : std_logic_vector(8 * n - 1 downto 0);  -- New vector of size n bytes
    variable t : byte_vector_t(0 to n - 1);             -- New byte_vector of size n bytes
  begin
    m := std_logic_vector(RESIZE(unsigned(x), 8 * n));

    for i in 0 to n - 1 loop
      t(i) := m(8 * (i + 1) - 1 downto 8 * i);
    end loop;

    return t;
  end;

  function conv_byte_enable(
    constant x : in std_logic_vector)
  return byte_enable_t is
    variable n : natural := x'length;               -- Size of x in bits (min size of 1)
    variable m : std_logic_vector(n - 1 downto 0);  -- New vector of size n bits
    variable t : byte_enable_t(0 to n - 1);         -- New byte_enable of size n bits
  begin
    m := x;

    for i in 0 to n - 1 loop
      t(i) := m(i);
    end loop;

    return t;
  end;

  function conv_vector(
    constant x : in byte_vector_t)
  return std_logic_vector is
    variable v          : std_logic_vector((8*x'length)-1 downto 0);
    variable next_byte  : byte_t;
    variable j          : natural;
  begin
    for i in x'range loop
      next_byte := x(i);
      j := i - x'low;
      v(j*8+7 downto j*8) := next_byte;
    end loop;

    return v;
  end;

  function conv_string_hex(
    constant x : in byte_vector_t)
  return string is
    variable j         : natural := 2 * x'length;
    variable s         : string(1 to j);
    variable next_byte : byte_t;
  begin
    for i in x'range loop
      next_byte := x(i);
      s(j)      := conv_string_hexdigit(next_byte(3 downto 0));
      s(j-1)    := conv_string_hexdigit(next_byte(7 downto 4));
      j         := j - 2;
    end loop;

    return s;
  end;

  function conv_string_hex(
    constant x : in std_logic_vector)
  return string is
    variable i : natural := x'length mod 4;
    variable j : natural := x'length + (4-i) mod 4;
    variable s : string(1 to j/4);
    variable y : std_logic_vector(j-1 downto 0) := (others => '0');
  begin
    y(x'length - 1 downto 0) := x;

    for k in (j/4)-1 downto 0 loop
      s(j/4 - k) := conv_string_hexdigit(y((k*4)+3 downto k*4));
    end loop;

    return s;
  end;

  function conv_string_bin(
    constant x : in byte_enable_t)
  return string is
    variable j : natural := x'length;
    variable s : string(1 to j);
  begin
    for i in x'range loop
      s(j) := conv_std_logic_char(x(i));
      j    := j-1;
    end loop;

    return s;
  end;

  function conv_string_bin(
    constant x : in std_logic_vector)
  return string is
    variable j : natural := x'length;
    variable s : string(1 to j);
    variable y : std_logic_vector(j-1 downto 0);
  begin
    y := x;

    for k in j-1 downto 0 loop
      s(j-k) := conv_std_logic_char(y(k));
    end loop;

    return s;
  end;


  -- Calculate the string width required to represent
  -- a decimal number in decimal format
  function conv_dec_width(
    decimal : in natural)
  return natural is
  begin
    assert not(decimal < 0)
      report "(conv_dec_width): Cannot use 32-bit natural"
      severity failure;

    case decimal is
      when 0 to 9 => return 1;
      when 10 to 99 => return 2;
      when 100 to 999 => return 3;
      when 1000 to 9999 => return 4;
      when 10000 to 99999 => return 5;
      when 100000 to 999999 => return 6;
      when 1000000 to 9999999 => return 7;
      when 10000000 to 99999999 => return 8;
      when 100000000 to 999999999 => return 9;
      when 1000000000 to integer'high => return 10;
      when others => return 11;
    end case;
  end function;


  -- ---------------------------------------------------------------
  -- Public AXI4 256 routines
  -- ---------------------------------------------------------------

  -- Data:    4-byte read (blocking)
  -- Address: 4-byte aligned
  -- Blocking read procedure waits until all data has been read before completing
  procedure axi4_256_sim_read_reg32(
           address      : in    axi4_256_addr_s;
           tag          : in    natural;
           data         : out   byte_vector_t;
           nxfered      : out   natural;
    signal axi_clk      : in    std_logic;
    signal axi4_256_m2s : out   axi4_256_m2s_t;
    signal axi4_256_s2m : in    axi4_256_s2m_t)
  is
    variable offset : natural := TO_INTEGER(UNSIGNED(address(AXI4_256_STRB_WIDTH_LOG2-1 downto 2)));
    variable addr_a : axi4_256_addr_s;
    variable data_a : byte_vector_t(0 to AXI4_256_STRB_WIDTH-1);
    variable xfer_a : natural;
  begin
    assert not(address(1 downto 0) /= "00")
      report "(axi4_256_sim_read_reg32): Address is not 4-byte aligned during 32-bit register read"
      severity error;

    assert not(data'length /= 4)
      report "(axi4_256_sim_read_reg32): Data is not 4-bytes"
      severity error;

    -- Convert 4-byte aligned address to AXI4_256_STRB_WIDTH-byte aligned address
    addr_a := address;
    addr_a(AXI4_256_STRB_WIDTH_LOG2-1 downto 0) := (others => '0');

    axi4_256_sim_read(addr_a, tag, data_a, xfer_a, axi_clk, axi4_256_m2s, axi4_256_s2m);

    -- Convert AXI4_256_STRB_WIDTH-byte aligned data to 4-byte data
    for i in 0 to 3 loop
      data(i) := data_a(4*offset+i);
    end loop;

    if (xfer_a = AXI4_256_STRB_WIDTH) then
      nxfered := 4;
    else
      nxfered := 0;
      assert false
        report "(axi4_256_sim_read_reg32): Bytes transferred = " & integer'image(xfer_a) &
               " (/=" & integer'image(AXI4_256_STRB_WIDTH) &
               ") during 4-byte register read."
        severity error;
    end if;
  end;

  -- Data:    4-byte write
  -- Address: 4-byte aligned
  procedure axi4_256_sim_write_reg32(
           address      : in    axi4_256_addr_s;
           tag          : in    natural;
           be           : in    byte_enable_t;
           data         : in    byte_vector_t;
           nxfered      : out   natural;
    signal axi_clk      : in    std_logic;
    signal axi4_256_m2s : out   axi4_256_m2s_t;
    signal axi4_256_s2m : in    axi4_256_s2m_t)
  is
    variable offset : natural := TO_INTEGER(UNSIGNED(address(AXI4_256_STRB_WIDTH_LOG2-1 downto 2)));
    variable addr_a : axi4_256_addr_s;
    variable be_a   : byte_enable_t(0 to AXI4_256_STRB_WIDTH-1) := (others => '0');
    variable data_a : byte_vector_t(0 to AXI4_256_STRB_WIDTH-1) := (others => (others => '0'));
    variable xfer_a : natural;
  begin
    assert not(address(1 downto 0) /= "00")
      report "(axi4_256_sim_write_reg32): Address is not 4-byte aligned during 32-bit register write"
      severity error;

    assert not(be'length /= 4)
      report "(axi4_256_sim_write_reg32): BE is not 4-bits"
      severity error;

    assert not(data'length /= 4)
      report "(axi4_256_sim_write_reg32): Data is not 4-bytes"
      severity error;

    -- Convert 4-byte aligned address to AXI4_256_STRB_WIDTH-byte aligned address
    addr_a := address;
    addr_a(AXI4_256_STRB_WIDTH_LOG2-1 downto 0) := (others => '0');

    -- Convert 4-bit be to AXI4_256_STRB_WIDTH-bit aligned be
    -- Convert 4-byte data to AXI4_256_STRB_WIDTH-byte aligned data
    for i in 0 to 3 loop
      be_a(4*offset+i)   := be(i);
      data_a(4*offset+i) := data(i);
    end loop;

    axi4_256_sim_write(addr_a, tag, be_a, data_a, xfer_a, axi_clk, axi4_256_m2s, axi4_256_s2m);

    if (xfer_a = AXI4_256_STRB_WIDTH) then
      nxfered := 4;
    else
      nxfered := 0;
      assert false
        report "(axi4_256_sim_write_reg32): Bytes transferred = " & integer'image(xfer_a) &
               " (/=" & integer'image(AXI4_256_STRB_WIDTH) &
               ") during 4-byte register write."
        severity error;
    end if;
  end;

  -- Data:    n-byte read (blocking)
  -- Address: 32-byte aligned
  -- Blocking read procedure waits until all data has been read before completing
  procedure axi4_256_sim_read(
           address      : in    axi4_256_addr_s;
           tag          : in    natural;
           data         : out   byte_vector_t;
           nxfered      : out   natural;
    signal axi_clk      : in    std_logic;
    signal axi4_256_m2s : out   axi4_256_m2s_t;
    signal axi4_256_s2m : in    axi4_256_s2m_t)
  is
    variable tag_out : natural;
  begin
    axi4_256_sim_burst_read_cmd(address, tag, data, nxfered, axi_clk, axi4_256_m2s, axi4_256_s2m);
    axi4_256_sim_burst_read_resp(data, tag_out, nxfered, axi_clk, axi4_256_m2s, axi4_256_s2m);
  end;

  -- Data:    32-byte read (non-blocking)
  -- Address: 32-byte aligned
  -- Non-Blocking read procedure. Produces read commands and accepts read responses
--  procedure axi4_256_sim_read_nb(
--    constant c_addr       : in    axi4_256_addr_s;
--    constant c_bl         : in    natural;
--    constant c_tag        : in    natural;
--    variable c_done       : out   boolean;
--    variable r_data       : out   byte_vector_t;
--    variable r_tag        : out   natural;
--    variable r_done       : out   boolean;
--    signal   axi_clk      : in    std_logic;
--    signal   axi4_256_m2s : out   axi4_256_m2s_t;
--    signal   axi4_256_s2m : in    axi4_256_s2m_t)
--  is
--  begin
--    axi4_256_sim_burst_read_nb(c_addr, c_bl, c_tag, c_done, r_data, r_tag, r_done, axi_clk, axi4_256_m2s, axi4_256_s2m);
--  end;

  -- Data:    n*32-byte reads (non-blocking)
  -- Address: 32-byte aligned
  -- Non-Blocking read procedure. Produces read commands and accepts read responses
  procedure axi4_256_sim_read_nb(
           address      : in    axi4_256_addr_s;
           tag          : in    natural;
           data         : out   byte_vector_t;
           nxfered      : out   natural;
    signal axi_clk      : in    std_logic;
    signal axi4_256_m2s : out   axi4_256_m2s_t;
    signal axi4_256_s2m : in    axi4_256_s2m_t)
  is
    variable size_int : natural := data'length; -- DMA size in bytes
    variable c_bl_rem : natural := (size_int + AXI4_256_STRB_WIDTH - 1) / AXI4_256_STRB_WIDTH; -- Transfer size in beats
    variable r_exp    : natural := c_bl_rem * AXI4_256_STRB_WIDTH;                             -- Transfer size in bytes
    variable c_bl     : natural := 8; -- Command burstlength = 8 beats
    variable r_cnt    : natural := 0;
    variable r_data   : byte_vector_t(0 to AXI4_256_STRB_WIDTH-1);
    variable c_addr   : axi4_256_addr_s := address;
    variable c_done   : boolean;
    variable r_done   : boolean;
    variable c_tag    : natural := tag;
    variable r_tag    : natural;
  begin
--    report "Performing NB read of size " & integer'image( size_int ) & " bytes.";

    -- Select full or remainder BL 
    while ( c_bl_rem /= 0 )or(r_cnt /= r_exp) loop
      if ( c_bl > c_bl_rem ) then
        c_bl := c_bl_rem;
      end if;

      axi4_256_m2s <= AXI4_256_FLOAT_M2S;

      -- Produce read commands
      -- Consume responses (1 beat)
      axi4_256_sim_burst_read_nb( c_addr, -- c_address
                           c_bl,          -- c_bl
                           c_tag,         -- c_tag
                           c_done,        -- c_done
                           r_data,        -- r_data
                           r_tag,         -- r_tag
                           r_done,        -- r_done
                           axi_clk,
                           axi4_256_m2s,
                           axi4_256_s2m );

      axi4_256_m2s <= AXI4_256_FLOAT_M2S;

      if c_done then
        c_bl_rem := c_bl_rem - c_bl;
        c_addr   := c_addr + std_logic_vector(to_unsigned(c_bl * AXI4_256_STRB_WIDTH,AXI4_256_ADDR_WIDTH));
--        report "(axi4_256_sim_read_nb): " & integer'image(c_bl_rem) & " Read commands remain to be issued.";
      end if;

      if r_done then
        for i in 0 to AXI4_256_STRB_WIDTH-1 loop
          if (( r_cnt + i ) < size_int) then
            data( r_cnt + i ) := r_data( i );
          end if;
        end loop;
        r_cnt := r_cnt + AXI4_256_STRB_WIDTH;
--        report "(axi4_256_sim_read_nb): Read " & integer'image(r_cnt) & " data bytes.";
      end if;
    end loop;

    nxfered := r_cnt;
--    report "NB Read completed";
  end;


  -- Address: 16-byte aligned
  -- Non-Blocking read command procedure. Produces read commands only
  procedure axi4_256_sim_read_cmd(
           address      : in    axi4_256_addr_s;
           tag          : in    natural;
           data         : out   byte_vector_t;
           nxfered      : out   natural;
    signal axi_clk      : in    std_logic;
    signal axi4_256_m2s : out   axi4_256_m2s_t;
    signal axi4_256_s2m : in    axi4_256_s2m_t)
  is
  begin
    axi4_256_sim_burst_read_cmd(address, tag, data, nxfered, axi_clk, axi4_256_m2s, axi4_256_s2m);
  end;

  -- Data:    n-byte read (non-blocking)
  -- Non-Blocking read response procedure. Waits for read response only
  procedure axi4_256_sim_read_resp(
           data         : out   byte_vector_t;
           tag          : out   natural;
           nxfered      : out   natural;
    signal axi_clk      : in    std_logic;
    signal axi4_256_m2s : out   axi4_256_m2s_t;
    signal axi4_256_s2m : in    axi4_256_s2m_t)
  is
  begin
    axi4_256_sim_burst_read_resp(data, tag, nxfered, axi_clk, axi4_256_m2s, axi4_256_s2m);
  end;

  -- Data:    n-byte write
  -- Address: 16-byte aligned
  procedure axi4_256_sim_write(
           address      : in    axi4_256_addr_s;
           tag          : in    natural;
           be           : in    byte_enable_t;
           data         : in    byte_vector_t;
           nxfered      : out   natural;
    signal axi_clk      : in    std_logic;
    signal axi4_256_m2s : out   axi4_256_m2s_t;
    signal axi4_256_s2m : in    axi4_256_s2m_t)
  is
  begin
    axi4_256_sim_burst_write(address, tag, be, data, nxfered, axi_clk, axi4_256_m2s, axi4_256_s2m);
  end;


  procedure axi4_256_sim_wait_cycles(
           cycles : in    natural;
    signal clk    : in    std_logic)
  is
  begin
    for i in 0 to (cycles-1) loop
      wait until rising_edge(clk);
    end loop;
  end;

end;
