library ieee;
use ieee.std_logic_1164.all;

entity tb_dma_demo_pcie is
  generic(
    sim_axi : boolean);
end;

architecture behav of tb_dma_demo_pcie is

  constant MODEL_NAME : string(1 to 11) := "ADM-XRC-KU1";

  -- Clocks
  constant REF_CLK_PER  : time := 1 us/300; -- 300 MHz reference clock
  signal refclk300_p : std_logic := '0';
  signal refclk300_n : std_logic := '1';
  constant PCIE_CLK_PER : time := 1 us/100; -- PCIe reference clock
  signal pcie100_p   : std_logic := '0';
  signal pcie100_n   : std_logic := '1';
  constant AVR_CLK_PER  : time := 1 us/50;  -- AVR uC clock
  signal avr_clk     : std_logic := '0';

  -- PCIe
  constant PCIE_SER_WIDTH : natural := 8;
  signal pci_t2h_p : std_logic_vector(PCIE_SER_WIDTH-1 downto 0);
  signal pci_t2h_n : std_logic_vector(PCIE_SER_WIDTH-1 downto 0);
  signal pci_h2t_p : std_logic_vector(PCIE_SER_WIDTH-1 downto 0);
  signal pci_h2t_n : std_logic_vector(PCIE_SER_WIDTH-1 downto 0);
  signal perst_n : std_logic := '0';

  -- AVR uC communication
  signal avr_b2t : std_logic;
  signal avr_t2b : std_logic;

  component dma_demo_pcie
  generic(
    sim_axi    : boolean := false);
  port(
    -- Reference clock
    refclk300_p : in    std_logic;
    refclk300_n : in    std_logic;
    -- PCIe
    perst_n   : in    std_logic;
    pcie100_p : in    std_logic;
    pcie100_n : in    std_logic;
    pci_t2h_p : out   std_logic_vector(7 downto 0);
    pci_t2h_n : out   std_logic_vector(7 downto 0);
    pci_h2t_p : in    std_logic_vector(7 downto 0);
    pci_h2t_n : in    std_logic_vector(7 downto 0);
    -- AVR
    avr_clk : in    std_logic;
    avr_b2t : in    std_logic;
    avr_t2b : out   std_logic);
  end component;

begin

  ----------------------------
  -- Board Clock Generation --
  ----------------------------

  -- Reference clocks
  refclk300_p <= not refclk300_p after REF_CLK_PER / 2;
  refclk300_n <= not refclk300_n after REF_CLK_PER / 2;

  -- Target PCIe clock (mgt226_clk0)
  pcie100_p <= not pcie100_p after PCIE_CLK_PER / 2;
  pcie100_n <= not pcie100_n after PCIE_CLK_PER / 2;

  -- AVR clock
  avr_clk <= not avr_clk after AVR_CLK_PER / 2;


  -- Report design constants
  report_setup_p : process
  begin
    report "Model Name : " & MODEL_NAME;
    report "Waiting for PCIe online...";
    wait until (perst_n = '1');
    report "PCIe link online";
    wait;
  end process;


  -- AXI only simulation
  -- Host application testbench in target PCIe interface
  sim_axi_g : if (sim_axi = true) generate
  begin
    pci_h2t_p <= (others => '-');
    pci_h2t_n <= (others => '-');

    avr_b2t <= '-';
  end generate;


  ------------------------------------
  -- Target Component Instantiation --
  ------------------------------------
  perst_n <= '1' after 100ns;

  uut : dma_demo_pcie
  generic map(
    sim_axi     => sim_axi) -- in    boolean := false;
  port map(
    -- 300MHz reference clock
    refclk300_p => refclk300_p, -- in    std_logic
    refclk300_n => refclk300_n, -- in    std_logic
    -- PCIe
    perst_n   => perst_n,    -- in    std_logic
    pcie100_p => pcie100_p,  -- in    std_logic
    pcie100_n => pcie100_n,  -- in    std_logic
    pci_t2h_p => pci_t2h_p,  -- out   std_logic_vector(7 downto 0)
    pci_t2h_n => pci_t2h_n,  -- out   std_logic_vector(7 downto 0)
    pci_h2t_p => pci_h2t_p,  -- in    std_logic_vector(7 downto 0)
    pci_h2t_n => pci_h2t_n,  -- in    std_logic_vector(7 downto 0)
    -- AVR uC communication
    avr_clk => avr_clk,   -- in    std_logic
    avr_b2t => avr_b2t,   -- in    std_logic
    avr_t2b => avr_t2b);  -- out   std_logic

end;