proc make_project { } {
  # Get the directory containing this script. Source file paths are relative to this.
  set origin_dir [file dirname [file normalize [info script]]]

  # Define versions of dependencies in one convenient place
  source "$origin_dir/version.tcl"

  set viv_ver [version -short]
  set viv_spl [split $viv_ver .]
  set viv_maj [lindex $viv_spl 0]
  set viv_min [lindex $viv_spl 1]

  if { $viv_maj < 2018 || ($viv_maj == 2018 && $viv_min < 3) } {
    send_msg_id {AD-MKXPR-001} {ERROR} "Vivado ${viv_ver} is not supported for this FPGA design. Minimum Vivado version is 2018.3."
  }

  # Create project
  set device {ku115_2i}
  set dir "itest-${device}"
  if { [info exists ::env(ADMXRC3_PKG_TEST)] } {
    append dir {-} [version -short]
  }
  set part {xcku115-flva1517-2-i}
  set name {itest}
  set project [create_project -force -part $part $name "$origin_dir/$dir"]
  current_project $project

  # Set the directory path for the new project
  set proj_dir [get_property directory $project]

  # Set project properties
  set_property "default_lib" "xil_defaultlib" $project
  set_property "part" $part $project
  set_property "simulator_language" "Mixed" $project
  set_property "target_language" "VHDL" $project

  set repo_paths [list \
    [file normalize "$origin_dir/../../../../fpga/repo/vivado-2019.2"] \
  ]
  set_property "ip_repo_paths" $repo_paths $project
  update_ip_catalog

  # Create 'sources_1' fileset (if not found)
  if {[string equal [get_filesets -quiet sources_1] ""]} {
    create_fileset -srcset sources_1
  }
  set source_set [get_filesets sources_1]

  # Add HDL source files
  set hdl_files [list \
    [file normalize "$origin_dir/../src/itest.vhd"] \
  ]
  add_files -norecurse -fileset $source_set $hdl_files

  # Create IP
  create_ip -vlnv "alphadata:user:mptl_tgt_ku_axi:${hsaxi_ver}" -module_name mptl_if
  set mptl_if [get_ips mptl_if]
  set_property -dict { \
    CONFIG.enable_bridge_master {true} \
    CONFIG.enable_dma {false} \
    CONFIG.enable_target_master {false} \
  } $mptl_if
  
  create_ip -vlnv "xilinx.com:ip:axi_dwidth_converter:*" -module_name axi4_128_to_32
  set axi4_128_to_32 [get_ips axi4_128_to_32]
  set_property -dict { \
    CONFIG.PROTOCOL {AXI4} \
    CONFIG.ADDR_WIDTH {64} \
    CONFIG.SI_DATA_WIDTH {128} \
    CONFIG.MI_DATA_WIDTH {32} \
    CONFIG.SI_ID_WIDTH {0} \
  } $axi4_128_to_32
  
  create_ip -vlnv "xilinx.com:ip:axi_protocol_converter:*" -module_name axi4_to_axi4l
  set axi4_to_axi4l [get_ips axi4_to_axi4l]
  set_property -dict { \
    CONFIG.SI_PROTOCOL {AXI4} \
    CONFIG.MI_PROTOCOL {AXI4LITE} \
    CONFIG.ADDR_WIDTH {64} \
    CONFIG.DATA_WIDTH {32} \
    CONFIG.ID_WIDTH {0} \
    CONFIG.TRANSLATION_MODE {2} \
  } $axi4_to_axi4l
  
  create_ip -vlnv "xilinx.com:ip:axi_bram_ctrl:*" -module_name axi4l_to_regs
  set axi4l_to_regs [get_ips axi4l_to_regs]
  set_property -dict { \
    CONFIG.PROTOCOL {AXI4LITE} \
    CONFIG.MEM_DEPTH {1024} \
    CONFIG.SUPPORTS_NARROW_BURST {0} \
    CONFIG.SINGLE_PORT_BRAM {1} \
  } $axi4l_to_regs

  # Add constraints files
  # Put target .xdc as LAST in list
  set constraint_set [get_filesets constrs_1]
  set constraint_files [list \
    [file normalize "$origin_dir/../src/mptl-admxrcku1.xdc"] \
    [file normalize "$origin_dir/../src/bitstream.xdc"] \
    [file normalize "$origin_dir/../src/refclk300.xdc"] \
    [file normalize "$origin_dir/../src/userid.xdc"] \
    [file normalize "$origin_dir/../src/itest.xdc"] \
  ]
  add_files -norecurse -fileset $constraint_set $constraint_files
  set_property "target_constrs_file" [lindex $constraint_files end] $constraint_set

  # Add utilities files (Tcl hook scripts etc.)
  set utils_set [get_filesets -quiet utils_1]
  if {[string equal $utils_set ""]} {
    # Utilities fileset introduced in Vivado 2018.3; fall back to using sources set if not present.
    set utils_set $source_set
  }
  #set utils_files [list \
  #  [file normalize "$origin_dir/write_bitstream_maxthread.tcl"] \
  #]
  #add_files -norecurse -fileset $utils_set $utils_files
  # Set used_in_XXX properties so that Vivado 2018.2 and earlier doesn't treat them as design sources.
  #set_property -dict {
  #  {used_in_synthesis} {false} \
  #  {used_in_implementation} {false} \
  #  {used_in_simulation} {false} \
  #} [get_files $utils_files]

  set synth_run [get_runs synth_1]
  set_property "needs_refresh" "1" $synth_run
  set_property "part" $part $synth_run
  current_run -synthesis $synth_run

  set impl_run [get_runs impl_1]
  set_property "needs_refresh" "1" $impl_run
  set_property "part" $part $impl_run
  set_property -name {steps.route_design.args.more options} -value -timing_summary -objects $impl_run
  current_run -implementation $impl_run

  update_compile_order -fileset $source_set
  if { [info exists sim_set] } {
    update_compile_order -fileset $sim_set
  }

  return $project
}

return [make_project]
