/*
** itest.cpp
**
** Test program for the Interrupt Test FPGA Design (itest)
**
** Notes
** =====
**
** 1. Uses the ADMXRC3UT library for opening the device. ADMXRC3UT is a set of
**    functions that contains commonly-used code in example / test programs,
**    but it is not part of the ADMXRC3 API.
**
**    ADMXRC3UT_OpenByIndex is similar to ADMXRC3_OpenEx from the ADMXRC3 API,
**    except that it also performs a check to see if the device is in an error
**    state and attempts to clear device errors if it is.
**
**    ADMXRC3UT_OpenBySerialNumber is similar to ADMXRC3_OpenEx from the
**    ADMXRC3 API, except that it searches for the device with a particular
**    serial number rather than opening by index. Like ADMXRC3UT_OpenByIndex,
**    it also checks to see if the device is in an error state and attempts to
**    clear device errors if it is.
**
** 2. This source file contains code for doing the actual test. In VxWorks,
**    the admxrc3ITestIndex & admxrc3ITestSN functions are entry points
**    that can be called from the VxWorks shell.
*/

#ifdef _WIN32

# include <windows.h>
# include <tchar.h>

typedef UINT8 uint8_t;
typedef UINT32 uint32_t;

#elif defined(__linux)

# include <stdint.h>

# define _T(x) x
# define _tmain main
# define _tcslen(x) strlen(x)
typedef char TCHAR;
# define _gettc getc
# define _totlower(x) tolower(x)

#elif defined(__VXWORKS__) || defined(__vxworks)

# include <vxWorks.h>

# if _WRS_VXWORKS_MAJOR >= 6
#   include <stdint.h>
# endif

# define _T(x) x
# define _tmain main
# define _tcslen(x) strlen(x)
typedef char TCHAR;
# define _gettc getc
# define _totlower(x) tolower(x)

#endif

#include <cstdio>
#include <cassert>
#include <cstring>
#include <string>
#include <iostream>
#include <algorithm>

#include <exappendian.h>
#include <exappthread.h>

#include <admxrc3.h>
#include <admxrc3ut.h>

#include "itest.h"

using namespace AlphaData::AppFramework;

#ifndef FALSE
# define FALSE 0
#endif
#ifndef TRUE
# define TRUE 1
#endif

#ifdef _UNICODE
typedef std::wstring SType;
typedef std::wstringstream SStreamType;
#define SSEMPTY(ss) ((ss).str(std::basic_string<wchar_t>()))
#else
typedef std::string SType;
typedef std::stringstream SStreamType;
#define SSEMPTY(ss) ((ss).str(std::string()))
#endif

#define MAX_NUM_DMA_ENGINE (32)

typedef struct _ITestRegisters {
  uint32_t IEnable;        /* 0x00 */
  uint32_t IEnableSet;     /* 0x04 */
  uint32_t IEnableClr;     /* 0x08 */
  uint32_t IStatus;        /* 0x0C */
  uint32_t Test;           /* 0x10 */
  uint32_t _Reserved1[3];
  uint32_t Period;         /* 0x20 */
  uint32_t _Reserved2[7];
} ITestRegisters;

static void writeRegister(uint32_t* pRegister, uint32_t val32);
static uint32_t readRegister(uint32_t* pRegister);
static void writeTESTRegister(ITestRegisters* pRegisters);

class CInterruptThread : protected CExAppThread {
public:
  CInterruptThread(
    CExAppMessageDisplay* pMessageDisplay,
    ADMXRC3_HANDLE hDevice,
    ITestRegisters* pRegisters)
    : CExAppThread(), m_pMessageDisplay(pMessageDisplay)
  {
    m_hDevice = hDevice;
    m_pRegisters = pRegisters;
#ifdef _WIN32
    m_hEvent = NULL;
#elif defined(__VXWORKS__) || defined(__vxworks)
    m_hSem = NULL;
#endif
  } 

  virtual ~CInterruptThread() {
    if (IsRunning()) {
      Shutdown();
    }

    Cleanup();
  }

  bool Initialize() {
    m_bTerminate = false;

#ifdef _WIN32
    m_hEvent = CreateEvent(NULL, FALSE, FALSE, NULL);
    if (NULL == m_hEvent) {
      DWORD dwLastError = GetLastError();
      m_pMessageDisplay->FormatError(_T("Failed to create Win32 event for FPGA interrupt notifications; LastError=%u(0x%X)\n"), dwLastError, dwLastError);
      return false;
    }
#elif defined(__VXWORKS__) || defined(__vxworks)
    m_hSem = semBCreate(SEM_Q_FIFO, SEM_EMPTY);
    if (NULL == m_hSem) {
      m_pMessageDisplay->FormatError(_T("Failed to create Win32 event for FPGA interrupt notifications; errno=%d\n"), errno);
      return false;
    }
#endif

    Start();

    return true;
  }

  void* Shutdown() {
    m_bTerminate = true;

#ifdef _WIN32
    // Set the Win32 Event, to make ThreadFunction quit.
    SetEvent(m_hEvent);
#elif defined(__VXWORKS__) || defined(__vxworks)
    // Set the Semaphore, to make ThreadFunction quit.
    semGive(m_hSem);
#elif defined(__linux)
    // Generate another interrupt, to make ThreadFunction quit.
    writeTESTRegister(m_pRegisters);
#endif

    return Join();
  }

protected:

#ifdef _WIN32

  //
  // Windows implementation uses Win32 Event registration technique.
  //

  void* ThreadFunction() {
    // Enabled FPGA internal interrupts
    const uint32_t enabled = 0x3U; // In a more complex application, this might not be a constant.

    ADMXRC3_STATUS status;
    uint32_t pending = 0;

    m_pMessageDisplay->FormatInfo(_T("[Interrupt Thread] Starting (Windows implementation)."));

    //
    // The registered Win32 event will be set to the signaled state by a falling
    // edge on the "firq_l" signal.
    //
    status = ADMXRC3_RegisterWin32Event(m_hDevice, ADMXRC3_EVENT_FPGAINTERRUPT(0), m_hEvent);
    if (ADMXRC3_SUCCESS != status) {
      m_pMessageDisplay->FormatError(_T("[Interrupt Thread] Failed to register Win32 event for FPGA interrupt notifications: %s(0x%X)."), ADMXRC3_GetStatusString(status, TRUE), status);
      goto done;
    }

    while (!m_bTerminate) {
      //
      // Clear only the pending internal interrupts that we sampled asserted.
      //
      // This will also force the "firq_l" signal (driven by the FPGA to the Bridge)
      // deasserted for a clock cycle, which will cause a falling edge and immediate
      // signaling of the Win32 Event if there happen to still be pending interrupts.
      //
      writeRegister(&m_pRegisters->IStatus, pending);

      WaitForSingleObject(m_hEvent, INFINITE);

      // Sample the pending internal interrupts in the FPGA.
      pending = readRegister(&m_pRegisters->IStatus) & enabled;

      if (pending & 0x1) {
        m_pMessageDisplay->FormatInfo(_T("[Interrupt Thread] Saw FPGA internal interrupt 0 (caused by TEST register write)."));

        // In a real application, you would do something here in response to the interrupt.
      }

      if (pending & 0x2) {
        m_pMessageDisplay->FormatInfo(_T("[Interrupt Thread] Saw FPGA internal interrupt 1 (caused by periodic timer)."));

        // In a real application, you would do something here in response to the interrupt.
      }
    }

done:
    m_pMessageDisplay->FormatInfo(_T("[Interrupt Thread] Terminating."));

    status = ADMXRC3_UnregisterWin32Event(m_hDevice, ADMXRC3_EVENT_FPGAINTERRUPT(0), m_hEvent);
    assert(ADMXRC3_SUCCESS == status);

    return NULL;
  }

#elif defined(__linux)

  //
  // Linux implementation uses ADXMRC3_StartNotificationWait ... ADXMRC3_FinishNotificationWait technique.
  // (This would also work in Windows.)
  //

  void* ThreadFunction() {
    // Enabled FPGA internal interrupts
    const uint32_t enabled = 0x3U; // In a more complex application, this might not be a constant.

    ADMXRC3_TICKET ticket;
    ADMXRC3_STATUS status;
    uint32_t pending = 0;

    m_pMessageDisplay->FormatInfo(_T("[Interrupt Thread] Starting (Linux implementation)."));

    ADMXRC3_InitializeTicket(&ticket);

    while (!m_bTerminate) {
      //
      // The notification wait will be completed by the next falling edge on the "firq_l" signal.
      //
      status = ADMXRC3_StartNotificationWait(m_hDevice, &ticket, ADMXRC3_EVENT_FPGAINTERRUPT(0));
      if (ADMXRC3_PENDING != status) {
        m_pMessageDisplay->FormatError(_T("[Interrupt Thread] Failed to start waiting for FPGA interrupt: %s(0x%X)."), ADMXRC3_GetStatusString(status, TRUE), status);
        break;
      }

      //
      // Clear only the pending internal interrupts that we sampled asserted.
      //
      // This will also force the "firq_l" signal (driven by the FPGA to the Bridge)
      // deasserted for a clock cycle, which will cause a falling edge and completion
      // of the notification wait if we happen to have called ADMXRC3_StartNotificationWait
      // after a falling edge has already occurred (i.e. we missed the falling edge).
      //
      writeRegister(&m_pRegisters->IStatus, pending);

      // Wait for completion of the notification wait.
      status = ADMXRC3_FinishNotificationWait(m_hDevice, &ticket, TRUE);
      if (ADMXRC3_SUCCESS != status) {
        m_pMessageDisplay->FormatError(_T("[Interrupt Thread] Failed to finish waiting for FPGA interrupt: %s(0x%X)."), ADMXRC3_GetStatusString(status, TRUE), status);
        break;
      }

      // Sample the pending internal interrupts in the FPGA.
      pending = readRegister(&m_pRegisters->IStatus) & enabled;

      if (pending & 0x1) {
        m_pMessageDisplay->FormatInfo(_T("[Interrupt Thread] Saw FPGA internal interrupt 0 (caused by TEST register write)."));

        // In a real application, you would do something here in response to the interrupt.
      }

      if (pending & 0x2) {
        m_pMessageDisplay->FormatInfo(_T("[Interrupt Thread] Saw FPGA internal interrupt 1 (caused by periodic timer)."));

        // In a real application, you would do something here in response to the interrupt.
      }
    }

    m_pMessageDisplay->FormatInfo(_T("[Interrupt Thread] Terminating."));

    return NULL;
  }

#elif defined(__VXWORKS__) || defined(__vxworks)

  //
  // VxWorks implementation uses Semaphore registration technique.
  //

  void* ThreadFunction() {
    // Enabled FPGA internal interrupts
    const uint32_t enabled = 0x3U; // In a more complex application, this might not be a constant.

    ADMXRC3_STATUS status;
    uint32_t pending = 0;

    m_pMessageDisplay->FormatInfo(_T("[Interrupt Thread] Starting (VxWorks implementation)."));

    //
    // The registered Semaphore will be set to the signaled state by a falling
    // edge on the "firq_l" signal.
    //
    status = ADMXRC3_RegisterVxwSem(m_hDevice, ADMXRC3_EVENT_FPGAINTERRUPT(0), m_hSem);
    if (ADMXRC3_SUCCESS != status) {
      m_pMessageDisplay->FormatError(_T("[Interrupt Thread] Failed to register Semaphore for FPGA interrupt notifications: %s(0x%X)."), ADMXRC3_GetStatusString(status, TRUE), status);
      goto done;
    }

    while (!m_bTerminate) {
      //
      // Clear only the pending internal interrupts that we sampled asserted.
      //
      // This will also force the "firq_l" signal (driven by the FPGA to the Bridge)
      // deasserted for a clock cycle, which will cause a falling edge and immediate
      // signaling of the Win32 Event if there happen to still be pending interrupts.
      //
      writeRegister(&m_pRegisters->IStatus, pending);

      semTake(m_hSem, WAIT_FOREVER);

      // Sample the pending internal interrupts in the FPGA.
      pending = readRegister(&m_pRegisters->IStatus) & enabled;

      if (pending & 0x1) {
        m_pMessageDisplay->FormatInfo(_T("[Interrupt Thread] Saw FPGA internal interrupt 0 (caused by TEST register write)."));

        // In a real application, you would do something here in response to the interrupt.
      }

      if (pending & 0x2) {
        m_pMessageDisplay->FormatInfo(_T("[Interrupt Thread] Saw FPGA internal interrupt 1 (caused by periodic timer)."));

        // In a real application, you would do something here in response to the interrupt.
      }
    }

done:
    m_pMessageDisplay->FormatInfo(_T("[Interrupt Thread] Terminating."));

    status = ADMXRC3_UnregisterVxwSem(m_hDevice, ADMXRC3_EVENT_FPGAINTERRUPT(0), m_hSem);
    assert(ADMXRC3_SUCCESS == status);

    return NULL;
  }

#endif

  void Cleanup() {
#ifdef _WIN32
    CloseHandle(m_hEvent);
    m_hEvent = NULL;
#elif defined(__VXWORKS__) || defined(__vxworks)
    semDelete(m_hSem);
    m_hSem = NULL;
#endif
  }

public:
  volatile bool m_bTerminate;

protected:
  CExAppMessageDisplay* m_pMessageDisplay;
  ADMXRC3_HANDLE m_hDevice;
  ITestRegisters* m_pRegisters;
#ifdef _WIN32
  HANDLE m_hEvent;
#elif defined(__VXWORKS__) || defined(__vxworks)
  SEM_ID m_hSem;
#endif
};

CExAppMessageDisplay g_itestMessageDisplay;

#ifdef _WIN32
static const TCHAR* g_pPathSeparator = _T("\\");
#else
static const TCHAR* g_pPathSeparator = _T("/");
#endif

static const TCHAR* g_pDesignName = _T("itest");

static void
writeRegister(
  uint32_t* pRegister,
  uint32_t val32)
{
  *((volatile uint32_t*)pRegister) = cpu_to_le32(val32);
}

static uint32_t
readRegister(
  uint32_t* pRegister)
{
  uint32_t val32 = *((volatile uint32_t*)pRegister);

  return le32_to_cpu(val32);
}

static void
writeTESTRegister(
  ITestRegisters* pRegisters)
{
  /* Generate a pulse on internal interrupt 0 in the FPGA */
  writeRegister(&pRegisters->Test, 0x1);

  readRegister(&pRegisters->Test); /* Read back to ensure that write has completed. */
}

static int /* ExitCode */
admxrc3ITest(
  ADMXRC3_HANDLE hDevice,
  ADMXRC3_HANDLE hDeviceExtra, // To be used by interrupt handler thread
  const TCHAR* pBitPath,
  const TCHAR* pBitFile)
{
  ITestRegisters* pRegisters = NULL; // Pointer to the Interrupt Test registers in the FPGA, mapped into our virtual address space
  CExAppMessageDisplay* pMessageDisplay = &g_itestMessageDisplay;
  int exitCode = ExitCodeSuccess;
  ADMXRC3_STATUS status;
  ADMXRC3_CARD_INFOEX cardInfo;
  SType sBitFile;
  bool bDone = false;
  CInterruptThread* pInterruptThread = NULL;

  // Get card information; needed below.
  status = ADMXRC3_GetCardInfoEx(hDevice, &cardInfo);
  if (ADMXRC3_SUCCESS != status) {
    pMessageDisplay->FormatError(_T("Failed to get card information: %s(0x%X)"), ADMXRC3_GetStatusString(status, TRUE), status);
    exitCode = ExitCodeGetCardInfoFailed;
    goto done;
  }

  // Check that we are running on a supported model
  switch (cardInfo.Model) {
  case ADMXRC3_MODEL_ADMXRCKU1:
  case ADMXRC3_MODEL_ADMXRCKU1AT8:
    break;

  default:
    pMessageDisplay->FormatError(_T("This program must be run on one of the following models: ADM-XRC-KU1, ADM-XRC-KU1A-T8. The device in use is: %s"), ADMXRC3UT_GetModelName(cardInfo.Model));
    exitCode = ExitCodeUnsupportedModel;
    goto done;
  }

  // Map the registers, addressable by Direct Slave, into our virtual address space
  status = ADMXRC3_MapWindow(hDevice, 0, 0, sizeof(*pRegisters), (void**)&pRegisters);
  if (ADMXRC3_SUCCESS != status) {
    pMessageDisplay->FormatError(_T("Failed to map registers into our virtual address space: %s(0x%X)"), ADMXRC3_GetStatusString(status, TRUE), status);
    exitCode = ExitCodeMapWindowFailed;
    goto done;
  }

  // Configure the target FPGA with the .bit file for the "itest" FPGA design.
  if (pBitFile == NULL) {
    SStreamType bitFile;
    //
    // Construct a .bit filename of the form:
    //   pBitPath + "/<design>-<device>[_<speed>][_<stepping>]/<design>.bit"
    //
    // Examples:
    //   pBitPath + "/itest-ku115_2e/itest.bit"
    //
    if (pBitPath == NULL) {
      pMessageDisplay->AddError(_T("'pBitFile' and 'pBitPath' parameters cannot both be NULL."));
      exitCode = ExitCodeBadBitPath;
      goto done;
    }
    ADMXRC3_FPGA_INFO fpgaInfo;
    status = ADMXRC3_GetFpgaInfo(hDevice, 0, &fpgaInfo);
    if (ADMXRC3_SUCCESS != status) {
      pMessageDisplay->FormatError(_T("Failed to get FPGA information: %s(0x%X)"), ADMXRC3_GetStatusString(status, TRUE), status);
      exitCode = ExitCodeFpgaInfoFailed;
      goto done;
    }
    bitFile << pBitPath << g_pPathSeparator << g_pDesignName << _T("-");
    switch (fpgaInfo.DeviceCode) {
    case ADMXRC3_FPGA_KU060:
      bitFile << _T("ku060");
      break;

    case ADMXRC3_FPGA_KU115:
      bitFile << _T("ku115");
      break;

    default:
      pMessageDisplay->FormatError(_T("Unrecognized FPGA device code (0x%X). Cannot construct .bit filename automatically."), fpgaInfo.DeviceCode);
      exitCode = ExitCodeUnrecognizedFpga;
      goto done;
    }
    if (_tcslen(fpgaInfo.SpeedGrade) > 0) {
      bitFile << _T("_") << fpgaInfo.SpeedGrade;
    }
    if (_tcslen(fpgaInfo.Stepping) > 0) {
      bitFile << _T("_") << fpgaInfo.Stepping;
    }
    bitFile << g_pPathSeparator << g_pDesignName << _T(".bit");
    sBitFile = bitFile.str();
    std::transform(sBitFile.begin(), sBitFile.end(), sBitFile.begin(), ::tolower);
    pBitFile = sBitFile.c_str();
  }

  pMessageDisplay->FormatInfo(_T("Configuring target FPGA with '%s' ..."), pBitFile);

  status = ADMXRC3_ConfigureFromFile(hDevice, 0, 0, pBitFile);
  if (ADMXRC3_SUCCESS != status) {
    pMessageDisplay->FormatError(_T("Failed to configure target FPGA 0 with '%s': %s(0x%X)"), pBitFile, ADMXRC3_GetStatusString(status, TRUE), status);
    exitCode = ExitCodeConfigureFpgaFailed;
    goto done;
  }

  pMessageDisplay->FormatInfo(_T("Creating interrupt handler thread ..."));

  pInterruptThread = new CInterruptThread(&g_itestMessageDisplay, hDeviceExtra, pRegisters);
  if (NULL == pInterruptThread) {
    pMessageDisplay->FormatError(_T("Failed to create interrupt handler thread."));
    exitCode = ExitCodeLaunchThreadFailed;
    goto done;
  }

  if (!pInterruptThread->Initialize()) {
    pMessageDisplay->FormatError(_T("Failed to initialize interrupt handler thread."));
    exitCode = ExitCodeInitializeThreadFailed;
    goto done;
  }

  pMessageDisplay->FormatInfo(_T("Setting periodic timer to interrupt every 5 s ..."));
  // Start periodic timer interrupt with period 5000 ms.
  writeRegister(&pRegisters->Period, 0x80000000 | 5000);
  readRegister(&pRegisters->Test); /* Read back to ensure that write has completed. */

  pMessageDisplay->FormatInfo(_T("Enabling internal FPGA interrupts 0 & 1 ..."));

  // Enable FPGA internal interrupts 0 & 1
  writeRegister(&pRegisters->IEnableSet, 0x3);
  readRegister(&pRegisters->IEnableSet); /* Read back to ensure that write has completed. */

  pMessageDisplay->FormatInfo(_T("Enter 'i' to generate FPGA internal interrupt 0, or enter 'q' to quit:"));

  do {
#ifdef _UNICODE
    wint_t c;
#else
    int c;
#endif
    
    c = _gettc(stdin);
    if (EOF == c) {
      bDone = true;
    } else {
      if (_totlower(c) == _T('q')) {
        bDone = true;
      } else if (_totlower(c) == _T('i')) {
        pMessageDisplay->FormatInfo(_T("Writing 0x1 to TEST register ..."));

        writeTESTRegister(pRegisters);
      }
    }
  } while (!bDone);

  pMessageDisplay->FormatInfo(_T("Terminating interrupt handler thread ..."));

  pInterruptThread->Shutdown();

done:
  if (NULL != pInterruptThread) {
    delete pInterruptThread;
  }

  if (NULL != pRegisters) {
    ADMXRC3_UnmapWindow(hDevice, (void*)pRegisters);
    pRegisters = NULL;
  }

  if (ADMXRC3_HANDLE_INVALID_VALUE != hDeviceExtra) {
    ADMXRC3_Close(hDeviceExtra);
    hDeviceExtra = ADMXRC3_HANDLE_INVALID_VALUE;
  }

  if (ADMXRC3_HANDLE_INVALID_VALUE != hDevice) {
    ADMXRC3_Close(hDevice);
    hDevice = ADMXRC3_HANDLE_INVALID_VALUE;
  }

  return exitCode;
}

int /* ExitCode */
admxrc3ITestIndex(
  const TCHAR* pBitPath,
  const TCHAR* pBitFile,
  unsigned int index)
{
  ADMXRC3_HANDLE hDevice = ADMXRC3_HANDLE_INVALID_VALUE;
  ADMXRC3_HANDLE hDeviceExtra = ADMXRC3_HANDLE_INVALID_VALUE;
  ADMXRC3_STATUS status;

  status = ADMXRC3UT_OpenByIndex(index, FALSE, TRUE, 0, &hDevice);
  if (ADMXRC3_SUCCESS != status) {
    g_itestMessageDisplay.FormatError(
      _T("Failed to open device with index %u: %s(0x%X)"),
      index, ADMXRC3_GetStatusString(status, TRUE), status);
    return ExitCodeCannotOpenIndex;
  }

  //
  // According to the rules about nonblocking operations in the ADMXRC3 API,
  // Each concurrent nonblocking operation in Linux should have its own device
  // handle, which cannot be reused for a 2nd nonblocking operation until the
  // first nonblocking operation is complete.
  //
  // To make this program portable between Linux and Windows, we get another device
  // handle to be used solely by the interrupt handling thread, which calls
  // ADMXRC3_StartNotificationWait ... ADMXRC3_FinishNotificationWait.
  //
  status = ADMXRC3UT_OpenByIndex(index, FALSE, TRUE, 0, &hDeviceExtra);
  if (ADMXRC3_SUCCESS != status) {
    ADMXRC3_Close(hDevice);

    g_itestMessageDisplay.FormatError(
      _T("Failed to open device with index %u: %s(0x%X)"),
      index, ADMXRC3_GetStatusString(status, TRUE), status);
    return ExitCodeCannotOpenIndex;
  }

  return admxrc3ITest(
    hDevice,
    hDeviceExtra,
    pBitPath,
    pBitFile);
}

int /* ExitCode */
admxrc3ITestSN(
  const TCHAR* pBitPath,
  const TCHAR* pBitFile,
  uint32_t serialNumber)
{
  ADMXRC3_HANDLE hDevice = ADMXRC3_HANDLE_INVALID_VALUE;
  ADMXRC3_HANDLE hDeviceExtra = ADMXRC3_HANDLE_INVALID_VALUE;
  ADMXRC3_STATUS status;

  status = ADMXRC3UT_OpenBySerialNumber(serialNumber, FALSE, TRUE, 0, &hDevice);
  if (ADMXRC3_SUCCESS != status) {
    g_itestMessageDisplay.FormatError(
      _T("Failed to open device with serial number %lu: %s(0x%X)"),
      (unsigned long)serialNumber, ADMXRC3_GetStatusString(status, TRUE), status);
    return ExitCodeCannotOpenSerial;
  }

  //
  // According to the rules about nonblocking operations in the ADMXRC3 API,
  // Each concurrent nonblocking operation in Linux should have its own device
  // handle, which cannot be reused for a 2nd nonblocking operation until the
  // first nonblocking operation is complete.
  //
  // To make this program portable between Linux and Windows, we get another device
  // handle to be used solely by the interrupt handling thread, which calls
  // ADMXRC3_StartNotificationWait ... ADMXRC3_FinishNotificationWait.
  //
  status = ADMXRC3UT_OpenBySerialNumber(serialNumber, FALSE, TRUE, 0, &hDeviceExtra);
  if (ADMXRC3_SUCCESS != status) {
    ADMXRC3_Close(hDevice);

    g_itestMessageDisplay.FormatError(
      _T("Failed to open device with serial number %lu: %s(0x%X)"),
      (unsigned long)serialNumber, ADMXRC3_GetStatusString(status, TRUE), status);
    return ExitCodeCannotOpenSerial;
  }

  return admxrc3ITest(
    hDevice,
    hDeviceExtra,
    pBitPath,
    pBitFile);
}
