--==============================================================================
--
--     File Name: axi4_128_pkg.vhd
--         Model: Synthesisable
--
--  Dependencies:
--
--       Company: Alpha Data
--        Design: Package axi4_128
--
--   Description: Package defining AXI4 128 profile.
--
--   Limitations:
--
--         Notes:
--
--    Disclaimer: THESE DESIGNS ARE PROVIDED "AS IS" WITH NO WARRANTY
--                WHATSOEVER AND ALPHA DATA SPECIFICALLY DISCLAIMS ANY
--                IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR
--                A PARTICULAR PURPOSE, OR AGAINST INFRINGEMENT.
--
--                Copyright (C) 2013-2015 Alpha Data, All rights reserved
--
--     Revisions: 0.1, KDC, 23 Mar 2013: Initial design
--                0.2, KDC, 10 Sep 2015:
--                     - Changed AXI4_128_ADDR_WIDTH from 32 to 64
--                     - Changed AXI4_128_LEN_WIDTH from 4 to 8
--                     - Changed AXI4_128_LOCK_WIDTH from 2 to 1
--                     - Added subtype axi4_128_region_s
--                     - Added awregion and arregion
--                     - Added wid
--
--==============================================================================

--
-- Exported functions:
--
--
--
-- Exported procedures:
--
--

library ieee;
use ieee.std_logic_1164.all;

package axi4_128 is

  -- Define AXI4 128 widths
  constant AXI4_128_RESP_WIDTH      : integer := 2;
  constant AXI4_128_ADDR_WIDTH      : integer := 64;
  constant AXI4_128_LEN_WIDTH       : integer := 8;
  constant AXI4_128_ID_WIDTH        : integer := 12;
  constant AXI4_128_DATA_WIDTH      : integer := 128;
  constant AXI4_128_DATA_WIDTH_LOG2 : integer := 7;
  constant AXI4_128_STRB_WIDTH      : integer := 16;
  constant AXI4_128_STRB_WIDTH_LOG2 : integer := 4;

  constant AXI4_128_SIZE_WIDTH      : integer := 3;
  constant AXI4_128_BURST_WIDTH     : integer := 2;
  constant AXI4_128_CACHE_WIDTH     : integer := 4;
  constant AXI4_128_PROT_WIDTH      : integer := 3;
  constant AXI4_128_LOCK_WIDTH      : integer := 1;
  constant AXI4_128_QOS_WIDTH       : integer := 4;
  constant AXI4_128_REGION_WIDTH    : integer := 4;


  -- Define AXI4 128 subtypes
  subtype axi4_128_resp_s is std_logic_vector(AXI4_128_RESP_WIDTH-1 downto 0);
  constant AXI4_128_RESP_OKAY   : axi4_128_resp_s := "00";
  constant AXI4_128_RESP_EXOKAY : axi4_128_resp_s := "01";
  constant AXI4_128_RESP_SLVERR : axi4_128_resp_s := "10";
  constant AXI4_128_RESP_DECERR : axi4_128_resp_s := "11";

  subtype axi4_size_s   is std_logic_vector(AXI4_128_SIZE_WIDTH-1 downto 0);
  subtype axi4_burst_s  is std_logic_vector(AXI4_128_BURST_WIDTH-1 downto 0);
  subtype axi4_cache_s  is std_logic_vector(AXI4_128_CACHE_WIDTH-1 downto 0);
  subtype axi4_prot_s   is std_logic_vector(AXI4_128_PROT_WIDTH-1 downto 0);
  subtype axi4_lock_s   is std_logic_vector(AXI4_128_LOCK_WIDTH-1 downto 0);
  subtype axi4_qos_s    is std_logic_vector(AXI4_128_QOS_WIDTH-1 downto 0);
  subtype axi4_region_s is std_logic_vector(AXI4_128_REGION_WIDTH-1 downto 0);

  constant AXI4_128_SIZE   : axi4_size_s   := "100";  -- 128 bits wide
  constant AXI4_128_BURST  : axi4_burst_s  := "01";   -- INCRemental
  constant AXI4_128_CACHE  : axi4_cache_s  := "0011"; -- Xilinx recommended value
  constant AXI4_128_PROT   : axi4_prot_s   := "000";  -- Xilinx recommended value
  constant AXI4_128_LOCK   : axi4_lock_s   := "0";    -- Xilinx recommended value
  constant AXI4_128_QOS    : axi4_qos_s    := "0000"; -- Xilinx recommended value
  constant AXI4_128_REGION : axi4_region_s := "0000"; -- Xilinx recommended value

  subtype axi4_128_addr_s   is std_logic_vector(AXI4_128_ADDR_WIDTH-1 downto 0);
  subtype axi4_128_len_s    is std_logic_vector(AXI4_128_LEN_WIDTH-1 downto 0);
  subtype axi4_128_id_s     is std_logic_vector(AXI4_128_ID_WIDTH-1 downto 0);
  subtype axi4_128_data_s   is std_logic_vector(AXI4_128_DATA_WIDTH-1 downto 0);
  subtype axi4_128_strb_s   is std_logic_vector(AXI4_128_STRB_WIDTH-1 downto 0);

  subtype axi4_128_size_s   is std_logic_vector(AXI4_128_SIZE_WIDTH-1 downto 0);
  subtype axi4_128_burst_s  is std_logic_vector(AXI4_128_BURST_WIDTH-1 downto 0);
  subtype axi4_128_cache_s  is std_logic_vector(AXI4_128_CACHE_WIDTH-1 downto 0);
  subtype axi4_128_prot_s   is std_logic_vector(AXI4_128_PROT_WIDTH-1 downto 0);
  subtype axi4_128_lock_s   is std_logic_vector(AXI4_128_LOCK_WIDTH-1 downto 0);
  subtype axi4_128_qos_s    is std_logic_vector(AXI4_128_QOS_WIDTH-1 downto 0);
  subtype axi4_128_region_s is std_logic_vector(AXI4_128_REGION_WIDTH-1 downto 0);


  -- AXI4 128 interface types
  -- Data flow from Master to Slave
  type axi4_128_m2s_t is record
    -- Write address channel
    awaddr   : axi4_128_addr_s; -- Write Address (16-byte aligned)
    awvalid  : std_logic;       -- Write Address Valid
    awid     : axi4_128_id_s;   -- Write Address ID
    awlen    : axi4_128_len_s;  -- Write Address burst length
    awsize   : axi4_128_size_s;
    awburst  : axi4_128_burst_s;
    awcache  : axi4_128_cache_s;
    awprot   : axi4_128_prot_s;
    awlock   : axi4_128_lock_s;
    awqos    : axi4_128_qos_s;
    awregion : axi4_128_region_s;
    -- Read address channel
    araddr   : axi4_128_addr_s; -- Read Address (16-byte aligned)
    arvalid  : std_logic;       -- Read Address Valid
    arid     : axi4_128_id_s;   -- Read Address ID
    arlen    : axi4_128_len_s;  -- Read Address burst length
    arsize   : axi4_128_size_s;
    arburst  : axi4_128_burst_s;
    arcache  : axi4_128_cache_s;
    arlock   : axi4_128_lock_s;
    arprot   : axi4_128_prot_s;
    arqos    : axi4_128_qos_s;
    arregion : axi4_128_region_s;
    -- Write data channel
    wdata    : axi4_128_data_s; -- Write Data
    wstrb    : axi4_128_strb_s; -- Write Data Strobes
    wvalid   : std_logic;       -- Write Data Valid
    wid      : axi4_128_id_s;   -- Write Data ID
    wlast    : std_logic;       -- Write Data Last
    -- Write response channel
    bready   : std_logic;       -- Write Response Ready
    -- Read data channel
    rready   : std_logic;       -- Read Data Ready
  end record;

  -- Data flow from Slave to Master
  type axi4_128_s2m_t is record
    -- Write address channel
    awready : std_logic;       -- Write Address Ready
    -- Read address channel
    arready : std_logic;       -- Read Address Ready
    -- Write data channel
    wready  : std_logic;       -- Write Data Ready
    -- Write response channel
    bresp   : axi4_128_resp_s; -- Write Response
    bvalid  : std_logic;       -- Write Response Valid
    bid     : axi4_128_id_s;   -- Write Response ID
    -- Read data channel
    rdata   : axi4_128_data_s; -- Read Data
    rresp   : axi4_128_resp_s; -- Read Data Response
    rvalid  : std_logic;       -- Read Data Valid
    rid     : axi4_128_id_s;   -- Read Data ID
    rlast   : std_logic;       -- Read Data Last
  end record;

  type axi4_128_m2s_t_array is array (natural range <>) of axi4_128_m2s_t;
  type axi4_128_s2m_t_array is array (natural range <>) of axi4_128_s2m_t;


  -- Define AXI4 128 interface type defaults
  constant AXI4_128_DEFAULT_M2S : axi4_128_m2s_t := (
    awaddr   => (others => '-'),
    awvalid  => '0',
    awid     => (others => '-'),
    awlen    => (others => '-'),
    awsize   => (others => '-'),
    awburst  => (others => '-'),
    awcache  => (others => '-'),
    awprot   => (others => '-'),
    awlock   => (others => '-'),
    awqos    => (others => '-'),
    awregion => (others => '-'),
    --
    araddr   => (others => '-'),
    arvalid  => '0',
    arid     => (others => '-'),
    arlen    => (others => '-'),
    arsize   => (others => '-'),
    arburst  => (others => '-'),
    arcache  => (others => '-'),
    arprot   => (others => '-'),
    arlock   => (others => '-'),
    arqos    => (others => '-'),
    arregion => (others => '-'),
    --
    wdata    => (others => '-'),
    wstrb    => (others => '-'),
    wvalid   => '0',
    wid      => (others => '-'),
    wlast    => '-',
    --
    bready   => '0',
    rready   => '0');

  constant AXI4_128_DEFAULT_S2M : axi4_128_s2m_t := (
    awready => '0',
    arready => '0',
    wready  => '0',
    --
    bresp   => (others => '-'),
    bvalid  => '0',
    bid     => (others => '-'),
    --
    rdata   => (others => '-'),
    rresp   => (others => '-'),
    rvalid  => '0',
    rid     => (others => '-'),
    rlast   => '-');

  constant AXI4_128_FLOAT_M2S : axi4_128_m2s_t := (
    awaddr   => (others => 'W'),
    awvalid  => 'L',
    awid     => (others => 'W'),
    awlen    => (others => 'W'),
    awsize   => (others => 'W'),
    awburst  => (others => 'W'),
    awcache  => (others => 'W'),
    awprot   => (others => 'W'),
    awlock   => (others => 'W'),
    awqos    => (others => 'W'),
    awregion => (others => 'W'),
    --
    araddr   => (others => 'W'),
    arvalid  => 'L',
    arid     => (others => 'W'),
    arlen    => (others => 'W'),
    arsize   => (others => 'W'),
    arburst  => (others => 'W'),
    arcache  => (others => 'W'),
    arprot   => (others => 'W'),
    arlock   => (others => 'W'),
    arqos    => (others => 'W'),
    arregion => (others => 'W'),
    --
    wdata    => (others => 'W'),
    wstrb    => (others => 'W'),
    wvalid   => 'L',
    wid      => (others => 'W'),
    wlast    => 'W',
    --
    bready   => 'L',
    rready   => 'L');

end package axi4_128;

package body axi4_128 is

end package body axi4_128;