proc make_project { } {
  # Get the directory containing this script. Source file paths are relative to this.
  set origin_dir [file dirname [file normalize [info script]]]

  # Define versions of dependencies in one convenient place
  source "$origin_dir/version.tcl"

  set viv_ver [version -short]
  set viv_spl [split $viv_ver .]
  set viv_maj [lindex $viv_spl 0]
  set viv_min [lindex $viv_spl 1]

  if { $viv_maj < 2018 || ($viv_maj == 2018 && $viv_min < 3) } {
    send_msg_id {AD-MKXPR-001} {ERROR} "Vivado ${viv_ver} is not supported for this FPGA design. Minimum Vivado version is 2018.3."
  }

  # Create project
  set device {ku115_2i}
  set dir "simple_ocp-${device}"
  if { [info exists ::env(ADMXRC3_PKG_TEST)] } {
    append dir {-} [version -short]
  }
  set part {xcku115-flva1517-2-i}
  set name {simple_ocp}
  set project [create_project -force -part $part $name "$origin_dir/$dir"]
  current_project $project

  # Set the directory path for the new project
  set proj_dir [get_property directory [current_project]]

  # Set project properties
  set_property "default_lib" "xil_defaultlib" $project
  set_property "part" $part $project
  set_property "simulator_language" "Mixed" $project
  set_property "target_language" "Verilog" $project

  set repo_paths [list \
    [file normalize "$origin_dir/../../../../fpga/repo/vivado-2019.2"] \
  ]
  set_property "ip_repo_paths" $repo_paths $project
  update_ip_catalog

  # Create 'sources_1' fileset (if not found)
  if {[string equal [get_filesets -quiet sources_1] ""]} {
    create_fileset -srcset sources_1
  }
  set source_set [get_filesets sources_1]

  # Set source set properties
  #set_property "top" "simple" $source_set

  # Add HDL source files
  set hdl_files [list \
    [file normalize "$origin_dir/../src/adb3_ocp_pkg.vhd"] \
    [file normalize "$origin_dir/../src/adb3_ocp_tb_pkg.vhd"] \
    [file normalize "$origin_dir/../src/tb_simple_ocp_stim.vhd"] \
    [file normalize "$origin_dir/../src/mptl_tgt_ku_ocp_sim.vhd"] \
    [file normalize "$origin_dir/../src/mptl_tgt_ku_ocp_wrap.vhd"] \
    [file normalize "$origin_dir/../src/adb3_ocp_simple_bus_if.vhd"] \
    [file normalize "$origin_dir/../src/simple_ocp.vhd"] \
  ]
  add_files -norecurse -fileset $source_set $hdl_files

  # Create IP
  create_ip -vlnv "alphadata:user:mptl_tgt_ku_slv:${hsocp_ver}" -module_name mptl_tgt_ku_slv_ip
  set mptl_tgt_ku_slv_ip [get_ips mptl_tgt_ku_slv_ip]
  set_property -dict { \
    CONFIG.break_out_ocp {false} \
    CONFIG.enable_bridge_master {true} \
    CONFIG.enable_dma {false} \
    CONFIG.enable_target_master {false} \
    CONFIG.wrapper_reset_signals {true} \
  } $mptl_tgt_ku_slv_ip

  # Add constraints files
  # Put target .xdc as LAST in list
  set constraint_set [get_filesets constrs_1]
  set constraint_files [list \
    [file normalize "$origin_dir/../src/refclk300.xdc"] \
    [file normalize "$origin_dir/../src/mptl-admxrcku1.xdc"] \
    [file normalize "$origin_dir/../src/bitstream.xdc"] \
    [file normalize "$origin_dir/../src/simple_ocp.xdc"] \
  ]
  add_files -norecurse -fileset $constraint_set $constraint_files
  set_property "target_constrs_file" [lindex $constraint_files end] $constraint_set

  # Add utilities files (Tcl hook scripts etc.)
  set utils_set [get_filesets -quiet utils_1]
  if {[string equal $utils_set ""]} {
    # Utilities fileset introduced in Vivado 2018.3; fall back to using sources set if not present.
    set utils_set $source_set
  }
  #set utils_files [list \
  #]
  #add_files -norecurse -fileset $utils_set $utils_files
  # Set used_in_XXX properties so that Vivado 2018.2 and earlier doesn't treat them as design sources.
  #set_property -dict {
  #  {used_in_synthesis} {false} \
  #  {used_in_implementation} {false} \
  #  {used_in_simulation} {false} \
  #} [get_files $utils_files]

  set sim_set [get_filesets sim_1]

  # Add simulation HDL source files
  set sim_files [list \
    [file normalize "$origin_dir/../src/tb_simple_ocp.vhd"] \
  ]
  add_files -norecurse -fileset $sim_set $sim_files

  # Set fileset properties
  set_property "source_set" "sources_1" $sim_set
  set_property "top" "tb_simple" $sim_set
  set_property "generic" "sim_ocp=true" $sim_set
  # XSIM incremental compilation seems unreliable for large projects
  set_property "incremental" "false" $sim_set
  set_property "runtime" "25 us" $sim_set
  # Set runtime on per-simulator basis
  foreach simulator [get_simulators] {
    set_property "target_simulator" "$simulator" $project
    set_property "${simulator}.simulate.runtime" "25 us" $sim_set
  }
  set_property "xsim.simulate.custom_tcl" [file normalize "$origin_dir/xsim_wave.tcl"] $sim_set
  set_property "modelsim.simulate.custom_udo" [file normalize "$origin_dir/vsim_wave.udo"] $sim_set
  # Return to using XSim
  set_property target_simulator {XSim} $project

  set synth_run [get_runs synth_1]
  #set_property strategy {Flow_PerfOptimized_high} $synth_run
  set_property "needs_refresh" "1" $synth_run
  set_property "part" $part $synth_run
  current_run -synthesis $synth_run

  set impl_run [get_runs impl_1]
  set_property "needs_refresh" "1" $impl_run
  set_property "part" $part $impl_run
  set_property -name {steps.route_design.args.more options} -value -timing_summary -objects $impl_run
  current_run -implementation $impl_run

  update_compile_order -fileset $source_set
  if { [info exists sim_set] } {
    update_compile_order -fileset $sim_set
  }

  return $project
}

return [make_project]
