--==============================================================================
--
--     File Name: adb3_ocp_pkg.vhd
--         Model: Synthesisable
--
--  Dependencies:
--
--       Company: Alpha Data
--        Design: Package adb3_ocp_pkg
--
--   Description: Package defining ADB3 OCP profiles.
--
--   Limitations:
--
--         Notes:
--
--    Disclaimer: THESE DESIGNS ARE PROVIDED "AS IS" WITH NO WARRANTY
--                WHATSOEVER AND ALPHA DATA SPECIFICALLY DISCLAIMS ANY
--                IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR
--                A PARTICULAR PURPOSE, OR AGAINST INFRINGEMENT.
--
--                Copyright (C) 2009-2015 Alpha Data, All rights reserved
--
--     Revisions: 0.1, KDC, 06 Nov 2009: Initial design
--                0.2, TW,  09 Feb 2011:
--                     - Added some more useful constants and types
--                0.3, KDC, 28 Dec 2011:
--                     - Moved make_defined and make_defined_s2m functions from
--                       adb3_target_pkg. Added new function make_defined_s2m_array
--                0.4, TW,  17 Jan 2012:
--                     - Added matching make_defined_m2s and
--                       make_defined_m2s_array functions.
--                0.5, KDC, 18 Jan 2012:
--                     - Changed ADB3 OCP type comments to use AXI4 terminology
--                     - Added ADB3 OCP Lite and ADB3 OCP Stream types
--                     - Changed default OCP record types to use (others => '-')
--                       instead of (others => '0') for address, data, etc
--                     - Added constant ADB3_OCP_FLOAT_M2S
--                     - Added functions adb3_ocp_cmd_is_wr, adb3_ocp_cmd_is_rd,
--                       adb3_ocp_cmd_is_idle and procedure adb3_ocp_l_wr
--                0.6, KDC, 13 Feb 2013:
--                     - Added functions to_std_logic_vector for adb3_ocp_l
--                     - Added functions to_adb3_ocp_l_m2sT, to_adb3_ocp_l_s2mT
--                0.7, KDC, 22 Oct 2013:
--                     - Added function make_defined_sl for std_logic. Changed
--                       function make_defined_slv to use it.
--                0.8, KDC, 13 Jan 2014:
--                     - Moved definitions adb3_ocp_addr_range_t, adb3_ocp_l_addr_range_t,
--                       adb3_ocp_addr_range_table_t, adb3_ocp_l_addr_range_table_t
--                       from adb3_ocp_comp_pkg
--                     - Changed constant names to upper case
--                0.9, KDC, 11 Feb 2015:
--                     - Corrected calculation of ocp_m2s_l.DataByteEn and
--                       ocp_m2s_l.Data indexes in procedure adb3_ocp_l_wr
--
--==============================================================================

--
-- Exported functions:
--
--   adb3_ocp_cmd_is_wr
--   adb3_ocp_cmd_is_rd
--   adb3_ocp_cmd_is_idle
--   to_std_logic
--   to_std_logic_vector (adb3_ocp_m2sT)
--   to_std_logic_vector (adb3_ocp_s2mT)
--   to_std_logic_vector (adb3_ocp_l_m2sT)
--   to_std_logic_vector (adb3_ocp_l_s2mT)
--   to_std_logic_vector (adb3_ocp_m2sT_array)
--   to_std_logic_vector (adb3_ocp_s2mT_array)
--   to_adb3_ocp_m2sT
--   to_adb3_ocp_s2mT
--   to_adb3_ocp_l_m2sT
--   to_adb3_ocp_l_s2mT
--   to_adb3_ocp_m2sT_array
--   to_adb3_ocp_s2mT_array
--   make_defined_m2s
--   make_defined_m2s_array
--   make_defined_s2m
--   make_defined_s2m_array
--   make_defined_slv
--   make_defined_sl
--
--
-- Exported procedures:
--
--   adb3_ocp_l_wr (std_logic_vector)
--   adb3_ocp_l_wr (std_logic)
--

library ieee;
use ieee.std_logic_1164.all;

package adb3_ocp is

  -- Define common widths
  constant ADB3_OCP_CMD_WIDTH  : integer := 3;
  constant ADB3_OCP_RESP_WIDTH : integer := 2;

  -- Define ADB3 OCP widths
  constant ADB3_OCP_ADDR_WIDTH      : integer := 64;
  constant ADB3_OCP_BURST_WIDTH     : integer := 12;
  constant ADB3_OCP_TAG_WIDTH       : integer := 8;
  constant ADB3_OCP_DATA_WIDTH      : integer := 128;
  constant ADB3_OCP_DATA_WIDTH_LOG2 : integer := 7;
  constant ADB3_OCP_BE_WIDTH        : integer := 16;
  constant ADB3_OCP_BE_WIDTH_LOG2   : integer := 4;

  -- Define ADB3 OCP Lite widths
  constant ADB3_OCP_L_ADDR_WIDTH      : integer := 32;
  constant ADB3_OCP_L_DATA_WIDTH      : integer := 32;
  constant ADB3_OCP_L_DATA_WIDTH_LOG2 : integer := 5;
  constant ADB3_OCP_L_BE_WIDTH        : integer := 4;
  constant ADB3_OCP_L_BE_WIDTH_LOG2   : integer := 2;


  -- Define common subtypes
  subtype ocp_CmdT is std_logic_vector(ADB3_OCP_CMD_WIDTH-1 downto 0);
  constant OCP_CMD_IDLE         : ocp_CmdT := "000";
  constant OCP_CMD_WRITE        : ocp_CmdT := "001";
  constant OCP_CMD_READ         : ocp_CmdT := "010";
  constant OCP_CMD_WRITENONPOST : ocp_CmdT := "101";

  subtype ocp_RespT is std_logic_vector(ADB3_OCP_RESP_WIDTH-1 downto 0);
  constant OCP_RESP_NONE   : ocp_RespT := "00";
  constant OCP_RESP_VALID  : ocp_RespT := "01";
  constant OCP_RESP_FAILED : ocp_RespT := "10";
  constant OCP_RESP_ERROR  : ocp_RespT := "11";

  -- Define ADB3 OCP subtypes
  subtype adb3_ocp_addr_s  is std_logic_vector(ADB3_OCP_ADDR_WIDTH-1 downto 0);
  subtype adb3_ocp_burst_s is std_logic_vector(ADB3_OCP_BURST_WIDTH-1 downto 0);
  subtype adb3_ocp_tag_s   is std_logic_vector(ADB3_OCP_TAG_WIDTH-1 downto 0);
  subtype adb3_ocp_data_s  is std_logic_vector(ADB3_OCP_DATA_WIDTH-1 downto 0);
  subtype adb3_ocp_be_s    is std_logic_vector(ADB3_OCP_BE_WIDTH-1 downto 0);

  -- Define ADB3 OCP Lite subtypes
  subtype adb3_ocp_l_addr_s is std_logic_vector(ADB3_OCP_L_ADDR_WIDTH-1 downto 0);
  subtype adb3_ocp_l_data_s is std_logic_vector(ADB3_OCP_L_DATA_WIDTH-1 downto 0);
  subtype adb3_ocp_l_be_s   is std_logic_vector(ADB3_OCP_L_BE_WIDTH-1 downto 0);


  -- ADB3 OCP interface types
  -- Compare with AXI4 signal names in ()
  -- Data flow from Master to Slave
  type adb3_ocp_m2sT is record
    -- Address channel
    Cmd         : ocp_CmdT;         -- Transaction type and validity (AWVALID/ARVALID)
    Addr        : adb3_ocp_addr_s;  -- Transaction byte address (16-byte aligned) (AWADDR/ARADDR)
    BurstLength : adb3_ocp_burst_s; -- Transaction burst length (AWLEN/ARLEN)
    Tag         : adb3_ocp_tag_s;   -- Transaction identifier (AWID/ARID)
    -- Write data channel
    DataValid   : std_logic;        -- Transfer valid (WVALID)
    Data        : adb3_ocp_data_s;  -- Transfer data (WDATA)
    DataByteEn  : adb3_ocp_be_s;    -- Transfer data byte enables (WSTRB)
    -- Read data channel
    RespAccept  : std_logic;        -- Transfer ready (RREADY)
  end record;

  -- Data flow from Slave to Master
  type adb3_ocp_s2mT is record
    -- Address channel
    CmdAccept   : std_logic;        -- Transaction ready (AWREADY/ARREADY)
    -- Write data channel
    DataAccept  : std_logic;        -- Transfer ready (WREADY)
    -- Read data channel
    Resp        : ocp_RespT;        -- Transfer type and validity (RRESP/RVALID)
    Data        : adb3_ocp_data_s;  -- Transfer data (RDATA)
    Tag         : adb3_ocp_tag_s;   -- Transaction identifier (RID)
  end record;

  type adb3_ocp_m2sT_array is array (natural range <>) of adb3_ocp_m2sT;
  type adb3_ocp_s2mT_array is array (natural range <>) of adb3_ocp_s2mT;

  constant ADB3_OCP_M2ST_SLV_LENGTH : natural := ADB3_OCP_CMD_WIDTH + ADB3_OCP_ADDR_WIDTH + ADB3_OCP_BURST_WIDTH + ADB3_OCP_DATA_WIDTH + ADB3_OCP_BE_WIDTH + 1 + 1 + ADB3_OCP_TAG_WIDTH;
  constant ADB3_OCP_S2MT_SLV_LENGTH : natural := 1 + 1 + ADB3_OCP_DATA_WIDTH + ADB3_OCP_RESP_WIDTH + ADB3_OCP_TAG_WIDTH;


  -- ADB3 OCP Lite interface types
  -- Compare with AXI4-Lite signal names in ()
  -- Data flow from Master to Slave
  type adb3_ocp_l_m2sT is record
    -- Address channel
    Cmd        : ocp_CmdT;          -- Transaction type and validity (AWVALID/ARVALID)
    Addr       : adb3_ocp_l_addr_s; -- Transaction byte address (4-byte aligned) (AWADDR/ARADDR)
    -- Write data channel
    Data       : adb3_ocp_l_data_s; -- Transfer data (WDATA)
    DataByteEn : adb3_ocp_l_be_s;   -- Transfer data byte enables (WSTRB)
  end record;

  -- Data flow from Slave to Master
  type adb3_ocp_l_s2mT is record
    -- Address channel
    CmdAccept  : std_logic;         -- Transaction ready (AWREADY/ARREADY)
    -- Read data channel
    Resp       : ocp_RespT;         -- Transfer type and validity (RRESP/RVALID)
    Data       : adb3_ocp_l_data_s; -- Transfer data (RDATA)
  end record;

  type adb3_ocp_l_m2sT_array is array (natural range <>) of adb3_ocp_l_m2sT;
  type adb3_ocp_l_s2mT_array is array (natural range <>) of adb3_ocp_l_s2mT;

  constant ADB3_OCP_L_M2ST_SLV_LENGTH : natural := ADB3_OCP_CMD_WIDTH + ADB3_OCP_L_ADDR_WIDTH + ADB3_OCP_L_DATA_WIDTH + ADB3_OCP_L_BE_WIDTH;
  constant ADB3_OCP_L_S2MT_SLV_LENGTH : natural := 1 + ADB3_OCP_L_DATA_WIDTH + ADB3_OCP_RESP_WIDTH;


  -- ADB3 OCP Stream interface types
  -- Compare with AXI4-Stream signal names in ()
  -- Data flow from Master to Slave
  type adb3_ocp_s_m2sT is record
    -- Write data channel
    DataValid  : std_logic;       -- Transfer valid (TVALID)
    Data       : adb3_ocp_data_s; -- Transfer data (TDATA)
    DataLast   : std_logic;       -- Last transfer in packet (TLAST)
  end record;

  -- Data flow from Slave to Master
  type adb3_ocp_s_s2mT is record
    -- Write data channel
    DataAccept : std_logic;       -- Ready for transfer (TREADY)
  end record;

  type adb3_ocp_s_m2sT_array is array (natural range <>) of adb3_ocp_s_m2sT;
  type adb3_ocp_s_s2mT_array is array (natural range <>) of adb3_ocp_s_s2mT;

  constant ADB3_OCP_S_M2ST_SLV_LENGTH : natural := 1 + ADB3_OCP_L_DATA_WIDTH + 1;
  constant ADB3_OCP_S_S2MT_SLV_LENGTH : natural := 1;


  -- Define ADB3 OCP interface type defaults
  constant ADB3_OCP_DEFAULT_M2S : adb3_ocp_m2sT :=
    (Cmd         => OCP_CMD_IDLE,
     Addr        => (others => '-'),
     BurstLength => (others => '-'),
     Tag         => (others => '-'),
     DataValid   => '0',
     Data        => (others => '-'),
     DataByteEn  => (others => '-'),
     RespAccept  => '0');

  constant ADB3_OCP_DEFAULT_S2M : adb3_ocp_s2mT :=
    (CmdAccept  => '0',
     DataAccept => '0',
     Resp       => OCP_RESP_NONE,
     Data       => (others => '-'),
     Tag        => (others => '-'));


  -- Define ADB3 OCP Lite interface type defaults
  constant ADB3_OCP_L_DEFAULT_M2S : adb3_ocp_l_m2sT :=
    (Cmd         => OCP_CMD_IDLE,
     Addr        => (others => '-'),
     Data        => (others => '-'),
     DataByteEn  => (others => '-'));

  constant ADB3_OCP_L_DEFAULT_S2M : adb3_ocp_l_s2mT :=
    (CmdAccept  => '0',
     Resp       => OCP_RESP_NONE,
     Data       => (others => '-'));


  -- Define ADB3 OCP Stream interface type defaults
  constant ADB3_OCP_S_DEFAULT_M2S : adb3_ocp_s_m2sT :=
    (Data        => (others => '-'),
     DataValid   => '0',
     DataLast    => '0');

  constant ADB3_OCP_S_DEFAULT_S2M : adb3_ocp_s_s2mT :=
    (DataAccept  => '0');


  -- Define ADB3 OCP full type
  constant ADB3_OCP_FLOAT_M2S : adb3_ocp_m2sT := (
    Cmd         => "LLL",
    Addr        => (others => 'W'),
    BurstLength => (others => 'W'),
    DataValid   => 'L',
    Data        => (others => 'W'),
    DataByteEn  => (others => 'W'),
    RespAccept  => 'L',
    Tag         => (others => 'W'));


  -- Define type tri-states
  constant ADB3_OCP_Z_M2S : adb3_ocp_m2sT :=
    (Cmd         => (others => 'Z'),
     Addr        => (others => 'Z'),
     BurstLength => (others => 'Z'),
     Tag         => (others => 'Z'),
     DataValid   => 'Z',
     Data        => (others => 'Z'),
     DataByteEn  => (others => 'Z'),
     RespAccept  => 'Z');

  constant ADB3_OCP_Z_S2M : adb3_ocp_s2mT :=
    (CmdAccept  => 'Z',
     DataAccept => 'Z',
     Resp       => (others => 'Z'),
     Data       => (others => 'Z'),
     Tag        => (others => 'Z'));


  -- Define Address range table types
  type adb3_ocp_addr_range_t is record
    addr_base : adb3_ocp_addr_s;
    addr_mask : adb3_ocp_addr_s;
  end record;

  type adb3_ocp_l_addr_range_t is record
    addr_base : adb3_ocp_l_addr_s;
    addr_mask : adb3_ocp_l_addr_s;
  end record;

  type adb3_ocp_addr_range_table_t   is array (natural range <>) of adb3_ocp_addr_range_t;
  type adb3_ocp_l_addr_range_table_t is array (natural range <>) of adb3_ocp_l_addr_range_t;


  -- Define ADB3 OCP procedures
  procedure adb3_ocp_l_wr(
    signal reg       : inout  std_logic_vector;
    signal ocp_m2s_l : in     adb3_ocp_l_m2sT;
    constant offset  : in     natural);

  procedure adb3_ocp_l_wr(
    signal reg_out   : out    std_logic_vector;
    signal reg_in    : in     std_logic_vector;
    signal ocp_m2s_l : in     adb3_ocp_l_m2sT;
    constant offset  : in     natural);

  procedure adb3_ocp_l_wr(
    signal reg       : inout  std_logic;
    signal ocp_m2s_l : in     adb3_ocp_l_m2sT;
    constant offset  : in     natural);

  procedure adb3_ocp_l_wr(
    signal reg_out   : out    std_logic;
    signal reg_in    : in     std_logic;
    signal ocp_m2s_l : in     adb3_ocp_l_m2sT;
    constant offset  : in     natural);


  -- Define ADB3 OCP functions
  function adb3_ocp_cmd_is_wr(cmd : ocp_CmdT) return std_logic;
  function adb3_ocp_cmd_is_rd(cmd : ocp_CmdT) return std_logic;
  function adb3_ocp_cmd_is_idle(cmd : ocp_CmdT) return std_logic;
  function to_std_logic(bool : boolean) return std_logic;
  function to_std_logic_vector(x : adb3_ocp_m2sT) return std_logic_vector;
  function to_std_logic_vector(x : adb3_ocp_s2mT) return std_logic_vector;
  function to_std_logic_vector(x : adb3_ocp_l_m2sT) return std_logic_vector;
  function to_std_logic_vector(x : adb3_ocp_l_s2mT) return std_logic_vector;
  function to_std_logic_vector(x : adb3_ocp_m2sT_array) return std_logic_vector;
  function to_std_logic_vector(x : adb3_ocp_s2mT_array) return std_logic_vector;

  function to_adb3_ocp_m2sT(x : std_logic_vector) return adb3_ocp_m2sT;
  function to_adb3_ocp_s2mT(x : std_logic_vector) return adb3_ocp_s2mT;
  function to_adb3_ocp_l_m2sT(x : std_logic_vector) return adb3_ocp_l_m2sT;
  function to_adb3_ocp_l_s2mT(x : std_logic_vector) return adb3_ocp_l_s2mT;
  function to_adb3_ocp_m2sT_array(x : std_logic_vector) return adb3_ocp_m2sT_array;
  function to_adb3_ocp_s2mT_array(x : std_logic_vector) return adb3_ocp_s2mT_array;

  function make_defined_s2m_array(constant x : in adb3_ocp_s2mT_array) return adb3_ocp_s2mT_array;
  function make_defined_s2m(constant x : in adb3_ocp_s2mT) return adb3_ocp_s2mT;
  function make_defined_m2s_array(constant x : in adb3_ocp_m2sT_array) return adb3_ocp_m2sT_array;
  function make_defined_m2s(constant x : in adb3_ocp_m2sT) return adb3_ocp_m2sT;
  function make_defined_slv(constant x : in std_logic_vector) return std_logic_vector;
  function make_defined_sl(constant x : in std_logic; constant z : in std_logic) return std_logic;

end package adb3_ocp;

package body adb3_ocp is

  -- Function to determine if OCP (full/lite) Cmd is write
  function adb3_ocp_cmd_is_wr(cmd : ocp_CmdT) return std_logic is
  begin
    if (to_X01(cmd(0)) = '1') then
      return '1';
    else
      return '0';
    end if;
  end;

  -- Function to determine if OCP (full/lite) Cmd is read
  function adb3_ocp_cmd_is_rd(cmd : ocp_CmdT) return std_logic is
  begin
    if (to_X01(cmd(1)) = '1') then
      return '1';
    else
      return '0';
    end if;
  end;

  -- Function to determine if OCP (full/lite) Cmd is idle
  function adb3_ocp_cmd_is_idle(cmd : ocp_CmdT) return std_logic is
    variable res_cmd : ocp_CmdT;
  begin
    for i in 0 to ADB3_OCP_CMD_WIDTH-1 loop
      res_cmd(i) := to_X01(cmd(i));
    end loop;

    if (res_cmd = OCP_CMD_IDLE) then
      return '1';
    else
      return '0';
    end if;
  end;


  -- Procedure used to write OCP lite Data to register (std_logic_vector) using DataByteEn
  procedure adb3_ocp_l_wr(
    signal reg       : inout  std_logic_vector;
    signal ocp_m2s_l : in     adb3_ocp_l_m2sT;
    constant offset  : in     natural)
  is
  begin
    for i in reg'range loop
      if (ocp_m2s_l.DataByteEn((i - reg'low + offset)/8) = '1') then
        reg(i) <= ocp_m2s_l.Data(i - reg'low + offset);
      end if;
    end loop;
  end procedure;

  -- Procedure used to write OCP lite Data to register (std_logic_vector) using DataByteEn
  -- Merges data according to DataByteEn
  procedure adb3_ocp_l_wr(
    signal reg_out   : out    std_logic_vector;
    signal reg_in    : in     std_logic_vector;
    signal ocp_m2s_l : in     adb3_ocp_l_m2sT;
    constant offset  : in     natural)
  is
  begin
    assert reg_in'low = reg_out'low report "adb3_ocp_l_wr: reg_in'low does not equal reg_out'low" severity error;
    assert reg_in'high = reg_out'high report "adb3_ocp_l_wr: reg_in'high does not equal reg_out'high" severity error;

    for i in reg_out'range loop
      if (ocp_m2s_l.DataByteEn((i - reg_out'low + offset)/8) = '1') then
        reg_out(i) <= ocp_m2s_l.Data(i - reg_out'low + offset);
      else
        reg_out(i) <= reg_in(i + offset);
      end if;
    end loop;
  end procedure;

  -- Procedure used to write OCP lite Data to register (std_logic) using DataByteEn
  procedure adb3_ocp_l_wr(
    signal reg       : inout  std_logic;
    signal ocp_m2s_l : in     adb3_ocp_l_m2sT;
    constant offset  : in     natural)
  is
  begin
    if (ocp_m2s_l.DataByteEn(offset/8) = '1') then
      reg <= ocp_m2s_l.Data(offset);
    end if;
  end procedure;

  -- Procedure used to write OCP lite Data to register (std_logic) using DataByteEn
  -- Merges data according to DataByteEn
  procedure adb3_ocp_l_wr(
    signal reg_out   : out    std_logic;
    signal reg_in    : in     std_logic;
    signal ocp_m2s_l : in     adb3_ocp_l_m2sT;
    constant offset  : in     natural)
  is
  begin
    if (ocp_m2s_l.DataByteEn(offset/8) = '1') then
      reg_out <= ocp_m2s_l.Data(offset);
    else
      reg_out <= reg_in;
    end if;
  end procedure;


  function to_std_logic (bool : boolean) return std_logic is
  begin
    if (bool) then
      return '1';
    else
      return '0';
    end if;
  end;


  -- Function to convert adb3_ocp_m2sT record to std_logic_vector
  function to_std_logic_vector(x : adb3_ocp_m2sT) return std_logic_vector is
    variable y : std_logic_vector(ADB3_OCP_M2ST_SLV_LENGTH-1 downto 0);
  begin
    y(2 downto 0) := x.Cmd;
    y(ADB3_OCP_ADDR_WIDTH-1 + 3 downto 3) := x.Addr;
    y(ADB3_OCP_BURST_WIDTH-1 + ADB3_OCP_ADDR_WIDTH+3 downto ADB3_OCP_ADDR_WIDTH+3) := x.BurstLength;
    y(ADB3_OCP_DATA_WIDTH-1 + ADB3_OCP_BURST_WIDTH+ADB3_OCP_ADDR_WIDTH+3 downto ADB3_OCP_BURST_WIDTH+ADB3_OCP_ADDR_WIDTH+3) := x.Data;
    y(ADB3_OCP_BE_WIDTH-1 + ADB3_OCP_DATA_WIDTH+ADB3_OCP_BURST_WIDTH+ADB3_OCP_ADDR_WIDTH+3 downto ADB3_OCP_DATA_WIDTH+ADB3_OCP_BURST_WIDTH+ADB3_OCP_ADDR_WIDTH+3) := x.DataByteEn;
    y(ADB3_OCP_BE_WIDTH+ADB3_OCP_DATA_WIDTH+ADB3_OCP_BURST_WIDTH+ADB3_OCP_ADDR_WIDTH+3) := x.DataValid;
    y(1+ADB3_OCP_BE_WIDTH+ADB3_OCP_DATA_WIDTH+ADB3_OCP_BURST_WIDTH+ADB3_OCP_ADDR_WIDTH+3) := x.RespAccept;
    y(ADB3_OCP_TAG_WIDTH-1 + 1+1+ADB3_OCP_BE_WIDTH+ADB3_OCP_DATA_WIDTH+ADB3_OCP_BURST_WIDTH+ADB3_OCP_ADDR_WIDTH+3 downto 1+1+ADB3_OCP_BE_WIDTH+ADB3_OCP_DATA_WIDTH+ADB3_OCP_BURST_WIDTH+ADB3_OCP_ADDR_WIDTH+3) := x.Tag;
    return y;
  end;

  -- Function to convert adb3_ocp_s2mT record to std_logic_vector
  function to_std_logic_vector(x : adb3_ocp_s2mT) return std_logic_vector is
    variable y : std_logic_vector(ADB3_OCP_S2MT_SLV_LENGTH-1 downto 0);
  begin
    y(0) := x.CmdAccept;
    y(1) := x.DataAccept;
    y(ADB3_OCP_DATA_WIDTH-1+2 downto 2) := x.Data;
    y(1 + ADB3_OCP_DATA_WIDTH+2 downto ADB3_OCP_DATA_WIDTH+2) := x.Resp;
    y(ADB3_OCP_TAG_WIDTH-1 + 2+ADB3_OCP_DATA_WIDTH+2 downto 2+ADB3_OCP_DATA_WIDTH+2) := x.Tag;
    return y;
  end;


  -- Function to convert adb3_ocp_l_m2sT record to std_logic_vector
  function to_std_logic_vector(x : adb3_ocp_l_m2sT) return std_logic_vector is
    variable y : std_logic_vector(ADB3_OCP_L_M2ST_SLV_LENGTH-1 downto 0);
  begin
    y(2 downto 0) := x.Cmd;
    y(ADB3_OCP_L_ADDR_WIDTH-1 + 3 downto 3) := x.Addr;
    y(ADB3_OCP_L_DATA_WIDTH-1 + ADB3_OCP_L_ADDR_WIDTH+3 downto ADB3_OCP_l_ADDR_WIDTH+3) := x.Data;
    y(ADB3_OCP_L_BE_WIDTH-1 + ADB3_OCP_L_DATA_WIDTH+ADB3_OCP_L_ADDR_WIDTH+3 downto ADB3_OCP_L_DATA_WIDTH+ADB3_OCP_L_ADDR_WIDTH+3) := x.DataByteEn;
    return y;
  end;

  -- Function to convert adb3_ocp_l_s2mT record to std_logic_vector
  function to_std_logic_vector(x : adb3_ocp_l_s2mT) return std_logic_vector is
    variable y : std_logic_vector(ADB3_OCP_L_S2MT_SLV_LENGTH-1 downto 0);
  begin
    y(0) := x.CmdAccept;
    y(ADB3_OCP_L_DATA_WIDTH-1+1 downto 1) := x.Data;
    y(1 + ADB3_OCP_L_DATA_WIDTH+1 downto ADB3_OCP_L_DATA_WIDTH+1) := x.Resp;
    return y;
  end;


  -- Function to convert adb3_ocp_m2sT_array records to std_logic_vector
  function to_std_logic_vector(x : adb3_ocp_m2sT_array) return std_logic_vector is
    variable y : std_logic_vector(ADB3_OCP_M2ST_SLV_LENGTH*x'length-1 downto 0);
  begin
    for i in 0 to x'length-1 loop
      y((i+1)*ADB3_OCP_M2ST_SLV_LENGTH-1 downto i*ADB3_OCP_M2ST_SLV_LENGTH) := to_std_logic_vector(x(i));
    end loop;
    return y;
  end;

  -- Function to convert adb3_ocp_s2mT_array records to std_logic_vector
  function to_std_logic_vector(x : adb3_ocp_s2mT_array) return std_logic_vector is
    variable y : std_logic_vector(ADB3_OCP_S2MT_SLV_LENGTH*x'length-1 downto 0);
  begin
    for i in 0 to x'length-1 loop
      y((i+1)*ADB3_OCP_S2MT_SLV_LENGTH-1 downto i*ADB3_OCP_S2MT_SLV_LENGTH) := to_std_logic_vector(x(i));
    end loop;
    return y;
  end;


  -- Function to convert std_logic_vector to adb3_ocp_m2sT record
  function to_adb3_ocp_m2sT(x : std_logic_vector) return adb3_ocp_m2sT is
    variable y : adb3_ocp_m2sT;
  begin
    y.Cmd         := x(2 downto 0);
    y.Addr        := x(ADB3_OCP_ADDR_WIDTH-1 + 3 downto 3);
    y.BurstLength := x(ADB3_OCP_BURST_WIDTH-1 + ADB3_OCP_ADDR_WIDTH+3 downto ADB3_OCP_ADDR_WIDTH+3);
    y.Data        := x(ADB3_OCP_DATA_WIDTH-1 + ADB3_OCP_BURST_WIDTH+ADB3_OCP_ADDR_WIDTH+3 downto ADB3_OCP_BURST_WIDTH+ADB3_OCP_ADDR_WIDTH+3);
    y.DataByteEn  := x(ADB3_OCP_BE_WIDTH-1 + ADB3_OCP_DATA_WIDTH+ADB3_OCP_BURST_WIDTH+ADB3_OCP_ADDR_WIDTH+3 downto ADB3_OCP_DATA_WIDTH+ADB3_OCP_BURST_WIDTH+ADB3_OCP_ADDR_WIDTH+3);
    y.DataValid   := x(ADB3_OCP_BE_WIDTH+ADB3_OCP_DATA_WIDTH+ADB3_OCP_BURST_WIDTH+ADB3_OCP_ADDR_WIDTH+3);
    y.RespAccept  := x(1+ADB3_OCP_BE_WIDTH+ADB3_OCP_DATA_WIDTH+ADB3_OCP_BURST_WIDTH+ADB3_OCP_ADDR_WIDTH+3);
    y.Tag         := x(ADB3_OCP_TAG_WIDTH-1 + 1+1+ADB3_OCP_BE_WIDTH+ADB3_OCP_DATA_WIDTH+ADB3_OCP_BURST_WIDTH+ADB3_OCP_ADDR_WIDTH+3 downto 1+1+ADB3_OCP_BE_WIDTH+ADB3_OCP_DATA_WIDTH+ADB3_OCP_BURST_WIDTH+ADB3_OCP_ADDR_WIDTH+3);
    return y;
  end;

  -- Function to convert std_logic_vector to adb3_ocp_s2mT record
  function to_adb3_ocp_s2mT(x : std_logic_vector) return adb3_ocp_s2mT is
    variable y : adb3_ocp_s2mT;
  begin
    y.CmdAccept  := x(0);
    y.DataAccept := x(1);
    y.Data       := x(ADB3_OCP_DATA_WIDTH-1+2 downto 2);
    y.Resp       := x(1 + ADB3_OCP_DATA_WIDTH+2 downto ADB3_OCP_DATA_WIDTH+2);
    y.Tag        := x(ADB3_OCP_TAG_WIDTH-1 + 2+ADB3_OCP_DATA_WIDTH+2 downto 2+ADB3_OCP_DATA_WIDTH+2);
    return y;
  end;


    -- Function to convert std_logic_vector to adb3_ocp_l_m2sT record
  function to_adb3_ocp_l_m2sT(x : std_logic_vector) return adb3_ocp_l_m2sT is
    variable y : adb3_ocp_l_m2sT;
  begin
    y.Cmd        := x(2 downto 0);
    y.Addr       := x(ADB3_OCP_L_ADDR_WIDTH-1 + 3 downto 3);
    y.Data       := x(ADB3_OCP_L_DATA_WIDTH-1 +ADB3_OCP_L_ADDR_WIDTH+3 downto ADB3_OCP_L_ADDR_WIDTH+3);
    y.DataByteEn := x(ADB3_OCP_L_BE_WIDTH-1 + ADB3_OCP_L_DATA_WIDTH+ADB3_OCP_L_ADDR_WIDTH+3 downto ADB3_OCP_L_DATA_WIDTH+ADB3_OCP_L_ADDR_WIDTH+3);
    return y;
  end;

  -- Function to convert std_logic_vector to adb3_ocp_l_s2mT record
  function to_adb3_ocp_l_s2mT(x : std_logic_vector) return adb3_ocp_l_s2mT is
    variable y : adb3_ocp_l_s2mT;
  begin
    y.CmdAccept := x(0);
    y.Data      := x(ADB3_OCP_L_DATA_WIDTH-1+1 downto 1);
    y.Resp      := x(1 + ADB3_OCP_L_DATA_WIDTH+1 downto ADB3_OCP_L_DATA_WIDTH+1);
    return y;
  end;


  -- Function to convert std_logic_vector to adb3_ocp_m2sT_array records
  function to_adb3_ocp_m2sT_array(x : std_logic_vector) return adb3_ocp_m2sT_array is
    variable y : adb3_ocp_m2sT_array(0 to (x'length/ADB3_OCP_M2ST_SLV_LENGTH) -1);
    variable x_i : std_logic_vector(ADB3_OCP_M2ST_SLV_LENGTH-1 downto 0);
  begin
    for i in 0 to (x'length/ADB3_OCP_M2ST_SLV_LENGTH)-1 loop
      x_i := x((i+1)*ADB3_OCP_M2ST_SLV_LENGTH-1 downto i*ADB3_OCP_M2ST_SLV_LENGTH);
      y(i) := to_adb3_ocp_m2sT(x_i);
    end loop;
    return y;
  end;

  -- Function to convert std_logic_vector to adb3_ocp_s2mT_array records
  function to_adb3_ocp_s2mT_array(x : std_logic_vector) return adb3_ocp_s2mT_array is
    variable y : adb3_ocp_s2mT_array(0 to (x'length/ADB3_OCP_S2MT_SLV_LENGTH) -1);
    variable x_i : std_logic_vector(ADB3_OCP_S2MT_SLV_LENGTH-1 downto 0);
  begin
    for i in 0 to (x'length/ADB3_OCP_S2MT_SLV_LENGTH)-1 loop
      x_i := x((i+1)*ADB3_OCP_S2MT_SLV_LENGTH-1 downto i*ADB3_OCP_S2MT_SLV_LENGTH);
      y(i) := to_adb3_ocp_s2mT(x_i);
    end loop;  -- i
    return y;
  end;


  -- Make sure adb3_ocp_s2mT record Data/Tag are defined
  function make_defined_s2m(
    constant x : in adb3_ocp_s2mT)
  return adb3_ocp_s2mT is
    variable m : adb3_ocp_s2mT;
  begin
    m.CmdAccept  := x.CmdAccept;
    m.DataAccept := x.DataAccept;
    m.Resp       := x.Resp;
    m.Data       := make_defined_slv(x.Data);
    m.Tag        := make_defined_slv(x.Tag);
    return m;
  end;

  -- Make sure adb3_ocp_s2mT_array records Data/Tag are defined
  function make_defined_s2m_array(
    constant x : in adb3_ocp_s2mT_array)
  return adb3_ocp_s2mT_array is
    variable m : adb3_ocp_s2mT_array(x'range);
  begin
    for i in x'range loop
      m(i) := make_defined_s2m(x(i));
    end loop;
    return m;
  end;


  -- Make sure adb3_ocp_m2sT record Addr/BurstLength/Data/DataByteEn/Tag are defined
  function make_defined_m2s(
    constant x : in adb3_ocp_m2sT)
  return adb3_ocp_m2sT is
    variable m : adb3_ocp_m2sT;
  begin
    m.Cmd         := x.Cmd;
    m.Addr        := make_defined_slv(x.Addr);
    m.BurstLength := make_defined_slv(x.BurstLength);
    m.Data        := make_defined_slv(x.Data);
    m.DataByteEn  := make_defined_slv(x.DataByteEn);
    m.RespAccept  := x.RespAccept;
    m.DataValid   := x.DataValid;
    m.Tag         := make_defined_slv(x.Tag);
    return m;
  end;

  -- Make sure adb3_ocp_m2sT_array records Addr/BurstLength/Data/DataByteEn/Tag are defined
  function make_defined_m2s_array(
    constant x : in adb3_ocp_m2sT_array)
  return adb3_ocp_m2sT_array is
    variable m : adb3_ocp_m2sT_array(x'range);
  begin
    for i in x'range loop
      m(i) := make_defined_m2s(x(i));
    end loop;
    return m;
  end;


  -- Convert any non '0' or '1' bits of std_logic_vector to '0'
  function make_defined_slv(
    constant x : in std_logic_vector)
  return std_logic_vector is
    variable m : std_logic_vector(x'range);
  begin
    for i in x'range loop
      m(i) := make_defined_sl(x(i),'0');
    end loop;
    return m;
  end;

  -- Convert a non '0' or '1' bit of std_logic to defined
  function make_defined_sl(
    constant x : in std_logic;
    constant z : in std_logic)
  return std_logic is
  begin
    if (x = '0')or(x = '1') then
      return x;
    else
      return z;
    end if;
  end;

end package body adb3_ocp;