library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

library work;
use work.adb3_ocp.all;

library unisim;
use unisim.vcomponents.all;

entity simple_ocp is
generic(
  sim_ocp : boolean := false);
port(
  -- Reference clock
  refclk300_p              : in    std_logic;
  refclk300_n              : in    std_logic;
  -- MPTL
  mptl_clk_p               : in    std_logic;
  mptl_clk_n               : in    std_logic;
  mptl_b2t_p               : in    std_logic_vector(3 downto 0);
  mptl_b2t_n               : in    std_logic_vector(3 downto 0);
  mptl_t2b_p               : out   std_logic_vector(3 downto 0);
  mptl_t2b_n               : out   std_logic_vector(3 downto 0);
  -- MPTL initialisation
  mptl_bridge_gtp_online_l : in    std_logic;
  mptl_target_configured_l : out   std_logic;
  mptl_target_gtp_online_l : out   std_logic);
end entity;

architecture admxrcku1 of simple_ocp is

  component mptl_tgt_ku_ocp_wrap
  generic(
    sim_ocp : boolean := false);
  port(
    -- OCP
    ocp_clk                  : in    std_logic;
    direct_slave_m2s         : out   adb3_ocp_m2sT;
    direct_slave_s2m         : in    adb3_ocp_s2mT;
    -- MPTL
    mptl_clk_p               : in    std_logic;
    mptl_clk_n               : in    std_logic;
    mptl_b2t_p               : in    std_logic_vector(3 downto 0);
    mptl_b2t_n               : in    std_logic_vector(3 downto 0);
    mptl_t2b_p               : out   std_logic_vector(3 downto 0);
    mptl_t2b_n               : out   std_logic_vector(3 downto 0);
    -- MPTL initialisation
    ocp_ready                : in    std_logic;
    mptl_ready               : out   std_logic;
    mptl_bridge_gtp_online_l : in    std_logic;
    mptl_target_configured_l : out   std_logic;
    mptl_target_gtp_online_l : out   std_logic;
    -- MPTL sideband
    mptl_gpio_t2b            : in    std_logic_vector(15 downto 0);
    mptl_gpio_b2t            : out   std_logic_vector(15 downto 0));
  end component;

  component adb3_ocp_simple_bus_if
  generic(
    addr_width   : natural range 1 to ADB3_OCP_ADDR_WIDTH;
    data_width   : natural range 8 to ADB3_OCP_DATA_WIDTH;
    read_latency : natural := 1);
  port(
    -- OCP
    ocp_rst : in    std_logic;
    ocp_clk : in    std_logic;
    ocp_m2s : in    adb3_ocp_m2sT;
    ocp_s2m : out   adb3_ocp_s2mT;
    -- Simple bus
    d       : out   std_logic_vector(data_width-1 downto 0);
    q       : in    std_logic_vector(data_width-1 downto 0);
    a       : out   std_logic_vector(addr_width-1 downto 0);
    w       : out   std_logic;
    r       : out   std_logic;
    we      : out   std_logic_vector((data_width/8)-1 downto 0));
  end component;


  ---------------------
  -- Local constants --
  ---------------------
  -- Constants package adb3_ocp
  -- Width of register data must be 8/16/32/64/128 bits
  -- Width of register byte addresses maximum is 22
  constant REG_DATA_WIDTH    : natural range 8 to ADB3_OCP_DATA_WIDTH := 32;
  constant REG_BE_WIDTH      : natural range 1 to ADB3_OCP_BE_WIDTH := 4;
  constant REG_BE_WIDTH_LOG2 : natural range 0 to ADB3_OCP_BE_WIDTH_LOG2 := 2;

  constant REG_ADDR_WIDTH    : natural range 1 to 22 := 3; -- Byte address width


  -----------------------------
  -- mptl_tgt_ku_ocp signals --
  -----------------------------
  signal refclk300_ibuf : std_logic;
  signal usr_clk        : std_logic;
  signal pon_rst_vec    : std_logic_vector(31 downto 0) := (others => '1');
  signal pon_rst        : std_logic;
  signal usr_rst        : std_logic;

  signal mptl_gpio_b2t : std_logic_vector(15 downto 0);
  signal mptl_gpio_t2b : std_logic_vector(15 downto 0) := (others => '0');
  signal ocp_ready     : std_logic;
  signal mptl_ready    : std_logic;

  signal direct_slave_m2s : adb3_ocp_m2sT;
  signal direct_slave_s2m : adb3_ocp_s2mT := ADB3_OCP_DEFAULT_S2M;

  ----------------------------
  -- adb3_ocp_simple_bus_if --
  ----------------------------
  signal ds_q     : std_logic_vector(REG_DATA_WIDTH-1 downto 0);
  signal ds_d     : std_logic_vector(REG_DATA_WIDTH-1 downto 0);
  signal ds_we    : std_logic_vector(REG_BE_WIDTH-1 downto 0);
  signal ds_wr    : std_logic;
  signal ds_rd    : std_logic;
  signal ds_addr  : std_logic_vector(REG_ADDR_WIDTH-1 downto 0); -- Byte address

  ----------------------
  -- Register signals --
  ----------------------
  signal reg_simple_wdata : std_logic_vector(REG_DATA_WIDTH-1 downto 0);
  signal reg_simple_rdata : std_logic_vector(REG_DATA_WIDTH-1 downto 0);

begin

  ----------------------
  -- User clock/reset --
  ----------------------

  -- Use reference clock (300MHz oscillator) as design reference clock
  ref_clk_dif_i : IBUFDS
  port map(
    I  => refclk300_p,
    IB => refclk300_n,
    O  => refclk300_ibuf);

  -- Divide reference clock by 2
  ref_clk_bufg_i : BUFGCE_DIV
  generic map(
    BUFGCE_DIVIDE => 2) -- integer := 1
  port map(
    CE  => '1',            -- in   std_ulogic
    CLR => '0',            -- in   std_ulogic
    I   => refclk300_ibuf, -- in   std_ulogic      
    O   => usr_clk);       -- out  std_ulogic

  -- Genereate reset pulse on power-on
  pon_rst_p : process(usr_clk)
  begin
    if rising_edge(usr_clk) then
      pon_rst_vec <= std_logic_vector(unsigned(pon_rst_vec) srl 1);
    end if;
  end process;

  pon_rst <= pon_rst_vec(0);


  --------------------
  -- MPTL interface --
  --------------------

  mptl_tgt_ku_ocp_wrap_i : mptl_tgt_ku_ocp_wrap
  generic map(
    sim_ocp                  => sim_ocp)
  port map(
    -- OCP
    ocp_clk                  => usr_clk,                  -- in    std_logic
    direct_slave_m2s         => direct_slave_m2s,         -- out   adb3_ocp_m2sT
    direct_slave_s2m         => direct_slave_s2m,         -- in    adb3_ocp_s2mT
    -- MPTL
    mptl_clk_p               => mptl_clk_p,               -- in    std_logic
    mptl_clk_n               => mptl_clk_n,               -- in    std_logic
    mptl_b2t_p               => mptl_b2t_p,               -- in    std_logic_vector(3 downto 0)
    mptl_b2t_n               => mptl_b2t_n,               -- in    std_logic_vector(3 downto 0)
    mptl_t2b_p               => mptl_t2b_p,               -- out   std_logic_vector(3 downto 0)
    mptl_t2b_n               => mptl_t2b_n,               -- out   std_logic_vector(3 downto 0)
    -- MPTL initialisation
    ocp_ready                => ocp_ready,                -- in    std_logic
    mptl_ready               => mptl_ready,               -- out   std_logic
    mptl_bridge_gtp_online_l => mptl_bridge_gtp_online_l, -- in    std_logic
    mptl_target_configured_l => mptl_target_configured_l, -- out   std_logic
    mptl_target_gtp_online_l => mptl_target_gtp_online_l, -- out   std_logic
    -- MPTL sideband
    mptl_gpio_t2b            => mptl_gpio_t2b,            -- in    std_logic_vector(15 downto 0)
    mptl_gpio_b2t            => mptl_gpio_b2t);           -- out   std_logic_vector(15 downto 0)

  -- Hold off MPTL target configured until all OCP clocks are stable
  ocp_ready <= NOT pon_rst;

  -- Keep rest of FPGA design in reset until MPTL interface is online
  usr_rst <= NOT mptl_ready;


  -------------------------------------
  -- Direct Slave parallel interface --
  -------------------------------------

  -- Address aliasing will occur as out of range address bits are not checked
  adb3_ocp_simple_bus_if_i : adb3_ocp_simple_bus_if
  generic map(
    addr_width   => REG_ADDR_WIDTH, -- natural range 1 to ADB3_OCP_ADDR_WIDTH
    data_width   => REG_DATA_WIDTH, -- natural range 8 to ADB3_OCP_DATA_WIDTH
    read_latency => 1)              -- natural := 1)
  port map(
    -- OCP
    ocp_rst => usr_rst,          -- in    std_logic
    ocp_clk => usr_clk,          -- in    std_logic
    ocp_m2s => direct_slave_m2s, -- in    adb3_ocp_m2sT
    ocp_s2m => direct_slave_s2m, -- out   adb3_ocp_s2mT
    -- Simple bus
    d       => ds_d,             -- out   std_logic_vector(data_width-1 downto 0)
    q       => ds_q,             -- in    std_logic_vector(data_width-1 downto 0)
    a       => ds_addr,          -- out   std_logic_vector(addr_width-1 downto 0)
    w       => ds_wr,            -- out   std_logic
    r       => ds_rd,            -- out   std_logic
    we      => ds_we);           -- out   std_logic_vector((data_width/8)-1 downto 0))


  ---------------
  -- Registers --
  ---------------

  -- ds_addr(REG_BE_WIDTH_LOG2-1 downto 0) is unused
  -- ds_we bit is active during OCP write when OCP byte enable is also active
  write_regs : process(usr_rst, usr_clk)
  begin
    if usr_rst = '1' then
      reg_simple_wdata <= (others => '0');
    elsif rising_edge(usr_clk) then
      if ds_addr(REG_ADDR_WIDTH-1 downto REG_BE_WIDTH_LOG2) = "0" then
        for i in 0 to REG_BE_WIDTH-1 loop
          if ds_we(i) = '1' then
            reg_simple_wdata(8*(i+1)-1 downto 8*i) <= ds_d(8*(i+1)-1 downto 8*i);
          end if;
        end loop;
      end if;
    end if;
  end process;

  -- ds_addr(REG_BE_WIDTH_LOG2-1 downto 0) is unused
  -- Read latency is 1 cycle
  read_regs : process(usr_rst, usr_clk)
  begin
    if usr_rst = '1' then
      ds_q <= (others => '0');
    elsif rising_edge(usr_clk) then
      if ds_addr(REG_ADDR_WIDTH-1 downto REG_BE_WIDTH_LOG2) = "0" then
        ds_q <= reg_simple_rdata;
      else
        ds_q <= (others => '-');
      end if;
    end if;
  end process;


  -- Nibble reverse data
  reg_simple_rdata <= reg_simple_wdata(3  downto 0)  &
                      reg_simple_wdata(7  downto 4)  &
                      reg_simple_wdata(11 downto 8)  &
                      reg_simple_wdata(15 downto 12) &
                      reg_simple_wdata(19 downto 16) &
                      reg_simple_wdata(23 downto 20) &
                      reg_simple_wdata(27 downto 24) &
                      reg_simple_wdata(31 downto 28);

end architecture;
