library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_misc.all;
use ieee.numeric_std.all;

library work;
use work.ddr4_sdram_model.all;
use work.ddr4_sdram_model_int.all;

entity ddr4_sdram_chip is
  generic(
    -- String prepended to messages to give some indication where the message comes from (for multi-chip modules)
    model_name          : in    string := "ddr4_sdram_chip";
    -- Selects the DDR4 SDRAM part to be modeled.
    part                : in    part_t;
    -- The value returned (in all DQ bits) of a memory location that has not previously been written or initialized.
    read_undef_val      : in    std_logic := 'U';
    -- Assumes that simulation begins from power on, so that power on timing used; other uses reset timing.
    init_from_power_on  : in    boolean := false;
    -- Demotes warnings about deviations from recommended initialization sequence to informational messages.
    relaxed_init_timing : in    boolean := true;
    -- true  => Use SystemVerilog version of ddr4_sdram_data module to work around 2016.3 XSIM access type issue.
    -- false => Use VHDL version of ddr4_sdram_data module.
    xsim_workaround     : in    boolean := true
  );
  port(
    reset_n             : in    std_logic;
    ck_t                : in    std_logic;
    ck_c                : in    std_logic;
    ten                 : in    std_logic := '0';
    
    cke                 : in    std_logic;
    cs_n                : in    std_logic;
    odt                 : in    std_logic;
    -- C0 functions as CKE1 if traditional twin-die part.
    -- C0 used to select die if stacked-die part.
    -- C0 ignored if neither twin-die nor stacked-die part.
    c0_cke1             : in    std_logic := 'U';
    -- C1 functions as CS1_n if traditional twin-die part.
    -- C1 used to select die if stacked-die part.
    -- C1 ignored if neither twin-die nor stacked-die part.
    c1_cs1_n            : in    std_logic := 'U';
    -- C2 functions as ODT1 if traditional twin-die part.
    -- C2 used to select die if stacked-die part.
    -- C2 ignored if neither twin-die nor stacked-die part.
    c2_odt1             : in    std_logic := 'U';
    act_n               : in    std_logic;
    -- A16 functions as RAS_n when ACT_n is high.
    -- A15 functions as CAS_n when ACT_n is high.
    -- A14 functions as WE_n when ACT_n is high.
    a                   : in    std_logic_vector(part.pinout.a_width - 1 downto 0);
    ba                  : in    std_logic_vector(part.pinout.ba_width - 1 downto 0);
    bg                  : in    std_logic_vector(part.pinout.bg_width - 1 downto 0);
    par                 : in    std_logic;
    
    -- x4: DM_DBI_n ignored.
    -- x8 & x16: DBI on same pins as DM and enabled via MR5.
    -- x8 only: DM_DBI_n[0] & TDQS_t are on same pin; TDQS function enabled via MR1 (takes precedence over DM & DBI functions).
    dm_dbi_n            : inout std_logic_vector(part.pinout.dm_dbi_width - 1 downto 0) := (others => 'U');
    -- x4 & x16: TDQS_c ignored.
    -- x8 only: TDQS_c function enabled via MR1; connect if used.
    tdqs_c              : in    std_logic := 'U';
    dq                  : inout std_logic_vector(4*part.pinout.dq_nibbles - 1 downto 0);
    dqs_t               : inout std_logic_vector(part.pinout.dqs_width - 1 downto 0);
    dqs_c               : inout std_logic_vector(part.pinout.dqs_width - 1 downto 0);
    
    alert_n             : out   std_logic
  );
end entity;

architecture behav of ddr4_sdram_chip is

  
begin

  -- Package with a single die.
  gen_single_die : if part.pinout.stack_type = stack_none generate
    
    U0 : ddr4_sdram_die
      generic map(
        model_name => model_name,
        part => part,
        read_undef_val => read_undef_val,
        init_from_power_on => init_from_power_on,
        relaxed_init_timing => relaxed_init_timing,
        xsim_workaround => xsim_workaround
      )
      port map(
        reset_n => reset_n,
        ck_t => ck_t,
        ck_c => ck_c,
        ten => ten,
        cke => cke,
        cs_n => cs_n,
        odt => odt,
        act_n => act_n,
        a => a,
        ba => ba,
        bg => bg,
        par => par,
        dm_dbi_n => dm_dbi_n,
        tdqs_c => tdqs_c,
        dq => dq,
        dqs_t => dqs_t,
        dqs_c => dqs_c,
        alert_n => alert_n
      );
        
        
  end generate;
  
  
  -- Traditional twin die package.
  -- For 2nd die, C0 functions as CKE, C1 functions as CS_n and C2 functions as ODT.
  gen_twin_die : if part.pinout.stack_type = stack_twin_die generate
  
    -- Internal ALERT_n for each die.
    signal alert0_n, alert1_n : std_logic;

    
  begin
  
    U0 : ddr4_sdram_die
      generic map(
        -- Generate a model name for the die as "<model_name>_die0".
        model_name => model_name & "_die0",
        part => part,
        read_undef_val => read_undef_val,
        init_from_power_on => init_from_power_on,
        relaxed_init_timing => relaxed_init_timing,
        xsim_workaround => xsim_workaround
      )
      port map(
        reset_n => reset_n,
        ck_t => ck_t,
        ck_c => ck_c,
        ten => ten,
        cke => cke,
        cs_n => cs_n,
        odt => odt,
        act_n => act_n,
        a => a,
        ba => ba,
        bg => bg,
        par => par,
        dm_dbi_n => dm_dbi_n,
        tdqs_c => tdqs_c,
        dq => dq,
        dqs_t => dqs_t,
        dqs_c => dqs_c,
        alert_n => alert0_n
      );
        
    U1 : ddr4_sdram_die
      generic map(
        -- Generate a model name for the die as "<model_name>_die1".
        model_name => model_name & "_die1",
        part => part,
        read_undef_val => read_undef_val,
        init_from_power_on => init_from_power_on,
        relaxed_init_timing => relaxed_init_timing,
        xsim_workaround => xsim_workaround
      )
      port map(
        reset_n => reset_n,
        ck_t => ck_t,
        ck_c => ck_c,
        ten => ten,
        cke => c0_cke1,
        cs_n => c1_cs1_n,
        odt => c2_odt1,
        act_n => act_n,
        a => a,
        ba => ba,
        bg => bg,
        par => par,
        dm_dbi_n => dm_dbi_n,
        tdqs_c => tdqs_c,
        dq => dq,
        dqs_t => dqs_t,
        dqs_c => dqs_c,
        alert_n => alert1_n
      );
    
    -- Combine alert_n from the dies
    alert_n <= alert0_n AND alert1_n;
    
    
  end generate;
  

  -- Stacked-die package (multiple dies, using C[2:0] for addressing individual dies).
  gen_stacked_die : if part.pinout.stack_type = stack_stacked generate

    constant num_die : natural := part.pinout.stack_size;
    
    -- Internal ALERT_n for each die.
    signal alert_int_n : std_logic_vector(num_die - 1 downto 0);

    
  begin
  
    gen_dies : for die in 0 to num_die - 1 generate
      
      -- "Die address" is C[2:0] value required for addressing this die
      constant die_addr : std_logic_vector(2 downto 0) := std_logic_vector(to_unsigned(die, 3));
      
      -- Internal CS_n for this die.
      signal cs_int_n : std_logic;
    
    
    begin
    
      -- Do CS_n generation this way instead of arith comparison so that simulator doesn't complain about Xs.
      -- Also causes X on C[2:0] to propagate X into per-die models so that they can complain it.
      cs_int_n <= (die_addr(0) xor c0_cke1) and (die_addr(1) xor c1_cs1_n) and (die_addr(2) xor c2_odt1);
      
      U0 : ddr4_sdram_die
        generic map(
          -- Generate a model name for the die as "<model_name>_die<N>".
          model_name => model_name & "_die" & natural'image(die),
          part => part,
          read_undef_val => read_undef_val,
          init_from_power_on => init_from_power_on,
          relaxed_init_timing => relaxed_init_timing,
          xsim_workaround => xsim_workaround
        )
        port map(
          reset_n => reset_n,
          ck_t => ck_t,
          ck_c => ck_c,
          ten => ten,
          cke => cke,
          cs_n => cs_int_n,
          odt => odt,
          act_n => act_n,
          a => a,
          ba => ba,
          bg => bg,
          par => par,
          dm_dbi_n => dm_dbi_n,
          tdqs_c => tdqs_c,
          dq => dq,
          dqs_t => dqs_t,
          dqs_c => dqs_c,
          alert_n => alert_int_n(die)
        );

    end generate;        
    
    -- Combine alert_n from the dies
    alert_n <= AND_reduce(alert_int_n);
    
    
  end generate;
  
  
end architecture;
