library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_misc.all;
use ieee.numeric_std.all;

library std;
use std.textio.all;

library work;
use work.ddr4_sdram_model.all;
use work.ddr4_sdram_model_int.all;

entity ddr4_sdram_data is
  generic(
    model_name          : in    string;
    part                : in    part_t;
    num_bank_total      : in    natural;
    read_undef_val      : in    std_logic := 'U'
  );
  port(
    reset_n             : in    std_logic; -- Should be _X01 version of signal
    ck_t                : in    std_logic;
    cke                 : in    std_logic; -- Should be _X01 version of signal
    cs_n                : in    std_logic; -- Should be _X01 version of signal
    act_n               : in    std_logic; -- Should be _X01 version of signal
    a_col               : in    std_logic_vector(part.pinout.col_width - 1 downto 0); -- Should be _01 version of signal
    a_ap                : in    std_logic; -- Should be _X01 version of signal
    a_bc                : in    std_logic; -- Should be _X01 version of signal
    ba                  : in    std_logic_vector(part.pinout.ba_width - 1 downto 0); -- Should be _01 version of signal
    bg                  : in    std_logic_vector(part.pinout.bg_width - 1 downto 0); -- Should be _01 version of signal
    ras_n               : in    std_logic; -- Should be _X01 version of signal
    cas_n               : in    std_logic; -- Should be _X01 version of signal
    we_n                : in    std_logic; -- Should be _X01 version of signal
    dq                  : inout std_logic_vector(4 * part.pinout.dq_nibbles - 1 downto 0);
    dm_dbi_n            : inout std_logic_vector(part.pinout.dm_dbi_width - 1 downto 0);
    dqs_t               : inout std_logic_vector(part.pinout.dqs_width - 1 downto 0);
    dqs_c               : inout std_logic_vector(part.pinout.dqs_width - 1 downto 0);
    bank_states         : in    bank_state_array_t;
    activated_rows      : in    bank_row_array_t;
    cl                  : in    cl_t;
    al                  : in    al_t;
    cwl                 : in    cwl_t;
    bl                  : in    burst_length_t;
    bt                  : in    burst_type_t;
    read_preamble       : in    read_preamble_t;
    rd_dbi_enabled      : in    boolean;
    wr_dbi_enabled      : in    boolean
  );
end entity;

architecture behav of ddr4_sdram_data is

  -- Min/max read or write latency
  constant min_rwl : natural := work.ddr4_sdram_model_int.min(min_rl, min_wl);
  constant max_rwl : natural := work.ddr4_sdram_model_int.max(max_rl, max_wl);

  -- Convert starting column address plus beat # (0..3 or 0..7) into a word address for addressing the column's RAM array
  function gen_rd_burst_addr(
    column : in    natural;
    beat   : in    natural;
    bl     : in    actual_burst_length_t;
    bt     : in    burst_type_t
  ) return natural is
    variable base, offset : natural;
    subtype offset_t is natural range 0 to 7;
    type natural_2d_array_t is array(natural range <>, natural range <>) of offset_t;
    constant bl4_seq_offsets : natural_2d_array_t(0 to 3, 0 to 3) := (
      (0, 1, 2, 3),
      (1, 2, 3, 0),
      (2, 3, 0, 1),
      (3, 0, 1, 2)
    );
    constant bl4_ilv_offsets : natural_2d_array_t(0 to 3, 0 to 3) := (
      (0, 1, 2, 3),
      (1, 0, 3, 2),
      (2, 3, 0, 1),
      (3, 2, 1, 0)
    );
    constant bl8_seq_offsets : natural_2d_array_t(0 to 7, 0 to 7) := (
      (0, 1, 2, 3, 4, 5, 6, 7),
      (1, 2, 3, 0, 5, 6, 7, 4),
      (2, 3, 0, 1, 6, 7, 4, 5),
      (3, 0, 1, 2, 7, 4, 5, 6),
      (4, 5, 6, 7, 0, 1, 2, 3),
      (5, 6, 7, 4, 1, 2, 3, 0),
      (6, 7, 4, 5, 2, 3, 0, 1),
      (7, 4, 5, 6, 3, 0, 1, 2)
    );
    constant bl8_ilv_offsets : natural_2d_array_t(0 to 7, 0 to 7) := (
      (0, 1, 2, 3, 4, 5, 6, 7),
      (1, 0, 3, 2, 5, 4, 7, 6),
      (2, 3, 0, 1, 6, 7, 4, 5),
      (3, 2, 1, 0, 7, 6, 5, 4),
      (4, 5, 6, 7, 0, 1, 2, 3),
      (5, 4, 7, 6, 1, 0, 3, 2),
      (6, 7, 4, 5, 2, 3, 0, 1),
      (7, 6, 5, 4, 3, 2, 1, 0)
    );
  begin
    if bl = actual_burst_length_8 then
      offset := column mod 8;
      base := column - offset;
      if bt = burst_type_nibble_seq then
        return base + bl8_seq_offsets(offset, beat);
      else
        return base + bl8_ilv_offsets(offset, beat);
      end if;
    else
      assert beat < 4 severity failure;
      offset := column mod 4;
      base := column - offset;
      if bt = burst_type_nibble_seq then
        return base + bl4_seq_offsets(offset, beat);
      else
        return base + bl4_ilv_offsets(offset, beat);
      end if;
    end if;
  end;
  
  -- Convert starting column address plus beat # (0..3 or 0..7) into a word address for addressing the column's RAM array
  function gen_wr_burst_addr(
    column : in    natural;
    beat   : in    natural;
    bl     : in    actual_burst_length_t
  ) return natural is
    variable base, offset : natural;
  begin
    if bl = actual_burst_length_8 then
      assert beat < 8 severity failure;
      offset := column mod 8;
    else
      assert beat < 4 severity failure;
      offset := column mod 4;
    end if;
    base := column - offset;
    return base + beat;
  end;
  
  type pipe_element_t is record
    valid      : boolean;
    is_write   : boolean;
    bank_index : natural;
    row        : natural;
    column     : natural;
    ap         : boolean;
    bl         : actual_burst_length_t;
    bt         : burst_type_t;
  end record;
  constant pipe_element_invalid : pipe_element_t := (
    valid => false,
    is_write => false,
    bank_index => 0,
    row => 0,
    column => 0,
    ap => false,
    bl => actual_burst_length_8,
    bt => burst_type_nibble_seq
  );
  type pipe_t is array(0 to max_rwl) of pipe_element_t;
  
  signal rl, wl : natural;
  signal otf_bl, actual_bl : actual_burst_length_t;
  signal bank_index : natural;
  signal cmd, write_cmd, autoprecharge : boolean;
  signal pipe : pipe_t;
  signal read_pipe_sel, write_pipe_sel : pipe_element_t;
  
  type burst_info_t is record
    valid      : boolean;
    bank_index : natural;
    row        : natural;
    column     : natural;
    bl         : actual_burst_length_t;
    bt         : burst_type_t;
    pos        : natural;
  end record;
  
  constant burst_info_invalid : burst_info_t := (
    valid => false,
    bank_index => 0,
    row => 0,
    column => 0,
    bl => actual_burst_length_8,
    bt => burst_type_nibble_seq,
    pos => 0);

  subtype memory_location_t is std_logic_vector(4 * part.pinout.dq_nibbles - 1 downto 0);
  subtype memory_dbi_t is std_logic_vector(part.pinout.dqs_width - 1 downto 0);
  
  signal read_burst, read_burst_f, read_burst_r : burst_info_t := burst_info_invalid;
  signal drive_dq : boolean := false;
  signal dq_out : memory_location_t;
  signal drive_dm_dbi : boolean := false;
  signal dm_dbi_out : memory_dbi_t;
  
  signal drive_dqs_pipe : boolean_array_t(0 to 5) := (others => false);
  signal drive_dqs, drive_dqs_r, drive_dqs_f : boolean := false;
  signal dqs_out_t : std_logic;
  signal generate_dqs_r, generate_dqs_f : boolean;
  
  signal write_burst, write_burst_f, write_burst_r : burst_info_t := burst_info_invalid;

  constant num_bank_per_group : natural := 2**part.pinout.ba_width;
  
  constant column_size : natural := 2**part.pinout.col_width;
  type memory_column_t is array(0 to column_size - 1) of memory_location_t;
  constant memory_column_undef : memory_column_t := (others => (others => read_undef_val));
  type memory_column_ptr_t is access memory_column_t;
  
  constant row_count : natural := 2**part.pinout.row_width;
  type memory_rows_t is array(0 to row_count - 1) of memory_column_ptr_t;
  type memory_rows_ptr_t is access memory_rows_t;
  
  type memory_banks_t is array(0 to num_bank_total - 1) of memory_rows_ptr_t;
  shared variable memory_array : memory_banks_t := (others => null);
  
  signal ignore_dm : boolean;

  
begin

  ignore_dm <= not part.pinout.dm_supported or wr_dbi_enabled; -- TO DO: also take into account TDQS enable in mode regs.
  
  cmd <= cke = '1' and cs_n = '0' and act_n = '1' and ras_n = '1' and cas_n = '0';
  write_cmd <= we_n = '0';
  autoprecharge <= a_ap = '1';
  
  otf_bl <= actual_burst_length_4 when a_bc = '1' else actual_burst_length_8;
  actual_bl <= actual_burst_length_8 when bl = burst_length_bl8 else
               actual_burst_length_4 when bl = burst_length_bc4 else
               otf_bl;
  
  bank_index <= to_integer(unsigned(bg)) * num_bank_per_group + to_integer(unsigned(ba));
  
  pipe(0) <= (
    valid => cmd,
    is_write => write_cmd,
    bank_index => bank_index,
    row => activated_rows(bank_index),
    column => to_integer(unsigned(a_col)),
    ap => autoprecharge,
    bl => actual_bl,
    bt => bt
  );
  
  pipeline_cmd : process(ck_t, reset_n)
  begin
    if reset_n = '0' then
      pipe(1 to pipe'high) <= (others => pipe_element_invalid);
    elsif rising_edge(ck_t) then
      pipe(1 to pipe'high) <= pipe(0 to pipe'high - 1);
    end if;
  end process;

  
  --
  -- Drive DQ on reads
  --

  -- Read latency is CL + AL, but subtract 1 because read_burst_r is delayed by 1 ck.
  rl <= cl + al - 1;
  read_pipe_sel <= pipe(rl);
  
  init_read_bursts : process(ck_t, reset_n)
  begin
    if reset_n = '0' then
      read_burst <= burst_info_invalid;
      read_burst_r <= burst_info_invalid;
    elsif rising_edge(ck_t) then
      if read_pipe_sel.valid and not read_pipe_sel.is_write then
        read_burst.valid <= true;
        read_burst.bank_index <= read_pipe_sel.bank_index;
        read_burst.row <= read_pipe_sel.row;
        read_burst.column <= read_pipe_sel.column;
        read_burst.bl <= read_pipe_sel.bl;
        read_burst.bt <= read_pipe_sel.bt;
        read_burst.pos <= 0;
      else
        if read_burst.valid then
          if read_burst.bl = actual_burst_length_4 then
            if read_burst.pos = 2 then
              read_burst.valid <= false;
            end if;
          else
            if read_burst.pos = 6 then
              read_burst.valid <= false;
            end if;
          end if;
          read_burst.pos <= read_burst.pos + 2;
        end if;
      end if;
      
      read_burst_r <= read_burst_f;
    end if;
    
    if reset_n = '0' then
      read_burst_f <= burst_info_invalid;
    elsif falling_edge(ck_t) then
      read_burst_f <= read_burst;
    end if;
  end process;
  
  generate_dq_out : process(reset_n, ck_t)
    variable col_addr : natural;
    variable rows_ptr : memory_rows_ptr_t;
    variable column_ptr : memory_column_ptr_t;
    variable dq_tmp : memory_location_t;
    variable dm_dbi_tmp : memory_dbi_t;
  begin
    if reset_n = '0' then
      dq_out <= (others => read_undef_val);
      drive_dq <= false;
      dm_dbi_out <= (others => read_undef_val);
      drive_dm_dbi <= false;
    else
      if rising_edge(ck_t) then
        if read_burst_f.valid then
          col_addr := gen_rd_burst_addr(read_burst_f.column, read_burst_f.pos, read_burst_f.bl, read_burst_f.bt);
          rows_ptr := memory_array(read_burst_f.bank_index);
          if null = rows_ptr then
            column_ptr := null;
          else
            column_ptr := rows_ptr.all(read_burst_f.row);
          end if;
          if null = column_ptr then
            dq_tmp := (others => read_undef_val);
            dm_dbi_tmp := (others => read_undef_val);
          else
            dq_tmp := column_ptr.all(col_addr);
            if rd_dbi_enabled then
              dm_dbi_tmp := compute_dbi(dq_tmp);
              for lane in 0 to part.pinout.dqs_width - 1 loop
                if dm_dbi_tmp(lane) = '0' then
                  dq_tmp(8 * (lane + 1) - 1 downto 8 * lane) := not dq_tmp(8 * (lane + 1) - 1 downto 8 * lane);
                end if;
              end loop;
            end if;
          end if;
          --report "read ck_t rise: col_addr=" & natural'image(col_addr) & " dq=" & to_string_binary(dq_tmp) & " dbi=" & to_string_binary(dm_dbi_tmp);
          dq_out <= dq_tmp;
          drive_dq <= true;
          dm_dbi_out <= dm_dbi_tmp;
          drive_dm_dbi <= rd_dbi_enabled;
        else
          drive_dq <= false;
          drive_dm_dbi <= false;
        end if;
      end if;
    
      if falling_edge(ck_t) then
        if read_burst_r.valid then
          col_addr := gen_rd_burst_addr(read_burst_r.column, read_burst_r.pos + 1, read_burst_r.bl, read_burst_r.bt);
          rows_ptr := memory_array(read_burst_r.bank_index);
          if null = rows_ptr then
            column_ptr := null;
          else
            column_ptr := rows_ptr.all(read_burst_r.row);
          end if;
          if null = column_ptr then
            dq_tmp := (others => read_undef_val);
            dm_dbi_tmp := (others => read_undef_val);
          else
            dq_tmp := column_ptr.all(col_addr);
            if rd_dbi_enabled then
              dm_dbi_tmp := compute_dbi(dq_tmp);
              for lane in 0 to part.pinout.dqs_width - 1 loop
                if dm_dbi_tmp(lane) = '0' then
                  dq_tmp(8 * (lane + 1) - 1 downto 8 * lane) := not dq_tmp(8 * (lane + 1) - 1 downto 8 * lane);
                end if;
              end loop;
            end if;
          end if;
          --report "read ck_t fall: col_addr=" & natural'image(col_addr) & " dq=" & to_string_binary(dq_tmp) & " dbi=" & to_string_binary(dm_dbi_tmp);
          dq_out <= dq_tmp;
          dm_dbi_out <= dm_dbi_tmp;
        end if;
      end if;
    end if;
  end process;
    
  dq <= dq_out when drive_dq else (others => 'Z');
  dm_dbi_n <= dm_dbi_out when drive_dm_dbi else (others => 'Z');
  

  --
  -- Drive DQS on reads
  --

  -- Generate DQS for read preamble and when driving DQ
  generate_drive_dqs : process(reset_n, ck_t)
  begin
    if reset_n = '0' then
      drive_dqs_pipe <= (others => false);
      drive_dqs_r <= false;
      drive_dqs_f <= false;
      dqs_out_t <= '1';
      generate_dqs_r <= false;
      generate_dqs_f <= false;
    else
      if rising_edge(ck_t) then
        drive_dqs_pipe(0) <= pipe(rl - 2).valid and not pipe(rl - 2).is_write;
        drive_dqs_pipe(1 to 5) <= drive_dqs_pipe(0 to 4);
        
        if read_preamble = read_preamble_2ck then
          drive_dqs_r <= or_reduce(drive_dqs_pipe(0 to 5));
        else
          drive_dqs_r <= or_reduce(drive_dqs_pipe(1 to 5));
        end if;
        
        generate_dqs_r <= or_reduce(drive_dqs_pipe(1 to 5));
        
        if generate_dqs_f then
          dqs_out_t <= '1';
        end if;
      end if;
      
      if falling_edge(ck_t) then
        drive_dqs_f <= drive_dqs_r;
        
        generate_dqs_f <= generate_dqs_r;
        
        if generate_dqs_r then
          dqs_out_t <= '0';
        end if;
      end if;
    end if;
  end process;

  drive_dqs <= drive_dqs_r or drive_dqs_f;
  dqs_t <= (others => dqs_out_t) when drive_dqs else (others => 'Z'); -- TO DO
  dqs_c <= (others => not dqs_out_t) when drive_dqs else (others => 'Z'); -- TO DO
  
  
  --
  -- Writes
  --
  
  -- Write latency is CWL + AL, but subtract 1 because write_burst_r is delayed by 1 ck.
  wl <= cwl + al - 1;
  write_pipe_sel <= pipe(wl);
  
  init_write_bursts : process(ck_t, reset_n)
  begin
    if reset_n = '0' then
      write_burst <= burst_info_invalid;
      write_burst_r <= burst_info_invalid;
    elsif rising_edge(ck_t) then
      if write_pipe_sel.valid and write_pipe_sel.is_write then
        write_burst.valid <= true;
        write_burst.bank_index <= write_pipe_sel.bank_index;
        write_burst.row <= write_pipe_sel.row;
        write_burst.column <= write_pipe_sel.column;
        write_burst.bl <= write_pipe_sel.bl;
        write_burst.bt <= write_pipe_sel.bt;
        write_burst.pos <= 0;
      else
        if write_burst.valid then
          if write_burst.bl = actual_burst_length_4 then
            if write_burst.pos = 2 then
              write_burst.valid <= false;
            end if;
          else
            if write_burst.pos = 6 then
              write_burst.valid <= false;
            end if;
          end if;
          write_burst.pos <= write_burst.pos + 2;
        end if;
      end if;
      
      write_burst_r <= write_burst_f;
    end if;
    
    if reset_n = '0' then
      write_burst_f <= burst_info_invalid;
    elsif falling_edge(ck_t) then
      write_burst_f <= write_burst;
    end if;
  end process;
  
  write_data : for lane in 0 to part.pinout.dqs_width - 1 generate
  
    dqs_rise : process(dqs_t(lane))
      variable col_addr : natural;
      variable rows_ptr : memory_rows_ptr_t;
      variable column_ptr : memory_column_ptr_t;
      variable dm_dbi_10 : std_logic;
    begin
      if rising_edge(dqs_t(lane)) then
        if write_burst_f.valid then
          col_addr := gen_wr_burst_addr(write_burst_f.column, write_burst_f.pos, write_burst_f.bl);
          rows_ptr := memory_array(write_burst_f.bank_index);
          if null = rows_ptr then
            report "write dqs_t rise: allocating rows for bank " & natural'image(write_burst_f.bank_index);
            rows_ptr := new memory_rows_t;
            memory_array(write_burst_f.bank_index) := rows_ptr;
          end if;
          column_ptr := rows_ptr.all(write_burst_f.row);
          if null = column_ptr then
            column_ptr := new memory_column_t;
            column_ptr.all := memory_column_undef;
            rows_ptr.all(write_burst_f.row) := column_ptr;
          end if;
          --report "write dqs_t rise: col_addr=" & natural'image(col_addr) & " dq=" & to_string_binary(dq) & " dm_dbi_n=" & to_string_binary(dm_dbi_n);
          dm_dbi_10 := make_10(dm_dbi_n(lane));
          if ignore_dm or dm_dbi_10 = '1' then
            if part.pinout.dqs_width = 1 then
              if wr_dbi_enabled and dm_dbi_10 = '0' then
                column_ptr.all(col_addr) := not make_X01(dq);
              else
                column_ptr.all(col_addr) := make_X01(dq);
              end if;
            else
              if wr_dbi_enabled and dm_dbi_10 = '0' then
                column_ptr.all(col_addr)(8 * (lane + 1) - 1 downto 8 * lane) := not make_X01(dq(8 * (lane + 1) - 1 downto 8 * lane));
              else
                column_ptr.all(col_addr)(8 * (lane + 1) - 1 downto 8 * lane) := make_X01(dq(8 * (lane + 1) - 1 downto 8 * lane));
              end if;
            end if;
          end if;
        end if;
      end if;
    end process;
    
    dqs_fall : process(dqs_t(lane))
      variable col_addr : natural;
      variable rows_ptr : memory_rows_ptr_t;
      variable column_ptr : memory_column_ptr_t;
      variable dm_dbi_10 : std_logic;
    begin
      if falling_edge(dqs_t(lane)) then
        if write_burst_r.valid then
          col_addr := gen_wr_burst_addr(write_burst_r.column, write_burst_r.pos + 1, write_burst_r.bl);
          rows_ptr := memory_array(write_burst_r.bank_index);
          if null = rows_ptr then
            report "write dqs_t fall: allocating rows for bank " & natural'image(write_burst_r.bank_index);
            rows_ptr := new memory_rows_t;
            memory_array(write_burst_r.bank_index) := rows_ptr;
          end if;
          column_ptr := rows_ptr.all(write_burst_r.row);
          if null = column_ptr then
            column_ptr := new memory_column_t;
            column_ptr.all := memory_column_undef;
            rows_ptr.all(write_burst_r.row) := column_ptr;
          end if;
          --report "write dqs_t fall: col_addr=" & natural'image(col_addr) & " dq=" & to_string_binary(dq) & " dm_dbi_n=" & to_string_binary(dm_dbi_n);
          dm_dbi_10 := make_10(dm_dbi_n(lane));
          if ignore_dm or dm_dbi_10 = '1' then
            if part.pinout.dqs_width = 1 then
              if wr_dbi_enabled and dm_dbi_10 = '0' then
                column_ptr.all(col_addr) := not make_X01(dq);
              else
                column_ptr.all(col_addr) := make_X01(dq);
              end if;
            else
              if wr_dbi_enabled and dm_dbi_10 = '0' then
                column_ptr.all(col_addr)(8 * (lane + 1) - 1 downto 8 * lane) := not make_X01(dq(8 * (lane + 1) - 1 downto 8 * lane));
              else
                column_ptr.all(col_addr)(8 * (lane + 1) - 1 downto 8 * lane) := make_X01(dq(8 * (lane + 1) - 1 downto 8 * lane));
              end if;
            end if;
          end if;
        end if;
      end if;
    end process;
    
  end generate;

  
end architecture;
