library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_misc.all;
use ieee.numeric_std.all;

library std;
use std.textio.all;

library work;
use work.ddr4_sdram_model.all;
use work.ddr4_sdram_model_int.all;

entity ddr4_sdram_power_state is
  generic(
    model_name          : in    string;
    part                : in    part_t
  );
  port(
    reset_n             : in    std_logic; -- Should be _X01 version of signal
    ck_t                : in    std_logic;
    cke                 : in    std_logic; -- Should be _X01 version of signal
    cke_prev            : in    std_logic; -- Should be _X01 version of signal
    cs_n                : in    std_logic; -- Should be _X01 version of signal
    act_n               : in    std_logic; -- Should be _X01 version of signal
    ras_n               : in    std_logic; -- Should be _X01 version of signal
    cas_n               : in    std_logic; -- Should be _X01 version of signal
    we_n                : in    std_logic; -- Should be _X01 version of signal
    power_state         : out   power_state_t
  );
end entity;

architecture behav of ddr4_sdram_power_state is

  signal i_power_state : power_state_t := power_state_powerdown;


begin

  power_state <= i_power_state;
  
  --
  -- Keep track of power state.
  -- Model enters power-down state via reset_n or PDE command; exits power-down via CKE high (DES command required).
  -- Enters self-refresh state via SRE command; exits self-refresh via CKE high (DES or NOP command required).
  --
  
  gen_power_state : process(reset_n, ck_t)
  begin
    if reset_n = '0' then
      i_power_state <= power_state_powerdown;
      if reset_n'event then
        report model_name & ": Forced into power-down state by RESET_n" severity note;
      end if;
    elsif rising_edge(ck_t) then
      case i_power_state is
        when power_state_powerdown =>
          if cke = '1' then
            i_power_state <= power_state_normal;
            report model_name & ": Exiting power-down state" severity note;
            if cs_n = '0' then
              report model_name & ": At power-down exit, only DES command permitted; current command is " &
                     classify_command(cke_prev, cke, cs_n, act_n, ras_n, cas_n, we_n)
                severity warning;
            end if;
          end if;
          
        when power_state_normal =>
          if cke = '0' then
            if cs_n = '1' then
              i_power_state <= power_state_powerdown;
              report model_name & ": Entering power-down state" severity note;
            else
              if act_n = '1' and ras_n = '0' and cas_n = '0' and we_n = '1' then
                i_power_state <= power_state_self_refresh;
                report model_name & ": Entering self-refresh state" severity note;
              else
                i_power_state <= power_state_powerdown;
                report model_name & ": Illegal CKE low command; treating as PDE. Expected PDE or SRE but current command is: " &
                       classify_command(cke_prev, cke, cs_n, act_n, ras_n, cas_n, we_n)
                  severity warning;
              end if;
            end if;
          end if;
          
        when power_state_self_refresh =>
          if cke = '1' then
            i_power_state <= power_state_normal;
            report model_name & ": Exiting self-refresh state" severity note;
            if not (cs_n = '1' or (cs_n = '0' and act_n = '1' and ras_n = '1' and cas_n = '1' and we_n = '1'))  then
              report model_name & ": At self-refresh exit, only NOP or DES command permitted; current command is " &
                     classify_command(cke_prev, cke, cs_n, act_n, ras_n, cas_n, we_n)
                severity warning;
            end if;
          end if;
      end case;
    end if;
  end process;
  
  
  --
  -- CKE must be stable for 10 ns and low at rising edge of reset_n.
  --

  check_cke_to_reset_setup : process(reset_n)
  begin
    if rising_edge(reset_n) then
      if not (cke'stable(part.timing.init.t_reset_cke) and cke = '0') then
        report model_name & ": CKE must be low and stable for " & time'image(part.timing.init.t_reset_cke) & " at RESET_n deassertion" severity warning;
      end if;
    end if;
  end process;
  
  
end architecture;
