library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_misc.all;
use ieee.numeric_std.all;

entity memory_exerciser_error_log is
  generic(
    phase_width        : in    natural;
    memory_data_width  : in    natural;
    memory_addr_width  : in    natural;
    count_width        : in    natural;
    index_width        : in    natural
  );
  port(                        
    rst                : in    std_logic;
    clk                : in    std_logic;
    -- Clears the error log
    sr                 : in    std_logic;
    -- Set initial address & phase for current phase
    init               : in    std_logic;
    init_phase         : in    std_logic_vector(phase_width - 1 downto 0);
    init_offset        : in    std_logic_vector(memory_addr_width - 1 downto 0);
    -- Expected and actual data
    chk_ce             : in    std_logic;
    chk_act_data       : in    std_logic_vector(memory_data_width - 1 downto 0);
    chk_exp_data       : in    std_logic_vector(memory_data_width - 1 downto 0);
    chk_done           : out   std_logic; -- Asserted when data has been checked
    -- Error readout interface
    rd_ce              : in    std_logic;
    rd_index           : in    std_logic_vector(index_width - 1 downto 0);
    rd_not_empty       : out   std_logic;
    rd_full            : out   std_logic;
    rd_overflow        : out   std_logic;
    rd_count           : out   std_logic_vector(count_width - 1 downto 0);
    rd_rec_phase       : out   std_logic_vector(phase_width - 1 downto 0);
    rd_rec_addr        : out   std_logic_vector(memory_addr_width - 1 downto 0);
    rd_rec_bad_bits    : out   std_logic_vector(memory_data_width - 1 downto 0);
    rd_rec_act_data    : out   std_logic_vector(memory_data_width - 1 downto 0)
  );
end entity;

architecture rtl of memory_exerciser_error_log is

  component memory_exerciser_data_check
    generic(
      memory_data_width  : in    natural;
      memory_addr_width  : in    natural
    );
    port(                        
      rst                : in    std_logic;
      clk                : in    std_logic;
      -- Initial address
      init               : in    std_logic;
      init_offset        : in    std_logic_vector(memory_addr_width - 1 downto 0);
      -- Expected and actual data
      ce                 : in    std_logic;
      actual             : in    std_logic_vector(memory_data_width - 1 downto 0);
      expected           : in    std_logic_vector(memory_data_width - 1 downto 0);
      -- Error reporting
      error_valid        : out   std_logic;
      error              : out   std_logic;
      error_addr         : out   std_logic_vector(memory_addr_width - 1 downto 0);
      error_bits         : out   std_logic_vector(memory_data_width - 1 downto 0);
      error_data         : out   std_logic_vector(memory_data_width - 1 downto 0);
      error_lanes        : out   std_logic_vector(memory_data_width / 8 - 1 downto 0)
    );
  end component;
  
  function choice(
    constant true_val  : in string;
    constant false_val : in string;
    constant cond      : in boolean)
  return string is
  begin
    if cond then
      return true_val;
    else
      return false_val;
    end if;
  end;
  
  signal error_valid : std_logic;
  signal error : std_logic;
  signal error_phase : std_logic_vector(phase_width - 1 downto 0);
  signal error_addr : std_logic_vector(memory_addr_width - 1 downto 0);
  signal error_bits : std_logic_vector(memory_data_width - 1 downto 0);
  signal error_data : std_logic_vector(memory_data_width - 1 downto 0);
    
  constant depth : natural := 2**index_width;
  constant record_width : natural := memory_data_width * 2 + phase_width + memory_addr_width;
  subtype record_t is std_logic_vector(record_width - 1 downto 0);
  type record_array_t is array(depth - 1 downto 0) of record_t;
  signal records : record_array_t;
  signal sel_record, sel_record_q : record_t;
  attribute RAM_STYLE : string;
  attribute RAM_STYLE of records : signal is choice("distributed", "block", index_width <= 8);
  
  signal wptr : std_logic_vector(index_width - 1 downto 0);
  signal count : std_logic_vector(count_width - 1 downto 0);
  signal not_empty : std_logic;
  signal full : std_logic;
  signal overflow : std_logic;
    
  
begin

  latch_phase : process(clk)
  begin
    if rising_edge(clk) then
      if init = '1' then
        error_phase <= init_phase;
      end if;
    end if;
  end process;
  
  data_check_inst : memory_exerciser_data_check
    generic map(
      memory_data_width => memory_data_width,
      memory_addr_width => memory_addr_width
    )
    port map(
      rst => rst,
      clk => clk,
      init => init,
      init_offset => init_offset,
      ce => chk_ce,
      actual => chk_act_data,
      expected => chk_exp_data,
      error => error,
      error_valid => error_valid,
      error_addr => error_addr,
      error_bits => error_bits,
      error_data => error_data,
      error_lanes => open
    );
    
  chk_done <= error_valid;
  
  store_errors : process(rst, clk)
  begin
    if rst = '1' then
      not_empty <= '0';
      full <= '0';
      overflow <= '0';
      count <= (others => '0');
    elsif rising_edge(clk) then
      if sr = '1' then
        not_empty <= '0';
        full <= '0';
        overflow <= '0';
        count <= (others => '0');
      else
        if error_valid = '1' and error = '1' and full = '0' then
          full <= AND_reduce(wptr);
        end if;
        if error_valid = '1' and error = '1' then
          not_empty <= '1';
          count <= count + 1;
          if full = '1' then
            overflow <= '1';
          end if;
        end if;
      end if;
    end if;
    
    if rising_edge(clk) then
      if sr = '1' then
        wptr <= (others => '0');
      else
        if error_valid = '1' and error = '1' and full = '0' then
          wptr <= wptr + 1;
        end if;
      end if;
      if error_valid = '1' and error = '1' and full = '0' then
        records(to_integer(unsigned(wptr))) <= error_addr & error_phase & error_bits & error_data;
      end if;
    end if;
  end process;
  
  read_record : process(clk)
  begin
    if rising_edge(clk) then
      if rd_ce = '1' then
        sel_record <= records(to_integer(unsigned(rd_index)));
      end if;
      sel_record_q <= sel_record;
    end if;
  end process;
  
  rd_not_empty <= not_empty;
  rd_full <= full;
  rd_overflow <= overflow;
  rd_count <= count;
  rd_rec_phase <= sel_record_q(memory_data_width * 2 + phase_width - 1 downto memory_data_width * 2);
  rd_rec_addr <= sel_record_q(memory_data_width * 2 + phase_width + memory_addr_width - 1 downto memory_data_width * 2 + phase_width);
  rd_rec_bad_bits <= sel_record_q(memory_data_width * 2 - 1 downto memory_data_width);
  rd_rec_act_data <= sel_record_q(memory_data_width - 1 downto 0);
  
  
end architecture;
