library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_misc.all;

entity memory_exerciser_rearranger is
  generic(
    memory_data_width  : in    natural;         -- Logical data width
    pack_factor        : in    natural;         -- Number of physical data words per logical data word
    -- true  => Rearrange data
    -- false => Don't rearrange data
    pack_rearranged    : in    boolean
  );
  port(
    data_in            : in    std_logic_vector(memory_data_width - 1 downto 0);
    mask_in            : in    std_logic_vector(memory_data_width / 8 - 1 downto 0);
    data_out           : out   std_logic_vector(memory_data_width - 1 downto 0);
    mask_out           : out   std_logic_vector(memory_data_width / 8 - 1 downto 0)
  );
end entity;

architecture rtl of memory_exerciser_rearranger is

  --
  -- The data generated by memory_exerciser_data_gen assumes that the physical data
  -- words in the burst are packed end to end in the logical data. For example, assuming e.g.
  -- 8 beats of 72-bit physical data (9 byte lanes) and logical data width of 576:
  --
  -- bit 575                                                                                             bit 0
  --    --------------------------------------------------------------------------------------------------------
  --    |7.8|7.7|7.6|7.5|7.4|7.3|7.2|7.1|7.0| .... |1.5|1.4|1.3|1.2|1.1|1.0|0.8|0.7|0.6|0.5|0.4|0.3|0.2|0.1|0.0|
  --    --------------------------------------------------------------------------------------------------------
  --     \_________________________________/       _______________________/ \_________________________________/
  --                      |                              |                                   |
  --            phys. word 7 in burst          phys. word 1 in burst               phys. word 0 in burst
  --                   72 bits                        72 bits                             72 bits
  --
  -- where x.y means byte lane y of beat x of physical data.
  --
  -- On the other hand, for some FPGA architectures, the MIG native interface expects write data as follows,
  -- for the same data width & burst parameters as used above:
  --
  -- bit 575                                                                                             bit 0
  --    --------------------------------------------------------------------------------------------------------
  --    |7.8|6.8|5.8|4.8|3.8|2.8|1.8|0.8| .... |7.1|6.1|5.1|4.1|3.1|2.1|1.1|0.1|7.0|6.0|5.0|4.0|3.0|2.0|1.0|0.0|
  --    --------------------------------------------------------------------------------------------------------
  --     \_____________________________/        \_____________________________/ \_____________________________/
  --                    |                                      |                               |
  --          all phys. byte lane 8                  all phys. byte lane 1           all phys. byte lane 0
  --                 64 bits                                64 bits                         64 bits
  --
  -- In other words, the 0th byte lane of every beat of physical data is
  -- packed into the 8 least significant bytes. Then the 1st byte lane of
  -- every beat occupies the next 8 least significant bytes.
  --
  -- This function rearranges data packed in the former manner into the
  -- latter manner, suitable for presentation to MIG.
  --
  function rearrange_to_mig(
    constant vec         : in std_logic_vector;
    constant pack_factor : in natural;
    constant lane_width  : in natural)
  return std_logic_vector is
    -- Example constants for 576-bit data, pack_factor = 8, lane_width = 8:
    --   phys_word_width = 72, num_lane_per_word = 9, lane_group_width = 64
    constant phys_word_width : natural := vec'length / pack_factor;
    constant num_lane_per_word : natural := phys_word_width / lane_width;
    constant lane_group_width : natural := pack_factor * lane_width;
    variable tmp_vec, tmp_rearranged : std_logic_vector(vec'length - 1 downto 0);
    variable returned : std_logic_vector(vec'range);
  begin
    tmp_vec := vec;
    for i in 0 to num_lane_per_word - 1 loop
      for j in 0 to pack_factor - 1 loop
        tmp_rearranged(lane_group_width * i + lane_width * (j + 1) - 1 downto lane_group_width * i + lane_width * j) :=
          tmp_vec(phys_word_width * j + lane_width * (i + 1) - 1 downto phys_word_width * j + lane_width * i);
      end loop;
    end loop;
    returned := tmp_rearranged;
    return returned;
  end;


begin

  data_out <= rearrange_to_mig(data_in, pack_factor, 8) when pack_rearranged else
              data_in;
  mask_out <= rearrange_to_mig(mask_in, pack_factor, 1) when pack_rearranged else
              mask_in;


end architecture;
