--
-- Acknowledgements
-- ================
--
-- This module uses the random number generation technique described in the paper
--
-- "The LUT-SR Family of Uniform Random Number Generators for FPGA Architectures"
-- by David B. Thomas, Member, IEEE, and Wayne Luk, Fellow, IEEE
--
-- IEEE TRANSACTIONS ON VERY LARGE SCALE INTEGRATION (VLSI) SYSTEMS, VOL. 21,
-- NO. 4, APRIL 2013.
--
-- The rng_n1024_r32_t6_k32_s2999b26 module was taken directly from the source
-- code provided at:
--
--   http://cas.ee.ic.ac.uk/people/dt10/research/rngs-fpga-lut_sr.html
--
-- Some minor modifications were performed to make it compile cleanly under the
-- Xilinx Vivado synthesiser and simulator.
--

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_misc.all;
use ieee.numeric_std.all;

entity memory_exerciser_rng is
  generic(
    -- Used as repeating 32 bits of initial state
    seed               : in    std_logic_vector(31 downto 0)
  );
  port(                        
    sr                 : in    std_logic;
    clk                : in    std_logic;
    ce                 : in    std_logic;
    ready              : out   std_logic;
    x                  : out   std_logic_vector(31 downto 0)
  );
end entity;

architecture mixed of memory_exerciser_rng is

  constant n : integer := 1024;
  constant r : integer := 32;
  constant seed_width : natural := seed'length;
  
  component rng_n1024_r32_t6_k32_s2999b26
    port(
      clk   : in std_logic;
      ce    : in std_logic;
      mode  : in std_logic;
      s_in  : in std_logic;
      s_out : out std_logic;
      rng   : out std_logic_vector(r - 1 downto 0)
    );
  end component;
  
  signal initialising : std_logic;
  signal seed_shreg : std_logic_vector(31 downto 0);
  signal init_count : std_logic_vector(9 downto 0);
  signal rng_ce : std_logic;
  
begin

  initialise : process(clk)
  begin
    if rising_edge(clk) then
      if sr = '1' then
        initialising <= '1';
        init_count <= (others => '0');
        seed_shreg <= seed;
      else
        if AND_reduce(init_count) = '1' then
          initialising <= '0';
        end if;
        if initialising = '1' then
          init_count <= init_count + 1;
          seed_shreg(seed_width - 1 downto 1) <= seed_shreg(seed_width - 2 downto 0);
          seed_shreg(0) <= seed_shreg(seed_width - 1);
        end if;
      end if;
    end if;
  end process;
  
  rng_ce <= not sr and (initialising or ce);
  ready <= not initialising;
  
  U0 : rng_n1024_r32_t6_k32_s2999b26
    port map(
      clk => clk,
      ce => rng_ce,
      mode => initialising,
      s_in => seed_shreg(0),
      s_out => open,
      rng => x);
      
end architecture;
