--
-- adb3_ocp_cmd_retime.vhd
--
-- Retimes the command path of an OCP bus by ending critical paths at
-- flip-flops. Implementation is equivalent to a FIFO capable of buffering
-- a single OCP command.
--
-- Maximum command rate, assuming slave side keeps CmdAccept asserted and
-- master side provides valid command on every cycle, is 1 command per 2 clock
-- cycles.
--

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_misc.all;

library work;
use work.adb3_ocp.all;

entity adb3_ocp_cmd_retime is
  port (
    rst       : in  std_logic; -- Async. reset (tie to 0 if unused)
    sr        : in  std_logic; -- Sync. reset (tie to 0 if unused)
    clk       : in  std_logic; -- OCP clock
    -- OCP slave command path
    s_cmd     : in  ocp_cmdT;
    s_addr    : in  std_logic_vector(ADB3_OCP_ADDR_WIDTH-1 downto 0);
    s_blen    : in  std_logic_vector(ADB3_OCP_BURST_WIDTH-1 downto 0);
    s_tag     : in  std_logic_vector(ADB3_OCP_TAG_WIDTH-1 downto 0);
    s_cmdacc  : out std_logic;
    -- OCP master command path
    m_cmd     : out ocp_cmdT;
    m_addr    : out std_logic_vector(ADB3_OCP_ADDR_WIDTH-1 downto 0);
    m_blen    : out std_logic_vector(ADB3_OCP_BURST_WIDTH-1 downto 0);
    m_tag     : out std_logic_vector(ADB3_OCP_TAG_WIDTH-1 downto 0);
    m_cmdacc  : in  std_logic);
end adb3_ocp_cmd_retime;

architecture rtl of adb3_ocp_cmd_retime is
  
  type state_t is (state_idle, state_full);
  signal state, n_state : state_t;
  signal n_m_cmd : ocp_cmdT;
  signal i_m_cmd : ocp_cmdT := OCP_CMD_IDLE;
  signal n_m_addr : std_logic_vector(ADB3_OCP_ADDR_WIDTH-1 downto 0);
  signal i_m_addr : std_logic_vector(ADB3_OCP_ADDR_WIDTH-1 downto 0) := (others => '-');
  signal n_m_blen : std_logic_vector(ADB3_OCP_BURST_WIDTH-1 downto 0);
  signal i_m_blen : std_logic_vector(ADB3_OCP_BURST_WIDTH-1 downto 0) := (others => '-');
  signal n_m_tag : std_logic_vector(ADB3_OCP_TAG_WIDTH-1 downto 0);
  signal i_m_tag : std_logic_vector(ADB3_OCP_TAG_WIDTH-1 downto 0) := (others => '-');
  signal n_s_cmdacc : std_logic;
  signal i_s_cmdacc : std_logic := '0';
  
begin

  m_cmd <= i_m_cmd;
  m_addr <= i_m_addr;
  m_blen <= i_m_blen;
  m_tag <= i_m_tag;
  s_cmdacc <= i_s_cmdacc;
  
  gen_state : process(rst, clk)
  begin
    if rst = '1' then
      state <= state_idle;
      i_m_cmd <= OCP_CMD_IDLE;
      i_s_cmdacc <= '0';
    elsif rising_edge(clk) then
      if sr = '1' then
        state <= state_idle;
        i_m_cmd <= OCP_CMD_IDLE;
        i_s_cmdacc <= '0';
      else
        state <= n_state;
        i_m_cmd <= n_m_cmd;
        i_s_cmdacc <= n_s_cmdacc;
      end if;
    end if;
      
    if rising_edge(clk) then
      i_m_addr <= n_m_addr;
      i_m_blen <= n_m_blen;
      i_m_tag <= n_m_tag;
    end if;
  end process;
  
  gen_n_state : process(
    state,
    s_cmd,
    s_addr,
    s_blen,
    s_tag,
    m_cmdacc,
    i_m_cmd,
    i_m_addr,
    i_m_blen,
    i_m_tag
    )
  begin
    case state is
      when state_idle =>
        if s_cmd /= OCP_CMD_IDLE then
          n_state <= state_full;
          n_m_cmd <= s_cmd;
          n_m_addr <= s_addr;
          n_m_blen <= s_blen;
          n_m_tag <= s_tag;
          n_s_cmdacc <= '0';
        else
          n_state <= state_idle;
          n_m_cmd <= OCP_CMD_IDLE;
          n_m_addr <= (others => '-');
          n_m_blen <= (others => '-');
          n_m_tag <= (others => '-');
          n_s_cmdacc <= '1';
        end if;
    
      when state_full =>
        if m_cmdacc = '1' then
          n_state <= state_idle;
          n_m_cmd <= OCP_CMD_IDLE;
          n_m_addr <= (others => '-');
          n_m_blen <= (others => '-');
          n_m_tag <= (others => '-');
          n_s_cmdacc <= '1';
        else
          n_state <= state_full;
          n_m_cmd <= i_m_cmd;
          n_m_addr <= i_m_addr;
          n_m_blen <= i_m_blen;
          n_m_tag <= i_m_tag;
          n_s_cmdacc <= '0';
        end if;
    end case;
  end process;
  
end rtl;
