--
-- ocp_demux_b128.vhd
-- Component of ADPKGOCP Library
--
-- (C) 2008-2009 Alpha Data Ltd.
--
-- ocp_demux_b128.vhd - OCP_IP interface for splitting the OCP address space
-- evenly among a number of slaves.
--


library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_misc.all;

library Work;
use work.adb3_ocp.all;

library UNISIM;
use UNISIM.VCOMPONENTS.all;

entity adb3_ocp_demux is
  generic (
    number_of_slaves  : integer := 2;
    demux_address_lsb : integer := 16);
  port(
    clk : in std_logic;
    master_m2s : in adb3_ocp_m2sT;
    master_s2m : out adb3_ocp_s2mT;
    slaves_m2s : out adb3_ocp_m2sT_array(0 to number_of_slaves-1);
    slaves_s2m : in  adb3_ocp_s2mT_array(0 to number_of_slaves-1));

end adb3_ocp_demux;

architecture mixed of adb3_ocp_demux is

  signal active_port     : std_logic_vector(7 downto 0) := (others => '0');
  signal master_s2m_i : adb3_ocp_s2mT;
  signal oor_read  : std_logic := '0';
  signal oor_write : std_logic := '0';
  signal oor_cmd_accept : std_logic := '0';
  signal busy : std_logic := '0';
  signal busy_timeout : std_logic_vector(11 downto 0) := (others => '0');
  
  
begin


  gen_slave_m2s_ports: for i in 0 to number_of_slaves-1 generate
    gen_i0: if i = number_of_slaves-1 generate
      slaves_m2s(i).Cmd <= master_m2s.Cmd when busy = '0' and CONV_INTEGER(master_m2s.Addr(demux_address_lsb+7 downto demux_address_lsb)) >= number_of_slaves-1 else ocp_Cmd_IDLE;
      slaves_m2s(i).DataValid <= master_m2s.DataValid when CONV_INTEGER(active_port) >= number_of_slaves-1 else '0';
      slaves_m2s(i).RespAccept <= master_m2s.RespAccept when CONV_INTEGER(active_port) >= number_of_slaves-1 else '0';
    end generate gen_i0;
    gen_igt0: if i < number_of_slaves-1 generate
      slaves_m2s(i).Cmd <= master_m2s.Cmd when  busy = '0' and CONV_INTEGER(master_m2s.Addr(demux_address_lsb+7 downto demux_address_lsb)) = i else ocp_Cmd_IDLE;
      slaves_m2s(i).DataValid <= master_m2s.DataValid when CONV_INTEGER(active_port) = i else '0';
      slaves_m2s(i).RespAccept <= master_m2s.RespAccept when CONV_INTEGER(active_port) = i else '0';
    end generate gen_igt0;    
    slaves_m2s(i).Addr <= EXT(master_m2s.Addr(demux_address_lsb-1 downto 0),64);
    slaves_m2s(i).BurstLength <= master_m2s.BurstLength;
    slaves_m2s(i).Data <= master_m2s.Data;
    slaves_m2s(i).DataByteEn <= master_m2s.DataByteEn;
   
    slaves_m2s(i).Tag <= master_m2s.Tag;   
  end generate gen_slave_m2s_ports;

  set_active_port: process (clk)
  begin  -- process set_active_port
    if clk'event and clk = '1' then  -- rising clock edge
      if master_m2s.Cmd /= ocp_Cmd_IDLE and master_s2m_i.CmdAccept = '1' and busy = '0' then
        if master_m2s.Addr(demux_address_lsb+7 downto demux_address_lsb) < number_of_slaves then
          active_port <= master_m2s.Addr(demux_address_lsb+7 downto demux_address_lsb);
        else
          active_port <= CONV_STD_LOGIC_VECTOR(number_of_slaves-1,8);
        end if;
        busy <= '1';
        busy_timeout <= (others => '0');
      end if;
      if busy = '1' then
        busy_timeout <= busy_timeout+1;
        if AND_reduce(busy_timeout) = '1' then
          busy <= '0';
        end if;
      end if;
      if (master_m2s.DataValid = '1' and master_s2m_i.DataAccept = '1') or (master_s2m_i.Resp(0) = '1' and master_m2s.RespAccept = '1') then
        busy <= '0';
      end if;
      
      
    end if;
  end process set_active_port;

   
 process (slaves_s2m,master_m2s,active_port) is
  begin  -- process
    master_s2m_i <= ADB3_OCP_DEFAULT_S2M;
    for i in 0 to number_of_slaves-1 loop
      if i = number_of_slaves-1 then
        if CONV_INTEGER(master_m2s.Addr(demux_address_lsb+7 downto demux_address_lsb)) >= number_of_slaves-1  then
          master_s2m_i.CmdAccept <= slaves_s2m(i).CmdAccept;         
        end if;
        if CONV_INTEGER(active_port) >= number_of_slaves-1 then
          master_s2m_i.DataAccept <= slaves_s2m(i).DataAccept; 
          master_s2m_i.Resp <= slaves_s2m(i).Resp;
        end if;
      end if;
      if i < number_of_slaves-1 then
        if CONV_INTEGER(master_m2s.Addr(demux_address_lsb+7 downto demux_address_lsb)) = i then
          master_s2m_i.CmdAccept <= slaves_s2m(i).CmdAccept;
          
        end if;
        if CONV_INTEGER(active_port) = i then
          master_s2m_i.DataAccept <= slaves_s2m(i).DataAccept;
          master_s2m_i.Resp <= slaves_s2m(i).Resp;
        end if;
      end if;
      if CONV_INTEGER(active_port) = i then
        master_s2m_i.Data <= slaves_s2m(i).Data;
        master_s2m_i.Tag <= slaves_s2m(i).Tag;
      end if;
    end loop;  -- i
  end process;
  
  
  --gen_s2m_ports: for i in 0 to number_of_slaves-1 generate
  --  gen_i0a: if i = number_of_slaves-1 generate
  --    master_s2m_i.CmdAccept <= slaves_s2m(i).CmdAccept when CONV_INTEGER(master_m2s.Addr(demux_address_lsb+7 downto demux_address_lsb)) >= number_of_slaves-1 else 'Z';
  --    master_s2m_i.DataAccept <= slaves_s2m(i).DataAccept when CONV_INTEGER(active_port) >= number_of_slaves-1  else 'Z';
  --    master_s2m_i.Resp <= slaves_s2m(i).Resp when CONV_INTEGER(active_port) >= number_of_slaves-1 else (others => 'Z');
  --  end generate gen_i0a;
  --  gen_igt0a: if i < number_of_slaves-1 generate
  --    master_s2m_i.CmdAccept <= slaves_s2m(i).CmdAccept when CONV_INTEGER(master_m2s.Addr(demux_address_lsb+7 downto demux_address_lsb)) = i else 'Z';
  --    master_s2m_i.DataAccept <= slaves_s2m(i).DataAccept when CONV_INTEGER(active_port) = i else 'Z';
  --    master_s2m_i.Resp <= slaves_s2m(i).Resp when CONV_INTEGER(active_port) = i else (others => 'Z');
  --  end generate gen_igt0a;    
  --  master_s2m_i.Data <= slaves_s2m(i).Data when CONV_INTEGER(active_port) = i else (others => 'Z'); 
  --  master_s2m_i.Tag <= slaves_s2m(i).Tag when CONV_INTEGER(active_port) = i else (others => 'Z');   
  --end generate gen_s2m_ports;

  
  master_s2m <= master_s2m_i;
  
end mixed;
