--------------------------------------------------------------------------------
--
-- Package Name      : adb3_ocp
--
--                    Copyright 2009 by
--                    Alpha Data Parallel Systems Ltd.
--                    All Rights Reserved.
--
-- Original Author  : Andrew McCormick
-- Created          : 
--
-- Modified By      : 
-- Date             :
-- Change Notes     :     
--
-- Description      : Package For ADB3 OCP Type
--
--
-- Dependencies     :
--
-- Disclaimer       : THESE DESIGNS ARE PROVIDED "AS IS" WITH NO WARRANTY
--                    WHATSOEVER AND ALPHA DATA SPECIFICALLY DISCLAIMS ANY
--                    WARRANTIES IMPLIED OF MERCHANTABILITY, FITNESS FOR A
--                    PARTICULAR PURPOSE, OR AGAINST INFRINGEMENT.
--
--------------------------------------------------------------------------------


library IEEE;
use IEEE.std_logic_1164.all;

package adb3_ocp is

  constant ADB3_OCP_DATA_WIDTH  : integer := 128;
  constant ADB3_OCP_ADDR_WIDTH  : integer := 64;
  constant ADB3_OCP_BURST_WIDTH : integer := 12;
  constant ADB3_OCP_TAG_WIDTH   : integer := 8;

  constant ADB3L_OCP_DATA_WIDTH  : integer := 32;
  constant ADB3L_OCP_ADDR_WIDTH  : integer := 32;
  constant ADB3L_OCP_TAG_WIDTH   : integer := 8;

  constant ADB3W_OCP_DATA_WIDTH  : integer := 256;
  
  subtype ocp_CmdT is std_logic_vector(2 downto 0);
  constant OCP_CMD_IDLE         : ocp_CmdT := "000";
  constant OCP_CMD_WRITE        : ocp_CmdT := "001";
  constant OCP_CMD_READ         : ocp_CmdT := "010";
  constant OCP_CMD_WRITENONPOST : ocp_CmdT := "101";

  subtype ocp_RespT is std_logic_vector(1 downto 0);
  constant OCP_RESP_NONE   : ocp_RespT := "00";
  constant OCP_RESP_VALID  : ocp_RespT := "01";
  constant OCP_RESP_FAILED : ocp_RespT := "10";
  constant OCP_RESP_ERROR  : ocp_RespT := "11";

  function to_std_logic(bool : boolean) return std_logic;


  -- Burst Capable bus for high performance data flows
  -- 128 bit wide data path
  type adb3_ocp_m2sT is record  -- Block Data Flow from Master to Slave...
    -- Request group...
    Cmd         : ocp_CmdT;   -- Command b'000=Idle b'010=Read b'001=Write
    Addr        : std_logic_vector(ADB3_OCP_ADDR_WIDTH-1 downto 0);
    BurstLength : std_logic_vector(ADB3_OCP_BURST_WIDTH-1 downto 0);
    -- DataHandshake group...
    Data        : std_logic_vector(ADB3_OCP_DATA_WIDTH-1 downto 0);
    DataByteEn  : std_logic_vector(ADB3_OCP_DATA_WIDTH/8-1 downto 0);
    DataValid   : std_logic;  -- Qualifies Data and the other signals in bundle
    -- Response group...
    RespAccept  : std_logic;  -- Master indication read data was accepted
    -- Tag group
    Tag         : std_logic_vector(ADB3_OCP_TAG_WIDTH-1 downto 0);
  end record adb3_ocp_m2sT;
  type adb3_ocp_s2mT is record  -- Block Data Flow from Slave to Master...
    -- Request group...
    CmdAccept  : std_logic;   -- Slave indication command/request accepted
    -- DataHandshake group...
    DataAccept : std_logic;             -- Slave indication write data accepted
    -- Response group...
    Data       : std_logic_vector(ADB3_OCP_DATA_WIDTH-1 downto 0);
    Resp       : ocp_RespT;   -- Slave read response b'00=None b'01=Valid
    -- Tag group
    Tag        : std_logic_vector(ADB3_OCP_TAG_WIDTH-1 downto 0);
  end record adb3_ocp_s2mT;

  --
  -- Define default tieoffs...
  constant ADB3_OCP_DEFAULT_M2S : adb3_ocp_m2sT :=
    (Cmd         => OCP_CMD_IDLE,
     Addr        => (others => '0'),
     BurstLength => (others => '0'),
     Data        => (others => '0'),
     DataByteEn  => (others => '0'),
     DataValid   => '0',
     RespAccept  => '0',
     Tag         => (others => '0')
     );
  constant ADB3_OCP_DEFAULT_S2M : adb3_ocp_s2mT :=
    (CmdAccept  => '0',
     DataAccept => '0',
     Data       => (others => '0'),
     Resp       => OCP_RESP_NONE,
     Tag        => (others => '0')
     );

  -- Define type tri-states
  constant ADB3_OCP_Z_M2S : adb3_ocp_m2sT :=
    (Cmd         => (others => 'Z'),
     Addr        => (others => 'Z'),
     BurstLength => (others => 'Z'),
     Data        => (others => 'Z'),
     DataByteEn  => (others => 'Z'),
     DataValid   => 'Z',
     RespAccept  => 'Z',
     Tag         => (others => 'Z'));

  constant ADB3_OCP_Z_S2M : adb3_ocp_s2mT :=
    (CmdAccept  => 'Z',
     DataAccept => 'Z',
     Data       => (others => 'Z'),
     Resp       => (others => 'Z'),
     Tag        => (others => 'Z'));



  type adb3_ocp_s2mT_array is array (natural range <>) of adb3_ocp_s2mT;
  type adb3_ocp_m2sT_array is array (natural range <>) of adb3_ocp_m2sT;

 

  -- Non-bursting bus for low performance - direct slave - register accesses
  -- 32 bit wide data path
  type adb3l_ocp_m2sT is record  -- Block Data Flow from Master to Slave...
    -- Request group...
    Cmd         : ocp_CmdT;   -- Command b'000=Idle b'010=Read b'001=Write
    Addr        : std_logic_vector(ADB3L_OCP_ADDR_WIDTH-1 downto 0);
    -- DataHandshake group... 
    Data        : std_logic_vector(ADB3L_OCP_DATA_WIDTH-1 downto 0);
    DataByteEn  : std_logic_vector(ADB3L_OCP_DATA_WIDTH/8-1 downto 0);
    DataValid   : std_logic;  -- Qualifies Data and the other signals in bundle
    -- Response group...
    RespAccept  : std_logic;  -- Master indication read data was accepted
  end record adb3l_ocp_m2sT;
  type adb3l_ocp_s2mT is record  -- Block Data Flow from Slave to Master...
    -- Request group...
    CmdAccept  : std_logic;   -- Slave indication command/request accepte
    -- DataHandshake group...
    DataAccept : std_logic;             -- Slave indication write data accepted
    -- Response group...
    Data       : std_logic_vector(ADB3L_OCP_DATA_WIDTH-1 downto 0);
    Resp       : ocp_RespT;   -- Slave read response b'00=None b'01=Valid
  end record adb3l_ocp_s2mT;

  --
  -- Define default tieoffs...
  constant ADB3L_OCP_DEFAULT_M2S : adb3l_ocp_m2sT :=
    (Cmd         => OCP_CMD_IDLE,
     Addr        => (others => '0'),
     Data        => (others => '0'),
     DataByteEn  => (others => '0'),
     DataValid   => '0',
     RespAccept  => '0');
  constant ADB3L_OCP_DEFAULT_S2M : adb3l_ocp_s2mT :=
    (CmdAccept  => '0',
     DataAccept => '0',
     Data       => (others => '0'),
     Resp       => OCP_RESP_NONE
     );

  -- Define type tri-states
  constant ADB3L_OCP_Z_M2S : adb3l_ocp_m2sT :=
    (Cmd         => (others => 'Z'),
     Addr        => (others => 'Z'),
     Data        => (others => 'Z'),
     DataByteEn  => (others => 'Z'),
     DataValid   => 'Z',
     RespAccept  => 'Z');

  constant ADB3L_OCP_Z_S2M : adb3l_ocp_s2mT :=
    (CmdAccept  => 'Z',
     DataAccept => 'Z',
     Data       => (others => 'Z'),
     Resp       => (others => 'Z'));



  type adb3l_ocp_s2mT_array is array (natural range <>) of adb3l_ocp_s2mT;
  type adb3l_ocp_m2sT_array is array (natural range <>) of adb3l_ocp_m2sT;


    -- Burst Capable bus for high performance data flows
  -- 256 bit wide data path
  type adb3w_ocp_m2sT is record  -- Block Data Flow from Master to Slave...
    -- Request group...
    Cmd         : ocp_CmdT;   -- Command b'000=Idle b'010=Read b'001=Write
    Addr        : std_logic_vector(ADB3_OCP_ADDR_WIDTH-1 downto 0);
    BurstLength : std_logic_vector(ADB3_OCP_BURST_WIDTH-1 downto 0);
    -- DataHandshake group...
    Data        : std_logic_vector(ADB3W_OCP_DATA_WIDTH-1 downto 0);
    DataByteEn  : std_logic_vector(ADB3W_OCP_DATA_WIDTH/8-1 downto 0);
    DataValid   : std_logic;  -- Qualifies Data and the other signals in bundle
    -- Response group...
    RespAccept  : std_logic;  -- Master indication read data was accepted
    -- Tag group
    Tag         : std_logic_vector(ADB3_OCP_TAG_WIDTH-1 downto 0);
  end record adb3w_ocp_m2sT;
  type adb3w_ocp_s2mT is record  -- Block Data Flow from Slave to Master...
    -- Request group...
    CmdAccept  : std_logic;   -- Slave indication command/request accepted
    -- DataHandshake group...
    DataAccept : std_logic;             -- Slave indication write data accepted
    -- Response group...
    Data       : std_logic_vector(ADB3W_OCP_DATA_WIDTH-1 downto 0);
    Resp       : ocp_RespT;   -- Slave read response b'00=None b'01=Valid
    -- Tag group
    Tag        : std_logic_vector(ADB3_OCP_TAG_WIDTH-1 downto 0);
  end record adb3w_ocp_s2mT;

  --
  -- Define default tieoffs...
  constant ADB3W_OCP_DEFAULT_M2S : adb3w_ocp_m2sT :=
    (Cmd         => OCP_CMD_IDLE,
     Addr        => (others => '0'),
     BurstLength => (others => '0'),
     Data        => (others => '0'),
     DataByteEn  => (others => '0'),
     DataValid   => '0',
     RespAccept  => '0',
     Tag         => (others => '0')
     );
  constant ADB3W_OCP_DEFAULT_S2M : adb3w_ocp_s2mT :=
    (CmdAccept  => '0',
     DataAccept => '0',
     Data       => (others => '0'),
     Resp       => OCP_RESP_NONE,
     Tag        => (others => '0')
     );

  -- Define type tri-states
  constant ADB3W_OCP_Z_M2S : adb3w_ocp_m2sT :=
    (Cmd         => (others => 'Z'),
     Addr        => (others => 'Z'),
     BurstLength => (others => 'Z'),
     Data        => (others => 'Z'),
     DataByteEn  => (others => 'Z'),
     DataValid   => 'Z',
     RespAccept  => 'Z',
     Tag         => (others => 'Z'));

  constant ADB3W_OCP_Z_S2M : adb3w_ocp_s2mT :=
    (CmdAccept  => 'Z',
     DataAccept => 'Z',
     Data       => (others => 'Z'),
     Resp       => (others => 'Z'),
     Tag        => (others => 'Z'));



  type adb3w_ocp_s2mT_array is array (natural range <>) of adb3w_ocp_s2mT;
  type adb3w_ocp_m2sT_array is array (natural range <>) of adb3w_ocp_m2sT;

 

  
  
  
  function to_std_logic_vector(x : adb3_ocp_m2sT) return std_logic_vector;
  function to_std_logic_vector(x : adb3_ocp_s2mT) return std_logic_vector;
  function to_std_logic_vector(x : adb3_ocp_m2sT_array) return std_logic_vector;
  function to_std_logic_vector(x : adb3_ocp_s2mT_array) return std_logic_vector;
  function to_std_logic_vector(x : adb3w_ocp_m2sT) return std_logic_vector;
  function to_std_logic_vector(x : adb3w_ocp_s2mT) return std_logic_vector;
  function to_std_logic_vector(x : adb3w_ocp_m2sT_array) return std_logic_vector;
  function to_std_logic_vector(x : adb3w_ocp_s2mT_array) return std_logic_vector;
  function to_adb3_ocp_m2sT(x : std_logic_vector) return adb3_ocp_m2sT;
  function to_adb3_ocp_s2mT(x : std_logic_vector) return adb3_ocp_s2mT;
  function to_adb3_ocp_m2sT_array(x : std_logic_vector) return adb3_ocp_m2sT_array;
  function to_adb3_ocp_s2mT_array(x : std_logic_vector) return adb3_ocp_s2mT_array;
  function to_adb3w_ocp_m2sT(x : std_logic_vector) return adb3w_ocp_m2sT;
  function to_adb3w_ocp_s2mT(x : std_logic_vector) return adb3w_ocp_s2mT;
  function to_adb3w_ocp_m2sT_array(x : std_logic_vector) return adb3w_ocp_m2sT_array;
  function to_adb3w_ocp_s2mT_array(x : std_logic_vector) return adb3w_ocp_s2mT_array;
  function to_std_logic_vector(x : adb3l_ocp_m2sT) return std_logic_vector;
  function to_std_logic_vector(x : adb3l_ocp_s2mT) return std_logic_vector;
  function to_adb3l_ocp_m2sT(x : std_logic_vector) return adb3l_ocp_m2sT;
  function to_adb3l_ocp_s2mT(x : std_logic_vector) return adb3l_ocp_s2mT;
  
  constant adb3_ocp_m2sT_slv_length  : natural := 3 + ADB3_OCP_ADDR_WIDTH + ADB3_OCP_BURST_WIDTH + ADB3_OCP_DATA_WIDTH + ADB3_OCP_DATA_WIDTH/8 + 1 + 1 + ADB3_OCP_TAG_WIDTH;
  constant adb3_ocp_s2mT_slv_length  : natural := 1 + 1 + ADB3_OCP_DATA_WIDTH + 2 + ADB3_OCP_TAG_WIDTH;
  constant adb3w_ocp_m2sT_slv_length  : natural := 3 + ADB3_OCP_ADDR_WIDTH + ADB3_OCP_BURST_WIDTH + ADB3W_OCP_DATA_WIDTH + ADB3W_OCP_DATA_WIDTH/8 + 1 + 1 + ADB3_OCP_TAG_WIDTH;
  constant adb3w_ocp_s2mT_slv_length  : natural := 1 + 1 + ADB3W_OCP_DATA_WIDTH + 2 + ADB3_OCP_TAG_WIDTH;
  constant adb3l_ocp_m2sT_slv_length  : natural := 3 + ADB3L_OCP_ADDR_WIDTH + ADB3L_OCP_DATA_WIDTH + ADB3L_OCP_DATA_WIDTH/8 + 1 + 1;
  constant adb3l_ocp_s2mT_slv_length  : natural := 1 + 1 + ADB3L_OCP_DATA_WIDTH + 2;
  
  type register64_array is array (natural range <>) of std_logic_vector(63 downto 0);
  type register32_array is array (natural range <>) of std_logic_vector(31 downto 0);

  type register32_be_array is array (natural range <>) of std_logic_vector(3 downto 0);


  component adb3_ocp_register_interface32
    generic (
      number_of_registers : natural := 1);
    port (
      -- OCP Interface
      ocp_clk          : in  std_logic;
      ocp_m2s          : in  adb3_ocp_m2sT;
      ocp_s2m          : out adb3_ocp_s2mT;
      -- simple register interface
      registers_out    : out register32_array(0 to number_of_registers-1);
      registers_back   : in  register32_array(0 to number_of_registers-1);
      register_written : out std_logic_vector(number_of_registers-1 downto 0)); 
  end component;

  component adb3_ocp_register_interface32_lite_be
    generic (
      number_of_registers : natural);
    port (
      ocp_clk          : in  std_logic;
      ocp_m2s          : in  adb3_ocp_m2sT;
      ocp_s2m          : out adb3_ocp_s2mT;
      registers_out    : out register32_array(0 to number_of_registers-1);
      registers_back   : in  register32_array(0 to number_of_registers-1);
      register_written : out std_logic_vector(number_of_registers-1 downto 0);
      registers_be     : out register32_be_array(0 to number_of_registers-1));
  end component;

  component adb3_ocp_register_interface64
    generic (
      number_of_registers : natural := 1);
    port (
      -- OCP Interface
      ocp_clk          : in  std_logic;
      ocp_m2s          : in  adb3_ocp_m2sT;
      ocp_s2m          : out adb3_ocp_s2mT;
      -- simple register interface
      registers_out    : out register64_array(0 to number_of_registers-1);
      registers_back   : in  register64_array(0 to number_of_registers-1);
      register_written : out std_logic_vector(number_of_registers-1 downto 0)); 
  end component;

  component adb3_ocp_simple_bus_interface32
    generic (
      read_latency : natural := 1);
    port (
      -- OCP Interface
      ocp_clk : in  std_logic;
      ocp_m2s : in  adb3_ocp_m2sT;
      ocp_s2m : out adb3_ocp_s2mT;
      -- simple bus interface
      d       : out std_logic_vector(31 downto 0);
      q       : in  std_logic_vector(31 downto 0);
      a       : out std_logic_vector(21 downto 0);
      w       : out std_logic;
      r       : out std_logic;
      be      : out std_logic_vector(3 downto 0));
  end component;

  component adb3_ocp_simple_bus_interface64
    generic (
      read_latency : natural);
    port (
      ocp_clk : in  std_logic;
      ocp_m2s : in  adb3_ocp_m2sT;
      ocp_s2m : out adb3_ocp_s2mT;
      -- simple bus interface
      d       : out std_logic_vector(63 downto 0);
      q       : in  std_logic_vector(63 downto 0);
      a       : out std_logic_vector(21 downto 0);
      w       : out std_logic;
      r       : out std_logic;
      be      : out std_logic_vector(7 downto 0));
  end component;

  component adb3_ocp_multiport2
    generic (
      round_robin_arb : boolean;
      use_timeout     : boolean;
      timeout_bits    : integer);
    port (
      ocp_m2s_memif : out adb3_ocp_m2sT;
      ocp_s2m_memif : in  adb3_ocp_s2mT;
      ocp_m2s0      : in  adb3_ocp_m2sT;
      ocp_s2m0      : out adb3_ocp_s2mT;
      ocp_m2s1      : in  adb3_ocp_m2sT;
      ocp_s2m1      : out adb3_ocp_s2mT;
      status        : out std_logic_vector(3 downto 0);
      Clk           : in  std_logic);
  end component;

  component adb3_ocp_demux
    generic (
      number_of_slaves  : integer;
      demux_address_lsb : integer);
    port (
      clk        : in  std_logic;
      master_m2s : in  adb3_ocp_m2sT;
      master_s2m : out adb3_ocp_s2mT;
      slaves_m2s : out adb3_ocp_m2sT_array(0 to number_of_slaves-1);
      slaves_s2m : in  adb3_ocp_s2mT_array(0 to number_of_slaves-1));
  end component;

  component adb3_ocp_register
    port (
      clk        : in  std_logic;
      master_m2s : in  adb3_ocp_m2sT;
      master_s2m : out adb3_ocp_s2mT;
      slave_m2s  : out adb3_ocp_m2sT;
      slave_s2m  : in  adb3_ocp_s2mT);
  end component;

  component adb3_ocp_cross_clock_dom
    port (
      master_clk : in  std_logic;
      slave_clk  : in  std_logic;
      master_m2s : in  adb3_ocp_m2sT;
      master_s2m : out adb3_ocp_s2mT;
      slave_m2s  : out adb3_ocp_m2sT;
      slave_s2m  : in  adb3_ocp_s2mT);
  end component;

  component adb3_ocp_dummy
    port (
      ocp_clk : in  std_logic;
      ocp_m2s : in  adb3_ocp_m2sT;
      ocp_s2m : out adb3_ocp_s2mT);
  end component;


  component adb3_ocp_recovery
    port (
      clk         : in  std_logic;
      src_m2s     : in  adb3_ocp_m2sT;
      src_s2m     : out adb3_ocp_s2mT;
      dst_m2s     : out adb3_ocp_m2sT;
      dst_s2m     : in  adb3_ocp_s2mT;
      dst_offline : in  std_logic;
      inflight    : out std_logic;
      clear       : in  std_Logic);
  end component;


  component adb3_to_adb3l is
    port (
      clk     : in  std_logic;
      pri_m2s : in  adb3_ocp_m2sT;
      pri_s2m : out adb3_ocp_s2mT;
      sec_m2s : out adb3l_ocp_m2sT;
      sec_s2m : in  adb3l_ocp_s2mT);
  end component adb3_to_adb3l;

   component adb3l_to_adb3 is
    port (
      pri_m2s : in  adb3l_ocp_m2sT;
      pri_s2m : out adb3l_ocp_s2mT;
      sec_m2s : out adb3_ocp_m2sT;
      sec_s2m : in  adb3_ocp_s2mT);
  end component adb3l_to_adb3;

    -- Define ADB3 OCP subtypes
  subtype adb3_ocp_addr_s  is std_logic_vector(ADB3_OCP_ADDR_WIDTH-1 downto 0);
  subtype adb3_ocp_burst_s is std_logic_vector(ADB3_OCP_BURST_WIDTH-1 downto 0);
  subtype adb3_ocp_tag_s   is std_logic_vector(ADB3_OCP_TAG_WIDTH-1 downto 0);
  subtype adb3_ocp_data_s  is std_logic_vector(ADB3_OCP_DATA_WIDTH-1 downto 0);
  subtype adb3_ocp_be_s    is std_logic_vector(ADB3_OCP_DATA_WIDTH/8-1 downto 0);

   constant ADB3_OCP_CMD_WIDTH  : integer := 3;
   constant ADB3_OCP_RESP_WIDTH : integer := 2;
  
end package adb3_ocp;

package body adb3_ocp is

  function to_std_logic (bool : boolean) return std_logic is
  begin
    if (bool) then return '1'; else return '0'; end if;
  end function to_std_logic;


  function to_std_logic_vector(x : adb3_ocp_m2sT) return std_logic_vector is
    variable y : std_logic_vector(adb3_ocp_m2sT_slv_length-1 downto 0);
  begin
    y(2 downto 0) := x.Cmd;
    y(ADB3_OCP_ADDR_WIDTH-1 + 3 downto 3) := x.Addr;
    y(ADB3_OCP_BURST_WIDTH-1 + ADB3_OCP_ADDR_WIDTH+3 downto ADB3_OCP_ADDR_WIDTH+3) := x.BurstLength;
    y(ADB3_OCP_DATA_WIDTH-1 + ADB3_OCP_BURST_WIDTH+ADB3_OCP_ADDR_WIDTH+3 downto ADB3_OCP_BURST_WIDTH+ADB3_OCP_ADDR_WIDTH+3) := x.Data;
    y(ADB3_OCP_DATA_WIDTH/8-1 + ADB3_OCP_DATA_WIDTH+ADB3_OCP_BURST_WIDTH+ADB3_OCP_ADDR_WIDTH+3 downto ADB3_OCP_DATA_WIDTH+ADB3_OCP_BURST_WIDTH+ADB3_OCP_ADDR_WIDTH+3) := x.DataByteEn;
    y(ADB3_OCP_DATA_WIDTH/8+ADB3_OCP_DATA_WIDTH+ADB3_OCP_BURST_WIDTH+ADB3_OCP_ADDR_WIDTH+3) := x.DataValid;
    y(1+ADB3_OCP_DATA_WIDTH/8+ADB3_OCP_DATA_WIDTH+ADB3_OCP_BURST_WIDTH+ADB3_OCP_ADDR_WIDTH+3) := x.RespAccept;
    y(ADB3_OCP_TAG_WIDTH-1 + 1+1+ADB3_OCP_DATA_WIDTH/8+ADB3_OCP_DATA_WIDTH+ADB3_OCP_BURST_WIDTH+ADB3_OCP_ADDR_WIDTH+3 downto 1+1+ADB3_OCP_DATA_WIDTH/8+ADB3_OCP_DATA_WIDTH+ADB3_OCP_BURST_WIDTH+ADB3_OCP_ADDR_WIDTH+3) := x.Tag;
    return y;
  end function to_std_logic_vector;

  function to_std_logic_vector(x : adb3_ocp_s2mT) return std_logic_vector is
    variable y : std_logic_vector(adb3_ocp_s2mT_slv_length-1 downto 0);
  begin
    y(0) := x.CmdAccept;
    y(1) := x.DataAccept;
    y(ADB3_OCP_DATA_WIDTH-1+2 downto 2) := x.Data;
    y(1 + ADB3_OCP_DATA_WIDTH+2 downto ADB3_OCP_DATA_WIDTH+2) := x.Resp;
    y(ADB3_OCP_TAG_WIDTH-1 + 2+ADB3_OCP_DATA_WIDTH+2 downto 2+ADB3_OCP_DATA_WIDTH+2) := x.Tag;
    return y;
  end function to_std_logic_vector;
  
  function to_std_logic_vector(x : adb3_ocp_m2sT_array) return std_logic_vector is
    variable y : std_logic_vector(adb3_ocp_m2sT_slv_length*x'length-1 downto 0);
  begin
    for i in 0 to x'length-1 loop
      y((i+1)*adb3_ocp_m2sT_slv_length-1 downto i*adb3_ocp_m2sT_slv_length) := to_std_logic_vector(x(i));
    end loop;  -- i
    return y;
  end function to_std_logic_vector;


  function to_std_logic_vector(x : adb3_ocp_s2mT_array) return std_logic_vector is
    variable y : std_logic_vector(adb3_ocp_s2mT_slv_length*x'length-1 downto 0);
  begin
    for i in 0 to x'length-1 loop
      y((i+1)*adb3_ocp_s2mT_slv_length-1 downto i*adb3_ocp_s2mT_slv_length) := to_std_logic_vector(x(i));
    end loop;  -- i
    return y;
  end function to_std_logic_vector;


  
  
  
  function to_adb3_ocp_m2sT(x : std_logic_vector) return adb3_ocp_m2sT is
    variable y : adb3_ocp_m2sT;
  begin
    y.Cmd := x(2 downto 0);
    y.Addr := x(ADB3_OCP_ADDR_WIDTH-1 + 3 downto 3);
    y.BurstLength := x(ADB3_OCP_BURST_WIDTH-1 + ADB3_OCP_ADDR_WIDTH+3 downto ADB3_OCP_ADDR_WIDTH+3);
    y.Data := x(ADB3_OCP_DATA_WIDTH-1 + ADB3_OCP_BURST_WIDTH+ADB3_OCP_ADDR_WIDTH+3 downto ADB3_OCP_BURST_WIDTH+ADB3_OCP_ADDR_WIDTH+3);
    y.DataByteEn := x(ADB3_OCP_DATA_WIDTH/8-1 + ADB3_OCP_DATA_WIDTH+ADB3_OCP_BURST_WIDTH+ADB3_OCP_ADDR_WIDTH+3 downto ADB3_OCP_DATA_WIDTH+ADB3_OCP_BURST_WIDTH+ADB3_OCP_ADDR_WIDTH+3);
    y.DataValid := x(ADB3_OCP_DATA_WIDTH/8+ADB3_OCP_DATA_WIDTH+ADB3_OCP_BURST_WIDTH+ADB3_OCP_ADDR_WIDTH+3);
    y.RespAccept := x(1+ADB3_OCP_DATA_WIDTH/8+ADB3_OCP_DATA_WIDTH+ADB3_OCP_BURST_WIDTH+ADB3_OCP_ADDR_WIDTH+3);
    y.Tag := x(ADB3_OCP_TAG_WIDTH-1 + 1+1+ADB3_OCP_DATA_WIDTH/8+ADB3_OCP_DATA_WIDTH+ADB3_OCP_BURST_WIDTH+ADB3_OCP_ADDR_WIDTH+3 downto 1+1+ADB3_OCP_DATA_WIDTH/8+ADB3_OCP_DATA_WIDTH+ADB3_OCP_BURST_WIDTH+ADB3_OCP_ADDR_WIDTH+3);
    return y;      
  end function to_adb3_ocp_m2sT;

  function to_adb3_ocp_s2mT(x : std_logic_vector) return adb3_ocp_s2mT is
    variable y : adb3_ocp_s2mT;
  begin
    y.CmdAccept := x(0);
    y.DataAccept := x(1);
    y.Data := x(ADB3_OCP_DATA_WIDTH-1+2 downto 2);
    y.Resp := x(1 + ADB3_OCP_DATA_WIDTH+2 downto ADB3_OCP_DATA_WIDTH+2);
    y.Tag := x(ADB3_OCP_TAG_WIDTH-1 + 2+ADB3_OCP_DATA_WIDTH+2 downto 2+ADB3_OCP_DATA_WIDTH+2);
    return y;
  end function to_adb3_ocp_s2mT;

  function to_adb3_ocp_m2sT_array(x : std_logic_vector) return adb3_ocp_m2sT_array is
    variable y : adb3_ocp_m2sT_array(0 to (x'length/adb3_ocp_m2sT_slv_length) -1);
    variable x_i : std_logic_vector(adb3_ocp_m2sT_slv_length-1 downto 0); 
  begin
    for i in 0 to (x'length/adb3_ocp_m2sT_slv_length)-1 loop
      x_i := x((i+1)*adb3_ocp_m2sT_slv_length-1 downto i*adb3_ocp_m2sT_slv_length);
      y(i) := to_adb3_ocp_m2sT(x_i);
    end loop;  -- i
    return y;
  end function to_adb3_ocp_m2sT_array;

  function to_adb3_ocp_s2mT_array(x : std_logic_vector) return adb3_ocp_s2mT_array is
    variable y : adb3_ocp_s2mT_array(0 to (x'length/adb3_ocp_s2mT_slv_length) -1);
    variable x_i : std_logic_vector(adb3_ocp_s2mT_slv_length-1 downto 0); 
  begin
    for i in 0 to (x'length/adb3_ocp_s2mT_slv_length)-1 loop
      x_i := x((i+1)*adb3_ocp_s2mT_slv_length-1 downto i*adb3_ocp_s2mT_slv_length);
      y(i) := to_adb3_ocp_s2mT(x_i);
    end loop;  -- i
    return y;
  end function to_adb3_ocp_s2mT_array;



  function to_std_logic_vector(x : adb3l_ocp_m2sT) return std_logic_vector is
    variable y : std_logic_vector(adb3l_ocp_m2sT_slv_length-1 downto 0);
  begin
    y(2 downto 0) := x.Cmd;
    y(ADB3L_OCP_ADDR_WIDTH-1 + 3 downto 3) := x.Addr;
    y(ADB3L_OCP_DATA_WIDTH-1 + ADB3L_OCP_ADDR_WIDTH+3 downto ADB3L_OCP_ADDR_WIDTH+3) := x.Data;
    y(ADB3L_OCP_DATA_WIDTH/8-1 + ADB3L_OCP_DATA_WIDTH+ADB3L_OCP_ADDR_WIDTH+3 downto ADB3L_OCP_DATA_WIDTH+ADB3L_OCP_ADDR_WIDTH+3) := x.DataByteEn;
    y(ADB3L_OCP_DATA_WIDTH/8+ADB3L_OCP_DATA_WIDTH+ADB3L_OCP_ADDR_WIDTH+3) := x.DataValid;
    y(1+ADB3L_OCP_DATA_WIDTH/8+ADB3L_OCP_DATA_WIDTH+ADB3L_OCP_ADDR_WIDTH+3) := x.RespAccept;
    return y;
  end function to_std_logic_vector;

  function to_std_logic_vector(x : adb3l_ocp_s2mT) return std_logic_vector is
    variable y : std_logic_vector(adb3l_ocp_s2mT_slv_length-1 downto 0);
  begin
    y(0) := x.CmdAccept;
    y(1) := x.DataAccept;
    y(ADB3L_OCP_DATA_WIDTH-1+2 downto 2) := x.Data;
    y(1 + ADB3L_OCP_DATA_WIDTH+2 downto ADB3L_OCP_DATA_WIDTH+2) := x.Resp;
    return y;
  end function to_std_logic_vector;







 function to_std_logic_vector(x : adb3w_ocp_m2sT) return std_logic_vector is
    variable y : std_logic_vector(adb3w_ocp_m2sT_slv_length-1 downto 0);
  begin
    y(2 downto 0) := x.Cmd;
    y(ADB3_OCP_ADDR_WIDTH-1 + 3 downto 3) := x.Addr;
    y(ADB3_OCP_BURST_WIDTH-1 + ADB3_OCP_ADDR_WIDTH+3 downto ADB3_OCP_ADDR_WIDTH+3) := x.BurstLength;
    y(ADB3W_OCP_DATA_WIDTH-1 + ADB3_OCP_BURST_WIDTH+ADB3_OCP_ADDR_WIDTH+3 downto ADB3_OCP_BURST_WIDTH+ADB3_OCP_ADDR_WIDTH+3) := x.Data;
    y(ADB3W_OCP_DATA_WIDTH/8-1 + ADB3W_OCP_DATA_WIDTH+ADB3_OCP_BURST_WIDTH+ADB3_OCP_ADDR_WIDTH+3 downto ADB3W_OCP_DATA_WIDTH+ADB3_OCP_BURST_WIDTH+ADB3_OCP_ADDR_WIDTH+3) := x.DataByteEn;
    y(ADB3W_OCP_DATA_WIDTH/8+ADB3W_OCP_DATA_WIDTH+ADB3_OCP_BURST_WIDTH+ADB3_OCP_ADDR_WIDTH+3) := x.DataValid;
    y(1+ADB3W_OCP_DATA_WIDTH/8+ADB3W_OCP_DATA_WIDTH+ADB3_OCP_BURST_WIDTH+ADB3_OCP_ADDR_WIDTH+3) := x.RespAccept;
    y(ADB3_OCP_TAG_WIDTH-1 + 1+1+ADB3W_OCP_DATA_WIDTH/8+ADB3W_OCP_DATA_WIDTH+ADB3_OCP_BURST_WIDTH+ADB3_OCP_ADDR_WIDTH+3 downto 1+1+ADB3W_OCP_DATA_WIDTH/8+ADB3W_OCP_DATA_WIDTH+ADB3_OCP_BURST_WIDTH+ADB3_OCP_ADDR_WIDTH+3) := x.Tag;
    return y;
  end function to_std_logic_vector;

  function to_std_logic_vector(x : adb3w_ocp_s2mT) return std_logic_vector is
    variable y : std_logic_vector(adb3w_ocp_s2mT_slv_length-1 downto 0);
  begin
    y(0) := x.CmdAccept;
    y(1) := x.DataAccept;
    y(ADB3W_OCP_DATA_WIDTH-1+2 downto 2) := x.Data;
    y(1 + ADB3W_OCP_DATA_WIDTH+2 downto ADB3W_OCP_DATA_WIDTH+2) := x.Resp;
    y(ADB3_OCP_TAG_WIDTH-1 + 2+ADB3W_OCP_DATA_WIDTH+2 downto 2+ADB3W_OCP_DATA_WIDTH+2) := x.Tag;
    return y;
  end function to_std_logic_vector;
  
  function to_std_logic_vector(x : adb3w_ocp_m2sT_array) return std_logic_vector is
    variable y : std_logic_vector(adb3w_ocp_m2sT_slv_length*x'length-1 downto 0);
  begin
    for i in 0 to x'length-1 loop
      y((i+1)*adb3w_ocp_m2sT_slv_length-1 downto i*adb3w_ocp_m2sT_slv_length) := to_std_logic_vector(x(i));
    end loop;  -- i
    return y;
  end function to_std_logic_vector;


  function to_std_logic_vector(x : adb3w_ocp_s2mT_array) return std_logic_vector is
    variable y : std_logic_vector(adb3w_ocp_s2mT_slv_length*x'length-1 downto 0);
  begin
    for i in 0 to x'length-1 loop
      y((i+1)*adb3w_ocp_s2mT_slv_length-1 downto i*adb3w_ocp_s2mT_slv_length) := to_std_logic_vector(x(i));
    end loop;  -- i
    return y;
  end function to_std_logic_vector;
    

function to_adb3w_ocp_m2sT(x : std_logic_vector) return adb3w_ocp_m2sT is
    variable y : adb3w_ocp_m2sT;
  begin
    y.Cmd := x(2 downto 0);
    y.Addr := x(ADB3_OCP_ADDR_WIDTH-1 + 3 downto 3);
    y.BurstLength := x(ADB3_OCP_BURST_WIDTH-1 + ADB3_OCP_ADDR_WIDTH+3 downto ADB3_OCP_ADDR_WIDTH+3);
    y.Data := x(ADB3W_OCP_DATA_WIDTH-1 + ADB3_OCP_BURST_WIDTH+ADB3_OCP_ADDR_WIDTH+3 downto ADB3_OCP_BURST_WIDTH+ADB3_OCP_ADDR_WIDTH+3);
    y.DataByteEn := x(ADB3W_OCP_DATA_WIDTH/8-1 + ADB3W_OCP_DATA_WIDTH+ADB3_OCP_BURST_WIDTH+ADB3_OCP_ADDR_WIDTH+3 downto ADB3W_OCP_DATA_WIDTH+ADB3_OCP_BURST_WIDTH+ADB3_OCP_ADDR_WIDTH+3);
    y.DataValid := x(ADB3W_OCP_DATA_WIDTH/8+ADB3W_OCP_DATA_WIDTH+ADB3_OCP_BURST_WIDTH+ADB3_OCP_ADDR_WIDTH+3);
    y.RespAccept := x(1+ADB3W_OCP_DATA_WIDTH/8+ADB3W_OCP_DATA_WIDTH+ADB3_OCP_BURST_WIDTH+ADB3_OCP_ADDR_WIDTH+3);
    y.Tag := x(ADB3_OCP_TAG_WIDTH-1 + 1+1+ADB3W_OCP_DATA_WIDTH/8+ADB3W_OCP_DATA_WIDTH+ADB3_OCP_BURST_WIDTH+ADB3_OCP_ADDR_WIDTH+3 downto 1+1+ADB3W_OCP_DATA_WIDTH/8+ADB3W_OCP_DATA_WIDTH+ADB3_OCP_BURST_WIDTH+ADB3_OCP_ADDR_WIDTH+3);
    return y;      
  end function to_adb3w_ocp_m2sT;

  function to_adb3w_ocp_s2mT(x : std_logic_vector) return adb3w_ocp_s2mT is
    variable y : adb3w_ocp_s2mT;
  begin
    y.CmdAccept := x(0);
    y.DataAccept := x(1);
    y.Data := x(ADB3W_OCP_DATA_WIDTH-1+2 downto 2);
    y.Resp := x(1 + ADB3W_OCP_DATA_WIDTH+2 downto ADB3W_OCP_DATA_WIDTH+2);
    y.Tag := x(ADB3_OCP_TAG_WIDTH-1 + 2+ADB3W_OCP_DATA_WIDTH+2 downto 2+ADB3W_OCP_DATA_WIDTH+2);
    return y;
  end function to_adb3w_ocp_s2mT;
    
  function to_adb3w_ocp_m2sT_array(x : std_logic_vector) return adb3w_ocp_m2sT_array is
    variable y : adb3w_ocp_m2sT_array(0 to (x'length/adb3w_ocp_m2sT_slv_length) -1);
    variable x_i : std_logic_vector(adb3w_ocp_m2sT_slv_length-1 downto 0); 
  begin
    for i in 0 to (x'length/adb3w_ocp_m2sT_slv_length)-1 loop
      x_i := x((i+1)*adb3w_ocp_m2sT_slv_length-1 downto i*adb3w_ocp_m2sT_slv_length);
      y(i) := to_adb3w_ocp_m2sT(x_i);
    end loop;  -- i
    return y;
  end function to_adb3w_ocp_m2sT_array;

  function to_adb3w_ocp_s2mT_array(x : std_logic_vector) return adb3w_ocp_s2mT_array is
    variable y : adb3w_ocp_s2mT_array(0 to (x'length/adb3w_ocp_s2mT_slv_length) -1);
    variable x_i : std_logic_vector(adb3w_ocp_s2mT_slv_length-1 downto 0); 
  begin
    for i in 0 to (x'length/adb3w_ocp_s2mT_slv_length)-1 loop
      x_i := x((i+1)*adb3w_ocp_s2mT_slv_length-1 downto i*adb3w_ocp_s2mT_slv_length);
      y(i) := to_adb3w_ocp_s2mT(x_i);
    end loop;  -- i
    return y;
  end function to_adb3w_ocp_s2mT_array;


    function to_adb3l_ocp_m2sT(x : std_logic_vector) return adb3l_ocp_m2sT is
    variable y : adb3l_ocp_m2sT;
  begin
    y.Cmd := x(2 downto 0);
    y.Addr := x(ADB3L_OCP_ADDR_WIDTH-1 + 3 downto 3); 
    y.Data := x(ADB3L_OCP_DATA_WIDTH-1 +ADB3L_OCP_ADDR_WIDTH+3 downto ADB3L_OCP_ADDR_WIDTH+3);
    y.DataByteEn := x(ADB3L_OCP_DATA_WIDTH/8-1 + ADB3L_OCP_DATA_WIDTH+ADB3L_OCP_ADDR_WIDTH+3 downto ADB3L_OCP_DATA_WIDTH+ADB3L_OCP_ADDR_WIDTH+3);
    y.DataValid := x(ADB3L_OCP_DATA_WIDTH/8+ADB3L_OCP_DATA_WIDTH+ADB3L_OCP_ADDR_WIDTH+3);
    y.RespAccept := x(1+ADB3L_OCP_DATA_WIDTH/8+ADB3L_OCP_DATA_WIDTH+ADB3L_OCP_ADDR_WIDTH+3);
    return y;      
  end function to_adb3l_ocp_m2sT;

  function to_adb3l_ocp_s2mT(x : std_logic_vector) return adb3l_ocp_s2mT is
    variable y : adb3l_ocp_s2mT;
  begin
    y.CmdAccept := x(0);
    y.DataAccept := x(1);
    y.Data := x(ADB3L_OCP_DATA_WIDTH-1+2 downto 2);
    y.Resp := x(1 + ADB3L_OCP_DATA_WIDTH+2 downto ADB3L_OCP_DATA_WIDTH+2);
    return y;
  end function to_adb3l_ocp_s2mT;
  
end package body adb3_ocp;

