--
-- adb3_ocp_reorder_base.vhd
--
-- Bridging component that reorders responses in an unordered stream of OCP
-- responses to match the order in which the OCP read commands that generated
-- them were received on the in-order side.
--
-- On the unordered side, OCP responses are assumed to be a stream where
-- responses (data & tag) for any in-flight read command can be interleaved
-- with responses for any other in-flight read command, with no ordering
-- guarantees.
--
-- On the in-order side, OCP responses are delivered in the order dictated by
-- the order of OCP read commands as they were received by the in-order side.
--

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_misc.all;
use ieee.numeric_std.all;
use ieee.math_real.all;

library work;
use work.adb3_ocp.all;

entity adb3_ocp_reorder_base is
  generic(
    -- Width of a data word, in bits
    data_width         : in natural;
    -- Log2 of max. number of OCP words in a read command accepted on in-order
    -- side.
    buffer_order       : in natural;
    -- Maximum number of reads in flight issued by in-order side OCP master.
    max_read_in_flight : in natural;
    -- If true, delays CmdAccept on in-order side when read in flight limit
    -- reached.
    enforce_max_read   : in boolean;
    -- If true, data buffer is adb3_ocp_reorder_dbuf_ur; otherwise
    -- adb3_ocp_reorder_dbuf_mf.
    unified_ram        : in boolean;
    -- If true, instantiates extra pipelining in DBUF output to mitigate logic
    -- complexity of multiplexors or large Tco of BlockRAMS.
    output_pipe        : in boolean);
  port(
    rst         : in  std_logic; -- Async. reset (tie to 0 if unused)
    sr          : in  std_logic; -- Sync. reset (tie to 0 if unused)
    clk         : in  std_logic; -- OCP clock
    -- In-order side, OCP slave port
    io_m2s_cmd  : in  ocp_CmdT;
    io_m2s_addr : in  std_logic_vector(ADB3_OCP_ADDR_WIDTH - 1 downto 0);
    io_m2s_blen : in  std_logic_vector(ADB3_OCP_BURST_WIDTH - 1 downto 0);
    io_m2s_data : in  std_logic_vector(data_width - 1 downto 0);
    io_m2s_be   : in  std_logic_vector(data_width / 8 - 1 downto 0);
    io_m2s_dval : in  std_logic;
    io_m2s_racc : in  std_logic;
    io_m2s_tag  : in  std_logic_vector(ADB3_OCP_TAG_WIDTH - 1 downto 0);
    io_s2m_cacc : out std_logic;
    io_s2m_dacc : out std_logic;
    io_s2m_data : out std_logic_vector(data_width - 1 downto 0);
    io_s2m_resp : out ocp_RespT;
    io_s2m_tag  : out std_logic_vector(ADB3_OCP_TAG_WIDTH - 1 downto 0);
    -- Unordered side, OCP master port
    uo_m2s_cmd  : out ocp_CmdT;
    uo_m2s_addr : out std_logic_vector(ADB3_OCP_ADDR_WIDTH - 1 downto 0);
    uo_m2s_blen : out std_logic_vector(ADB3_OCP_BURST_WIDTH - 1 downto 0);
    uo_m2s_data : out std_logic_vector(data_width - 1 downto 0);
    uo_m2s_be   : out std_logic_vector(data_width / 8 - 1 downto 0);
    uo_m2s_dval : out std_logic;
    uo_m2s_racc : out std_logic;
    uo_m2s_tag  : out std_logic_vector(ADB3_OCP_TAG_WIDTH - 1 downto 0);
    uo_s2m_cacc : in  std_logic;
    uo_s2m_dacc : in  std_logic;
    uo_s2m_data : in  std_logic_vector(data_width - 1 downto 0);
    uo_s2m_resp : in  ocp_RespT;
    uo_s2m_tag  : in  std_logic_vector(ADB3_OCP_TAG_WIDTH - 1 downto 0);
    -- Status fields
    --    [7:0]  Number of read commands currently in flight
    --    [9:0]  RFSM state: 00 => IDLE, 01 => READING
    --      [10] Command FIFO 'valid' flag
    --      [11] Command FIFO 'full' flag
    -- Following 3 fields valid only when RFSM state is READING:
    --   [19:12] Tag of current in-order side read
    --   [31:20] Burst length remaining in current in-order side read
    --   [35:32] Slot number associated with current in-order side read
    --   [47:36] (reserved)
    -- Error flags in bits [63:48]
    --      [48] In-order side OCP master attempted to exceed
    --           max_read_in_flight
    --      [49] Attempt to allocate and free a particular slot in same clock
    --           cycle
    --      [50] A read command was accepted but there was no free slot to
    --           store it
    --      [51] A read command was accepted with burst length >
    --           2**buffer_order
    --      [52] The command FIFO overflowed
    --      [53] The command FIFO underflowed
    --      [54] A response was received on the unordered side whose tag could
    --           not be matched to any slot.
    --      [55] Overflow detected in a slot's data buffer
    --      [56] Underflow detected in a slot's data buffer
    --   [63:57] (reserved)
    status    : out std_logic_vector(63 downto 0);
    -- A '1' in any bit clears corresponding 'error' bit in status[63:48]
    clr_error : in  std_logic_vector(15 downto 0));
end entity;

architecture rtl of adb3_ocp_reorder_base is

  function maximum(
    constant a, b : natural)
  return natural is
  begin
    if a > b then
      return a;
    else
      return b;
    end if;
  end;
  
  -- Convert a one-hot vector to a binary-encoded index
  function one_hot_to_index(
    constant width : natural;
    constant value_1h : std_logic_vector)
  return std_logic_vector is
    variable tmp : std_logic_vector(width - 1 downto 0);
  begin
    tmp := (others => '0');
    for i in value_1h'range loop
      if value_1h(i) = '1' then
        tmp := tmp OR std_logic_vector(to_unsigned(i, width));
      end if;
    end loop;
    return tmp;
  end;
  
  -- Convert a binary-coded index to a one-hot vector
  function index_to_one_hot(
    constant width : natural;
    constant value_binary : std_logic_vector)
  return std_logic_vector is
    variable tmp : std_logic_vector(width - 1 downto 0);
  begin
    for i in 0 to width - 1 loop
      if value_binary'length = 0 OR value_binary = i then
        tmp(i) := '1';
      else
        tmp(i) := '0';
      end if;
    end loop;
    return tmp;
  end;
  
  -- Return '1' if less than or equal to 2
  function less_than_equal_to_two(
    constant value_binary : std_logic_vector)
  return std_logic is
    variable tmp : std_logic_vector(value_binary'length - 1 downto 0);
  begin
    tmp := value_binary;
    return NOT OR_reduce(tmp(tmp'high downto 2) & tmp(0));
  end;
  
  -- Return '1' if less than or equal to 1
  function less_than_equal_to_one(
    constant value_binary : std_logic_vector)
  return std_logic is
    variable tmp : std_logic_vector(value_binary'length - 1 downto 0);
  begin
    tmp := value_binary;
    return NOT OR_reduce(tmp(tmp'high downto 1));
  end;
  
  component adb3_ocp_cmd_retime
  port(
    rst       : in  std_logic;
    sr        : in  std_logic;
    clk       : in  std_logic;
    s_cmd     : in  ocp_cmdT;
    s_addr    : in  std_logic_vector(ADB3_OCP_ADDR_WIDTH-1 downto 0);
    s_blen    : in  std_logic_vector(ADB3_OCP_BURST_WIDTH-1 downto 0);
    s_tag     : in  std_logic_vector(ADB3_OCP_TAG_WIDTH-1 downto 0);
    s_cmdacc  : out std_logic;
    m_cmd     : out ocp_cmdT;
    m_addr    : out std_logic_vector(ADB3_OCP_ADDR_WIDTH-1 downto 0);
    m_blen    : out std_logic_vector(ADB3_OCP_BURST_WIDTH-1 downto 0);
    m_tag     : out std_logic_vector(ADB3_OCP_TAG_WIDTH-1 downto 0);
    m_cmdacc  : in  std_logic);
  end component;
  
  component adb3_ocp_reorder_sfifo
  generic(
    ORDER       : in    natural range 2 to 6;
    WIDTH       : in    natural;          
    FAST_SR     : in    boolean := false);
  port(
    rst         : in    std_logic;
    clk         : in    std_logic;
    sr          : in    std_logic;
    w           : in    std_logic;
    wd          : in    std_logic_vector(WIDTH - 1 downto 0);
    wfull       : out   std_logic;
    r           : in    std_logic;
    rvalid      : out   std_logic;
    rq          : out   std_logic_vector(WIDTH - 1 downto 0));
  end component;
  
  component adb3_ocp_reorder_dbuf_mf
  generic(
    data_width      : in  natural;
    num_slot        : in  natural;
    slot_order      : in  natural;
    output_pipe     : in  boolean);
  port(
    rst             : in  std_logic;
    sr              : in  std_logic;
    clk             : in  std_logic;
    uo_resp         : in  ocp_RespT;
    uo_data         : in  std_logic_vector(data_width - 1 downto 0);
    uo_tag          : in  std_logic_vector(ADB3_OCP_TAG_WIDTH - 1 downto 0);
    uo_resp_accept  : out std_logic;
    read_slot_1hot  : in  std_logic_vector(num_slot - 1 downto 0);
    read_slot_valid : in  std_logic;
    read            : in  std_logic;
    read_data       : out std_logic_vector(data_width - 1 downto 0);
    read_valid      : out std_logic;
    alloc           : in  std_logic;
    alloc_tag       : in  std_logic_vector(ADB3_OCP_TAG_WIDTH - 1 downto 0);
    alloc_length    : in  std_logic_vector(ADB3_OCP_BURST_WIDTH - 1 downto 0);
    alloc_slot_1hot : in  std_logic_vector(num_slot - 1 downto 0);
    free            : in  std_logic;
    free_slot_1hot  : in  std_logic_vector(num_slot - 1 downto 0);
    error_pulse     : out std_logic_vector(7 downto 0));
  end component;

  component adb3_ocp_reorder_dbuf_ur
  generic(
    data_width      : in  natural;
    num_slot        : in  natural;
    slot_order      : in  natural;
    output_pipe     : in  boolean);
  port(
    rst             : in  std_logic;
    sr              : in  std_logic;
    clk             : in  std_logic;
    uo_resp         : in  ocp_RespT;
    uo_data         : in  std_logic_vector(data_width - 1 downto 0);
    uo_tag          : in  std_logic_vector(ADB3_OCP_TAG_WIDTH - 1 downto 0);
    uo_resp_accept  : out std_logic;
    read_slot_1hot  : in  std_logic_vector(num_slot - 1 downto 0);
    read_slot_valid : in  std_logic;
    read            : in  std_logic;
    read_data       : out std_logic_vector(data_width - 1 downto 0);
    read_valid      : out std_logic;
    alloc           : in  std_logic;
    alloc_tag       : in  std_logic_vector(ADB3_OCP_TAG_WIDTH - 1 downto 0);
    alloc_length    : in  std_logic_vector(ADB3_OCP_BURST_WIDTH - 1 downto 0);
    alloc_slot_1hot : in  std_logic_vector(num_slot - 1 downto 0);
    free            : in  std_logic;
    free_slot_1hot  : in  std_logic_vector(num_slot - 1 downto 0);
    error_pulse     : out std_logic_vector(7 downto 0));
  end component;
  
  -- Maximum burst length of any read command accepted from in-order OCP
  -- master
  constant max_read_length : natural := 2**buffer_order;
  
  -- -------------------------------------------------------------------------
  -- Slot management (SM) signals.
  -- -------------------------------------------------------------------------
  
  signal sm_slots_occupied : std_logic_vector(max_read_in_flight - 1 downto 0) := (others => '0');
  signal sm_free_slot_1hot : std_logic_vector(max_read_in_flight - 1 downto 0) := (others => '0');
  signal sm_slots_not_full : std_logic := '0';
  signal sm_allow_command : std_logic := '0';
  
  -- Asserted when a word of response data has been matched to a slot
  signal tag_slot_valid : std_logic; -- Qualifies "tag_slot_1hot", "resp_tag_q", "resp_data_q"
  -- Indicates to which slot a word of response data has been matched
  signal tag_slot_1hot : std_logic_vector(max_read_in_flight - 1 downto 0);
  
  -- -------------------------------------------------------------------------
  -- Command FIFO (CFIFO) signals
  -- -------------------------------------------------------------------------
  
  -- Data format is
  -- [             7 :             0] Tag
  -- [ CFIFO_ORDER+7 :             8] Slot number
  -- [CFIFO_ORDER+19 : CFIFO_ORDER+8] Burst length
  constant CFIFO_ORDER : natural := natural(floor(log2(real(2 * max_read_in_flight - 1) + 0.5)));
  constant CFIFO_WIDTH : natural := CFIFO_ORDER + ADB3_OCP_TAG_WIDTH + ADB3_OCP_BURST_WIDTH;
  signal cfifo_write_data : std_logic_vector(CFIFO_WIDTH - 1 downto 0);
  signal cfifo_write : std_logic;
  signal cfifo_full : std_logic;
  signal cfifo_read_data : std_logic_vector(CFIFO_WIDTH - 1 downto 0);
  signal cfifo_read_blen : std_logic_vector(ADB3_OCP_BURST_WIDTH - 1 downto 0);
  signal cfifo_read_tag : std_logic_vector(ADB3_OCP_TAG_WIDTH - 1 downto 0);
  signal cfifo_read_slot : std_logic_vector(CFIFO_ORDER - 1 downto 0);
  signal cfifo_read : std_logic;
  signal cfifo_valid : std_logic;
  
  -- -------------------------------------------------------------------------
  -- Data Buffer (DBUF) signals
  -- -------------------------------------------------------------------------
  
  -- Indicates DBUF driving valid data on 'dbuf_read_data'
  signal dbuf_read_valid : std_logic;
  -- Asserted to read next word from DBUF
  signal dbuf_read : std_logic; -- Qualifies "dbuf_read_pos"
  -- DBUF read data.
  signal dbuf_read_data : std_logic_vector(data_width - 1 downto 0);
  -- DBUF slot management
  signal dbuf_alloc : std_logic;
  signal dbuf_alloc_tag : std_logic_vector(ADB3_OCP_TAG_WIDTH - 1 downto 0);
  signal dbuf_alloc_length : std_logic_vector(ADB3_OCP_BURST_WIDTH - 1 downto 0);
  signal dbuf_alloc_slot_1hot : std_logic_vector(max_read_in_flight - 1 downto 0);
  signal dbuf_free : std_logic;
  signal dbuf_free_slot_1hot : std_logic_vector(max_read_in_flight - 1 downto 0);
  signal dbuf_error_pulse : std_logic_vector(7 downto 0);
  
  -- -------------------------------------------------------------------------
  -- Read FSM (RFSM) signals
  -- -------------------------------------------------------------------------
  
  type rfsm_state_t is (rfsm_state_idle, rfsm_state_reading);
  signal rfsm_state, n_rfsm_state : rfsm_state_t := rfsm_state_idle;
  signal rfsm_blen, n_rfsm_blen : std_logic_vector(ADB3_OCP_BURST_WIDTH - 1 downto 0);
  signal rfsm_blen_last, n_rfsm_blen_last : std_logic;
  signal rfsm_slot_valid : std_logic := '0';
  signal n_rfsm_slot_valid : std_logic;
  signal rfsm_slot, n_rfsm_slot : std_logic_vector(CFIFO_ORDER - 1 downto 0);
  signal rfsm_slot_1hot, n_rfsm_slot_1hot : std_logic_vector(max_read_in_flight - 1 downto 0);
  signal rfsm_tag, n_rfsm_tag : std_logic_vector(ADB3_OCP_TAG_WIDTH - 1 downto 0);
    
  -- -------------------------------------------------------------------------
  -- Misc. signals
  -- -------------------------------------------------------------------------
  
  -- Asserted when a command from in-order side is accepted
  signal read_cmd_accepted : std_logic;
  -- Asserted when the last read response for a command side is accepted is by the in-order side
  signal read_cmd_completed : std_logic;
  -- Asserted when read data is accepted by in-order side
  signal io_resp_accepted : std_logic;
  -- Asserted when read data from unordered side is accepted
  signal uo_resp_accepted : std_logic;
  
  -- -------------------------------------------------------------------------
  -- Debug/status signals
  -- -------------------------------------------------------------------------
  
  -- Read in flight counter, only used for debug/status.
  signal num_rd_in_flight : std_logic_vector(7 downto 0);
  
  -- Internal signal for driving 'status' & 'error' ports
  signal i_status : std_logic_vector(47 downto 0) := (others => '0');
  signal i_error : std_logic_vector(15 downto 0) := (others => '0');

  -- Internal version of unordered side OCP bus; command path may be retimed
  signal i_uo_m2s_cmd : ocp_CmdT;
  signal i_uo_m2s_addr : std_logic_vector(ADB3_OCP_ADDR_WIDTH - 1 downto 0);
  signal i_uo_m2s_blen : std_logic_vector(ADB3_OCP_BURST_WIDTH - 1 downto 0);
  signal i_uo_m2s_data : std_logic_vector(data_width - 1 downto 0);
  signal i_uo_m2s_be : std_logic_vector(data_width / 8 - 1 downto 0);
  signal i_uo_m2s_dval : std_logic;
  signal i_uo_m2s_racc : std_logic;
  signal i_uo_m2s_tag : std_logic_vector(ADB3_OCP_TAG_WIDTH - 1 downto 0);
  signal i_uo_s2m_cacc : std_logic;
  signal i_uo_s2m_dacc : std_logic;
  signal i_uo_s2m_data : std_logic_vector(data_width - 1 downto 0);
  signal i_uo_s2m_resp : ocp_RespT;
  signal i_uo_s2m_tag : std_logic_vector(ADB3_OCP_TAG_WIDTH - 1 downto 0);
  
begin

  -- -------------------------------------------------------------------------
  -- Debug/status
  -- -------------------------------------------------------------------------
  
  -- Reads in flight counter
  count_read_in_flight : process(rst, clk)
    variable tmp : std_logic_vector(1 downto 0);
  begin
    if rst = '1' then
      num_rd_in_flight <= (others => '0');
    elsif rising_edge(clk) then
      if sr = '1' then
        num_rd_in_flight <= (others => '0');
      else
        tmp(0) := read_cmd_accepted;
        tmp(1) := read_cmd_completed;
        case tmp is
          when "00" | "11" =>
            null;
            
          when "01" =>
            num_rd_in_flight <= num_rd_in_flight + 1;
          
          when "10" => 
            num_rd_in_flight <= num_rd_in_flight - 1;
            
          when others =>
            -- Illegal
            num_rd_in_flight <= (others => '-');
        end case;
      end if;
    end if;
  end process;
  
  -- Generate status vector
  gen_status : process(rst, clk)
  begin
    if rising_edge(clk) then
      i_status(7 downto 0) <= num_rd_in_flight;
      
      case rfsm_state is
        when rfsm_state_idle =>
          i_status(9 downto 8) <= "00";
          
        when rfsm_state_reading =>
          i_status(9 downto 8) <= "01";
      end case;
      
      i_status(10) <= cfifo_valid;
      i_status(11) <= cfifo_full;
      
      i_status(19 downto 12) <= rfsm_tag;
      i_status(31 downto 20) <= rfsm_blen;
      i_status(35 downto 32) <= std_logic_vector(resize(unsigned(rfsm_slot), 4));

      status <= i_error & i_status;
    end if;
    
    if rst = '1' then
      i_error(8 downto 6) <= "000";
    elsif rising_edge(clk) then
      if sr = '1' then
        i_error(8 downto 6) <= "000";
      else
        i_error(8 downto 6) <= (i_error(8 downto 6) AND NOT clr_error(8 downto 6)) OR dbuf_error_pulse(2 downto 0);
      end if;
    end if;
  end process;
  
  
  -- -------------------------------------------------------------------------
  -- Misc.
  -- -------------------------------------------------------------------------
  
  -- If enforce_max_read, logic complexity of in-order OCP command path is
  -- increased, so instantiate something to retime it.
  retime_uo_cmd : if enforce_max_read generate
    U0 : adb3_ocp_cmd_retime
    port map(
      rst => rst,
      sr => sr,
      clk => clk,
      m_cmd => uo_m2s_cmd,
      m_addr => uo_m2s_addr,
      m_blen => uo_m2s_blen,
      m_tag => uo_m2s_tag,
      m_cmdacc => uo_s2m_cacc,
      s_cmd => i_uo_m2s_cmd,
      s_addr => i_uo_m2s_addr,
      s_blen => i_uo_m2s_blen,
      s_tag => i_uo_m2s_tag,
      s_cmdacc => i_uo_s2m_cacc);
      
    uo_m2s_data <= i_uo_m2s_data;
    uo_m2s_be <= i_uo_m2s_be;
    uo_m2s_dval <= i_uo_m2s_dval;
    uo_m2s_racc <= i_uo_m2s_racc;
    i_uo_s2m_data <= uo_s2m_data;
    i_uo_s2m_tag <= uo_s2m_tag;
    i_uo_s2m_resp <= uo_s2m_resp;
    i_uo_s2m_dacc <= uo_s2m_dacc;
  end generate;
  
  -- If NOT enforce_max_read, trivially pass i_uo_m2s -> uo_m2 and uo_s2m ->
  -- i_uo_s2ms (no retiming).
  no_retime_uo_cmd : if NOT enforce_max_read generate
    uo_m2s_cmd <= i_uo_m2s_cmd;
    uo_m2s_addr <= i_uo_m2s_addr;
    uo_m2s_blen <= i_uo_m2s_blen;
    uo_m2s_data <= i_uo_m2s_data;
    uo_m2s_be <= i_uo_m2s_be;
    uo_m2s_dval <= i_uo_m2s_dval;
    uo_m2s_racc <= i_uo_m2s_racc;
    uo_m2s_tag <= i_uo_m2s_tag;    
    i_uo_s2m_cacc <= uo_s2m_cacc;
    i_uo_s2m_dacc <= uo_s2m_dacc;
    i_uo_s2m_data <= uo_s2m_data;
    i_uo_s2m_resp <= uo_s2m_resp;
    i_uo_s2m_tag <= uo_s2m_tag;
  end generate;
  
  -- Pass commands from in-order M2S to unordered M2S
  i_uo_m2s_cmd <= io_m2s_cmd when sm_allow_command = '1' else OCP_CMD_IDLE;
  i_uo_m2s_addr <= io_m2s_addr;
  i_uo_m2s_tag <= io_m2s_tag;
  i_uo_m2s_blen <= io_m2s_blen;
    
  -- Pass write data unmodified from in-order M2S to unordered M2S
  i_uo_m2s_data <= io_m2s_data;
  i_uo_m2s_be <= io_m2s_be;
  i_uo_m2s_dval <= io_m2s_dval;
  
  -- Pass command accept & data accept from unordered S2M to in-order S2M
  io_s2m_cacc <= i_uo_s2m_cacc AND sm_allow_command;
  io_s2m_dacc <= i_uo_s2m_dacc;
  
  -- Generate some handy flags
  read_cmd_accepted <= '1' when io_m2s_cmd = OCP_CMD_READ AND sm_allow_command = '1' AND i_uo_s2m_cacc = '1' else
                       '0';
  io_resp_accepted <= '1' when dbuf_read_valid = '1' AND rfsm_state = rfsm_state_reading AND io_m2s_racc = '1' else
                      '0';
  uo_resp_accepted <= '1' when i_uo_m2s_racc = '1' AND i_uo_s2m_resp = OCP_RESP_VALID else
                      '0';
  read_cmd_completed <= '1' when dbuf_read_valid = '1' AND rfsm_state = rfsm_state_reading AND io_m2s_racc = '1' AND rfsm_blen_last = '1' else
                       '0';
  
  
  -- -------------------------------------------------------------------------
  -- Slot management.
  -- -------------------------------------------------------------------------
  
  -- If enforce_max_read is true, delay CmdAccept to in-order OCP master when
  -- number of reads has reached the limit.
  sm_allow_command <= sm_slots_not_full when enforce_max_read else
                      '1';
  
  -- Generate free_slot_1hot; represents slot to be occupied by next read
  -- command accepted on in-order side.
  -- It is critical that this vector is immediately updated when a command is
  -- accepted, but not critical that it is immediately updated when a slot
  -- becomes free.    
  gen_sm_free_slot : process(clk,rst)
    -- Slots free in next clock cycle
    variable n_slots_free : std_logic_vector(max_read_in_flight - 1 downto 0);
  begin
    if rst = '1' then
      sm_free_slot_1hot <= (others => '0');
      sm_slots_not_full <= '0';
      i_error(0) <= '0';
    elsif rising_edge(clk) then
      if sr = '1' then
        sm_free_slot_1hot <= (others => '0');
        sm_slots_not_full <= '0';
        i_error(0) <= '0';
      else
        -- Find a free slot, taking into account that a slots might
        -- be freed or allocated in this clock cycle.
        n_slots_free := NOT sm_slots_occupied;
        if read_cmd_accepted = '1' then
          -- A slot being occupied in this clock cycle
          n_slots_free := n_slots_free AND NOT sm_free_slot_1hot;
        end if;
        if read_cmd_completed = '1' then
          -- A slot being freed in this clock cycle
          n_slots_free := n_slots_free OR rfsm_slot_1hot;
        end if;
        
        -- Priority encoder to find a free slot; could be done more
        -- efficiently.
        sm_free_slot_1hot(0) <= n_slots_free(0);
        for i in 1 to max_read_in_flight - 1 loop
          sm_free_slot_1hot(i) <= n_slots_free(i) AND NOT OR_reduce(n_slots_free(i - 1 downto 0));
        end loop;
        
        -- Used to prevent in-order side OCP master exceeding
        -- 'max_reads_in_flight'.
        sm_slots_not_full <= OR_reduce(n_slots_free);
        
        -- Sanity check that in-order side OCP master is honouring
        -- 'max_read_in_flight'
        if OR_reduce(sm_free_slot_1hot) = '0' AND io_m2s_cmd = OCP_CMD_READ then
          i_error(0) <= '1';
          assert false
            report "adb3_ocp_reorder_base: In-order side OCP master attempting to exceed max_read_in_flight(" & natural'image(max_read_in_flight) & ")"
            severity error;
        elsif clr_error(0) = '1' then
          i_error(0) <= '0';
        end if;
      end if;  
    end if;
  end process;
  
  -- Main slot management process; monitors in-order side of OCP port to
  -- determine when a command is received, and stores it in a free slot.
  -- Also monitors when the in-order side accepts responses so that it can mark
  -- a slot free when the read command associated with that slot has been
  -- completed.
  implement_slots : process(rst, clk)
    variable case_tmp : std_logic_vector(1 downto 0);
  begin
    if rst = '1' then
      sm_slots_occupied <= (others => '0');
      i_error(3 downto 1) <= "000";
    elsif rising_edge(clk) then
      for i in 0 to max_read_in_flight - 1 loop
        if sr = '1' then
          sm_slots_occupied(i) <= '0';
          i_error(3 downto 1) <= "000";
        else
          if read_cmd_accepted = '1' AND sm_free_slot_1hot(i) = '1' then
            case_tmp(0) := '1'; -- Read command accepted on unordered side
          else
            case_tmp(0) := '0';
          end if;
          if io_resp_accepted = '1' AND rfsm_slot_1hot(i) = '1' AND rfsm_blen_last = '1' then
            case_tmp(1) := '1'; -- Final response of command accepted on in-order side
          else
            case_tmp(1) := '0';
          end if;
          case case_tmp is
            when "01" => 
              -- Read command accepted from in-order side, so mark its slot as
              -- occupied and store its tag & burst length.
              sm_slots_occupied(i) <= '1';

            when "10" =>
              -- Read command finished on in-order side, so mark its slot as unoccupied
              sm_slots_occupied(i) <= '0';
              
            when "00" =>
              -- Nothing happening
              
            when others =>
              -- Illegal, should never happen
              sm_slots_occupied(i) <= '-';
          end case;
          
          if case_tmp = "11" then
            i_error(1) <= '1';
            assert false
              report "adb3_ocp_reorder_base: Slot " & natural'image(i) & ": attempt to occupy and free at the same time"
              severity error;
          elsif clr_error(1) = '1' then
            i_error(1) <= '0';
          end if;
          
          -- Complain if a command is accepted and there is no free slot in
          -- which to store it.
          if read_cmd_accepted = '1' AND OR_reduce(sm_free_slot_1hot) = '0' then
            i_error(2) <= '1';
--synthesis translate_off
            assert false
              report "adb3_ocp_reorder_base: Read command accepted but no free slot"
              severity error;
--synthesis translate_on
          elsif clr_error(2) = '1' then
            i_error(2) <= '0';
          end if;
          
          -- Complain if burst length of an accepted read command is more than
          -- 2**buffer_order.
          if read_cmd_accepted = '1' AND io_m2s_blen > max_read_length then
            i_error(3) <= '1';
--synthesis translate_off
            assert false
              report "adb3_ocp_reorder_base: Read command accepted with BurstLength(" & integer'image(conv_integer(io_m2s_blen)) & ") > 2**buffer_order(" & natural'image(max_read_length) & ")"
              severity error;
--synthesis translate_on
          elsif clr_error(3) = '1' then
            i_error(3) <= '0';
          end if;
        end if;
      end loop;
    end if;
  end process;

  
  -- -------------------------------------------------------------------------
  -- Command FIFO
  -- Stores commands accepted from in-order side, so that they can be popped
  -- and processed in-order.
  -- -------------------------------------------------------------------------
  
  gen_cfifo_write : process(rst, clk)
  begin
    if rising_edge(clk) then
      cfifo_write_data <= io_m2s_blen & one_hot_to_index(CFIFO_ORDER, sm_free_slot_1hot) & io_m2s_tag;
      cfifo_write <= read_cmd_accepted;
    end if;
    
    -- Sanity checks
    if rst = '1' then
      i_error(5 downto 4) <= "00";
    elsif rising_edge(clk) then
      if sr = '1' then
        i_error(5 downto 4) <= "00";
      else
        if cfifo_write = '1' AND cfifo_full = '1' then
          i_error(4) <= '1';
          assert false
            report "adb3_ocp_reorder_base: Command FIFO overflow"
            severity error;
        elsif clr_error(4) = '1' then
          i_error(4) <= '0';
        end if;
        
        if cfifo_read = '1' AND cfifo_valid = '0' then
          i_error(5) <= '1';
          assert false
            report "adb3_ocp_reorder_base: Command FIFO underflow"
            severity error;
        elsif clr_error(5) = '1' then
          i_error(5) <= '0';
        end if;
      end if;
    end if;
  end process;
  
  cfifo : adb3_ocp_reorder_sfifo
    generic map(
      WIDTH => CFIFO_WIDTH,
      ORDER => maximum(2, CFIFO_ORDER))
    port map(
      rst => rst,
      sr => sr,
      clk => clk,
      wd => cfifo_write_data,
      w => cfifo_write,
      wfull => cfifo_full,
      r => cfifo_read,
      rq => cfifo_read_data,
      rvalid => cfifo_valid);
  
  -- Break "cfifo_read_data" out into its fields
  cfifo_read_tag <= cfifo_read_data(ADB3_OCP_TAG_WIDTH - 1 downto 0);
  cfifo_read_slot <= cfifo_read_data(CFIFO_ORDER + ADB3_OCP_TAG_WIDTH - 1 downto ADB3_OCP_TAG_WIDTH);
  cfifo_read_blen <= cfifo_read_data(CFIFO_ORDER + ADB3_OCP_TAG_WIDTH + ADB3_OCP_BURST_WIDTH - 1 downto CFIFO_ORDER + ADB3_OCP_TAG_WIDTH);

  
  -- -------------------------------------------------------------------------
  -- Data buffer (DBUF)
  -- -------------------------------------------------------------------------
  
  dbuf_alloc <= read_cmd_accepted;
  dbuf_alloc_tag <= io_m2s_tag;
  dbuf_alloc_length <= io_m2s_blen;
  dbuf_alloc_slot_1hot <= sm_free_slot_1hot;
  
  dbuf_free <= read_cmd_completed AND rfsm_blen_last;
  dbuf_free_slot_1hot <= rfsm_slot_1hot;
  
  dbuf_ur : if unified_ram generate
    U0 : adb3_ocp_reorder_dbuf_ur
    generic map(
      data_width => data_width,
      num_slot => max_read_in_flight,
      slot_order => buffer_order,
      output_pipe => output_pipe)
    port map(
      rst => rst,
      sr => sr,
      clk => clk,
      uo_data => i_uo_s2m_data,
      uo_resp => i_uo_s2m_resp,
      uo_tag => i_uo_s2m_tag,
      uo_resp_accept => i_uo_m2s_racc,
      read_slot_1hot => rfsm_slot_1hot,
      read_slot_valid => rfsm_slot_valid,
      read => dbuf_read,
      read_data => dbuf_read_data,
      read_valid => dbuf_read_valid,
      alloc => dbuf_alloc,
      alloc_tag => dbuf_alloc_tag,
      alloc_length => dbuf_alloc_length,
      alloc_slot_1hot => dbuf_alloc_slot_1hot,
      free => dbuf_free,
      free_slot_1hot => dbuf_free_slot_1hot,
      error_pulse => dbuf_error_pulse);
  end generate;
       
  dbuf_mf : if NOT unified_ram generate
    U0 : adb3_ocp_reorder_dbuf_mf
    generic map(
      data_width => data_width,
      num_slot => max_read_in_flight,
      slot_order => buffer_order,
      output_pipe => output_pipe)
    port map(
      rst => rst,
      sr => sr,
      clk => clk,
      uo_data => i_uo_s2m_data,
      uo_resp => i_uo_s2m_resp,
      uo_tag => i_uo_s2m_tag,
      uo_resp_accept => i_uo_m2s_racc,
      read_slot_1hot => rfsm_slot_1hot,
      read_slot_valid => rfsm_slot_valid,
      read => dbuf_read,
      read_data => dbuf_read_data,
      read_valid => dbuf_read_valid,
      alloc => dbuf_alloc,
      alloc_tag => dbuf_alloc_tag,
      alloc_length => dbuf_alloc_length,
      alloc_slot_1hot => dbuf_alloc_slot_1hot,
      free => dbuf_free,
      free_slot_1hot => dbuf_free_slot_1hot,
      error_pulse => dbuf_error_pulse);
  end generate;
       
       
  -- -------------------------------------------------------------------------
  -- Response FSM
  -- Generates responses to in-order side, using commands from CFIFO
  -- and data from DBUF.
  -- -------------------------------------------------------------------------

  dbuf_read <= '1' when dbuf_read_valid = '1' AND rfsm_state = rfsm_state_reading AND io_m2s_racc = '1' else
               '0';
  
  io_s2m_resp <= OCP_RESP_VALID when dbuf_read_valid = '1' AND rfsm_state = rfsm_state_reading else
                 OCP_RESP_NONE;
  io_s2m_data <= dbuf_read_data;
  io_s2m_tag <= rfsm_tag;
  
  response_fsm_state : process(rst, clk)
  begin
    if rst = '1' then
      rfsm_state <= rfsm_state_idle;
      rfsm_blen <= (others => '-');
      rfsm_blen_last <= '-';
      rfsm_slot_valid <= '0';
      rfsm_slot <= (others => '-');
      rfsm_slot_1hot <= (others => '-');
      rfsm_tag <= (others => '-');
    elsif rising_edge(clk) then
      if sr = '1' then
        rfsm_state <= rfsm_state_idle;
        rfsm_blen <= (others => '-');
        rfsm_blen_last <= '-';
        rfsm_slot_valid <= '0';
        rfsm_slot <= (others => '-');
        rfsm_slot_1hot <= (others => '-');
        rfsm_tag <= (others => '-');
      else
        rfsm_state <= n_rfsm_state;
        rfsm_blen <= n_rfsm_blen;
        rfsm_blen_last <= n_rfsm_blen_last;
        rfsm_slot_valid <= n_rfsm_slot_valid;
        rfsm_slot <= n_rfsm_slot;
        rfsm_slot_1hot <= n_rfsm_slot_1hot;
        rfsm_tag <= n_rfsm_tag;
      end if;
    end if;
  end process;
  
  cfifo_read <= '1' when rfsm_state = rfsm_state_idle AND cfifo_valid = '1' else
                '0';
  
  response_fsm_next_state : process(
    rfsm_state,
    dbuf_read,
    rfsm_blen,
    rfsm_slot,
    rfsm_slot_1hot,
    rfsm_blen_last,
    cfifo_valid,
    cfifo_read_tag,
    cfifo_read_slot,
    cfifo_read_blen,
    rfsm_tag
    )
  begin
    case rfsm_state is
      when rfsm_state_idle =>
        if cfifo_valid = '1' then
          n_rfsm_state <= rfsm_state_reading;
          n_rfsm_blen <= cfifo_read_blen;
          n_rfsm_blen_last <= less_than_equal_to_one(cfifo_read_blen);
          n_rfsm_slot_valid <= '1';
          n_rfsm_slot <= cfifo_read_slot;
          -- Without to_01(..), can get warning of X in arithmetic operand due to delta cycles in simulation.
          n_rfsm_slot_1hot <= index_to_one_hot(max_read_in_flight, std_logic_vector(to_01(unsigned(cfifo_read_slot))));
          n_rfsm_tag <= cfifo_read_tag;
        else
          n_rfsm_state <= rfsm_state_idle;
          n_rfsm_blen <= (others => '-');
          n_rfsm_blen_last <= '-';
          n_rfsm_slot_valid <= '0';
          n_rfsm_slot <= (others => '-');
          n_rfsm_slot_1hot <= (others => '-');
          n_rfsm_tag <= (others => '-');
        end if;
        
      when rfsm_state_reading =>
        if dbuf_read = '1' then
          if rfsm_blen_last = '1' then
            n_rfsm_state <= rfsm_state_idle;
            n_rfsm_blen <= (others => '-');
            n_rfsm_blen_last <= '-';
            n_rfsm_slot_valid <= '0'; -- Must go to zero for at least a cycle when slot changes
            n_rfsm_slot <= (others => '-');
            n_rfsm_slot_1hot <= (others => '-');
            n_rfsm_tag <= (others => '-');
          else
            n_rfsm_state <= rfsm_state_reading;
            n_rfsm_blen <= rfsm_blen - 1;
            n_rfsm_blen_last <= less_than_equal_to_two(rfsm_blen);
            n_rfsm_slot_valid <= '1';
            n_rfsm_slot <= rfsm_slot;
            n_rfsm_slot_1hot <= rfsm_slot_1hot;
            n_rfsm_tag <= rfsm_tag;
          end if;
        else 
          n_rfsm_state <= rfsm_state_reading;
          n_rfsm_blen <= rfsm_blen;
          n_rfsm_blen_last <= rfsm_blen_last;
          n_rfsm_slot_valid <= '1';
          n_rfsm_slot <= rfsm_slot;
          n_rfsm_slot_1hot <= rfsm_slot_1hot;
          n_rfsm_tag <= rfsm_tag;
        end if;
    end case;
  end process;
  
end rtl;
