--------------------------------------------------------------------------------
--
-- Module Name      : adb3_to_adb3l
--
--                    Copyright 2009 by
--                    Alpha Data Parallel Systems Ltd.
--                    All Rights Reserved.
--
-- Original Author  : Andrew McCormick
-- Created          : 
--
-- Modified By      : 
-- Date             :
-- Change Notes     :     
--
-- Description      : Conversion module connecting ADB3 Master and ADB3L Slave
--
--
-- Dependencies     : adb3_ocp
--
-- Disclaimer       : THESE DESIGNS ARE PROVIDED "AS IS" WITH NO WARRANTY
--                    WHATSOEVER AND ALPHA DATA SPECIFICALLY DISCLAIMS ANY
--                    WARRANTIES IMPLIED OF MERCHANTABILITY, FITNESS FOR A
--                    PARTICULAR PURPOSE, OR AGAINST INFRINGEMENT.
--
--------------------------------------------------------------------------------


library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_misc.all;

library Work;
use work.adb3_ocp.all;

entity adb3_to_adb3l is
  port (
    clk     : in  std_logic;
    pri_m2s : in  adb3_ocp_m2sT;
    pri_s2m : out adb3_ocp_s2mT;
    sec_m2s : out adb3l_ocp_m2sT;
    sec_s2m : in  adb3l_ocp_s2mT);
end entity adb3_to_adb3l;

architecture rtl of adb3_to_adb3l is

  type state_type is (idle,waiting_data,writing,read_req,read_resp);
  signal fsm_state : state_type := idle;
  signal cmd_accept : std_logic := '1';
  signal data_accept : std_logic := '0';
  signal resp_accept : std_logic := '0';
  
  
  signal write_words_active : std_logic_vector(3 downto 0) := "0000";
  signal write_cmd_active : std_logic := '0';
  signal wdata_reg : std_logic_vector(127 downto 0);
  signal wdata_be_reg : std_logic_vector(15 downto 0);
  signal burst_count : std_logic_vector(11 downto 0) := (others => '0');
  signal addr : std_logic_vector(31 downto 0) := (others => '0');
  
  signal read_words_active : std_logic_vector(3 downto 0) := "0000";
  signal read_words_received : std_logic_vector(3 downto 0) := "0000";
  signal rdata_reg : std_logic_vector(127 downto 0);
  signal resp_valid : std_logic := '0';
  signal tag : std_logic_vector(7 downto 0);
  
begin  -- architecture rtl

  process (clk) is
  begin  -- process
    if rising_edge(clk) then            -- rising clock edge
      case fsm_state is
        when idle =>
          if pri_m2s.Cmd = OCP_CMD_WRITE then
            burst_count <= pri_m2s.BurstLength;
            fsm_state <= waiting_data;
            data_accept <= '1';
            cmd_accept <= '0';
            resp_accept <= '0';
            resp_valid <= '0';
            addr <= pri_m2s.Addr(31 downto 0);
          elsif pri_m2s.Cmd = OCP_CMD_READ then
            burst_count <= pri_m2s.BurstLength;
            fsm_state <= read_req;
            data_accept <= '0';
            cmd_accept <= '0';
            resp_accept <= '1';
            resp_valid <= '0';
            addr <= pri_m2s.Addr(31 downto 0);
            read_words_active <= "1111";
            read_words_received <= "0000";
            tag <= pri_m2s.Tag;
          end if;
        when waiting_data =>
          if pri_m2s.DataValid = '1' then
            fsm_state <= writing;
            data_accept <= '0';
            write_words_active <= OR_reduce(pri_m2s.DataByteEn(15 downto 12)) &
                                  OR_reduce(pri_m2s.DataByteEn(11 downto 8)) &
                                  OR_reduce(pri_m2s.DataByteEn(7 downto 4)) &
                                  OR_reduce(pri_m2s.DataByteEn(3 downto 0));
            wdata_reg <= pri_m2s.Data;
            wdata_be_reg <= pri_m2s.DataByteEn;
            burst_count <= burst_count-1;
            write_cmd_active <='1';
          end if;
        when writing =>
          if write_words_active(0) = '0' or sec_s2m.DataAccept = '1' then
            write_words_active <= '0' & write_words_active(3 downto 1);
            wdata_reg <= X"00000000" & wdata_reg(127 downto 32);
            wdata_be_reg <= X"0" & wdata_be_reg(15 downto 4);
            addr <= addr +4;
            if OR_reduce(write_words_active(3 downto 1)) = '0' then
              if OR_reduce(burst_count) = '0' then
                fsm_state <= idle;
                data_accept <= '0';
                cmd_accept <= '1';
                resp_valid <= '0';
                resp_accept <= '0';
              else
                fsm_state <= waiting_data;
                data_accept <= '1';
                cmd_accept <= '0';
                resp_valid <= '0';
                resp_accept <= '0';
              end if;
              write_cmd_active <='0';
            else
              write_cmd_active <='1';
            end if;
          elsif sec_s2m.CmdAccept = '1' then
            write_cmd_active <= '0';
          end if;    
        when read_req =>
          if sec_s2m.CmdAccept = '1' then
            read_words_active <= '0' & read_words_active(3 downto 1);
            addr <= addr +4;
          end if;
          if sec_s2m.Resp = OCP_RESP_VALID then
            read_words_received <= '1' & read_words_received(3 downto 1);
            rdata_reg <= sec_s2m.Data & rdata_reg(127 downto 32);
            if AND_reduce(read_words_received(3 downto 1)) = '1' then
              fsm_state <= read_resp;
              resp_accept <= '0';
              data_accept <= '0';
              cmd_accept <= '0';
              burst_count <= burst_count -1;
              resp_valid <= '1';
            end if;
          end if;
        when read_resp =>
          if pri_m2s.RespAccept = '1' then
            if OR_reduce(burst_count) = '0' then
              fsm_state <= idle;
              cmd_accept <= '1';
              data_accept <= '0';
              resp_accept <= '0';
              resp_valid <= '0';
            else
              fsm_state <= read_req;
              resp_accept <= '1';
              data_accept <= '0';
              cmd_accept <= '0';
              resp_valid <= '0';
              read_words_active <= "1111";
              read_words_received <= "0000";
            end if;
          end if;      
        when others =>
          fsm_state <= idle;
      end case;
      
    
    end if;
  end process;

  sec_m2s.Cmd <= OCP_CMD_WRITE when write_cmd_active = '1' and write_words_active(0) ='1' else
                 OCP_CMD_READ when read_words_active(0) = '1' else
                 OCP_CMD_IDLE;
  sec_m2s.Addr <= addr;
  sec_m2s.Data <= wdata_reg(31 downto 0);
  sec_m2s.DataByteEn <= wdata_be_reg(3 downto 0);
  sec_m2s.DataValid <= write_words_active(0);
  sec_m2s.RespAccept <= resp_accept;

  pri_s2m.CmdAccept <= cmd_accept;
  pri_s2m.DataAccept <= data_accept;
  pri_s2m.Data <= rdata_reg;
  pri_s2m.Resp <= OCP_RESP_VALID when resp_valid = '1' else OCP_RESP_NONE;
  pri_s2m.Tag <= tag;
  

  
end architecture rtl;
