--
-- adb3w_ocp_reorder.vhd
--
-- Bridging component that reorders responses in an unordered stream of OCP
-- responses to match the order in which the OCP read commands that generated
-- them were received on the in-order side.
--
-- On the unordered side, OCP responses are assumed to be a stream where
-- responses (data & tag) for any in-flight read command can be interleaved
-- with responses for any other in-flight read command, with no ordering
-- guarantees.
--
-- On the in-order side, OCP responses are delivered in the order dictated by
-- the order of OCP read commands as they were received by the in-order side.
--

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_misc.all;
use ieee.numeric_std.all;
use ieee.math_real.all;

library work;
use work.adb3_ocp.all;

entity adb3w_ocp_reorder is
  generic(
    -- Log2 of max. number of OCP words in a read command accepted on in-order
    -- side.
    buffer_order       : in natural;
    -- Maximum number of reads in flight issued by in-order side OCP master.
    max_read_in_flight : in natural;
    -- If true, delays CmdAccept on in-order side when read in flight limit
    -- reached.
    enforce_max_read   : in boolean := false;
    -- If true, data buffer is adb3_ocp_reorder_dbuf_ur; otherwise
    -- adb3_ocp_reorder_dbuf_mf.
    unified_ram        : in boolean := true;
    -- If true, instantiates extra pipelining in DBUF output to mitigate logic
    -- complexity of multiplexors or large Tco of BlockRAMS.
    output_pipe        : in boolean := false);
  port(
    rst       : in  std_logic; -- Async. reset (tie to 0 if unused)
    sr        : in  std_logic; -- Sync. reset (tie to 0 if unused)
    clk       : in  std_logic; -- OCP clock
    -- In-order side, OCP slave port
    io_m2s    : in  adb3w_ocp_m2sT;
    io_s2m    : out adb3w_ocp_s2mT;
    -- Unordered side, OCP master port
    uo_m2s    : out adb3w_ocp_m2sT;
    uo_s2m    : in  adb3w_ocp_s2mT;
    -- Status fields
    --    [7:0]  Number of read commands currently in flight
    --    [9:0]  RFSM state: 00 => IDLE, 01 => READING
    --      [10] Command FIFO 'valid' flag
    --      [11] Command FIFO 'full' flag
    -- Following 3 fields valid only when RFSM state is READING:
    --   [19:12] Tag of current in-order side read
    --   [31:20] Burst length remaining in current in-order side read
    --   [35:32] Slot number associated with current in-order side read
    --   [47:36] (reserved)
    -- Error flags in bits [63:48]
    --      [48] In-order side OCP master attempted to exceed
    --           max_read_in_flight
    --      [49] Attempt to allocate and free a particular slot in same clock
    --           cycle
    --      [50] A read command was accepted but there was no free slot to
    --           store it
    --      [51] A read command was accepted with burst length >
    --           2**buffer_order
    --      [52] The command FIFO overflowed
    --      [53] The command FIFO underflowed
    --      [54] A response was received on the unordered side whose tag could
    --           not be matched to any slot.
    --      [55] Overflow detected in a slot's data buffer
    --      [56] Underflow detected in a slot's data buffer
    --   [63:57] (reserved)
    status    : out std_logic_vector(63 downto 0);
    -- A '1' in any bit clears corresponding 'error' bit in status[63:48]
    clr_error : in  std_logic_vector(15 downto 0));
end entity;

architecture rtl of adb3w_ocp_reorder is

  component adb3_ocp_reorder_base
    generic (
      -- Width of a data word, in bits
      data_width         : in natural;
      -- Log2 of max. number of OCP words in a read command accepted on in-order
      -- side.
      buffer_order       : in natural;
      -- Maximum number of reads in flight issued by in-order side OCP master.
      max_read_in_flight : in natural;
      -- If true, delays CmdAccept on in-order side when read in flight limit
      -- reached.
      enforce_max_read   : in boolean;
      -- If true, data buffer is adb3_ocp_reorder_dbuf_ur; otherwise
      -- adb3_ocp_reorder_dbuf_mf.
      unified_ram        : in boolean;
      -- If true, instantiates extra pipelining in DBUF output to mitigate logic
      -- complexity of multiplexors or large Tco of BlockRAMS.
      output_pipe        : in boolean);
    port (
      rst         : in  std_logic; -- Async. reset (tie to 0 if unused)
      sr          : in  std_logic; -- Sync. reset (tie to 0 if unused)
      clk         : in  std_logic; -- OCP clock
      -- In-order side, OCP slave port
      io_m2s_cmd  : in  ocp_CmdT;
      io_m2s_addr : in  std_logic_vector(ADB3_OCP_ADDR_WIDTH - 1 downto 0);
      io_m2s_blen : in  std_logic_vector(ADB3_OCP_BURST_WIDTH - 1 downto 0);
      io_m2s_data : in  std_logic_vector(data_width - 1 downto 0);
      io_m2s_be   : in  std_logic_vector(data_width / 8 - 1 downto 0);
      io_m2s_dval : in  std_logic;
      io_m2s_racc : in  std_logic;
      io_m2s_tag  : in  std_logic_vector(ADB3_OCP_TAG_WIDTH - 1 downto 0);
      io_s2m_cacc : out std_logic;
      io_s2m_dacc : out std_logic;
      io_s2m_data : out std_logic_vector(data_width - 1 downto 0);
      io_s2m_resp : out ocp_RespT;
      io_s2m_tag  : out std_logic_vector(ADB3_OCP_TAG_WIDTH - 1 downto 0);
      -- Unordered side, OCP master port
      uo_m2s_cmd  : out ocp_CmdT;
      uo_m2s_addr : out std_logic_vector(ADB3_OCP_ADDR_WIDTH - 1 downto 0);
      uo_m2s_blen : out std_logic_vector(ADB3_OCP_BURST_WIDTH - 1 downto 0);
      uo_m2s_data : out std_logic_vector(data_width - 1 downto 0);
      uo_m2s_be   : out std_logic_vector(data_width / 8 - 1 downto 0);
      uo_m2s_dval : out std_logic;
      uo_m2s_racc : out std_logic;
      uo_m2s_tag  : out std_logic_vector(ADB3_OCP_TAG_WIDTH - 1 downto 0);
      uo_s2m_cacc : in  std_logic;
      uo_s2m_dacc : in  std_logic;
      uo_s2m_data : in  std_logic_vector(data_width - 1 downto 0);
      uo_s2m_resp : in  ocp_RespT;
      uo_s2m_tag  : in  std_logic_vector(ADB3_OCP_TAG_WIDTH - 1 downto 0);
      -- Status fields
      --    [7:0]  Number of read commands currently in flight
      --    [9:0]  RFSM state: 00 => IDLE, 01 => READING
      --      [10] Command FIFO 'valid' flag
      --      [11] Command FIFO 'full' flag
      -- Following 3 fields valid only when RFSM state is READING:
      --   [19:12] Tag of current in-order side read
      --   [31:20] Burst length remaining in current in-order side read
      --   [35:32] Slot number associated with current in-order side read
      --   [47:36] (reserved)
      -- Error flags in bits [63:48]
      --      [48] In-order side OCP master attempted to exceed
      --           max_read_in_flight
      --      [49] Attempt to allocate and free a particular slot in same clock
      --           cycle
      --      [50] A read command was accepted but there was no free slot to
      --           store it
      --      [51] A read command was accepted with burst length >
      --           2**buffer_order
      --      [52] The command FIFO overflowed
      --      [53] The command FIFO underflowed
      --      [54] A response was received on the unordered side whose tag could
      --           not be matched to any slot.
      --      [55] Overflow detected in a slot's data buffer
      --      [56] Underflow detected in a slot's data buffer
      --   [63:57] (reserved)
      status    : out std_logic_vector(63 downto 0);
      -- A '1' in any bit clears corresponding 'error' bit in status[63:48]
      clr_error : in  std_logic_vector(15 downto 0));
  end component;

  signal io_m2s_cmd : ocp_CmdT;
  signal io_m2s_addr : std_logic_vector(ADB3_OCP_ADDR_WIDTH - 1 downto 0);
  signal io_m2s_blen : std_logic_vector(ADB3_OCP_BURST_WIDTH - 1 downto 0);
  signal io_m2s_data : std_logic_vector(ADB3W_OCP_DATA_WIDTH - 1 downto 0);
  signal io_m2s_be : std_logic_vector(ADB3W_OCP_DATA_WIDTH / 8 - 1 downto 0);
  signal io_m2s_dval : std_logic;
  signal io_m2s_racc : std_logic;
  signal io_m2s_tag : std_logic_vector(ADB3_OCP_TAG_WIDTH - 1 downto 0);
  signal io_s2m_cacc : std_logic;
  signal io_s2m_dacc : std_logic;
  signal io_s2m_data : std_logic_vector(ADB3W_OCP_DATA_WIDTH - 1 downto 0);
  signal io_s2m_resp : ocp_RespT;
  signal io_s2m_tag : std_logic_vector(ADB3_OCP_TAG_WIDTH - 1 downto 0);

  signal uo_m2s_cmd : ocp_CmdT;
  signal uo_m2s_addr : std_logic_vector(ADB3_OCP_ADDR_WIDTH - 1 downto 0);
  signal uo_m2s_blen : std_logic_vector(ADB3_OCP_BURST_WIDTH - 1 downto 0);
  signal uo_m2s_data : std_logic_vector(ADB3W_OCP_DATA_WIDTH - 1 downto 0);
  signal uo_m2s_be : std_logic_vector(ADB3W_OCP_DATA_WIDTH / 8 - 1 downto 0);
  signal uo_m2s_dval : std_logic;
  signal uo_m2s_racc : std_logic;
  signal uo_m2s_tag : std_logic_vector(ADB3_OCP_TAG_WIDTH - 1 downto 0);
  signal uo_s2m_cacc : std_logic;
  signal uo_s2m_dacc : std_logic;
  signal uo_s2m_data : std_logic_vector(ADB3W_OCP_DATA_WIDTH - 1 downto 0);
  signal uo_s2m_resp : ocp_RespT;
  signal uo_s2m_tag : std_logic_vector(ADB3_OCP_TAG_WIDTH - 1 downto 0);

begin

  U0 : adb3_ocp_reorder_base
    generic map(
      data_width => ADB3W_OCP_DATA_WIDTH,
      buffer_order => buffer_order,
      max_read_in_flight => max_read_in_flight,
      enforce_max_read => enforce_max_read,
      unified_ram => unified_ram,
      output_pipe => output_pipe)
    port map(
      rst => rst,
      sr => sr,
      clk => clk,
      io_m2s_cmd => io_m2s_cmd,
      io_m2s_addr => io_m2s_addr,
      io_m2s_blen => io_m2s_blen,
      io_m2s_data => io_m2s_data,
      io_m2s_be => io_m2s_be,
      io_m2s_dval => io_m2s_dval,
      io_m2s_racc => io_m2s_racc,
      io_m2s_tag => io_m2s_tag,
      io_s2m_cacc => io_s2m_cacc,
      io_s2m_dacc => io_s2m_dacc,
      io_s2m_data => io_s2m_data,
      io_s2m_resp => io_s2m_resp,
      io_s2m_tag => io_s2m_tag,
      uo_m2s_cmd => uo_m2s_cmd,
      uo_m2s_addr => uo_m2s_addr,
      uo_m2s_blen => uo_m2s_blen,
      uo_m2s_data => uo_m2s_data,
      uo_m2s_be => uo_m2s_be,
      uo_m2s_dval => uo_m2s_dval,
      uo_m2s_racc => uo_m2s_racc,
      uo_m2s_tag => uo_m2s_tag,
      uo_s2m_cacc => uo_s2m_cacc,
      uo_s2m_dacc => uo_s2m_dacc,
      uo_s2m_data => uo_s2m_data,
      uo_s2m_resp => uo_s2m_resp,
      uo_s2m_tag => uo_s2m_tag,
      status => status,
      clr_error => clr_error);

  io_m2s_cmd <= io_m2s.Cmd;
  io_m2s_addr <= io_m2s.Addr;
  io_m2s_blen <= io_m2s.BurstLength;
  io_m2s_data <= io_m2s.Data;
  io_m2s_be <= io_m2s.DataByteEn;
  io_m2s_dval <= io_m2s.DataValid;
  io_m2s_racc <= io_m2s.RespAccept;
  io_m2s_tag <= io_m2s.Tag;

  io_s2m.CmdAccept <= io_s2m_cacc;
  io_s2m.DataAccept <= io_s2m_dacc;
  io_s2m.Data <= io_s2m_data;
  io_s2m.Resp <= io_s2m_resp;
  io_s2m.Tag <= io_s2m_tag;

  uo_m2s.Cmd <= uo_m2s_cmd;
  uo_m2s.Addr <= uo_m2s_addr;
  uo_m2s.BurstLength <= uo_m2s_blen;
  uo_m2s.Data <= uo_m2s_data;
  uo_m2s.DataByteEn <= uo_m2s_be;
  uo_m2s.DataValid <= uo_m2s_dval;
  uo_m2s.RespAccept <= uo_m2s_racc;
  uo_m2s.Tag <= uo_m2s_tag;

  uo_s2m_cacc <= uo_s2m.CmdAccept;
  uo_s2m_dacc <= uo_s2m.DataAccept;
  uo_s2m_data <= uo_s2m.Data;
  uo_s2m_resp <= uo_s2m.Resp;
  uo_s2m_tag <= uo_s2m.Tag;

end rtl;
