--
-- AXI4 Slave to OCP Master Protocol Converter
-- For Direct Master 
--
-- Assumptions made:
--
--  1a. OCP writes are performed by downstream slave in same order as OCP write commands.
--
--  1b. It is OK to issue write response (BID) as soon as last word of AXI4 write data has been
--     accepted.
--
--   2. OCP read responses are returned by downstream slave in same order as OCP read commands.
--      This makes mapping of OCP tag back to AXI4 RID unnecessary, as AXI4 ARID can be placed
--      into a FIFO and read out as RID. Another consequence is that OCP tag generation is trivial
--      and can be accomplished simply by cycling through a counter.
--
-- Limitations:
--
--   4. AXI4 WRAP and FIXED burst types are not supported.
--  
--   5. AXI4 narrow bursts are not supported.
--  
--   6. AXI4 locked transactions are treated as unlocked transactions.
--  
--   7. AXI4 cache attributes are ignored.
--  
--   8. AXI4 unaligned addresses are not supported.
--

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_misc.all;
use ieee.numeric_std.all;

library work;
use work.adb3_ocp.all;
use work.adb3w_ocp_axi4.all;

entity axi4_to_adb3w_ocp_dm is
  generic(
    minimize_dead_cycles : in    boolean := true;    -- true => eliminate dead cycles where possible, at cost of area & speed
    arb_mode             : in    string := "RR";     -- One of { "RR", "RDPRI", "WRPRI" }
    cf_implementation    : in    string := "FIFO";   -- One of { "FIFO", "REG" }
    cf_order             : in    natural := 4;       -- Ignored if not FIFO
    rwf_implementation   : in    string := "FIFO";   -- One of { "FIFO", "REG" }
    rwf_order            : in    natural := 4;       -- Ignored if not FIFO
    rrf_implementation   : in    string := "FIFO";   -- One of { "FIFO", "REG" }
    rrf_order            : in    natural := 4;       -- Ignored if not FIFO
    wwf_implementation   : in    string := "FIFO";   -- One of { "FIFO", "REG" }
    wwf_order            : in    natural := 4;       -- Ignored if not FIFO
    wdf_implementation   : in    string := "FIFO";   -- One of { "FIFO", "REG" }
    wdf_order            : in    natural := 4;       -- Ignored if not FIFO
    wrf_implementation   : in    string := "FIFO";   -- One of { "FIFO", "REG" }
    wrf_order            : in    natural := 4);      -- Ignored if not FIFO
  port(
    clk                  : in    std_logic;
    rst                  : in    std_logic;
    axi_m2s              : in    axi4_256_m2sT;
    axi_s2m              : out   axi4_256_s2mT;
    ocp_m2s              : out   adb3w_ocp_m2sT;
    ocp_s2m              : in    adb3w_ocp_s2mT);
end entity;

architecture rtl of axi4_to_adb3w_ocp_dm is

  component axi4_to_adb3w_ocp_dm_fifo
    generic(
      implementation : in    string := "FIFO";                      -- One of { "FIFO", "REG" }
      order          : in    natural range 2 to 6;                  -- Depth is 2**order
      width          : in    natural;                               -- width of FIFO, in bits
      afull_free     : in    natural := 1);                         -- When 'afull_free' or fewer free locations, 'afull' is asserted
    port(
      rst            : in    std_logic;                             -- Asynchronous reset
      clk            : in    std_logic;                             -- Clock
      wr             : in    std_logic;                             -- '1' => write value on 'd' into FIFO
      d              : in    std_logic_vector(width - 1 downto 0);  -- Item to write into FIFO, qualified by 'wr'
      adv            : in    std_logic;                             -- '1' => present next item on 'q' to next item
      full           : out   std_logic;                             -- '1' => full
      not_full       : out   std_logic;                             -- Inverse of 'full'
      afull          : out   std_logic;                             -- '1' => almost full
      not_afull      : out   std_logic;                             -- Inverse of 'afull'
      valid          : out   std_logic;                             -- '1' => not empty
      q              : out   std_logic_vector(width - 1 downto 0)); -- Current item, qualified by 'valid'
  end component;

  -- Arbiter (ARB)
  signal arb_last_was_read : std_logic := '0';
  signal arb_arready : std_logic := '0';
  signal arb_awready : std_logic := '0';
  
  -- Command FIFO (CF)
  -- m2s.Cmd (3), m2s.Addr (64), m2s.BurstLen (9)
  constant cf_width : natural := 3 + 64 + 9;
  signal cf_wr : std_logic;
  signal cf_d : std_logic_vector(cf_width - 1 downto 0);
  signal cf_adv : std_logic;
  signal cf_not_full : std_logic;
  signal cf_valid : std_logic;
  signal cf_q : std_logic_vector(cf_width - 1 downto 0);
  signal cf_q_cmd : std_logic_vector(2 downto 0);
  signal cf_q_addr : std_logic_vector(63 downto 0);
  signal cf_q_blen : std_logic_vector(8 downto 0);
  signal cf_generated_tag : std_logic_vector(7 downto 0);
  
  -- Read Worker FIFO (RWF)
  -- LEN(8) + ID (4)
  constant rwf_width : natural := 8 + 4;
  signal rwf_wr : std_logic;
  signal rwf_d : std_logic_vector(rwf_width - 1 downto 0);
  signal rwf_adv : std_logic;
  signal rwf_not_full : std_logic;
  signal rwf_valid : std_logic;
  signal rwf_q : std_logic_vector(rwf_width - 1 downto 0);
  signal rwf_q_id : std_logic_vector(3 downto 0);
  signal rwf_q_len : std_logic_vector(7 downto 0);

  -- Read Response FIFO (RRF)
  -- RDATA(256) + RRESP (2)
  constant rrf_width : natural := 256 + 2;
  signal rrf_wr : std_logic;
  signal rrf_d : std_logic_vector(rrf_width - 1 downto 0);
  signal rrf_adv : std_logic;
  signal rrf_not_full : std_logic;
  signal rrf_valid : std_logic;
  signal rrf_q : std_logic_vector(rrf_width - 1 downto 0);
  signal rrf_q_data : std_logic_vector(255 downto 0);
  signal rrf_q_resp : std_logic_vector(1 downto 0);
  
  -- Read worker (RW)
  type rw_state_t is (rw_state_idle, rw_state_reading);
  signal rw_state : rw_state_t := rw_state_idle;
  signal n_rw_state : rw_state_t;
  signal rw_id, n_rw_id : std_logic_vector(3 downto 0);
  signal rw_last, n_rw_last : std_logic;
  signal rw_len, n_rw_len : std_logic_vector(7 downto 0);
  
  -- Write Worker FIFO (WWF)
  -- LEN(8) + ID (4)
  constant wwf_width : natural := 8 + 4;
  signal wwf_wr : std_logic;
  signal wwf_d : std_logic_vector(wwf_width - 1 downto 0);
  signal wwf_adv : std_logic;
  signal wwf_not_full : std_logic;
  signal wwf_valid : std_logic;
  signal wwf_q : std_logic_vector(wwf_width - 1 downto 0);
  signal wwf_q_id : std_logic_vector(3 downto 0);
  signal wwf_q_len : std_logic_vector(7 downto 0);

  -- Write Data FIFO (WDF)
  -- WDATA(256) + WSTRB (32)
  constant wdf_width : natural := 256 + 32;
  signal wdf_wr : std_logic;
  signal wdf_d : std_logic_vector(wdf_width - 1 downto 0);
  signal wdf_adv : std_logic;
  signal wdf_not_full : std_logic;
  signal wdf_valid : std_logic;
  signal wdf_q : std_logic_vector(wdf_width - 1 downto 0);
  signal wdf_q_data : std_logic_vector(255 downto 0);
  signal wdf_q_strb : std_logic_vector(31 downto 0);

  -- Write Worker (WW)
  type ww_state_t is (ww_state_idle, ww_state_writing);
  signal ww_state : ww_state_t := ww_state_idle;
  signal n_ww_state : ww_state_t;
  signal ww_id, n_ww_id : std_logic_vector(3 downto 0);
  signal ww_last, n_ww_last : std_logic;
  signal ww_len, n_ww_len : std_logic_vector(7 downto 0);
  
  -- Write Response FIFO (WRF)
  -- BID (4)
  constant wrf_width : natural := 4;
  signal wrf_wr : std_logic;
  signal wrf_d : std_logic_vector(wrf_width - 1 downto 0);
  signal wrf_adv : std_logic;
  signal wrf_not_full : std_logic;
  signal wrf_not_afull : std_logic;
  signal wrf_valid : std_logic;
  signal wrf_q : std_logic_vector(wrf_width - 1 downto 0);
  signal wrf_q_id : std_logic_vector(3 downto 0);

  
begin

  assert arb_mode = "RR" report "axi4_to_adb3w_ocp_dm: Currently, only RR arbitration mode is supported" severity failure;

  
  --
  -- Arbiter
  -- Inputs attributes from AXI4 AR & AW subchannels into Command FIFO.
  --
  
  arbiter : process(rst, clk)
    variable tmp2 : std_logic_vector(1 downto 0);
    variable tmp3 : std_logic_vector(2 downto 0);
  begin
    if rst = '1' then
      arb_arready <= '0';
      arb_awready <= '0';
      arb_last_was_read <= '0';
      cf_d <= (others => '-');
      rwf_d <= (others => '-');
      wwf_d <= (others => '-');
    elsif rising_edge(clk) then
      -- Round-robin arbitration; favours the subchannel that was not the last one selected.
      -- Rate of command acceptance is 1 per 2 clock cycles at most.
      arb_arready <= axi_m2s.arvalid and (not arb_last_was_read or not axi_m2s.awvalid) and cf_not_full and not cf_wr and wwf_not_full and rwf_not_full;
      arb_awready <= axi_m2s.awvalid and (arb_last_was_read or not axi_m2s.arvalid) and cf_not_full and not cf_wr and wwf_not_full and rwf_not_full;
      
      tmp2 := arb_awready & arb_arready;
      case tmp2 is
        when "00" =>
          null;
          
        when "01" =>
          arb_last_was_read <= '1';
          
        when "10" =>
          arb_last_was_read <= '0';
          
        when "11" =>
          arb_last_was_read <= '-';
          report "axi4_to_adb3w_ocp_dm: Internal error detected: accepting a read and a write from AXI4 port in same cycle" severity error;
          
        when others =>
          assert false severity error; -- Should never get here
          arb_last_was_read <= '-';
      end case;
      
      tmp3 := arb_last_was_read & axi_m2s.awvalid & axi_m2s.arvalid;
      case tmp3 is
        when "000" | "100" =>
          cf_d <= (others => '-');
          rwf_d <= (others => '-');
          wwf_d <= (others => '-');
          
        when "001" | "101" | "011" =>
          cf_d(2 downto 0) <= OCP_CMD_READ;
          cf_d(66 downto 3) <= axi_m2s.araddr;
          cf_d(75 downto 67) <= std_logic_vector(resize(unsigned(axi_m2s.arlen), 9)) + 1;
          rwf_d(3 downto 0) <= axi_m2s.arid;
          rwf_d(11 downto 4) <= axi_m2s.arlen;
          wwf_d <= (others => '-');
          
        when "010" | "110" | "111" =>
          cf_d(2 downto 0) <= OCP_CMD_WRITE;
          cf_d(66 downto 3) <= axi_m2s.awaddr;
          cf_d(75 downto 67) <= std_logic_vector(resize(unsigned(axi_m2s.awlen), 9)) + 1;
          rwf_d <= (others => '-');
          wwf_d(3 downto 0) <= axi_m2s.awid;
          wwf_d(11 downto 4) <= axi_m2s.awlen;
          
        when others =>
          assert false severity error; -- Should never get here
          cf_d <= (others => '-');
          rwf_d <= (others => '-');
          wwf_d <= (others => '-');
      end case;
    end if;
  end process;
  
  axi_s2m.arready <= arb_arready;
  axi_s2m.awready <= arb_awready;
  
  
  --
  -- Command FIFO
  -- Inputs commands from of AR & AW subchannels and presents them on OCP port.
  --
  
  cf_wr <= arb_arready or arb_awready;
  
  cf_inst : axi4_to_adb3w_ocp_dm_fifo
    generic map(
      implementation => cf_implementation,
      order => cf_order,
      width => cf_width)
    port map(
      rst => rst,
      clk => clk,
      wr => cf_wr,
      d => cf_d,
      adv => cf_adv,
      not_full => cf_not_full,
      q => cf_q,
      valid => cf_valid);
  
  cf_adv <= ocp_s2m.CmdAccept and cf_valid;
  cf_q_cmd <= cf_q(2 downto 0);
  cf_q_addr <= cf_q(66 downto 3);
  cf_q_blen <= cf_q(75 downto 67);
  
  ocp_m2s.Cmd <= cf_q_cmd when cf_valid = '1' else OCP_CMD_IDLE;
  ocp_m2s.Addr <= cf_q_addr;
  ocp_m2s.BurstLength <= std_logic_vector(resize(unsigned(cf_q_blen), ADB3_OCP_BURST_WIDTH));
  ocp_m2s.Tag <= cf_generated_tag;
  
  tag_generator : process(rst, clk)
  begin
    if rst = '1' then
      cf_generated_tag <= (others => '0');
    elsif rising_edge(clk) then
      if cf_valid = '1' and ocp_s2m.CmdAccept = '1' and cf_q_cmd = OCP_CMD_READ then
        cf_generated_tag <= cf_generated_tag + 1;
      end if;
    end if;
  end process;
  
  
  --
  -- Read Worker FIFO
  -- Buffers ID & LEN from AXI4 AR subchannel for Read Worker.
  --

  -- rwf_d is generated in arbiter process above
  rwf_wr <= arb_arready;
  
  rwf_inst : axi4_to_adb3w_ocp_dm_fifo
    generic map(
      implementation => rwf_implementation,
      order => rwf_order,
      width => rwf_width)
    port map(
      rst => rst,
      clk => clk,
      wr => rwf_wr,
      d => rwf_d,
      adv => rwf_adv,
      not_full => rwf_not_full,
      q => rwf_q,
      valid => rwf_valid);

  -- rwf_adv is generated in combinatorial part of Read Worker state machine      
  rwf_q_id <= rwf_q(3 downto 0);
  rwf_q_len <= rwf_q(11 downto 4);
  
  
  --
  -- Read Response FIFO
  -- Buffers read Data & Resp from OCP port and presents them on AXI4 R subchannel as RDATA & RRESP.
  -- OCP Resp corresponding to any kind of error is converted to AXI4 SLVERR ("10").
  --

  rrf_wr <= '1' when ocp_s2m.Resp /= OCP_RESP_NONE and rrf_not_full = '1' else '0';
  rrf_d(255 downto 0) <= ocp_s2m.Data;
  rrf_d(257 downto 256) <= "10" when ocp_s2m.Resp = OCP_RESP_FAILED or ocp_s2m.Resp = OCP_RESP_ERROR else
                           "00" when ocp_s2m.Resp = OCP_RESP_VALID else
                           "--";
  
  ocp_m2s.RespAccept <= rrf_not_full;
  
  rrf_inst : axi4_to_adb3w_ocp_dm_fifo
    generic map(
      implementation => rrf_implementation,
      order => rrf_order,
      width => rrf_width)
    port map(
      rst => rst,
      clk => clk,
      wr => rrf_wr,
      d => rrf_d,
      adv => rrf_adv,
      not_full => rrf_not_full,
      q => rrf_q,
      valid => rrf_valid);
  
  rrf_adv <= '1' when rrf_valid = '1' and rw_state = rw_state_reading and axi_m2s.rready = '1' else '0';
  rrf_q_data <= rrf_q(255 downto 0);
  rrf_q_resp <= rrf_q(257 downto 256);

  axi_s2m.rdata <= rrf_q_data;
  axi_s2m.rresp <= rrf_q_resp;
  axi_s2m.rvalid <= '1' when rrf_valid = '1' and rw_state = rw_state_reading else '0';

  
  --
  -- Read Worker
  -- Gets ID from Read Worker FIFO and presents on AXI4 R subchannel as RID.
  -- Tracks presentation of RDATA, RID & RRESP on AXI4 R subchannel in order to detect
  -- end of burst and generate RLAST on AXI4 R subchannel.
  --
  
  gen_rw_state : process(rst, clk)
  begin
    if rst = '1' then
      rw_state <= rw_state_idle;
      rw_id <= (others => '-');
      rw_last <= '-';
      rw_len <= (others => '-');
    elsif rising_edge(clk) then
      rw_state <= n_rw_state;
      rw_id <= n_rw_id;
      rw_last <= n_rw_last;
      rw_len <= n_rw_len;
    end if;
  end process;
  
  gen_n_rw_state : process(
    rw_state,
    rw_id,
    rw_last,
    rw_len,
    rwf_valid,
    rwf_q_id,
    rwf_q_len,
    rrf_adv,
    axi_m2s)
  begin
    -- Establish defaults
    rwf_adv <= '0';
    
    case rw_state is
      when rw_state_idle =>
        if rwf_valid = '1' then
          n_rw_state <= rw_state_reading;
          n_rw_id <= rwf_q_id;
          n_rw_last <= not or_reduce(rwf_q_len);
          n_rw_len <= rwf_q_len;
          rwf_adv <= '1';
        else
          n_rw_state <= rw_state_idle;
          n_rw_id <= (others => '-');
          n_rw_last <= '-';
          n_rw_len <= (others => '-');
        end if;
        
      when rw_state_reading =>
        if rrf_adv = '1' then
          if rw_last = '1' then 
            if minimize_dead_cycles and rwf_valid = '1' then
              n_rw_state <= rw_state_reading;
              n_rw_id <= rwf_q_id;
              n_rw_last <= not or_reduce(rwf_q_len);
              n_rw_len <= rwf_q_len;
              rwf_adv <= '1';
            else
              n_rw_state <= rw_state_idle;
              n_rw_id <= (others => '-');
              n_rw_last <= '-';
              n_rw_len <= (others => '-');
            end if;
          else
          n_rw_state <= rw_state_reading;
          n_rw_id <= rw_id;
          n_rw_last <= not or_reduce(rw_len(rw_len'high downto 1));
          n_rw_len <= rw_len - 1;
          end if;
        else
          n_rw_state <= rw_state_reading;
          n_rw_id <= rw_id;
          n_rw_last <= rw_last;
          n_rw_len <= rw_len;
        end if;
    end case;
  end process;
  
  axi_s2m.rid <= rw_id;
  axi_s2m.rlast <= rw_last;

  
  --
  -- Write Worker FIFO
  -- Buffers ID & LEN tuples from AXI4 AW subchannel for Write Worker.
  --

  -- wwf_d is generated in arbiter process above
  wwf_wr <= arb_awready;
  
  wwf_inst : axi4_to_adb3w_ocp_dm_fifo
    generic map(
      implementation => wwf_implementation,
      order => wwf_order,
      width => wwf_width)
    port map(
      rst => rst,
      clk => clk,
      wr => wwf_wr,
      d => wwf_d,
      adv => wwf_adv,
      not_full => wwf_not_full,
      q => wwf_q,
      valid => wwf_valid);

  -- wwf_adv is generated in combinatorial part of Write Worker state machine      
  wwf_q_id <= wwf_q(3 downto 0);
  wwf_q_len <= wwf_q(11 downto 4);
  
  
  --
  -- Write Data FIFO
  -- Accepts WDATA & WSTRB from AXI4 W subchannel and presents them to OCP port.
  --

  wdf_wr <= axi_m2s.wvalid and wdf_not_full;
  wdf_d(255 downto 0) <= axi_m2s.wdata;
  wdf_d(287 downto 256) <= axi_m2s.wstrb;
  -- axi_m2s.wlast is ignored.
  
  axi_s2m.wready <= wdf_not_full;
  
  wdf_inst : axi4_to_adb3w_ocp_dm_fifo
    generic map(
      implementation => wdf_implementation,
      order => wdf_order,
      width => wdf_width)
    port map(
      rst => rst,
      clk => clk,
      wr => wdf_wr,
      d => wdf_d,
      adv => wdf_adv,
      not_full => wdf_not_full,
      q => wdf_q,
      valid => wdf_valid);
  
  wdf_adv <= '1' when wdf_valid = '1' and ww_state = ww_state_writing and ocp_s2m.DataAccept = '1' else '0';
  wdf_q_data <= wdf_q(255 downto 0);
  wdf_q_strb <= wdf_q(287 downto 256);
  
  ocp_m2s.Data <= wdf_q_data;
  ocp_m2s.DataByteEn <= wdf_q_strb;
  ocp_m2s.DataValid <= '1' when wdf_valid = '1' and ww_state = ww_state_writing else '0';
    
  
  --
  -- Write Worker
  -- Gets ID + LEN from Write Worker FIFO and outputs BID to Write Response FIFO.
  -- Tracks presentation of Data + DataByteEn on OCP port in order to detect end of burst.
  --
  
  gen_ww_state : process(rst, clk)
  begin
    if rst = '1' then
      ww_state <= ww_state_idle;
      ww_id <= (others => '-');
      ww_last <= '-';
      ww_len <= (others => '-');
    elsif rising_edge(clk) then
      ww_state <= n_ww_state;
      ww_id <= n_ww_id;
      ww_last <= n_ww_last;
      ww_len <= n_ww_len;
    end if;
  end process;
  
  gen_n_ww_state : process(
    ww_state,
    ww_id,
    ww_last,
    ww_len,
    wwf_valid,
    wwf_q_id,
    wwf_q_len,
    wdf_adv,
    wrf_not_full,
    wrf_not_afull,
    ocp_s2m)
  begin
    -- Establish defaults
    wwf_adv <= '0';
    wrf_wr <= '0';
    
    case ww_state is
      when ww_state_idle =>
        if wwf_valid = '1' and wrf_not_full = '1' then
          n_ww_state <= ww_state_writing;
          n_ww_id <= wwf_q_id;
          n_ww_last <= not or_reduce(wwf_q_len);
          n_ww_len <= wwf_q_len;
          wwf_adv <= '1';
        else
          n_ww_state <= ww_state_idle;
          n_ww_id <= (others => '-');
          n_ww_last <= '-';
          n_ww_len <= (others => '-');
        end if;
        
      when ww_state_writing =>
        if wdf_adv = '1' then
          if ww_last = '1' then 
            if minimize_dead_cycles and wwf_valid = '1' and wrf_not_afull = '1' then
              n_ww_state <= ww_state_writing;
              n_ww_id <= wwf_q_id;
              n_ww_last <= not or_reduce(wwf_q_len);
              n_ww_len <= wwf_q_len;
              wwf_adv <= '1';
            else
              n_ww_state <= ww_state_idle;
              n_ww_id <= (others => '-');
              n_ww_last <= '-';
              n_ww_len <= (others => '-');
            end if;
            wrf_wr <= '1';
          else
          n_ww_state <= ww_state_writing;
          n_ww_id <= ww_id;
          n_ww_last <= not or_reduce(ww_len(ww_len'high downto 1));
          n_ww_len <= ww_len - 1;
          end if;
        else
          n_ww_state <= ww_state_writing;
          n_ww_id <= ww_id;
          n_ww_last <= ww_last;
          n_ww_len <= ww_len;
        end if;
    end case;
  end process;
  
  
  --
  -- Write Response FIFO (WRF)
  -- Accepts BID from Write Worker and presents it on AXI4 B subchannel.
  -- BRESP is always "00" (no error).
  --

  -- wrf_wr is generated by Write Worker above  
  wrf_d(3 downto 0) <= ww_id;
  
  wrf_inst : axi4_to_adb3w_ocp_dm_fifo
    generic map(
      implementation => wrf_implementation,
      order => wrf_order,
      width => wrf_width)
    port map(
      rst => rst,
      clk => clk,
      wr => wrf_wr,
      d => wrf_d,
      adv => wrf_adv,
      not_full => wrf_not_full,
      not_afull => wrf_not_afull,
      q => wrf_q,
      valid => wrf_valid);
  
  wrf_adv <= axi_m2s.bready and wrf_valid;
  wrf_q_id <= wrf_q(3 downto 0);

  axi_s2m.bvalid <= wrf_valid;
  axi_s2m.bid <= wrf_q_id;
  axi_s2m.bresp <= "00";

  
end architecture rtl;


library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_misc.all;
use ieee.numeric_std.all;

entity axi4_to_adb3w_ocp_dm_fifo is
  generic(
    implementation : in    string := "FIFO";                      -- One of "FIFO", "REG" or "NONE;
    order          : in    natural range 2 to 6;                  -- Depth is 2**order
    width          : in    natural;                               -- width of FIFO, in bits
    afull_free     : in    natural := 1);                         -- When 'afull_free' or fewer free locations, 'afull' is asserted
  port(
    rst            : in    std_logic;                             -- Asynchronous reset
    clk            : in    std_logic;                             -- Clock
    wr             : in    std_logic;                             -- '1' => write value on 'd' into FIFO
    d              : in    std_logic_vector(width - 1 downto 0);  -- Item to write into FIFO, qualified by 'wr'
    adv            : in    std_logic;                             -- '1' => present next item on 'q' to next item
    full           : out   std_logic;                             -- '1' => full
    not_full       : out   std_logic;                             -- Inverse of 'full'
    afull          : out   std_logic;                             -- '1' => almost full
    not_afull      : out   std_logic;                             -- Inverse of 'afull'
    valid          : out   std_logic;                             -- '1' => not empty
    q              : out   std_logic_vector(width - 1 downto 0)); -- Current item, qualified by 'valid'
end;

architecture rtl of axi4_to_adb3w_ocp_dm_fifo is

begin

  assert implementation = "REG" or implementation = "FIFO"
    report "axi4_to_adb3w_ocp_dm_fifo: implementation '" & implementation & "' is not supported"
    severity failure;
    
  gen_fifo : if implementation = "FIFO" generate
  
    constant DEPTH : natural := 2**order;
    constant FULL_LEVEL : natural := DEPTH - 1;
    constant AFULL_LEVEL : natural := FULL_LEVEL - afull_free;

    subtype ram_data_t is std_logic_vector(width - 1 downto 0);
    type ram_array_t is array(0 to DEPTH - 1) of ram_data_t;

    signal wptr : std_logic_vector(order - 1 downto 0) := (others => '0');
    signal rptr : std_logic_vector(order - 1 downto 0) := (others => '0');
    signal rn : std_logic_vector(order - 1 downto 0) := (others => '0');
    signal not_empty : std_logic;
    signal i_full : std_logic := '0';
    signal i_not_full : std_logic := '1';
    signal i_afull : std_logic := '0';
    signal i_not_afull : std_logic := '1';

    signal qvalid : std_logic := '0';
    signal qce : std_logic;
    signal rd : std_logic;
    signal i_q : std_logic_vector(width - 1 downto 0);

    signal ram : ram_array_t;
    attribute RAM_STYLE : string;
    attribute RAM_STYLE of ram : signal is "distributed";
    signal sel_ram : ram_data_t;

  begin

    valid <= qvalid;
    full <= i_full;
    not_full <= i_not_full;
    afull <= i_afull;
    not_afull <= i_not_afull;
    q <= i_q;

    write_logic : process(rst, clk)
      variable tmp : std_logic_vector(order - 1 downto 0);
    begin
      if rst = '1' then
        wptr <= (others => '0');
        i_not_full <= '1';
        i_full <= '0';
        i_not_afull <= '1';
        i_afull <= '0';
      elsif rising_edge(clk) then
        if wr = '1' then
          wptr <= wptr + 1;
        end if;

        if wr = '1' then
          assert conv_integer(wptr - rptr) < FULL_LEVEL severity error;
          tmp := wptr - rptr + 1;
        else
          tmp := wptr - rptr;
        end if;
        if tmp >= FULL_LEVEL then
          i_full <= '1';
          i_not_full <= '0';
        else
          i_full <= '0';
          i_not_full <= '1';
        end if;
        if tmp >= AFULL_LEVEL then
          i_afull <= '1';
          i_not_afull <= '0';
        else
          i_afull <= '0';
          i_not_afull <= '1';
        end if;
      end if;

      if rising_edge(clk) then
        if wr = '1' then
          ram(conv_integer(wptr)) <= d;
        end if;
      end if;
    end process;

    not_empty <= OR_reduce(rn);
    qce <= adv OR NOT qvalid;
    rd <= (adv OR NOT qvalid) AND not_empty;

    read_logic : process(rst, clk)
    begin
      if rst = '1' then
        rptr <= (others => '0');
      elsif rising_edge(clk) then
        if rd = '1' then
          rptr <= rptr + 1;
        end if;
      end if;

      if rst = '1' then
        null;
      elsif rising_edge(clk) then
        if rd = '1' then
          assert conv_integer(wptr - rptr) > 0 severity failure;
        end if;
      end if;

      if rst = '1' then
        rn <= (others => '0');
      elsif rising_edge(clk) then
        if rd = '1' then
          rn <= wptr - rptr - 1;
        else
          rn <= wptr - rptr;
        end if;
      end if;
    end process;

    output_valid : process(rst, clk)
    begin
      if rst = '1' then
        qvalid <= '0';
      elsif rising_edge(clk) then
        if qce = '1' then
          qvalid <= not_empty;
        end if;
      end if;
    end process;

    sel_ram <= ram(conv_integer(rptr));

    output_data : process(clk)
    begin
      if rising_edge(clk) then
        if qce = '1' then
          -- Assign i_q to undefined on FIFO empty to assist in simulation debug only
          if not_empty = '1' then
            i_q <= sel_ram;
          else
            i_q <= (others => '-');
          end if;
        end if;
      end if;
    end process;

  end generate;

  gen_reg : if implementation = "REG" generate

    signal i_full : std_logic := '0';
    signal i_not_full : std_logic := '0';
    
  begin

    full <= i_full;
    not_full <= i_not_full;
    afull <= '1';
    not_afull <= '0';
    valid <= i_full;
    
    gem_full : process(rst, clk)
      variable tmp : std_logic_vector(1 downto 0);
    begin
      if rst = '1' then
        i_full <= '0';
        i_not_full <= '1';
        q <= (others => '-');
      elsif rising_edge(clk) then
        tmp := wr & adv;
        case tmp is
          when "00" | "11" =>
            null;
            
          when "01" =>
            i_full <= '0';
            i_not_full <= '1';
            
          when "10" =>
            i_full <= '1';
            i_not_full <= '0';
            
          when others =>
            assert false severity error; -- Should never get here
            i_full <= '-';
            i_not_full <= '-';
        end case;
        
        if wr = '1' then
          assert i_full = '0' severity error;
          q <= d;
        end if;
        
        if adv = '1' then
          assert i_full = '1' severity error;
        end if;
      end if;
    end process;
    
  end generate;
end;