--==============================================================================
--
--     File Name: adb3_ocp_dm_adapter_base.vhd
--         Model: Synthesisable
--
--  Dependencies: adb3_ocp (package)
--                adb3_ocp_dm_int (package)
--
--       Company: Alpha Data
--        Design: 
--
--   Description: See header for adb3_ocp_dm_adapter.vhd &
--                adb3_ocp_w_dm_adapter.vhd for details.
--
--   Limitations:
--
--         Notes: 
--
--    Disclaimer: THESE DESIGNS ARE PROVIDED "AS IS" WITH NO WARRANTY
--                WHATSOEVER AND ALPHA DATA SPECIFICALLY DISCLAIMS ANY
--                IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR
--                A PARTICULAR PURPOSE, OR AGAINST INFRINGEMENT.
--
--                Copyright (C) 2014 Alpha Data, All rights reserved
--
--     Revisions: 0.1, TW, 20 Feb 2014: Initial design, created by generalizing
--                  adb3_ocp_dm_adapter.vhd to accept a DATA_WIDTH parameter.
--
--==============================================================================

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_misc.all;
use ieee.numeric_std.all;
use ieee.math_real.all;

library work;
use work.adb3_ocp.all;
use work.adb3_ocp_dm_int.all;

entity adb3_ocp_dm_adapter_base is
generic(
  -- log2(size of the DM-side command FIFO)
  DCFIFO_ORDER    : in    natural range 4 to 8;
  -- log2(size of the OCP-side command FIFO)
  OCFIFO_ORDER    : in    natural range 4 to 8;
  -- log2(size of the DM-side data & byte enable FIFO)
  DWFIFO_ORDER    : in    natural range 4 to 10;
  -- log2(size of the OCP-side data & byte enable FIFO)
  OWFIFO_ORDER    : in    natural range 4 to 8;
  -- true  => Generated write bursts are also compatible with MPTL payload.
  -- false => Generated write bursts in general are not compatible with MPTL
  --          payload.
  MPTL_COMPATIBLE : in    boolean;
  -- true  => For write bursts, non-OCP word address bits of master_m2s_addr
  --          indicate the first valid byte lane of the first word of the burst
  --          and master_m2s_tag (normally ignored by slave) indicates the last
  --          valid byte of the the last word of the burst.
  --
  --          For read bursts, non-OCP word address bits of master_m2s_addr
  --          are 0 and master_m2s.Tag functions normally.
  --
  --          NOTE: See adb3_ocp_dm_adapter / adb3_ocp_w_dm_adapter for
  --          detailed Addr / Tag usage, because it depends upon DATA_WIDTH.
  --
  -- false => Non-OCP word address bits bits of master_m2s_addr are zero and
  --          master_m2s_tag must be considered invalid unless the burst is a
  --          read.
  PCIE_HINTS      : in    boolean;
  -- OCP data width
  DATA_WIDTH      : in    natural);
port(
  ocp_rst         : in    std_logic;
  ocp_clk         : in    std_logic;
  -- OCP slave port
  slave_m2s_cmd   : in    ocp_CmdT;
  slave_m2s_addr  : in    adb3_ocp_addr_s;
  slave_m2s_blen  : in    adb3_ocp_burst_s;
  slave_m2s_tag   : in    adb3_ocp_tag_s;
  slave_m2s_dval  : in    std_logic;
  slave_m2s_data  : in    std_logic_vector(DATA_WIDTH - 1 downto 0);
  slave_m2s_be    : in    std_logic_vector(DATA_WIDTH / 8 - 1 downto 0);
  slave_m2s_racc  : in    std_logic;
  slave_s2m_cacc  : out   std_logic;
  slave_s2m_dacc  : out   std_logic;
  slave_s2m_resp  : out   ocp_RespT;
  slave_s2m_data  : out   std_logic_vector(DATA_WIDTH - 1 downto 0);
  slave_s2m_tag   : out   adb3_ocp_tag_s;
  -- OCP master port
  master_m2s_cmd  : out   ocp_CmdT;
  master_m2s_addr : out   adb3_ocp_addr_s;
  master_m2s_blen : out   adb3_ocp_burst_s;
  master_m2s_tag  : out   adb3_ocp_tag_s;
  master_m2s_dval : out   std_logic;
  master_m2s_data : out   std_logic_vector(DATA_WIDTH - 1 downto 0);
  master_m2s_be   : out   std_logic_vector(DATA_WIDTH / 8 - 1 downto 0);
  master_m2s_racc : out   std_logic;
  master_s2m_cacc : in    std_logic;
  master_s2m_dacc : in    std_logic;
  master_s2m_resp : in    ocp_RespT;
  master_s2m_data : in    std_logic_vector(DATA_WIDTH - 1 downto 0);
  master_s2m_tag  : in    adb3_ocp_tag_s;
  -- Control ports
  ctl_max_rd_req  : in    std_logic_vector(2 downto 0); -- Maximum read request, encoded as per PCI Express specification
  ctl_max_payload : in    std_logic_vector(2 downto 0); -- Maximum payload, encoded as per PCI Express specification
  -- Debug information
  --
  --    [63:0] CHOP
  --         [0] data_enable_q
  --         [1] cmd_enable_q
  --       [4:2] max_rd_req_q
  --       [7:5] max_payload_q
  --      [11:8] cp_state:
  --             0000 => idle
  --             0001 => read1
  --             0010 => read2
  --             0011 => write_normal
  --             0100 => write_general
  --     [23:12] cp_remaining;
  --     [31:24] cp_addr;
  --     [43:32] cp_start_addr;
  --     [52:44] cp_bcount;
  --        [53] cp_bfirst;
  --        [54] cp_blast;
  --        [55] cp_bboundary;
  --     [63:56] cp_dm_tag;
  --
  --   [75:64] OWFIFO:
  --     [73:64] rn
  --        [74] qvalid
  --        [75] nafull
  --
  --   [87:76] OCFIFO:
  --     [85:76] rn
  --        [86] qvalid
  --        [87] nafull
  --
  --   [99:88] RFIFO:
  --     [97:88] rn
  --        [98] qvalid
  --        [99] nafull
  --
  -- [111:100] DCFIFO:
  --    [99:100] rn
  --       [110] qvalid
  --       [111] nafull
  --
  -- [123:112] DWFIFO:
  --    [121:112] rn
  --       [122] qvalid
  --       [123] nafull
  --   
  debug_info      : out   std_logic_vector(127 downto 0));
end entity;


architecture syn of adb3_ocp_dm_adapter_base is

  -- Number of byte lanes.
  constant BE_WIDTH : natural := DATA_WIDTH / 8;
  -- Log base 2 of number of byte lanes.
  constant LOG2_BE_WIDTH : natural := natural(floor(log2(real(BE_WIDTH) + 0.5)));
  
  -- Maximum write burst length that Chopper will output for any single
  -- write command to DM port. This depends upon WFIFO_ORDER because a burst's
  -- length is *not known* until all byte enables for that burst have been
  -- analyzed (and written into DM data & byte enable FIFO).
  --
  -- It is therefore possible that 2**MAX_WR_BURST_ORDER is less than
  -- the value specified by 'ctl_max_payload', but that's OK.
  --
  -- For example, if WFIFO_ORDER is 6, MAX_WR_BURST_ORDER is 5, which
  -- equates to a maximum payload of 2^5 * 16 = 512 bytes. It is not recommended
  constant MAX_WR_BURST_ORDER : natural := DWFIFO_ORDER - 1;

  subtype data_word_t is std_logic_vector(DATA_WIDTH - 1 downto 0);
  subtype be_word_t is std_logic_vector(DATA_WIDTH / 8 - 1 downto 0);
  
  signal owf_data_adv : std_logic;  
  signal owf_data : data_word_t;
  signal owf_be : be_word_t;
  signal owf_data_valid : std_logic;
  
  signal ocf_cmd_adv : std_logic;
  signal ocf_cmd : ocp_CmdT;
  signal ocf_addr : adb3_ocp_addr_s;
  signal ocf_blen : adb3_ocp_burst_s;
  signal ocf_tag : adb3_ocp_tag_s;
  signal ocf_cmd_valid : std_logic;
  
  signal chop_cmd : ocp_CmdT;
  signal chop_addr : adb3_ocp_addr_s;
  signal chop_blen : adb3_ocp_burst_s;
  signal chop_ocp_tag : adb3_ocp_tag_s;
  signal chop_dm_tag : adb3_ocp_tag_s;
  signal chop_cmd_valid : std_logic;
  signal chop_data : data_word_t;
  signal chop_be : be_word_t;
  signal chop_data_valid : std_logic;

  signal dm_resp_valid : std_logic;
  signal dm_resp_acc : std_logic;
  
  signal tm_resp_acc : std_logic;
  signal tm_resp : ocp_RespT;
  signal tm_data : data_word_t;
  signal tm_tag : adb3_ocp_tag_s;
  signal tm_resp_valid : std_logic;

  signal cf_nafull : std_logic;
  
  signal wf_nafull : std_logic;
  
  
begin

  -- Check that data width is a supported value.
  assert DATA_WIDTH = 64 OR DATA_WIDTH = 128 OR DATA_WIDTH = 256
    report "adb3_ocp_dm_adapter_base_base: invalid DATA_WIDTH (" & natural'image(DATA_WIDTH) & ")"
    severity failure;
    
  -- Display a warning if DWFIFO_ORDER is larger than it needs to be (wastes RAM resources).
  assert DWFIFO_ORDER <= 13 - LOG2_BE_WIDTH
    report "adb3_ocp_dm_adapter_base: to avoid wasting RAM resources, DWFIFO_ORDER (" &
           natural'image(DWFIFO_ORDER) & ") should be no greater than " & natural'image(13 - LOG2_BE_WIDTH) &
           " for " & natural'image(BE_WIDTH) & " byte lanes"
    severity warning;
  
  -- Display a warning if chosen DWFIFO_ORDER limits generated PCI-E write payload.
  assert DWFIFO_ORDER >= 13 - LOG2_BE_WIDTH
    report "adb3_ocp_dm_adapter_base: DWFIFO_ORDER (" & natural'image(DWFIFO_ORDER) &
           ") limits generated PCI-E write payload to " & natural'image(2**(DWFIFO_ORDER - 1) * BE_WIDTH) & " bytes"
    severity warning;
  
  --
  -- OCP-side write data / byte enable FIFO
  --
  
  owfifo_0 : adb3_ocp_dm_adapter_owfifo
  generic map(
    ORDER => OWFIFO_ORDER,
    DATA_WIDTH => DATA_WIDTH)
  port map(
    ocp_rst => ocp_rst,
    ocp_clk => ocp_clk,
    ocp_data_valid => slave_m2s_dval,
    ocp_data => slave_m2s_data,
    ocp_be => slave_m2s_be,
    ocp_data_acc => slave_s2m_dacc,
    owf_data_adv => owf_data_adv,
    owf_data => owf_data,
    owf_be => owf_be,
    owf_data_valid => owf_data_valid,
    debug_info => debug_info(75 downto 64));
    
    
  --
  -- OCP-side command FIFO
  --
  
  ocfifo_0 : adb3_ocp_dm_adapter_ocfifo
  generic map(
    ORDER => OWFIFO_ORDER)
  port map(
    ocp_rst => ocp_rst,
    ocp_clk => ocp_clk,
    ocp_cmd => slave_m2s_cmd,
    ocp_addr => slave_m2s_addr,
    ocp_blen => slave_m2s_blen,
    ocp_tag => slave_m2s_tag,
    ocp_cmd_acc => slave_s2m_cacc,
    ocf_cmd_adv => ocf_cmd_adv,
    ocf_cmd => ocf_cmd,
    ocf_addr => ocf_addr,
    ocf_blen => ocf_blen,
    ocf_tag => ocf_tag,
    ocf_cmd_valid => ocf_cmd_valid,
    debug_info => debug_info(87 downto 76));
    
    
  --
  -- OCP to DM command Chopper
  --
  
  chop_0 : adb3_ocp_dm_adapter_chop
  generic map(
    MAX_WR_BURST_ORDER => MAX_WR_BURST_ORDER,
    MPTL_COMPATIBLE => MPTL_COMPATIBLE,
    PCIE_HINTS => PCIE_HINTS,
    DATA_WIDTH => DATA_WIDTH)
  port map(
    ocp_rst => ocp_rst,
    ocp_clk => ocp_clk,
    ocf_cmd_valid => ocf_cmd_valid,
    ocf_cmd => ocf_cmd,
    ocf_addr => ocf_addr,
    ocf_blen => ocf_blen,
    ocf_tag => ocf_tag,
    ocf_cmd_adv => ocf_cmd_adv,
    owf_data => owf_data,
    owf_be => owf_be,
    owf_data_valid => owf_data_valid,
    owf_data_adv => owf_data_adv,
    chop_cmd => chop_cmd,
    chop_addr => chop_addr,
    chop_blen => chop_blen,
    chop_ocp_tag => chop_ocp_tag,
    chop_dm_tag => chop_dm_tag,
    chop_cmd_valid => chop_cmd_valid,
    chop_data => chop_data,
    chop_be => chop_be,
    chop_data_valid => chop_data_valid,
    ctl_cmd_enable => cf_nafull,
    ctl_data_enable => wf_nafull,
    ctl_max_rd_req => ctl_max_rd_req,
    ctl_max_payload => ctl_max_payload,
    debug_info => debug_info(63 downto 0));
  
  dm_resp_valid <= '1' when tm_resp_acc = '1' AND master_s2m_resp /= OCP_RESP_NONE else
                   '0';
  
  
  --
  -- Response tag mapper
  --
  
  tmap_0 : adb3_ocp_dm_adapter_tmap
  generic map(
    DATA_WIDTH => DATA_WIDTH)
  port map(
    ocp_rst => ocp_rst,
    ocp_clk => ocp_clk,
    chop_cmd => chop_cmd,
    chop_addr => chop_addr,
    chop_blen => chop_blen,
    chop_ocp_tag => chop_ocp_tag,
    chop_dm_tag => chop_dm_tag,
    chop_cmd_valid => chop_cmd_valid,
    dm_resp_valid => dm_resp_valid,
    dm_resp => master_s2m_resp,
    dm_data => master_s2m_data,
    dm_tag => master_s2m_tag,
    tm_resp_valid => tm_resp_valid,
    tm_resp => tm_resp,
    tm_data => tm_data,
    tm_tag => tm_tag);
  
  
  --
  -- OCP-side response FIFO
  --
  
  master_m2s_racc <= tm_resp_acc;
  
  resp_fifo_0 : adb3_ocp_dm_adapter_rfifo
  generic map(
    DATA_WIDTH => DATA_WIDTH)
  port map(
    ocp_rst => ocp_rst,
    ocp_clk => ocp_clk,
    tm_resp_valid => tm_resp_valid,
    tm_resp => tm_resp,
    tm_data => tm_data,
    tm_tag => tm_tag,
    tm_resp_acc => tm_resp_acc,
    ocp_resp => slave_s2m_resp,
    ocp_data => slave_s2m_data,
    ocp_tag => slave_s2m_tag,
    ocp_resp_acc => slave_m2s_racc,
    debug_info => debug_info(99 downto 88));
  
  
  --
  -- DM-side command FIFO
  --
  
  dcfifo_0 : adb3_ocp_dm_adapter_dcfifo
  generic map(
    ORDER => DCFIFO_ORDER)
  port map(
    ocp_rst => ocp_rst,
    ocp_clk => ocp_clk,
    chop_cmd => chop_cmd,
    chop_addr => chop_addr,
    chop_blen => chop_blen,
    chop_dm_tag => chop_dm_tag,
    chop_cmd_valid => chop_cmd_valid,
    dm_cmd => master_m2s_cmd,
    dm_addr => master_m2s_addr,
    dm_blen => master_m2s_blen,
    dm_tag => master_m2s_tag,
    dm_cmd_acc => master_s2m_cacc,
    nafull => cf_nafull,
    debug_info => debug_info(111 downto 100));

  
  --
  -- DM-side write data / byte enable FIFO
  --
  
  dwfifo_0 : adb3_ocp_dm_adapter_dwfifo
  generic map(
    ORDER => DWFIFO_ORDER,
    DATA_WIDTH => DATA_WIDTH)
  port map(
    ocp_rst => ocp_rst,
    ocp_clk => ocp_clk,
    chop_data => chop_data,
    chop_be => chop_be,
    chop_data_valid => chop_data_valid,
    dm_data_acc => master_s2m_dacc,
    dm_data => master_m2s_data,
    dm_be => master_m2s_be,
    dm_data_valid => master_m2s_dval,
    nafull => wf_nafull,
    debug_info => debug_info(123 downto 112));
    

  --
  -- Unused debug / status bits
  --
  
  debug_info(127 downto 124) <= (others => '0');
  
end;
