--==============================================================================
--
--     File Name: adb3_ocp_dm_adapter_sfifo.vhd
--         Model: Synthesisable
--
--  Dependencies:
--
--       Company: Alpha Data
--        Design: adb3_ocp_dm_adapter
--
--   Description: Synchronous FIFO used in adb3_ocp_dm_adapter.
--
--   Limitations: 
--
--         Notes: 
--
--    Disclaimer: THESE DESIGNS ARE PROVIDED "AS IS" WITH NO WARRANTY
--                WHATSOEVER AND ALPHA DATA SPECIFICALLY DISCLAIMS ANY
--                IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR
--                A PARTICULAR PURPOSE, OR AGAINST INFRINGEMENT.
--
--                Copyright (C) 2013 Alpha Data, All rights reserved
--
--     Revisions: 0.1, TW, 26 Nov 2013: Initial design
--
--==============================================================================

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_misc.all;
use ieee.numeric_std.all;

--
-- A synchronous FIFO used in the 'adb3_ocp_dm_adapter' module.
--
-- 'ORDER' is log2 of the FIFO depth, e.g. ORDER = 4 => depth is 16 locations.
--
-- 'WIDTH' is the width of FIFO locations, in bits.
--
-- 'EMPTY_VAL' is the value driven on 'rq(EMPTY_LO:EMPTY_HI-1)' when 'valid' is
-- not asserted; map both 'EMPTY_LO' and 'EMPTY_HI' to 0 if this feature is
-- not required.
--
-- The FIFO is "almost full" (nafull = '0') at approximately 75% of depth.
--

entity adb3_ocp_dm_adapter_sfifo is
generic(
  ORDER       : in    natural range 4 to 10;
  WIDTH       : in    natural;
  EMPTY_VAL   : in    std_logic := '0';
  EMPTY_LO    : in    natural := 0;
  EMPTY_HI    : in    natural := 0);
port(
  rst         : in    std_logic;                             -- Asynchronous reset
  sr          : in    std_logic;                             -- Synchronous reset
  clk         : in    std_logic;                             -- Clock
  wen         : in    std_logic;                             -- '1' => write value on 'wd' into FIFO
  wd          : in    std_logic_vector(WIDTH - 1 downto 0);  -- Item to write into FIFO, qualified by 'wen'
  wafull      : out   std_logic;                             -- '1' => almost full
  wnafull     : out   std_logic;                             -- Inverse of 'wafull'
  radv        : in    std_logic;                             -- '1' => present next item on 'rq'
  rvalid      : out   std_logic;                             -- '1' => not empty
  rq          : out   std_logic_vector(WIDTH - 1 downto 0);  -- Current item, qualified by 'rvalid'
  -- Debug information:
  --    [8:0] rn
  --      [9] qvalid
  --     [10] ovf;
  --     [11] udf;
  debug_info  : out   std_logic_vector(11 downto 0));
end;

architecture syn of adb3_ocp_dm_adapter_sfifo is

  function determine_ram_style(
    constant order : in natural)
  return string is
  begin
    if order > 8 then
      return "BLOCK";
    else
      return "DISTRIBUTED";
    end if;
  end;
  
  constant DEPTH : natural := 2**ORDER;
  constant MAX_LEVEL : natural := DEPTH - 1;

  -- Generate initial value for rq output
  constant RQ_TOP  : std_logic_vector(WIDTH - 1 downto EMPTY_HI)    := (others => '-');
  constant RQ_MID  : std_logic_vector(EMPTY_HI - 1 downto EMPTY_LO) := (others => EMPTY_VAL);
  constant RQ_LOW  : std_logic_vector(EMPTY_LO - 1 downto 0)        := (others => '-');
  constant RQ_INIT : std_logic_vector(WIDTH - 1 downto 0) := RQ_TOP & RQ_MID & RQ_LOW;

  subtype ram_data_t is std_logic_vector(WIDTH - 1 downto 0);
  type ram_array_t is array(0 to DEPTH - 1) of ram_data_t;

  signal wptr : std_logic_vector(ORDER - 1 downto 0) := (others => '0');
  signal rptr : std_logic_vector(ORDER - 1 downto 0) := (others => '0');
  signal rn : std_logic_vector(ORDER - 1 downto 0) := (others => '0');
  signal not_empty : std_logic;
  signal i_full : std_logic := '0';
  signal i_nfull : std_logic := '1';
  signal ovf : std_logic := '0';
  signal udf : std_logic := '0';

  signal qvalid : std_logic := '0';
  signal qce : std_logic;
  signal rd : std_logic;
  signal i_rq : std_logic_vector(WIDTH - 1 downto 0) := RQ_INIT;

  signal ram : ram_array_t;
  attribute RAM_STYLE : string;
  attribute RAM_STYLE of ram : signal is determine_ram_style(ORDER);
  signal sel_ram : ram_data_t;

begin

  debug_state : process(clk)
  begin
    if rising_edge(clk) then
      debug_info(8 downto 0) <= std_logic_vector(resize(unsigned(rn), 9));
      debug_info(9) <= qvalid;
      debug_info(10) <= ovf;
      debug_info(11) <= udf;
    end if;
  end process;
  
  rvalid <= qvalid;
  wafull <= i_full;
  wnafull <= i_nfull;
  rq     <= i_rq;

  write_logic : process(rst, clk)
    variable tmp : std_logic_vector(ORDER - 1 downto 0);
    variable full : std_logic;
  begin
    if rst = '1' then
      wptr <= (others => '0');
      i_full <= '0';
      i_nfull <= '1';
      ovf <= '0';
    elsif rising_edge(clk) then
      if sr = '1' then
        wptr <= (others => '0');
        i_full <= '0';
        i_nfull <= '1';
        ovf <= '0';
      else
        if wen = '1' then
          wptr <= wptr + 1;
        end if;

        if wen = '1' then
          assert conv_integer(wptr - rptr) < MAX_LEVEL report "FIFO overflow" severity error;
          tmp := wptr - rptr + 1;
        else
          tmp := wptr - rptr;
        end if;
        full := AND_REDUCE(tmp(ORDER - 1 downto ORDER - 2));
        i_full <= full;
        i_nfull <= NOT full;
        
        tmp := wptr - rptr;
        if wen = '1' and AND_reduce(wptr - rptr) = '1' then
          ovf <= '1';
        end if;
      end if;
    end if;

    if rising_edge(clk) then
      if wen = '1' then
        ram(conv_integer(wptr)) <= wd;
      end if;
    end if;
  end process;

  not_empty <= OR_reduce(rn);
  qce <= radv OR NOT qvalid;
  rd <= (radv OR NOT qvalid) AND not_empty;

  read_logic : process(rst, clk)
  begin
    if rising_edge(clk) then
      -- Asserting 'radv' when 'qvalid' not asserted is an error
      assert NOT (radv = '1' AND qvalid /= '1') report "radv asserted without rvalid" severity error;
    end if;

    if rst = '1' then
      rptr <= (others => '0');
    elsif rising_edge(clk) then
      if sr = '1' then
        rptr <= (others => '0');
      else
        if rd = '1' then
          rptr <= rptr + 1;
        end if;
      end if;
    end if;

    if rst = '1' then
      null;
    elsif rising_edge(clk) then
      if sr = '1' then
        null;
      else
        if rd = '1' then
          assert conv_integer(wptr - rptr) > 0 severity failure;
        end if;
      end if;
    end if;

    if rising_edge(clk) then
      if rd = '1' then
        rn <= wptr - rptr - 1;
      else
        rn <= wptr - rptr;
      end if;
    end if;
  end process;

  output_valid : process(rst, clk)
  begin
    if rst = '1' then
      qvalid <= '0';
      udf <= '0';
    elsif rising_edge(clk) then
      if sr = '1' then
        qvalid <= '0';
        udf <= '0';
      else
        if qce = '1' then
          qvalid <= not_empty;
        end if;
        
        if radv = '1' AND qvalid = '0' then
          udf <= '1';
        end if;
      end if;
    end if;
  end process;

  sel_ram <= ram(conv_integer(rptr));

  output_data : for i in 0 to WIDTH - 1 generate
    with_reset : if EMPTY_LO <= i AND i < EMPTY_HI generate
      U0 : process(rst, clk)
      begin
        if rst = '1' then
          i_rq(i) <= EMPTY_VAL;
        elsif rising_edge(clk) then
          if sr = '1' then
            i_rq(i) <= EMPTY_VAL;
          else
            if qce = '1' then
              if not_empty = '1' then
                i_rq(i) <= sel_ram(i);
              else
                i_rq(i) <= EMPTY_VAL;
              end if;
            end if;
          end if;
        end if;
      end process;
    end generate;

    without_reset : if NOT (EMPTY_LO <= i AND i < EMPTY_HI) generate
      U0 : process(clk)
      begin
        if rising_edge(clk) then
          if qce = '1' then
            -- Assign i_rq to undefined on FIFO empty to assist in simulation debug only
            if not_empty = '1' then
              i_rq(i) <= sel_ram(i);
            else
              i_rq(i) <= '-';
            end if;
          end if;
        end if;
      end process;
    end generate;
  end generate;

end;
