library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_misc.all;
use ieee.numeric_std.all;
use ieee.math_real.all;

library work;

library unisim;
use unisim.vcomponents.all;


entity avr2_proc_interface_v1 is
  generic (
    vpd_max_size        : natural := 512;
    sns_max_size        : natural := 64;
    mez_max_size        : natural := 512
  );
  port (
    -- Physical ports
    rx_ser_data         : in  std_logic;
    tx_ser_data         : out std_logic;
    ck_ser_data         : in  std_logic;
    -- Clock, Reset, Interrupt
    rst                 : in  std_logic;
    reg_clk             : in  std_logic;
    irq_evnt_srvc       : out std_logic;
    irq_coms_rx         : out std_logic;
    irq_coms_tx         : out std_logic;
    irq_alarm_thrm_crit : out std_logic; -- Thermal Alarm, Critical ( BoardMan2 will take action )
    irq_alarm_thrm_warn : out std_logic; -- Thermal Alarm, Warning
    irq_alarm_sppl_crit : out std_logic; -- Supply Alarm, Critical ( ... )
    irq_alarm_sppl_warn : out std_logic; -- Supply Alarm, Warning
    -- Serial Number Output
    serial_no           : out std_logic_vector(31 downto 0);
    serial_no_valid     : out std_logic;
    -- VPD Forwarding
    vpd_forward_data    : out std_logic_vector(7 downto 0);
    vpd_forward_addr    : out std_logic_vector(4 downto 0);
    vpd_forward_be      : out std_logic_vector(15 downto 0);
    -- Register interface
    write               : in    std_logic;
    be                  : in    std_logic_vector(15 downto 0);
    data                : in    std_logic_vector(127 downto 0);
    status              : out   std_logic_vector(127 downto 0);
    avr_bram_vpd_a      : in    std_Logic_vector(8 downto 4);
    avr_bram_vpd_q      : out   std_logic_vector(127 downto 0);
    avr_bram_mez_a      : in    std_Logic_vector(8 downto 4);
    avr_bram_mez_q      : out   std_logic_vector(127 downto 0);
    avr_bram_sns_a      : in    std_Logic_vector(8 downto 4);
    avr_bram_sns_q      : out   std_logic_vector(127 downto 0)
  );
end entity;

architecture rtl of avr2_proc_interface_v1 is

  signal tx_d               : std_logic_vector( 9 downto 0 );
  signal tx_w               : std_logic := '0';
  signal tx_d_mux               : std_logic_vector( 9 downto 0 );
  signal tx_w_mux               : std_logic := '0';
  signal tx_d_auto               : std_logic_vector( 9 downto 0 );
  signal tx_w_auto               : std_logic := '0';
  signal tx_d_latch               : std_logic_vector( 9 downto 0 );
  signal tx_w_latch               : std_logic := '0';

  signal tx_d_auto_counter  : unsigned(7 downto 0) := (others => '1');
  signal tx_d_auto_valid : std_logic := '1';
  
  signal tx_e               : std_logic;
  signal tx_f               : std_logic;
  signal rx_d               : std_logic_vector( 9 downto 0 );
  signal rx_r               : std_logic := '0';
  signal rx_e               : std_logic;
  signal rx_f               : std_logic;
  
  signal rx_intrnl_data     : std_logic_vector( 7 downto 0 );
  signal rx_intrnl_empty    : std_logic;
  signal rx_intrnl_soh      : std_logic;
  signal rx_intrnl_eot      : std_logic;
  signal rx_intrnl_rd       : std_logic;
  signal rx_intrnl_full     : std_logic;
  
  signal vpd_update         : std_logic := '0';
  signal sns_update         : std_logic := '0';
  
  signal rx_intrnl_rd_pipe_0  : std_logic := '0';
  
  procedure reg_write( 
    signal   reg     : inout   std_logic_vector;
    signal   data    : in      std_logic_vector;
    signal   byte_en : in      std_logic_vector;
    constant in_off  : in      natural
  ) is
  begin
    for i in reg'left downto reg'right loop
      if byte_en( (i+in_off) / 8 ) = '1' then
        reg(i) <= data(i+in_off);
      end if;
    end loop;
  end procedure;  
  
  procedure reg_write( 
    signal   reg     : inout   std_logic;
    signal   data    : in      std_logic_vector;
    signal   byte_en : in      std_logic_vector;
    constant in_off  : in      natural
  ) is
  begin
    if byte_en( (in_off) / 8 ) = '1' then
      reg <= data(in_off);
    end if;
  end procedure;  
  
  subtype rom_word is std_logic_vector(127 downto 0);
  
  constant vpd_column_left  : natural := natural(ceil(log2(real(vpd_max_size / 16)))) - 1;
  constant sns_column_left  : natural := natural(ceil(log2(real(sns_max_size / 16)))) - 1;
  constant mez_column_left  : natural := natural(ceil(log2(real(mez_max_size / 16)))) - 1;
  
  type vpd_data_rom is array ( vpd_max_size / 16 - 1 downto 0 ) of rom_word;
  type sns_data_rom is array ( sns_max_size / 16 - 1 downto 0 ) of rom_word; 
  type mez_data_rom is array ( mez_max_size / 16 - 1 downto 0 ) of rom_word; 
  
  --type vpd_data_rom is array ( 15 downto 0 ) of vpd_rom_column;
  --type sns_data_rom is array ( 15 downto 0 ) of sns_rom_column;
  --type mez_data_rom is array ( 15 downto 0 ) of mez_rom_column;
  
  signal vpd_rom : vpd_data_rom := ( others => X"CDCDCDCD_CDCDCDCD_CDCDCDCD_CDCDCDCD" );
  signal mez_rom : mez_data_rom := ( others => X"CDCDCDCD_CDCDCDCD_CDCDCDCD_CDCDCDCD" );
  signal sns_rom : sns_data_rom := ( others => X"CDCDCDCD_CDCDCDCD_CDCDCDCD_CDCDCDCD" );
  
  signal vpd_addresss     : std_logic_vector( vpd_column_left downto 0 );
  signal sns_addresss     : std_logic_vector( sns_column_left downto 0 );
  signal mez_addresss     : std_logic_vector( mez_column_left downto 0 );
        
  signal rx_data_pipe_0   : std_logic_vector( 9 downto 0 );
  
  signal driver_fifo_wr   : std_logic := '0';
  
  type rx_decode_state is (
    layer_2_soh_detect,
    layer_3_auto,
    layer_3_auto_ext,
    layer_3_auto_ext_sensor,
    layer_3_auto_ext_vpd,
    layer_3_auto_ext_mez0,
    layer_3_pass_through,
    layer_3_unknown
  );
  
  signal rx_decode : rx_decode_state := layer_2_soh_detect;
  signal n_rx_decode : rx_decode_state;
  signal p_rx_decode : rx_decode_state := layer_2_soh_detect;
  
  signal n_mode_service_detect      : std_logic;
  signal n_mode_application_detect  : std_logic;
  signal mode_service_detect      : std_logic := '0';
  signal mode_application_detect  : std_logic := '0';
  signal firmware_checksum_error_detect : std_logic := '0';
  signal n_firmware_checksum_error_detect : std_logic := '0';
  
  signal reg_irq_evnt_srvc       : std_logic := '0';
  signal reg_irq_coms_rx         : std_logic := '0';
  signal reg_irq_coms_tx         : std_logic := '0';
  signal reg_irq_alarm_thrm_crit : std_logic := '0';
  signal reg_irq_alarm_thrm_warn : std_logic := '0';
  signal reg_irq_alarm_sppl_crit : std_logic := '0';
  signal reg_irq_alarm_sppl_warn : std_logic := '0';
  
  signal alarm_detect_vw         : std_logic := '0';
  signal alarm_detect_vc         : std_logic := '0';
  signal alarm_detect_tw         : std_logic := '0';
  signal alarm_detect_tc         : std_logic := '0';
  signal n_alarm_detect_vw       : std_logic;
  signal n_alarm_detect_vc       : std_logic;
  signal n_alarm_detect_tw       : std_logic;
  signal n_alarm_detect_tc       : std_logic;
  
  signal rx_data_pipe_0_ext      : std_logic_vector( 31 downto 0 ) := ( others => '-' );
  signal rx_d_ext                : std_logic_vector( 31 downto 0 );
  
  signal tx_reset                : std_logic;
  
  signal rst_or_prst             : std_logic;
  signal por_vec                 : std_logic_vector( 2 downto 0 ) := ( others => '1' );
  signal por_rst                 : std_logic := '1';
  
  signal drv_rx_vec              : std_logic_vector( 2 downto 0 ) := ( others => '1' );
  signal drv_rx_rst              : std_logic := '1';
  
  signal reg_serial_no           : std_logic_vector(31 downto 0) := ( others => '-' );
  signal reg_serial_no_valid     : std_logic := '0';
  
  signal tx_error                : std_logic;
  signal service_mode            : std_logic := '0';
  
  signal rom_addr                : std_logic_vector( 15 downto 0 ) := ( others => '-' );
  signal n_rom_addr              : std_logic_vector( 15 downto 0 );
  
  signal rom_wr                  : std_logic_vector( 2 downto 0 ) := ( others => '-' );


  signal vpd_addr : std_logic_vector(7  downto 0) := (others => '0');
  signal vpd_sel : std_logic_vector(15 downto 0) := (others => '0');
  signal vpd_data : std_logic_vector(7 downto 0) := (others => '0');
  
  signal vpd_updated              : std_logic := '0';
  signal sns_updated              : std_logic := '0';
  signal mez_updated              : std_logic := '0';
  
  signal firmware_checksum_error  : std_logic := '0';
  signal firmware_checksum_error_clr  : std_logic := '0';
  
  signal n_vpd_updated            : std_logic;
  signal n_sns_updated            : std_logic;
  signal n_mez_updated            : std_logic;
  
  signal sns_updated_ack          : std_logic := '0';
  signal vpd_updated_ack          : std_logic := '0';
  signal mez_updated_ack          : std_logic := '0';
  
  signal driver_rd_cnt            : std_logic_vector( 11 downto 0 ) := (others => '0');
  
  signal rx_error                 : std_logic;

  signal rx_rst_sart              : std_logic := '0';
  signal rx_rst_driver            : std_logic := '0';
  
  signal reg_status               : std_logic_vector( 127 downto 0 );
  
  signal rst_or_prst_or_driver_rst    : std_logic;

  signal crd_stat : std_logic_vector(31 downto 0) := (others => '0');

  signal force_gen1 : std_logic := '0';
  
begin

  -- TODO:
  -- - Forward VPD
  --   -- Decode serial number

  serial_no       <= reg_serial_no;
  serial_no_valid <= reg_serial_no_valid;

  rst_or_prst <= rst or por_rst;
  
  p_tx_por : process( reg_clk )
  begin
    if rising_edge( reg_clk ) then
      if rst = '1' then
        por_rst <= '1';
        por_vec <= ( others => '1' );
      else
        if por_vec = 0 then
          por_vec <= por_vec;
          por_rst <= '0';
        else
          por_vec <= por_vec - 1;
          por_rst <= por_rst;
        end if;
      end if;
    end if;
  end process;
  
  p_rx_drv_rst : process( reg_clk )
  begin
    if rising_edge( reg_clk ) then
      if rx_rst_driver = '1' then
        drv_rx_rst <= '1';
        drv_rx_vec <= ( others => '1' );
      else
        if drv_rx_vec = 0 then
          drv_rx_vec <= drv_rx_vec;
          drv_rx_rst <= '0';
        else
          drv_rx_vec <= drv_rx_vec - 1;
          drv_rx_rst <= drv_rx_rst;
        end if;
      end if;
    end if;
  end process;  

  irq_evnt_srvc       <= reg_irq_evnt_srvc;
  irq_coms_rx         <= reg_irq_coms_rx;
  irq_coms_tx         <= reg_irq_coms_tx;
  irq_alarm_thrm_crit <= reg_irq_alarm_thrm_crit;
  irq_alarm_thrm_warn <= reg_irq_alarm_thrm_warn;
  irq_alarm_sppl_crit <= reg_irq_alarm_sppl_crit;
  irq_alarm_sppl_warn <= reg_irq_alarm_sppl_warn;
  
  assert ( vpd_max_size rem 16 ) = 0
         report "(0) vpd_max_size must be greater than zero, a multiple of 16, and a power of 2 " severity failure;
  assert ( sns_max_size rem 16 ) = 0
         report "(1) sns_max_size must be greater than zero, a multiple of 16, and a power of 2 " severity failure;
  assert vpd_max_size /= 0
         report "(2) vpd_max_size must be greater than zero, a multiple of 16, and a power of 2 " severity failure;
  assert sns_max_size /= 0
         report "(3) sns_max_size must be greater than zero, a multiple of 16, and a power of 2 " severity failure;
  assert ( vpd_max_size / 16 ) = 2 ** natural(ceil(log2(real(vpd_max_size / 16))))
         report "(4) vpd_max_size must be greater than zero, a multiple of 16, and a power of 2 " severity failure;
  assert ( sns_max_size / 16 ) = 2 ** natural(ceil(log2(real(sns_max_size / 16))))
         report "(5) sns_max_size must be greater than zero, a multiple of 16, and a power of 2 " severity failure;

  vpd_addresss <= avr_bram_vpd_a( vpd_addresss'left + 4 downto 4 );
  sns_addresss <= avr_bram_sns_a( sns_addresss'left + 4 downto 4 );
  mez_addresss <= avr_bram_mez_a( mez_addresss'left + 4 downto 4 );

  --p_avr_rom : process ( reg_clk )
  --begin
  --  if rising_edge( reg_clk ) then

  --    avr_bram_vpd_q <= vpd_rom(15)( to_integer( unsigned( vpd_addresss ) ) )
  --                    & vpd_rom(14)( to_integer( unsigned( vpd_addresss ) ) )
  --                    & vpd_rom(13)( to_integer( unsigned( vpd_addresss ) ) )
  --                    & vpd_rom(12)( to_integer( unsigned( vpd_addresss ) ) )
  --                    & vpd_rom(11)( to_integer( unsigned( vpd_addresss ) ) )
  --                    & vpd_rom(10)( to_integer( unsigned( vpd_addresss ) ) )
  --                    & vpd_rom(09)( to_integer( unsigned( vpd_addresss ) ) )
  --                    & vpd_rom(08)( to_integer( unsigned( vpd_addresss ) ) )
  --                    & vpd_rom(07)( to_integer( unsigned( vpd_addresss ) ) )
  --                    & vpd_rom(06)( to_integer( unsigned( vpd_addresss ) ) )
  --                    & vpd_rom(05)( to_integer( unsigned( vpd_addresss ) ) )
  --                    & vpd_rom(04)( to_integer( unsigned( vpd_addresss ) ) )
  --                    & vpd_rom(03)( to_integer( unsigned( vpd_addresss ) ) )
  --                    & vpd_rom(02)( to_integer( unsigned( vpd_addresss ) ) )
  --                    & vpd_rom(01)( to_integer( unsigned( vpd_addresss ) ) )
  --                    & vpd_rom(00)( to_integer( unsigned( vpd_addresss ) ) );

  --    avr_bram_sns_q <= sns_rom(15)( to_integer( unsigned( sns_addresss ) ) )
  --                    & sns_rom(14)( to_integer( unsigned( sns_addresss ) ) )
  --                    & sns_rom(13)( to_integer( unsigned( sns_addresss ) ) )
  --                    & sns_rom(12)( to_integer( unsigned( sns_addresss ) ) )
  --                    & sns_rom(11)( to_integer( unsigned( sns_addresss ) ) )
  --                    & sns_rom(10)( to_integer( unsigned( sns_addresss ) ) )
  --                    & sns_rom(09)( to_integer( unsigned( sns_addresss ) ) )
  --                    & sns_rom(08)( to_integer( unsigned( sns_addresss ) ) )
  --                    & sns_rom(07)( to_integer( unsigned( sns_addresss ) ) )
  --                    & sns_rom(06)( to_integer( unsigned( sns_addresss ) ) )
  --                    & sns_rom(05)( to_integer( unsigned( sns_addresss ) ) )
  --                    & sns_rom(04)( to_integer( unsigned( sns_addresss ) ) )
  --                    & sns_rom(03)( to_integer( unsigned( sns_addresss ) ) )
  --                    & sns_rom(02)( to_integer( unsigned( sns_addresss ) ) )
  --                    & sns_rom(01)( to_integer( unsigned( sns_addresss ) ) )
  --                    & sns_rom(00)( to_integer( unsigned( sns_addresss ) ) );

  --    avr_bram_mez_q <= mez_rom(15)( to_integer( unsigned( mez_addresss ) ) )
  --                    & mez_rom(14)( to_integer( unsigned( mez_addresss ) ) )
  --                    & mez_rom(13)( to_integer( unsigned( mez_addresss ) ) )
  --                    & mez_rom(12)( to_integer( unsigned( mez_addresss ) ) )
  --                    & mez_rom(11)( to_integer( unsigned( mez_addresss ) ) )
  --                    & mez_rom(10)( to_integer( unsigned( mez_addresss ) ) )
  --                    & mez_rom(09)( to_integer( unsigned( mez_addresss ) ) )
  --                    & mez_rom(08)( to_integer( unsigned( mez_addresss ) ) )
  --                    & mez_rom(07)( to_integer( unsigned( mez_addresss ) ) )
  --                    & mez_rom(06)( to_integer( unsigned( mez_addresss ) ) )
  --                    & mez_rom(05)( to_integer( unsigned( mez_addresss ) ) )
  --                    & mez_rom(04)( to_integer( unsigned( mez_addresss ) ) )
  --                    & mez_rom(03)( to_integer( unsigned( mez_addresss ) ) )
  --                    & mez_rom(02)( to_integer( unsigned( mez_addresss ) ) )
  --                    & mez_rom(01)( to_integer( unsigned( mez_addresss ) ) )                   
  --                    & mez_rom(00)( to_integer( unsigned( mez_addresss ) ) );
  --  end if;
  --end process;
  
  p_avr_reg : process ( reg_clk )
  begin
    if rising_edge( reg_clk ) then
      if write = '1' then
	  
	  reg_write( firmware_checksum_error_clr, data, be, 35 );
        reg_write( rx_rst_sart, data, be, 33 );
        reg_write( rx_rst_driver, data, be, 32 );
        reg_write( sns_updated_ack, data, be, 31 );
        reg_write( vpd_updated_ack, data, be, 30 );
        reg_write( mez_updated_ack, data, be, 29 );
        reg_write( tx_reset, data, be, 18 );
        reg_write( rx_r, data, be, 11 );
        reg_write( tx_w, data, be, 10 );
        reg_write( tx_d, data, be, 0 );
      else
        tx_w <= '0';
        rx_r <= '0';
        tx_reset <= '0';
        rx_rst_sart <= '0';
        rx_rst_driver <= '0';
        sns_updated_ack <= '0';
        vpd_updated_ack <= '0';
        mez_updated_ack <= '0';
		firmware_checksum_error_clr <= '0';
      end if;
    end if;
  end process;
  
  status <= reg_status;
  reg_status( reg_status'left downto 96 ) <= ( others => '0' );
  reg_status( 95 downto 64 ) <= crd_stat;
  reg_status( 63) <= force_gen1;
  reg_status( 62 downto 36 ) <= ( others => '0' );
  
  reg_status( 35 ) <= firmware_checksum_error;
  reg_status( 34 ) <= rx_error;
  reg_status( 33 ) <= rx_rst_sart;
  reg_status( 32 ) <= rx_rst_driver;
  
  reg_status( 31 ) <= sns_updated;
  reg_status( 30 ) <= vpd_updated;
  reg_status( 29 ) <= mez_updated;
  reg_status( 28 downto 19 ) <= driver_rd_cnt( 9 downto 0 );
  reg_status( 18 ) <= tx_reset;
  reg_status( 17 ) <= service_mode;
  reg_status( 16 ) <= tx_error;
  reg_status( 15 ) <= tx_e;
  reg_status( 14 ) <= tx_f;
  reg_status( 13 ) <= rx_f;
  reg_status( 12 ) <= rx_e;
  reg_status( 11 ) <= rx_r; -- Should never be seen as '1' as write is self clearning
  reg_status( 10 ) <= tx_w; -- Should never be seen as '1' as write is self clearning
  reg_status( 9 downto 0 ) <= rx_d;

  i_common_uart : entity work.sysmon2_layer1 port map
    (
      ser_rx      => rx_ser_data,
      ser_tx      => tx_ser_data,
      ser_ck      => ck_ser_data,
      clk         => reg_clk,
      rst         => rst_or_prst,
      tx_rst      => tx_reset,
      tx_data     => tx_d_mux( 7 downto 0 ),
      tx_soh      => tx_d_mux( 8 ),
      tx_eot      => tx_d_mux( 9 ),
      tx_err      => tx_error,
      tx_wr       => tx_w_mux,
      tx_full     => tx_f,
      tx_empty    => tx_e,
      rx_rst      => rx_rst_sart,
      rx_data     => rx_intrnl_data,
      rx_soh      => rx_intrnl_soh,
      rx_eot      => rx_intrnl_eot,
      rx_rd       => rx_intrnl_rd,
      rx_err      => rx_error,
      rx_full     => rx_intrnl_full,
      rx_empty    => rx_intrnl_empty
    );  
    
  rx_intrnl_rd <= not rx_intrnl_empty;
  
  p_layer_2_3_sync : process( reg_clk )
  begin
    if rising_edge( reg_clk ) then 
      rx_data_pipe_0 <= rx_intrnl_eot & rx_intrnl_soh & rx_intrnl_data;
      p_rx_decode <= rx_decode;
      rx_intrnl_rd_pipe_0 <= rx_intrnl_rd;
      if rst_or_prst = '1' then
        rx_decode <= layer_2_soh_detect;
        mode_service_detect <= '0';
        firmware_checksum_error_detect <= '0';
        mode_application_detect <= '0';
        alarm_detect_vw <= '0';
        alarm_detect_vc <= '0';
        alarm_detect_tw <= '0';
        alarm_detect_tc <= '0';
        reg_irq_evnt_srvc <= '0';
        reg_irq_coms_rx <= '0';
        reg_irq_coms_tx <= '0';
        reg_irq_alarm_thrm_crit <= '0';
        reg_irq_alarm_thrm_warn <= '0';
        reg_irq_alarm_sppl_crit <= '0';
        reg_irq_alarm_sppl_warn <= '0';
        rom_addr <= ( others => '0' );
        firmware_checksum_error <= '0';
      else
        -- Update state machine signals
        rx_decode <= n_rx_decode; 
        mode_service_detect <= n_mode_service_detect;
        firmware_checksum_error_detect <= n_firmware_checksum_error_detect;
        mode_application_detect <= n_mode_application_detect;
        alarm_detect_vw <= n_alarm_detect_vw;
        alarm_detect_vc <= n_alarm_detect_vc;
        alarm_detect_tw <= n_alarm_detect_tw;
        alarm_detect_tc <= n_alarm_detect_tc;
        -- Update interrupts
        reg_irq_evnt_srvc <= mode_service_detect or mode_application_detect;
        reg_irq_coms_rx <= not rx_e;
        reg_irq_coms_tx <= tx_e or tx_error;
        reg_irq_alarm_thrm_crit <= alarm_detect_tc;
        reg_irq_alarm_thrm_warn <= alarm_detect_tw;
        reg_irq_alarm_sppl_crit <= alarm_detect_vc;
        reg_irq_alarm_sppl_warn <= alarm_detect_vw;
        rom_addr <= n_rom_addr;
        firmware_checksum_error <= ( firmware_checksum_error or firmware_checksum_error_detect ) and not firmware_checksum_error_clr;
      end if;
      service_mode <= ( service_mode or mode_service_detect ) and not mode_application_detect;
      vpd_updated <= n_vpd_updated;
      sns_updated <= n_sns_updated;
      mez_updated <= n_mez_updated;       
    end if;
  end process;
  
  p_layer_2_3_async : process( rx_decode, rx_intrnl_soh, rx_intrnl_empty, rx_intrnl_data, rx_intrnl_eot )
  begin
    case rx_decode is
    
      when layer_2_soh_detect =>
        if rx_intrnl_soh = '1' and rx_intrnl_eot = '0' and rx_intrnl_empty = '0' then
          case rx_intrnl_data is
            when X"00" => -- Auto
              n_rx_decode <= layer_3_auto;
            when X"02" => -- Auto Ext
              n_rx_decode <= layer_3_auto_ext;
            when X"03" => -- User ( passthrough entire packet )
              n_rx_decode <= layer_3_pass_through;
            when X"05" => -- User Ext ( passthrough entire packet )
              n_rx_decode <= layer_3_pass_through;
            when X"06" => -- Boot Flash ( passthrough entire packet )
              n_rx_decode <= layer_3_pass_through;
            when others =>
              n_rx_decode <= layer_3_unknown;
          end case;
        else
          n_rx_decode <= layer_2_soh_detect;
        end if;
        n_mode_service_detect <= '0'; n_mode_application_detect <= '0'; n_firmware_checksum_error_detect <= '0';
        n_alarm_detect_vw <= '0'; n_alarm_detect_vc <= '0'; n_alarm_detect_tw <= '0'; n_alarm_detect_tc <= '0'; 

      when layer_3_auto =>
        if rx_intrnl_empty = '0' then
          case rx_intrnl_data is
            when X"00" => -- Service Mode detect
              n_mode_service_detect <= '1';
              n_mode_application_detect <= '0';
              n_firmware_checksum_error_detect <= '0';
            when X"02" => -- Application Mode detect
              n_mode_service_detect <= '0';
              n_mode_application_detect <= '1';
              n_firmware_checksum_error_detect <= '0';
            when X"03" => -- Application Checksum Error
              n_mode_service_detect <= '0';
              n_mode_application_detect <= '0';
              n_firmware_checksum_error_detect <= '1';
            when others =>
              n_mode_service_detect <= '0';
              n_mode_application_detect <= '0';
              n_firmware_checksum_error_detect <= '0';
          end case;
          -- Only one extra byte expected, so return to looking for SOH
          n_rx_decode <= layer_2_soh_detect; 
        else
          n_mode_service_detect <= '0';
          n_mode_application_detect <= '0';
          n_firmware_checksum_error_detect <= '0';
          n_rx_decode <= layer_3_auto;
        end if;
        n_alarm_detect_vw <= '0'; n_alarm_detect_vc <= '0'; n_alarm_detect_tw <= '0'; n_alarm_detect_tc <= '0'; 

      when layer_3_auto_ext =>
        if rx_intrnl_empty = '0' then
          case rx_intrnl_data is
            when X"00" => -- Temperature Warning
              n_alarm_detect_vw <= '0'; n_alarm_detect_vc <= '0'; n_alarm_detect_tw <= '1'; n_alarm_detect_tc <= '0';
            when X"02" => -- Temperature Critical
              n_alarm_detect_vw <= '0'; n_alarm_detect_vc <= '0'; n_alarm_detect_tw <= '0'; n_alarm_detect_tc <= '1';
            when X"03" => -- Voltage WarningMez. 
              n_alarm_detect_vw <= '1'; n_alarm_detect_vc <= '0'; n_alarm_detect_tw <= '0'; n_alarm_detect_tc <= '0';
            when X"05" => -- Voltage Critical
              n_alarm_detect_vw <= '0'; n_alarm_detect_vc <= '1'; n_alarm_detect_tw <= '0'; n_alarm_detect_tc <= '0';
            when others =>
              n_alarm_detect_vw <= '0'; n_alarm_detect_vc <= '0'; n_alarm_detect_tw <= '0'; n_alarm_detect_tc <= '0'; 
          end case;
          if rx_intrnl_eot = '1' then
            n_rx_decode <= layer_2_soh_detect; 
          else
            case rx_intrnl_data is
              when X"06" => -- Sensor Data
                n_rx_decode <= layer_3_auto_ext_sensor;
              when X"07" => -- VPD Data
                n_rx_decode <= layer_3_auto_ext_vpd;
              when X"08" => -- Module Data
                n_rx_decode <= layer_3_auto_ext_mez0;
              when others =>
                n_rx_decode <= layer_3_auto_ext; 
            end case;
          end if;
        else
          n_rx_decode <= layer_3_auto_ext;
          n_alarm_detect_vw <= '0'; n_alarm_detect_vc <= '0'; n_alarm_detect_tw <= '0'; n_alarm_detect_tc <= '0'; 
        end if;
        n_mode_service_detect <= '0'; n_mode_application_detect <= '0'; n_firmware_checksum_error_detect <= '0';
        
      when layer_3_auto_ext_sensor =>
        if rx_intrnl_empty = '0' then
          if rx_intrnl_eot = '1' then
            n_rx_decode <= layer_2_soh_detect; 
          else
            n_rx_decode <= layer_3_auto_ext_sensor; 
          end if;
        else
          n_rx_decode <= layer_3_auto_ext_sensor;
        end if;
        n_mode_service_detect <= '0'; n_mode_application_detect <= '0'; n_firmware_checksum_error_detect <= '0';
        n_alarm_detect_vw <= '0'; n_alarm_detect_vc <= '0'; n_alarm_detect_tw <= '0'; n_alarm_detect_tc <= '0'; 
       
      when layer_3_auto_ext_vpd =>
        if rx_intrnl_empty = '0' then
          if rx_intrnl_eot = '1' then
            n_rx_decode <= layer_2_soh_detect; 
          else
            n_rx_decode <= layer_3_auto_ext_vpd; 
          end if;
        else
          n_rx_decode <= layer_3_auto_ext_vpd;
        end if;
        n_mode_service_detect <= '0'; n_mode_application_detect <= '0'; n_firmware_checksum_error_detect <= '0';
        n_alarm_detect_vw <= '0'; n_alarm_detect_vc <= '0'; n_alarm_detect_tw <= '0'; n_alarm_detect_tc <= '0'; 

      when layer_3_auto_ext_mez0 =>
        if rx_intrnl_empty = '0' then
          if rx_intrnl_eot = '1' then
            n_rx_decode <= layer_2_soh_detect; 
          else
            n_rx_decode <= layer_3_auto_ext_mez0; 
          end if;
        else
          n_rx_decode <= layer_3_auto_ext_mez0;
        end if;
        n_mode_service_detect <= '0'; n_mode_application_detect <= '0'; n_firmware_checksum_error_detect <= '0';
        n_alarm_detect_vw <= '0'; n_alarm_detect_vc <= '0'; n_alarm_detect_tw <= '0'; n_alarm_detect_tc <= '0'; 

      when layer_3_pass_through =>
        if rx_intrnl_empty = '0' then
          if rx_intrnl_eot = '1' then
            n_rx_decode <= layer_2_soh_detect;
          else
            n_rx_decode <= layer_3_pass_through;
          end if;        
        else
          n_rx_decode <= layer_3_pass_through;
        end if;
        n_mode_service_detect <= '0'; n_mode_application_detect <= '0'; n_firmware_checksum_error_detect <= '0';
        n_alarm_detect_vw <= '0'; n_alarm_detect_vc <= '0'; n_alarm_detect_tw <= '0'; n_alarm_detect_tc <= '0'; 
      
      when layer_3_unknown =>
        n_rx_decode <= layer_2_soh_detect;
        n_mode_service_detect <= '0'; n_mode_application_detect <= '0'; n_firmware_checksum_error_detect <= '0';
        n_alarm_detect_vw <= '0'; n_alarm_detect_vc <= '0'; n_alarm_detect_tw <= '0'; n_alarm_detect_tc <= '0'; 

      when others =>
        n_rx_decode <= layer_2_soh_detect;
        n_mode_service_detect <= '0'; n_mode_application_detect <= '0'; n_firmware_checksum_error_detect <= '0';
        n_alarm_detect_vw <= '0'; n_alarm_detect_vc <= '0'; n_alarm_detect_tw <= '0'; n_alarm_detect_tc <= '0'; 
        
    end case;
  end process;
  
  n_sns_updated <= '1' when ( p_rx_decode = layer_3_auto_ext_sensor ) and ( rx_decode = layer_2_soh_detect ) else sns_updated and not sns_updated_ack;
  n_vpd_updated <= '1' when ( p_rx_decode = layer_3_auto_ext_vpd )    and ( rx_decode = layer_2_soh_detect ) else vpd_updated and not vpd_updated_ack;
  n_mez_updated <= '1' when ( p_rx_decode = layer_3_auto_ext_mez0 )   and ( rx_decode = layer_2_soh_detect ) else mez_updated and not mez_updated_ack;
  
  driver_fifo_wr <= rx_intrnl_rd_pipe_0 when ( rx_f = '0' ) and ( ( rx_decode = layer_3_pass_through ) or ( p_rx_decode = layer_3_pass_through ) ) else '0';
  
  rom_wr(0) <= rx_intrnl_rd_pipe_0 when ( p_rx_decode = layer_3_auto_ext_sensor ) else '0';  -- ( rx_f = '0' ) and 
  rom_wr(1) <= rx_intrnl_rd_pipe_0 when ( p_rx_decode = layer_3_auto_ext_vpd ) else '0';     -- ( rx_f = '0' ) and 
  rom_wr(2) <= rx_intrnl_rd_pipe_0 when ( p_rx_decode = layer_3_auto_ext_mez0 ) else '0';    -- ( rx_f = '0' ) and 
  
  n_rom_addr <= ( others => '0' ) when ( rx_decode = layer_2_soh_detect ) else
                rom_addr + 1 when (rom_wr /= "000" ) else rom_addr; -- and ( rx_f = '0' )

  p_data_ram_wr : process( reg_clk )
  begin
    if rising_edge( reg_clk ) then
      vpd_addr <= (others => '0');
      vpd_addr(vpd_column_left downto 0) <= rom_addr( vpd_column_left + 4 downto 4 );
      vpd_data <= rx_intrnl_data;
      for i in 0 to 15 loop
        vpd_sel(i) <= '0';
        if rom_addr( 3 downto 0 ) = i then
          --if rom_wr(0) = '1' then
          --  sns_rom(i)(to_integer( unsigned( rom_addr( vpd_column_left + 4 downto 4 ) ) ) ) <= rx_data_pipe_0( 7 downto 0 );    
          --end if;
          if rom_wr(1) = '1' then
            --vpd_rom(i)(to_integer( unsigned( rom_addr( vpd_column_left + 4 downto 4 ) ) ) ) <= rx_data_pipe_0( 7 downto 0 );
            vpd_sel(i) <= '1';
          end if;
          --if rom_wr(2) = '1' then
          --   mez_rom(i)(to_integer( unsigned( rom_addr( vpd_column_left + 4 downto 4 ) ) ) ) <= rx_data_pipe_0( 7 downto 0 );
          --end if;
        end if;
      end loop;
    end if;
  end process;


  gen_roms: for i in 0 to 15 generate
    sns_rom_proc : process( reg_clk )
    begin
      if rising_edge(reg_clk) then
        if rom_addr( 3 downto 0 ) = i and rom_wr(0) = '1' then
          sns_rom(to_integer( unsigned( rom_addr( vpd_column_left + 4 downto 4 ) ) ) )(8*i+7 downto 8*i) <= rx_data_pipe_0( 7 downto 0 );      
        end if;
        avr_bram_sns_q(8*i+7 downto 8*i) <= sns_rom( to_integer( unsigned( sns_addresss ) ) )(8*i+7 downto 8*i);
      end if; 
    end process;

    vpd_rom_proc : process( reg_clk )
    begin
      if rising_edge(reg_clk) then
        if rom_addr( 3 downto 0 ) = i and rom_wr(1) = '1' then
          vpd_rom(to_integer( unsigned( rom_addr( vpd_column_left + 4 downto 4 ) ) ) )(8*i+7 downto 8*i) <= rx_data_pipe_0( 7 downto 0 );      
        end if;
        avr_bram_vpd_q(8*i+7 downto 8*i) <= vpd_rom( to_integer( unsigned( vpd_addresss ) ) )(8*i+7 downto 8*i);
      end if; 
    end process;

    mez_rom_proc : process( reg_clk )
    begin
      if rising_edge(reg_clk) then
        if rom_addr( 3 downto 0 ) = i and rom_wr(2) = '1' then
          mez_rom(to_integer( unsigned( rom_addr( vpd_column_left + 4 downto 4 ) ) ) )(8*i+7 downto 8*i) <= rx_data_pipe_0( 7 downto 0 );      
        end if;
        avr_bram_mez_q(8*i+7 downto 8*i) <= mez_rom( to_integer( unsigned( mez_addresss ) ) )(8*i+7 downto 8*i);
      end if; 
    end process;
    
  end generate;
  

  p_data_filter : process( reg_clk )
    constant crd_stat_addr : std_logic_vector(11 downto 0) := X"030";
    constant fg1_addr : std_logic_vector(11 downto 0) := X"06C";
  begin
    if rising_edge( reg_clk ) then
      if rom_wr(0) = '1' then
        if rom_addr(vpd_column_left+4 downto 2) = crd_stat_addr(vpd_column_left+4 downto 2) then
          case rom_addr(1 downto 0) is
            when "00" =>
              crd_stat(7 downto 0) <= rx_data_pipe_0( 7 downto 0 );
            when "01" =>
              crd_stat(15 downto 8) <= rx_data_pipe_0( 7 downto 0 );
            when "10" =>
              crd_stat(23 downto 16) <= rx_data_pipe_0( 7 downto 0 );
            when "11" =>
              crd_stat(31 downto 24) <= rx_data_pipe_0( 7 downto 0 );
            when others => null;
          end case;
        end if;
      end if;
      if rom_wr(2) = '1' then
        if rom_addr(vpd_column_left+4 downto 0) = fg1_addr(vpd_column_left+4 downto 0) then
          force_gen1 <= rx_data_pipe_0(0);
        end if;
      end if;
    end if;
  end process;

  vpd_forward_data  <= vpd_data;
  vpd_forward_addr <= vpd_addr(4 downto 0);
  vpd_forward_be <= vpd_sel;
  
  rx_data_pipe_0_ext( 9 downto 0 ) <= rx_data_pipe_0;
  rx_data_pipe_0_ext( 31 downto 10 ) <= ( others => '-' );
  
  rx_d <= rx_d_ext( 9 downto 0 );

  i_rx_driver : entity work.dpram_sfifo_level10
      generic map (
        width => 10)
      port map (
        clk    => reg_clk,
        rst    => rst_or_prst_or_driver_rst,
        wadv   => driver_fifo_wr,
        wdata  => rx_data_pipe_0_ext(9 downto 0),
        nfull  => rx_f,
        radv   => rx_r,
        rempty => rx_e,
        rdata  => rx_d_ext(9 downto 0),
        rlevel => driver_rd_cnt(10 downto 0));
  
  --i_rx_driver : unisim.vcomponents.fifo18e1 generic map (
  --  almost_empty_offset => X"0080",   -- Sets the almost empty threshold
  --  almost_full_offset => X"0080",    -- Sets almost full threshold
  --  data_width => 18,                 -- Sets data width to 4-36
  --  do_reg => 1,                      -- Enable output register (1-0) Must be 1 if EN_SYN = FALSE
  --  en_syn => FALSE ,                  -- Specifies FIFO as dual-clock (FALSE) or Synchronous (TRUE)
  --  fifo_mode => "FIFO18",            -- Sets mode to FIFO18 or FIFO18_36
  --  first_word_fall_through => TRUE,  -- Sets the FIFO FWFT to FALSE, TRUE
  --  init => x"000000000",             -- Initial values on output tx_port
  --  sim_device => "7SERIES",          -- Must be set to "7SERIES" for simulation behavior
  --  srval => x"000000000"             -- Set/Reset value for output tx_port
  --) port map (
  --  -- Read Data: 32-bit (each) output: Read output data
  --  do => rx_d_ext,         -- 32-bit output: Data output
  --  dop => open,                -- 4-bit output: Parity data output
  --  -- Status: 1-bit (each) output: Flags and other FIFO status outputs
  --  almostempty => open,        -- 1-bit output: Almost empty flag
  --  almostfull => rx_f,      -- 1-bit output: Almost full flag
  --  empty => rx_e,         -- 1-bit output: Empty flag
  --  full => open,               -- 1-bit output: Full flag
  --  rdcount => driver_rd_cnt,            -- 12-bit output: Read count
  --  rderr => open,              -- 1-bit output: Read error
  --  wrcount => open,            -- 12-bit output: Write count
  --  wrerr => open,              -- 1-bit output: Write error
  --  -- Read Control Signals: 1-bit (each) input: Read clock, enable and reset input signals
  --  rdclk => reg_clk,               -- 1-bit input: Read clock
  --  rden => rx_r,            -- 1-bit input: Read enable
  --  regce => '1',               -- 1-bit input: Clock enable
  --  rst => rst_or_prst_or_driver_rst,       -- 1-bit input: Asynchronous Reset
  --  rstreg => rst_or_prst_or_driver_rst,    -- 1-bit input: Output register set/reset
  --  -- Write Control Signals: 1-bit (each) input: Write clock and enable input signals
  --  wrclk => reg_clk,               -- 1-bit input: Write clock
  --  wren => driver_fifo_wr,              -- 1-bit input: Write enable
  --  -- Write Data: 32-bit (each) input: Write input data
  --  di => rx_data_pipe_0_ext,          -- 32-bit input: Data input
  --  dip => ( others => '0' )    -- 4-bit input: Parity input
  --);   
  
  rst_or_prst_or_driver_rst <= drv_rx_rst or rst_or_prst;
 
  tx_d_mux <= tx_d_auto when tx_d_auto_valid = '1' else
              tx_d_latch when tx_w_latch = '1' else
              tx_d;
  tx_w_mux <= tx_w_auto when tx_d_auto_valid = '1' else
              tx_w_latch when tx_w_latch = '1' else
              tx_w;

  p_tx_auto : process( reg_clk )
  begin
    if rising_edge( reg_clk ) then
      tx_w_auto <= '0';
      if por_rst = '1' or tx_reset = '1' then
        tx_d_auto_counter <= (others => '1');
        tx_d_auto_valid <= '1';
        tx_d_auto <= "0000000000";
        tx_w_latch <= '0';
        tx_d_latch <= "0000000000";
      else
        if tx_d_auto_valid = '1' then
          tx_d_auto_counter <= tx_d_auto_counter-1;

          if tx_d_auto_counter = 1 then
            tx_d_auto_valid <= '0';
          end if;
          if tx_d_auto_counter = 3 then
            tx_d_auto <= "0100000101";
            tx_w_auto <= '1';
          end if;
          if tx_d_auto_counter = 2 then
            tx_d_auto <= "1010000111";
            tx_w_auto <= '1';
          end if;
          if tx_w = '1' then
            tx_w_latch <='1';
            tx_d_latch <= tx_d;
          end if;
        else
          tx_w_latch <= '0';
        end if;        
      end if;
    end if;
  end process;
  

  
end architecture;
