--------------------------------------------------------------------------------
--
-- Package Name      : adb3_dma
--
--                    Copyright 2009 by
--                    Alpha Data Parallel Systems Ltd.
--                    All Rights Reserved.
--
-- Original Author  : Andrew McCormick
-- Created          : 
--
-- Modified By      : 
-- Date             :
-- Change Notes     :     
--
-- Description      : 
--
--
-- Dependencies     : adb3_ocp
--
-- Disclaimer       : THESE DESIGNS ARE PROVIDED "AS IS" WITH NO WARRANTY
--                    WHATSOEVER AND ALPHA DATA SPECIFICALLY DISCLAIMS ANY
--                    WARRANTIES IMPLIED OF MERCHANTABILITY, FITNESS FOR A
--                    PARTICULAR PURPOSE, OR AGAINST INFRINGEMENT.
--
--------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_misc.all;

library work;
use work.adb3_ocp.all;

package adb3_dma is

  constant adb3_dma_xfer_bits_used   : natural := 39;
  constant adb3_local_addr_bits_used : natural := 39;

  type adb3_dma_descriptor_type is record
    start_addr              : std_logic_vector(63 downto 0);
    xfer_length             : std_logic_vector(adb3_dma_xfer_bits_used-1 downto 0);
    next_descriptor         : std_logic_vector(63 downto 0);
    read_data               : std_logic;
    last_descriptor         : std_logic;
    irq_on_pcie_completion  : std_logic;
    irq_on_mptl_completion  : std_logic;
    irq_now                 : std_logic;
    wait_irq_ack            : std_logic;
    fixed_addr              : std_logic;
    force_flush             : std_logic;
    local_addr              : std_logic_vector(adb3_local_addr_bits_used-1 downto 0);
    local_length_adjustment : std_logic_vector(5 downto 0);
    nonirqmode              : std_logic;
  end record;

  type adb3_dma_worker_command_type is record
    start_addr        : std_logic_vector(63 downto 0);
    xfer_length       : std_logic_vector(adb3_dma_xfer_bits_used-1 downto 0);
    read_data         : std_logic;
    irq_on_completion : std_logic;
    last_descriptor   : std_logic;
    fixed_addr        : std_logic;
  end record;

  constant ADB3_DMA_WORKER_COMMAND_TYPE_DEFAULT : adb3_dma_worker_command_type := 
   (
		start_addr => (others => '0'),
 	   xfer_length => (others => '0'),
      read_data => '0',
      irq_on_completion => '0',
      last_descriptor => '0',
      fixed_addr => '0'
   );

  type adb3_dma_control_register_type is record
    -- Abort and Cleanup Strobes
    abort                         : std_logic;
    cleanup                       : std_logic;
    enable                        : std_logic;
    -- Descriptor Fetch Address and WE/Start Strobe 
    descriptor_host_addr          : std_logic_vector(63 downto 0);
    descriptor_host_we            : std_logic;
    descriptor_host_wait_ack      : std_logic;
    -- Descriptor Fetch from FPGA with WE/Start Strobe
    -- (descriptors for Host Side addresses though)
    descriptor_fpga_addr          : std_logic_vector(63 downto 0);
    descriptor_fpga_we            : std_logic;
    descriptor_fpga_wait_ack      : std_logic;
    -- Maximum Burst Sizes
    host_write_worker_max_burst   : std_logic_vector(9 downto 0);
    host_read_worker_max_burst    : std_logic_vector(9 downto 0);
    fpga_write_worker_max_burst   : std_logic_vector(9 downto 0);
    fpga_read_worker_max_burst    : std_logic_vector(9 downto 0);
    -- BYTE Rotations
    alignment_rotate              : std_logic_vector(3 downto 0);
    alignment_rotate_we           : std_logic;
    ignore_first_word             : std_logic;
    -- IRQ Acknowledge
    irq_enable                    : std_logic_vector(7 downto 0);
    irq_enable_we                 : std_logic;
    irq_ack_register              : std_logic_vector(7 downto 0);
    irq_ack_we                    : std_logic;
  end record;

  constant ADB3_DMA_CONTROL_REGISTER_DEFAULT : adb3_dma_control_register_type :=
    (
      -- Abort and Cleanup Strobes
      abort                         => '0',
      cleanup                       => '0',
      enable                        => '0',
      -- Descriptor Fetch Address and WE/Start Strobe 
      descriptor_host_addr          => (others => '0'),
      descriptor_host_we            => '0',
      descriptor_host_wait_ack      => '0',
      -- Descriptor Fetch from FPGA with WE/Start Strobe
      -- (descriptors for Host Side addresses though)
      descriptor_fpga_addr          => (others => '0'),
      descriptor_fpga_we            => '0',
      descriptor_fpga_wait_ack      => '0',
      -- Maximum Burst Sizes
      host_write_worker_max_burst   => "0000001000",
      host_read_worker_max_burst    => "0000100000",
      fpga_write_worker_max_burst   => "0000001000",
      fpga_read_worker_max_burst    => "0000001000",
      -- BYTE Rotations
      alignment_rotate              => (others => '0'),
      alignment_rotate_we           => '0',
      ignore_first_word             => '0',
      -- IRQ Acknowledge
      irq_enable                    => (others => '0'),
      irq_enable_we                 => '0',
      irq_ack_register              => (others => '0'),
      irq_ack_we                    => '0'
      );

  
  type adb3_dma_status_register_type is record
    fsm_idle     : std_logic_vector(13 downto 0);
    fsm_aborting : std_logic_vector(13 downto 0);
    fsm_fail     : std_logic_vector(3 downto 0);
    illegal_we   : std_logic;
    irq_status   : std_logic_vector(7 downto 0);
    df_waiting   : std_logic_vector(1 downto 0);
    fifo_info    : std_logic_vector(31 downto 0);
    debug        : std_logic_vector(63 downto 0);
  end record;

  type dma_engine_configuration is record
    registering_for_speed : boolean;
    fpga_descriptors      : boolean;
    fpga_to_pci_only      : boolean;
    pci_to_fpga_only      : boolean;
  end record;

  constant DMA_ENGINE_CONFIGURATION_DEFAULT : dma_engine_configuration :=
    (registering_for_speed => true,
     fpga_descriptors      => true,
     fpga_to_pci_only      => false,
     pci_to_fpga_only      => false);

  type dma_engine_configuration_array is array (natural range <>) of dma_engine_configuration;




  function slv_to_adb3_dma_descriptor (dma_descriptor_slv : std_logic_vector(255 downto 0))
    return adb3_dma_descriptor_type;

  function slv_to_adb3_dma_worker_command (slv : std_logic_vector(adb3_dma_xfer_bits_used-1+64+4 downto 0))
    return adb3_dma_worker_command_type;

  function adb3_dma_worker_command_to_slv (cmd : adb3_dma_worker_command_type)
    return std_logic_vector;

  

  component descriptor_fetch
    generic (
      constant registering_for_speed : boolean := true);
    port (
      clk                   : in  std_logic;
      abort                 : in  std_logic;
      cleanup               : in  std_logic;
      fsm_idle              : out std_logic;
      fsm_aborting          : out std_logic;
      fsm_fail              : out std_logic;
      new_desc_addr         : in  std_logic_vector(63 downto 0);
      new_desc_addr_we      : in  std_logic;
      ocp_m2s               : out adb3_ocp_m2sT;
      ocp_s2m               : in  adb3_ocp_s2mT;
      pcie_worker_cmd       : out adb3_dma_worker_command_type;
      pcie_worker_cmd_valid : out std_logic;
      pcie_worker_ready     : in  std_logic;
      mptl_worker_cmd       : out adb3_dma_worker_command_type;
      mptl_worker_cmd_valid : out std_logic;
      mptl_worker_ready     : in  std_logic;
      irq                   : out std_logic;
      irq_ack               : in  std_logic;
      waiting               : out std_logic;
      waiting_ack           : in  std_logic;
      write_finished   : in  std_logic;
      alignment        : out std_logic_vector(3 downto 0);
      alignment_ifw    : out std_logic;
      alignment_set    : out std_logic);
  end component;

  component write_worker
    generic (
      addr_bits             : natural := 64;
      left_bits             : natural := 39;
      buffer_descriptors    : boolean := true;
      registering_for_speed : boolean := true);
    port (
      clk              : in  std_logic;
      abort            : in  std_logic;
      cleanup          : in  std_logic;
      fsm_idle         : out std_logic_vector(1 downto 0);
      fsm_aborting     : out std_logic_vector(1 downto 0);
      ocp_m2s          : out adb3_ocp_m2sT;
      ocp_s2m          : in  adb3_ocp_s2mT;
      worker_cmd       : in  adb3_dma_worker_command_type;
      worker_cmd_valid : in  std_logic;
      worker_ready     : out std_logic;
      worker_completed     : out std_logic;
      max_burst        : in  std_logic_vector(9 downto 0);
      irq              : out std_logic;
      irq_ack          : in  std_logic;
      fifo_data        : in  std_logic_vector(127 downto 0);
      fifo_empty       : in  std_logic;
      fifo_level       : in  std_logic_vector(8 downto 0);
      fifo_radv        : out std_logic);
  end component;

  component alignment_write_fifo
    port (
      clk               : in  std_logic;
      data_in           : in  std_logic_vector(127 downto 0);
      push              : in  std_logic;
      nfull             : out std_logic;
      data_out          : out std_logic_vector(127 downto 0);
      radv              : in  std_logic;
      empty             : out std_logic;
      level             : out std_logic_vector(8 downto 0);
      fail              : out std_logic;
      cleanup           : in  std_logic;
      rotate            : in  std_logic_vector(3 downto 0);
      rotate_we         : in  std_logic;
      ignore_first_word : in  std_logic);
  end component;

  component read_worker
    generic (
      check_tags            : boolean;
      addr_bits             : natural := 64;
      left_bits             : natural := 39;
      buffer_descriptors    : boolean := true;
      registering_for_speed : boolean := true;
      chipscope             : boolean := false);
    port (
      clk              : in  std_logic;
      abort            : in  std_logic;
      cleanup          : in  std_logic;
      fsm_idle         : out std_logic_vector(2 downto 0);
      fsm_aborting     : out std_logic_vector(2 downto 0);
      fsm_error        : out std_logic;
      ocp_m2s          : out adb3_ocp_m2sT;
      ocp_s2m          : in  adb3_ocp_s2mT;
      worker_cmd       : in  adb3_dma_worker_command_type;
      worker_cmd_valid : in  std_logic;
      worker_ready     : out std_logic;
      max_burst        : in  std_logic_vector(9 downto 0);
      need_extra_write : in  std_logic;
      irq              : out std_logic;
      irq_ack          : in  std_logic;
      fifo_data        : out std_logic_vector(127 downto 0);
      fifo_nfull       : in  std_logic;
      fifo_wadv        : out std_logic);
  end component;

  component alignment_read_fifo
    port (
      clk               : in  std_logic;
      data_in           : in  std_logic_vector(127 downto 0);
      push              : in  std_logic;
      nfull             : out std_logic;
      data_out          : out std_logic_vector(127 downto 0);
      radv              : in  std_logic;
      empty             : out std_logic;
      fail              : out std_logic := '0';
      cleanup           : in  std_logic;
      level             : out std_logic_vector(8 downto 0);
      rotate            : in  std_logic_vector(3 downto 0);
      rotate_we         : in  std_logic;
      ignore_first_word : in  std_logic);
  end component;

  component dma_ocp_mux
    port (
      clk            : in  std_logic;
      enable         : in  std_logic;
      external_m2s   : out adb3_ocp_m2sT;
      external_s2m   : in  adb3_ocp_s2mT;
      desc_fetch_m2s : in  adb3_ocp_m2sT;
      desc_fetch_s2m : out adb3_ocp_s2mT;
      reader_m2s     : in  adb3_ocp_m2sT;
      reader_s2m     : out adb3_ocp_s2mT;
      writer_m2s     : in  adb3_ocp_m2sT;
      writer_s2m     : out adb3_ocp_s2mT);
  end component;

  component adb3_dma_engine
    generic (
      registering_for_speed : boolean := true;
      fpga_descriptors      : boolean := true;
      fpga_to_pci_only      : boolean := false;
      pci_to_fpga_only      : boolean := false;
      read_worker_chipscope : boolean := false);  
    port (
      clk          : in  std_logic;
      irq          : out std_logic;
      pcie_ocp_m2s : out adb3_ocp_m2sT;
      pcie_ocp_s2m : in  adb3_ocp_s2mT;
      mptl_ocp_m2s : out adb3_ocp_m2sT;
      mptl_ocp_s2m : in  adb3_ocp_s2mT;
      control      : in  adb3_dma_control_register_type;
      status       : out adb3_dma_status_register_type);
  end component;
  
  
end adb3_dma;

package body adb3_dma is

  function slv_to_adb3_dma_descriptor (dma_descriptor_slv : std_logic_vector(255 downto 0))
    return adb3_dma_descriptor_type
  is
    variable desc : adb3_dma_descriptor_type;
  begin
    desc.start_addr              := dma_descriptor_slv(63 downto 0);
    desc.xfer_length             := dma_descriptor_slv(adb3_dma_xfer_bits_used+64-1 downto 64);
    desc.next_descriptor         := dma_descriptor_slv(191 downto 128);
    desc.local_addr              := dma_descriptor_slv(adb3_local_addr_bits_used+192-1 downto 192);
    desc.read_data               := dma_descriptor_slv(240);
    desc.last_descriptor         := dma_descriptor_slv(241);
    desc.irq_on_pcie_completion  := dma_descriptor_slv(242);
    desc.irq_on_mptl_completion  := dma_descriptor_slv(243);
    desc.irq_now                 := dma_descriptor_slv(244);
    desc.wait_irq_ack            := dma_descriptor_slv(245);
    desc.fixed_addr              := dma_descriptor_slv(246);
    desc.force_flush             := dma_descriptor_slv(247);
    desc.local_length_adjustment := dma_descriptor_slv(253 downto 248);
    desc.nonirqmode              := dma_descriptor_slv(255);
    return desc;
  end function;


  function slv_to_adb3_dma_worker_command (slv : std_logic_vector(adb3_dma_xfer_bits_used-1+64+4 downto 0))
    return adb3_dma_worker_command_type
  is
    variable cmd : adb3_dma_worker_command_type;
  begin
    cmd.start_addr        := slv(63 downto 0);
    cmd.xfer_length       := slv(adb3_dma_xfer_bits_used+68-1 downto 68);
    cmd.read_data         := slv(64);
    cmd.irq_on_completion := slv(65);
    cmd.last_descriptor   := slv(66);
    cmd.fixed_addr        := slv(67);
    return cmd;
  end function;

  function adb3_dma_worker_command_to_slv (cmd : adb3_dma_worker_command_type)
    return std_logic_vector
  is
    variable slv : std_logic_vector(adb3_dma_xfer_bits_used-1+64+4 downto 0);
  begin
    slv(63 downto 0)                            := cmd.start_addr;
    slv(adb3_dma_xfer_bits_used+68-1 downto 68) := cmd.xfer_length;
    slv(64)                                     := cmd.read_data;
    slv(65)                                     := cmd.irq_on_completion;
    slv(66)                                     := cmd.last_descriptor;
    slv(67)                                     := cmd.fixed_addr;
    return slv;
  end function;
  
end adb3_dma;
