--------------------------------------------------------------------------------
--
-- Module Name      : alignment_read_fifo
--
--                    Copyright 2009 by
--                    Alpha Data Parallel Systems Ltd.
--                    All Rights Reserved.
--
-- Original Author  : Andrew McCormick
-- Created          : 
--
-- Modified By      : 
-- Date             :
-- Change Notes     :     
--
-- Description      : 
--
--
-- Dependencies     : fifo
--
-- Disclaimer       : THESE DESIGNS ARE PROVIDED "AS IS" WITH NO WARRANTY
--                    WHATSOEVER AND ALPHA DATA SPECIFICALLY DISCLAIMS ANY
--                    WARRANTIES IMPLIED OF MERCHANTABILITY, FITNESS FOR A
--                    PARTICULAR PURPOSE, OR AGAINST INFRINGEMENT.
--
--------------------------------------------------------------------------------


library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_misc.all;

library work;
use work.fifo.all;


entity alignment_read_fifo256 is
  port (
    -- FIFO Ports
    clk       : in  std_logic;
    data_in   : in  std_logic_vector(255 downto 0);
    push      : in  std_logic;
    nfull     : out std_logic;
    data_out  : out std_logic_vector(255 downto 0);
    radv      : in  std_logic;
    empty     : out std_logic;
    fail      : out std_logic := '0';
    cleanup   : in  std_logic;
    level     : out std_logic_vector(8 downto 0);
    -- Alignment control
    rotate    : in  std_logic_vector(4 downto 0);
    rotate_we : in  std_logic;
    ignore_first_word : in std_logic);

end alignment_read_fifo256;

architecture synth of alignment_read_fifo256 is
  
  constant USE_COREGEN_FIFO : boolean := false;
  constant USE_FAST_FIFO    : boolean := true;

  signal rflush    : std_logic := '0';
  signal re        : std_logic;
  signal rlevel    : std_logic_vector(15 downto 0);
  signal local_rst : std_logic := '1';
  signal rst       : std_logic := '0';
  signal rst1       : std_logic := '0';

  signal wd : std_logic_vector(255 downto 0);

  signal al_reg0, al_reg1,al_reg0_r1             : std_logic_vector(255 downto 0) := (others => '0');
  signal al_reg0_state, al_reg1_state : std_logic                      := '0';
  signal wadv,wadv_e                         : std_logic                      := '0';
  signal push_reg                     : std_logic                      := '0';

  signal rot : std_logic := '0';
  signal rot_r1 : std_logic := '0';

  signal rudf, wovf : std_logic := '0';

  signal rot_regs : std_logic_vector(511 downto 0);

  -- For optional Compilation
  -- Instantiate CoreGen FIFO instead of SDK FIFO for speed
  --component sdk_sfifo128
  --  port (
  --    clk        : in  std_logic;
  --    din        : in  std_logic_vector(127 downto 0);
  --    rd_en      : in  std_logic;
  --    wr_en      : in  std_logic;
  --    data_count : out std_logic_vector(9 downto 0);
  --    dout       : out std_logic_vector(127 downto 0);
  --    empty      : out std_logic;
  --    full       : out std_logic;
  --    prog_full  : out std_logic);
  --end component;

  component dpram_sfifo_level
    generic (
      width : integer := 64);
    port (
      clk    : in  std_logic;
      rst    : in  std_logic;
      wadv   : in  std_logic;
      wdata  : in  std_logic_vector(width-1 downto 0);
      nfull  : out std_logic;
      radv   : in  std_logic;
      rempty : out std_logic;
      rdata  : out std_logic_vector(width-1 downto 0);
      rlevel : out std_logic_vector(9 downto 0));
  end component;

  signal rotate1    : std_logic_vector(4 downto 0);
  signal rotate_we1 : std_logic;

  type rot_regs_bytes_type is array (0 to 7) of std_logic_vector(39 downto 0);
  signal rot_one_hot : std_logic_vector(7 downto 0);
  type rot_regs_prod_type is array (0 to 7) of std_logic_vector(31 downto 0);

  signal rot_regs_bytes : rot_regs_bytes_type;
  signal rot_regs_reduced : std_logic_vector(319 downto 0);
  signal rot_regs_prod0 : rot_regs_prod_type;
  signal rot_regs_prod1 : rot_regs_prod_type;
  signal regs_rotated : std_logic_vector(255 downto 0);
  
begin  -- synth



  -- Generate a Power on Reset for FIFOs
 
  lrst0 : process (clk)
  begin  -- process lrst0
    if clk'event and clk = '1' then     -- rising clock edge
      local_rst <= cleanup;
      rst1       <= local_rst;-- or cleanup;
      rst <= rst1;
    end if;
  end process lrst0;
  
  
  gen_sdk_fifo0 : if not USE_COREGEN_FIFO and not USE_FAST_FIFO generate
    sfifo_1 : sfifo
      generic map (
        width  => 256,
        order  => 9,
        wpfl   => 256,
        rpel   => 2,
        family => family_virtex5,
        safety => false)
      port map (
        rst    => rst,
        k      => clk,
        wd     => wd,
        wadv   => wadv,
        we     => open,
        wpf    => nfull,
        wf     => open,
        wovf   => wovf,
        rq     => data_out,
        radv   => radv,
        rflush => rflush,
        re     => re,
        rpe    => open,
        rf     => open,
        rudf   => rudf,
        rlevel => rlevel);
  end generate gen_sdk_fifo0;

  --gen_coregen_fifo0 : if USE_COREGEN_FIFO generate
  --  cg_fifo0 : sdk_sfifo128
  --    port map (
  --      clk        => clk,
  --      din        => wd,
  --      rd_en      => radv,
  --      wr_en      => wadv,
  --      data_count => rlevel(9 downto 0),
  --      dout       => data_out,
  --      empty      => re,
  --      full       => open,
  --      prog_full  => nfull);
  --end generate gen_coregen_fifo0;

  gen_fast_fifo0 : if USE_FAST_FIFO generate
    dpram_sfifo_level_1 : dpram_sfifo_level
      generic map (
        width => 256)
      port map (
        clk    => clk,
        rst    => rst,
        wadv   => wadv,
        wdata  => wd,
        nfull  => nfull,
        radv   => radv,
        rempty => re,
        rdata  => data_out,
        rlevel => rlevel(9 downto 0));    
  end generate gen_fast_fifo0;


  process (clk)
 
  begin  -- process   
    if clk'event and clk = '1' then
      rotate1 <= rotate;
      rotate_we1 <= rotate_we;
      
      if rotate_we1 = '1' or cleanup = '1' then
        al_reg0_state <= '0';
        al_reg1_state <= '0';
      elsif push = '1' then
        al_reg0_state <= '1';
        al_reg1_state <= al_reg0_state;
        al_reg0       <= data_in;
        al_reg1       <= al_reg0;
      end if;

      push_reg <= push;

      if rot = '0' or ignore_first_word = '0' then
        wadv_e <= al_reg0_state and push_reg;
      else
        wadv_e <= al_reg0_state and al_reg1_state and push_reg;
      end if;

      if rotate_we1 = '1' then
        rot <= OR_reduce(rotate1);
      end if;
      
      for i in 0 to 7 loop
        if i = rotate(2 downto 0) then
          rot_one_hot(i) <= '1';
        else
          rot_one_hot(i) <= '0';
        end if;
      end loop;  -- i

      al_reg0_r1 <= al_reg0;
      rot_r1 <= rot;
      wadv <= wadv_e;
      
      if rot_r1 = '0' then
        wd <= al_reg0_r1;
      else
        wd <= regs_rotated;
      end if;

      if cleanup = '1' then
        fail <= '0';
      --elsif wovf = '1' or rudf = '1' then
      elsif wovf = '1' then
        fail <= '1';
      end if;
      
    end if;
  end process;

  rot_regs <= al_reg0 & al_reg1;

  level <= rlevel(8 downto 0) when rlevel(9) = '0' else (others => '1');
  empty <= re;

  process (clk) is
  begin  -- process
    if rising_edge(clk) then            -- rising clock edge
       case rotate1(4 downto 3) is
        when "00" => rot_regs_reduced <= rot_regs(511 downto 192);
        when "01" => rot_regs_reduced <= rot_regs(447 downto 128);
        when "10" => rot_regs_reduced <= rot_regs(383 downto 64);
        when "11" => rot_regs_reduced <= rot_regs(319 downto 0);          
        when others => null;
      end case;
    end if;
  end process;

  gen_i_loop: for i in 0 to 7 generate
    gen_j_loop: for j in 0 to 39 generate
      rot_regs_bytes(i)(j) <= rot_regs_reduced(8*j+i);
    end generate gen_j_loop;
    rot_regs_prod0(i) <= rot_regs_bytes(i)(23 downto 0) * rot_one_hot;
    rot_regs_prod1(i) <= rot_regs_bytes(i)(39 downto 16) * rot_one_hot;
    gen_j_loop2: for j in 0 to 15 generate
      regs_rotated(8*j+i) <= rot_regs_prod0(i)(j+8);
      regs_rotated(8*j+128+i) <= rot_regs_prod1(i)(j+8);
    end generate gen_j_loop2;
    
  end generate gen_i_loop;

  
end synth;







