--------------------------------------------------------------------------------
--
-- Module Name      : dpram_sfifo
--
--                    Copyright 2009 by
--                    Alpha Data Parallel Systems Ltd.
--                    All Rights Reserved.
--
-- Original Author  : Andrew McCormick
-- Created          : 
--
-- Modified By      : 
-- Date             :
-- Change Notes     :     
--
-- Description      :
-- Dual Port RAM Synchronous FIFO
--
-- Configurable data width
--
-- Use behavioural BRAM description to allow tools to identify
-- and use DPBRAMs appropriately
--
--
--
-- Dependencies     : adb3_ocp
--
-- Disclaimer       : THESE DESIGNS ARE PROVIDED "AS IS" WITH NO WARRANTY
--                    WHATSOEVER AND ALPHA DATA SPECIFICALLY DISCLAIMS ANY
--                    WARRANTIES IMPLIED OF MERCHANTABILITY, FITNESS FOR A
--                    PARTICULAR PURPOSE, OR AGAINST INFRINGEMENT.
--
----------------------------------------------------------------------------------


library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_misc.all;

entity dpram_sfifo_ooo is
  
  generic (
    width : integer := 64;
    order : natural := 9);

  port (
    clk   : in  std_logic;
    rst    : in  std_logic;
    wadv   : in  std_logic_vector(7 downto 0);
    wdata  : in  std_logic_vector(width-1 downto 0);
    nfull  : out std_logic_vector(7 downto 0);
    radv   : in  std_logic_vector(7 downto 0);
    rempty : out std_logic_vector(7 downto 0);
    rdata  : out std_logic_vector(width-1 downto 0));

end dpram_sfifo_ooo;

architecture rtl of dpram_sfifo_ooo is

  constant depth : natural := 2**order;
  constant torder : natural := order-3;
  
  type ram_type is array (depth-1 downto 0) of std_logic_vector(width-1 downto 0);

  signal ram : ram_type := (others => (others => '0'));
  signal waddr : std_logic_vector(order-1 downto 0) := (others => '0');
  signal raddr : std_logic_vector(order-1 downto 0) := (others => '0');
  signal ram_reg,ram_out,wdata_reg  : std_logic_vector(width-1 downto 0) := (others => '0');
  signal rd_ram : std_logic := '0';
  signal wadv_reg : std_logic := '0';
  signal wadv_array_reg : std_logic_vector(7 downto 0) := (others => '0');

  signal waddr_array : std_logic_vector(8*torder-1 downto 0) := (others => '0');
  signal raddr_array : std_logic_vector(8*torder-1 downto 0) := (others => '0');
  --signal waddr_latch,waddr_rclk : std_logic_vector(8 downto 0) := (others => '0');
  
  --signal waddr_rclk_valid : std_logic := '0';
 

  signal level : std_logic_vector(8*torder-1 downto 0) := (others => '0');
 
  
begin  -- rtl

  -- Model Block Rams, with output register
  process (clk)
  begin  -- process
    if clk'event and clk = '1' then  -- rising clock edge
      if wadv_reg = '1' then
        ram(conv_integer(waddr)) <= wdata_reg;
      end if;
    end if;
  end process;
  
  process (clk)
  begin  -- process
    if clk'event and clk = '1' then  -- rising clock edge
      ram_reg <= ram(conv_integer(raddr));
      ram_out <= ram_reg;
    end if;
  end process;


  gen_fsms: for i in 0 to 7 generate

    p1: process (clk)
    begin  -- process write    
      if clk'event and clk = '1' then  -- rising clock edge

        if rst = '1' then                   -- asynchronous reset
          waddr_array(torder*i+torder-1 downto torder*i) <= (others => '0'); 
          raddr_array(torder*i+torder-1 downto torder*i) <= (others => '0');
          level(torder*i+torder-1 downto torder*i) <= (others => '0');
          rempty(i) <= '1';
          
        else
          if wadv(i) = '1' then
            waddr_array(torder*i+torder-1 downto torder*i) <= waddr_array(torder*i+torder-1 downto torder*i)+1;
          end if;
          
          if radv(i) = '1' then
            raddr_array(torder*i+torder-1 downto torder*i) <= raddr_array(torder*i+torder-1 downto torder*i)+1;
          end if;
          if wadv_array_reg(i) = '1' and radv(i) = '0' then
            level(torder*i+torder-1 downto torder*i) <= level(torder*i+torder-1 downto torder*i)+1;
            rempty(i) <= '0';
          elsif wadv_array_reg(i) = '0' and radv(i) = '1' then
            level(torder*i+torder-1 downto torder*i) <= level(torder*i+torder-1 downto torder*i)-1;
            if level(torder*i+torder-1 downto torder*i) = 1 then
              rempty(i) <= '1';
            end if;
          end if;
        end if;
      end if;
  end process p1;

  p2: process (clk) is
      begin  -- process p2
    if rising_edge(clk) then            -- rising clock edge
      nfull(i) <= AND_reduce(level(torder*i+torder-1 downto torder*i+3));
      wadv_array_reg(i) <= wadv(i);
      wdata_reg <= wdata;
    end if;
  end process p2;
    
    
  end generate gen_fsms;

  process (clk) is
    variable waddr_msbs : std_logic_vector(2 downto 0);
    variable raddr_msbs : std_logic_vector(2 downto 0);
    variable waddr_lsbs : std_logic_vector(order-4 downto 0);
    variable raddr_lsbs : std_logic_vector(order-4 downto 0);
  begin  -- process
    if rising_edge(clk) then            -- rising clock edge
      waddr_msbs := "000";
      raddr_msbs := "000";
      for i in 0 to 7 loop
        if wadv(i) = '1' then
          waddr_msbs := CONV_STD_LOGIC_VECTOR(i,3);
          waddr_lsbs := waddr_array(torder*i+torder-1 downto torder*i);
        end if;
      end loop;  -- i

      for i in 0 to 7 loop
        if radv(i) = '1' then
          raddr_msbs := CONV_STD_LOGIC_VECTOR(i,3);
          raddr_lsbs := raddr_array(torder*i+torder-1 downto torder*i);
        end if;
      end loop;  -- i

      wadv_reg <= OR_reduce(wadv);
      waddr <= waddr_msbs & waddr_lsbs;
      raddr <= raddr_msbs & raddr_lsbs;
      
    end if;
  end process;
  
   rdata <= ram_out;
  
end rtl;

