--
-- rrpe_8.vhd
--
-- (c) Alpha Data Parallel Systems, 2005
---------------------------------------------------------------------------
-- Description
--    8-input round-robin priority encoder
--
---------------------------------------------------------------------------
-- Notes:
--    Design adapted from 
--    "Design and Implementation of a Fast Crossbar Scheduler"
--    by Pankaj Gupta & Nick McKeown
---------------------------------------------------------------------------
library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_misc.all;

library work;
use work.pcie_config.ALL;
use work.pcie.ALL;
use work.adb3_ocp.ALL;
use work.pciegen3_pkg.all;

-- synopsys translate_off
library unisim;               
use unisim.vcomponents.all;   
-- synopsys translate_on

entity dma_arbiter is
  generic(
    maxChannels : integer := 5
  );
   port (
      -- no reset
      Ck                : in  std_logic;
   
      -- Inputs
      PtrUpdate         : in  std_logic;
      Req               : in  std_logic_vector(maxChannels-1 downto 0);
      
      -- output
      AnySet            : out std_logic;
      LPVal             : out std_logic_vector(maxChannels-1 downto 0)
   );

end dma_arbiter;


---------------------------------------------------------------------------
architecture rtl of dma_arbiter is

------------------------------------
-- internal signals
------------------------------------
signal LPVal_int  : std_logic_vector(4 downto 0) := (others => '0');
signal AnySet_int : std_logic := '0';
signal ThermoPtr  : std_logic_vector(4 downto 0) := (others => '0');
signal ReqMask    : std_logic_vector(4 downto 0) := (others => '0');
signal ThermoReq  : std_logic_vector(4 downto 0) := (others => '0');
signal MS_AnySet  : std_logic := '0';
signal MS_LPVal   : std_logic_vector(4 downto 0) := (others => '0');
signal LS_AnySet  : std_logic := '0';
signal LS_LPVal   : std_logic_vector(4 downto 0) := (0=> '1',others => '0');


------------------------------------
-- procedure: priority8
-- eight bit priority encoder
------------------------------------
procedure priorityn ( signal Inputs : in  std_logic_vector(4 downto 0);
                      signal AnySet : out std_logic;
                      signal LPVal  : out std_logic_vector(4 downto 0)) is
begin
  if Inputs(0) = '1' then
    AnySet <= '1';
    LPVal  <= "00001";
  elsif Inputs(1) = '1' then
    AnySet <= '1';
    LPVal  <= "00010";
  elsif Inputs(2) = '1' then
    AnySet <= '1';
    LPVal  <= "00100";
  elsif Inputs(3) = '1' then
    AnySet <= '1';
    LPVal  <= "01000";
  elsif Inputs(4) = '1' then
    AnySet <= '1';
    LPVal  <= "10000";
  else
    AnySet <= '0';
    LPVal  <= "00001";
  end if;
end priorityn;
------------------------------------



------------------------------------
begin

--
-- ls_pe uses the "raw" request inputs
-- ms_pe uses the thermo encoded version.  This has the
-- higher priority, so the mux outputs this value if any
-- are set.

  ThermoReq <= ReqMask and (not (ThermoPtr));
  
  
  ReqMask(maxChannels-1 downto 0) <= Req;
  
  ls_pe : priorityn( Inputs => ReqMask,
                     AnySet => LS_AnySet,
                     LPVal  => LS_LPVal);
  
  ms_pe : priorityn( Inputs => ThermoReq,
                     AnySet => MS_AnySet,
                     LPVal  => MS_LPVal);


  mux_proc : process(MS_AnySet,MS_LPVal,LS_LPVal,LS_AnySet)
  -- mux_proc : process(Ck)
  begin
   if MS_AnySet = '1' then
    AnySet_int <= MS_AnySet;
    LPVal_int  <= MS_LPVal;
   else
      AnySet_int <= LS_AnySet;
      LPVal_int  <= LS_LPVal;
   end if;
  end process mux_proc;

--
-- "thermometer" encode the priority pointer
-- The logic equations are pasted from the tcl script 
-- generated "thermo_enc8.txt"
--

  thermo_n : process(Ck)
  begin
    if( rising_edge(ck) )then
      if( PtrUpdate = '1' )then
  			case LPVal_int is
  				when "00001"	=> ThermoPtr <= "00001";
  				when "00010"	=> ThermoPtr <= "00011";
  				when "00100"	=> ThermoPtr <= "00111";
  				when "01000"	=> ThermoPtr <= "01111";
  				when "10000"	=> ThermoPtr <= "11111";
  				when others   => ThermoPtr <= "00000";
  			end case;
      end if;
    end if;
  end process thermo_n;
  
--
-- concurrent assignment of the outputs
--
  LPVal  <= LPVal_int(maxChannels-1 downto 0);
  AnySet <= AnySet_int; 

end rtl;
