--------------------------------------------------------------------------------
-- Company:         Alpha-Data Parallel Systems
-- Engineer:        Alan Crawford
--
-- Create Date:
-- Design Name:
-- Module Name:     pcieg3_bar_rx_if
-- Project Name:    ADB3G3
-- Target Device:   V7
-- Tool versions:
-- Description:     Interfaces to the Gen3 Core BAR I/F and provides the
--                  required number of OCP ports
--
-- Dependencies:    PCIe/OCP Packages for signal records
--
-- Revision:
-- Revision 0.01 - File Created
-- Additional Comments:
--
--------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.all;
use IEEE.STD_LOGIC_ARITH.all;
use IEEE.STD_LOGIC_UNSIGNED.all;

library work;
use work.pcie.all;
use work.adb3_ocp.all;
use work.pciegen3_pkg.all;
use work.pcie_config.all;

entity pcieg3_bar2_rx_if is
  generic(
    datawidthTX         : integer := 256
  );
  port(
    pcie_clk            : in  std_logic;
    pcie_rst            : in  std_logic;
    m_m2s_o             : out adb3w_ocp_m2sT;
    m_s2m_i             : in  adb3w_ocp_s2mT;

    rx_data_out         : out std_logic_vector(datawidthTX+datawidthTX/32+1 downto 0);
    rx_data_pop         : in  std_logic;
    rx_data_pkt_rdy     : out std_logic;
    bin_responses       : in  std_logic;
    pause_rx            : out std_logic;
    rx_tag_value        : out std_logic_vector(7 downto 0);
    rx_tag_clear        : out std_logic;
    rx_tag_info         : in  std_logic_vector(71 downto 0);    
    rx_tag_valid        : in  std_logic   
  );
end pcieg3_bar2_rx_if;

architecture synth of pcieg3_bar2_rx_if is
  signal rx_tag_clear_int         : std_logic;
  signal rx_tag_value_int         : std_logic_vector(7 downto 0);    
                                  
  signal valid_rx_data            : std_logic;
  signal rx_data_tag              : std_logic_vector(263 downto 0);
  signal rx_data_tag_fifo         : std_logic_vector(263 downto 0);
  signal rx_push                  : std_logic;
  signal rx_full                  : std_logic;
  signal rx_pop                   : std_logic := '0';
  signal rx_pop_int               : std_logic := '0';
  signal rx_tag_fifo              : std_logic_vector(7 downto 0);
  signal rx_empty                 : std_logic;
                                  
  signal rx_sequencer             : std_logic_vector(4 downto 0);  -- 7 downto 0
  signal rx_descriptor            : std_logic_vector(127 downto 0);
  signal push_descriptor          : std_logic;
  signal cpld_data_in             : std_logic_vector(datawidthTX+datawidthTX/32+1 downto 0);
  signal push_cpld                : std_logic;
  signal rx_cpld_empty            : std_logic;
  signal rx_cpld_full             : std_logic;
  signal rx_count                 : std_logic_vector(7 downto 0);
  signal data_available           : std_logic_vector(7 downto 0);
                                  
  signal end_address_dw           : std_logic_vector(2 downto 0);
                                  
  signal store_data               : std_logic_vector(255 downto 0);
  signal start_half               : std_logic;
  signal push_cpld256             : std_logic;
  signal cpld_data_in256          : std_logic_vector(265 downto 0);
  signal flush_in_progress        : std_logic := '0';
  signal bin_response_del         : std_logic := '0';
  signal flush_tag_clear          : std_logic := '0';
  signal flush_tag_info_reg       : std_logic_vector(73 downto 0) := (others => '0');
  signal flush_tag_value_int      : std_logic_vector(2 downto 0) := "000";
  signal flush_seq                : std_logic_vector(4 downto 0) := "00000";
  signal flush_length             : std_logic_vector(7 downto 0) := (others => '0');
  signal flush_descriptor         : std_logic := '0';
  signal flush_push               : std_logic := '0';
  signal flush_pkt_sent           : std_logic := '0';
  signal flush_descriptor_header  : std_logic_vector(127 downto 0);
  signal flush_pkt_seq            : std_logic_vector(3 downto 0) := "0000";

  signal i_rx_data_out            : std_logic_vector(datawidthTX+datawidthTX/32+1 downto 0);

  signal local_rst,local_sync_rst : std_logic := '1';
  attribute keep : string;
  attribute max_fanout : string;
  attribute equivalent_register_removal : string;
  attribute keep of local_rst : signal is "TRUE";
  attribute equivalent_register_removal of local_rst : signal is "no";
  attribute max_fanout of local_rst : signal is "1";
  attribute keep of local_sync_rst : signal is "TRUE";
  attribute equivalent_register_removal of local_sync_rst : signal is "no";
  signal local_rst_r1: std_logic := '1';
  attribute shreg_extract : string;
  attribute keep of local_rst_r1 : signal is "TRUE";  
  attribute equivalent_register_removal of local_rst_r1 : signal is "no";
  attribute max_fanout of local_sync_rst : signal is "16";
  attribute shreg_extract of local_rst_r1 : signal is "no";
  attribute shreg_extract of local_sync_rst : signal is "no";
begin

  process (pcie_clk,pcie_rst) is
  begin  -- process
    if pcie_rst = '1' then
      local_rst <= '1';
    elsif rising_edge(pcie_clk) then            -- rising clock edge
      local_rst <= '0';
    end if;
  end process;
  
  process (pcie_clk) is
  begin  -- process
    if rising_edge(pcie_clk) then       -- rising clock edge
      local_rst_r1 <= local_rst;
      local_sync_rst <= local_rst_r1;
    end if;
  end process;
  
  rx_data_out <= i_rx_data_out;
  rx_data_tag <= m_s2m_i.tag & m_s2m_i.data;

   bar_if_cmd_fifo : packet_sfifo
    generic map(
          RAM_TYPE    => "AUTO",
          ORDER       => 7,
          WIDTH       => 256+8,
          AFULL_LEVEL => 128
    )
    port map(
      rst         => pcie_rst,
      sr          => '0',
      clk         => pcie_clk,
      wen         => rx_push,
      wd          => rx_data_tag,
      wfull       => rx_full,
      wafull      => open,
      radv        => rx_pop,
      rq          => rx_data_tag_fifo, 
      rlevel      => data_available,
      rempty      => rx_empty,
      debug_info  => open);

 
--   bar_if_cmd_fifo : pcie_pkt_fifo4
--     generic map(
--       fifo_depth      => 4,
--       fifo_width      => 256+8
--       )
--     port map(
--       clk             => pcie_clk,
--       sr              => pcie_rst,
--       max_pkt_size    => "1000",
--       din             => rx_data_tag,
--       dout            => rx_data_tag_fifo,
--       push            => rx_push,
--       pop             => rx_pop ,
--       full           => rx_full,
--       empty          => rx_empty,
--       last           => open,
--       free_pkt       => open,
--       data_available => data_available
--       );
-- 

  m_m2s_o.respaccept  <= (not rx_full) and (not bin_responses);
  m_m2s_o.Cmd         <= (others => '0');
  m_m2s_o.Addr        <= (others => '0');
  m_m2s_o.BurstLength <= (others => '0');
  m_m2s_o.Data        <= (others => '0');
  m_m2s_o.DataByteEn  <= (others => '0');
  m_m2s_o.Tag         <= (others => '0');
  m_m2s_o.DataValid   <= '0'; 


  gen_push : process(m_s2m_i, rx_full,bin_responses)
  begin
    if(((rx_full = '0')and(m_s2m_i.resp = "01"))and(bin_responses = '0'))then
      rx_push <= '1';
    else
      rx_push <= '0';
    end if;
  end process gen_push;

  rx_descriptor <= x"0000_0000" &                     -- data if in DW aligned mode              32  127:96
                   '0' &                              -- Force ECRC                               1      95
                   rx_tag_info(60 downto 58) &        -- Packet attributes                        3   94:92
                   rx_tag_info(57 downto 55) &        -- TC                                       3   91:89
                   '0' &                              -- use Completer ID in core                 1      88
                   x"0000" &                          -- unused when previous bit is '0'         16   87:72
                   rx_tag_info(68 downto 61) &        -- PCIE TAG                                 8   71:64
                   rx_tag_info(54 downto 39) &        -- Requester ID                            16   63:48
                   '0' &                              -- reserved                                 1      47 
                   '0' &                              -- Poison TLP Not used                      1      46
                   "000" &                            -- completer status (Successful)            3   45:43
                   '0'& rx_tag_info(38 downto 29) &   -- Length in DW of completion              11   42:32
                   "00" &                             -- reserved                                 2   31:30
                   '0' &                              -- locked read (NOT used)                   1      29
                   rx_tag_info(20 downto 8) &         -- Byte Count                              13   28:16
                   "000000" &                         -- reserved                                 6   15:10
                   "00" & --rx_tag_info(56 downto 55) &        -- address type                             2    9: 8 
                   '0' &                              -- reserved                                 1       7
                   rx_tag_info(28 downto 22);         -- low address                              7    6: 0
  
  end_address_dw <= rx_tag_info(71 downto 69);

  seq_rx : process(pcie_clk)
  begin
    
    if pcie_clk'event and pcie_clk = '1' then
      if(local_sync_rst = '1')then
        rx_sequencer  <= (others => '0');
        rx_count      <= (others => '0');
      else
        rx_sequencer  <= rx_sequencer(3 downto 0)& '0';      -- 6 downto 0
        if((rx_empty = '0')and(rx_sequencer = "00000")and(bin_responses = '0'))then  -- 0000000
          rx_sequencer(0) <= '1';
        end if;
        rx_pop_int      <= '0';
        rx_tag_clear_int    <= rx_sequencer(0);                  -- 1
        push_descriptor <= rx_sequencer(0);                  -- 1
        if(rx_sequencer(0) = '1')then                        --1
          rx_count <= rx_tag_info(7 downto 0);
        end if;

        if(rx_sequencer(1) = '1')then          -- 2
          if((rx_count > data_available)or(rx_cpld_full = '1'))then
            rx_sequencer(2 downto 1) <= "01";  --2 downto 1
          end if;
        end if;
        if(rx_sequencer(2) = '1')then          -- 3
          if((rx_count > "00000001"))then
            rx_sequencer(3 downto 2) <= "01";  -- 4 downto 3
            rx_count                 <= rx_count - 1;
            rx_pop_int               <= '1';
          elsif((rx_count = "00000001"))then
            rx_sequencer(3 downto 2) <= "10";  -- 4 downto 3
            rx_count                 <= rx_count - 1;
            rx_pop_int               <= '1';
          end if;
        end if;
        if(rx_pop = '1')then
        end if;
      end if;
    end if;
  end process seq_rx;

  rx_pop <= rx_pop_int;

  process (rx_data_tag_fifo,rx_empty)
  begin
    if(rx_empty = '1')then
      rx_tag_value_int <= (others => '0');
    else
      rx_tag_value_int <= rx_data_tag_fifo(263 downto 256);
    end if;
  end process;

  push_cpld <= ((rx_pop  or push_descriptor) and (not bin_responses)) or flush_push;


    sel_op_data256 : process(push_descriptor, rx_descriptor, rx_data_tag_fifo, store_data, start_half,bin_responses,flush_descriptor,flush_descriptor_header)
    begin
      if(bin_responses = '0')then
        if(push_descriptor = '1')then
          cpld_data_in(255 downto 0) <= x"0000_0000_0000_0000_0000_0000_0000_0000" & rx_descriptor;
        else
          cpld_data_in(255 downto 0) <=  rx_data_tag_fifo(255 downto 0);
        end if;
      else
        if(flush_descriptor = '1')then
          cpld_data_in(255 downto 0) <= x"0000_0000_0000_0000_0000_0000_0000_0000" & flush_descriptor_header;
        else
          cpld_data_in(255 downto 0) <=  (others => '1');
        end if;
      end if;
    end process sel_op_data256;

    sel_op_flags256 : process(push_descriptor,bin_responses,flush_descriptor,flush_length, rx_descriptor, rx_data_tag_fifo, rx_pop, rx_count, end_address_dw)
    begin
      if(bin_responses = '0')then
        if(rx_count > "000000000")then
          if(push_descriptor = '1')then
            cpld_data_in(265 downto 256) <= "10" & X"0F";
          else
            cpld_data_in(265 downto 256) <= "00" & X"FF";
          end if;
        else
          case (end_address_dw(2 downto 0)) is
            when "000"  => cpld_data_in(265 downto 256) <= "01" & X"01";
            when "001"  => cpld_data_in(265 downto 256) <= "01" & X"03";
            when "010"  => cpld_data_in(265 downto 256) <= "01" & X"07";
            when "011"  => cpld_data_in(265 downto 256) <= "01" & X"0F";
            when "100"  => cpld_data_in(265 downto 256) <= "01" & X"1F";
            when "101"  => cpld_data_in(265 downto 256) <= "01" & X"3F";
            when "110"  => cpld_data_in(265 downto 256) <= "01" & X"7F";
            when others => cpld_data_in(265 downto 256) <= "01" & X"FF";
          end case;
        end if;
      else
        if(flush_length > "000000000")then
          if(flush_descriptor = '1')then
            cpld_data_in(265 downto 256) <= "10" & X"0F";
          else
            cpld_data_in(265 downto 256) <= "00" & X"FF";
          end if;
        else
          case (end_address_dw(2 downto 0)) is
            when "000"  => cpld_data_in(265 downto 256) <= "01" & X"01";
            when "001"  => cpld_data_in(265 downto 256) <= "01" & X"03";
            when "010"  => cpld_data_in(265 downto 256) <= "01" & X"07";
            when "011"  => cpld_data_in(265 downto 256) <= "01" & X"0F";
            when "100"  => cpld_data_in(265 downto 256) <= "01" & X"1F";
            when "101"  => cpld_data_in(265 downto 256) <= "01" & X"3F";
            when "110"  => cpld_data_in(265 downto 256) <= "01" & X"7F";
            when others => cpld_data_in(265 downto 256) <= "01" & X"FF";
          end case;
        end if;
      end if;
    end process sel_op_flags256;

     bar_if_cpld_fifo : packet_sfifo
    generic map(
          RAM_TYPE    => "AUTO",
          ORDER       => 4,
          WIDTH       => datawidthTX+datawidthTX/32+2,
          AFULL_LEVEL => 14
    )
    port map(
      rst         => pcie_rst,
      sr          => '0',
      clk         => pcie_clk,
      wen         => push_cpld,
      wd          => cpld_data_in,
      wfull       => open,
      wafull      => rx_cpld_full,
      radv        => rx_data_pop,
      rq          => i_rx_data_out, 
      rlevel      => open,
      rempty      => rx_cpld_empty,
      debug_info  => open);
   
    test : pcie_pkt_counter 
      port map(
        clk            => pcie_clk, 
        sr             => pcie_rst,
        eop_in         => cpld_data_in(datawidthTX+datawidthTX/32),
        eop_out        => i_rx_data_out(datawidthTX+datawidthTX/32),
        push           => push_cpld,
        pop            => rx_data_pop,
        pkt_available  => rx_data_pkt_rdy
      ); 
   flush_responses : process(pcie_clk)
   begin
    if(rising_edge(pcie_clk))then
      if(local_sync_rst = '1')then  
        pause_rx <= '0';
      else
        flush_tag_clear <= '0';
        bin_response_del <= bin_responses;
        if((bin_responses = '1')and(bin_response_del = '0')and(flush_in_progress = '0'))then
          pause_rx <= '1';
          flush_tag_value_int   <= "000";
          flush_seq <= "00001";
          flush_in_progress <= '1';
        end if;
        if(flush_seq /= "00000")then
          if(flush_seq = "00001")then
            flush_seq <= "00010";
          end if;
  
          if(flush_seq = "00100")then
            flush_descriptor_header <= rx_descriptor;
            flush_seq <= "01000";
          end if;
          if(flush_seq = "00010")then
            if(rx_tag_valid = '0')then
              flush_seq <= "10000";
            else
              flush_seq <= "00100";
            end if;
          end if;
          if(flush_seq = "01000")then
            if(flush_pkt_sent = '1')then
              flush_seq <= "10000";
              flush_tag_clear <= '1';
            end if;
          end if;
          if(flush_seq = "10000")then
            if(flush_tag_value_int = "111")then
              pause_rx <= '0';
              if(flush_in_progress = '0')then
                flush_seq <= "00000";
              else
                flush_seq <= "00001";
                flush_tag_value_int   <= "000";
              end if;
            else 
              flush_tag_value_int <= flush_tag_value_int + '1';
              flush_seq <= "00001";
            end if;
          end if;
        end if;
        if((bin_responses = '0')and(bin_response_del = '1'))then
          flush_in_progress <= '0';
          flush_seq <= "00000";
        end if;
      end if;
    end if;
  end process;

  gen_flush_pkt : process(pcie_clk)
  begin
    if(rising_edge(pcie_clk))then
      flush_push        <= '0';
      flush_pkt_sent    <= '0';       
      if(flush_seq(3) = '1')then
        if(flush_pkt_seq = "0000")then
          flush_descriptor  <= '1';
          flush_pkt_seq <= "0001";
          flush_length      <= rx_tag_info(7 downto 0); 
        end if;
        if(flush_pkt_seq = "0001")then
          flush_push        <= '1';
          flush_pkt_seq <= "0010";
        end if;
        if(flush_pkt_seq = "0010")then
          flush_pkt_seq <= "0100";
          flush_descriptor  <= '0';
        end if;
        if(flush_pkt_seq = "0100")then
          if(rx_cpld_full = '0')then
            if(flush_length = "000000000")then
              flush_pkt_seq <= "1000";
              flush_pkt_sent    <= '1';       
            else
              flush_push        <= '1';
              flush_length      <= flush_length - '1';
            end if;
          end if;
        end if;
        if(flush_pkt_seq = "1000")then
          flush_pkt_seq <= "0000";
        end if;
        
      end if;
    end if;
  end process;
  
  
  process(bin_responses,flush_tag_value_int,flush_tag_clear,rx_tag_value_int, rx_tag_clear_int)
  begin
    if(bin_responses = '1')then
      rx_tag_value    <= "00000"&flush_tag_value_int;
      rx_tag_clear    <= flush_tag_clear;
    else      
      rx_tag_value    <= rx_tag_value_int;
      rx_tag_clear    <= rx_tag_clear_int;
    end if;
  end process;
        
end architecture;
