library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.numeric_std.all;

library work;
use work.uart_pkg.all;

entity ser_rx_sync is
generic
(
  data_size     : natural   := 8;
  stop_bits     : natural   := 1;
  msb_first     : boolean   := false
);
port
(
  rst         : in    std_logic;
  clk         : in    std_logic;
  ce          : in    std_logic;
  rx_d        : out   std_logic_vector( data_size - 1 downto 0 );
  rx_vld      : out   std_logic;
  rx_ser      : in    std_logic
);
end entity;

architecture rtl of ser_rx_sync is

  type state_type is
  ( 
    rx_idle_state, 
    rx_start_state,
    rx_bit_state,
    rx_stop_state,
    rx_wait_state
  );

  signal bit_cnt     : std_logic_vector( data_size - 1 downto 0 ) := ( others => '-' );
  signal n_bit_cnt   : std_logic_vector( data_size - 1 downto 0 );
  
  signal rx_state    : state_type := rx_idle_state;
  signal n_rx_state  : state_type;
  
  signal o_rx_d       : std_logic_vector( data_size - 1 downto 0 ) := ( others => '-' );
  signal o_rx_vld     : std_logic := '0';
  
  signal n_rx_d       : std_logic_vector( data_size - 1 downto 0 );
  signal n_rx_vld     : std_logic;
  
begin

  rx_d   <= o_rx_d;  
  rx_vld <= o_rx_vld;

  p_clked : process( clk )
  begin
    if rising_edge( clk ) then
      if rst = '1' then
        bit_cnt <= ( others => '-' );
        rx_state <= rx_idle_state;
        o_rx_d   <= ( others => '-' );
        o_rx_vld <= '0';
      else
        if ce = '1' then
          bit_cnt <= n_bit_cnt;
          rx_state <= n_rx_state;
          o_rx_d   <= n_rx_d;  
          o_rx_vld <= n_rx_vld;
        end if;
      end if;
    end if;
  end process;
  
  p_async : process( rx_state, bit_cnt, o_rx_d, rx_ser )
  begin
    case rx_state is
      when rx_idle_state =>
        if rx_ser = '0' then
          n_rx_state <= rx_bit_state;
        else
          n_rx_state <= rx_idle_state;
        end if;
        n_bit_cnt <= std_logic_vector( to_unsigned( 1, n_bit_cnt'length ) );
        n_rx_d <= ( others => '-' );
        n_rx_vld <= '0';
      when rx_bit_state =>
        if msb_first then
          n_rx_d <= o_rx_d( o_rx_d'left - 1 downto 0 ) & rx_ser;
        else
          n_rx_d <= rx_ser & o_rx_d( o_rx_d'left downto 1 );
        end if;
        if bit_cnt( bit_cnt'left ) = '1' then 
          n_rx_state <= rx_stop_state;
          n_bit_cnt <= std_logic_vector( to_unsigned( 1, n_bit_cnt'length ) );
          n_rx_vld <= '1';
        else
          n_rx_state <= rx_bit_state;
          n_bit_cnt <= bit_cnt( bit_cnt'left - 1 downto 0 ) & "0";
          n_rx_vld <= '0';
        end if;
      when rx_stop_state =>
        n_bit_cnt <= bit_cnt( bit_cnt'left - 1 downto 0 ) & "0";
        n_rx_vld <= '0';
        n_rx_d <= ( others => '-' );
        if bit_cnt( stop_bits-1 ) = '1' then 
          n_rx_state <= rx_idle_state;
        else
          n_rx_state <= rx_stop_state;
        end if;
      when others =>
        n_bit_cnt <= ( others => '-' );
        n_rx_state <= rx_idle_state;
        n_rx_d   <= ( others => '-' );
        n_rx_vld <= '0';
    end case;
  end process;
  
end architecture;
