library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.numeric_std.all;

library work;
use work.uart_pkg.all;

entity ser_tx_sync is
generic
(
  data_size     : natural   := 8;
  stop_bits     : natural   := 1;
  inter_gap     : natural   := 3;
  msb_first     : boolean   := false
);
port
(
  rst         : in    std_logic;
  clk         : in    std_logic;
  ce          : in    std_logic;
  tx_d        : in    std_logic_vector( data_size - 1 downto 0 );
  tx_dv       : in    std_logic;
  tx_rdy      : out   std_logic;
  tx_ser      : out   std_logic
);
end entity;

architecture rtl of ser_tx_sync is

  type state_type is
  ( 
    tx_idle_state, 
    tx_start_state,
    tx_bit_state,
    tx_stop_state,
    tx_wait_state
  );
  
  signal reg_tx_d    : std_logic_vector( data_size - 1 downto 0 );
  signal n_reg_tx_d  : std_logic_vector( data_size - 1 downto 0 );
  signal bit_cnt     : std_logic_vector( data_size - 1 downto 0 );
  
  signal tx_state    : state_type := tx_idle_state;
  signal n_tx_state  : state_type;
  
  signal o_tx_rdy   : std_logic;
  signal o_tx_ser   : std_logic := '1';
  signal n_o_tx_rdy : std_logic;
  signal n_o_tx_ser : std_logic;
  signal n_bit_cnt  : std_logic_vector( data_size - 1 downto 0 );
  
begin

  tx_ser <= o_tx_ser;
  tx_rdy <= o_tx_rdy;

  p_clked : process( clk )
  begin
    if rising_edge( clk ) then
      if rst = '1' then
        o_tx_ser <= '1';
        o_tx_rdy <= '0';
        bit_cnt <= ( others => '-' );
        reg_tx_d <= ( others => '-' );
        tx_state <= tx_idle_state;
      else
        if ce = '1' then
          o_tx_ser <= n_o_tx_ser;
          o_tx_rdy <= n_o_tx_rdy;
          bit_cnt <= n_bit_cnt;
          reg_tx_d <= n_reg_tx_d;
          tx_state <= n_tx_state;
        end if;
      end if;
    end if;
  end process;
  
  p_async : process( tx_state, tx_dv, tx_d, bit_cnt, reg_tx_d )
  begin
    case tx_state is
      when tx_idle_state =>
        if tx_dv = '1' then
          -- Send the start bit and drop ready
          n_o_tx_ser <= '0';
          n_o_tx_rdy <= '0';
          n_tx_state <= tx_bit_state;
        else
          n_o_tx_ser <= '1';
          n_o_tx_rdy <= '1';
          n_tx_state <= tx_idle_state;
        end if;
        n_reg_tx_d <= tx_d;
        n_bit_cnt <= std_logic_vector( to_unsigned( 1, n_bit_cnt'length ) );
      when tx_bit_state =>
        n_o_tx_rdy <= '0';
        if msb_first then
          n_reg_tx_d <= reg_tx_d( tx_d'left - 1 downto 0 ) & "0";
          n_o_tx_ser <= reg_tx_d(tx_d'left);
        else
          n_reg_tx_d <= "0" & reg_tx_d( tx_d'left downto 1 );
          n_o_tx_ser <= reg_tx_d(0);
        end if;
        if bit_cnt( bit_cnt'left ) = '1' then 
          n_tx_state <= tx_stop_state;
          n_bit_cnt <= std_logic_vector( to_unsigned( 1, n_bit_cnt'length ) );
        else
          n_tx_state <= tx_bit_state;
          n_bit_cnt <= bit_cnt( bit_cnt'left - 1 downto 0 ) & "0";
        end if;
      when tx_stop_state =>
        n_o_tx_rdy <= '0';
        n_o_tx_ser <= '0';
        n_reg_tx_d <= ( others => '-' );
        if bit_cnt( stop_bits - 1 ) = '1' then 
          n_tx_state <= tx_wait_state;
          n_bit_cnt <= std_logic_vector( to_unsigned( 1, n_bit_cnt'length ) );
        else
          n_tx_state <= tx_stop_state;
          n_bit_cnt <= bit_cnt( bit_cnt'left - 1 downto 0 ) & "0";
        end if;
      when tx_wait_state =>
        n_o_tx_rdy <= '0';
        n_o_tx_ser <= '1';
        n_bit_cnt <= bit_cnt( bit_cnt'left - 1 downto 0 ) & "0";
        n_reg_tx_d <= ( others => '-' );
        if bit_cnt( inter_gap - 1 ) = '1' then 
          n_tx_state <= tx_idle_state;
        else
          n_tx_state <= tx_wait_state;
        end if;
      when others =>
        n_o_tx_rdy <= '0';
        n_o_tx_ser <= '1';
        n_bit_cnt <= ( others => '-' );
        n_reg_tx_d <= ( others => '-' );
        n_tx_state <= tx_idle_state;
    end case;
  end process;
  
end architecture;
