library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;

library work;
use work.uart_pkg.all;

library unisim;
use unisim.vcomponents.all;

entity uart_sync is
  generic
  (
    data_size     : natural   := 8;
    msb_first     : boolean   := false
  );
  port
  (
    -- External Interface
    ser_rx                    : in  std_logic;
    ser_tx                    : out std_logic;
    ser_ck                    : in  std_logic;
    -- Internal Interface
    clk                       : in  std_logic;
    tx_rst                    : in  std_logic;
    tx_data                   : in  std_logic_vector( data_size - 1 downto 0 );
    tx_wr                     : in  std_logic;
    tx_full                   : out std_logic;
    tx_empty                  : out std_logic;
    rx_rst                    : in  std_logic;
    rx_data                   : out std_logic_vector( data_size - 1 downto 0 );
    rx_rd                     : in  std_logic;
    rx_full                   : out std_logic;
    rx_empty                  : out std_logic
  );
end entity;

architecture rtl of uart_sync is

  signal o_tx_full     : std_logic;
  signal o_tx_empty    : std_logic;
  signal o_rx_data     : std_logic_vector( 32 - 1 downto 0 );
  signal i_rx_rd       : std_logic;
  signal o_rx_full     : std_logic;
  signal o_rx_empty    : std_logic;
  
  signal ser_tx_data   : std_logic_vector( 32 - 1 downto 0 );
  signal ser_tx_rd     : std_logic;
  signal ser_tx_rdy    : std_logic;
  
  signal ser_rx_data   : std_logic_vector( 32 - 1 downto 0 );
  signal ser_rx_wr     : std_logic;
  signal ser_rx_rdy    : std_logic;
  
  signal tx_data_ext   : std_logic_vector( 32 - 1 downto 0 ) := ( others => '-' );
  
  signal ser_ck_r1 : std_logic := '0';
  signal ser_ck_r2 : std_logic := '0';
  signal ser_ck_ten : std_logic := '0';
  signal ser_ck_ren : std_logic := '0';

  signal ser_rx_r1 : std_logic := '0';
  signal ser_rx_r2 : std_logic := '0';
  
begin

  rising_edge_ser_ck0: process (clk) is
  begin  -- process rising_edge_ser_ck0
    if rising_edge(clk) then
      ser_ck_r1 <= ser_ck;
      ser_ck_r2 <= ser_ck_r1;
      ser_ck_ten <= ser_ck_r1 and not ser_ck_r2;
      ser_ck_ren <= (not ser_ck_r1) and ser_ck_r2;
      
      ser_rx_r1 <= ser_rx;
      ser_rx_r2 <= ser_rx_r1;
    end if;
  end process rising_edge_ser_ck0;
 
  tx_full     <= o_tx_full;-- or tx_por; 
  tx_empty    <= o_tx_empty;
  
  
  tx_data_ext(tx_data'left downto 0) <= tx_data;
  tx_data_ext(tx_data_ext'left downto tx_data'length) <= ( others => '-' );
  
  rx_data     <= o_rx_data( data_size - 1 downto 0 ); 
  i_rx_rd     <= rx_rd and not rx_rst;   
  rx_full     <= o_rx_full; 
  rx_empty    <= o_rx_empty or rx_rst;

  i_txfifo : entity work.small_sfifo32
      generic map (
        width => 9)
      port map (
        clk    => clk,
        rst    => tx_rst,
        wadv   => tx_wr,
        wdata  => tx_data_ext(8 downto 0),
        nfull  => o_tx_full,
        radv   => ser_tx_rd,
        rempty => o_tx_empty,
        rdata  => ser_tx_data(8 downto 0));
  
  --i_txfifo : unisim.vcomponents.fifo18e1 generic map (
  --  almost_empty_offset => X"0080",   -- Sets the almost empty threshold
  --  almost_full_offset => X"0080",    -- Sets almost full threshold
  --  data_width => 9,                  -- Sets data width to 4-36
  --  do_reg => 1,                      -- Enable output register (1-0) Must be 1 if EN_SYN = FALSE
  --  en_syn => FALSE,                  -- Specifies FIFO as dual-clock (FALSE) or Synchronous (TRUE)
  --  fifo_mode => "FIFO18",            -- Sets mode to FIFO18 or FIFO18_36
  --  first_word_fall_through => TRUE,  -- Sets the FIFO FWFT to FALSE, TRUE
  --  init => x"000000000",             -- Initial values on output port
  --  sim_device => "7SERIES",          -- Must be set to "7SERIES" for simulation behavior
  --  srval => x"000000000"             -- Set/Reset value for output port
  --) port map (
  --  -- Read Data: 32-bit (each) output: Read output data
  --  do => ser_tx_data,          -- 32-bit output: Data output
  --  dop => open,                -- 4-bit output: Parity data output
  --  -- Status: 1-bit (each) output: Flags and other FIFO status outputs
  --  almostempty => open,        -- 1-bit output: Almost empty flag
  --  almostfull => o_tx_full,    -- 1-bit output: Almost full flag
  --  empty => o_tx_empty,        -- 1-bit output: Empty flag
  --  full => open,               -- 1-bit output: Full flag
  --  rdcount => open,            -- 12-bit output: Read count
  --  rderr => open,              -- 1-bit output: Read error
  --  wrcount => open,            -- 12-bit output: Write count
  --  wrerr => open,              -- 1-bit output: Write error
  --  -- Read Control Signals: 1-bit (each) input: Read clock, enable and reset input signals
  --  rdclk => clk,            -- 1-bit input: Read clock
  --  rden => ser_tx_rd,       -- 1-bit input: Read enable
  --  regce => ser_ck_en,               -- 1-bit input: Clock enable
  --  rst => tx_rst,               -- 1-bit input: Asynchronous Reset
  --  rstreg => tx_rst,            -- 1-bit input: Output register set/reset
  --  -- Write Control Signals: 1-bit (each) input: Write clock and enable input signals
  --  wrclk => clk,            -- 1-bit input: Write clock
  --  wren => tx_wr,           -- 1-bit input: Write enable
  --  -- Write Data: 32-bit (each) input: Write input data
  --  di => tx_data_ext,              -- 32-bit input: Data input
  --  dip => ( others => '0' )    -- 4-bit input: Parity input
  --);
  
  ser_tx_rd <= ser_ck_ten and ser_tx_rdy and not o_tx_empty and not tx_rst;


  
  
  i_tx : entity work.ser_tx_sync generic map (
    data_size   => 8,
    stop_bits   => 1,
    inter_gap   => 8,
    msb_first   => msb_first
  ) port map (
    rst         => tx_rst,
    clk         => clk,
    ce          => ser_ck_ten,
    tx_d        => ser_tx_data( data_size - 1 downto 0 ),
    tx_dv       => ser_tx_rd,
    tx_rdy      => ser_tx_rdy,
    tx_ser      => ser_tx
  );
  
  i_rx : entity work.ser_rx_sync generic map (
    data_size   => 8,
    stop_bits   => 1,
    msb_first   => msb_first
  ) port map (
    rst       => rx_rst,
    clk       => clk,
    ce        => ser_ck_ren,
    rx_d      => ser_rx_data( data_size - 1 downto 0 ),
    rx_vld    => ser_rx_rdy,
    rx_ser    => ser_rx_r2
  );
  
  ser_rx_data( ser_rx_data'left downto data_size ) <= ( others => '-' );
  
  ser_rx_wr <= ser_ck_ren and ser_rx_rdy and not o_rx_full and not tx_rst;

  i_rxfifo : entity work.small_sfifo32
    generic map (
      width => 9)
    port map (
        clk    => clk,
        rst    => rx_rst,
        wadv   => ser_rx_wr,
        wdata  => ser_rx_data(8 downto 0),
        nfull  => o_rx_full,
        radv   => i_rx_rd,
        rempty => o_rx_empty,
        rdata  => o_rx_data(8 downto 0));
  
  --i_rxfifo : unisim.vcomponents.fifo18e1 generic map (
  --  almost_empty_offset => X"0080",   -- Sets the almost empty threshold
  --  almost_full_offset => X"0080",    -- Sets almost full threshold
  --  data_width => 9,          -- Sets data width to 4-36
  --  do_reg => 1,                      -- Enable output register (1-0) Must be 1 if EN_SYN = FALSE
  --  en_syn => FALSE,                  -- Specifies FIFO as dual-clock (FALSE) or Synchronous (TRUE)
  --  fifo_mode => "FIFO18",            -- Sets mode to FIFO18 or FIFO18_36
  --  first_word_fall_through => TRUE,  -- Sets the FIFO FWFT to FALSE, TRUE
  --  init => X"000000000",             -- Initial values on output port
  --  sim_device => "7SERIES",          -- Must be set to "7SERIES" for simulation behavior
  --  srval => X"000000000"             -- Set/Reset value for output port
  --) port map (
  --  -- Read Data: 32-bit (each) output: Read output data
  --  do => o_rx_data,            -- 32-bit output: Data output
  --  dop => open,                -- 4-bit output: Parity data output
  --  -- Status: 1-bit (each) output: Flags and other FIFO status outputs
  --  almostempty => open,        -- 1-bit output: Almost empty flag
  --  almostfull => o_rx_full,    -- 1-bit output: Almost full flag
  --  empty => o_rx_empty,       -- 1-bit output: Empty flag
  --  full => open,               -- 1-bit output: Full flag
  --  rdcount => open,            -- 12-bit output: Read count
  --  rderr => open,              -- 1-bit output: Read error
  --  wrcount => open,            -- 12-bit output: Write count
  --  wrerr => open,              -- 1-bit output: Write error
  --  -- Read Control Signals: 1-bit (each) input: Read clock, enable and reset input signals
  --  rdclk => clk,               -- 1-bit input: Read clock
  --  rden => i_rx_rd,            -- 1-bit input: Read enable
  --  regce => '1',               -- 1-bit input: Clock enable
  --  rst => rx_rst,                 -- 1-bit input: Asynchronous Reset
  --  rstreg => rx_rst,              -- 1-bit input: Output register set/reset
  --  -- Write Control Signals: 1-bit (each) input: Write clock and enable input signals
  --  wrclk => clk,            -- 1-bit input: Write clock
  --  wren => ser_rx_wr,          -- 1-bit input: Write enable
  --  -- Write Data: 32-bit (each) input: Write input data
  --  di => ser_rx_data,          -- 32-bit input: Data input
  --  dip => ( others => '0' )    -- 4-bit input: Parity input
  --);

end architecture rtl;
