--
-- adb3_ocp_reorder_dbuf_ur.vhd
--
-- Multi-slot buffering for adb3_ocp_reorder component.
--
-- This implementation uses a unified RAM that is large enough for
-- 2**slot_order * num_slot OCP data words. The data for a given slot
-- dynamically occupies a range of locations in the RAM, with the RAM addresses
-- chosen to put the data into the correct order. Data is then read out
-- strictly in ascending address, resulting in the required data order.
-- That is, read responses are returned in same order as the corresponding read
-- commands were issued on the in-order side.
--

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_misc.all;
use ieee.numeric_std.all;
use ieee.math_real.all;

library work;
use work.adb3_ocp.all;

entity adb3_ocp_reorder_dbuf_ur is
  generic(
    data_width      : in natural;   -- OCP data width, in bits
    num_slot        : in natural;   -- No. of slots
    slot_order      : in natural;   -- Log2 of no. of OCP data words that each slot must be able to hold
    output_pipe     : in boolean);  -- true => extra pipeline stage to mitigate BlockRAM clock-to-out
  port(
    rst             : in  std_logic;
    sr              : in  std_logic;
    clk             : in  std_logic;
    -- This is a subset of unordered side OCP signals
    uo_resp         : in  ocp_RespT;
    uo_data         : in  std_logic_vector(data_width - 1 downto 0);
    uo_tag          : in  std_logic_vector(ADB3_OCP_TAG_WIDTH - 1 downto 0);
    uo_resp_accept  : out std_logic;
    -- Read side
    read_slot_1hot  : in  std_logic_vector(num_slot - 1 downto 0);
    read_slot_valid : in  std_logic;
    read            : in  std_logic;
    read_data       : out std_logic_vector(data_width - 1 downto 0);
    read_valid      : out std_logic;
    -- Slot management
    alloc           : in  std_logic;
    alloc_tag       : in  std_logic_vector(ADB3_OCP_TAG_WIDTH - 1 downto 0);
    alloc_length    : in  std_logic_vector(ADB3_OCP_BURST_WIDTH - 1 downto 0);
    alloc_slot_1hot : in  std_logic_vector(num_slot - 1 downto 0);
    free            : in  std_logic;
    free_slot_1hot  : in  std_logic_vector(num_slot - 1 downto 0);
    error_pulse     : out std_logic_vector(7 downto 0));
end adb3_ocp_reorder_dbuf_ur;

architecture rtl of adb3_ocp_reorder_dbuf_ur is

  -- Return a binary representation of a std_logic_vector value.
  function conv_string_binary(
    constant v : std_logic_vector)
  return string is
    variable tmp : std_logic_vector(1 to v'length);
    variable str : string(1 to v'length);
  begin
    tmp := v;
    for i in str'range loop
      str(i) := std_logic'image(tmp(i))(1);
    end loop;
    return str;
  end function;

  -- Return a hexadecimal string representation of a std_logic_vector value.
  function conv_string_hex(
    constant v : std_logic_vector)
  return string is
    variable n : natural := ((v'length + 3) / 4) * 4;
    variable ndigit : natural := n / 4;
    variable tmp : std_logic_vector(n - 1 downto 0);
    variable digit : std_logic_vector(3 downto 0);
    variable str : string(1 to ndigit);
  begin
    tmp(v'length - 1 downto 0) := v;
    if n > v'length then
      tmp(n - 1 downto v'length) := (others => '0');
    end if;
    for i in str'range loop
      digit := tmp(4 * i - 1 downto 4 * (i - 1));
      case digit is
        when "0000" =>
          str(ndigit - i + 1) := '0';

        when "0001" =>
          str(ndigit - i + 1) := '1';

        when "0010" =>
          str(ndigit - i + 1) := '2';

        when "0011" =>
          str(ndigit - i + 1) := '3';

        when "0100" =>
          str(ndigit - i + 1) := '4';

        when "0101" =>
          str(ndigit - i + 1) := '5';

        when "0110" =>
          str(ndigit - i + 1) := '6';

        when "0111" =>
          str(ndigit - i + 1) := '7';

        when "1000" =>
          str(ndigit - i + 1) := '8';

        when "1001" =>
          str(ndigit - i + 1) := '9';

        when "1010" =>
          str(ndigit - i + 1) := 'A';

        when "1011" =>
          str(ndigit - i + 1) := 'B';

        when "1100" =>
          str(ndigit - i + 1) := 'C';

        when "1101" =>
          str(ndigit - i + 1) := 'D';

        when "1110" =>
          str(ndigit - i + 1) := 'E';

        when "1111" =>
          str(ndigit - i + 1) := 'F';

        when others =>
          str(ndigit - i + 1) := 'X';
      end case;
    end loop;
    return str;
  end function;

  -- Figure out how much RAM we need
  constant slot_size : natural := 2**slot_order;
  constant required_ram_size : natural := slot_size * num_slot;
  constant ram_order : natural := natural(floor(log2(real(2 * required_ram_size - 1) + 0.5)));
  constant ram_size : natural := 2**ram_order;

  -- Type for holding a RAM address
  subtype ram_addr_t is std_logic_vector(ram_order - 1 downto 0);
  -- Type for holding a RAM data word
  subtype ram_data_t is std_logic_vector(data_width - 1 downto 0);
  -- Type for inferring a RAM
  type ram_data_array_t is array(0 to ram_size - 1) of ram_data_t;

  -- Type for holding a one-hot vector
  subtype one_hot_t is std_logic_vector(num_slot - 1 downto 0);

  -- -------------------------------------------------------------------------
  -- Slot management (SM) signals.
  -- -------------------------------------------------------------------------

  type read_command_slot_t is record
    -- Tag value for command
    tag : std_logic_vector(ADB3_OCP_TAG_WIDTH - 1 downto 0);
    -- Position in RAM at which data from unordered side will be written
    write_pos : ram_addr_t;
  end record;
  type read_command_slot_array_t is array(natural range <>) of read_command_slot_t;
  constant UNOCCUPIED_SLOT : read_command_slot_t := (
    tag => (others => '-'),
    write_pos => (others => '-'));

  signal sm_slots : read_command_slot_array_t(0 to num_slot - 1) := (others => UNOCCUPIED_SLOT);
  signal sm_slots_occupied : std_logic_vector(num_slot - 1 downto 0) := (others => '0');

  -- -------------------------------------------------------------------------
  -- Matching incoming tag to a slot
  -- -------------------------------------------------------------------------

  signal tag_match : std_logic;
  signal tag_match_1hot : one_hot_t;
  signal tag_match_valid : std_logic := '0';
  signal tag_q : std_logic_vector(ADB3_OCP_TAG_WIDTH - 1 downto 0);

  -- -------------------------------------------------------------------------
  -- Data RAM signals, write side
  -- -------------------------------------------------------------------------

  signal ram_acc_ptr : ram_addr_t := (others => '0');
  signal ram_write_addr : ram_addr_t;
  signal ram_write : std_logic;
  signal ram_write_1hot : one_hot_t;
  signal ram_write_data : ram_data_t;
  signal ram : ram_data_array_t;
  attribute RAM_STYLE : string;
  attribute RAM_STYLE of ram : signal is "block";

  -- -------------------------------------------------------------------------
  -- Data RAM signals, read side
  -- -------------------------------------------------------------------------

  signal ram_read : std_logic;
  signal ram_read_adv : std_logic;
  signal ram_read_ptr : ram_addr_t := (others => '0');
  signal sel_write_pos : ram_addr_t;
  signal ram_level : ram_addr_t := (others => '0');
  signal ram_level_nonzero : std_logic;
  signal ram_read_data : ram_data_t;
  signal ram_read_data_valid : std_logic := '0';
  signal ram_pipe_en : std_logic;
  signal ram_pipe_data_q : ram_data_t;
  signal ram_pipe_data_q_valid : std_logic := '0';

  -- -------------------------------------------------------------------------
  -- Debug/status
  -- -------------------------------------------------------------------------

  signal i_error_pulse : std_logic_vector(7 downto 0) := (others => '0');

begin

  error_pulse <= i_error_pulse;

  uo_resp_accept <= '1'; -- OK to assume that in-order side master won't request more data than we can buffer?

  implement_slots : process(rst, clk)
    variable case_tmp : std_logic_vector(1 downto 0);
  begin
    if rst = '1' then
      sm_slots <= (others => UNOCCUPIED_SLOT);
      sm_slots_occupied <= (others => '0');
    elsif rising_edge(clk) then
      for i in 0 to num_slot - 1 loop
        if sr = '1' then
          sm_slots(i) <= UNOCCUPIED_SLOT;
          sm_slots_occupied(i) <= '0';
        else
          case_tmp(0) := alloc AND alloc_slot_1hot(i);
          case_tmp(1) := free AND free_slot_1hot(i);
          case case_tmp is
            when "01" =>
              -- synthesis translate_off
              assert sm_slots_occupied(i) = '0'
                report "adb3_ocp_reorder_dbuf_ur: Attempt to allocate already occupied slot, alloc_slot_1h=" & conv_string_binary(alloc_slot_1hot)
                severity error;
              -- synthesis translate_on
              sm_slots_occupied(i) <= '1';
              sm_slots(i).tag <= alloc_tag;
              sm_slots(i).write_pos <= ram_acc_ptr;

            when "10" =>
              -- synthesis translate_off
              assert sm_slots_occupied(i) = '1'
                report "adb3_ocp_reorder_dbuf_ur: Attempt to free unoccupied slot, alloc_slot_1h=" & conv_string_binary(free_slot_1hot)
                severity error;
              -- synthesis translate_on;
              sm_slots_occupied(i) <= '0';
              sm_slots(i).write_pos <= (others => '-');

            when "00" =>
              null;

            when "11" =>
              -- Illegal, should never happen
              -- synthesis translate_off
              assert false report "adb3_ocp_reorder_dbuf_ur: Slot " & natural'image(i) & ": attempt to occupy and free at the same time" severity error;
              -- synthesis translate_on
              sm_slots_occupied(i) <= '-';
              sm_slots(i).tag <= (others => '-');
              sm_slots(i).write_pos <= (others => '-');

            when others =>
              -- Illegal, should never happen
              sm_slots_occupied(i) <= '-';
              sm_slots(i).tag <= (others => '-');
              sm_slots(i).write_pos <= (others => '-');
          end case;

          case_tmp(0) := alloc AND alloc_slot_1hot(i);
          case_tmp(1) := tag_match_valid AND tag_match_1hot(i);
          case case_tmp is
            when "01" =>
              sm_slots(i).write_pos <= ram_acc_ptr;

            when "10" =>
              sm_slots(i).write_pos <= sm_slots(i).write_pos + 1;

            when "00" =>
              null;

            when others =>
              -- Illegal, should never happen
              sm_slots(i).write_pos <= (others => '-');
          end case;
        end if;
      end loop;
    end if;
  end process;


  -- -------------------------------------------------------------------------
  -- Match the tag in the response from the unordered side to a particular
  -- slot.
  -- -------------------------------------------------------------------------

  match_tag_to_slot : process(rst, clk)
    variable matched : std_logic;
  begin
    if rst = '1' then
      tag_match <= '-';
      tag_match_1hot <= (others => '-');
      tag_match_valid <= '0';
      tag_q <= (others => '-');
    elsif rising_edge(clk) then
      if sr = '1' then
        tag_match <= '-';
        tag_match_1hot <= (others => '-');
        tag_match_valid <= '0';
        tag_q <= (others => '-');
      else
        if uo_resp = OCP_RESP_VALID then
          matched := '0';
          for i in 0 to num_slot - 1 loop
            if sm_slots_occupied(i) = '1' AND sm_slots(i).tag = uo_tag then
              tag_match_1hot(i) <= '1';
              matched := matched OR '1';
            else
              tag_match_1hot(i) <= '0';
            end if;
          end loop;
          tag_match <= matched;
          tag_match_valid <= '1';
          tag_q <= uo_tag;
        else
          tag_match_1hot <= (others => '-');
          tag_match <= '-';
          tag_match_valid <= '0';
          tag_q <= (others => '-');
        end if;

        -- Sanity check
        if tag_match_valid = '1' AND tag_match = '0' then
          -- synthesis translate_off
          assert false
            report "adb3_ocp_reorder_dbuf_ur: Unordered side response with tag 0x" & conv_string_hex(tag_q) & " not matched to any slot"
            severity error;
          -- synthesis translate_on
          i_error_pulse(0) <= '1';
        else
          i_error_pulse(1) <= '0';
        end if;
      end if;
    end if;
  end process;


  -- -------------------------------------------------------------------------
  -- RAM accounting pointer
  -- This represents the extent to which locations in the RAM have been
  -- reserved for data from the unordered side. It is incremented by the
  -- burst length specified when a slot is allocated.
  -- -------------------------------------------------------------------------

  gen_ram_acc_ptr : process(rst, clk)
    variable tmp : std_logic_vector(ADB3_OCP_BURST_WIDTH - 1 downto 0);
  begin
    if rst = '1' then
      ram_acc_ptr <= (others => '0');
    elsif rising_edge(clk) then
      if sr = '1' then
        ram_acc_ptr <= (others => '0');
      else
        if alloc = '1' then
          tmp := ram_acc_ptr + alloc_length;
          ram_acc_ptr <= std_logic_vector(resize(unsigned(tmp), ram_order));
        end if;
      end if;
    end if;
  end process;


  -- -------------------------------------------------------------------------
  -- Data RAM, write side
  -- -------------------------------------------------------------------------

  ram_write <= tag_match_valid;

  -- The address the RAM at which to write word of data from unordered side
  -- depends upon which slot was matched to the tag.
  gen_ram_write_addr : process(
    tag_match_1hot,
    sm_slots)
    variable tmp : ram_addr_t;
  begin
    tmp := (others => '0');
    for i in 0 to num_slot - 1 loop
      if tag_match_1hot(i) = '1' then
        tmp := tmp OR sm_slots(i).write_pos;
      end if;
    end loop;
    ram_write_addr <= tmp;

  end process;

  write_to_ram : process(clk)
  begin
    if rising_edge(clk) then
      ram_write_data <= uo_data; -- Because ram_write is 1 cycle delayed

      if ram_write = '1' then
        ram(conv_integer(ram_write_addr)) <= ram_write_data;
      end if;
    end if;
  end process;

  -- -------------------------------------------------------------------------
  -- Data RAM, read side
  -- -------------------------------------------------------------------------
  -- synthesis translate_off
  -- Some sanity checks
  read_sanity_check : process(clk)
  begin
    if rising_edge(clk) then
      if read = '1' then
        assert read_slot_valid = '1'
          report "adb3_ocp_reorder_dbuf_ur: read asserted without read_slot_valid"
          severity error;
        if read_slot_valid = '1' then
          assert NOT is_x(read_slot_1hot)
            report "adb3_ocp_reorder_dbuf_ur: read asserted with read_slot_1hot invalid"
            severity error;
        end if;
      end if;
    end if;
  end process;
  -- synthesis translate_on

  -- Select the write_pos value for the current slot
  select_write_pos : process(
    read_slot_1hot,
    sm_slots)
    variable tmp : ram_addr_t;
  begin
    tmp := (others => '0');
    for i in 0 to num_slot - 1 loop
      if read_slot_1hot(i) = '1' then
        tmp := tmp OR sm_slots(i).write_pos;
      end if;
    end loop;
    sel_write_pos <= tmp;
  end process;

  -- Calculate number of words available for reading in RAM
  gen_ram_level : process(rst, clk)
  begin
    if rst = '1' then
      ram_level <= (others => '0');
    elsif rising_edge(clk) then
      if sr = '1' then
        ram_level <= (others => '0');
      else
        if read_slot_valid = '1' then
          -- Not changing slots, calculate ram_level normally
          if ram_read_adv = '1' then
            ram_level <= sel_write_pos - ram_read_ptr - 1;
          else
            ram_level <= sel_write_pos - ram_read_ptr;
          end if;
        else
          -- If changing slots, ram_level must not depend on sel_write_pos
          if ram_read_adv = '1' then
            ram_level <= ram_level - 1;
          end if;
        end if;
      end if;
    end if;
  end process;

  ram_level_nonzero <= OR_reduce(ram_level);

  ram_read_adv <= ram_level_nonzero AND (read OR                     -- Current word is being read
                                         NOT ram_read_data_valid OR  -- RAM output is not valid
                                         NOT ram_pipe_data_q_valid); -- RAM output (pipelined) is not valid

  ram_read <= read OR                      -- Current word is being read
              NOT ram_read_data_valid OR   -- RAM output is not valid
              NOT ram_pipe_data_q_valid;   -- RAM output (pipelined) is not valid

  read_from_ram : process(rst, clk)
  begin
    if rising_edge(clk) then
      if ram_read = '1' then
        ram_read_data <= ram(conv_integer(ram_read_ptr));
      end if;
    end if;

    if rst = '1' then
      ram_read_ptr <= (others => '0');
      ram_read_data_valid <= '0';
    elsif rising_edge(clk) then
      if sr = '1' then
        ram_read_ptr <= (others => '0');
        ram_read_data_valid <= '0';
      else
        if ram_read_adv = '1' then
          ram_read_ptr <= ram_read_ptr + 1;
        end if;
        if ram_read = '1' then
          ram_read_data_valid <= ram_level_nonzero;
        end if;
      end if;
    end if;
  end process;

  gen_ram_pipe : if output_pipe generate

    --
    -- This results in an extra level of pipelining right after the BlockRAM
    -- data outputs.
    --

    ram_pipe_en <= read OR                    -- Current word is being read
                   NOT ram_pipe_data_q_valid; -- Output is not valid

    ram_pipe : process(rst, clk)
    begin
      if rising_edge(clk) then
        if ram_pipe_en = '1' then
          ram_pipe_data_q <= ram_read_data;
        end if;
      end if;

      if rst = '1' then
        ram_pipe_data_q_valid <= '0';
      elsif rising_edge(clk) then
        if sr = '1' then
          ram_pipe_data_q_valid <= '0';
        else
          if ram_pipe_en = '1' then
            ram_pipe_data_q_valid <= ram_read_data_valid;
          end if;
        end if;
      end if;
    end process;

    read_data <= ram_pipe_data_q;
    read_valid <= ram_pipe_data_q_valid;

  end generate;

  gen_no_ram_pipe : if NOT output_pipe generate

    -- This results in no extra pipelining of the BlockRAM outputs.
    
    -- A constant '1' will cause 'ram_pipe_data_q_valid' to have no effect in above logic
    ram_pipe_data_q_valid <= '1';
    
    -- Unused if NOT output_pipe.
    ram_pipe_data_q <= (others => '-');

    read_data <= ram_read_data;
    read_valid <= ram_read_data_valid;

  end generate;

end rtl;
