--
-- adb3_ocp_reorder_sfifo.vhd
--
-- Synchronous FIFO used by adb3_ocp_reorder module.
--
-- Depth is exactly equal to 2^ORDER.
--
-- Number of words in FIFO at any given moment is given by (rn + rvalid). The
-- RAM array stores up to DEPTH-1 words and the output register holds one word
-- if 'valid' is asserted.
--

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_misc.all;
use ieee.numeric_std.all;

entity adb3_ocp_reorder_sfifo is
generic(
  -- Expect to use distributed RAM, which puts limits on ORDER
  ORDER       : in    natural range 2 to 9;
  -- Width of a FIFO data word, in bits
  WIDTH       : in    natural;          
  -- If true, 'sr' pulse width can be as short as 1 cycle  
  FAST_SR     : in    boolean := false);
port(
  rst         : in    std_logic;                             -- Async. reset
  clk         : in    std_logic;                             -- Clock
  sr          : in    std_logic;                             -- Sync. reset
  w           : in    std_logic;                             -- '1' => write value on 'wd' into FIFO
  wd          : in    std_logic_vector(WIDTH - 1 downto 0);  -- Data word to write, qualified by 'w'
  wfull       : out   std_logic;                             -- '1' => FIFO is full
  r           : in    std_logic;                             -- '1' => Advance 'rq' to next data word
  rvalid      : out   std_logic;                             -- '1' => FIFO not empty, i.e. 'rq' is valid
  rq          : out   std_logic_vector(WIDTH - 1 downto 0)); -- Current data word, qualified by 'rvalid'
end adb3_ocp_reorder_sfifo;

architecture rtl of adb3_ocp_reorder_sfifo is

  constant DEPTH : natural := 2**ORDER;

  -- RAM write pointer 
  signal wptr : std_logic_vector(ORDER - 1 downto 0) := (others => '0');
  
  -- RAM read pointer
  signal rptr : std_logic_vector(ORDER - 1 downto 0) := (others => '0');
  
  -- Number of words in FIFO, not including data output reg.
  signal rn : std_logic_vector(ORDER - 1 downto 0) := (others => '0');
  signal rn_nonzero : std_logic;
  
  -- '1' => FIFO is not empty
  signal valid : std_logic := '0';
  
  -- '1' => FIFO is full 
  signal full : std_logic := '0';
  
  -- Clock enable for data output register
  signal reg_ce : std_logic; 
  
  -- When asserted, advances RAM read pointer
  signal adv_rptr : std_logic; 
  
  -- Current data word being output by FIFO
  signal q : std_logic_vector(WIDTH - 1 downto 0); 

  -- The RAM, holds up to DEPTH-1 words
  subtype ram_data_t is std_logic_vector(WIDTH - 1 downto 0);
  type ram_array_t is array(0 to DEPTH - 1) of ram_data_t;
  signal ram : ram_array_t;
  attribute RAM_STYLE : string;
  attribute RAM_STYLE of ram : signal is "distributed";
  signal sel_ram : ram_data_t;

begin

  wfull <= full;
  rq <= q;
  rvalid <= valid;

  write_logic : process(rst, clk)
    variable tmp : std_logic_vector(ORDER - 1 downto 0);
  begin
    if rst = '1' then
      wptr <= (others => '0');
      full <= '0';
    elsif rising_edge(clk) then
      if sr = '1' then
        wptr <= (others => '0');
        full <= '0';
      else
        if w = '1' then
          wptr <= wptr + 1;
        end if;

        if w = '1' then
          assert full = '0' report "adb3_ocp_reorder_sfifo overflow" severity error;
          tmp := wptr - rptr + 1;
        else
          tmp := wptr - rptr;
        end if;
        full <= AND_REDUCE(tmp);
      end if;
    end if;

    if rising_edge(clk) then
      if w = '1' then
        ram(conv_integer(wptr)) <= wd;
      end if;
    end if;
  end process;

  rn_nonzero <= OR_reduce(rn);
  reg_ce <= r OR NOT valid;
  adv_rptr <= (r OR NOT valid) AND rn_nonzero;

  read_logic : process(rst, clk)
  begin
    if rst = '1' then
      rptr <= (others => '0');
    elsif rising_edge(clk) then
      if sr = '1' then
        rptr <= (others => '0');
      else
        if adv_rptr = '1' then
          rptr <= rptr + 1;
        end if;
      end if;
    end if;

    if rising_edge(clk) then
      if FAST_SR AND sr = '1' then
        rn <= (others => '0');
      else
        if adv_rptr = '1' then
          rn <= wptr - rptr - 1;
        else
          rn <= wptr - rptr;
        end if;
      end if;
    end if;
    
    if rst = '1' then
      null;
    elsif rising_edge(clk) then
      if sr = '1' then
        null;
      else
        if r = '1' then
          assert valid = '1' report "adb3_ocp_reorder_sfifo underflow" severity error;
        end if;
      end if;
    end if;
  end process;

  generate_valid : process(rst, clk)
  begin
    if rst = '1' then
      valid <= '0';
    elsif rising_edge(clk) then
      if sr = '1' then
        valid <= '0';
      else
        if reg_ce = '1' then
          valid <= rn_nonzero;
        end if;
      end if;
    end if;
  end process;

  sel_ram <= ram(conv_integer(rptr));

  generate_q : process(clk)
  begin
    if rising_edge(clk) then
      if reg_ce = '1' then
        if rn_nonzero = '1' then
          q <= sel_ram;
        else
          -- Assign 'q' to undefined on FIFO empty to assist in simulation debug
          q <= (others => '-');
        end if;
      end if;
    end if;
  end process;

end rtl;
