--==============================================================================
--
--     File Name: adb3_ocp_dm_adapter.vhd
--         Model: Synthesisable
--
--  Dependencies: adb3_ocp (package)
--                adb3_ocp_dm (package)
--
--       Company: Alpha Data
--        Design: 
--
--   Description: Adapter to convert general OCP commands into OCP commands
--                that satisfy the constraints of the Direct Master OCP port.
--
--                For OCP reads, the constraints on commands issued to the
--                Direct Master OCP port ('master_m2s' & 'master_s2m') are:
--                 R1. OCP read command length shall be less than or equal to
--                     the 'max_rd_req' value.
--                 R2. No OCP read burst shall cross a 4 kiB address boundary.
--                 R3. The OCP commands shall correspond one-to-one with PCI
--                     Express read transactions.
--
--                For OCP writes, the constraints on commands issued to the
--                Direct Master OCP port ('master_m2s' & 'master_s2m') are:
--                 W1. OCP write command length shall be less than or equal to
--                     the 'max_payload' value.
--                 W2. No OCP write burst shall cross a 4 kiB address boundary.
--                 W3. The byte enables accompanying OCP write data words shall
--                     satisfy PCI Express first+last 32-bit word byte enable
--                     rules.
--                 W4. The OCP commands shall correspond one-to-one with PCI
--                     Express write transactions.
--
--                The above constraints DO NOT apply to OCP commands accepted
--                on the OCP slave port ('slave_m2s' & 'slave_s2m').
--
--   Limitations:
--
--         Notes: 
--
--    Disclaimer: THESE DESIGNS ARE PROVIDED "AS IS" WITH NO WARRANTY
--                WHATSOEVER AND ALPHA DATA SPECIFICALLY DISCLAIMS ANY
--                IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR
--                A PARTICULAR PURPOSE, OR AGAINST INFRINGEMENT.
--
--                Copyright (C) 2013-2014 Alpha Data, All rights reserved
--
--     Revisions: 0.1, TW, 14 Nov 2013: Initial design
--                0.2, TW, 19 Feb 2013:
--                  - Changed to be a wrapper around adb3_ocp_dm_adapter_base.
--                  - MPTL compatibility is now a generic, rather than a
--                    control input.
--                  - Changed maximum allowed DWFIFO_ORDER to 9, to enable
--                    maximum possible PCI-E payload of 4096 kiB to be
--                    generated on master interface.
--
--==============================================================================

library ieee;
use ieee.std_logic_1164.all;

library work;
use work.adb3_ocp.all;
use work.adb3_ocp_dm_int.all;

entity adb3_ocp_dm_adapter is
generic(
  -- log2(Size of the DM-side command FIFO)
  DCFIFO_ORDER    : in    natural range 4 to 8 := 4;
  -- log2(Size of the OCP-side command FIFO)
  OCFIFO_ORDER    : in    natural range 4 to 8 := 4;
  -- log2(Size of the DM-side data & byte enable FIFO)
  DWFIFO_ORDER    : in    natural range 4 to 9 := 6;
  -- log2(Size of the OCP-side data & byte enable FIFO)
  OWFIFO_ORDER    : in    natural range 4 to 8 := 4;
  -- true  => Generated write bursts are also compatible with MPTL payload.
  -- false => Generated write bursts in general are not compatible with MPTL
  --          payload.
  MPTL_COMPATIBLE : in    boolean := true;
  -- true  => For write bursts, non-OCP word address bits of master_m2s.Addr
  --          indicate the first valid byte lane of the first word of the burst
  --          and master_m2s.Tag (normally ignored by slave) indicates the last
  --          valid byte of the the last word of the burst.
  --
  --          master_m2s.Tag[5:4] is "--" (reserved)
  --          master_m2s.Tag[7] is '1' => indicates OCP command comes from Direct Master
  --          if (byte enables are arbitrary) then
  --            master_m2s.Addr[2:0] is "000"
  --            master_m2s.Addr[3] indicates first valid QWORD lane
  --            master_m2s.Tag[2:0] is "111" (reserved)
  --            master_m2s.Tag[3] indicates last valid QWORD lane
  --            master_m2s.Tag[6] is '1' => indicates arbitrary byte enables and BL = 1
  --          else
  --            master_m2s.Addr[3:0] indicates first valid byte lane
  --            master_m2s.Tag[3:0] indicates last valid byte lane
  --            master_m2s.Tag[6] is '1' => indicates contiguous byte enables
  --
  --          For read bursts, master_m2s.Addr[3:0] is "0000" and
  --          master_m2s.Tag functions normally.
  --
  -- false => Non-OCP word address bits of master_m2s.Addr are zero and
  --          master_m2s.Tag must be considered invalid unless the burst is a
  --          read.
  PCIE_HINTS      : in    boolean := false);
port(
  ocp_rst         : in    std_logic;
  ocp_clk         : in    std_logic;
  -- OCP slave port
  slave_m2s       : in    adb3_ocp_m2sT;
  slave_s2m       : out   adb3_ocp_s2mT;
  -- OCP master ports
  master_m2s      : out   adb3_ocp_m2sT;
  master_s2m      : in    adb3_ocp_s2mT;
  -- Control ports
  ctl_max_rd_req  : in    std_logic_vector(2 downto 0); -- Maximum read request, encoded as per PCI Express specification
  ctl_max_payload : in    std_logic_vector(2 downto 0); -- Maximum payload, encoded as per PCI Express specification
  -- Debug information
  --
  --    [63:0] CHOP
  --         [0] data_enable_q
  --         [1] cmd_enable_q
  --       [4:2] max_rd_req_q
  --       [7:5] max_payload_q
  --      [11:8] cp_state:
  --             0000 => idle
  --             0001 => read1
  --             0010 => read2
  --             0011 => write_normal
  --             0100 => write_general
  --     [23:12] cp_remaining;
  --     [31:24] cp_addr;
  --     [43:32] cp_start_addr;
  --     [52:44] cp_bcount;
  --        [53] cp_bfirst;
  --        [54] cp_blast;
  --        [55] cp_bboundary;
  --     [63:56] cp_dm_tag;
  --
  --   [75:64] OWFIFO:
  --     [73:64] rn
  --        [74] qvalid
  --        [75] nafull
  --
  --   [87:76] OCFIFO:
  --     [85:76] rn
  --        [86] qvalid
  --        [87] nafull
  --
  --   [99:88] RFIFO:
  --     [97:88] rn
  --        [98] qvalid
  --        [99] nafull
  --
  -- [111:100] DCFIFO:
  --    [99:100] rn
  --       [110] qvalid
  --       [111] nafull
  --
  -- [123:112] DWFIFO:
  --    [121:112] rn
  --       [122] qvalid
  --       [123] nafull
  --   
  debug_info      : out   std_logic_vector(127 downto 0));
end;


architecture wrap of adb3_ocp_dm_adapter is
  
begin

  U0 : adb3_ocp_dm_adapter_base
  generic map(
    DCFIFO_ORDER => DCFIFO_ORDER,
    OCFIFO_ORDER => OCFIFO_ORDER,
    DWFIFO_ORDER => DWFIFO_ORDER,
    OWFIFO_ORDER => OWFIFO_ORDER,
    MPTL_COMPATIBLE => MPTL_COMPATIBLE,
    PCIE_HINTS => PCIE_HINTS,
    DATA_WIDTH => ADB3_OCP_DATA_WIDTH)
  port map(
    ocp_rst => ocp_rst,
    ocp_clk => ocp_clk,
    slave_m2s_cmd => slave_m2s.Cmd,
    slave_m2s_addr => slave_m2s.Addr,
    slave_m2s_blen => slave_m2s.BurstLength,
    slave_m2s_tag => slave_m2s.Tag,
    slave_m2s_dval => slave_m2s.DataValid,
    slave_m2s_data => slave_m2s.Data,
    slave_m2s_be => slave_m2s.DataByteEn,
    slave_m2s_racc => slave_m2s.RespAccept,
    slave_s2m_cacc => slave_s2m.CmdAccept,
    slave_s2m_dacc => slave_s2m.DataAccept,
    slave_s2m_resp => slave_s2m.Resp,
    slave_s2m_data => slave_s2m.Data,
    slave_s2m_tag => slave_s2m.Tag,
    master_m2s_cmd => master_m2s.Cmd,
    master_m2s_addr => master_m2s.Addr,
    master_m2s_blen => master_m2s.BurstLength,
    master_m2s_tag => master_m2s.Tag,
    master_m2s_dval => master_m2s.DataValid,
    master_m2s_data => master_m2s.Data,
    master_m2s_be => master_m2s.DataByteEn,
    master_m2s_racc => master_m2s.RespAccept,
    master_s2m_cacc => master_s2m.CmdAccept,
    master_s2m_dacc => master_s2m.DataAccept,
    master_s2m_resp => master_s2m.Resp,
    master_s2m_data => master_s2m.Data,
    master_s2m_tag => master_s2m.Tag,
    ctl_max_rd_req => ctl_max_rd_req,
    ctl_max_payload => ctl_max_payload,
    debug_info => debug_info);
  
end;
