--
-- adb3_ocp_readen_read_fifo_interface128
-- Component of ADB3_OCP Library
--
-- (C) 2010 Alpha Data Ltd.
--
-- Read OCP FIFO with detection and management of "non-destructive" reads
-- indicated by Tag field bit 7.
-- For any "end of DMA chain" reads which cross a boundary and
-- result in 2 reads to same address, the first burst request tag
-- will have bit 7 set, indicating that the burst should not assert radv for
-- the last word of data

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_misc.all;

library work;
use work.adb3_ocp.all;

entity adb3_ocp_w_readen_read_fifo_interface256 is
  port (
    -- OCP Interface
    ocp_clk : in std_logic;
    ocp_m2s : in adb3w_ocp_m2sT;
    ocp_s2m : out adb3w_ocp_s2mT;
    -- FIFO Interface
    q    : in  std_logic_vector(255 downto 0);
    qv   : in  std_logic;
    radv : out std_logic);

end adb3_ocp_w_readen_read_fifo_interface256;

architecture rtl of adb3_ocp_w_readen_read_fifo_interface256 is

  signal nd_tag_inflight : std_logic := '0';
  signal data_requested_count : std_logic_vector(11 downto 0) := (others => '0');
  signal fifo_requested_count : std_logic_vector(11 downto 0) := (others => '0');
  signal data_requested : std_logic := '0';
  signal fifo_requested : std_logic := '0';

  component small_sfifo
    generic (
      width : integer := 64);
    port (
      clk    : in  std_logic;
      rst    : in  std_logic;
      wadv   : in  std_logic;
      wdata  : in  std_logic_vector(width-1 downto 0);
      nfull  : out std_logic;
      radv   : in  std_logic;
      rempty : out std_logic;
      rdata  : out std_logic_vector(width-1 downto 0));
  end component;

  signal tag_fifo_in,tag_fifo_out : std_logic_vector(15 downto 0);
  signal tag_burst_out : std_logic_vector(11 downto 0);
  signal tag_tag_out : std_logic_vector(7 downto 0);
  signal tag_fifo_wadv,tag_fifo_radv,tag_fifo_nfull,tag_fifo_empty : std_logic := '0';
  signal this_burst_count  : std_logic_vector(7 downto 0) := (others => '0');

  signal dadv : std_logic;
  
  signal ocp_m2s_respaccept : std_logic;
  signal q_reg : std_Logic_vector(255 downto 0);
  signal resp_reg : std_logic;  

  signal data_requested_limit_reached : std_logic := '0';
  
begin

  ocp_s2m.CmdAccept  <= (qv and not nd_tag_inflight and not data_requested_limit_reached) and (not tag_fifo_nfull) when ocp_m2s.Cmd = OCP_CMD_READ else '0';
  ocp_s2m.DataAccept <= '0';
  ocp_s2m.Tag        <= tag_tag_out;  
  ocp_s2m.Data       <= q_reg;
  ocp_s2m.Resp       <= '0' & resp_reg;

-- If using indirect connection, must use OCP respAccept
--  ocp_m2s_respaccept <= ocp_m2s.RespAccept
--  q_reg <= q;
--  resp_reg <= (qv and data_requested and not tag_fifo_empty);
-- If connected directly to Gen3 DMA engine, only ever used to read, respaccept is always '1'
-- So hardwire internally to improve timing, and register data
  ocp_m2s_respaccept <= '1';
  process (ocp_clk) is
    begin  -- process track_data
      if rising_edge(ocp_clk) then
         q_reg <= q;
         resp_reg <= (qv and data_requested and not tag_fifo_empty);
         tag_tag_out <= tag_fifo_out(7 downto 0);
      end if;
    end process;
  radv <= ocp_m2s_respaccept and fifo_requested and qv and not tag_fifo_empty;
  dadv <= ocp_m2s_respaccept and data_requested and qv and not tag_fifo_empty;

  track_data: process (ocp_clk) is
  begin  -- process track_data
    if rising_edge(ocp_clk) then            -- rising clock edge
      data_requested_limit_reached <= OR_reduce(data_requested_count(11 downto 6));
      -- Track data requested for OCP
      if qv = '1' and data_requested = '1' and ocp_m2s_respaccept = '1' and tag_fifo_empty = '0' then
        if ocp_m2s.Cmd = OCP_CMD_READ and qv = '1' and nd_tag_inflight = '0' and tag_fifo_nfull = '0' and data_requested_limit_reached = '0' then
          --data_requested_count <= data_requested_count + EXT(ocp_m2s.BurstLength,16) -1;
          data_requested_count <= data_requested_count + ocp_m2s.BurstLength -1;
          
          data_requested <= '1';
        else
          data_requested_count <= data_requested_count -1;
          if data_requested_count = X"001" then
            data_requested <= '0';
          end if;
        end if;
      else
        if ocp_m2s.Cmd = OCP_CMD_READ and qv = '1' and nd_tag_inflight = '0' and tag_fifo_nfull = '0' and data_requested_limit_reached = '0' then
          --data_requested_count <= data_requested_count + EXT(ocp_m2s.BurstLength,16);
          data_requested_count <= data_requested_count + ocp_m2s.BurstLength;
          data_requested <= '1';
        end if;
      end if;
      
      -- Track data requested from FIFO
      if ocp_m2s_respaccept = '1' and fifo_requested = '1' and qv = '1' and tag_fifo_empty = '0' then
        if ocp_m2s.Cmd = OCP_CMD_READ and qv = '1' and nd_tag_inflight = '0' and tag_fifo_nfull = '0' and data_requested_limit_reached = '0' then
          if ocp_m2s.Tag(7) = '1' then
            --fifo_requested_count <= fifo_requested_count + EXT(ocp_m2s.BurstLength,16) -2;
            --if fifo_requested_count + EXT(ocp_m2s.BurstLength-1,16) = X"0001" then
            fifo_requested_count <= fifo_requested_count + ocp_m2s.BurstLength -2;
            if fifo_requested_count + ocp_m2s.BurstLength = X"002" then                        
              fifo_requested <= '0';
            else
              fifo_requested <= '1';
            end if;
          else
            --fifo_requested_count <= fifo_requested_count + EXT(ocp_m2s.BurstLength,16) -1;                      
            fifo_requested_count <= fifo_requested_count + ocp_m2s.BurstLength -1;
            fifo_requested <= '1';
          end if;
        else
           fifo_requested_count <= fifo_requested_count -1;
           if fifo_requested_count = X"001" then
              fifo_requested <= '0';
           end if;
        end if;
      else
        if ocp_m2s.Cmd = OCP_CMD_READ and qv = '1' and nd_tag_inflight = '0' and tag_fifo_nfull = '0' and data_requested_limit_reached = '0' then
          if ocp_m2s.Tag(7) = '1' then
            --fifo_requested_count <= fifo_requested_count + EXT(ocp_m2s.BurstLength,16) -1;
            --if fifo_requested_count + EXT(ocp_m2s.BurstLength-1,16) = X"0000" then
            fifo_requested_count <= fifo_requested_count + ocp_m2s.BurstLength -1;                      
            if fifo_requested_count + ocp_m2s.BurstLength = X"001" then
              fifo_requested <= '0';
            else
              fifo_requested <= '1';
            end if;
          else
            --fifo_requested_count <= fifo_requested_count + EXT(ocp_m2s.BurstLength,16);
            fifo_requested_count <= fifo_requested_count + ocp_m2s.BurstLength;                       
            fifo_requested <= '1';
          end if;
        end if;
      end if;
      
      -- nd_tag_inflight flag
      -- blocks new read requests until all data requested has been read from
      -- FIFO up to end of current read 
      if ocp_m2s.Cmd = OCP_CMD_READ and qv = '1' and nd_tag_inflight = '0' and tag_fifo_nfull = '0' and ocp_m2s.Tag(7) = '1' and data_requested_limit_reached = '0' then
        nd_tag_inflight <= '1';
      elsif data_requested = '0' then
        nd_tag_inflight <= '0';
      end if;
      
      if ocp_m2s.Cmd = OCP_CMD_READ and qv = '1' and nd_tag_inflight = '0' and tag_fifo_nfull = '0' and data_requested_limit_reached = '0' then
        tag_fifo_wadv <= '1';
      else
        tag_fifo_wadv <= '0';
      end if;
      -- Assumes bursts less than 256 beats (8kB)
      tag_fifo_in <= ocp_m2s.BurstLength(7 downto 0) & ocp_m2s.Tag;
        
      if dadv = '1' then
        if this_burst_count = tag_burst_out(7 downto 0)-1 then
          this_burst_count <= (others => '0');
        else
          this_burst_count <= this_burst_count+1;
        end if;
      end if;
      
    end if;
  end process track_data;

  tag_fifo_radv <= dadv when this_burst_count = tag_burst_out(7 downto 0)-1  else '0';
  
  --tag_tag_out <= tag_fifo_out(7 downto 0);
  tag_burst_out <= "0000" & tag_fifo_out(15 downto 8);
  
  
  tag_fifo : small_sfifo
    generic map (
      width => 16)
    port map (
      clk    => ocp_clk,
      rst    => '0',
      wadv   => tag_fifo_wadv,
      wdata  => tag_fifo_in,
      nfull  => tag_fifo_nfull,
      radv   => tag_fifo_radv,
      rempty => tag_fifo_empty,
      rdata  => tag_fifo_out);

  
  
  
end rtl;
