--
-- Behavioural simulation model of ADM-PCIE-KU3-BCHI, Target FPGA side
--
-- IMPORTANT: Simulator resolution must be 1 ps or better.
--

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_misc.all;
use ieee.numeric_std.all;

entity adb3_admxrcku1_bm_x8_axi4 is
  generic(
    number_of_dma_engines    : natural range 0 to 4 := 1;
    number_of_pcie_dm_ports  : natural range 0 to 1 := 0;
    number_of_bar01_dm_ports : natural range 0 to 1 := 0;
    dma_engine0_config       : natural range 0 to 4 := 0;
    dma_engine1_config       : natural range 0 to 4 := 0;
    dma_engine2_config       : natural range 0 to 4 := 0;
    dma_engine3_config       : natural range 0 to 4 := 0;
    disable_model_io         : boolean := false;
    use_axi4_pcie_legalizer  : boolean := false;
    bar2_size                : natural range 15 to 21 := 15;
    enable_dma_abort         : boolean := false;
    enable_interrupt         : boolean := false;
    use_icap                 : boolean := true;
    use_diff_ref_clock       : boolean := true;
    pcie_dm_port_id_width    : natural range 1 to 4 := 1
  );
  port(
    -- PCIE
    perst_n                  : in    std_logic;
    pcie100_p                : in    std_logic;
    pcie100_n                : in    std_logic;
    refclk200_p              : in    std_logic := '0';
    refclk200_n              : in    std_logic := '1';
    refclk200_in             : in    std_logic := '0';
    refclk                   : out   std_logic;
    pci_exp_txn              : out   std_logic_vector(7 downto 0);
    pci_exp_txp              : out   std_logic_vector(7 downto 0);
    pci_exp_rxn              : in    std_logic_vector(7 downto 0);
    pci_exp_rxp              : in    std_logic_vector(7 downto 0);
    -- Model-specific I/O pins
    model_inout_i            : in    std_logic_vector(55 downto 0);
    model_inout_o            : out   std_logic_vector(55 downto 0);
    model_inout_t            : out   std_logic_vector(55 downto 0);
    -- VPD transport interface to 2nd core
    vpd_out                  : out   std_logic_vector(47 downto 0);

    -- Signals that are replicated between Target and Host via pcie_exp_{txn,txp,rxn,rxp}
    aclk : out std_logic;
    aresetn : out std_logic;
    ds_axi_araddr : out std_logic_vector(63 downto 0);
    ds_axi_arburst : out std_logic_vector(1 downto 0);
    ds_axi_arcache : out std_logic_vector(3 downto 0);
    ds_axi_arlen : out std_logic_vector(7 downto 0);
    ds_axi_arlock : out std_logic_vector(0 downto 0);
    ds_axi_arprot : out std_logic_vector(2 downto 0);
    ds_axi_arqos : out std_logic_vector(3 downto 0);
    ds_axi_arready : in std_logic := '0';
    ds_axi_arregion : out std_logic_vector(3 downto 0);
    ds_axi_arsize : out std_logic_vector(2 downto 0);
    ds_axi_arvalid : out std_logic;
    ds_axi_awaddr : out std_logic_vector(63 downto 0);
    ds_axi_awburst : out std_logic_vector(1 downto 0);
    ds_axi_awcache : out std_logic_vector(3 downto 0);
    ds_axi_awlen : out std_logic_vector(7 downto 0);
    ds_axi_awlock : out std_logic_vector(0 downto 0);
    ds_axi_awprot : out std_logic_vector(2 downto 0);
    ds_axi_awqos : out std_logic_vector(3 downto 0);
    ds_axi_awready : in std_logic := '0';
    ds_axi_awregion : out std_logic_vector(3 downto 0);
    ds_axi_awsize : out std_logic_vector(2 downto 0);
    ds_axi_awvalid : out std_logic;
    ds_axi_rdata : in std_logic_vector(255 downto 0) := (others => '0');
    ds_axi_rlast : in std_logic := '0';
    ds_axi_rready : out std_logic;
    ds_axi_rresp : in std_logic_vector(1 downto 0) := (others => '0');
    ds_axi_rvalid : in std_logic := '0';
    ds_axi_wdata : out std_logic_vector(255 downto 0);
    ds_axi_wlast : out std_logic;
    ds_axi_wready : in std_logic := '0';
    ds_axi_wstrb : out std_logic_vector(31 downto 0);
    ds_axi_wvalid : out std_logic;
    ds_axi_bready : out std_logic;
    ds_axi_bresp : in std_logic_vector(1 downto 0) := (others => '0');
    ds_axi_bvalid : in std_logic := '0';
    dma0_axi_araddr : out std_logic_vector(38 downto 0);
    dma0_axi_arburst : out std_logic_vector(1 downto 0);
    dma0_axi_arcache : out std_logic_vector(3 downto 0);
    dma0_axi_arlen : out std_logic_vector(7 downto 0);
    dma0_axi_arlock : out std_logic_vector(0 downto 0);
    dma0_axi_arprot : out std_logic_vector(2 downto 0);
    dma0_axi_arqos : out std_logic_vector(3 downto 0);
    dma0_axi_arready : in std_logic := '0';
    dma0_axi_arregion : out std_logic_vector(3 downto 0);
    dma0_axi_arsize : out std_logic_vector(2 downto 0);
    dma0_axi_arvalid : out std_logic;
    dma0_axi_awaddr : out std_logic_vector(38 downto 0);
    dma0_axi_awburst : out std_logic_vector(1 downto 0);
    dma0_axi_awcache : out std_logic_vector(3 downto 0);
    dma0_axi_awlen : out std_logic_vector(7 downto 0);
    dma0_axi_awlock : out std_logic_vector(0 downto 0);
    dma0_axi_awprot : out std_logic_vector(2 downto 0);
    dma0_axi_awqos : out std_logic_vector(3 downto 0);
    dma0_axi_awready : in std_logic := '0';
    dma0_axi_awregion : out std_logic_vector(3 downto 0);
    dma0_axi_awsize : out std_logic_vector(2 downto 0);
    dma0_axi_awvalid : out std_logic;
    dma0_axi_rdata : in std_logic_vector(255 downto 0) := (others => '0');
    dma0_axi_rlast : in std_logic := '0';
    dma0_axi_rready : out std_logic;
    dma0_axi_rresp : in std_logic_vector(1 downto 0) := (others => '0');
    dma0_axi_rvalid : in std_logic := '0';
    dma0_axi_wdata : out std_logic_vector(255 downto 0);
    dma0_axi_wstrb : out std_logic_vector(31 downto 0);
    dma0_axi_wlast : out std_logic;
    dma0_axi_wready : in std_logic := '0';
    dma0_axi_wvalid : out std_logic;
    dma0_axi_bready : out std_logic;
    dma0_axi_bresp : in std_logic_vector(1 downto 0) := (others => '0');
    dma0_axi_bvalid : in std_logic := '0';
    dma1_axi_araddr : out std_logic_vector(38 downto 0);
    dma1_axi_arburst : out std_logic_vector(1 downto 0);
    dma1_axi_arcache : out std_logic_vector(3 downto 0);
    dma1_axi_arlen : out std_logic_vector(7 downto 0);
    dma1_axi_arlock : out std_logic_vector(0 downto 0);
    dma1_axi_arprot : out std_logic_vector(2 downto 0);
    dma1_axi_arqos : out std_logic_vector(3 downto 0);
    dma1_axi_arready : in std_logic := '0';
    dma1_axi_arregion : out std_logic_vector(3 downto 0);
    dma1_axi_arsize : out std_logic_vector(2 downto 0);
    dma1_axi_arvalid : out std_logic;
    dma1_axi_awaddr : out std_logic_vector(38 downto 0);
    dma1_axi_awburst : out std_logic_vector(1 downto 0);
    dma1_axi_awcache : out std_logic_vector(3 downto 0);
    dma1_axi_awlen : out std_logic_vector(7 downto 0);
    dma1_axi_awlock : out std_logic_vector(0 downto 0);
    dma1_axi_awprot : out std_logic_vector(2 downto 0);
    dma1_axi_awqos : out std_logic_vector(3 downto 0);
    dma1_axi_awready : in std_logic := '0';
    dma1_axi_awregion : out std_logic_vector(3 downto 0);
    dma1_axi_awsize : out std_logic_vector(2 downto 0);
    dma1_axi_awvalid : out std_logic;
    dma1_axi_rdata : in std_logic_vector(255 downto 0) := (others => '0');
    dma1_axi_rlast : in std_logic := '0';
    dma1_axi_rready : out std_logic;
    dma1_axi_rresp : in std_logic_vector(1 downto 0) := (others => '0');
    dma1_axi_rvalid : in std_logic := '0';
    dma1_axi_wdata : out std_logic_vector(255 downto 0);
    dma1_axi_wstrb : out std_logic_vector(31 downto 0);
    dma1_axi_wlast : out std_logic;
    dma1_axi_wready : in std_logic := '0';
    dma1_axi_wvalid : out std_logic;
    dma1_axi_bready : out std_logic;
    dma1_axi_bresp : in std_logic_vector(1 downto 0) := (others => '0');
    dma1_axi_bvalid : in std_logic := '0';
    dma2_axi_araddr : out std_logic_vector(38 downto 0);
    dma2_axi_arburst : out std_logic_vector(1 downto 0);
    dma2_axi_arcache : out std_logic_vector(3 downto 0);
    dma2_axi_arlen : out std_logic_vector(7 downto 0);
    dma2_axi_arlock : out std_logic_vector(0 downto 0);
    dma2_axi_arprot : out std_logic_vector(2 downto 0);
    dma2_axi_arqos : out std_logic_vector(3 downto 0);
    dma2_axi_arready : in std_logic := '0';
    dma2_axi_arregion : out std_logic_vector(3 downto 0);
    dma2_axi_arsize : out std_logic_vector(2 downto 0);
    dma2_axi_arvalid : out std_logic;
    dma2_axi_awaddr : out std_logic_vector(38 downto 0);
    dma2_axi_awburst : out std_logic_vector(1 downto 0);
    dma2_axi_awcache : out std_logic_vector(3 downto 0);
    dma2_axi_awlen : out std_logic_vector(7 downto 0);
    dma2_axi_awlock : out std_logic_vector(0 downto 0);
    dma2_axi_awprot : out std_logic_vector(2 downto 0);
    dma2_axi_awqos : out std_logic_vector(3 downto 0);
    dma2_axi_awready : in std_logic := '0';
    dma2_axi_awregion : out std_logic_vector(3 downto 0);
    dma2_axi_awsize : out std_logic_vector(2 downto 0);
    dma2_axi_awvalid : out std_logic;
    dma2_axi_rdata : in std_logic_vector(255 downto 0) := (others => '0');
    dma2_axi_rlast : in std_logic := '0';
    dma2_axi_rready : out std_logic;
    dma2_axi_rresp : in std_logic_vector(1 downto 0) := (others => '0');
    dma2_axi_rvalid : in std_logic := '0';
    dma2_axi_wdata : out std_logic_vector(255 downto 0);
    dma2_axi_wstrb : out std_logic_vector(31 downto 0);
    dma2_axi_wlast : out std_logic;
    dma2_axi_wready : in std_logic := '0';
    dma2_axi_wvalid : out std_logic;
    dma2_axi_bready : out std_logic;
    dma2_axi_bresp : in std_logic_vector(1 downto 0) := (others => '0');
    dma2_axi_bvalid : in std_logic := '0';
    dma3_axi_araddr : out std_logic_vector(38 downto 0);
    dma3_axi_arburst : out std_logic_vector(1 downto 0);
    dma3_axi_arcache : out std_logic_vector(3 downto 0);
    dma3_axi_arlen : out std_logic_vector(7 downto 0);
    dma3_axi_arlock : out std_logic_vector(0 downto 0);
    dma3_axi_arprot : out std_logic_vector(2 downto 0);
    dma3_axi_arqos : out std_logic_vector(3 downto 0);
    dma3_axi_arready : in std_logic := '0';
    dma3_axi_arregion : out std_logic_vector(3 downto 0);
    dma3_axi_arsize : out std_logic_vector(2 downto 0);
    dma3_axi_arvalid : out std_logic;
    dma3_axi_awaddr : out std_logic_vector(38 downto 0);
    dma3_axi_awburst : out std_logic_vector(1 downto 0);
    dma3_axi_awcache : out std_logic_vector(3 downto 0);
    dma3_axi_awlen : out std_logic_vector(7 downto 0);
    dma3_axi_awlock : out std_logic_vector(0 downto 0);
    dma3_axi_awprot : out std_logic_vector(2 downto 0);
    dma3_axi_awqos : out std_logic_vector(3 downto 0);
    dma3_axi_awready : in std_logic := '0';
    dma3_axi_awregion : out std_logic_vector(3 downto 0);
    dma3_axi_awsize : out std_logic_vector(2 downto 0);
    dma3_axi_awvalid : out std_logic;
    dma3_axi_rdata : in std_logic_vector(255 downto 0) := (others => '0');
    dma3_axi_rlast : in std_logic := '0';
    dma3_axi_rready : out std_logic;
    dma3_axi_rresp : in std_logic_vector(1 downto 0) := (others => '0');
    dma3_axi_rvalid : in std_logic := '0';
    dma3_axi_wdata : out std_logic_vector(255 downto 0);
    dma3_axi_wstrb : out std_logic_vector(31 downto 0);
    dma3_axi_wlast : out std_logic;
    dma3_axi_wready : in std_logic := '0';
    dma3_axi_wvalid : out std_logic;
    dma3_axi_bready : out std_logic;
    dma3_axi_bresp : in std_logic_vector(1 downto 0) := (others => '0');
    dma3_axi_bvalid : in std_logic := '0';
    dm_axi_araddr : in std_logic_vector(63 downto 0) := (others => '0');
    dm_axi_arburst : in std_logic_vector(1 downto 0) := (others => '0');
    dm_axi_arcache : in std_logic_vector(3 downto 0) := (others => '0');
    dm_axi_arid : in std_logic_vector(pcie_dm_port_id_width-1 downto 0) := (others => '0');
    dm_axi_arlen : in std_logic_vector(7 downto 0) := (others => '0');
    dm_axi_arlock : in std_logic_vector(0 downto 0) := (others => '0');
    dm_axi_arprot : in std_logic_vector(2 downto 0) := (others => '0');
    dm_axi_arqos : in std_logic_vector(3 downto 0) := (others => '0');
    dm_axi_arready : out std_logic;
    dm_axi_arregion : in std_logic_vector(3 downto 0) := (others => '0');
    dm_axi_arsize : in std_logic_vector(2 downto 0) := (others => '0');
    dm_axi_arvalid : in std_logic := '0';
    dm_axi_awaddr : in std_logic_vector(63 downto 0) := (others => '0');
    dm_axi_awburst : in std_logic_vector(1 downto 0) := (others => '0');
    dm_axi_awcache : in std_logic_vector(3 downto 0) := (others => '0');
    dm_axi_awid : in std_logic_vector(pcie_dm_port_id_width-1 downto 0) := (others => '0');
    dm_axi_awlen : in std_logic_vector(7 downto 0) := (others => '0');
    dm_axi_awlock : in std_logic_vector(0 downto 0) := (others => '0');
    dm_axi_awprot : in std_logic_vector(2 downto 0) := (others => '0');
    dm_axi_awqos : in std_logic_vector(3 downto 0) := (others => '0');
    dm_axi_awready : out std_logic;
    dm_axi_awregion : in std_logic_vector(3 downto 0) := (others => '0');
    dm_axi_awsize : in std_logic_vector(2 downto 0) := (others => '0');
    dm_axi_awvalid : in std_logic := '0';
    dm_axi_rdata : out std_logic_vector(255 downto 0);
    dm_axi_rid : out std_logic_vector(pcie_dm_port_id_width-1 downto 0);
    dm_axi_rlast : out std_logic;
    dm_axi_rready : in std_logic := '0';
    dm_axi_rresp : out std_logic_vector(1 downto 0);
    dm_axi_rvalid : out std_logic;
    dm_axi_wdata : in std_logic_vector(255 downto 0) := (others => '0');
    dm_axi_wlast : in std_logic := '0';
    dm_axi_wready : out std_logic;
    dm_axi_wstrb : in std_logic_vector(31 downto 0) := (others => '0');
    dm_axi_wvalid : in std_logic := '0';
    dm_axi_bid : out std_logic_vector(pcie_dm_port_id_width-1 downto 0);
    dm_axi_bready : in std_logic := '0';
    dm_axi_bresp : out std_logic_vector(1 downto 0);
    dm_axi_bvalid : out std_logic;
    dmb_axi_araddr : in std_logic_vector(12 downto 0) := (others => '0');
    dmb_axi_arready : out std_logic;
    dmb_axi_arvalid : in std_logic := '0';
    dmb_axi_awaddr : in std_logic_vector(12 downto 0) := (others => '0');
    dmb_axi_awready : out std_logic;
    dmb_axi_awvalid : in std_logic := '0';
    dmb_axi_rdata : out std_logic_vector(31 downto 0);
    dmb_axi_rready : in std_logic := '0';
    dmb_axi_rresp : out std_logic_vector(1 downto 0);
    dmb_axi_rvalid : out std_logic;
    dmb_axi_wdata : in std_logic_vector(31 downto 0) := (others => '0');
    dmb_axi_wready : out std_logic;
    dmb_axi_wstrb : in std_logic_vector(3 downto 0) := (others => '0');
    dmb_axi_wvalid : in std_logic := '0';
    dmb_axi_bready : in std_logic := '0';
    dmb_axi_bresp : out std_logic_vector(1 downto 0);
    dmb_axi_bvalid : out std_logic;
    dma0_m_axis_tdata : out std_logic_vector(255 downto 0);
    dma0_m_axis_tready : in std_logic := '0';
    dma0_m_axis_tvalid : out std_logic;
    dma0_s_axis_tdata : in std_logic_vector(255 downto 0) := (others => '0');
    dma0_s_axis_tready : out std_logic;
    dma0_s_axis_tvalid : in std_logic := '0';
    dma1_m_axis_tdata : out std_logic_vector(255 downto 0);
    dma1_m_axis_tready : in std_logic := '0';
    dma1_m_axis_tvalid : out std_logic;
    dma1_s_axis_tdata : in std_logic_vector(255 downto 0) := (others => '0');
    dma1_s_axis_tready : out std_logic;
    dma1_s_axis_tvalid : in std_logic := '0';
    dma2_m_axis_tdata : out std_logic_vector(255 downto 0);
    dma2_m_axis_tready : in std_logic := '0';
    dma2_m_axis_tvalid : out std_logic;
    dma2_s_axis_tdata : in std_logic_vector(255 downto 0) := (others => '0');
    dma2_s_axis_tready : out std_logic;
    dma2_s_axis_tvalid : in std_logic := '0';
    dma3_m_axis_tdata : out std_logic_vector(255 downto 0);
    dma3_m_axis_tready : in std_logic := '0';
    dma3_m_axis_tvalid : out std_logic;
    dma3_s_axis_tdata : in std_logic_vector(255 downto 0) := (others => '0');
    dma3_s_axis_tready : out std_logic;
    dma3_s_axis_tvalid : in std_logic := '0';
    interrupt : in std_logic := '0';
    dma0_abort : in std_logic := '0';
    dma1_abort : in std_logic := '0';
    dma2_abort : in std_logic := '0';
    dma3_abort : in std_logic := '0';
    core_status : out std_logic_vector(63 downto 0)
  );
end entity;


architecture behav of adb3_admxrcku1_bm_x8_axi4 is

  constant handshake_quantum : time := 1 ps;
  -- If true, handshake protocol delays assignments to handshake signals by 'handshake_quantum'.
  constant handshake_use_quantum : boolean := false;
  
  constant transmit_quantum : time := 1 ps;
  -- If true, transmit process waits for 'transmit_quantum' between each item.
  constant transmit_use_quantum : boolean := true;
  
  
  -- Receive a header, which may include a single-bit std_logic value along with signal index
  procedure receive_header(
    signal   pci_exp_rxn : in    std_logic_vector(7 downto 0);
    signal   pci_exp_rxp : in    std_logic_vector(6 downto 0);
    signal   pci_exp_txp : out   std_logic_vector(7 downto 7);
    variable index       : out   natural;
    variable val         : out   std_logic)
  is
    variable index_vec : std_logic_vector(11 downto 0);
  begin
    -- Header, giving index of signal & std_logic value
    wait until pci_exp_rxp(6) = '1';
    val := pci_exp_rxp(5);
    index_vec(11 downto 8) := pci_exp_rxp(3 downto 0);
    index_vec(7 downto 0) := pci_exp_rxn;
    index := to_integer(unsigned(index_vec));
    if handshake_use_quantum then
      pci_exp_txp(7) <= '1' after handshake_quantum;
      wait until pci_exp_rxp(6) = '0';
      pci_exp_txp(7) <= '0' after handshake_quantum;
    else
      pci_exp_txp(7) <= '1';
      wait until pci_exp_rxp(6) = '0';
      pci_exp_txp(7) <= '0';
    end if;
  end;

  -- Receive a std_logic_vector value
  procedure receive(
    signal   pci_exp_rxn : in    std_logic_vector(7 downto 0);
    signal   pci_exp_rxp : in    std_logic_vector(6 downto 0);
    signal   pci_exp_txp : out   std_logic_vector(7 downto 7);
    signal   vec         : out   std_logic_vector)
  is
    constant n : natural := vec'length;
    constant nword : natural := n / 14;
    constant remainder : natural := n mod 14;
    variable tmp : std_logic_vector(13 downto 0);
    variable low : natural;
  begin
    -- Payload, giving a series of 14-bit vectors making up the std_logic_vector value
    low := 0;
    for i in 1 to nword loop
      wait until pci_exp_rxp(6) = '1';
      vec(low + 7 downto low) <= pci_exp_rxn;
      vec(low + 13 downto low + 8) <= pci_exp_rxp(5 downto 0);
      if handshake_use_quantum then
        pci_exp_txp(7) <= '1' after handshake_quantum;
        wait until pci_exp_rxp(6) = '0';
        pci_exp_txp(7) <= '0' after handshake_quantum;
      else
        pci_exp_txp(7) <= '1';
        wait until pci_exp_rxp(6) = '0';
        pci_exp_txp(7) <= '0';
      end if;
      low := low + 14;
    end loop;
    if remainder > 0 then
      -- Remainder of payload, i.e. final slice (less than 14 bits long) of std_logic_vector value
      wait until pci_exp_rxp(6) = '1';
      tmp(7 downto 0) := pci_exp_rxn;
      tmp(13 downto 8) := pci_exp_rxp(5 downto 0);
      vec(low + remainder - 1 downto low) <= tmp(remainder - 1 downto 0);
      if handshake_use_quantum then
        pci_exp_txp(7) <= '1' after handshake_quantum;
        wait until pci_exp_rxp(6) = '0';
        pci_exp_txp(7) <= '0' after handshake_quantum;
      else
        pci_exp_txp(7) <= '1';
        wait until pci_exp_rxp(6) = '0';
        pci_exp_txp(7) <= '0';
      end if;
    end if;
  end;

  -- Transmit a std_logic signal
  procedure transmit(
    signal   pci_exp_txn : out   std_logic_vector(7 downto 0);
    signal   pci_exp_txp : out   std_logic_vector(6 downto 0);
    signal   pci_exp_rxp : in    std_logic_vector(7 downto 7);
    constant index       : in    natural;
    signal   val         : in    std_logic)
  is
    constant index_vec : std_logic_vector(11 downto 0) := std_logic_vector(to_unsigned(index, 12));
  begin
    -- Header, giving index of signal & std_logic value
    pci_exp_txp(5 downto 0) <= val & '0' & index_vec(11 downto 8);
    pci_exp_txn(7 downto 0) <= index_vec(7 downto 0);
    if handshake_use_quantum then
      pci_exp_txp(6) <= '1' after handshake_quantum;
      wait until pci_exp_rxp(7) = '1';
      pci_exp_txp(6) <= '0' after handshake_quantum;
    else
      pci_exp_txp(6) <= '1';
      wait until pci_exp_rxp(7) = '1';
      pci_exp_txp(6) <= '0';
    end if;
    wait until pci_exp_rxp(7) = '0';
  end;

  -- Transmit a std_logic_vector signal, 14 bits at a time
  procedure transmit(
    signal   pci_exp_txn : out   std_logic_vector(7 downto 0);
    signal   pci_exp_txp : out   std_logic_vector(6 downto 0);
    signal   pci_exp_rxp : in    std_logic_vector(7 downto 7);
    constant index       : in    natural;
    signal   vec         : in    std_logic_vector)
  is
    constant index_vec : std_logic_vector(11 downto 0) := std_logic_vector(to_unsigned(index, 12));
    constant n : natural := vec'length;
    constant nword : natural := n / 14;
    constant remainder : natural := n mod 14;
    variable tmp : std_logic_vector(13 downto 0);
    variable low : natural;
  begin
    -- Header, giving index of signal
    pci_exp_txp(5 downto 0) <= "00" & index_vec(11 downto 8);
    pci_exp_txn(7 downto 0) <= index_vec(7 downto 0);
    if handshake_use_quantum then
      pci_exp_txp(6) <= '1' after handshake_quantum;
      wait until pci_exp_rxp(7) = '1';
      pci_exp_txp(6) <= '0' after handshake_quantum;
    else
      pci_exp_txp(6) <= '1';
      wait until pci_exp_rxp(7) = '1';
      pci_exp_txp(6) <= '0';
    end if;
    wait until pci_exp_rxp(7) = '0';
    -- Payload, giving a series of 14-bit vectors making up the std_logic_vector value
    low := 0;
    for i in 1 to nword loop
      pci_exp_txp(5 downto 0) <= vec(low + 13 downto low + 8);
      pci_exp_txn(7 downto 0) <= vec(low + 7 downto low);
      if handshake_use_quantum then
        pci_exp_txp(6) <= '1' after handshake_quantum;
        wait until pci_exp_rxp(7) = '1';
        pci_exp_txp(6) <= '0' after handshake_quantum;
      else
        pci_exp_txp(6) <= '1';
        wait until pci_exp_rxp(7) = '1';
        pci_exp_txp(6) <= '0';
      end if;
      wait until pci_exp_rxp(7) = '0';
      low := low + 14;
    end loop;
    if remainder > 0 then
      -- Remainder of payload, i.e. final slice (less than 14 bits long) of std_logic_vector value
      tmp := std_logic_vector(resize(unsigned(vec(low + remainder - 1 downto low)), 14));
      pci_exp_txp(5 downto 0) <= tmp(13 downto 8);
      pci_exp_txn(7 downto 0) <= tmp(7 downto 0);
      if handshake_use_quantum then
        pci_exp_txp(6) <= '1' after handshake_quantum;
        wait until pci_exp_rxp(7) = '1';
        pci_exp_txp(6) <= '0' after handshake_quantum;
      else
        pci_exp_txp(6) <= '1';
        wait until pci_exp_rxp(7) = '1';
        pci_exp_txp(6) <= '0';
      end if;
      wait until pci_exp_rxp(7) = '0';
    end if;
  end;

  type port_in_enum_t is (
    port_ds_axi_arready,
    port_ds_axi_awready,
    port_ds_axi_rdata,
    port_ds_axi_rlast,
    port_ds_axi_rresp,
    port_ds_axi_rvalid,
    port_ds_axi_wready,
    port_ds_axi_bresp,
    port_ds_axi_bvalid,
    port_dma0_axi_arready,
    port_dma0_axi_awready,
    port_dma0_axi_rdata,
    port_dma0_axi_rlast,
    port_dma0_axi_rresp,
    port_dma0_axi_rvalid,
    port_dma0_axi_wready,
    port_dma0_axi_bresp,
    port_dma0_axi_bvalid,
    port_dma1_axi_arready,
    port_dma1_axi_awready,
    port_dma1_axi_rdata,
    port_dma1_axi_rlast,
    port_dma1_axi_rresp,
    port_dma1_axi_rvalid,
    port_dma1_axi_wready,
    port_dma1_axi_bresp,
    port_dma1_axi_bvalid,
    port_dma2_axi_arready,
    port_dma2_axi_awready,
    port_dma2_axi_rdata,
    port_dma2_axi_rlast,
    port_dma2_axi_rresp,
    port_dma2_axi_rvalid,
    port_dma2_axi_wready,
    port_dma2_axi_bresp,
    port_dma2_axi_bvalid,
    port_dma3_axi_arready,
    port_dma3_axi_awready,
    port_dma3_axi_rdata,
    port_dma3_axi_rlast,
    port_dma3_axi_rresp,
    port_dma3_axi_rvalid,
    port_dma3_axi_wready,
    port_dma3_axi_bresp,
    port_dma3_axi_bvalid,
    port_dm_axi_araddr,
    port_dm_axi_arburst,
    port_dm_axi_arcache,
    port_dm_axi_arid,
    port_dm_axi_arlen,
    port_dm_axi_arlock,
    port_dm_axi_arprot,
    port_dm_axi_arqos,
    port_dm_axi_arregion,
    port_dm_axi_arsize,
    port_dm_axi_arvalid,
    port_dm_axi_awaddr,
    port_dm_axi_awburst,
    port_dm_axi_awcache,
    port_dm_axi_awid,
    port_dm_axi_awlen,
    port_dm_axi_awlock,
    port_dm_axi_awprot,
    port_dm_axi_awqos,
    port_dm_axi_awregion,
    port_dm_axi_awsize,
    port_dm_axi_awvalid,
    port_dm_axi_rready,
    port_dm_axi_wdata,
    port_dm_axi_wlast,
    port_dm_axi_wstrb,
    port_dm_axi_wvalid,
    port_dm_axi_bready,
    port_dmb_axi_araddr,
    port_dmb_axi_arvalid,
    port_dmb_axi_awaddr,
    port_dmb_axi_awvalid,
    port_dmb_axi_rready,
    port_dmb_axi_wdata,
    port_dmb_axi_wstrb,
    port_dmb_axi_wvalid,
    port_dmb_axi_bready,
    port_dma0_m_axis_tready,
    port_dma0_s_axis_tdata,
    port_dma0_s_axis_tvalid,
    port_dma1_m_axis_tready,
    port_dma1_s_axis_tdata,
    port_dma1_s_axis_tvalid,
    port_dma2_m_axis_tready,
    port_dma2_s_axis_tdata,
    port_dma2_s_axis_tvalid,
    port_dma3_m_axis_tready,
    port_dma3_s_axis_tdata,
    port_dma3_s_axis_tvalid,
    port_interrupt,
    port_dma0_abort,
    port_dma1_abort,
    port_dma2_abort,
    port_dma3_abort
  );

  type port_out_enum_t is (
    port_aclk,
    port_aresetn,
    port_ds_axi_araddr,
    port_ds_axi_arburst,
    port_ds_axi_arcache,
    port_ds_axi_arlen,
    port_ds_axi_arlock,
    port_ds_axi_arprot,
    port_ds_axi_arqos,
    port_ds_axi_arregion,
    port_ds_axi_arsize,
    port_ds_axi_arvalid,
    port_ds_axi_awaddr,
    port_ds_axi_awburst,
    port_ds_axi_awcache,
    port_ds_axi_awlen,
    port_ds_axi_awlock,
    port_ds_axi_awprot,
    port_ds_axi_awqos,
    port_ds_axi_awregion,
    port_ds_axi_awsize,
    port_ds_axi_awvalid,
    port_ds_axi_rready,
    port_ds_axi_wdata,
    port_ds_axi_wlast,
    port_ds_axi_wstrb,
    port_ds_axi_wvalid,
    port_ds_axi_bready,
    port_dma0_axi_araddr,
    port_dma0_axi_arburst,
    port_dma0_axi_arcache,
    port_dma0_axi_arlen,
    port_dma0_axi_arlock,
    port_dma0_axi_arprot,
    port_dma0_axi_arqos,
    port_dma0_axi_arregion,
    port_dma0_axi_arsize,
    port_dma0_axi_arvalid,
    port_dma0_axi_awaddr,
    port_dma0_axi_awburst,
    port_dma0_axi_awcache,
    port_dma0_axi_awlen,
    port_dma0_axi_awlock,
    port_dma0_axi_awprot,
    port_dma0_axi_awqos,
    port_dma0_axi_awregion,
    port_dma0_axi_awsize,
    port_dma0_axi_awvalid,
    port_dma0_axi_rready,
    port_dma0_axi_wdata,
    port_dma0_axi_wstrb,
    port_dma0_axi_wlast,
    port_dma0_axi_wvalid,
    port_dma0_axi_bready,
    port_dma1_axi_araddr,
    port_dma1_axi_arburst,
    port_dma1_axi_arcache,
    port_dma1_axi_arlen,
    port_dma1_axi_arlock,
    port_dma1_axi_arprot,
    port_dma1_axi_arqos,
    port_dma1_axi_arregion,
    port_dma1_axi_arsize,
    port_dma1_axi_arvalid,
    port_dma1_axi_awaddr,
    port_dma1_axi_awburst,
    port_dma1_axi_awcache,
    port_dma1_axi_awlen,
    port_dma1_axi_awlock,
    port_dma1_axi_awprot,
    port_dma1_axi_awqos,
    port_dma1_axi_awregion,
    port_dma1_axi_awsize,
    port_dma1_axi_awvalid,
    port_dma1_axi_rready,
    port_dma1_axi_wdata,
    port_dma1_axi_wstrb,
    port_dma1_axi_wlast,
    port_dma1_axi_wvalid,
    port_dma1_axi_bready,
    port_dma2_axi_araddr,
    port_dma2_axi_arburst,
    port_dma2_axi_arcache,
    port_dma2_axi_arlen,
    port_dma2_axi_arlock,
    port_dma2_axi_arprot,
    port_dma2_axi_arqos,
    port_dma2_axi_arregion,
    port_dma2_axi_arsize,
    port_dma2_axi_arvalid,
    port_dma2_axi_awaddr,
    port_dma2_axi_awburst,
    port_dma2_axi_awcache,
    port_dma2_axi_awlen,
    port_dma2_axi_awlock,
    port_dma2_axi_awprot,
    port_dma2_axi_awqos,
    port_dma2_axi_awregion,
    port_dma2_axi_awsize,
    port_dma2_axi_awvalid,
    port_dma2_axi_rready,
    port_dma2_axi_wdata,
    port_dma2_axi_wstrb,
    port_dma2_axi_wlast,
    port_dma2_axi_wvalid,
    port_dma2_axi_bready,
    port_dma3_axi_araddr,
    port_dma3_axi_arburst,
    port_dma3_axi_arcache,
    port_dma3_axi_arlen,
    port_dma3_axi_arlock,
    port_dma3_axi_arprot,
    port_dma3_axi_arqos,
    port_dma3_axi_arregion,
    port_dma3_axi_arsize,
    port_dma3_axi_arvalid,
    port_dma3_axi_awaddr,
    port_dma3_axi_awburst,
    port_dma3_axi_awcache,
    port_dma3_axi_awlen,
    port_dma3_axi_awlock,
    port_dma3_axi_awprot,
    port_dma3_axi_awqos,
    port_dma3_axi_awregion,
    port_dma3_axi_awsize,
    port_dma3_axi_awvalid,
    port_dma3_axi_rready,
    port_dma3_axi_wdata,
    port_dma3_axi_wstrb,
    port_dma3_axi_wlast,
    port_dma3_axi_wvalid,
    port_dma3_axi_bready,
    port_dm_axi_arready,
    port_dm_axi_awready,
    port_dm_axi_rdata,
    port_dm_axi_rid,
    port_dm_axi_rlast,
    port_dm_axi_rresp,
    port_dm_axi_rvalid,
    port_dm_axi_wready,
    port_dm_axi_bid,
    port_dm_axi_bresp,
    port_dm_axi_bvalid,
    port_dmb_axi_arready,
    port_dmb_axi_awready,
    port_dmb_axi_rdata,
    port_dmb_axi_rresp,
    port_dmb_axi_rvalid,
    port_dmb_axi_wready,
    port_dmb_axi_bresp,
    port_dmb_axi_bvalid,
    port_dma0_m_axis_tdata,
    port_dma0_m_axis_tvalid,
    port_dma0_s_axis_tready,
    port_dma1_m_axis_tdata,
    port_dma1_m_axis_tvalid,
    port_dma1_s_axis_tready,
    port_dma2_m_axis_tdata,
    port_dma2_m_axis_tvalid,
    port_dma2_s_axis_tready,
    port_dma3_m_axis_tdata,
    port_dma3_m_axis_tvalid,
    port_dma3_s_axis_tready,
    port_core_status
  );

  constant num_port_in : natural := port_in_enum_t'pos(port_in_enum_t'high) + 1;
  type boolean_array_t is array(natural range <>) of boolean;
  signal pending_in : boolean_array_t(0 to num_port_in - 1) := (others => false);
  signal ack_in : boolean_array_t(0 to num_port_in - 1) := (others => false);

  constant inertial_delay : time := 1 ns;
  signal i_ds_axi_arready : std_logic;
  signal i_ds_axi_awready : std_logic;
  signal i_ds_axi_rdata : std_logic_vector(255 downto 0);
  signal i_ds_axi_rlast : std_logic;
  signal i_ds_axi_rresp : std_logic_vector(1 downto 0);
  signal i_ds_axi_rvalid : std_logic;
  signal i_ds_axi_wready : std_logic;
  signal i_ds_axi_bresp : std_logic_vector(1 downto 0);
  signal i_ds_axi_bvalid : std_logic;
  signal i_dma0_axi_arready : std_logic;
  signal i_dma0_axi_awready : std_logic;
  signal i_dma0_axi_rdata : std_logic_vector(255 downto 0);
  signal i_dma0_axi_rlast : std_logic;
  signal i_dma0_axi_rresp : std_logic_vector(1 downto 0);
  signal i_dma0_axi_rvalid : std_logic;
  signal i_dma0_axi_wready : std_logic;
  signal i_dma0_axi_bresp : std_logic_vector(1 downto 0);
  signal i_dma0_axi_bvalid : std_logic;
  signal i_dma1_axi_arready : std_logic;
  signal i_dma1_axi_awready : std_logic;
  signal i_dma1_axi_rdata : std_logic_vector(255 downto 0);
  signal i_dma1_axi_rlast : std_logic;
  signal i_dma1_axi_rresp : std_logic_vector(1 downto 0);
  signal i_dma1_axi_rvalid : std_logic;
  signal i_dma1_axi_wready : std_logic;
  signal i_dma1_axi_bresp : std_logic_vector(1 downto 0);
  signal i_dma1_axi_bvalid : std_logic;
  signal i_dma2_axi_arready : std_logic;
  signal i_dma2_axi_awready : std_logic;
  signal i_dma2_axi_rdata : std_logic_vector(255 downto 0);
  signal i_dma2_axi_rlast : std_logic;
  signal i_dma2_axi_rresp : std_logic_vector(1 downto 0);
  signal i_dma2_axi_rvalid : std_logic;
  signal i_dma2_axi_wready : std_logic;
  signal i_dma2_axi_bresp : std_logic_vector(1 downto 0);
  signal i_dma2_axi_bvalid : std_logic;
  signal i_dma3_axi_arready : std_logic;
  signal i_dma3_axi_awready : std_logic;
  signal i_dma3_axi_rdata : std_logic_vector(255 downto 0);
  signal i_dma3_axi_rlast : std_logic;
  signal i_dma3_axi_rresp : std_logic_vector(1 downto 0);
  signal i_dma3_axi_rvalid : std_logic;
  signal i_dma3_axi_wready : std_logic;
  signal i_dma3_axi_bresp : std_logic_vector(1 downto 0);
  signal i_dma3_axi_bvalid : std_logic;
  signal i_dm_axi_araddr : std_logic_vector(63 downto 0);
  signal i_dm_axi_arburst : std_logic_vector(1 downto 0);
  signal i_dm_axi_arcache : std_logic_vector(3 downto 0);
  signal i_dm_axi_arid : std_logic_vector(pcie_dm_port_id_width-1 downto 0);
  signal i_dm_axi_arlen : std_logic_vector(7 downto 0);
  signal i_dm_axi_arlock : std_logic_vector(0 downto 0);
  signal i_dm_axi_arprot : std_logic_vector(2 downto 0);
  signal i_dm_axi_arqos : std_logic_vector(3 downto 0);
  signal i_dm_axi_arregion : std_logic_vector(3 downto 0);
  signal i_dm_axi_arsize : std_logic_vector(2 downto 0);
  signal i_dm_axi_arvalid : std_logic;
  signal i_dm_axi_awaddr : std_logic_vector(63 downto 0);
  signal i_dm_axi_awburst : std_logic_vector(1 downto 0);
  signal i_dm_axi_awcache : std_logic_vector(3 downto 0);
  signal i_dm_axi_awid : std_logic_vector(pcie_dm_port_id_width-1 downto 0);
  signal i_dm_axi_awlen : std_logic_vector(7 downto 0);
  signal i_dm_axi_awlock : std_logic_vector(0 downto 0);
  signal i_dm_axi_awprot : std_logic_vector(2 downto 0);
  signal i_dm_axi_awqos : std_logic_vector(3 downto 0);
  signal i_dm_axi_awregion : std_logic_vector(3 downto 0);
  signal i_dm_axi_awsize : std_logic_vector(2 downto 0);
  signal i_dm_axi_awvalid : std_logic;
  signal i_dm_axi_rready : std_logic;
  signal i_dm_axi_wdata : std_logic_vector(255 downto 0);
  signal i_dm_axi_wlast : std_logic;
  signal i_dm_axi_wstrb : std_logic_vector(31 downto 0);
  signal i_dm_axi_wvalid : std_logic;
  signal i_dm_axi_bready : std_logic;
  signal i_dmb_axi_araddr : std_logic_vector(12 downto 0);
  signal i_dmb_axi_arvalid : std_logic;
  signal i_dmb_axi_awaddr : std_logic_vector(12 downto 0);
  signal i_dmb_axi_awvalid : std_logic;
  signal i_dmb_axi_rready : std_logic;
  signal i_dmb_axi_wdata : std_logic_vector(31 downto 0);
  signal i_dmb_axi_wstrb : std_logic_vector(3 downto 0);
  signal i_dmb_axi_wvalid : std_logic;
  signal i_dmb_axi_bready : std_logic;
  signal i_dma0_m_axis_tready : std_logic;
  signal i_dma0_s_axis_tdata : std_logic_vector(255 downto 0);
  signal i_dma0_s_axis_tvalid : std_logic;
  signal i_dma1_m_axis_tready : std_logic;
  signal i_dma1_s_axis_tdata : std_logic_vector(255 downto 0);
  signal i_dma1_s_axis_tvalid : std_logic;
  signal i_dma2_m_axis_tready : std_logic;
  signal i_dma2_s_axis_tdata : std_logic_vector(255 downto 0);
  signal i_dma2_s_axis_tvalid : std_logic;
  signal i_dma3_m_axis_tready : std_logic;
  signal i_dma3_s_axis_tdata : std_logic_vector(255 downto 0);
  signal i_dma3_s_axis_tvalid : std_logic;
  signal i_interrupt : std_logic;
  signal i_dma0_abort : std_logic;
  signal i_dma1_abort : std_logic;
  signal i_dma2_abort : std_logic;
  signal i_dma3_abort : std_logic;


begin

  model_inout_o <= (others => '-');
  model_inout_t <= (others => '1');


  refclk <= refclk200_p when use_diff_ref_clock else refclk200_in;


  vpd_out <= (others => '0');

  
  i_ds_axi_arready <= ds_axi_arready after inertial_delay;
  i_ds_axi_awready <= ds_axi_awready after inertial_delay;
  i_ds_axi_rdata <= ds_axi_rdata after inertial_delay;
  i_ds_axi_rlast <= ds_axi_rlast after inertial_delay;
  i_ds_axi_rresp <= ds_axi_rresp after inertial_delay;
  i_ds_axi_rvalid <= ds_axi_rvalid after inertial_delay;
  i_ds_axi_wready <= ds_axi_wready after inertial_delay;
  i_ds_axi_bresp <= ds_axi_bresp after inertial_delay;
  i_ds_axi_bvalid <= ds_axi_bvalid after inertial_delay;
  i_dma0_axi_arready <= dma0_axi_arready after inertial_delay;
  i_dma0_axi_awready <= dma0_axi_awready after inertial_delay;
  i_dma0_axi_rdata <= dma0_axi_rdata after inertial_delay;
  i_dma0_axi_rlast <= dma0_axi_rlast after inertial_delay;
  i_dma0_axi_rresp <= dma0_axi_rresp after inertial_delay;
  i_dma0_axi_rvalid <= dma0_axi_rvalid after inertial_delay;
  i_dma0_axi_wready <= dma0_axi_wready after inertial_delay;
  i_dma0_axi_bresp <= dma0_axi_bresp after inertial_delay;
  i_dma0_axi_bvalid <= dma0_axi_bvalid after inertial_delay;
  i_dma1_axi_arready <= dma1_axi_arready after inertial_delay;
  i_dma1_axi_awready <= dma1_axi_awready after inertial_delay;
  i_dma1_axi_rdata <= dma1_axi_rdata after inertial_delay;
  i_dma1_axi_rlast <= dma1_axi_rlast after inertial_delay;
  i_dma1_axi_rresp <= dma1_axi_rresp after inertial_delay;
  i_dma1_axi_rvalid <= dma1_axi_rvalid after inertial_delay;
  i_dma1_axi_wready <= dma1_axi_wready after inertial_delay;
  i_dma1_axi_bresp <= dma1_axi_bresp after inertial_delay;
  i_dma1_axi_bvalid <= dma1_axi_bvalid after inertial_delay;
  i_dma2_axi_arready <= dma2_axi_arready after inertial_delay;
  i_dma2_axi_awready <= dma2_axi_awready after inertial_delay;
  i_dma2_axi_rdata <= dma2_axi_rdata after inertial_delay;
  i_dma2_axi_rlast <= dma2_axi_rlast after inertial_delay;
  i_dma2_axi_rresp <= dma2_axi_rresp after inertial_delay;
  i_dma2_axi_rvalid <= dma2_axi_rvalid after inertial_delay;
  i_dma2_axi_wready <= dma2_axi_wready after inertial_delay;
  i_dma2_axi_bresp <= dma2_axi_bresp after inertial_delay;
  i_dma2_axi_bvalid <= dma2_axi_bvalid after inertial_delay;
  i_dma3_axi_arready <= dma3_axi_arready after inertial_delay;
  i_dma3_axi_awready <= dma3_axi_awready after inertial_delay;
  i_dma3_axi_rdata <= dma3_axi_rdata after inertial_delay;
  i_dma3_axi_rlast <= dma3_axi_rlast after inertial_delay;
  i_dma3_axi_rresp <= dma3_axi_rresp after inertial_delay;
  i_dma3_axi_rvalid <= dma3_axi_rvalid after inertial_delay;
  i_dma3_axi_wready <= dma3_axi_wready after inertial_delay;
  i_dma3_axi_bresp <= dma3_axi_bresp after inertial_delay;
  i_dma3_axi_bvalid <= dma3_axi_bvalid after inertial_delay;
  i_dm_axi_araddr <= dm_axi_araddr after inertial_delay;
  i_dm_axi_arburst <= dm_axi_arburst after inertial_delay;
  i_dm_axi_arcache <= dm_axi_arcache after inertial_delay;
  i_dm_axi_arid <= dm_axi_arid after inertial_delay;
  i_dm_axi_arlen <= dm_axi_arlen after inertial_delay;
  i_dm_axi_arlock <= dm_axi_arlock after inertial_delay;
  i_dm_axi_arprot <= dm_axi_arprot after inertial_delay;
  i_dm_axi_arqos <= dm_axi_arqos after inertial_delay;
  i_dm_axi_arregion <= dm_axi_arregion after inertial_delay;
  i_dm_axi_arsize <= dm_axi_arsize after inertial_delay;
  i_dm_axi_arvalid <= dm_axi_arvalid after inertial_delay;
  i_dm_axi_awaddr <= dm_axi_awaddr after inertial_delay;
  i_dm_axi_awburst <= dm_axi_awburst after inertial_delay;
  i_dm_axi_awcache <= dm_axi_awcache after inertial_delay;
  i_dm_axi_awid <= dm_axi_awid after inertial_delay;
  i_dm_axi_awlen <= dm_axi_awlen after inertial_delay;
  i_dm_axi_awlock <= dm_axi_awlock after inertial_delay;
  i_dm_axi_awprot <= dm_axi_awprot after inertial_delay;
  i_dm_axi_awqos <= dm_axi_awqos after inertial_delay;
  i_dm_axi_awregion <= dm_axi_awregion after inertial_delay;
  i_dm_axi_awsize <= dm_axi_awsize after inertial_delay;
  i_dm_axi_awvalid <= dm_axi_awvalid after inertial_delay;
  i_dm_axi_rready <= dm_axi_rready after inertial_delay;
  i_dm_axi_wdata <= dm_axi_wdata after inertial_delay;
  i_dm_axi_wlast <= dm_axi_wlast after inertial_delay;
  i_dm_axi_wstrb <= dm_axi_wstrb after inertial_delay;
  i_dm_axi_wvalid <= dm_axi_wvalid after inertial_delay;
  i_dm_axi_bready <= dm_axi_bready after inertial_delay;
  i_dmb_axi_araddr <= dmb_axi_araddr after inertial_delay;
  i_dmb_axi_arvalid <= dmb_axi_arvalid after inertial_delay;
  i_dmb_axi_awaddr <= dmb_axi_awaddr after inertial_delay;
  i_dmb_axi_awvalid <= dmb_axi_awvalid after inertial_delay;
  i_dmb_axi_rready <= dmb_axi_rready after inertial_delay;
  i_dmb_axi_wdata <= dmb_axi_wdata after inertial_delay;
  i_dmb_axi_wstrb <= dmb_axi_wstrb after inertial_delay;
  i_dmb_axi_wvalid <= dmb_axi_wvalid after inertial_delay;
  i_dmb_axi_bready <= dmb_axi_bready after inertial_delay;
  i_dma0_m_axis_tready <= dma0_m_axis_tready after inertial_delay;
  i_dma0_s_axis_tdata <= dma0_s_axis_tdata after inertial_delay;
  i_dma0_s_axis_tvalid <= dma0_s_axis_tvalid after inertial_delay;
  i_dma1_m_axis_tready <= dma1_m_axis_tready after inertial_delay;
  i_dma1_s_axis_tdata <= dma1_s_axis_tdata after inertial_delay;
  i_dma1_s_axis_tvalid <= dma1_s_axis_tvalid after inertial_delay;
  i_dma2_m_axis_tready <= dma2_m_axis_tready after inertial_delay;
  i_dma2_s_axis_tdata <= dma2_s_axis_tdata after inertial_delay;
  i_dma2_s_axis_tvalid <= dma2_s_axis_tvalid after inertial_delay;
  i_dma3_m_axis_tready <= dma3_m_axis_tready after inertial_delay;
  i_dma3_s_axis_tdata <= dma3_s_axis_tdata after inertial_delay;
  i_dma3_s_axis_tvalid <= dma3_s_axis_tvalid after inertial_delay;
  i_interrupt <= interrupt after inertial_delay;
  i_dma0_abort <= dma0_abort after inertial_delay;
  i_dma1_abort <= dma1_abort after inertial_delay;
  i_dma2_abort <= dma2_abort after inertial_delay;
  i_dma3_abort <= dma3_abort after inertial_delay;


  propagate_ds_axi_arready : process
    constant index : natural := port_in_enum_t'pos(port_ds_axi_arready);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_ds_axi_arready;
    end loop;
  end process;

  propagate_ds_axi_awready : process
    constant index : natural := port_in_enum_t'pos(port_ds_axi_awready);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_ds_axi_awready;
    end loop;
  end process;

  propagate_ds_axi_rdata : process
    constant index : natural := port_in_enum_t'pos(port_ds_axi_rdata);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_ds_axi_rdata;
    end loop;
  end process;

  propagate_ds_axi_rlast : process
    constant index : natural := port_in_enum_t'pos(port_ds_axi_rlast);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_ds_axi_rlast;
    end loop;
  end process;

  propagate_ds_axi_rresp : process
    constant index : natural := port_in_enum_t'pos(port_ds_axi_rresp);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_ds_axi_rresp;
    end loop;
  end process;

  propagate_ds_axi_rvalid : process
    constant index : natural := port_in_enum_t'pos(port_ds_axi_rvalid);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_ds_axi_rvalid;
    end loop;
  end process;

  propagate_ds_axi_wready : process
    constant index : natural := port_in_enum_t'pos(port_ds_axi_wready);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_ds_axi_wready;
    end loop;
  end process;

  propagate_ds_axi_bresp : process
    constant index : natural := port_in_enum_t'pos(port_ds_axi_bresp);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_ds_axi_bresp;
    end loop;
  end process;

  propagate_ds_axi_bvalid : process
    constant index : natural := port_in_enum_t'pos(port_ds_axi_bvalid);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_ds_axi_bvalid;
    end loop;
  end process;

  propagate_dma0_axi_arready : process
    constant index : natural := port_in_enum_t'pos(port_dma0_axi_arready);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dma0_axi_arready;
    end loop;
  end process;

  propagate_dma0_axi_awready : process
    constant index : natural := port_in_enum_t'pos(port_dma0_axi_awready);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dma0_axi_awready;
    end loop;
  end process;

  propagate_dma0_axi_rdata : process
    constant index : natural := port_in_enum_t'pos(port_dma0_axi_rdata);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dma0_axi_rdata;
    end loop;
  end process;

  propagate_dma0_axi_rlast : process
    constant index : natural := port_in_enum_t'pos(port_dma0_axi_rlast);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dma0_axi_rlast;
    end loop;
  end process;

  propagate_dma0_axi_rresp : process
    constant index : natural := port_in_enum_t'pos(port_dma0_axi_rresp);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dma0_axi_rresp;
    end loop;
  end process;

  propagate_dma0_axi_rvalid : process
    constant index : natural := port_in_enum_t'pos(port_dma0_axi_rvalid);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dma0_axi_rvalid;
    end loop;
  end process;

  propagate_dma0_axi_wready : process
    constant index : natural := port_in_enum_t'pos(port_dma0_axi_wready);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dma0_axi_wready;
    end loop;
  end process;

  propagate_dma0_axi_bresp : process
    constant index : natural := port_in_enum_t'pos(port_dma0_axi_bresp);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dma0_axi_bresp;
    end loop;
  end process;

  propagate_dma0_axi_bvalid : process
    constant index : natural := port_in_enum_t'pos(port_dma0_axi_bvalid);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dma0_axi_bvalid;
    end loop;
  end process;

  propagate_dma1_axi_arready : process
    constant index : natural := port_in_enum_t'pos(port_dma1_axi_arready);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dma1_axi_arready;
    end loop;
  end process;

  propagate_dma1_axi_awready : process
    constant index : natural := port_in_enum_t'pos(port_dma1_axi_awready);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dma1_axi_awready;
    end loop;
  end process;

  propagate_dma1_axi_rdata : process
    constant index : natural := port_in_enum_t'pos(port_dma1_axi_rdata);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dma1_axi_rdata;
    end loop;
  end process;

  propagate_dma1_axi_rlast : process
    constant index : natural := port_in_enum_t'pos(port_dma1_axi_rlast);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dma1_axi_rlast;
    end loop;
  end process;

  propagate_dma1_axi_rresp : process
    constant index : natural := port_in_enum_t'pos(port_dma1_axi_rresp);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dma1_axi_rresp;
    end loop;
  end process;

  propagate_dma1_axi_rvalid : process
    constant index : natural := port_in_enum_t'pos(port_dma1_axi_rvalid);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dma1_axi_rvalid;
    end loop;
  end process;

  propagate_dma1_axi_wready : process
    constant index : natural := port_in_enum_t'pos(port_dma1_axi_wready);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dma1_axi_wready;
    end loop;
  end process;

  propagate_dma1_axi_bresp : process
    constant index : natural := port_in_enum_t'pos(port_dma1_axi_bresp);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dma1_axi_bresp;
    end loop;
  end process;

  propagate_dma1_axi_bvalid : process
    constant index : natural := port_in_enum_t'pos(port_dma1_axi_bvalid);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dma1_axi_bvalid;
    end loop;
  end process;

  propagate_dma2_axi_arready : process
    constant index : natural := port_in_enum_t'pos(port_dma2_axi_arready);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dma2_axi_arready;
    end loop;
  end process;

  propagate_dma2_axi_awready : process
    constant index : natural := port_in_enum_t'pos(port_dma2_axi_awready);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dma2_axi_awready;
    end loop;
  end process;

  propagate_dma2_axi_rdata : process
    constant index : natural := port_in_enum_t'pos(port_dma2_axi_rdata);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dma2_axi_rdata;
    end loop;
  end process;

  propagate_dma2_axi_rlast : process
    constant index : natural := port_in_enum_t'pos(port_dma2_axi_rlast);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dma2_axi_rlast;
    end loop;
  end process;

  propagate_dma2_axi_rresp : process
    constant index : natural := port_in_enum_t'pos(port_dma2_axi_rresp);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dma2_axi_rresp;
    end loop;
  end process;

  propagate_dma2_axi_rvalid : process
    constant index : natural := port_in_enum_t'pos(port_dma2_axi_rvalid);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dma2_axi_rvalid;
    end loop;
  end process;

  propagate_dma2_axi_wready : process
    constant index : natural := port_in_enum_t'pos(port_dma2_axi_wready);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dma2_axi_wready;
    end loop;
  end process;

  propagate_dma2_axi_bresp : process
    constant index : natural := port_in_enum_t'pos(port_dma2_axi_bresp);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dma2_axi_bresp;
    end loop;
  end process;

  propagate_dma2_axi_bvalid : process
    constant index : natural := port_in_enum_t'pos(port_dma2_axi_bvalid);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dma2_axi_bvalid;
    end loop;
  end process;

  propagate_dma3_axi_arready : process
    constant index : natural := port_in_enum_t'pos(port_dma3_axi_arready);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dma3_axi_arready;
    end loop;
  end process;

  propagate_dma3_axi_awready : process
    constant index : natural := port_in_enum_t'pos(port_dma3_axi_awready);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dma3_axi_awready;
    end loop;
  end process;

  propagate_dma3_axi_rdata : process
    constant index : natural := port_in_enum_t'pos(port_dma3_axi_rdata);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dma3_axi_rdata;
    end loop;
  end process;

  propagate_dma3_axi_rlast : process
    constant index : natural := port_in_enum_t'pos(port_dma3_axi_rlast);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dma3_axi_rlast;
    end loop;
  end process;

  propagate_dma3_axi_rresp : process
    constant index : natural := port_in_enum_t'pos(port_dma3_axi_rresp);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dma3_axi_rresp;
    end loop;
  end process;

  propagate_dma3_axi_rvalid : process
    constant index : natural := port_in_enum_t'pos(port_dma3_axi_rvalid);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dma3_axi_rvalid;
    end loop;
  end process;

  propagate_dma3_axi_wready : process
    constant index : natural := port_in_enum_t'pos(port_dma3_axi_wready);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dma3_axi_wready;
    end loop;
  end process;

  propagate_dma3_axi_bresp : process
    constant index : natural := port_in_enum_t'pos(port_dma3_axi_bresp);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dma3_axi_bresp;
    end loop;
  end process;

  propagate_dma3_axi_bvalid : process
    constant index : natural := port_in_enum_t'pos(port_dma3_axi_bvalid);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dma3_axi_bvalid;
    end loop;
  end process;

  propagate_dm_axi_araddr : process
    constant index : natural := port_in_enum_t'pos(port_dm_axi_araddr);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dm_axi_araddr;
    end loop;
  end process;

  propagate_dm_axi_arburst : process
    constant index : natural := port_in_enum_t'pos(port_dm_axi_arburst);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dm_axi_arburst;
    end loop;
  end process;

  propagate_dm_axi_arcache : process
    constant index : natural := port_in_enum_t'pos(port_dm_axi_arcache);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dm_axi_arcache;
    end loop;
  end process;

  propagate_dm_axi_arid : process
    constant index : natural := port_in_enum_t'pos(port_dm_axi_arid);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dm_axi_arid;
    end loop;
  end process;

  propagate_dm_axi_arlen : process
    constant index : natural := port_in_enum_t'pos(port_dm_axi_arlen);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dm_axi_arlen;
    end loop;
  end process;

  propagate_dm_axi_arlock : process
    constant index : natural := port_in_enum_t'pos(port_dm_axi_arlock);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dm_axi_arlock;
    end loop;
  end process;

  propagate_dm_axi_arprot : process
    constant index : natural := port_in_enum_t'pos(port_dm_axi_arprot);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dm_axi_arprot;
    end loop;
  end process;

  propagate_dm_axi_arqos : process
    constant index : natural := port_in_enum_t'pos(port_dm_axi_arqos);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dm_axi_arqos;
    end loop;
  end process;

  propagate_dm_axi_arregion : process
    constant index : natural := port_in_enum_t'pos(port_dm_axi_arregion);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dm_axi_arregion;
    end loop;
  end process;

  propagate_dm_axi_arsize : process
    constant index : natural := port_in_enum_t'pos(port_dm_axi_arsize);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dm_axi_arsize;
    end loop;
  end process;

  propagate_dm_axi_arvalid : process
    constant index : natural := port_in_enum_t'pos(port_dm_axi_arvalid);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dm_axi_arvalid;
    end loop;
  end process;

  propagate_dm_axi_awaddr : process
    constant index : natural := port_in_enum_t'pos(port_dm_axi_awaddr);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dm_axi_awaddr;
    end loop;
  end process;

  propagate_dm_axi_awburst : process
    constant index : natural := port_in_enum_t'pos(port_dm_axi_awburst);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dm_axi_awburst;
    end loop;
  end process;

  propagate_dm_axi_awcache : process
    constant index : natural := port_in_enum_t'pos(port_dm_axi_awcache);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dm_axi_awcache;
    end loop;
  end process;

  propagate_dm_axi_awid : process
    constant index : natural := port_in_enum_t'pos(port_dm_axi_awid);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dm_axi_awid;
    end loop;
  end process;

  propagate_dm_axi_awlen : process
    constant index : natural := port_in_enum_t'pos(port_dm_axi_awlen);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dm_axi_awlen;
    end loop;
  end process;

  propagate_dm_axi_awlock : process
    constant index : natural := port_in_enum_t'pos(port_dm_axi_awlock);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dm_axi_awlock;
    end loop;
  end process;

  propagate_dm_axi_awprot : process
    constant index : natural := port_in_enum_t'pos(port_dm_axi_awprot);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dm_axi_awprot;
    end loop;
  end process;

  propagate_dm_axi_awqos : process
    constant index : natural := port_in_enum_t'pos(port_dm_axi_awqos);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dm_axi_awqos;
    end loop;
  end process;

  propagate_dm_axi_awregion : process
    constant index : natural := port_in_enum_t'pos(port_dm_axi_awregion);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dm_axi_awregion;
    end loop;
  end process;

  propagate_dm_axi_awsize : process
    constant index : natural := port_in_enum_t'pos(port_dm_axi_awsize);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dm_axi_awsize;
    end loop;
  end process;

  propagate_dm_axi_awvalid : process
    constant index : natural := port_in_enum_t'pos(port_dm_axi_awvalid);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dm_axi_awvalid;
    end loop;
  end process;

  propagate_dm_axi_rready : process
    constant index : natural := port_in_enum_t'pos(port_dm_axi_rready);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dm_axi_rready;
    end loop;
  end process;

  propagate_dm_axi_wdata : process
    constant index : natural := port_in_enum_t'pos(port_dm_axi_wdata);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dm_axi_wdata;
    end loop;
  end process;

  propagate_dm_axi_wlast : process
    constant index : natural := port_in_enum_t'pos(port_dm_axi_wlast);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dm_axi_wlast;
    end loop;
  end process;

  propagate_dm_axi_wstrb : process
    constant index : natural := port_in_enum_t'pos(port_dm_axi_wstrb);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dm_axi_wstrb;
    end loop;
  end process;

  propagate_dm_axi_wvalid : process
    constant index : natural := port_in_enum_t'pos(port_dm_axi_wvalid);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dm_axi_wvalid;
    end loop;
  end process;

  propagate_dm_axi_bready : process
    constant index : natural := port_in_enum_t'pos(port_dm_axi_bready);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dm_axi_bready;
    end loop;
  end process;

  propagate_dmb_axi_araddr : process
    constant index : natural := port_in_enum_t'pos(port_dmb_axi_araddr);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dmb_axi_araddr;
    end loop;
  end process;

  propagate_dmb_axi_arvalid : process
    constant index : natural := port_in_enum_t'pos(port_dmb_axi_arvalid);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dmb_axi_arvalid;
    end loop;
  end process;

  propagate_dmb_axi_awaddr : process
    constant index : natural := port_in_enum_t'pos(port_dmb_axi_awaddr);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dmb_axi_awaddr;
    end loop;
  end process;

  propagate_dmb_axi_awvalid : process
    constant index : natural := port_in_enum_t'pos(port_dmb_axi_awvalid);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dmb_axi_awvalid;
    end loop;
  end process;

  propagate_dmb_axi_rready : process
    constant index : natural := port_in_enum_t'pos(port_dmb_axi_rready);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dmb_axi_rready;
    end loop;
  end process;

  propagate_dmb_axi_wdata : process
    constant index : natural := port_in_enum_t'pos(port_dmb_axi_wdata);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dmb_axi_wdata;
    end loop;
  end process;

  propagate_dmb_axi_wstrb : process
    constant index : natural := port_in_enum_t'pos(port_dmb_axi_wstrb);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dmb_axi_wstrb;
    end loop;
  end process;

  propagate_dmb_axi_wvalid : process
    constant index : natural := port_in_enum_t'pos(port_dmb_axi_wvalid);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dmb_axi_wvalid;
    end loop;
  end process;

  propagate_dmb_axi_bready : process
    constant index : natural := port_in_enum_t'pos(port_dmb_axi_bready);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dmb_axi_bready;
    end loop;
  end process;

  propagate_dma0_m_axis_tready : process
    constant index : natural := port_in_enum_t'pos(port_dma0_m_axis_tready);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dma0_m_axis_tready;
    end loop;
  end process;

  propagate_dma0_s_axis_tdata : process
    constant index : natural := port_in_enum_t'pos(port_dma0_s_axis_tdata);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dma0_s_axis_tdata;
    end loop;
  end process;

  propagate_dma0_s_axis_tvalid : process
    constant index : natural := port_in_enum_t'pos(port_dma0_s_axis_tvalid);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dma0_s_axis_tvalid;
    end loop;
  end process;

  propagate_dma1_m_axis_tready : process
    constant index : natural := port_in_enum_t'pos(port_dma1_m_axis_tready);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dma1_m_axis_tready;
    end loop;
  end process;

  propagate_dma1_s_axis_tdata : process
    constant index : natural := port_in_enum_t'pos(port_dma1_s_axis_tdata);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dma1_s_axis_tdata;
    end loop;
  end process;

  propagate_dma1_s_axis_tvalid : process
    constant index : natural := port_in_enum_t'pos(port_dma1_s_axis_tvalid);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dma1_s_axis_tvalid;
    end loop;
  end process;

  propagate_dma2_m_axis_tready : process
    constant index : natural := port_in_enum_t'pos(port_dma2_m_axis_tready);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dma2_m_axis_tready;
    end loop;
  end process;

  propagate_dma2_s_axis_tdata : process
    constant index : natural := port_in_enum_t'pos(port_dma2_s_axis_tdata);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dma2_s_axis_tdata;
    end loop;
  end process;

  propagate_dma2_s_axis_tvalid : process
    constant index : natural := port_in_enum_t'pos(port_dma2_s_axis_tvalid);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dma2_s_axis_tvalid;
    end loop;
  end process;

  propagate_dma3_m_axis_tready : process
    constant index : natural := port_in_enum_t'pos(port_dma3_m_axis_tready);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dma3_m_axis_tready;
    end loop;
  end process;

  propagate_dma3_s_axis_tdata : process
    constant index : natural := port_in_enum_t'pos(port_dma3_s_axis_tdata);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dma3_s_axis_tdata;
    end loop;
  end process;

  propagate_dma3_s_axis_tvalid : process
    constant index : natural := port_in_enum_t'pos(port_dma3_s_axis_tvalid);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dma3_s_axis_tvalid;
    end loop;
  end process;

  propagate_interrupt : process
    constant index : natural := port_in_enum_t'pos(port_interrupt);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_interrupt;
    end loop;
  end process;

  propagate_dma0_abort : process
    constant index : natural := port_in_enum_t'pos(port_dma0_abort);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dma0_abort;
    end loop;
  end process;

  propagate_dma1_abort : process
    constant index : natural := port_in_enum_t'pos(port_dma1_abort);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dma1_abort;
    end loop;
  end process;

  propagate_dma2_abort : process
    constant index : natural := port_in_enum_t'pos(port_dma2_abort);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dma2_abort;
    end loop;
  end process;

  propagate_dma3_abort : process
    constant index : natural := port_in_enum_t'pos(port_dma3_abort);
  begin
    loop
      pending_in(index) <= true;
      wait until ack_in(index);
      pending_in(index) <= false;
      wait until not ack_in(index);
      wait on i_dma3_abort;
    end loop;
  end process;


  transmit_signals : process
    variable port_in : port_in_enum_t;
    variable work_done : boolean;
  begin
    loop
      loop
        work_done := false;
        for index in pending_in'range loop
          if pending_in(index) then
            work_done := true;

            -- Send current value of port to other side of link
            port_in := port_in_enum_t'val(index);
            case port_in is
              when port_ds_axi_arready =>
                transmit(pci_exp_txn(7 downto 0), pci_exp_txp(6 downto 0), pci_exp_rxp(7 downto 7), index, i_ds_axi_arready);

              when port_ds_axi_awready =>
                transmit(pci_exp_txn(7 downto 0), pci_exp_txp(6 downto 0), pci_exp_rxp(7 downto 7), index, i_ds_axi_awready);

              when port_ds_axi_rdata =>
                transmit(pci_exp_txn(7 downto 0), pci_exp_txp(6 downto 0), pci_exp_rxp(7 downto 7), index, i_ds_axi_rdata);

              when port_ds_axi_rlast =>
                transmit(pci_exp_txn(7 downto 0), pci_exp_txp(6 downto 0), pci_exp_rxp(7 downto 7), index, i_ds_axi_rlast);

              when port_ds_axi_rresp =>
                transmit(pci_exp_txn(7 downto 0), pci_exp_txp(6 downto 0), pci_exp_rxp(7 downto 7), index, i_ds_axi_rresp);

              when port_ds_axi_rvalid =>
                transmit(pci_exp_txn(7 downto 0), pci_exp_txp(6 downto 0), pci_exp_rxp(7 downto 7), index, i_ds_axi_rvalid);

              when port_ds_axi_wready =>
                transmit(pci_exp_txn(7 downto 0), pci_exp_txp(6 downto 0), pci_exp_rxp(7 downto 7), index, i_ds_axi_wready);

              when port_ds_axi_bresp =>
                transmit(pci_exp_txn(7 downto 0), pci_exp_txp(6 downto 0), pci_exp_rxp(7 downto 7), index, i_ds_axi_bresp);

              when port_ds_axi_bvalid =>
                transmit(pci_exp_txn(7 downto 0), pci_exp_txp(6 downto 0), pci_exp_rxp(7 downto 7), index, i_ds_axi_bvalid);

              when port_dma0_axi_arready =>
                transmit(pci_exp_txn(7 downto 0), pci_exp_txp(6 downto 0), pci_exp_rxp(7 downto 7), index, i_dma0_axi_arready);

              when port_dma0_axi_awready =>
                transmit(pci_exp_txn(7 downto 0), pci_exp_txp(6 downto 0), pci_exp_rxp(7 downto 7), index, i_dma0_axi_awready);

              when port_dma0_axi_rdata =>
                transmit(pci_exp_txn(7 downto 0), pci_exp_txp(6 downto 0), pci_exp_rxp(7 downto 7), index, i_dma0_axi_rdata);

              when port_dma0_axi_rlast =>
                transmit(pci_exp_txn(7 downto 0), pci_exp_txp(6 downto 0), pci_exp_rxp(7 downto 7), index, i_dma0_axi_rlast);

              when port_dma0_axi_rresp =>
                transmit(pci_exp_txn(7 downto 0), pci_exp_txp(6 downto 0), pci_exp_rxp(7 downto 7), index, i_dma0_axi_rresp);

              when port_dma0_axi_rvalid =>
                transmit(pci_exp_txn(7 downto 0), pci_exp_txp(6 downto 0), pci_exp_rxp(7 downto 7), index, i_dma0_axi_rvalid);

              when port_dma0_axi_wready =>
                transmit(pci_exp_txn(7 downto 0), pci_exp_txp(6 downto 0), pci_exp_rxp(7 downto 7), index, i_dma0_axi_wready);

              when port_dma0_axi_bresp =>
                transmit(pci_exp_txn(7 downto 0), pci_exp_txp(6 downto 0), pci_exp_rxp(7 downto 7), index, i_dma0_axi_bresp);

              when port_dma0_axi_bvalid =>
                transmit(pci_exp_txn(7 downto 0), pci_exp_txp(6 downto 0), pci_exp_rxp(7 downto 7), index, i_dma0_axi_bvalid);

              when port_dma1_axi_arready =>
                transmit(pci_exp_txn(7 downto 0), pci_exp_txp(6 downto 0), pci_exp_rxp(7 downto 7), index, i_dma1_axi_arready);

              when port_dma1_axi_awready =>
                transmit(pci_exp_txn(7 downto 0), pci_exp_txp(6 downto 0), pci_exp_rxp(7 downto 7), index, i_dma1_axi_awready);

              when port_dma1_axi_rdata =>
                transmit(pci_exp_txn(7 downto 0), pci_exp_txp(6 downto 0), pci_exp_rxp(7 downto 7), index, i_dma1_axi_rdata);

              when port_dma1_axi_rlast =>
                transmit(pci_exp_txn(7 downto 0), pci_exp_txp(6 downto 0), pci_exp_rxp(7 downto 7), index, i_dma1_axi_rlast);

              when port_dma1_axi_rresp =>
                transmit(pci_exp_txn(7 downto 0), pci_exp_txp(6 downto 0), pci_exp_rxp(7 downto 7), index, i_dma1_axi_rresp);

              when port_dma1_axi_rvalid =>
                transmit(pci_exp_txn(7 downto 0), pci_exp_txp(6 downto 0), pci_exp_rxp(7 downto 7), index, i_dma1_axi_rvalid);

              when port_dma1_axi_wready =>
                transmit(pci_exp_txn(7 downto 0), pci_exp_txp(6 downto 0), pci_exp_rxp(7 downto 7), index, i_dma1_axi_wready);

              when port_dma1_axi_bresp =>
                transmit(pci_exp_txn(7 downto 0), pci_exp_txp(6 downto 0), pci_exp_rxp(7 downto 7), index, i_dma1_axi_bresp);

              when port_dma1_axi_bvalid =>
                transmit(pci_exp_txn(7 downto 0), pci_exp_txp(6 downto 0), pci_exp_rxp(7 downto 7), index, i_dma1_axi_bvalid);

              when port_dma2_axi_arready =>
                transmit(pci_exp_txn(7 downto 0), pci_exp_txp(6 downto 0), pci_exp_rxp(7 downto 7), index, i_dma2_axi_arready);

              when port_dma2_axi_awready =>
                transmit(pci_exp_txn(7 downto 0), pci_exp_txp(6 downto 0), pci_exp_rxp(7 downto 7), index, i_dma2_axi_awready);

              when port_dma2_axi_rdata =>
                transmit(pci_exp_txn(7 downto 0), pci_exp_txp(6 downto 0), pci_exp_rxp(7 downto 7), index, i_dma2_axi_rdata);

              when port_dma2_axi_rlast =>
                transmit(pci_exp_txn(7 downto 0), pci_exp_txp(6 downto 0), pci_exp_rxp(7 downto 7), index, i_dma2_axi_rlast);

              when port_dma2_axi_rresp =>
                transmit(pci_exp_txn(7 downto 0), pci_exp_txp(6 downto 0), pci_exp_rxp(7 downto 7), index, i_dma2_axi_rresp);

              when port_dma2_axi_rvalid =>
                transmit(pci_exp_txn(7 downto 0), pci_exp_txp(6 downto 0), pci_exp_rxp(7 downto 7), index, i_dma2_axi_rvalid);

              when port_dma2_axi_wready =>
                transmit(pci_exp_txn(7 downto 0), pci_exp_txp(6 downto 0), pci_exp_rxp(7 downto 7), index, i_dma2_axi_wready);

              when port_dma2_axi_bresp =>
                transmit(pci_exp_txn(7 downto 0), pci_exp_txp(6 downto 0), pci_exp_rxp(7 downto 7), index, i_dma2_axi_bresp);

              when port_dma2_axi_bvalid =>
                transmit(pci_exp_txn(7 downto 0), pci_exp_txp(6 downto 0), pci_exp_rxp(7 downto 7), index, i_dma2_axi_bvalid);

              when port_dma3_axi_arready =>
                transmit(pci_exp_txn(7 downto 0), pci_exp_txp(6 downto 0), pci_exp_rxp(7 downto 7), index, i_dma3_axi_arready);

              when port_dma3_axi_awready =>
                transmit(pci_exp_txn(7 downto 0), pci_exp_txp(6 downto 0), pci_exp_rxp(7 downto 7), index, i_dma3_axi_awready);

              when port_dma3_axi_rdata =>
                transmit(pci_exp_txn(7 downto 0), pci_exp_txp(6 downto 0), pci_exp_rxp(7 downto 7), index, i_dma3_axi_rdata);

              when port_dma3_axi_rlast =>
                transmit(pci_exp_txn(7 downto 0), pci_exp_txp(6 downto 0), pci_exp_rxp(7 downto 7), index, i_dma3_axi_rlast);

              when port_dma3_axi_rresp =>
                transmit(pci_exp_txn(7 downto 0), pci_exp_txp(6 downto 0), pci_exp_rxp(7 downto 7), index, i_dma3_axi_rresp);

              when port_dma3_axi_rvalid =>
                transmit(pci_exp_txn(7 downto 0), pci_exp_txp(6 downto 0), pci_exp_rxp(7 downto 7), index, i_dma3_axi_rvalid);

              when port_dma3_axi_wready =>
                transmit(pci_exp_txn(7 downto 0), pci_exp_txp(6 downto 0), pci_exp_rxp(7 downto 7), index, i_dma3_axi_wready);

              when port_dma3_axi_bresp =>
                transmit(pci_exp_txn(7 downto 0), pci_exp_txp(6 downto 0), pci_exp_rxp(7 downto 7), index, i_dma3_axi_bresp);

              when port_dma3_axi_bvalid =>
                transmit(pci_exp_txn(7 downto 0), pci_exp_txp(6 downto 0), pci_exp_rxp(7 downto 7), index, i_dma3_axi_bvalid);

              when port_dm_axi_araddr =>
                transmit(pci_exp_txn(7 downto 0), pci_exp_txp(6 downto 0), pci_exp_rxp(7 downto 7), index, i_dm_axi_araddr);

              when port_dm_axi_arburst =>
                transmit(pci_exp_txn(7 downto 0), pci_exp_txp(6 downto 0), pci_exp_rxp(7 downto 7), index, i_dm_axi_arburst);

              when port_dm_axi_arcache =>
                transmit(pci_exp_txn(7 downto 0), pci_exp_txp(6 downto 0), pci_exp_rxp(7 downto 7), index, i_dm_axi_arcache);

              when port_dm_axi_arid =>
                transmit(pci_exp_txn(7 downto 0), pci_exp_txp(6 downto 0), pci_exp_rxp(7 downto 7), index, i_dm_axi_arid);

              when port_dm_axi_arlen =>
                transmit(pci_exp_txn(7 downto 0), pci_exp_txp(6 downto 0), pci_exp_rxp(7 downto 7), index, i_dm_axi_arlen);

              when port_dm_axi_arlock =>
                transmit(pci_exp_txn(7 downto 0), pci_exp_txp(6 downto 0), pci_exp_rxp(7 downto 7), index, i_dm_axi_arlock);

              when port_dm_axi_arprot =>
                transmit(pci_exp_txn(7 downto 0), pci_exp_txp(6 downto 0), pci_exp_rxp(7 downto 7), index, i_dm_axi_arprot);

              when port_dm_axi_arqos =>
                transmit(pci_exp_txn(7 downto 0), pci_exp_txp(6 downto 0), pci_exp_rxp(7 downto 7), index, i_dm_axi_arqos);

              when port_dm_axi_arregion =>
                transmit(pci_exp_txn(7 downto 0), pci_exp_txp(6 downto 0), pci_exp_rxp(7 downto 7), index, i_dm_axi_arregion);

              when port_dm_axi_arsize =>
                transmit(pci_exp_txn(7 downto 0), pci_exp_txp(6 downto 0), pci_exp_rxp(7 downto 7), index, i_dm_axi_arsize);

              when port_dm_axi_arvalid =>
                transmit(pci_exp_txn(7 downto 0), pci_exp_txp(6 downto 0), pci_exp_rxp(7 downto 7), index, i_dm_axi_arvalid);

              when port_dm_axi_awaddr =>
                transmit(pci_exp_txn(7 downto 0), pci_exp_txp(6 downto 0), pci_exp_rxp(7 downto 7), index, i_dm_axi_awaddr);

              when port_dm_axi_awburst =>
                transmit(pci_exp_txn(7 downto 0), pci_exp_txp(6 downto 0), pci_exp_rxp(7 downto 7), index, i_dm_axi_awburst);

              when port_dm_axi_awcache =>
                transmit(pci_exp_txn(7 downto 0), pci_exp_txp(6 downto 0), pci_exp_rxp(7 downto 7), index, i_dm_axi_awcache);

              when port_dm_axi_awid =>
                transmit(pci_exp_txn(7 downto 0), pci_exp_txp(6 downto 0), pci_exp_rxp(7 downto 7), index, i_dm_axi_awid);

              when port_dm_axi_awlen =>
                transmit(pci_exp_txn(7 downto 0), pci_exp_txp(6 downto 0), pci_exp_rxp(7 downto 7), index, i_dm_axi_awlen);

              when port_dm_axi_awlock =>
                transmit(pci_exp_txn(7 downto 0), pci_exp_txp(6 downto 0), pci_exp_rxp(7 downto 7), index, i_dm_axi_awlock);

              when port_dm_axi_awprot =>
                transmit(pci_exp_txn(7 downto 0), pci_exp_txp(6 downto 0), pci_exp_rxp(7 downto 7), index, i_dm_axi_awprot);

              when port_dm_axi_awqos =>
                transmit(pci_exp_txn(7 downto 0), pci_exp_txp(6 downto 0), pci_exp_rxp(7 downto 7), index, i_dm_axi_awqos);

              when port_dm_axi_awregion =>
                transmit(pci_exp_txn(7 downto 0), pci_exp_txp(6 downto 0), pci_exp_rxp(7 downto 7), index, i_dm_axi_awregion);

              when port_dm_axi_awsize =>
                transmit(pci_exp_txn(7 downto 0), pci_exp_txp(6 downto 0), pci_exp_rxp(7 downto 7), index, i_dm_axi_awsize);

              when port_dm_axi_awvalid =>
                transmit(pci_exp_txn(7 downto 0), pci_exp_txp(6 downto 0), pci_exp_rxp(7 downto 7), index, i_dm_axi_awvalid);

              when port_dm_axi_rready =>
                transmit(pci_exp_txn(7 downto 0), pci_exp_txp(6 downto 0), pci_exp_rxp(7 downto 7), index, i_dm_axi_rready);

              when port_dm_axi_wdata =>
                transmit(pci_exp_txn(7 downto 0), pci_exp_txp(6 downto 0), pci_exp_rxp(7 downto 7), index, i_dm_axi_wdata);

              when port_dm_axi_wlast =>
                transmit(pci_exp_txn(7 downto 0), pci_exp_txp(6 downto 0), pci_exp_rxp(7 downto 7), index, i_dm_axi_wlast);

              when port_dm_axi_wstrb =>
                transmit(pci_exp_txn(7 downto 0), pci_exp_txp(6 downto 0), pci_exp_rxp(7 downto 7), index, i_dm_axi_wstrb);

              when port_dm_axi_wvalid =>
                transmit(pci_exp_txn(7 downto 0), pci_exp_txp(6 downto 0), pci_exp_rxp(7 downto 7), index, i_dm_axi_wvalid);

              when port_dm_axi_bready =>
                transmit(pci_exp_txn(7 downto 0), pci_exp_txp(6 downto 0), pci_exp_rxp(7 downto 7), index, i_dm_axi_bready);

              when port_dmb_axi_araddr =>
                transmit(pci_exp_txn(7 downto 0), pci_exp_txp(6 downto 0), pci_exp_rxp(7 downto 7), index, i_dmb_axi_araddr);

              when port_dmb_axi_arvalid =>
                transmit(pci_exp_txn(7 downto 0), pci_exp_txp(6 downto 0), pci_exp_rxp(7 downto 7), index, i_dmb_axi_arvalid);

              when port_dmb_axi_awaddr =>
                transmit(pci_exp_txn(7 downto 0), pci_exp_txp(6 downto 0), pci_exp_rxp(7 downto 7), index, i_dmb_axi_awaddr);

              when port_dmb_axi_awvalid =>
                transmit(pci_exp_txn(7 downto 0), pci_exp_txp(6 downto 0), pci_exp_rxp(7 downto 7), index, i_dmb_axi_awvalid);

              when port_dmb_axi_rready =>
                transmit(pci_exp_txn(7 downto 0), pci_exp_txp(6 downto 0), pci_exp_rxp(7 downto 7), index, i_dmb_axi_rready);

              when port_dmb_axi_wdata =>
                transmit(pci_exp_txn(7 downto 0), pci_exp_txp(6 downto 0), pci_exp_rxp(7 downto 7), index, i_dmb_axi_wdata);

              when port_dmb_axi_wstrb =>
                transmit(pci_exp_txn(7 downto 0), pci_exp_txp(6 downto 0), pci_exp_rxp(7 downto 7), index, i_dmb_axi_wstrb);

              when port_dmb_axi_wvalid =>
                transmit(pci_exp_txn(7 downto 0), pci_exp_txp(6 downto 0), pci_exp_rxp(7 downto 7), index, i_dmb_axi_wvalid);

              when port_dmb_axi_bready =>
                transmit(pci_exp_txn(7 downto 0), pci_exp_txp(6 downto 0), pci_exp_rxp(7 downto 7), index, i_dmb_axi_bready);

              when port_dma0_m_axis_tready =>
                transmit(pci_exp_txn(7 downto 0), pci_exp_txp(6 downto 0), pci_exp_rxp(7 downto 7), index, i_dma0_m_axis_tready);

              when port_dma0_s_axis_tdata =>
                transmit(pci_exp_txn(7 downto 0), pci_exp_txp(6 downto 0), pci_exp_rxp(7 downto 7), index, i_dma0_s_axis_tdata);

              when port_dma0_s_axis_tvalid =>
                transmit(pci_exp_txn(7 downto 0), pci_exp_txp(6 downto 0), pci_exp_rxp(7 downto 7), index, i_dma0_s_axis_tvalid);

              when port_dma1_m_axis_tready =>
                transmit(pci_exp_txn(7 downto 0), pci_exp_txp(6 downto 0), pci_exp_rxp(7 downto 7), index, i_dma1_m_axis_tready);

              when port_dma1_s_axis_tdata =>
                transmit(pci_exp_txn(7 downto 0), pci_exp_txp(6 downto 0), pci_exp_rxp(7 downto 7), index, i_dma1_s_axis_tdata);

              when port_dma1_s_axis_tvalid =>
                transmit(pci_exp_txn(7 downto 0), pci_exp_txp(6 downto 0), pci_exp_rxp(7 downto 7), index, i_dma1_s_axis_tvalid);

              when port_dma2_m_axis_tready =>
                transmit(pci_exp_txn(7 downto 0), pci_exp_txp(6 downto 0), pci_exp_rxp(7 downto 7), index, i_dma2_m_axis_tready);

              when port_dma2_s_axis_tdata =>
                transmit(pci_exp_txn(7 downto 0), pci_exp_txp(6 downto 0), pci_exp_rxp(7 downto 7), index, i_dma2_s_axis_tdata);

              when port_dma2_s_axis_tvalid =>
                transmit(pci_exp_txn(7 downto 0), pci_exp_txp(6 downto 0), pci_exp_rxp(7 downto 7), index, i_dma2_s_axis_tvalid);

              when port_dma3_m_axis_tready =>
                transmit(pci_exp_txn(7 downto 0), pci_exp_txp(6 downto 0), pci_exp_rxp(7 downto 7), index, i_dma3_m_axis_tready);

              when port_dma3_s_axis_tdata =>
                transmit(pci_exp_txn(7 downto 0), pci_exp_txp(6 downto 0), pci_exp_rxp(7 downto 7), index, i_dma3_s_axis_tdata);

              when port_dma3_s_axis_tvalid =>
                transmit(pci_exp_txn(7 downto 0), pci_exp_txp(6 downto 0), pci_exp_rxp(7 downto 7), index, i_dma3_s_axis_tvalid);

              when port_interrupt =>
                transmit(pci_exp_txn(7 downto 0), pci_exp_txp(6 downto 0), pci_exp_rxp(7 downto 7), index, i_interrupt);

              when port_dma0_abort =>
                transmit(pci_exp_txn(7 downto 0), pci_exp_txp(6 downto 0), pci_exp_rxp(7 downto 7), index, i_dma0_abort);

              when port_dma1_abort =>
                transmit(pci_exp_txn(7 downto 0), pci_exp_txp(6 downto 0), pci_exp_rxp(7 downto 7), index, i_dma1_abort);

              when port_dma2_abort =>
                transmit(pci_exp_txn(7 downto 0), pci_exp_txp(6 downto 0), pci_exp_rxp(7 downto 7), index, i_dma2_abort);

              when port_dma3_abort =>
                transmit(pci_exp_txn(7 downto 0), pci_exp_txp(6 downto 0), pci_exp_rxp(7 downto 7), index, i_dma3_abort);

              when others =>
                report "transmit: case " & port_in_enum_t'image(port_in) & " is not handled." severity failure;
            end case;

            -- Indicate that signal has been updated
            if handshake_use_quantum then
              ack_in(index) <= true after handshake_quantum;
            else
              ack_in(index) <= true;
            end if;
            
            -- Complete the handshake
            wait until not pending_in(index);
            if handshake_use_quantum then
              ack_in(index) <= false after handshake_quantum;
            else
              ack_in(index) <= false;
            end if;

            if transmit_use_quantum then
              -- Avoid excessive numbers of delta cycles at a given simulation time.
              wait for transmit_quantum;
            end if;
          end if;
        end loop;
        exit when not work_done;
      end loop;
      wait on pending_in;
    end loop;
  end process;


  receive_signals : process
    variable index : natural;
    variable val : std_logic;
    variable port_out : port_out_enum_t;
  begin
    loop
      receive_header(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), index, val);
      port_out := port_out_enum_t'val(index);
      case port_out is
        when port_aclk =>
          aclk <= val;

        when port_aresetn =>
          aresetn <= val;

        when port_ds_axi_araddr =>
          receive(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), ds_axi_araddr);

        when port_ds_axi_arburst =>
          receive(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), ds_axi_arburst);

        when port_ds_axi_arcache =>
          receive(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), ds_axi_arcache);

        when port_ds_axi_arlen =>
          receive(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), ds_axi_arlen);

        when port_ds_axi_arlock =>
          receive(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), ds_axi_arlock);

        when port_ds_axi_arprot =>
          receive(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), ds_axi_arprot);

        when port_ds_axi_arqos =>
          receive(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), ds_axi_arqos);

        when port_ds_axi_arregion =>
          receive(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), ds_axi_arregion);

        when port_ds_axi_arsize =>
          receive(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), ds_axi_arsize);

        when port_ds_axi_arvalid =>
          ds_axi_arvalid <= val;

        when port_ds_axi_awaddr =>
          receive(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), ds_axi_awaddr);

        when port_ds_axi_awburst =>
          receive(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), ds_axi_awburst);

        when port_ds_axi_awcache =>
          receive(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), ds_axi_awcache);

        when port_ds_axi_awlen =>
          receive(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), ds_axi_awlen);

        when port_ds_axi_awlock =>
          receive(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), ds_axi_awlock);

        when port_ds_axi_awprot =>
          receive(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), ds_axi_awprot);

        when port_ds_axi_awqos =>
          receive(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), ds_axi_awqos);

        when port_ds_axi_awregion =>
          receive(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), ds_axi_awregion);

        when port_ds_axi_awsize =>
          receive(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), ds_axi_awsize);

        when port_ds_axi_awvalid =>
          ds_axi_awvalid <= val;

        when port_ds_axi_rready =>
          ds_axi_rready <= val;

        when port_ds_axi_wdata =>
          receive(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), ds_axi_wdata);

        when port_ds_axi_wlast =>
          ds_axi_wlast <= val;

        when port_ds_axi_wstrb =>
          receive(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), ds_axi_wstrb);

        when port_ds_axi_wvalid =>
          ds_axi_wvalid <= val;

        when port_ds_axi_bready =>
          ds_axi_bready <= val;

        when port_dma0_axi_araddr =>
          receive(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), dma0_axi_araddr);

        when port_dma0_axi_arburst =>
          receive(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), dma0_axi_arburst);

        when port_dma0_axi_arcache =>
          receive(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), dma0_axi_arcache);

        when port_dma0_axi_arlen =>
          receive(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), dma0_axi_arlen);

        when port_dma0_axi_arlock =>
          receive(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), dma0_axi_arlock);

        when port_dma0_axi_arprot =>
          receive(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), dma0_axi_arprot);

        when port_dma0_axi_arqos =>
          receive(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), dma0_axi_arqos);

        when port_dma0_axi_arregion =>
          receive(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), dma0_axi_arregion);

        when port_dma0_axi_arsize =>
          receive(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), dma0_axi_arsize);

        when port_dma0_axi_arvalid =>
          dma0_axi_arvalid <= val;

        when port_dma0_axi_awaddr =>
          receive(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), dma0_axi_awaddr);

        when port_dma0_axi_awburst =>
          receive(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), dma0_axi_awburst);

        when port_dma0_axi_awcache =>
          receive(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), dma0_axi_awcache);

        when port_dma0_axi_awlen =>
          receive(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), dma0_axi_awlen);

        when port_dma0_axi_awlock =>
          receive(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), dma0_axi_awlock);

        when port_dma0_axi_awprot =>
          receive(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), dma0_axi_awprot);

        when port_dma0_axi_awqos =>
          receive(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), dma0_axi_awqos);

        when port_dma0_axi_awregion =>
          receive(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), dma0_axi_awregion);

        when port_dma0_axi_awsize =>
          receive(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), dma0_axi_awsize);

        when port_dma0_axi_awvalid =>
          dma0_axi_awvalid <= val;

        when port_dma0_axi_rready =>
          dma0_axi_rready <= val;

        when port_dma0_axi_wdata =>
          receive(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), dma0_axi_wdata);

        when port_dma0_axi_wstrb =>
          receive(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), dma0_axi_wstrb);

        when port_dma0_axi_wlast =>
          dma0_axi_wlast <= val;

        when port_dma0_axi_wvalid =>
          dma0_axi_wvalid <= val;

        when port_dma0_axi_bready =>
          dma0_axi_bready <= val;

        when port_dma1_axi_araddr =>
          receive(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), dma1_axi_araddr);

        when port_dma1_axi_arburst =>
          receive(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), dma1_axi_arburst);

        when port_dma1_axi_arcache =>
          receive(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), dma1_axi_arcache);

        when port_dma1_axi_arlen =>
          receive(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), dma1_axi_arlen);

        when port_dma1_axi_arlock =>
          receive(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), dma1_axi_arlock);

        when port_dma1_axi_arprot =>
          receive(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), dma1_axi_arprot);

        when port_dma1_axi_arqos =>
          receive(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), dma1_axi_arqos);

        when port_dma1_axi_arregion =>
          receive(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), dma1_axi_arregion);

        when port_dma1_axi_arsize =>
          receive(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), dma1_axi_arsize);

        when port_dma1_axi_arvalid =>
          dma1_axi_arvalid <= val;

        when port_dma1_axi_awaddr =>
          receive(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), dma1_axi_awaddr);

        when port_dma1_axi_awburst =>
          receive(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), dma1_axi_awburst);

        when port_dma1_axi_awcache =>
          receive(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), dma1_axi_awcache);

        when port_dma1_axi_awlen =>
          receive(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), dma1_axi_awlen);

        when port_dma1_axi_awlock =>
          receive(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), dma1_axi_awlock);

        when port_dma1_axi_awprot =>
          receive(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), dma1_axi_awprot);

        when port_dma1_axi_awqos =>
          receive(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), dma1_axi_awqos);

        when port_dma1_axi_awregion =>
          receive(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), dma1_axi_awregion);

        when port_dma1_axi_awsize =>
          receive(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), dma1_axi_awsize);

        when port_dma1_axi_awvalid =>
          dma1_axi_awvalid <= val;

        when port_dma1_axi_rready =>
          dma1_axi_rready <= val;

        when port_dma1_axi_wdata =>
          receive(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), dma1_axi_wdata);

        when port_dma1_axi_wstrb =>
          receive(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), dma1_axi_wstrb);

        when port_dma1_axi_wlast =>
          dma1_axi_wlast <= val;

        when port_dma1_axi_wvalid =>
          dma1_axi_wvalid <= val;

        when port_dma1_axi_bready =>
          dma1_axi_bready <= val;

        when port_dma2_axi_araddr =>
          receive(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), dma2_axi_araddr);

        when port_dma2_axi_arburst =>
          receive(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), dma2_axi_arburst);

        when port_dma2_axi_arcache =>
          receive(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), dma2_axi_arcache);

        when port_dma2_axi_arlen =>
          receive(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), dma2_axi_arlen);

        when port_dma2_axi_arlock =>
          receive(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), dma2_axi_arlock);

        when port_dma2_axi_arprot =>
          receive(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), dma2_axi_arprot);

        when port_dma2_axi_arqos =>
          receive(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), dma2_axi_arqos);

        when port_dma2_axi_arregion =>
          receive(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), dma2_axi_arregion);

        when port_dma2_axi_arsize =>
          receive(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), dma2_axi_arsize);

        when port_dma2_axi_arvalid =>
          dma2_axi_arvalid <= val;

        when port_dma2_axi_awaddr =>
          receive(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), dma2_axi_awaddr);

        when port_dma2_axi_awburst =>
          receive(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), dma2_axi_awburst);

        when port_dma2_axi_awcache =>
          receive(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), dma2_axi_awcache);

        when port_dma2_axi_awlen =>
          receive(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), dma2_axi_awlen);

        when port_dma2_axi_awlock =>
          receive(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), dma2_axi_awlock);

        when port_dma2_axi_awprot =>
          receive(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), dma2_axi_awprot);

        when port_dma2_axi_awqos =>
          receive(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), dma2_axi_awqos);

        when port_dma2_axi_awregion =>
          receive(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), dma2_axi_awregion);

        when port_dma2_axi_awsize =>
          receive(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), dma2_axi_awsize);

        when port_dma2_axi_awvalid =>
          dma2_axi_awvalid <= val;

        when port_dma2_axi_rready =>
          dma2_axi_rready <= val;

        when port_dma2_axi_wdata =>
          receive(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), dma2_axi_wdata);

        when port_dma2_axi_wstrb =>
          receive(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), dma2_axi_wstrb);

        when port_dma2_axi_wlast =>
          dma2_axi_wlast <= val;

        when port_dma2_axi_wvalid =>
          dma2_axi_wvalid <= val;

        when port_dma2_axi_bready =>
          dma2_axi_bready <= val;

        when port_dma3_axi_araddr =>
          receive(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), dma3_axi_araddr);

        when port_dma3_axi_arburst =>
          receive(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), dma3_axi_arburst);

        when port_dma3_axi_arcache =>
          receive(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), dma3_axi_arcache);

        when port_dma3_axi_arlen =>
          receive(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), dma3_axi_arlen);

        when port_dma3_axi_arlock =>
          receive(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), dma3_axi_arlock);

        when port_dma3_axi_arprot =>
          receive(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), dma3_axi_arprot);

        when port_dma3_axi_arqos =>
          receive(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), dma3_axi_arqos);

        when port_dma3_axi_arregion =>
          receive(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), dma3_axi_arregion);

        when port_dma3_axi_arsize =>
          receive(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), dma3_axi_arsize);

        when port_dma3_axi_arvalid =>
          dma3_axi_arvalid <= val;

        when port_dma3_axi_awaddr =>
          receive(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), dma3_axi_awaddr);

        when port_dma3_axi_awburst =>
          receive(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), dma3_axi_awburst);

        when port_dma3_axi_awcache =>
          receive(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), dma3_axi_awcache);

        when port_dma3_axi_awlen =>
          receive(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), dma3_axi_awlen);

        when port_dma3_axi_awlock =>
          receive(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), dma3_axi_awlock);

        when port_dma3_axi_awprot =>
          receive(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), dma3_axi_awprot);

        when port_dma3_axi_awqos =>
          receive(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), dma3_axi_awqos);

        when port_dma3_axi_awregion =>
          receive(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), dma3_axi_awregion);

        when port_dma3_axi_awsize =>
          receive(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), dma3_axi_awsize);

        when port_dma3_axi_awvalid =>
          dma3_axi_awvalid <= val;

        when port_dma3_axi_rready =>
          dma3_axi_rready <= val;

        when port_dma3_axi_wdata =>
          receive(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), dma3_axi_wdata);

        when port_dma3_axi_wstrb =>
          receive(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), dma3_axi_wstrb);

        when port_dma3_axi_wlast =>
          dma3_axi_wlast <= val;

        when port_dma3_axi_wvalid =>
          dma3_axi_wvalid <= val;

        when port_dma3_axi_bready =>
          dma3_axi_bready <= val;

        when port_dm_axi_arready =>
          dm_axi_arready <= val;

        when port_dm_axi_awready =>
          dm_axi_awready <= val;

        when port_dm_axi_rdata =>
          receive(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), dm_axi_rdata);

        when port_dm_axi_rid =>
          receive(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), dm_axi_rid);

        when port_dm_axi_rlast =>
          dm_axi_rlast <= val;

        when port_dm_axi_rresp =>
          receive(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), dm_axi_rresp);

        when port_dm_axi_rvalid =>
          dm_axi_rvalid <= val;

        when port_dm_axi_wready =>
          dm_axi_wready <= val;

        when port_dm_axi_bid =>
          receive(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), dm_axi_bid);

        when port_dm_axi_bresp =>
          receive(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), dm_axi_bresp);

        when port_dm_axi_bvalid =>
          dm_axi_bvalid <= val;

        when port_dmb_axi_arready =>
          dmb_axi_arready <= val;

        when port_dmb_axi_awready =>
          dmb_axi_awready <= val;

        when port_dmb_axi_rdata =>
          receive(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), dmb_axi_rdata);

        when port_dmb_axi_rresp =>
          receive(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), dmb_axi_rresp);

        when port_dmb_axi_rvalid =>
          dmb_axi_rvalid <= val;

        when port_dmb_axi_wready =>
          dmb_axi_wready <= val;

        when port_dmb_axi_bresp =>
          receive(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), dmb_axi_bresp);

        when port_dmb_axi_bvalid =>
          dmb_axi_bvalid <= val;

        when port_dma0_m_axis_tdata =>
          receive(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), dma0_m_axis_tdata);

        when port_dma0_m_axis_tvalid =>
          dma0_m_axis_tvalid <= val;

        when port_dma0_s_axis_tready =>
          dma0_s_axis_tready <= val;

        when port_dma1_m_axis_tdata =>
          receive(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), dma1_m_axis_tdata);

        when port_dma1_m_axis_tvalid =>
          dma1_m_axis_tvalid <= val;

        when port_dma1_s_axis_tready =>
          dma1_s_axis_tready <= val;

        when port_dma2_m_axis_tdata =>
          receive(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), dma2_m_axis_tdata);

        when port_dma2_m_axis_tvalid =>
          dma2_m_axis_tvalid <= val;

        when port_dma2_s_axis_tready =>
          dma2_s_axis_tready <= val;

        when port_dma3_m_axis_tdata =>
          receive(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), dma3_m_axis_tdata);

        when port_dma3_m_axis_tvalid =>
          dma3_m_axis_tvalid <= val;

        when port_dma3_s_axis_tready =>
          dma3_s_axis_tready <= val;

        when port_core_status =>
          receive(pci_exp_rxn(7 downto 0), pci_exp_rxp(6 downto 0), pci_exp_txp(7 downto 7), core_status);

        when others =>
          report "receive: case " & port_out_enum_t'image(port_out) & " is not handled." severity failure;
      end case;
    end loop;
  end process;

end architecture;
