-------------------------------------------------------------------------------
-- Module Name      : read_worker
--
--                    Copyright 2009 by
--                    Alpha Data Parallel Systems Ltd.
--                    All Rights Reserved.
--
-- Original Author  : Andrew McCormick
-- Created          : 
--
-- Modified By      : 
-- Date             :
-- Change Notes     :     
--
-- Description      : 
--
--
-- Dependencies     : adb3_ocp,adb3_dma
--
-- Disclaimer       : THESE DESIGNS ARE PROVIDED "AS IS" WITH NO WARRANTY
--                    WHATSOEVER AND ALPHA DATA SPECIFICALLY DISCLAIMS ANY
--                    WARRANTIES IMPLIED OF MERCHANTABILITY, FITNESS FOR A
--                    PARTICULAR PURPOSE, OR AGAINST INFRINGEMENT.
--
--------------------------------------------------------------------------------

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_misc.all;

library work;
use work.adb3_ocp.all;
use work.adb3_dma.all;


entity read_worker256 is
  generic (
    check_tags : boolean := true;
    addr_bits  : natural := 64;
    left_bits  : natural := 39;
    buffer_descriptors : boolean := true;
    registering_for_speed : boolean := true;
    chipscope : boolean := false);
  port (
    clk              : in  std_logic;
    -- ABORT/Cleanup Control
    abort            : in  std_logic;
    cleanup          : in  std_logic;
    -- FSM Status
    fsm_idle         : out std_logic_vector(2 downto 0);
    fsm_aborting     : out std_logic_vector(2 downto 0);
    fsm_error        : out std_logic;
    -- OCP Master Port (to MUX to PCIe)
    ocp_m2s          : out adb3w_ocp_m2sT;
    ocp_s2m          : in  adb3w_ocp_s2mT;
    -- Worker DMA Engine Control
    worker_cmd       : in  adb3_dma_worker_command_type;
    worker_cmd_valid : in  std_logic;
    worker_ready     : out std_logic;
    -- Burst Length Configuration
    max_burst        : in  std_logic_vector(8 downto 0);
    -- Need Extra Write if write FIFO rotation
    need_extra_write : in  std_logic;
    -- IRQ Control
    irq              : out std_logic;
    irq_ack          : in  std_logic;
    -- Data Path
    fifo_data        : out std_logic_vector(255 downto 0);
    fifo_nfull       : in  std_logic;
    fifo_wadv        : out std_logic
    );

end read_worker256;

architecture rtl of read_worker256 is

  -- State Machine for converting Read Descriptor Commands into Short OCP Bursts 
  type state_type is (IDLE,
                      IS_ADDR_ALIGNED,
                      FIRST_BURST_WAIT,
                      IS_FIRST_BURST_LEFT,
                      GEN_FIRST_BURST,
                      IS_BURST_LEFT,
                      GEN_BURST,
                      IS_SHORT_BURST_LEFT,
                      GEN_SHORT_BURST,
                      GEN_IRQ,
                      S_ABORT);
  signal state : state_type := S_ABORT;



  -- State Machine for Requestsing Short OCP Reads
  type req_state_type is (REQ_IDLE,
                          REQ_READ_BURST_DESC,
                          REQ_GEN_IRQ,
                          REQ_GEN_READ_CMD,
                          REQ_ABORT);

  signal req_state : req_state_type := REQ_ABORT;

  -- State Machine for Completing OCP Reads
  type comp_state_type is (COMP_IDLE,
                           COMP_READ_IFF,
                           COMP_GEN_IRQ,
                           COMP_GEN_FLUSH,
                           COMP_ACCEPT_DATA,
                           COMP_ABORT,
                           COMP_ERROR);

  
  signal comp_state : comp_state_type := COMP_ABORT;
  signal comp_state_r1 : comp_state_type := COMP_ABORT;

  signal ocp_s2m_resp_r1  : std_logic_vector(1 downto 0);
  signal ocp_s2m_tag_r1 : std_logic_vector(7 downto 0) := (others => '0');
  signal ocp_s2m_data_r1 : std_logic_vector(255 downto 0) := (others => '0');

  component small_sfifo
    generic (
      width : integer := 64);
    port (
      clk    : in  std_logic;
      rst    : in  std_logic;
      wadv   : in  std_logic;
      wdata  : in  std_logic_vector(width-1 downto 0);
      nfull  : out std_logic;
      radv   : in  std_logic;
      rempty : out std_logic;
      rdata  : out std_logic_vector(width-1 downto 0));
  end component;

  component small_sfifo32
    generic (
      width : integer := 64);
    port (
      clk    : in  std_logic;
      rst    : in  std_logic;
      wadv   : in  std_logic;
      wdata  : in  std_logic_vector(width-1 downto 0);
      nfull  : out std_logic;
      radv   : in  std_logic;
      rempty : out std_logic;
      rdata  : out std_logic_vector(width-1 downto 0));
  end component;

   component small_sfifo_reg_op
    generic (
      width : integer := 64);
    port (
      clk    : in  std_logic;
      rst    : in  std_logic;
      wadv   : in  std_logic;
      wdata  : in  std_logic_vector(width-1 downto 0);
      nfull  : out std_logic;
      radv   : in  std_logic;
      rempty : out std_logic;
      rdata  : out std_logic_vector(width-1 downto 0));
  end component;

  component dpram_sfifo
    generic (
      width : integer);
    port (
      clk    : in  std_logic;
      rst    : in  std_logic;
      wadv   : in  std_logic;
      wdata  : in  std_logic_vector(width-1 downto 0);
      nfull  : out std_logic;
      radv   : in  std_logic;
      rempty : out std_logic;
      rdata  : out std_logic_vector(width-1 downto 0));
  end component;


  signal cmd_fifo_d, cmd_fifo_q : std_logic_vector(106 downto 0) := (others => '0');
  signal worker_cmd_q           : adb3_dma_worker_command_type := ADB3_DMA_WORKER_COMMAND_TYPE_DEFAULT ;
  signal cmd_fifo_nfull         : std_logic := '0';
  signal cmd_fifo_empty         : std_logic := '1';
  signal cmd_fifo_radv          : std_logic := '0';
  signal cmd_fifo_wadv          : std_logic := '0';
  signal cmd_fifo_nfull_reg         : std_logic:= '0';

  signal cmd_fifo_empty_r1         : std_logic := '1';
  signal cmd_fifo_empty_r2         : std_logic := '1';
  
  signal current_cmd : adb3_dma_worker_command_type := ADB3_DMA_WORKER_COMMAND_TYPE_DEFAULT;

  signal addr0, addr1, addr2, addr3, addr4 : std_logic_vector(13 downto 0) := (others => '0');
  signal left0, left0_reg, left1, left2, left3, left4 : std_logic_vector(13 downto 0) := (others => '0');
  signal left_inc                          : std_logic := '0';
  signal addr_valid, left_valid            : std_logic := '0';

  --signal addr, xfer_left, xfer_left_reg : std_logic_vector(63 downto 0);
  signal xfer_left, xfer_left_reg : std_logic_vector(left_bits-1 downto 0) := (others => '0');
  signal addr,addr_reg        : std_logic_vector(63 downto 0) := (others => '0');

  signal lots_left     : std_logic := '0';
  signal max_burst_reg1 : std_logic_vector(8 downto 0);-- := "0000000000";
  signal max_burst_reg2 : std_logic_vector(8 downto 0);-- := "0000001001";
  signal max_burst_reg3 : std_logic_vector(8 downto 0);-- := "0000010010";
  signal max_burst_reg4 : std_logic_vector(8 downto 0);-- := "0000100011";

  signal max_read_inflight : std_logic_vector(5 downto 0);
  signal read_inflight_count : std_logic_vector(6 downto 0) := (others => '0');
  signal read_inflight_full : std_Logic;
  
  signal inc_addr      : std_logic := '1';

  signal irq_cmd : std_logic := '0';
  signal size    : std_logic_vector(8 downto 0) := (others => '0');
  signal flush_cmd  : std_logic := '0';
  signal need_flush : std_logic := '0';
  signal need_flush_calc : std_logic_vector(5 downto 0) := (others => '0');
  
  
  signal burst_desc_fifo_ready : std_logic := '0';
  signal burst_desc_fifo_nfull : std_logic := '0';
   signal burst_desc_fifo_nfull_reg : std_logic := '0';
  signal burst_desc_fifo_wadv  : std_logic := '0';
  signal burst_desc_fifo_radv  : std_logic := '0';
  signal burst_desc_fifo_empty : std_logic := '1';
  signal burst_desc_fifo_empty_r1 : std_logic := '1';
  signal burst_desc_fifo_empty_r2 : std_logic := '1';
  
  signal burst_desc_fifo_d : std_logic_vector(75 downto 0) := (others => '0');
  signal burst_desc_fifo_q : std_logic_vector(75 downto 0) := (others => '0');
  signal burst_desc_fifo_q_r1 :  std_logic_vector(75 downto 0) := (others => '0');
  signal burst_desc_fifo_q_r2 :  std_logic_vector(75 downto 0) := (others => '0');

  signal r_irq_cmd : std_logic := '0';
  signal r_size    : std_logic_vector(8 downto 0) := (others => '0');
  signal r_addr    : std_logic_vector(63 downto 0) := (others => '0');
  signal r_flush_cmd   : std_logic := '0';

  signal r_burst_left : std_logic_vector(8 downto 0) := (others => '0');
  signal fifo_radv_i  : std_logic := '0';


  signal irq_state : std_logic_vector(2 downto 0) := "000";

  signal local_rst : std_logic := '1';
  signal rst       : std_logic := '0';

  signal mcmd         : ocp_CmdT                                          := OCP_CMD_IDLE;
  signal maddr        : std_logic_vector(ADB3_OCP_ADDR_WIDTH-1 downto 0)  := (others => '0');
  signal mburstlength : std_logic_vector(ADB3_OCP_BURST_WIDTH-1 downto 0) := (others => '0');
  signal mrespaccept  : std_logic                                         := '0';
  signal mrespaccept_state  : std_logic                                         := '0';
  signal mtag         : std_logic_vector(2 downto 0) := "000";
  signal current_tag  : std_logic_vector(2 downto 0)                      := "000";

  signal inflight_data, inflight_q                                    : std_logic_vector(13 downto 0) := (others => '0');
  signal inflight_radv, inflight_wadv, inflight_nfull, inflight_nfull_r1, inflight_empty : std_logic                     := '0';
  signal inflight_q_r1                                    : std_logic_vector(13 downto 0) := (others => '0');
 
  signal this_irq   : std_logic := '0';
  signal this_tag   : std_logic_vector(2 downto 0) := (others => '0');
  signal this_burst : std_logic_vector(9 downto 0) := (others => '0');
  signal this_flush : std_logic := '0';
  signal last_in_this_burst : std_logic := '0';
  
  signal fifo_data_i : std_logic_vector(255 downto 0) := (others => '0');
  signal fifo_wadv_i : std_logic := '0';

  signal fifo_data_ii0 : std_logic_vector(255 downto 0) := (others => '0');
  signal fifo_wadv_ii0 : std_logic := '0';
  signal fifo_data_ii1 : std_logic_vector(255 downto 0) := (others => '0');
  signal fifo_wadv_ii1 : std_logic := '0';
  signal fifo_data_ii2 : std_logic_vector(255 downto 0) := (others => '0');
  signal fifo_wadv_ii2 : std_logic := '0';
  signal fifo_data_ii3 : std_logic_vector(255 downto 0) := (others => '0');
  signal fifo_wadv_ii3 : std_logic := '0';

  signal fifo_wadv_ii4 : std_logic := '0';
  signal fifo_wadv_ii5 : std_logic := '0';
  signal fifo_wadv_ii6 : std_logic := '0';
  signal fifo_wadv_ii7 : std_logic := '0';

  signal fifo_data_ii  : std_logic_vector(255 downto 0) := (others => '0');
  signal fifo_data_iii : std_logic_vector(255 downto 0) := (others => '0');
  signal fifo_wadv_iii : std_logic := '0';
  signal fifo_wadv_ooo : std_logic := '0';

  signal ooo_fifo_d, ooo_fifo_q0, ooo_fifo_q1,ooo_fifo_q2, ooo_fifo_q3                                       : std_logic_vector(258 downto 0) := (others => '0');
  signal ooo_fifo_radv0, ooo_fifo_wadv0,ooo_fifo_nfull0,ooo_fifo_nfull0_r1 : std_logic := '0';
  signal ooo_fifo_radv1, ooo_fifo_wadv1,ooo_fifo_nfull1,ooo_fifo_nfull1_r1 : std_logic := '0';
  signal ooo_fifo_radv2, ooo_fifo_wadv2,ooo_fifo_nfull2,ooo_fifo_nfull2_r1 : std_logic := '0';
  signal ooo_fifo_radv3, ooo_fifo_wadv3,ooo_fifo_nfull3,ooo_fifo_nfull3_r1 : std_logic := '0';
  signal ooo_fifo_radv4, ooo_fifo_wadv4,ooo_fifo_nfull4,ooo_fifo_nfull4_r1 : std_logic := '0';
  signal ooo_fifo_radv5, ooo_fifo_wadv5,ooo_fifo_nfull5,ooo_fifo_nfull5_r1 : std_logic := '0';
  signal ooo_fifo_radv6, ooo_fifo_wadv6,ooo_fifo_nfull6,ooo_fifo_nfull6_r1 : std_logic := '0';
  signal ooo_fifo_radv7, ooo_fifo_wadv7,ooo_fifo_nfull7,ooo_fifo_nfull7_r1 : std_logic := '0';
  signal ooo_fifo_empty0,ooo_fifo_empty1 : std_logic := '1';
  signal ooo_fifo_empty_q0,ooo_fifo_empty_q1 : std_logic := '1';
  signal ooo_fifo_empty0_reg,ooo_fifo_empty1_reg : std_logic := '1';
  signal ooo_fifo_radv0_reg,ooo_fifo_radv1_reg : std_logic := '0';
  signal ooo_fifo_empty2,ooo_fifo_empty3 : std_logic := '1';
  signal ooo_fifo_empty_q2,ooo_fifo_empty_q3 : std_logic := '1';
  signal ooo_fifo_empty2_reg,ooo_fifo_empty3_reg : std_logic := '1';
  signal ooo_fifo_radv2_reg,ooo_fifo_radv3_reg : std_logic := '0';
  signal ooo_fifo_empty4,ooo_fifo_empty5 : std_logic := '1';
  signal ooo_fifo_empty_q4,ooo_fifo_empty_q5 : std_logic := '1';
  signal ooo_fifo_empty4_reg,ooo_fifo_empty5_reg : std_logic := '1';
  signal ooo_fifo_radv4_reg,ooo_fifo_radv5_reg : std_logic := '0';
  signal ooo_fifo_empty6,ooo_fifo_empty7 : std_logic := '1';
  signal ooo_fifo_empty_q6,ooo_fifo_empty_q7 : std_logic := '1';
  signal ooo_fifo_empty6_reg,ooo_fifo_empty7_reg : std_logic := '1';
  signal ooo_fifo_radv6_reg,ooo_fifo_radv7_reg : std_logic := '0';
  
  signal ooo_fifo_flush0 : std_logic := '0';
  signal ooo_fifo_flush0_r1 : std_logic := '0';
  signal ooo_fifo_flush1 : std_logic := '0';
  signal ooo_fifo_flush1_r1 : std_logic := '0';
  signal ooo_fifo_empty_state0 : std_logic := '0';
  signal ooo_fifo_empty_state1 : std_logic := '0';
  signal ooo_fifo_flush2 : std_logic := '0';
  signal ooo_fifo_flush2_r1 : std_logic := '0';
  signal ooo_fifo_flush3 : std_logic := '0';
  signal ooo_fifo_flush3_r1 : std_logic := '0';
  signal ooo_fifo_empty_state2 : std_logic := '0';
  signal ooo_fifo_empty_state3 : std_logic := '0';
  signal ooo_fifo_flush4 : std_logic := '0';
  signal ooo_fifo_flush4_r1 : std_logic := '0';
  signal ooo_fifo_flush5 : std_logic := '0';
  signal ooo_fifo_flush5_r1 : std_logic := '0';
  signal ooo_fifo_empty_state4 : std_logic := '0';
  signal ooo_fifo_empty_state5 : std_logic := '0';
  signal ooo_fifo_flush6 : std_logic := '0';
  signal ooo_fifo_flush6_r1 : std_logic := '0';
  signal ooo_fifo_flush7 : std_logic := '0';
  signal ooo_fifo_flush7_r1 : std_logic := '0';
  signal ooo_fifo_empty_state6 : std_logic := '0';
  signal ooo_fifo_empty_state7 : std_logic := '0';
  
  signal ooo_fifo_wadv0_r1,ooo_fifo_wadv1_r1 : std_logic := '0';
  signal ooo_fifo_wadv2_r1,ooo_fifo_wadv3_r1 : std_logic := '0';
  signal ooo_fifo_wadv4_r1,ooo_fifo_wadv5_r1 : std_logic := '0';
  signal ooo_fifo_wadv6_r1,ooo_fifo_wadv7_r1 : std_logic := '0';
  signal ooo_fifo_d_r1 : std_logic_vector(255 downto 0) := (others => '0');

  signal ooo_wadv_pending : std_logic_vector(4 downto 0) := (others => '0');
 
  
  signal fifo_nfull_reg : std_logic := '0';

  -- Run Continouously if SIZE = 0
  --signal run_continuous : std_logic := '0';
  signal flush_counter : std_logic_vector(1 downto 0) := "00";


  
  signal state_aborting : std_logic := '0'; 

  signal abort1 : std_logic := '0';
  signal abort2 : std_logic := '0';
  signal cleanup1 : std_logic := '0';
  
  signal not_first_descriptor,last_descriptor : std_logic := '0';

  signal extra_read_sent : std_logic := '0';
  signal non_mod_read : std_logic := '0';
  signal r_non_mod_read : std_logic := '0';
--   attribute use_dsp48 : string;
--   attribute use_dsp48 of left1 : signal is "yes";
--   attribute use_dsp48 of left2 : signal is "yes";
--   attribute use_dsp48 of addr1 : signal is "yes";
--   attribute use_dsp48 of addr2 : signal is "yes";
--   attribute use_dsp48 of addr3 : signal is "yes";
--   attribute use_dsp48 of addr4 : signal is "yes";
  
  attribute shreg_extract : string;
  attribute shreg_extract of ooo_fifo_d_r1 : signal is "no";
  attribute shreg_extract of burst_desc_fifo_q_r1 : signal is "no";
  attribute shreg_extract of burst_desc_fifo_q_r2 : signal is "no";
  attribute shreg_extract of burst_desc_fifo_empty_r1 : signal is "no";
  attribute shreg_extract of burst_desc_fifo_empty_r2 : signal is "no";
  attribute shreg_extract of ooo_fifo_flush0_r1 : signal is "no";
  attribute shreg_extract of ooo_fifo_flush1_r1 : signal is "no";
  attribute shreg_extract of ooo_fifo_flush2_r1 : signal is "no";
  attribute shreg_extract of ooo_fifo_flush3_r1 : signal is "no";

  attribute max_fanout : string;
  attribute max_fanout of cleanup : signal is "16";
  attribute max_fanout of abort : signal is "16";
  attribute max_fanout of rst : signal is "16";

  signal first_burst_addr_mask : std_logic_vector(8 downto 0) := (others => '0');
  signal first_burst_length : std_logic_vector(8 downto 0) := (others => '0');


  component icon_rw
  PORT (
    CONTROL0 : INOUT STD_LOGIC_VECTOR(35 DOWNTO 0));
  end component;
  
  component ila_rw
  PORT (
    CONTROL : INOUT STD_LOGIC_VECTOR(35 DOWNTO 0);
    CLK : IN STD_LOGIC;
    TRIG0 : IN STD_LOGIC_VECTOR(31 DOWNTO 0));
  end component;

  signal control0 : std_logic_vector(35 downto 0);
  signal trig0 : std_logic_vector(31 downto 0);
  
  signal fsm_error1 : std_logic;
  signal fsm_error_count : std_logic_vector(10 downto 0);

  signal ocp_s2m_i : adb3w_ocp_s2mT;
  signal resp_fifo_in,resp_fifo_out : std_logic_vector(258 downto 0) := (others => '0');
  signal resp_fifo_nfull,resp_fifo_wadv,resp_fifo_empty,resp_fifo_radv : std_logic := '0';

  signal ooo_fifo_wadv_r1 : std_logic_vector(7 downto 0);
  signal ooo_fifo_radv : std_logic_vector(7 downto 0);
   signal ooo_fifo_radv_q : std_logic_vector(7 downto 0);
   signal ooo_fifo_radv_flush : std_logic_vector(7 downto 0) := (others => '0');
  signal ooo_fifo_nfull : std_logic_vector(7 downto 0);
  signal ooo_fifo_q_e3 : std_logic_vector(255 downto 0) := (others => '0');
  signal ooo_fifo_empty : std_logic_vector(7 downto 0);
  
  signal ooo_error_detected : std_logic := '0';

  signal comp_counter : std_logic_vector(15 downto 0) := (others => '0');
  signal active_counter : std_logic_vector(31 downto 0) := (others => '0');
  signal all_idle_sr : std_logic_vector(31 downto 0) := (others => '0');
  signal all_idle : std_logic;
  
begin  -- rtl

  
  
  ocp_m2s.Cmd         <= mcmd;
  ocp_m2s.Addr        <= maddr;
  ocp_m2s.BurstLength <= mburstlength;
  ocp_m2s.Data        <= (others => '0');
  ocp_m2s.DataByteEn  <= (others => '0');
  ocp_m2s.DataValid   <= '0';
  ocp_m2s.RespAccept  <= '1'; --mrespaccept;
  ocp_m2s.Tag         <= non_mod_read & "00" & mtag & "00";

  -- Generate a Power on Reset for FIFOs
  lrst0 : process (clk)
  begin  -- process lrst0
    if clk'event and clk = '1' then     -- rising clock edge
      local_rst <= '0';

      rst       <= local_rst or cleanup1;

      abort1 <= abort;
      abort2 <= abort1;
      cleanup1 <= cleanup;
    end if;
  end process lrst0;
 
  
  -- Small buffer to hold multiple descriptors
  gen_buf_desc0: if buffer_descriptors generate
    small_sfifo0 : small_sfifo
    generic map (
      width => 107)
    port map (
      clk    => clk,
      rst    => rst,
      wadv   => cmd_fifo_wadv,
      wdata  => cmd_fifo_d,
      nfull  => cmd_fifo_nfull,
      radv   => cmd_fifo_radv,
      rempty => cmd_fifo_empty,
      rdata  => cmd_fifo_q);
    
    cmd_fifo_empty_r2 <= cmd_fifo_empty_r1;
    process (clk) is
    begin  -- process
      if rising_edge(clk) then          -- rising clock edge
        worker_cmd_q <= slv_to_adb3_dma_worker_command(cmd_fifo_q);
        cmd_fifo_empty_r1 <= cmd_fifo_empty;
        
      end if;
    end process;
    
    
  end generate gen_buf_desc0;
  
  gen_no_buf_desc0: if not buffer_descriptors generate
    worker_cmd_q <= worker_cmd;
    cmd_fifo_empty <= not worker_cmd_valid;
    cmd_fifo_nfull <= '0';
  end generate gen_no_buf_desc0;
 

  gen_rfs0: if not registering_for_speed generate
    cmd_fifo_d <= adb3_dma_worker_command_to_slv(worker_cmd);
    cmd_fifo_wadv <= worker_cmd_valid;
  end generate gen_rfs0;
  
  
  -- FSM to break descriptor requests into OCP Burst Commands
  fsm0 : process (clk)
  begin  -- process fsm0
    if clk'event and clk = '1' then     -- rising clock edge

      if registering_for_speed then
        cmd_fifo_d <= adb3_dma_worker_command_to_slv(worker_cmd);
        cmd_fifo_wadv <= worker_cmd_valid;
        cmd_fifo_nfull_reg    <= cmd_fifo_nfull;
        worker_ready          <= (not cmd_fifo_nfull_reg) and (not state_aborting);
        burst_desc_fifo_nfull_reg <= burst_desc_fifo_nfull;
        burst_desc_fifo_ready <= not burst_desc_fifo_nfull_reg;
      else
        worker_ready          <= (not cmd_fifo_nfull) and (not state_aborting);
        burst_desc_fifo_ready <= not burst_desc_fifo_nfull;
      end if;     

      if state = S_ABORT then
        state_aborting <= '1';
      else
        state_aborting <= '0';
      end if;
      
      cmd_fifo_radv         <= '0';
      max_burst_reg1         <= max_burst;
      max_burst_reg2         <= max_burst;
      max_burst_reg3         <= max_burst;
      max_burst_reg4         <= max_burst;
      burst_desc_fifo_wadv  <= '0';
      irq_cmd               <= '0';
      flush_cmd             <= '0';
      extra_read_sent       <= '0';
      need_flush            <= not need_flush_calc(5);
      xfer_left_reg         <= xfer_left;
      left0_reg             <= left0;

      if cleanup1 = '1' then            -- prevent merging of flush counter
                                        -- across multiple DMA engines
        flush_counter <= (others => '0');
      else
        flush_counter <= flush_counter+1;
      end if;

      for i in 1 to 8 loop
        if max_burst_reg1(i) = '1' then
          for j in 8 downto i loop
            first_burst_addr_mask(j) <= '0';
          end loop;  -- j
          for j in i-1 downto 0 loop
            first_burst_addr_mask(j) <= '1';
          end loop;  -- j
        end if;          
      end loop;  -- i
      first_burst_length <= max_burst_reg2 - (first_burst_addr_mask and addr0(13 downto 5));

      if max_burst_reg1(8 downto 7) /= "00" then
        max_read_inflight <= "000001";
      elsif max_burst_reg1(6) = '1' then
        max_read_inflight <= "000011";
      elsif max_burst_reg1(5) = '1' then
        max_read_inflight <= "000111";
      elsif max_burst_reg1(4) = '1' then
        max_read_inflight <= "001111";
      elsif max_burst_reg1(3) = '1' then
        max_read_inflight <= "011111";
      else
        max_read_inflight <= "111111";
      end if;
      
      
      case state is
        when IDLE =>
          if abort = '1' then
            state <= S_ABORT;
            left0 <= (others => '-');
          else
            if cmd_fifo_empty_r2 = '0' then
              current_cmd   <= worker_cmd_q;
              cmd_fifo_radv <= '1';
              addr0         <= '0' & worker_cmd_q.start_addr(12 downto 0);
              left0         <= '0' & worker_cmd_q.xfer_length(12 downto 0);
              state         <= IS_ADDR_ALIGNED;
              left_inc      <= '0';
              inc_addr      <= not worker_cmd_q.fixed_addr;
              last_descriptor <= worker_cmd_q.last_descriptor;
              if worker_cmd_q.last_descriptor = '1'  then
                if need_extra_write = '1' then
                  need_flush_calc <= "011111";
                else
                  if worker_cmd_q.xfer_length(4 downto 0) = "00000" or worker_cmd_q.start_addr(4 downto 0) = "00000" then
                    need_flush_calc <= (others => '1');
                  else
                    need_flush_calc <= "011111";
                  end if;
                end if;
              else
                need_flush_calc <= (others => '1');
              end if;
            end if;
          end if;
        when IS_ADDR_ALIGNED =>
          if abort = '1' then
            state <= S_ABORT;
            left0 <= (others => '-');
          else
            if OR_reduce(addr(4 downto 0)) = '1' then
              if not_first_descriptor = '0' then
                addr0    <= addr0(13 downto 5) & "00000";
                left0    <= left0 + EXT(addr(4 downto 0), 14);
                left_inc <= '1';
              else
                addr0    <= (addr0(13 downto 5) + 1) & "00000";
                if (left0 + EXT(addr(4 downto 0), 14)) >= 32 or left1 /= "00000000000000" or left2 /= "00000000000000" then
                  left0    <= left0 + EXT(addr(4 downto 0), 14) - 32;
                else
                  left0 <= (others => '0');
                end if;
                left_inc <= '0';
              end if;            
             
             
            end if;
            state          <= FIRST_BURST_WAIT;
            not_first_descriptor <= not last_descriptor;
            --run_continuous <= not OR_reduce(current_cmd.xfer_length);
          end if;
          when FIRST_BURST_WAIT =>
          if abort = '1' then
            state <= S_ABORT;
            left0 <= (others => '-');
          else
            state          <= IS_FIRST_BURST_LEFT;
            left_inc <= '0';
            if left0(13) = '1' then
              left0(13) <= '0';
            end if;
            if addr0(13) = '1' then
              addr0(13) <= '0';
            end if;
          end if;
        when IS_FIRST_BURST_LEFT =>
          if abort = '1' then
            state <= S_ABORT;
            left0 <= (others => '-');
          else
            if addr_valid = '0' or left_valid = '0' then
              if addr0(13) = '1' then
                addr0(13) <= '0';
              end if;
              if left0(13) = '1' then
                left0(13) <= '0';
              end if;
            else
              left_inc <= '0';
              --if lots_left = '1' or run_continuous = '1' or  xfer_left(13 downto 4) >= max_burst_reg1 then
              if lots_left = '1' or  xfer_left(13 downto 5) >= first_burst_length then
                state <= GEN_FIRST_BURST;
              else
                state <= IS_SHORT_BURST_LEFT;
              end if;
            end if;
          end if;
        when GEN_FIRST_BURST =>
          if abort = '1' then
            state <= S_ABORT;
            left0 <= (others => '-');
          else
            if burst_desc_fifo_ready = '1' then
              if inc_addr = '1' then
                addr0 <= (addr0(13 downto 5) & "00000") + (first_burst_length & "00000");
              end if;
              left0                <= left0_reg - (first_burst_length & "00000");
              size                 <= first_burst_length;
              state                <= IS_BURST_LEFT;
              addr_reg(addr_bits-1 downto 0)             <= addr(addr_bits-1 downto 0);
              burst_desc_fifo_wadv <= '1';
            end if;
          end if;        
        when IS_BURST_LEFT =>
          if abort = '1' then
            state <= S_ABORT;
            left0 <= (others => '-');
          else
            if addr_valid = '0' or left_valid = '0' then
              if addr0(13) = '1' then
                addr0(13) <= '0';
              end if;
              if left0(13) = '1' then
                left0(13) <= '0';
              end if;
            else
              left_inc <= '0';
              --if lots_left = '1' or run_continuous = '1' or  xfer_left(13 downto 4) >= max_burst_reg1 then
              if lots_left = '1' or  xfer_left(13 downto 5) >= max_burst_reg1 then
                state <= GEN_BURST;
              else
                state <= IS_SHORT_BURST_LEFT;
              end if;
            end if;
          end if;
        when GEN_BURST =>
          if abort = '1' then
            state <= S_ABORT;
            left0 <= (others => '-');
          else
            if burst_desc_fifo_ready = '1' then
              if inc_addr = '1' then
                addr0 <= (addr0(13 downto 5) & "00000") + (max_burst_reg2 & "00000");
              end if;
              left0                <= left0_reg - (max_burst_reg3 & "00000");
              size                 <= max_burst_reg4;
              state                <= IS_BURST_LEFT;
              addr_reg(addr_bits-1 downto 0)             <= addr(addr_bits-1 downto 0);
              burst_desc_fifo_wadv <= '1';
            end if;
          end if;
        when IS_SHORT_BURST_LEFT =>
          if abort = '1' then
            state <= S_ABORT;
            left0 <= (others => '-');
          else
            if addr_valid = '0' or left_valid = '0' then
              if addr0(13) = '1' then
                addr0(13) <= '0';
              end if;
              if left0(13) = '1' then
                left0(13) <= '0';
              end if;
            else
              if OR_reduce(xfer_left(13 downto 0)) = '1' then
                state <= GEN_SHORT_BURST;
              else
                state <= GEN_IRQ;
              end if;
            end if;
          end if;
        when GEN_SHORT_BURST =>
          if abort = '1' then
            state <= S_ABORT;
            left0 <= (others => '-');
          else
            if burst_desc_fifo_ready = '1' then
              if OR_reduce(left0_reg(4 downto 0)) = '1' then
                size <= xfer_left_reg(13 downto 5)+1;
                extra_read_sent <= last_descriptor;
              else
                size <= xfer_left_reg(13 downto 5);
              end if;
              left0                <= (others => '0');
              state                <= GEN_IRQ;
              addr_reg(addr_bits-1 downto 0)             <= addr(addr_bits-1 downto 0);
              burst_desc_fifo_wadv <= '1';
            end if;
          end if;
        when GEN_IRQ =>
          if abort = '1' then
            state <= S_ABORT;
            left0 <= (others => '-');
          else
            if current_cmd.irq_on_completion = '1' or need_flush = '1' then
              if burst_desc_fifo_ready = '1' then
                irq_cmd              <= current_cmd.irq_on_completion;
                flush_cmd            <= need_flush;
                size                 <= (others => '0');
                burst_desc_fifo_wadv <= '1';
                state                <= IDLE;
              end if;
            else
              state <= IDLE;
            end if;
          end if;
        when S_ABORT =>
          if flush_counter = "00" and cmd_fifo_empty_r2 = '0' then
            cmd_fifo_radv <= '1';
          else
            cmd_fifo_radv <= '0';
          end if;
          worker_ready         <= '0';
          burst_desc_fifo_wadv <= '0';
          if cleanup = '1' then
            state <= IDLE;
          end if;
          --run_continuous <= '0';
          left0 <= (others => '0');
          not_first_descriptor <= '0';
        when others =>
          state <= IDLE;
      end case;

      -- Efficiently handle 64 bit arithmetic without 64 bit long carry chains
      if state = IDLE then
        addr1 <= '0' & worker_cmd_q.start_addr(25 downto 13);
        addr2 <= '0' & worker_cmd_q.start_addr(38 downto 26);
        if addr_bits = 64 then
          addr3 <= '0' & worker_cmd_q.start_addr(51 downto 39);
          addr4 <= "00" & worker_cmd_q.start_addr(63 downto 52);
        end if;
        left1 <= '0' & worker_cmd_q.xfer_length(25 downto 13);
        left2 <= '0' & worker_cmd_q.xfer_length(38 downto 26);
      else
        if addr0(13) = '1' then
          addr1 <= addr1+1;
        end if;
        if addr1(13) = '1' then
          addr2     <= addr2+1;
          addr1(13) <= '0';
        end if;
        if addr_bits = 64 then
          if addr2(13) = '1' then
            addr3     <= addr3+1;
            addr2(13) <= '0';
          end if;
          if addr3(13) = '1' then
            addr4     <= addr4+1;
            addr3(13) <= '0';
          end if;
        end if;
        
        if left_inc = '0' then
          if left0(13) = '1' then
            left1 <= left1-1;
          end if;
          if left1(13) = '1' then
            left2     <= left2-1;
            left1(13) <= '0';
          end if;
        else
          if left0(13) = '1' then
            left1 <= left1+1;
          end if;
          if left1(13) = '1' then
            left2     <= left2+1;
            left1(13) <= '0';
          end if;
        end if;
        if left_bits = 39 then
          lots_left <= OR_reduce(left1(12 downto 0)) or OR_reduce(left2(12 downto 0));
        elsif left_bits = 26 then
          lots_left <= OR_reduce(left1(12 downto 0));
        else
          lots_left <= '0';
        end if;
        
      end if;
      
    end if;
  end process fsm0;

  gen_addr_bits64: if addr_bits = 64 generate
    addr_valid <= not (addr0(13) or addr1(13) or addr2(13) or addr3(13));
    addr       <= addr4(11 downto 0) & addr3(12 downto 0) & addr2(12 downto 0) & addr1(12 downto 0) & addr0(12 downto 0);
  end generate gen_addr_bits64;

  gen_addr_bits39: if addr_bits = 39 generate
    addr_valid <= not (addr0(13) or addr1(13));
    addr       <= EXT(addr2(12 downto 0) & addr1(12 downto 0) & addr0(12 downto 0),64);
  end generate gen_addr_bits39;

  gen_left_bits39: if left_bits = 39 generate
    xfer_left  <= left2(12 downto 0) & left1(12 downto 0) & left0(12 downto 0);
    left_valid <= not (left0(13) or left1(13));
  end generate gen_left_bits39;

  gen_left_bits26: if left_bits = 26 generate
    xfer_left  <= left1(12 downto 0) & left0(12 downto 0);
    left_valid <= not (left0(13));
  end generate gen_left_bits26;

  gen_left_bits14: if left_bits = 14 generate
    xfer_left  <= EXT(left0(12 downto 0),39);
    left_valid <= '1';
  end generate gen_left_bits14;
  

  burst_desc_fifo_d <= extra_read_sent & flush_cmd & irq_cmd & size & addr_reg;

  burst_desc_fifo0 : small_sfifo
    generic map (
      width => 76)
    port map (
      clk    => clk,
      rst    => rst,
      wadv   => burst_desc_fifo_wadv,
      wdata  => burst_desc_fifo_d,
      nfull  => burst_desc_fifo_nfull,
      radv   => burst_desc_fifo_radv,
      rempty => burst_desc_fifo_empty,
      rdata  => burst_desc_fifo_q);

 
  process (clk) is
  begin  -- process
    if rising_edge(clk) then            -- rising clock edge
      burst_desc_fifo_q_r1 <= burst_desc_fifo_q;
      burst_desc_fifo_empty_r1 <= burst_desc_fifo_empty or burst_desc_fifo_radv;
      burst_desc_fifo_q_r2 <= burst_desc_fifo_q_r1;
      burst_desc_fifo_empty_r2 <= burst_desc_fifo_empty_r1;
    end if;
  end process;

  -- FSM for generating OCP read requests from burst descriptions
  fsm1 : process (clk)
  begin  -- process fsm1
    if clk'event and clk = '1' then
      inflight_wadv        <= '0';
      inflight_data        <= (others => '0');
      fifo_nfull_reg       <= fifo_nfull;
      burst_desc_fifo_radv <= '0';

      maddr <= (others => '0');
      maddr(addr_bits-1 downto 0)         <= r_addr(addr_bits-1 downto 0);
      non_mod_read <= r_non_mod_read;
      case req_state is
        when REQ_IDLE =>
          if abort1 = '1' then
            req_state <= REQ_ABORT;
          elsif burst_desc_fifo_empty_r2 = '0' then
            r_flush_cmd          <= burst_desc_fifo_q_r2(74);
            r_irq_cmd            <= burst_desc_fifo_q_r2(73);
            r_size               <= burst_desc_fifo_q_r2(72 downto 64);          
--              r_addr            else
            --maddr(addr_bits-1 downto 0) <= burst_desc_fifo_q_r2(addr_bits-1 downto 0);
            r_addr(addr_bits-1 downto 0) <= burst_desc_fifo_q_r2(addr_bits-1 downto 0);
            r_non_mod_read <=  burst_desc_fifo_q_r2(75);
            --mburstlength  <= "00" & burst_desc_fifo_q(73 downto 64);
            burst_desc_fifo_radv <= '1';
            req_state            <= REQ_READ_BURST_DESC;
          end if;
          mcmd         <= OCP_CMD_IDLE;
          --maddr        <= (others => '0');
          mburstlength <= (others => '0');
        when REQ_READ_BURST_DESC =>
          mcmd         <= OCP_CMD_IDLE;
          if abort1 = '1' then
            req_state <= REQ_ABORT;
          else
            if r_irq_cmd = '1' or r_flush_cmd = '1' then
              req_state <= REQ_GEN_IRQ;
            elsif fifo_nfull_reg = '0' and inflight_nfull_r1 = '0' and read_inflight_full = '0' then
              req_state     <= REQ_GEN_READ_CMD;
              mcmd          <= OCP_CMD_READ;             
              mburstlength  <= "000" & r_size;            
              mtag          <= current_tag;
              inflight_wadv <= '1';
              inflight_data <= '0' & current_tag & '0' & r_size;
              current_tag   <= current_tag+1;
            end if;
          end if;
        when REQ_GEN_IRQ =>
          if abort1 = '1' then
            req_state <= REQ_ABORT;
          else
            req_state     <= REQ_IDLE;
            inflight_wadv <= '1';
            inflight_data <= '1' & "000" & "00000000" & r_flush_cmd & r_irq_cmd;
          end if;
          
        when REQ_GEN_READ_CMD =>
          if abort1 = '1' then
            req_state <= REQ_ABORT;
          else
            if ocp_s2m.CmdAccept = '1' then
              mcmd      <= OCP_CMD_IDLE;
              req_state <= REQ_IDLE;
            end if;
          end if;
        when REQ_ABORT =>
          mcmd         <= OCP_CMD_IDLE;
--          maddr        <= (others => '0');
          mburstlength <= (others => '0');
          if burst_desc_fifo_empty_r2 = '0' and flush_counter = "00" then
            burst_desc_fifo_radv <= '1';
          end if;
          if cleanup = '1' then
            req_state <= REQ_IDLE;
          end if;
          current_tag <= (others => '0');
        when others =>
          req_state <= REQ_IDLE;
      end case;

    end if;
  end process fsm1;

  limit_read_inflight: process (clk) is
  begin  -- process limit_read_inflight
    if rising_edge(clk) then            -- rising clock edge
      if cleanup = '1' then
        read_inflight_count <= (others => '0');
      elsif inflight_wadv = '1' and inflight_radv = '0'  then
        read_inflight_count <= read_inflight_count +1;
      elsif inflight_wadv = '0' and inflight_radv = '1'  then
        read_inflight_count <= read_inflight_count -1;
      end if;
      if max_read_inflight = read_inflight_count then
        read_inflight_full <= '1';
      else
        read_inflight_full <= '0';
      end if;
    end if;
  end process limit_read_inflight;

  
  inflight_fifo0 : small_sfifo
    generic map (
      width => 14)
    port map (
      clk    => clk,
      rst    => rst,
      wadv   => inflight_wadv,
      wdata  => inflight_data,
      nfull  => inflight_nfull,
      radv   => inflight_radv,
      rempty => inflight_empty,
      rdata  => inflight_q);


  fsm2 : process (clk)
  begin  -- process fsm2
    if clk'event and clk = '1' then     -- rising clock edge
      inflight_radv <= '0';
  --    fifo_wadv_i   <= '0';
  --    fifo_wadv_ii0   <= '0';
  --    fifo_wadv_ii1   <= '0';
  --    fifo_wadv_ii2   <= '0';
  --    fifo_wadv_ii3   <= '0';
  --    fifo_wadv_ii4   <= '0';
  --    fifo_wadv_ii5   <= '0';
  --    fifo_wadv_ii6   <= '0';
  --    fifo_wadv_ii7   <= '0';
  --    ooo_fifo_flush0 <= '0';
  --    ooo_fifo_flush0_r1 <= ooo_fifo_flush0;
  --    ooo_fifo_flush1 <= '0';
  --    ooo_fifo_flush1_r1 <= ooo_fifo_flush1;
  --    ooo_fifo_flush2 <= '0';
  --    ooo_fifo_flush2_r1 <= ooo_fifo_flush2;
  --    ooo_fifo_flush3 <= '0';
  --    ooo_fifo_flush3_r1 <= ooo_fifo_flush3;
  --    ooo_fifo_flush4 <= '0';
  --    ooo_fifo_flush4_r1 <= ooo_fifo_flush4;
  --    ooo_fifo_flush5 <= '0';
  --    ooo_fifo_flush5_r1 <= ooo_fifo_flush5;
  --    ooo_fifo_flush6 <= '0';
  --    ooo_fifo_flush6_r1 <= ooo_fifo_flush6;
  --    ooo_fifo_flush7 <= '0';
  --    ooo_fifo_flush7_r1 <= ooo_fifo_flush7;
  --    ooo_fifo_wadv0 <= '0';
  --    ooo_fifo_wadv1 <= '0';
  --    ooo_fifo_wadv2 <= '0';
  --    ooo_fifo_wadv3 <= '0';
  --    ooo_fifo_wadv4 <= '0';
  --    ooo_fifo_wadv5 <= '0';
  --    ooo_fifo_wadv6 <= '0';
  --    ooo_fifo_wadv7 <= '0';
  --    ooo_fifo_nfull0_r1 <= ooo_fifo_nfull0;
  --    ooo_fifo_nfull1_r1 <= ooo_fifo_nfull1;
  --    ooo_fifo_nfull2_r1 <= ooo_fifo_nfull2;
  --    ooo_fifo_nfull3_r1 <= ooo_fifo_nfull3;
  --    ooo_fifo_nfull4_r1 <= ooo_fifo_nfull4;
  --    ooo_fifo_nfull5_r1 <= ooo_fifo_nfull5;
  --    ooo_fifo_nfull6_r1 <= ooo_fifo_nfull6;
  --    ooo_fifo_nfull7_r1 <= ooo_fifo_nfull7;
      inflight_q_r1 <= inflight_q;
      inflight_nfull_r1 <= inflight_nfull;
      
  --    ooo_fifo_wadv0_r1 <= ooo_fifo_wadv0;
  --    ooo_fifo_wadv1_r1 <= ooo_fifo_wadv1;
  --    ooo_fifo_wadv2_r1 <= ooo_fifo_wadv2;
  --    ooo_fifo_wadv3_r1 <= ooo_fifo_wadv3;
  --    ooo_fifo_wadv4_r1 <= ooo_fifo_wadv4;
  --    ooo_fifo_wadv5_r1 <= ooo_fifo_wadv5;
  --    ooo_fifo_wadv6_r1 <= ooo_fifo_wadv6;
  --    ooo_fifo_wadv7_r1 <= ooo_fifo_wadv7;
  --    ooo_wadv_pending <=  ooo_wadv_pending(3 downto 0) & (ooo_fifo_wadv0 or ooo_fifo_wadv1 or ooo_fifo_wadv2 or ooo_fifo_wadv3 or ooo_fifo_wadv4 or ooo_fifo_wadv5 or ooo_fifo_wadv6 or ooo_fifo_wadv7);

      
  --    ooo_fifo_d_r1 <= ooo_fifo_d;
  --    ooo_fifo_empty_state0 <= '0';
  --    ooo_fifo_empty_state1 <= '0';
  --    ooo_fifo_empty_state2 <= '0';
  --    ooo_fifo_empty_state3 <= '0';
  --    ooo_fifo_empty_state4 <= '0';
  --    ooo_fifo_empty_state5 <= '0';
  --    ooo_fifo_empty_state6 <= '0';
  --    ooo_fifo_empty_state7 <= '0';
  --    --ooo_fifo_d    <= ocp_s2m.Tag(4 downto 2) & ocp_s2m.Data;
  --    ooo_fifo_d    <= ocp_s2m_tag_r1(4 downto 2) & ocp_s2m_data_r1;
      
  --    fifo_data_i <= ocp_s2m_Data_r1;
  --    --fifo_data_ii0 <= ooo_fifo_q0(255 downto 0);
  --    --fifo_data_ii1 <= ooo_fifo_q1(255 downto 0);
  --    --fifo_data_ii2 <= ooo_fifo_q2(255 downto 0);
  --    --fifo_data_ii3 <= ooo_fifo_q3(255 downto 0);
      
      fifo_wadv_ii0 <= '0';
      ooo_fifo_radv_flush <= (others => '0');
      
      case comp_state is
      when COMP_IDLE =>
          mrespaccept_state <= '0';
          if abort2 = '1' then
            comp_state <= COMP_ABORT;
            mrespaccept_state <= '1';
          elsif inflight_empty = '0' then
            inflight_radv <= '1';
            this_irq      <= inflight_q(13) and inflight_q(0);
            this_flush    <= inflight_q(13) and inflight_q(1);
            this_tag      <= inflight_q(12 downto 10);
            --this_burst    <= inflight_q(9 downto 0);
            comp_state    <= COMP_READ_IFF;
            mrespaccept_state <= '0';
          end if;
          last_in_this_burst <= '0';
        when COMP_READ_IFF =>
          last_in_this_burst <= '0';
          this_burst    <= inflight_q_r1(9 downto 0);
          if inflight_q_r1(9 downto 0) = "0000000001" then
            last_in_this_burst <= '1';
          end if;
          if abort2 = '1' then
            comp_state <= COMP_ABORT;
             mrespaccept_state <= '1';
          elsif this_irq = '1' then
            comp_state <= COMP_GEN_IRQ;
             mrespaccept_state <= '0';
          elsif this_flush = '1' then
            comp_state <= COMP_GEN_FLUSH;
             mrespaccept_state <= '0';        
          else
            comp_state <= COMP_ACCEPT_DATA;
            mrespaccept_state <= '1';
          end if;
        when COMP_GEN_IRQ =>
          if abort2 = '1' then
            comp_state <= COMP_ABORT;
            mrespaccept_state <= '1';
          else
            if this_flush = '1' then
              comp_state <= COMP_GEN_FLUSH;
              mrespaccept_state <= '0';
            else
              comp_state <= COMP_IDLE;
              mrespaccept_state <= '0';
            end if;
          end if;
        when COMP_GEN_FLUSH =>
          if abort2 = '1' then
            comp_state <= COMP_ABORT;
            mrespaccept_state <= '1';
          else
            comp_state <= COMP_IDLE;
            mrespaccept_state <= '0';
            fifo_wadv_ii0 <= '1';
            
          end if;        
        when COMP_ACCEPT_DATA =>
          if abort2 = '1' then
            comp_state <= COMP_ABORT;
            mrespaccept_state <= '1';
          elsif ooo_fifo_empty(CONV_INTEGER(this_tag)) = '0' then          
            if last_in_this_burst = '1' then
              if inflight_empty = '0' then
                inflight_radv <= '1';
                this_irq      <= inflight_q(13) and inflight_q(0);
                this_flush    <= inflight_q(13) and inflight_q(1);
                this_tag      <= inflight_q(12 downto 10);
                --this_burst    <= inflight_q(9 downto 0);
                comp_state    <= COMP_READ_IFF;
                mrespaccept_state <= '0';
                last_in_this_burst <= '0';
              else
                comp_state <= COMP_IDLE;
                mrespaccept_state <= '0';             
              end if;
            end if;
            this_burst <= this_burst-1;
            if this_burst = "0000000010" then
              last_in_this_burst <= '1';
            end if;          
          else
            if this_burst = "0000000001" then
              last_in_this_burst <= '1';
            end if; 
          end if;
        when COMP_ABORT =>
          if inflight_empty = '0' and flush_counter = "00" then
            inflight_radv <= '1';
          end if;
          if flush_counter = "00" then
            ooo_fifo_radv_flush <= not ooo_fifo_empty;
          end if;
          if cleanup1 = '1' then
            comp_state <= COMP_IDLE;
            mrespaccept_state <= '0';
          end if;
        when COMP_ERROR =>
          if inflight_empty = '0' and flush_counter = "00" then
            inflight_radv <= '1';
          end if;
          if abort2 = '1' then
            comp_state <= COMP_ABORT;
            mrespaccept_state <= '1';
          elsif cleanup = '1' then
            comp_state <= COMP_IDLE;
            mrespaccept_state <= '0';
          end if;
        when others =>
          comp_state <= COMP_IDLE;
          mrespaccept_state <= '0';
      end case;

      --if irq_ack = '1' or comp_state = COMP_ABORT then
      if cleanup1 = '1' then
        irq_state <= "000";
      elsif irq_ack = '1' and irq_state /= "000" then
        irq_state <= irq_state -1;
      elsif comp_state = COMP_GEN_IRQ then
        irq_state <= irq_state+1;
      end if;
      

      --comp_state_r1 <= comp_state;
      --ocp_s2m_resp_r1 <= ocp_s2m_i.Resp;
      --ocp_s2m_tag_r1 <= ocp_s2m_i.Tag;
      --ocp_s2m_data_r1 <= ocp_s2m_i.Data;
      --ooo_error_detected <= '0';
      --if comp_state_r1 = COMP_ACCEPT_DATA  and ocp_s2m_resp_r1 = OCP_RESP_VALID then
      --  if (ocp_s2m_tag_r1(4 downto 2) /= this_tag) and check_tags then
      --    if (ocp_s2m_tag_r1(4 downto 2) = "000" and ooo_fifo_nfull0_r1 = '1') or (ocp_s2m_tag_r1(4 downto 2) = "001" and ooo_fifo_nfull1_r1 = '1') or (ocp_s2m_tag_r1(4 downto 2) = "010" and ooo_fifo_nfull2_r1 = '1') or (ocp_s2m_tag_r1(4 downto 2) = "011" and ooo_fifo_nfull3_r1 = '1') or (ocp_s2m_tag_r1(4 downto 2) = "100" and ooo_fifo_nfull4_r1 = '1') or (ocp_s2m_tag_r1(4 downto 2) = "101" and ooo_fifo_nfull5_r1 = '1') or (ocp_s2m_tag_r1(4 downto 2) = "110" and ooo_fifo_nfull6_r1 = '1') or (ocp_s2m_tag_r1(4 downto 2) = "111" and ooo_fifo_nfull7_r1 = '1')then
      --      ooo_error_detected <= '1';
    
      --    else
      --      ooo_fifo_wadv0 <= (not ocp_s2m_Tag_r1(2)) and (not ocp_s2m_Tag_r1(3)) and (not ocp_s2m_Tag_r1(4));
      --      ooo_fifo_wadv1 <= ocp_s2m_Tag_r1(2) and (not ocp_s2m_Tag_r1(3)) and (not ocp_s2m_Tag_r1(4));
      --      ooo_fifo_wadv2 <= (not ocp_s2m_Tag_r1(2)) and ocp_s2m_Tag_r1(3) and (not ocp_s2m_Tag_r1(4));
      --      ooo_fifo_wadv3 <= ocp_s2m_Tag_r1(2) and ocp_s2m_Tag_r1(3) and (not ocp_s2m_Tag_r1(4));
      --      ooo_fifo_wadv4 <= (not ocp_s2m_Tag_r1(2)) and (not ocp_s2m_Tag_r1(3)) and (ocp_s2m_Tag_r1(4));
      --      ooo_fifo_wadv5 <= ocp_s2m_Tag_r1(2) and (not ocp_s2m_Tag_r1(3)) and (ocp_s2m_Tag_r1(4));
      --      ooo_fifo_wadv6 <= (not ocp_s2m_Tag_r1(2)) and ocp_s2m_Tag_r1(3) and (ocp_s2m_Tag_r1(4));
      --      ooo_fifo_wadv7 <= ocp_s2m_Tag_r1(2) and ocp_s2m_Tag_r1(3) and (ocp_s2m_Tag_r1(4));
      --    end if;
      --  else
      --    fifo_wadv_i <= '1';           
      --  end if;       
      --end if;
      
      --if registering_for_speed then
      --  --if fifo_wadv_i = '1' then
      --  --  fifo_data_iii <= fifo_data_i;
      --  --elsif fifo_wadv_ii0 = '1' then       
      --  --  fifo_data_iii <= fifo_data_ii0;
      --  --elsif fifo_wadv_ii1 = '1' then       
      --  --  fifo_data_iii <= fifo_data_ii1;
      --  --elsif fifo_wadv_ii2 = '1' then       
      --  --  fifo_data_iii <= fifo_data_ii2;
      --  --else
      --  --  fifo_data_iii <= fifo_data_ii3;
      --  --end if;
      --  fifo_data_ii <= fifo_data_i;
      --  fifo_data_iii <= fifo_data_ii;
      --  fifo_wadv_iii <= fifo_wadv_i or fifo_wadv_ii0 or fifo_wadv_ii1 or fifo_wadv_ii2 or fifo_wadv_ii3 or fifo_wadv_ii4 or fifo_wadv_ii5 or fifo_wadv_ii6 or fifo_wadv_ii7;
      --  fifo_wadv_ooo <= fifo_wadv_ii0 or fifo_wadv_ii1 or fifo_wadv_ii2 or fifo_wadv_ii3 or fifo_wadv_ii4 or fifo_wadv_ii5 or fifo_wadv_ii6 or fifo_wadv_ii7;
      --  --fifo_data <= fifo_data_iii;
      --  fifo_wadv <= fifo_wadv_iii;
      --end if;
      --ooo_fifo_empty0_reg <= ooo_fifo_empty0;
      --ooo_fifo_radv0_reg  <=  ooo_fifo_radv0;
      --ooo_fifo_empty1_reg <= ooo_fifo_empty1;
      --ooo_fifo_radv1_reg  <=  ooo_fifo_radv1;
      --ooo_fifo_empty2_reg <= ooo_fifo_empty2;
      --ooo_fifo_radv2_reg  <=  ooo_fifo_radv2;
      --ooo_fifo_empty3_reg <= ooo_fifo_empty3;
      --ooo_fifo_radv3_reg  <=  ooo_fifo_radv3;
      --ooo_fifo_empty4_reg <= ooo_fifo_empty4;
      --ooo_fifo_radv4_reg  <=  ooo_fifo_radv4;
      --ooo_fifo_empty5_reg <= ooo_fifo_empty5;
      --ooo_fifo_radv5_reg  <=  ooo_fifo_radv5;
      --ooo_fifo_empty6_reg <= ooo_fifo_empty6;
      --ooo_fifo_radv6_reg  <=  ooo_fifo_radv6;
      --ooo_fifo_empty7_reg <= ooo_fifo_empty7;
      --ooo_fifo_radv7_reg  <=  ooo_fifo_radv7;
    end if;
  end process fsm2;

  --fifo_data <= fifo_data_iii when fifo_wadv_ooo = '0' else ooo_fifo_q_e3(255 downto 0);
  
  
  --gen_rfs1: if not registering_for_speed generate
  --  fifo_data <= fifo_data_i when fifo_wadv_i = '1' else
  --               fifo_data_ii0 when fifo_wadv_ii0 = '1' else
  --               fifo_data_ii1 when fifo_wadv_ii1 = '1' else
  --               fifo_data_ii2 when fifo_wadv_ii2 = '1' else
  --               fifo_data_ii3;
  --  fifo_wadv <= fifo_wadv_i or fifo_wadv_ii0 or fifo_wadv_ii1 or fifo_wadv_ii2 or fifo_wadv_ii3;      
  --end generate gen_rfs1;
  
--   mrespaccept <= '1' when comp_state = COMP_ERROR
--                  or comp_state = COMP_ABORT
--                  or comp_state COMP_ACCEPT_DATA else '0';
  --mrespaccept <= mrespaccept_state;
  --ooo_fifo_empty_q0 <= ooo_fifo_empty0_reg or ooo_fifo_radv0_reg;
  --ooo_fifo_empty_q1 <= ooo_fifo_empty1_reg or ooo_fifo_radv1_reg;
  --ooo_fifo_empty_q2 <= ooo_fifo_empty2_reg or ooo_fifo_radv2_reg;
  --ooo_fifo_empty_q3 <= ooo_fifo_empty3_reg or ooo_fifo_radv3_reg;
  --ooo_fifo_empty_q4 <= ooo_fifo_empty4_reg or ooo_fifo_radv4_reg;
  --ooo_fifo_empty_q5 <= ooo_fifo_empty5_reg or ooo_fifo_radv5_reg;
  --ooo_fifo_empty_q6 <= ooo_fifo_empty6_reg or ooo_fifo_radv6_reg;
  --ooo_fifo_empty_q7 <= ooo_fifo_empty7_reg or ooo_fifo_radv7_reg;
  ----ooo_fifo_radv <= not ooo_fifo_empty_q when comp_state = COMP_EMPTY_OOO_FIFO else ooo_fifo_flush;
  --ooo_fifo_radv0 <= ((not ooo_fifo_empty_q0) and ooo_fifo_empty_state0)  or ooo_fifo_flush0_r1;
  --ooo_fifo_radv1 <= ((not ooo_fifo_empty_q1) and ooo_fifo_empty_state1)  or ooo_fifo_flush1_r1;
  --ooo_fifo_radv2 <= ((not ooo_fifo_empty_q2) and ooo_fifo_empty_state2)  or ooo_fifo_flush2_r1;
  --ooo_fifo_radv3 <= ((not ooo_fifo_empty_q3) and ooo_fifo_empty_state3)  or ooo_fifo_flush3_r1;
  --ooo_fifo_radv4 <= ((not ooo_fifo_empty_q4) and ooo_fifo_empty_state4)  or ooo_fifo_flush4_r1;
  --ooo_fifo_radv5 <= ((not ooo_fifo_empty_q5) and ooo_fifo_empty_state5)  or ooo_fifo_flush5_r1;
  --ooo_fifo_radv6 <= ((not ooo_fifo_empty_q6) and ooo_fifo_empty_state6)  or ooo_fifo_flush6_r1;
  --ooo_fifo_radv7 <= ((not ooo_fifo_empty_q7) and ooo_fifo_empty_state7)  or ooo_fifo_flush7_r1;

  fsm3: process (clk) is
  begin  -- process fsm2
    if rising_edge(clk) then            -- rising clock edge
      ooo_fifo_d_r1 <= ocp_s2m.Data;
      if ocp_s2m.Resp = OCP_RESP_VALID then
         case ocp_s2m.Tag(4 downto 2) is
        when "000" => ooo_fifo_wadv_r1 <= "00000001";
        when "001" => ooo_fifo_wadv_r1 <= "00000010";
        when "010" => ooo_fifo_wadv_r1 <= "00000100";
        when "011" => ooo_fifo_wadv_r1 <= "00001000";
        when "100" => ooo_fifo_wadv_r1 <= "00010000";
        when "101" => ooo_fifo_wadv_r1 <= "00100000";
        when "110" => ooo_fifo_wadv_r1 <= "01000000";
        when "111" => ooo_fifo_wadv_r1 <= "10000000";
        when others =>
           ooo_fifo_wadv_r1 <= "00000000";
      end case;
      else
         ooo_fifo_wadv_r1 <= "00000000";
      end if;
     
                          
      fifo_wadv_ii1 <= OR_reduce(ooo_fifo_radv) or fifo_wadv_ii0;
      fifo_wadv_ii2 <= fifo_wadv_ii1;
      fifo_wadv_ii3 <= fifo_wadv_ii2;
      fifo_wadv <= fifo_wadv_ii3;
      fifo_data <= ooo_fifo_q_e3;
      
    end if;
  end process fsm3;

  gen_read_ooo: for i in 0 to 7 generate
    ooo_fifo_radv_q(i) <= mrespaccept_state and (not ooo_fifo_empty(i)) when this_tag = i else '0';
  end generate gen_read_ooo;
  
  ooo_fifo_radv <= ooo_fifo_radv_q or ooo_fifo_radv_flush;
--  gen_ooo_fifo : if check_tags generate

     --ooo_fifo_wadv_r1 <= ooo_fifo_wadv7_r1 & ooo_fifo_wadv6_r1 & ooo_fifo_wadv5_r1 & ooo_fifo_wadv4_r1 & ooo_fifo_wadv3_r1 & ooo_fifo_wadv2_r1 & ooo_fifo_wadv1_r1 & ooo_fifo_wadv0_r1;
     --ooo_fifo_radv <= ooo_fifo_radv7 & ooo_fifo_radv6 & ooo_fifo_radv5 & ooo_fifo_radv4 & ooo_fifo_radv3 & ooo_fifo_radv2 & ooo_fifo_radv1 & ooo_fifo_radv0;
     
    
    dpram_sfifo_ooo_1: entity work.dpram_sfifo_ooo
      generic map (
        width => 256,
        order => 9)
      port map (
        clk    => clk,
        rst    => rst,
        wadv   => ooo_fifo_wadv_r1,
        wdata  => ooo_fifo_d_r1,
        nfull  => ooo_fifo_nfull,
        radv   => ooo_fifo_radv,
        rempty => ooo_fifo_empty,
        rdata  => ooo_fifo_q_e3);
    
     --ooo_fifo_empty0 <= ooo_fifo_empty(0);
     --ooo_fifo_empty1 <= ooo_fifo_empty(1);
     --ooo_fifo_empty2 <= ooo_fifo_empty(2);
     --ooo_fifo_empty3 <= ooo_fifo_empty(3);
     --ooo_fifo_empty4 <= ooo_fifo_empty(4);
     --ooo_fifo_empty5 <= ooo_fifo_empty(5);
     --ooo_fifo_empty6 <= ooo_fifo_empty(6);
     --ooo_fifo_empty7 <= ooo_fifo_empty(7);
     
     
    --out_of_order_fifo0 : dpram_sfifo
    --  generic map (
    --    width => 259)
    --  port map (
    --    clk    => clk,
    --    rst    => rst,
    --    wadv   => ooo_fifo_wadv0_r1,
    --    wdata  => ooo_fifo_d_r1,
    --    nfull  => ooo_fifo_nfull0,
    --    radv   => ooo_fifo_radv0,
    --    rempty => ooo_fifo_empty0,
    --    rdata  => ooo_fifo_q0);

    --out_of_order_fifo1 : dpram_sfifo
    --  generic map (
    --    width => 259)
    --  port map (
    --    clk    => clk,
    --    rst    => rst,
    --    wadv   => ooo_fifo_wadv1_r1,
    --    wdata  => ooo_fifo_d_r1,
    --    nfull  => ooo_fifo_nfull1,
    --    radv   => ooo_fifo_radv1,
    --    rempty => ooo_fifo_empty1,
    --    rdata  => ooo_fifo_q1);


    --out_of_order_fifo2 : dpram_sfifo
    --  generic map (
    --    width => 259)
    --  port map (
    --    clk    => clk,
    --    rst    => rst,
    --    wadv   => ooo_fifo_wadv2_r1,
    --    wdata  => ooo_fifo_d_r1,
    --    nfull  => ooo_fifo_nfull2,
    --    radv   => ooo_fifo_radv2,
    --    rempty => ooo_fifo_empty2,
    --    rdata  => ooo_fifo_q2);

    --out_of_order_fifo3 : dpram_sfifo
    --  generic map (
    --    width => 259)
    --  port map (
    --    clk    => clk,
    --    rst    => rst,
    --    wadv   => ooo_fifo_wadv3_r1,
    --    wdata  => ooo_fifo_d_r1,
    --    nfull  => ooo_fifo_nfull3,
    --    radv   => ooo_fifo_radv3,
    --    rempty => ooo_fifo_empty3,
    --    rdata  => ooo_fifo_q3);
--  end generate gen_ooo_fifo;


  --resp_fifo_in(255 downto 0) <= ocp_s2m.Data;
  --resp_fifo_in(258 downto 256) <= ocp_s2m.Tag(4 downto 2);
  --resp_fifo_wadv <= '1' when ocp_s2m.Resp = OCP_RESP_VALID else '0';
  
  --resp_fifo0 : dpram_sfifo
  --    generic map (
  --      width => 259)
  --    port map (
  --      clk    => clk,
  --      rst    => rst,
  --      wadv   => resp_fifo_wadv,
  --      wdata  => resp_fifo_in,
  --      nfull  => resp_fifo_nfull,
  --      radv   => resp_fifo_radv,
  --      rempty => resp_fifo_empty,
  --      rdata  => resp_fifo_out);

  --ocp_s2m_i.Data <= resp_fifo_out(255 downto 0);
  --ocp_s2m_i.Tag <= "000" & resp_fifo_out(258 downto 256) & "00";
  --ocp_s2m_i.Resp <= OCP_RESP_VALID when resp_fifo_empty = '0' else OCP_RESP_NONE;
  --resp_fifo_radv <= mrespaccept and (not resp_fifo_empty);

  
  
  
  irq <= '0' when irq_state = "000" else '1';





  status_out0 : process (clk)
  begin  -- process status_out0
    if clk'event and clk = '1' then     -- rising clock edge
      if state = IDLE then
        fsm_idle(0) <= cmd_fifo_empty;
      else
        fsm_idle(0) <= '0';
      end if;
      if state = S_ABORT then
        fsm_aborting(0) <= '1';
      else
        fsm_aborting(0) <= '0';
      end if;

      if req_state = REQ_IDLE then
        fsm_idle(1) <= '1';
      else
        fsm_idle(1) <= '0';
      end if;
      if req_state = REQ_ABORT then
        fsm_aborting(1) <= '1';
      else
        fsm_aborting(1) <= '0';
      end if;

      if comp_state = COMP_IDLE then
        fsm_idle(2) <= '1';
      else
        fsm_idle(2) <= '0';
      end if;
      if comp_state = COMP_ABORT then
        fsm_aborting(2) <= '1';
      else
        fsm_aborting(2) <= '0';
      end if;

     -- fsm_error <= comp_counter(15) or active_counter(31);
      fsm_error <= OR_reduce(ooo_fifo_nfull);

      --if state = IDLE and req_state = REQ_IDLE and comp_state = COMP_IDLE and cmd_fifo_empty = '1' then
      --  all_idle_sr(0) <= '1';
      --else
      --  all_idle_sr(0) <= '0';
      --end if;
      --all_idle_sr(31 downto 1) <= all_idle_sr(30 downto 0);
      --all_idle <= all_idle_sr(0) and all_idle_sr(31);
      --if all_idle = '1' then
      --  comp_counter <= (others => '0');
      --  active_counter <= (others => '0');
      --else
      --  if ocp_s2m.Resp = OCP_RESP_VALID then
      --    comp_counter <=comp_counter+1;
      --  end if;
      --  active_counter <=active_counter+1;
      --end if;

      
       --if comp_state = COMP_ERROR then
      --   fsm_error <= '1';
       --else
       --  fsm_error <= '0';
      -- end if;
      --fsm_error <=  not (ooo_fifo_empty0 and  ooo_fifo_empty1 and  ooo_fifo_empty2 and ooo_fifo_empty3) or resp_fifo_nfull;
      --if fsm_error1 = '1' then
      --  fsm_error_count <= fsm_error_count+1;
      --else
      --  fsm_error_count <= (others => '0');
      --end if;
      --fsm_error <= fsm_error_count(10);
    end if;
  end process status_out0;


  gen_chipcsope: if chipscope generate
    icon0: icon_rw
      port map (
        CONTROL0 => CONTROL0);

    ila0: ila_rw
      port map (
        CONTROL => CONTROL0,
        CLK => CLK,
        TRIG0 => TRIG0);

    trig0(0) <= ooo_fifo_empty_q0;
    trig0(1) <= ooo_fifo_empty_q1;
    trig0(2) <= ooo_fifo_empty_q2;
    trig0(3) <= ooo_fifo_empty_q3;
    trig0(7 downto 4) <= "0000" when comp_state = COMP_IDLE else
                         "0001" when comp_state = COMP_READ_IFF else
                         "0010" when comp_state = COMP_GEN_IRQ else
                         "0011" when comp_state = COMP_GEN_FLUSH else
                         "0100" when comp_state = COMP_ACCEPT_DATA else
                         --"0100" when comp_state = COMP_EMPTY_OOO_FIFO0 else
                         --"0101" when comp_state = COMP_EMPTY_OOO_FIFO1 else
                         --"0110" when comp_state = COMP_EMPTY_OOO_FIFO2 else
                         --"0111" when comp_state = COMP_EMPTY_OOO_FIFO3 else
                         "1000" when comp_state = COMP_ABORT else
                         "1111" when comp_state = COMP_ERROR;
                         
    trig0(8) <= fifo_wadv_ii0;
    trig0(9) <= fifo_wadv_ii1;
    trig0(10) <= fifo_wadv_ii2;
    trig0(11) <= fifo_wadv_ii3;
    trig0(12) <= ooo_fifo_wadv0;
    trig0(13) <= ooo_fifo_wadv1;
    trig0(14) <= ooo_fifo_wadv2;
    trig0(15) <= ooo_fifo_wadv3;
    trig0(16) <= last_in_this_burst;
    trig0(17) <= mrespaccept_state;
    trig0(18) <= fifo_wadv_i;
    trig0(19) <= '1';
    trig0(20) <= ooo_fifo_radv0_reg;
    trig0(21) <= ooo_fifo_radv1_reg;
    trig0(22) <= ooo_fifo_radv2_reg;
    trig0(23) <= ooo_fifo_radv3_reg;
    trig0(26 downto 24) <=  ocp_s2m_i.Tag(4 downto 2);
    trig0(29 downto 27) <= this_tag;
    trig0(31 downto 30) <= ocp_s2m_i.Resp;

    
    
    
    
  end generate gen_chipcsope;
  
end rtl;

