--------------------------------------------------------------------------------
-- Company:         Alpha-Data Parallel Systems
-- Engineer:        Alan Crawford
--
-- Create Date:
-- Design Name:
-- Module Name:     pcieg3_bar_if
-- Project Name:    ADB3G3
-- Target Device:   V7
-- Tool versions:
-- Description:     Interfaces to the Gen3 Core BAR I/F and provides the
--                  required number of OCP ports
--
-- Dependencies:    PCIe/OCP Packages for signal records
--
-- Revision:
-- Revision 0.01 - File Created
-- Additional Comments:
--
--------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.all;
use IEEE.STD_LOGIC_ARITH.all;
use IEEE.STD_LOGIC_UNSIGNED.all;

library work;
use work.pcie.all;
use work.adb3_ocp.all;
use work.pciegen3_pkg.all;
use work.pcie_config.all;

entity pcieg3_bar2_decoder is
  generic(
    max_timeout         : std_logic_vector(15 downto 0) := x"0FFF";
    datawidthRX         : integer := 256;
    datawidthTX         : integer := 256;
    single_tag_request  : boolean := false
  );
  port(
    pcie_clk            : in  std_logic;
    pcie_rst            : in  std_logic;
    us_strobe           : in  std_logic;
    bar_mask            : in  std_logic_vector(63 downto 0);
    bar_page            : in  std_logic_vector(63 downto 0);
    -- basic core CMD and data I/F
    data                : in  std_logic_vector(datawidthRX-1 downto 0);
    data_valid          : in  std_logic;
    data_ben            : in  std_logic_vector(datawidthRX/8-1 downto 0);
    data_user           : in  std_logic_vector(84 downto 0);

    cmd_push            : in  std_logic;
    data_push           : in  std_logic;

    bar_full            : out std_logic;

    rx_data_out         : out std_logic_vector(datawidthTX+datawidthTX/32+1 downto 0);
    rx_data_pop         : in  std_logic;
    rx_data_pkt_rdy     : out std_logic;

    m_m2s_o             : out adb3w_ocp_m2sT;
    m_s2m_i             : in  adb3w_ocp_s2mT;
    
    timeout             : out std_logic;
    clr_timeout         : in  std_logic
  );
end pcieg3_bar2_decoder;

architecture synth of pcieg3_bar2_decoder is

  signal m_m2s_o_cp               : adb3w_ocp_m2sT;
  signal data_empty               : std_logic;
  signal cmd_empty                : std_logic;
  signal bar_data_fifo_in         : std_logic_vector(datawidthRX+datawidthRX/8-1 downto 0);
  signal data_out                 : std_logic_vector(datawidthRX+datawidthRX/8-1 downto 0);
  signal cmd_data_in              : std_logic_vector(135 downto 0);
  signal cmd_data_out             : std_logic_vector(135 downto 0);
  signal temp                     : std_logic;
  signal temp2                    : std_logic;
  signal pip                      : std_logic;
  signal pip_sequencer            : std_logic_vector(4 downto 0);

  signal descriptor               : std_logic_vector(135 downto 0);

  signal load_descriptor          : std_logic := '0';
  signal cmd_pop                  : std_logic;

  signal cmd_entry                : std_logic_vector(86 downto 0);
  signal ocp_m2s_all              : std_logic_vector(86 downto 0);
  signal ocp_cmd                  : std_logic_vector(2 downto 0);
  signal ocp_address              : std_logic_vector(63 downto 0);
  signal ocp_tag                  : std_logic_vector(7 downto 0);
  signal ocp_cmd_empty            : std_logic;
  signal ocp_cmd_full             : std_logic;

  signal flush_tag_value          : std_logic_vector(7 downto 0);
  signal flush_tag_clear          : std_logic;
  signal flush_tag_info           : std_logic_vector(74 downto 0);

  alias descriptor_lbe_in         : std_logic_vector(3 downto 0) is cmd_data_out(135 downto 132);
  alias descriptor_fbe_in         : std_logic_vector(3 downto 0) is cmd_data_out(131 downto 128);
  alias descriptor_attr_in        : std_logic_vector(2 downto 0) is cmd_data_out(126 downto 124);
  alias descriptor_tc_in          : std_logic_vector(2 downto 0) is cmd_data_out(123 downto 121);
  alias descriptor_bar_in         : std_logic_vector(5 downto 0) is cmd_data_out(120 downto 115);
  alias descriptor_barid_in       : std_logic_vector(2 downto 0) is cmd_data_out(114 downto 112);
  alias descriptor_func_in        : std_logic_vector(7 downto 0) is cmd_data_out(111 downto 104);
  alias descriptor_tag_in         : std_logic_vector(7 downto 0) is cmd_data_out(103 downto 96);
  alias descriptor_id_in          : std_logic_vector(15 downto 0) is cmd_data_out(95 downto 80);
  alias descriptor_type_in        : std_logic_vector(3 downto 0) is cmd_data_out(78 downto 75);
  alias descriptor_length_in      : std_logic_vector(10 downto 0) is cmd_data_out(74 downto 64);
  alias descriptor_address_in     : std_logic_vector(61 downto 0) is cmd_data_out(63 downto 2);
  alias descriptor_addr_type_in   : std_logic_vector(1 downto 0) is cmd_data_out(1 downto 0);

  alias descriptor_lbe_reg        : std_logic_vector(3 downto 0) is descriptor(135 downto 132);
  alias descriptor_fbe_reg        : std_logic_vector(3 downto 0) is descriptor(131 downto 128);
  alias descriptor_attr_reg       : std_logic_vector(2 downto 0) is descriptor(126 downto 124);
  alias descriptor_tc_reg         : std_logic_vector(2 downto 0) is descriptor(123 downto 121);
  alias descriptor_bar_reg        : std_logic_vector(5 downto 0) is descriptor(120 downto 115);
  alias descriptor_barid_reg      : std_logic_vector(2 downto 0) is descriptor(114 downto 112);
  alias descriptor_func_reg       : std_logic_vector(7 downto 0) is descriptor(111 downto 104);
  alias descriptor_tag_reg        : std_logic_vector(7 downto 0) is descriptor(103 downto 96);
  alias descriptor_id_reg         : std_logic_vector(15 downto 0) is descriptor(95 downto 80);
  alias descriptor_type_reg       : std_logic_vector(3 downto 0) is descriptor(78 downto 75);
  alias descriptor_length_reg     : std_logic_vector(10 downto 0) is descriptor(74 downto 64);
  alias descriptor_address_reg    : std_logic_vector(61 downto 0) is descriptor(63 downto 2);
  alias descriptor_addr_type_reg  : std_logic_vector(1 downto 0) is descriptor(1 downto 0);

  signal tx_tag_info              : std_logic_vector(71 downto 0);
  signal tx_tag_request           : std_logic;
  signal tx_tag_valid             : std_logic;
  signal tx_tag_number            : std_logic_vector(2 downto 0);
  signal rx_tag_value             : std_logic_vector(7 downto 0);
  signal rx_tag_clear             : std_logic;
  signal rx_tag_valid             : std_logic;
  signal rx_tag_info              : std_logic_vector(71 downto 0);

  signal flush_requests           : std_logic := '0';
  signal block_pcie_requests      : std_logic := '0';
  signal bin_ocp_responses        : std_logic := '0';

  signal byte_count               : std_logic_vector(13 downto 0);
  signal low_address              : std_logic_vector(6 downto 0);
  signal requester_id             : std_logic_vector(15 downto 0);
  signal attr                     : std_logic_vector(2 downto 0);
  signal tc                       : std_logic_vector(2 downto 0);
  signal addr_type                : std_logic_vector(1 downto 0);
  signal start_addr               : std_logic_vector(13 downto 0);
  signal size                     : std_logic_vector(13 downto 0);

  signal tag_input_data           : std_logic_vector(54 downto 0);
  signal ocp_length               : std_logic_vector(8 downto 0);
  signal end_address              : std_logic_vector(13 downto 0);
  signal cq_lbe_calc              : std_logic_vector(2 downto 0);
  signal cq_fbe_calc              : std_logic_vector(2 downto 0);
  signal bc_offset                : std_logic_vector(2 downto 0);
  signal lbe_set                  : std_logic;
  signal pause_rx                 : std_logic := '0';
  constant max_pkt_size           : std_logic_vector(3 downto 0) := "1000";

  signal conv_pop                 : std_logic;
  signal data_out256              : std_logic_vector(datawidthTX+datawidthTX/8-1 downto 0);

  signal sequencer                : std_logic_vector(4 downto 0);
  signal conv_push_wide           : std_logic;
  signal conv_pop_wide            : std_logic;
  signal bar_data_fifo_in_narrow  : std_logic_vector(143 downto 0);
  signal data_out_wide            : std_logic_vector(287 downto 0);
  signal data_empty_wide          : std_logic;
  signal temp2_wide               : std_logic;
  signal temp_reg                 : std_logic_vector(datawidthRX+datawidthRX/8-1 downto 0);

  signal data_push_r1             : std_logic := '0';
  signal data_pop                 : std_logic := '0';
  signal pop_cmd                  : std_logic := '0';
  signal tag_timeout_enable       : std_logic := '1'; -- disable timeouts
  signal clear_tags               : std_logic := '0';
  
  signal local_rst,local_sync_rst : std_logic := '1';
  attribute keep : string;
  attribute max_fanout : string;
  attribute equivalent_register_removal : string;
  attribute keep of local_rst : signal is "TRUE";
  attribute equivalent_register_removal of local_rst : signal is "no";
  attribute max_fanout of local_rst : signal is "1";
  attribute keep of local_sync_rst : signal is "TRUE";
  attribute equivalent_register_removal of local_sync_rst : signal is "no";

  signal local_rst_r1: std_logic := '1';
  attribute shreg_extract : string;
  attribute keep of local_rst_r1 : signal is "TRUE";  
  attribute equivalent_register_removal of local_rst_r1 : signal is "no";
  attribute max_fanout of local_sync_rst : signal is "16";
  attribute shreg_extract of local_rst_r1 : signal is "no";
  attribute shreg_extract of local_sync_rst : signal is "no";
  
begin

  process (pcie_clk,pcie_rst) is
  begin  -- process
    if pcie_rst = '1' then
      local_rst <= '1';
    elsif rising_edge(pcie_clk) then            -- rising clock edge
      local_rst <= '0';
    end if;
  end process;
  
  process (pcie_clk) is
  begin  -- process
    if rising_edge(pcie_clk) then       -- rising clock edge
      local_rst_r1 <= local_rst;
      local_sync_rst <= local_rst_r1;
    end if;
  end process;

  
  cmd_data_in <= data_user(7 downto 0) & data(127 downto 0);

    bar_if_cmd_fifo : packet_sfifo
    generic map(
          RAM_TYPE    => "AUTO",
          ORDER       => 4,
          WIDTH       => 136,
          AFULL_LEVEL => 8
    )
    port map(
      rst         => pcie_rst,
      sr          => '0',
      clk         => pcie_clk,
      wen         => cmd_push,
      wd          => cmd_data_in,
      wfull       => open,
      wafull      => temp,
      radv        => cmd_pop,
      rq          => cmd_data_out, 
      rlevel      => open,
      rempty      => cmd_empty,
      debug_info  => open);
    
  process (pcie_clk) is
  begin 
    if rising_edge(pcie_clk) then
      bar_data_fifo_in <= data & data_ben;
      data_push_r1 <= data_push;
    end if;
  end process;
    
    downsize_data_fifo : packet_sfifo
    generic map(
          RAM_TYPE    => "AUTO",
          ORDER       => 7,
          WIDTH       => 288,
          AFULL_LEVEL => 128
    )
    port map(
      rst         => pcie_rst,
      sr          => '0',
      clk         => pcie_clk,
      wen         => data_push_r1,
      wd          => bar_data_fifo_in,
      wfull       => open,
      wafull      => temp2,
      radv        => data_pop,
      rq          => data_out_wide, 
      rlevel      => open,
      rempty      => data_empty_wide,
      debug_info  => open);

  gen_ocp_data_accept_pop : process(flush_requests,m_s2m_i.dataaccept,data_empty_wide)
  begin
    if(flush_requests = '1')then
      data_pop <= not data_empty_wide;
    else
      data_pop <= m_s2m_i.dataaccept;
    end if;
  end process;
      
  m_m2s_o.data       <= data_out_wide(287 downto 32);
  m_m2s_o.databyteen <= data_out_wide(31 downto 0);
  m_m2s_o.datavalid  <= not data_empty_wide and (not flush_requests);
  bar_full           <= not (not temp and not temp2) or (pause_rx) or(clear_tags);-- or block_pcie_requests;

  store_descriptor : process (pcie_clk)
  begin
    if pcie_clk'event and pcie_clk = '1' then
      if(local_sync_rst = '1')then
        descriptor       <= (others => '0');
      else
        if(load_descriptor = '1')then
          descriptor       <= cmd_data_out;
        end if;
      end if;
    end if;
  end process store_descriptor;

-- Async read in the descriptor (minimise down time between transfers
-- keep accepting transfers while in the request - should not be set if read request

  capture_descriptor : process(pcie_clk)
  begin
    
    if((pcie_clk'event) and (pcie_clk = '1'))then
      if(local_sync_rst = '1')then
        cmd_pop         <= '0';
        load_descriptor <= '0';
        pip             <= '0';
        pip_sequencer   <= (others => '0');
        tx_tag_request  <= '0';
      else
        cmd_pop        <= '0';
        tx_tag_request <= '0';
        pip_sequencer  <= pip_sequencer(3 downto 0)&'0';
        if((cmd_empty = '0')and (ocp_cmd_full = '0')and(cmd_pop = '0')and(load_descriptor = '0')and(pip = '0')and((descriptor_type_in = "0001")or(tx_tag_valid = '1')))then
--         if(((cmd_empty = '0')and (ocp_cmd_full = '0')and(cmd_pop = '0')and(load_descriptor = '0')and(pip = '0')and((descriptor_type_in = "0001")or(tx_tag_valid = '1')))and(block_pcie_requests = '0'))then
          pip              <= '1';
          pip_sequencer(0) <= '1';
          cmd_pop          <= '1';
          load_descriptor  <= '1';
        else
          load_descriptor <= '0';
        end if;
        if(pip_sequencer(3) = '1')then
          pip <= '0';
        end if;
        if(descriptor_type_reg = "0000")then
          tx_tag_request <= pip_sequencer(1);
        end if;
      end if;
    end if;
  end process capture_descriptor;

  islbeset : process(descriptor_length_reg)
  begin
    if(descriptor_length_reg = "00000000001")then
      lbe_set <= '0';
    else
      lbe_set <= '1';
    end if;
  end process islbeset;

  calcfbeoffset : process(lbe_set, descriptor_fbe_reg)
  begin
    case(lbe_set&descriptor_fbe_reg)is
      when "00001" => cq_fbe_calc <= "011";
      when "00010" => cq_fbe_calc <= "011";
      when "00011" => cq_fbe_calc <= "010";
      when "00100" => cq_fbe_calc <= "011";
      when "00101" => cq_fbe_calc <= "001";
      when "00110" => cq_fbe_calc <= "010";
      when "00111" => cq_fbe_calc <= "001";
      when "01000" => cq_fbe_calc <= "011";
      when "01001" => cq_fbe_calc <= "000";
      when "01010" => cq_fbe_calc <= "001";
      when "01011" => cq_fbe_calc <= "000";
      when "01100" => cq_fbe_calc <= "010";
      when "01101" => cq_fbe_calc <= "000";
      when "01110" => cq_fbe_calc <= "001";
      when "01111" => cq_fbe_calc <= "000";
      when "10010" => cq_fbe_calc <= "001";
      when "10110" => cq_fbe_calc <= "001";
      when "11010" => cq_fbe_calc <= "001";
      when "11110" => cq_fbe_calc <= "001";
      when "10100" => cq_fbe_calc <= "010";
      when "11100" => cq_fbe_calc <= "010";
      when "11000" => cq_fbe_calc <= "011";
      when others  => cq_fbe_calc <= "000";
    end case;
  end process calcfbeoffset;

  calclbeoffset : process(descriptor_lbe_reg)
  begin
    case(descriptor_lbe_reg)is
      when "0001" => cq_lbe_calc <= "011";
      when "0010" => cq_lbe_calc <= "010";
      when "0011" => cq_lbe_calc <= "010";
      when "0100" => cq_lbe_calc <= "001";
      when "0101" => cq_lbe_calc <= "001";
      when "0110" => cq_lbe_calc <= "001";
      when "0111" => cq_lbe_calc <= "001";
      when "1000" => cq_lbe_calc <= "000";
      when "1001" => cq_lbe_calc <= "000";
      when "1010" => cq_lbe_calc <= "000";
      when "1011" => cq_lbe_calc <= "000";
      when "1100" => cq_lbe_calc <= "000";
      when "1101" => cq_lbe_calc <= "000";
      when "1110" => cq_lbe_calc <= "000";
      when "1111" => cq_lbe_calc <= "000";
      when others => cq_lbe_calc <= "000";
    end case;

  end process calclbeoffset;

  bc_offset <= cq_fbe_calc + cq_lbe_calc;

  start_addr <= "000000000"& descriptor_address_reg(2 downto 0)& "00";

  end_address <= start_addr + size - "00000000000001"; 

  size <= '0' & descriptor_length_reg(10 downto 0)&"00";

  calc_tag_info : process(pcie_clk)
    variable size_m1 : std_logic_vector(13 downto 0);
  begin
    if((pcie_clk'event) and (pcie_clk = '1')) then
        case(descriptor_fbe_reg) is
          when "0010" => low_address(1 downto 0) <= "01";
          when "0110" => low_address(1 downto 0) <= "01";
          when "1010" => low_address(1 downto 0) <= "01";
          when "1110" => low_address(1 downto 0) <= "01";
          when "0100" => low_address(1 downto 0) <= "10";
          when "1100" => low_address(1 downto 0) <= "10";
          when "1000" => low_address(1 downto 0) <= "11";
          when others => low_address(1 downto 0) <= "00";
        end case;
      low_address(6 downto 2) <= descriptor_address_reg(4 downto 0);
      case(descriptor_fbe_reg)is
        when "0001" => low_Address(1 downto 0) <= "00";
        when "0010" => low_Address(1 downto 0) <= "01";
        when "0011" => low_Address(1 downto 0) <= "00";
        when "0100" => low_Address(1 downto 0) <= "10";
        when "0101" => low_Address(1 downto 0) <= "00";
        when "0110" => low_Address(1 downto 0) <= "01";
        when "0111" => low_Address(1 downto 0) <= "00";
        when "1000" => low_Address(1 downto 0) <= "11";
        when "1001" => low_Address(1 downto 0) <= "00";
        when "1010" => low_Address(1 downto 0) <= "01";
        when "1011" => low_Address(1 downto 0) <= "00";
        when "1100" => low_Address(1 downto 0) <= "10";
        when "1101" => low_Address(1 downto 0) <= "00";
        when "1110" => low_Address(1 downto 0) <= "01";
        when "1111" => low_Address(1 downto 0) <= "00";
        when others => low_Address(1 downto 0) <= "00";
      end case;
      ocp_length <= end_address(datawidthRX/128+11 downto datawidthRX/128+3)+'1';
      byte_count <= "0"&descriptor_length_reg&"00" - ("00000000000"&bc_offset);
    end if;
  end process calc_tag_info;
  
  ocp_address <= (bar_mask and (descriptor_address_reg(61 downto 3) & "00000")) or bar_page;
  ocp_tag     <= "00000"&tx_tag_number;

  tx_tag_info <= end_address(4 downto 2) &
                 descriptor_tag_reg &
                 descriptor_attr_reg &
                 descriptor_tc_reg &
                 descriptor_id_reg &
                 descriptor_length_reg(9 downto 0) &
                 low_address(6 downto 0) &
                 byte_count(13 downto 0) &
                 ocp_length(7 downto 0);

  get_ocp_cmd : process(descriptor_type_reg)
  begin
    if(descriptor_type_reg = "0000")then
      ocp_cmd <= OCP_CMD_READ;
    else
      ocp_cmd <= OCP_CMD_WRITE;
    end if;
  end process get_ocp_cmd;

  cmd_entry <= ocp_cmd & ocp_address & ocp_tag & "000" & ocp_length;

    ocp_cmd_fifo : packet_sfifo
    generic map(
          RAM_TYPE    => "AUTO",
          ORDER       => 4,
          WIDTH       => 87,
          AFULL_LEVEL => 8
    )
    port map(
      rst         => pcie_rst,
      sr          => '0',
      clk         => pcie_clk,
      wen         => pip_sequencer(2),
      wd          => cmd_entry,
      wfull       => open,
      wafull      => ocp_cmd_full,
      radv        => pop_cmd,
      rq          => ocp_m2s_all, 
      rlevel      => open,
      rempty      => ocp_cmd_empty,
      debug_info  => open);


  gen_ocp_cmd_accept_pop :process (flush_requests, m_s2m_i.cmdaccept,ocp_cmd_empty)
  begin
    if(flush_requests = '1')then
      pop_cmd <= not ocp_cmd_empty;
    else
      pop_cmd <= m_s2m_i.cmdaccept;
    end if;
  end process;
  
  gen_ocp_cmd : process(ocp_m2s_all, ocp_cmd_empty, flush_requests)
  begin
    if((ocp_cmd_empty = '0')and(flush_requests = '0'))then
      m_m2s_o.Cmd <= ocp_m2s_all(86 downto 84);
    else
      m_m2s_o.Cmd <= OCP_CMD_IDLE;
    end if;
  end process gen_ocp_cmd;

  m_m2s_o.Addr        <= ocp_m2s_all(83 downto 24) & "0000";
  m_m2s_o.Tag         <= ocp_m2s_all(19 downto 12);
  m_m2s_o.Burstlength <= ocp_m2s_all(11 downto 0);

  rx : pcieg3_bar2_rx_if
    generic map(
      datawidthTX       => 256
    )
    port map(
      pcie_clk          => pcie_clk,
      pcie_rst          => pcie_rst,
      m_m2s_o           => m_m2s_o_cp,
      m_s2m_i           => m_s2m_i,
      rx_data_out       => rx_data_out,
      rx_data_pop       => rx_data_pop,
      rx_data_pkt_rdy   => rx_data_pkt_rdy,
      bin_responses     => flush_requests,
      pause_rx          => pause_rx,
      rx_tag_value      => rx_tag_value,
      rx_tag_clear      => rx_tag_clear,
      rx_tag_info       => rx_tag_info,
      rx_tag_valid      => rx_tag_valid
    );

  m_m2s_o.respaccept <= m_m2s_o_cp.respaccept;
  
  bar_tags_manager : bar_tag_manager
    generic map(
      max_timeout         => max_timeout,
      single_req          => single_tag_request
    )                     
    port map(             
      pcie_rst            => pcie_rst,
      pcie_clk            => pcie_clk,
      us_counter          => us_strobe,
      tx_tag_request      => tx_tag_request,
      tx_tag_data_in      => tx_tag_info,
      tx_tag_number       => tx_tag_number,
      tx_tag_valid        => tx_tag_valid,
      rx_tag_value        => rx_tag_value(2 downto 0),
      rx_tag_clear        => rx_tag_clear,
      rx_tag_valid        => rx_tag_valid,
      rx_tag_info         => rx_tag_info,
      tag_timeout_enable  => tag_timeout_enable,
      clr_timeout         => clr_timeout,
      tag_timeout         => block_pcie_requests,
      flush_tags          => flush_requests,
      clear_tags          => clear_tags,
      tag_error           => open,
      status              => open
    );
    
    bin_ocp_responses <= block_pcie_requests;
    timeout           <= block_pcie_requests;

end architecture;
