--------------------------------------------------------------------------------
-- Company:         Alpha-Data Parallel Systems Ltd
-- Engineer:        Alan Crawford
--
-- Create Date:     2014
-- Design Name:     pcie_g3_bridge
-- Module Name:     pcie_g3_dma_merger
-- Project Name:    ADB3G3
-- Target Device:   V7
--
-- Description:     Arbitration mechanism for DMA engines
--
-- Dependencies:    PCIe/OCP Packages for signal records
--
-- Revision:        1.0 - Release Version
-- Additional Comments:
--
--                  Design adapted from 
--                  "Design and Implementation of a Fast Crossbar Scheduler"
--                  by Pankaj Gupta & Nick McKeown
--------------------------------------------------------------------------------
library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_misc.all;

library work;
use work.pcie_config.ALL;
use work.pcie.ALL;
use work.adb3_ocp.ALL;
use work.pciegen3_pkg.all;

-- synopsys translate_off
library unisim;               
use unisim.vcomponents.all;   
-- synopsys translate_on

entity dma_merger is
  generic(
    channelCount         : integer := 5;
    dataWidth            : integer := 256
    );
   port (
      -- no reset
      Ck                 : in  std_logic;
   
      -- Inputs 
      srcReq             : in  std_logic_vector(channelCount-1 downto 0);
      srcData            : in  channelArray;
      srcPop             : out std_logic_vector(channelCount-1 downto 0);
      -- output
      opData             : out std_logic_vector(dataWidth+dataWidth/32+3 downto 0);
      opPush             : out std_logic;
      opFull             : in std_logic
   );

end dma_merger;

architecture synth of dma_merger is

  type MaskedchannelArray is array (0 to 15) of  std_logic_vector(dataWidth+dataWidth/32+3 downto 0);
  signal maskedDataChannels  : MaskedchannelArray;
  
  signal opDataInt           : std_logic_vector(dataWidth+dataWidth/32+3 downto 0) := (others => '0');
  signal updateArb           : std_logic := '0';
  signal updateArbInt        : std_logic := '0';
  signal pktWaiting          : std_logic := '0';
  signal srcPopInt           : std_logic_vector(channelCount-1 downto 0) := (others => '0');
  signal LPVal               : std_logic_vector(channelCount-1 downto 0) := (others => '0');
  signal idleReq             : std_logic_vector(channelCount-1 downto 0) := (others => '0');
--   signal pop                 : std_logic := '0';
  signal reqInt              : std_logic_vector(channelCount-1 downto 0) := (others => '0');
  signal end_of_pkt          : std_logic := '0';
  signal del_end_of_pkt      : std_logic := '0';
  signal pip                 : std_logic := '0';
  signal updateArbdel        : std_logic := '0';
  signal opPushInt           : std_logic := '0';
  signal pipsingle           : std_logic := '0';
  signal no_change           : std_logic := '0';
begin

  iArb : dma_arbiter
    generic map(
      maxChannels => channelCount
    )
    port map(
      Ck                => Ck,
      PtrUpdate         => updateArb,
      Req               => reqInt,
      AnySet            => pktWaiting,
      LPVal             => LPVal
    );


  process(Ck)
  begin
    if(rising_edge(Ck))then
      if(updateArb = '1')then
        srcPopInt <= LPVal;
      end if;
    end if;
  end process;   
  process(srcPopInt,LPVal)
  begin
    if(srcPopInt = LPVal)then
      no_change <= '1';
    else
      no_change <= '0';
    end if;
  end process;   
   
  process(pip,srcPopInt,opFull)
  begin
      if((pip = '1')and(opFull = '0'))then 
        srcPop <= srcPopInt;
        opPush <= '1';
      else
        srcPop <= (others => '0');
        opPush <= '0';
      end if;
  end process;

  reqInt <= srcReq;

  process(Ck)
  begin
    if(rising_edge(Ck))then
      pip <= pip;
      updateArbdel <= updateArb;
--       if((updateArb = '1')and(pktWaiting = '1')and(opFull = '0'))then
      if((((no_change ='1')and(updateArbdel = '1'))or ((no_change = '0')and(updateArb = '1')))and(pktWaiting = '1')and(opFull = '0'))then
        pip <= '1';
      else
        if((end_of_pkt = '1')and((pktWaiting = '0')or(no_change='0')))then 
          pip<='0';
        end if;
      end if;
      if((pipsingle = '1')and(pip = '1'))then
        pip<='0';
      end if;
    end if;
  end process;
  
  process(opDataInt,opFull)
  begin
    pipsingle <='0';
    if((opDataInt(dataWidth+dataWidth/32+3 downto dataWidth+dataWidth/32+2)= "11")and(opFull='0'))then
      pipsingle <= '1';
    end if;
  end process;

  process(opDataInt,pip,opFull)
  begin
    if((opDataInt(dataWidth+dataWidth/32+2)='1') and (pip='1')and(opFull='0'))then
      end_of_pkt <= '1';
    else
      end_of_pkt <= '0';
    end if;
  end process;


  process(pktWaiting,pip,end_of_pkt,updateArbDel,opFull,pipsingle)
  begin
    updateArb <= '0';
    if(
      ((pktWaiting = '1')and((pip = '0')or(end_of_pkt = '1')or(pipsingle = '1'))and(opFull = '0')and(updateArbdel = '0'))
    )then
      updateArb <= '1';
    end if;
  end process;

  maskChannels : for chIndex in 0 to 15 generate
    activechannels : if chIndex<channelCount generate
      process(srcPopInt,srcData)
      begin
        if(srcPopInt(chIndex) = '1')then
          maskedDataChannels(chIndex) <= srcData(chIndex);
        else
          maskedDataChannels(chIndex) <= (others => '0');
        end if;
      end process;
    end generate activeChannels;
    inactiveChannels : if chIndex >= channelCount generate
      maskedDataChannels(chIndex) <= (others => '0');
    end generate inactiveChannels;
  end generate maskChannels;
  
  opDataInt <= maskedDataChannels(0) or 
         maskedDataChannels(1) or 
         maskedDataChannels(2) or 
         maskedDataChannels(3) or 
         maskedDataChannels(4) ;
         
  
  opData <= opDataInt;

end synth;


