--------------------------------------------------------------------------------
-- Company:         Alpha-Data Parallel Systems
-- Engineer:        Alan Crawford
--
-- Create Date:
-- Design Name:
-- Module Name:     pcie_pkt_fifo
-- Project Name:    adb3
-- Target Device:   V5FX
-- Tool versions:
-- Description:     Basic FIFO for all data paths
--
-- Dependencies:    PCIe/OCP Packages for signal records
--
-- Revision:
-- Revision 0.01 - File Created
-- Additional Comments:
--
--------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

library work;
use work.pcie.ALL;
use work.adb3_ocp.ALL;

entity pcie_pkt_fifo4 is
  generic(
    fifo_depth          : integer := 6;
    fifo_width          : integer := 131;
    ram_style_def       : string  := "distributed"
    );
  port(
    clk                 : in  std_logic;
--     ar                  : in  std_logic;
    sr                  : in  std_logic;
    max_pkt_size        : in  std_logic_vector(fifo_depth-1 downto 0);

    din                 : in  std_logic_vector(fifo_width-1 downto 0);
    dout                : out std_logic_vector(fifo_width-1 downto 0) := (others => '0');

    push                : in  std_logic;
    pop                 : in  std_logic;

    full                : out std_logic := '0';
    empty               : out std_logic := '1';
    last                : out std_logic := '0';
    free_pkt            : out std_logic := '1';
    data_available      : out std_logic_vector(fifo_depth-1 downto 0)
    );


end pcie_pkt_fifo4;

architecture synth of pcie_pkt_fifo4 is
type    data_fifo_t         is array(2**fifo_depth-1 downto 0) of std_logic_vector(fifo_width-1 downto 0);
signal  data_fifo           : data_fifo_t := (others => (others => '0'));
signal  wr_ptr              : std_logic_vector(fifo_depth-1 downto 0) := (others => '0');
signal  rd_ptr              : std_logic_vector(fifo_depth-1 downto 0) := (others => '0');
signal  full_int            : std_logic := '0';
signal  empty_int           : std_logic := '1';
signal  space_free          : std_logic_vector(fifo_depth-1 downto 0) := (others => '1');
signal  packet_count        : std_logic_vector(6 downto 0) := "0000000";
signal  dout_int            : std_logic_vector(fifo_width-1 downto 0) := (others => '0');

signal  pop_int             : std_logic := '0';
signal  load                : std_logic := '0';
signal  valid               : std_logic := '0';

signal  push_del            : std_logic_vector(1 downto 0) := (others => '0');
signal  inp_eop             : std_logic_vector(1 downto 0) := (others => '0');

attribute ram_style: string;
attribute ram_style of data_fifo: signal is ram_style_def;

signal max_pkt_size_p1        : std_logic_vector(fifo_depth-1 downto 0) := (others => '0');
signal free_pkt_i : std_logic := '1';

signal pkt_available_i       : std_logic := '0';

  attribute shreg_extract : string;
  attribute shreg_extract of push_del : signal is "no";
  attribute shreg_extract of inp_eop : signal is "no";
  attribute shreg_extract of free_pkt_i : signal is "no";
  attribute shreg_extract of free_pkt : signal is "no";

begin

  full <= full_int;
  empty <=(not valid);

  push_gen : process(clk)
  begin
    if(rising_edge(clk))then
      if(push = '1')then
        data_fifo(conv_integer(wr_ptr)) <= din;
      end if;
    end if;
  end process push_gen;


  op_reg1 : process(clk)
  begin
    if(rising_edge(clk))then
      if(load = '1')then
        dout_int <= data_fifo(conv_integer(rd_ptr));
      end if;
    end if;
  end process op_reg1;

  op_reg2 : process(clk)
  begin
    if(rising_edge(clk))then
      if(empty_int = '1')then
        if(pop = '1')then
          valid <= '0';
        else
          valid <= valid;
        end if;
      else
        if(pop_int = '1')then
          valid <= '1';
        else
          valid <= valid;
        end if;
      end if;
    end if;
  end process op_reg2;

  pop_int <= load;

  op_reg3: process(valid, pop, empty_int)
  begin
    if(((valid = '0') or (pop = '1')) and (empty_int ='0'))then
      load <= '1';
    else
      load <= '0';
    end if;
  end process op_reg3;

  dout <= dout_int;

  gen_pkt_avail : process(packet_count)
  begin
    if(packet_count > "0000000")then
      pkt_available_i <= '1';
    else
      pkt_available_i <= '0';
    end if;
  end process gen_pkt_avail;



  data_ptr_gen : process(clk)
  variable pushpop : std_logic_vector(1 downto 0);
  begin
    if(rising_edge(clk))then
        if((rd_ptr=wr_ptr)and(valid='1')and(push ='0'))then
          last <= '1';
        else
          last <= '0';
        end if;
        pushpop := push & pop_int;
        case(pushpop)is
          when "10" =>  if(full_int = '0')then
                          wr_ptr <= wr_ptr + '1';
                          rd_ptr <= rd_ptr;
                          empty_int <= '0';
                          space_free <= space_free - '1';
                          if((wr_ptr+'1') = rd_ptr)then
                            full_int <= '1';
                            assert false report "FIFO4 Is full" severity note;
                          else
                            full_int <= '0';
                          end if;
                        else
                          space_free <= space_free;
                          wr_ptr <= wr_ptr;
                          rd_ptr <= rd_ptr;
                          full_int   <= full_int;
                          empty_int  <= empty_int;
                        end if;

          when "01" =>  if(empty_int = '0')then
                          rd_ptr <= rd_ptr + '1';
                          wr_ptr <= wr_ptr;
                          full_int   <= '0';
                          space_free <= space_free + '1';
                          if(wr_ptr = (rd_ptr + '1'))then
                            empty_int  <= '1';
                            assert false report "FIFO4 Is empty" severity note;
                          else
                            empty_int <= '0';
                          end if;
                        else
                          space_free <= space_free;
                          wr_ptr <= wr_ptr;
                          rd_ptr <= rd_ptr;
                          full_int   <= full_int;
                          empty_int  <= empty_int;
                        end if;
          when "11" =>  rd_ptr <= rd_ptr + '1';
                        wr_ptr <= wr_ptr + '1';
                        full_int   <= full_int;
                        empty_int  <= empty_int;
                        space_free <= space_free;

          when others => rd_ptr <= rd_ptr;
                         wr_ptr <= wr_ptr;
                         full_int   <= full_int;
                         empty_int  <= empty_int;
                         space_free <= space_free;
        end case;
        push_del  <= push_del(0) & push;
        inp_eop   <= inp_eop(0) & din(fifo_width-2);
        pushpop := push_del(1) & pop;
        case(pushpop)is
          when "10" =>  packet_count <= packet_count + '1';
          when "01" =>  packet_count <= packet_count - '1';
          when "11" =>  packet_count <= packet_count;
                        
          when others => packet_count <= packet_count;
        end case;
    end if;
  end process data_ptr_gen;

--   gen_pkt_free : process(space_free, max_pkt_size)
  gen_pkt_free : process(clk)
  begin
    if(rising_edge(clk))then
      max_pkt_size_p1 <= max_pkt_size+2;
      if(space_free < max_pkt_size_p1)then
        free_pkt_i <= '0';
      else
        free_pkt_i <= '1';
      end if;
      free_pkt <= free_pkt_i;
      
    end if;
  end process gen_pkt_free;
--   pkt_available <= pkt_available_i;

  data_available <= packet_count(fifo_depth-1 downto 0);
end synth;

