--------------------------------------------------------------------------------
-- Company:         Alpha-Data Parallel Systems
-- Engineer:        Alan Crawford
--
-- Create Date:
-- Design Name:
-- Module Name:     pcieg3_bar_if
-- Project Name:    ADB3G3
-- Target Device:   V7
-- Tool versions:
-- Description:     Interfaces to the Gen3 Core BAR I/F and provides the
--                  required number of OCP ports
--
-- Dependencies:    PCIe/OCP Packages for signal records
--
-- Revision:
-- Revision 0.01 - File Created
-- Additional Comments:
--
--------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

library work;
use work.pcie.ALL;
use work.adb3_ocp.ALL;
use work.pciegen3_pkg.all;
use work.pcie_config.all;

entity pcieg3_bar_if is
  generic(
    max_timeout  : std_logic_vector(15 downto 0) := x"0FFF";
    datawidthRX   : integer := axi4_data_widthRX;
    datawidthTX   : integer := axi4_data_widthTX;
    datawidthBAR2 : integer := 128
    );
  port(
    pcie_clk              : in  std_logic;
    pcie_rst              : in  std_logic;
    us_strobe             : in  std_logic;
    -- Bar MAsks from config block
    bar_masks             : in  pcie_bar_mask_t;
    -- BAR Requester I/F
    cq_c2b                : in  cq_c2b_t;
    cq_b2c                : out cq_b2c_t;
    -- BAR Completion I/F
    cc_b2c                : out cc_b2c_t;
    cc_c2b                : in  cc_c2b_t;
    -- BAR OCP ports (no_masters = 3)
    m_m2s_o               : out adb3_ocp_m2sT_array(0 to 1);
    m_s2m_i               : in  adb3_ocp_s2mT_array(0 to 1);
    m_m2s_ow              : out adb3w_ocp_m2sT;
    m_s2m_iw              : in  adb3w_ocp_s2mT;
    timeout_bar23         : out std_logic;
    clr_timeout_bar23     : in  std_logic
    );
end pcieg3_bar_if;

architecture synth of pcieg3_bar_if is

signal m_axis_cq_tdata            : std_logic_vector(datawidthRX-1 downto 0);
-- signal m_axis_cq_tparity          : std_logic_vector(datawidthRX/8-1 downto 0);
signal m_axis_cq_tvalid           : std_logic;
signal m_axis_cq_tlast            : std_logic;
signal m_axis_cq_tkeep            : std_logic_vector(datawidthRX/32-1 downto 0);
signal m_axis_cq_tuser            : std_logic_vector(84 downto 0);
-- signal pcie_cq_np_req_count       : std_logic_vector(5 downto 0);

signal m_axis_cq_tready           : std_logic := '0';



-- Internal signals
-- signal reg_descriptor             : std_logic_vector(127 downto 0);
signal bar_detect                 : std_logic_vector(3 downto 0) := "0000";
signal reg_data_valid             : std_logic := '0';
signal reg_cmd_valid              : std_logic := '0';

signal reg_tdata                  : std_logic_vector(datawidthRX-1 downto 0);
signal reg_tdata_valid            : std_logic;
signal reg_tdata_ben              : std_logic_vector(datawidthRX/8-1 downto 0);

signal bar2_data_fifo_in          : std_logic_vector(datawidthRX+datawidthRX/8-1 downto 0);


signal bar0_full                  : std_logic := '0';
signal bar1_full                  : std_logic := '0';
signal bar2_full                  : std_logic := '0';
-- signal bar4_full                  : std_logic := '0';

signal push_bar0_cmd              : std_logic := '0';
signal push_bar1_cmd              : std_logic := '0';
signal push_bar2_cmd              : std_logic := '0';
-- signal push_bar4_cmd              : std_logic := '0';

signal push_bar0_data             : std_logic := '0';
signal push_bar1_data             : std_logic := '0';
signal push_bar2_data             : std_logic := '0';
-- signal push_bar4_data             : std_logic := '0';

signal tx_tag_bar0_info              : std_logic_vector(72 downto 0);
signal tx_tag_bar0_request           : std_logic;
signal tx_tag_bar0_valid             : std_logic;
signal tx_tag_bar0_number            : std_logic_vector(7 downto 0);

signal tx_tag_bar1_info              : std_logic_vector(72 downto 0);
signal tx_tag_bar1_request           : std_logic;
signal tx_tag_bar1_valid             : std_logic;
signal tx_tag_bar1_number            : std_logic_vector(7 downto 0);

signal tx_tag_bar2_info              : std_logic_vector(72 downto 0);
signal tx_tag_bar2_request           : std_logic;
signal tx_tag_bar2_valid             : std_logic;
signal tx_tag_bar2_number            : std_logic_vector(7 downto 0);

signal rx_tag_bar0_value          : std_logic_vector(7 downto 0);
signal rx_tag_bar0_clear          : std_logic;
signal rx_tag_bar0_valid          : std_logic;
signal rx_tag_bar0_info           : std_logic_vector(72 downto 0);
signal rx_tag_bar0_check          : std_logic;

signal rx_bar0_data_out           : std_logic_vector(datawidthTX+datawidthTX/32+1 downto 0);
signal rx_bar1_data_out           : std_logic_vector(datawidthTX+datawidthTX/32+1 downto 0);
signal rx_bar2_data_out           : std_logic_vector(datawidthTX+datawidthTX/32+1 downto 0);
signal rx_bar0_data_pop           : std_logic := '0';
signal rx_bar1_data_pop           : std_logic := '0';
signal rx_bar2_data_pop           : std_logic := '1';
signal rx_bar0_data_pkt_rdy       : std_logic;
signal rx_bar1_data_pkt_rdy       : std_logic;
signal rx_bar2_data_pkt_rdy       : std_logic;

signal rx_tag_bar1_value          : std_logic_vector(7 downto 0);
signal rx_tag_bar1_clear          : std_logic;
signal rx_tag_bar1_valid          : std_logic;
signal rx_tag_bar1_info           : std_logic_vector(72 downto 0);
signal rx_tag_bar1_check          : std_logic;

signal rx_tag_bar2_value          : std_logic_vector(7 downto 0);
signal rx_tag_bar2_clear          : std_logic;
signal rx_tag_bar2_valid          : std_logic;
signal rx_tag_bar2_info           : std_logic_vector(72 downto 0);
signal rx_tag_bar2_check          : std_logic;

signal temp_reg_full              : std_logic := '0';
signal sequencer                  : std_logic_vector(3 downto 0) := "0000"; -- 0000 - idle, 0001 - CMD in not accepted, 0010 - CMD Accepted, 0100 - Data in, 1000 - Pause end
signal  rx_bar_data_out       : std_logic_vector(datawidthTX+datawidthTX/32+1 downto 0);
signal  rx_bar_data_push      : std_logic;
signal  rx_bar_data_space     : std_logic := '0';

subtype bar_maskT is std_logic_vector(63 downto 0);
type  bar_mask_arrayT is array(0 to 3) of bar_maskT;

signal bar_mask : bar_mask_arrayT;
signal bar_page : bar_mask_arrayT;

signal local_rst,local_sync_rst : std_logic := '1';
  attribute keep : string;
  attribute max_fanout : string;
  attribute equivalent_register_removal : string;
  attribute keep of local_rst : signal is "TRUE";
  attribute equivalent_register_removal of local_rst : signal is "no";
  attribute max_fanout of local_rst : signal is "1";
  attribute keep of local_sync_rst : signal is "TRUE";
  attribute equivalent_register_removal of local_sync_rst : signal is "no";
  signal local_rst_r1: std_logic := '1';
  attribute shreg_extract : string;
  attribute keep of local_rst_r1 : signal is "TRUE";  
  attribute equivalent_register_removal of local_rst_r1 : signal is "no";
  attribute max_fanout of local_sync_rst : signal is "16";
  attribute shreg_extract of local_rst_r1 : signal is "no";
  attribute shreg_extract of local_sync_rst : signal is "no";
begin

  process (pcie_clk,pcie_rst) is
  begin  -- process
    if pcie_rst = '1' then
      local_rst <= '1';
    elsif rising_edge(pcie_clk) then            -- rising clock edge
      local_rst <= '0';
    end if;
  end process;
  
  process (pcie_clk) is
  begin  -- process
    if rising_edge(pcie_clk) then       -- rising clock edge
      local_rst_r1 <= local_rst;
      local_sync_rst <= local_rst_r1;
    end if;
  end process;
    
bar_mask(0) <= bar_masks.bar0_mask;
bar_mask(1) <= bar_masks.bar1_mask;
bar_mask(2) <= bar_masks.bar23_mask;
bar_mask(3) <= bar_masks.bar45_mask;

bar_page(0) <= bar_masks.bar0_page;
bar_page(1) <= bar_masks.bar1_page;
bar_page(2) <= bar_masks.bar23_page;
bar_page(3) <= bar_masks.bar45_page;



  cq_b2c.m_axis_cq_tready    <= m_axis_cq_tready;

  m_axis_cq_tdata            <= cq_c2b.m_axis_cq_tdata;
--   m_axis_cq_tparity          <= cq_c2b.m_axis_cq_tparity;
  m_axis_cq_tvalid           <= cq_c2b.m_axis_cq_tvalid;
  m_axis_cq_tkeep            <= cq_c2b.m_axis_cq_tkeep;
  m_axis_cq_tuser            <= cq_c2b.m_axis_cq_tuser;
  m_axis_cq_tlast            <= cq_c2b.m_axis_cq_tlast;
--   pcie_cq_np_req_count       <= cq_c2b.pcie_cq_np_req_count;


sequence_ctl : process (pcie_clk)
begin
  if pcie_clk'event and pcie_clk = '1' then
    if( local_sync_rst = '1')then
      sequencer         <= (others => '0');
    else
      if((m_axis_cq_tvalid = '1')and(m_axis_cq_tuser(40) = '1'))then
        sequencer <= "0001";
      end if;
      if(sequencer = "0001")then
        case bar_detect is
          when "0001" =>  if(bar0_full = '0')then
                            if(m_axis_cq_tlast = '1')then
                              sequencer <= "1000";
                            else
                              sequencer <= "0010";
                            end if;
                          end if;
          when "0010" => if(bar1_full = '0')then
                           if(m_axis_cq_tlast = '1')then
                             sequencer <= "1000";
                           else
                             sequencer <= "0010";
                           end if;
                         end if;
          when "0100" => if(bar2_full = '0')then
                           if(m_axis_cq_tlast = '1')then
                             sequencer <= "1000";
                           else
                             sequencer <= "0010";
                           end if;
                         end if;
          when others => sequencer <= "0000";
        end case;
      end if;
      if(sequencer = "0010")then
        sequencer <= "0100";
      end if;
      if(sequencer = "0100")then
        if(m_axis_cq_tlast='1')then
          sequencer <= "0000";
        end if;
      end if;
      if(sequencer = "1000")then
        sequencer <= "0000";
      end if;
    end if;
  end if;
end process;
      

tready_ctl : process(sequencer)
begin
  if((sequencer = "0010") or (sequencer = "0100") or (sequencer = "1000"))then
    m_axis_cq_tready <= '1';  
  else
    m_axis_cq_tready <= '0';  
  end if;
end process;

-- ------------------------------------------------------------------------
-- Decode BARS on each SOP from the core
-- ------------------------------------------------------------------------
Decode_bars : process (pcie_clk)
begin
  if pcie_clk'event and pcie_clk = '1' then
    if( local_sync_rst = '1')then
      bar_detect         <= (others => '0');
    else
      --bar_detect         <= (others => '0');
      if((m_axis_cq_tvalid = '1')and(m_axis_cq_tuser(40) = '1'))then
        case m_axis_cq_tdata(114 downto 112) is
          when "000"   => bar_detect <= "0001"; -- bar 0
          when "001"   => bar_detect <= "0010"; -- bar 1
          when "010"   => bar_detect <= "0100"; -- bar 2/3
          when "011"   => bar_detect <= "0100"; -- Invalid for 64 bit bar
          when "100"   => bar_detect <= "0100"; -- bar 4/5
          when "101"   => bar_detect <= "0100"; -- Invalid for 64 bit bar
          when others  => bar_detect <= "0000"; -- don't use expansion rom
        end case;
      end if;
    end if;
  end if;
end process Decode_bars;

-- ------------------------------------------------------------------------
-- Register all outputs of the core when the transfer is valid
-- Set valid flag in case of interrupted transfers cause gaps in the
-- incoming data
-- ------------------------------------------------------------------------
reg_data_256 : if(datawidthRX = 256)generate
  reg_cq_tdata : process (pcie_clk)
  begin
    if pcie_clk'event and pcie_clk = '1' then
      if( local_sync_rst = '1')then
        reg_tdata         <= (others => '0');
        reg_tdata_valid   <= '0';
        reg_tdata_ben     <= (others => '0');
--         reg_descriptor <= (others => '0');
      else
        if((m_axis_cq_tvalid = '1')and(m_axis_cq_tready = '1'))then
          reg_tdata       <= m_axis_cq_tdata;
          if(m_axis_cq_tlast = '1')then

            if(m_axis_cq_tkeep(7) = '1')then
              reg_tdata_ben(31 downto 28) <= m_axis_cq_tuser(39 downto 36);
            else
              reg_tdata_ben(31 downto 28) <= "0000";
            end if;
            if(m_axis_cq_tkeep(6) = '1')then
              reg_tdata_ben(27 downto 24) <= m_axis_cq_tuser(35 downto 32);
            else
              reg_tdata_ben(27 downto 24) <= "0000";
            end if;
            if(m_axis_cq_tkeep(5) = '1')then
              reg_tdata_ben(23 downto 20) <= m_axis_cq_tuser(31 downto 28);
            else
              reg_tdata_ben(23 downto 20) <= "0000";
            end if;
            if(m_axis_cq_tkeep(4) = '1')then
              reg_tdata_ben(19 downto 16) <= m_axis_cq_tuser(27 downto 24);
            else
              reg_tdata_ben(19 downto 16) <= "0000";
            end if;
            if(m_axis_cq_tkeep(3) = '1')then
              reg_tdata_ben(15 downto 12) <= m_axis_cq_tuser(23 downto 20);
            else
              reg_tdata_ben(15 downto 12) <= "0000";
            end if;
            if(m_axis_cq_tkeep(2) = '1')then
              reg_tdata_ben(11 downto 8) <= m_axis_cq_tuser(19 downto 16);
            else
              reg_tdata_ben(11 downto 8) <= "0000";
            end if;
            if(m_axis_cq_tkeep(1) = '1')then
              reg_tdata_ben(7 downto 4) <= m_axis_cq_tuser(15 downto 12);
            else
              reg_tdata_ben(7 downto 4) <= "0000";
            end if;
            if(m_axis_cq_tkeep(0) = '1')then
              reg_tdata_ben(3 downto 0) <= m_axis_cq_tuser(11 downto 8);
            else
              reg_tdata_ben(3 downto 0) <= "0000";
            end if;
          else
            reg_tdata_ben <= m_axis_cq_tuser(39 downto 8);
          end if;
          if(m_axis_cq_tuser(40) = '1')then
            reg_data_valid  <= '0';
            reg_cmd_valid   <= '1';
          else
            reg_data_valid  <= '1';
            reg_cmd_valid   <= '0';
          end if;
        else
          reg_data_valid  <= '0';
          reg_cmd_valid   <= '0';
        end if;
      end if;
    end if;
  end process reg_cq_tdata;
end generate reg_data_256;

reg_data_128 : if(datawidthRX = 128)generate
  reg_cq_tdata : process (pcie_clk)
  begin
    if pcie_clk'event and pcie_clk = '1' then
      if( local_sync_rst = '1')then
        reg_tdata         <= (others => '0');
        reg_tdata_valid   <= '0';
        reg_tdata_ben     <= (others => '0');
--         reg_descriptor <= (others => '0');
      else
        if((m_axis_cq_tvalid = '1')and(m_axis_cq_tready = '1'))then
          reg_tdata       <= m_axis_cq_tdata;
          if(m_axis_cq_tlast = '1')then
            if(m_axis_cq_tkeep(3) = '1')then
              reg_tdata_ben(15 downto 12) <= m_axis_cq_tuser(23 downto 20);
            else
              reg_tdata_ben(15 downto 12) <= "0000";
            end if;
            if(m_axis_cq_tkeep(2) = '1')then
              reg_tdata_ben(11 downto 8) <= m_axis_cq_tuser(19 downto 16);
            else
              reg_tdata_ben(11 downto 8) <= "0000";
            end if;
            if(m_axis_cq_tkeep(1) = '1')then
              reg_tdata_ben(7 downto 4) <= m_axis_cq_tuser(15 downto 12);
            else
              reg_tdata_ben(7 downto 4) <= "0000";
            end if;
            if(m_axis_cq_tkeep(0) = '1')then
              reg_tdata_ben(3 downto 0) <= m_axis_cq_tuser(11 downto 8);
            else
              reg_tdata_ben(3 downto 0) <= "0000";
            end if;
          else
            reg_tdata_ben <= m_axis_cq_tuser(datawidthRX/8+7 downto 8);
          end if;
          if(m_axis_cq_tuser(40) = '1')then
            reg_data_valid  <= '0';
            reg_cmd_valid   <= '1';
          else
            reg_data_valid  <= '1';
            reg_cmd_valid   <= '0';
          end if;
        else
          reg_data_valid  <= '0';
          reg_cmd_valid   <= '0';
        end if;
      end if;
    end if;
  end process reg_cq_tdata;
end generate reg_data_128;


-- ------------------------------------------------------------------------
-- decode whether the bridge can accept the next request
-- The descriptor shall be registered so the BARS can be decoded but
-- the data will be held off until the descriptor can be transferred
-- to the desired BAR
-- not that the barX_full flags shall be active until the command has been
-- transferred AND there is space for a maximum length write
-- ------------------------------------------------------------------------
-- gen_tready : process (pcie_clk)
-- begin
--   if pcie_clk'event and pcie_clk = '1' then
--     if( pcie_rst = '1')then
--       m_axis_cq_tready   <= '0';
--     else
--       case bar_detect is
--         when "0001" =>  if(bar0_full = '1')then
--                           m_axis_cq_tready <= '0';
--                         else
--                           m_axis_cq_tready <= '1';
--                         end if;
--         when "0010" =>  if(bar1_full = '1')then
--                           m_axis_cq_tready <= '0';
--                         else
--                           m_axis_cq_tready <= '1';
--                         end if;
--         when "0100" =>  if(bar2_full = '1')then
--                           m_axis_cq_tready <= '0';
--                         else
--                           m_axis_cq_tready <= '1';
--                         end if;
-- --         when "1000" =>  if(bar4_full = '1')then
-- --                           m_axis_cq_tready <= '0';
-- --                         else
-- --                           m_axis_cq_tready <= '1';
-- --                         end if;
--         when others =>  m_axis_cq_tready <= '1';
--       end case;
--     end if;
--   end if;
-- end process gen_tready;
-- 
-- ------------------------------------------------------------------------
-- Direct Packets to decoded destinations
-- splitting the data and command descriptors in the process
-- ------------------------------------------------------------------------
split_bars : process (reg_data_valid,bar_detect,reg_cmd_valid)
begin
  push_bar0_cmd <= '0';
  push_bar1_cmd <= '0';
  push_bar2_cmd <= '0';
  case bar_detect is
    when "0001" => push_bar0_cmd <= reg_cmd_valid;
    when "0010" => push_bar1_cmd <= reg_cmd_valid;
    when "0100" => push_bar2_cmd <= reg_cmd_valid;
    when "1000" => 
--      assert false report "Illegal BAR Decode Detected" severity error;
    when others => push_bar0_cmd <= '0';
  end case;
  push_bar0_data <= '0';
  push_bar1_data <= '0';
  push_bar2_data <= '0';
  case bar_detect is
    when "0001" => push_bar0_data <= reg_data_valid;
    when "0010" => push_bar1_data <= reg_data_valid;
    when "0100" => push_bar2_data <= reg_data_valid;
    when "1000" => 
--      assert false report "Illegal BAR Decode Detected" severity error;
    when others => push_bar0_data <= '0';
  end case;
end process split_bars;

bar0 : pcieg3_bar_decoder
  generic map(
    datawidthRX => datawidthRX,
    datawidthTX => datawidthTX,
    single_tag_request => false
    )
  port map(
    pcie_clk              => pcie_clk,
    pcie_rst              => pcie_rst,
    bar_mask              => bar_mask(0),
    bar_page              => bar_page(0),
   
    data                  => reg_tdata,
    data_valid            => reg_tdata_valid,
    data_ben              => reg_tdata_ben,
    data_user             => m_axis_cq_tuser,
    cmd_push              => push_bar0_cmd,
    data_push             => push_bar0_data,

    rx_data_out           => rx_bar0_data_out,
    rx_data_pop           => rx_bar0_data_pop,
    rx_data_pkt_rdy       => rx_bar0_data_pkt_rdy,

    bar_full              => bar0_full,
    m_m2s_o               => m_m2s_o(0),
    m_s2m_i               => m_s2m_i(0)
    );




bar1 : pcieg3_bar_decoder
  generic map(
    datawidthRX => datawidthRX,
    datawidthTX => datawidthTX,
    single_tag_request => false
    )
  port map(
    pcie_clk              => pcie_clk,
    pcie_rst              => pcie_rst,
    bar_mask              => bar_mask(1),
    bar_page              => bar_page(1),
    data                  => reg_tdata,
    data_valid            => reg_tdata_valid,
    data_ben              => reg_tdata_ben,
    data_user             => m_axis_cq_tuser,
    cmd_push              => push_bar1_cmd,
    data_push             => push_bar1_data,
    bar_full              => bar1_full,

    rx_data_out           => rx_bar1_data_out,
    rx_data_pop           => rx_bar1_data_pop,
    rx_data_pkt_rdy       => rx_bar1_data_pkt_rdy,

    m_m2s_o               => m_m2s_o(1),
    m_s2m_i               => m_s2m_i(1)
    );


bar23 : pcieg3_bar2_decoder
  generic map(
    max_timeout => max_timeout,
    datawidthRX => 256,
    datawidthTX => 256,
    single_tag_request => false
    )
  port map(
    pcie_clk              => pcie_clk,
    pcie_rst              => pcie_rst,
    us_strobe             => us_strobe,
    bar_mask              => bar_mask(2),
    bar_page              => bar_page(2),
    data                  => reg_tdata,
    data_valid            => reg_tdata_valid,
    data_ben              => reg_tdata_ben,
    data_user             => m_axis_cq_tuser,
    cmd_push              => push_bar2_cmd,
    data_push             => push_bar2_data,
    bar_full              => bar2_full,

 
    rx_data_out           => rx_bar2_data_out,
    rx_data_pop           => rx_bar2_data_pop,
    rx_data_pkt_rdy       => rx_bar2_data_pkt_rdy,

    m_m2s_o               => m_m2s_ow,
    m_s2m_i               => m_s2m_iw,
    timeout               => timeout_bar23,
    clr_timeout          => clr_timeout_bar23
    );

 arbiter : pcieg3_bar_arbiter
  generic map(
    datawidthTX => datawidthTX
    )
  port map (
    pcie_clk              => pcie_clk,
    pcie_rst              => pcie_rst,
    rx_bar0_data_out      => rx_bar0_data_out,
    rx_bar0_data_pop      => rx_bar0_data_pop,
    rx_bar0_data_pkt_rdy  => rx_bar0_data_pkt_rdy,
    rx_bar1_data_out      => rx_bar1_data_out,
    rx_bar1_data_pop      => rx_bar1_data_pop,
    rx_bar1_data_pkt_rdy  => rx_bar1_data_pkt_rdy,
    rx_bar2_data_out      => rx_bar2_data_out,
    rx_bar2_data_pop      => rx_bar2_data_pop,
    rx_bar2_data_pkt_rdy  => rx_bar2_data_pkt_rdy,


    rx_bar_data_out       => rx_bar_data_out,
    rx_bar_data_push      => rx_bar_data_push,
    rx_bar_data_space     => rx_bar_data_space

    );

cc_if : pcieg3_cc_interface
  generic map(
    datawidthTX => datawidthTX
    )
  port map(
    pcie_clk              => pcie_clk,
    pcie_rst              => pcie_rst,
    rx_bar_data_out       => rx_bar_data_out,
    rx_bar_data_push      => rx_bar_data_push,
    rx_bar_data_space     => rx_bar_data_space,
    cc_b2c                => cc_b2c,
    cc_c2b                => cc_c2b

    );



end architecture;
