--
-- Small Dual Port RAM Synchronous FIFO
--
-- Configurable data width, fixed depth of 16
--

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_misc.all;

entity small_sfifo_reg_op is
  
  generic (
    width : integer := 64);

  port (
    clk    : in  std_logic;
    rst    : in  std_logic;
    wadv   : in  std_logic;
    wdata  : in  std_logic_vector(width-1 downto 0);
    nfull  : out std_logic;
    radv   : in  std_logic;
    rempty : out std_logic;
    rdata  : out std_logic_vector(width-1 downto 0));

end small_sfifo_reg_op;

architecture rtl of small_sfifo_reg_op is

  type fifo_type is array (0 to 15) of std_logic_vector(width-1 downto 0);

  signal ofifo_regs             : fifo_type                          := (others => (others => '0'));
  signal ofifo_rptr, ofifo_wptr : std_logic_vector(3 downto 0)       := "0000";
  signal ofifo_level            : std_logic_vector(4 downto 0)       := "00000";
  signal ofifo_out, ofifo_in    : std_logic_vector(width-1 downto 0) := (others => '0');
  signal ofifo_empty            : std_logic                          := '1';
  signal ofifo_wr, ofifo_rd     : std_logic                          := '0';

  attribute ram_style               : string;
  attribute ram_style of ofifo_regs : signal is "block";
  
begin  -- rtl

  ofifo_wr <= wadv;
  ofifo_in <= wdata;


  ofifo : process (clk)
  begin  -- process ofifo
    if clk'event and clk = '1' then     -- rising clock edge
      if rst = '1' then                 -- asynchronous reset
        ofifo_level <= "00000";
        ofifo_rptr  <= "0000";
        ofifo_wptr  <= "0000";
        ofifo_empty <= '1';
        nfull       <= '0';
      else
        if ofifo_wr = '1' and ofifo_rd = '0' then
          ofifo_level <= ofifo_level+1;
          ofifo_empty <= '0';
        elsif ofifo_wr = '0' and ofifo_rd = '1' then
          ofifo_level <= ofifo_level-1;
          if ofifo_level = "00001" then
            ofifo_empty <= '1';
          end if;
        end if;
        if ofifo_wr = '1' then
          ofifo_regs(conv_integer(ofifo_wptr)) <= ofifo_in;
          ofifo_wptr                           <= ofifo_wptr+1;
        end if;
        if ofifo_rd = '1' and ofifo_level /= "00000" then
          ofifo_rptr <= ofifo_rptr+1;
        end if;
        nfull <= ofifo_level(3) or ofifo_level(4);

        -- Register Output to Allow BRAM Synthesis
        ofifo_out <= ofifo_regs(conv_integer(ofifo_rptr));
      end if;
    end if;

  end process ofifo;

  ofifo_rd <= radv and not ofifo_empty;
  rempty   <= ofifo_empty;
  rdata    <= ofifo_out;
  
end rtl;

