--------------------------------------------------------------------------------
--
-- Module Name      : mptl128_to_adb3_ocp_resp
--
--                    Copyright 2009 by
--                    Alpha Data Parallel Systems Ltd.
--                    All Rights Reserved.
--
-- Original Author  : Andrew McCormick
-- Created          : 
--
-- Modified By      : 
-- Date             :
-- Change Notes     :     
--
-- Description      :  Bridge module for converting MPTL packets into ADB3_OCP slave responses
--
--
-- Dependencies     : adb3_ocp, mptl
--
-- Disclaimer       : THESE DESIGNS ARE PROVIDED "AS IS" WITH NO WARRANTY
--                    WHATSOEVER AND ALPHA DATA SPECIFICALLY DISCLAIMS ANY
--                    WARRANTIES IMPLIED OF MERCHANTABILITY, FITNESS FOR A
--                    PARTICULAR PURPOSE, OR AGAINST INFRINGEMENT.
--
--------------------------------------------------------------------------------


library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_misc.all;

library work;
use work.adb3_ocp.all;
use work.mptl.all;

entity mptl128_to_adb3_ocp_resp is

  generic (
    fifo_order : natural := 9);
  port (
    -- OCP Interface
    ocp_clk            : in  std_logic;
    ocp_m2s            : in  adb3_ocp_m2sT;
    ocp_s2m            : out adb3_ocp_s2mT;
    -- MPTL interface
    mptl128_data        : in  std_logic_vector(127 downto 0);
    mptl128_valid       : in  std_logic;
    online             : in  std_logic;
--    debug : out std_logic_vector(7 downto 0);
    -- Last read from FIFO
    remote_last_packet : out std_logic_vector(7 downto 0):= (others => '0'));


end mptl128_to_adb3_ocp_resp;

architecture rtl of mptl128_to_adb3_ocp_resp is
  
  type state_type is (idle, parse_burst_data);
  signal state         : state_type := idle;
 
  signal radv   : std_logic;
  signal rempty : std_logic;
  signal rdata  : std_logic_vector(127 downto 0) := (others => '0');
  signal radv_reg   : std_logic;

  signal packet_length : std_logic_vector(4 downto 0);

  signal resp : std_logic_vector(1 downto 0) := OCP_RESP_NONE;
  signal tag  : std_logic_vector(ADB3_OCP_TAG_WIDTH-1 downto 0);
 

  
  component dpram_sfifo
    generic (
      width : integer;
      order : natural);
    port (
      clk    : in  std_logic;
      rst    : in  std_logic;
      wadv   : in  std_logic;
      wdata  : in  std_logic_vector(width-1 downto 0);
      nfull  : out std_logic;
      radv   : in  std_logic;
      rempty : out std_logic;
      rdata  : out std_logic_vector(width-1 downto 0));
  end component;

  signal mptl128_valid_reg : std_logic := '0';
  signal mptl128_data_reg : std_logic_vector(127 downto 0) := (others => '0');

  signal fifo_rst : std_logic := '1';
  signal fifo_rst1 : std_logic := '1';
  signal fifo_rst2 : std_logic := '1';
  
  component small_sfifo32 is
    generic (
      width : integer);
    port (
      clk    : in  std_logic;
      rst    : in  std_logic;
      wadv   : in  std_logic;
      wdata  : in  std_logic_vector(width-1 downto 0);
      nfull  : out std_logic;
      radv   : in  std_logic;
      rempty : out std_logic;
      rdata  : out std_logic_vector(width-1 downto 0));
  end component small_sfifo32;


  signal wadv1   : std_logic;
  signal wdata1  : std_logic_vector(128+8-1 downto 0);
  signal nfull1  : std_logic;
  signal nfull1_reg  : std_logic := '0';
  signal radv1   : std_logic;
  signal rempty1 : std_logic;
  signal rdata1  : std_logic_vector(128+8-1 downto 0);

  signal dps_nfull : std_logic;
  
  attribute shreg_extract : string;
  attribute shreg_extract of mptl128_valid_reg : signal is "no";
  attribute shreg_extract of mptl128_data_reg : signal is "no";
  
begin  -- rtl


  -- Don't drive cmd/data accept 
  ocp_s2m.Data <= rdata1(127 downto 0);
  ocp_s2m.Resp <= resp;
  ocp_s2m.Tag  <= rdata1(135 downto 128);
  ocp_s2m.CmdAccept <= 'Z';
  ocp_s2m.DataAccept <= 'Z';

  reg_mptl64_data0: process (ocp_clk)
  begin  -- process reg_mptl64_data0
    if ocp_clk'event and ocp_clk = '1' then 
      mptl128_valid_reg <= mptl128_valid;
      mptl128_data_reg <= mptl128_data;
--       fifo_rst <= fifo_rst1;
--       fifo_rst1 <= fifo_rst2;
--       fifo_rst2 <= not online;     
    end if;
  end process reg_mptl64_data0;

  fifo_rst <= not online;
  
  input_sfifo0 : dpram_sfifo
    generic map (
      width => 128,
      order => fifo_order)
    port map (
      clk    => ocp_clk,
      rst    => fifo_rst,
      wadv   => mptl128_valid_reg,
      wdata  => mptl128_data_reg,
      nfull  => dps_nfull,
      radv   => radv,
      rempty => rempty,
      rdata  => rdata);

  -- FSM
  fsm0 : process (ocp_clk)
  begin  -- process fsm0
    if ocp_clk'event and ocp_clk = '1' then  -- rising clock edge
      case state is
        when idle =>
          if rempty = '0' then
            if rdata(3 downto 0) = MPTL_READ_DATA then
              state <= parse_burst_data;
              remote_last_packet <= rdata(23 downto 16);            
            end if;
            packet_length      <= rdata(12 downto 8);            
            tag <= rdata(47 downto 40);
          end if;
          if online = '0' then
            remote_last_packet <= (others => '0'); 
          end if;
        when parse_burst_data =>
          if rempty = '0' or online = '0' then
            if packet_length = "00010" then
              if nfull1_reg = '0' then
                state <= idle;
              end if;
            else
              if nfull1_reg = '0' then
                packet_length <= packet_length-1;
              end if;    
            end if;
          end if;
          
        when others =>
          state <= idle;
      end case;
    end if;
  end process fsm0;

  fsm_outputs0 : process (state, rempty, ocp_m2s,rdata,packet_length, online,nfull1_reg)
  begin  -- process fsm_outputs0
    -- Default value
    radv  <= '0';
    wadv1 <= '0';
    case state is
      when idle =>
        if rempty = '0' and online = '1' then                
          radv <= '1';            
        end if;
      when parse_burst_data =>
        if rempty = '0' then
          if nfull1_reg = '0' then
             radv <= '1';
             wadv1 <= '1';
          end if;
        end if;
      when others =>
        wadv1 <= '0';
        radv       <= '0';
    end case; 
  end process fsm_outputs0;


  process (ocp_clk) is
  begin  -- process
    if rising_edge(ocp_clk) then        -- rising clock edge
      nfull1_reg <= nfull1;   
    end if;
  end process;

  wdata1 <= tag & rdata; 

  small_sfifo_1: small_sfifo32
    generic map (
      width => 136)
    port map (
      clk    => ocp_clk,
      rst    => fifo_rst,
      wadv   => wadv1,
      wdata  => wdata1,
      nfull  => nfull1,
      radv   => radv1,
      rempty => rempty1,
      rdata  => rdata1);

  resp <= '0' & not rempty1;
  radv1 <= ocp_m2s.RespAccept and not rempty1;
  
--   debug(0) <= nfull1_reg;
--   debug(1) <= rempty1;
--   debug(2) <= radv1;
--   debug(3) <= wadv1;
--   debug(4) <= '1' when state = idle else '0';
--   debug(5) <= radv;
--   debug(6) <= mptl128_valid_reg;
--   debug(7) <= dps_nfull;
  

  
end rtl;
