--
-- dpbram0.vhd - entity "dpbram0"
--
-- SYNTHESIZABLE
--
-- Parameterisable dual port BlockRAM
--
-- Constraints on parameters
-- =========================
--
-- 1.  'width0' must be less than or equal to 'width1'
-- 2.  'width0' and 'width1' must be valid port data widths for a BlockRAM
--     of the specified family
-- 3.  'order0' and 'order1' must be valid port address widths for a 
--     BlockRAM of the specified family
-- 4.  The number of bits of BlockRAM storage required by the wider of the
--     two ports must be valid for a BlockRAM of the specified family.
-- 5.  The length of 'init' must the same as the number of bits of BlockRAM
--     sotrage required by the wider of the two ports.
--

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_misc.all;

library work;
use work.dpmem.all;

entity dpbram0 is
  generic(
    order0      : in  natural;           -- number of address bits in port 0
    width0      : in  natural;           -- width of port 0
    order1      : in  natural;           -- number of address bits in port 1
    width1      : in  natural;           -- width of port 1
    -- synopsys translate_off
    init        : in  std_logic_vector;  -- initial contents of RAM
    -- synopsys translate_on
    family      : in  family_t);         -- FPGA family to target
  port(
    rst : in  std_logic;
    k0  : in  std_logic;
    sr0 : in  std_logic;
    en0 : in  std_logic;
    w0  : in  std_logic;
    a0  : in  std_logic_vector(order0 - 1 downto 0);
    d0  : in  std_logic_vector(width0 - 1 downto 0);
    q0  : out std_logic_vector(width0 - 1 downto 0);
    k1  : in  std_logic;
    sr1 : in  std_logic;
    en1 : in  std_logic;
    w1  : in  std_logic;
    a1  : in  std_logic_vector(order1 - 1 downto 0);
    d1  : in  std_logic_vector(width1 - 1 downto 0);
    q1  : out std_logic_vector(width1 - 1 downto 0));
end;

architecture infer of dpbram0 is

  type ram_type is array ((2**order0)-1 downto 0) of std_logic_vector(width0-1 downto 0);
  shared variable RAM : ram_type;

begin  -- architecture infer

  assert order0 = order1 report "Only equal depth memory ports supported" severity error;
  assert width0 = width1 report "Only equal width memory ports supported" severity error;
  
  process (k0)
  begin
    if rising_edge(k0) then
      if EN0 = '1' then
        if W0 = '1' then
          RAM(conv_integer(A0)) := D0;
        end if;
        Q0 <= RAM(conv_integer(A0));
      end if;
    end if;
  end process;
  
  process (k1)
  begin
    if rising_edge(k1) then
      if EN1 = '1' then
        if W1 = '1' then
          RAM(conv_integer(A1)) := D1;
        end if;
        Q1 <= RAM(conv_integer(A1));
      end if;
    end if;
  end process;

  

end architecture infer;
