--------------------------------------------------------------------------------
--
-- Module Name      : adb3_ocp_to_mptl128
--
--                    Copyright 2009 by
--                    Alpha Data Parallel Systems Ltd.
--                    All Rights Reserved.
--
-- Original Author  : Andrew McCormick
-- Created          : 
--
-- Modified By      : 
-- Date             :
-- Change Notes     :     
--
-- Description      : Bridge module for converting ADB3_OCP transactions into MPTL packets
--
--
-- Dependencies     : adb3_ocp, mptl
--
-- Disclaimer       : THESE DESIGNS ARE PROVIDED "AS IS" WITH NO WARRANTY
--                    WHATSOEVER AND ALPHA DATA SPECIFICALLY DISCLAIMS ANY
--                    WARRANTIES IMPLIED OF MERCHANTABILITY, FITNESS FOR A
--                    PARTICULAR PURPOSE, OR AGAINST INFRINGEMENT.
--
--------------------------------------------------------------------------------



library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_misc.all;

library work;
use work.adb3_ocp.all;
use work.mptl.all;

entity adb3_ocp_to_mptl128 is

  generic (
    channel : integer := 0;
    limit_inflight_packets : boolean := false);
  port (
    -- OCP Interface
    ocp_clk            : in  std_logic;
    ocp_m2s            : in  adb3_ocp_m2sT;
    ocp_s2m            : out adb3_ocp_s2mT;
    -- MPTL interface
    mptl128_data        : out std_logic_vector(127 downto 0);
    mptl128_valid       : out std_logic;
    mptl128_accept      : in  std_logic;
    online             : in  std_logic;
    -- Last packet consumed by
    remote_last_packet : in  std_logic_vector(7 downto 0));

end adb3_ocp_to_mptl128;

architecture rtl of adb3_ocp_to_mptl128 is

  constant IMPROVE_TIMING_ON_MPTL128 : boolean := true; 

  signal current_packet    : std_logic_vector(7 downto 0) := (others => '0');
  signal remote_fifo_level : std_logic_vector(7 downto 0) := (others => '0');
  signal can_accept_data   : std_logic                    := '1';
  signal can_accept_data_q   : std_logic                    := '1';

  type state_type is (idle, send_read_request, send_single_write, send_burst_cmd, send_burst_data, send_burst_data_wait);
  signal state         : state_type := idle;
  signal state_is_idle : std_logic := '1';
  signal beat_counter2 : std_logic;
  signal beat_counter4 : std_logic_vector(1 downto 0);
  signal beat_counter  : std_logic_vector(4 downto 0);

  signal cmd_accept, data_accept : std_logic;

  signal packet_header, next_packet_header, prev_packet_header : std_logic_vector(63 downto 0);
  signal mptl128_hvalid, prev_mptl128_hvalid                     : std_logic;

  signal mptl128_avalid, mptl128_dvalid : std_logic := '0';

  signal burst_length_left           : std_logic_vector(ADB3_OCP_BURST_WIDTH-1 downto 0) := (others => '0');
  signal next_burst_packet_size      : std_logic_vector(4 downto 0)                    := (others => '0');
  signal beat_limit, beat_limit_calc : std_logic_vector(4 downto 0)                    := (others => '0');
  signal next_burst_last, last_burst : std_logic                                       := '0';
  signal be_latched                  : std_logic                                       := '0';

  signal packet_addr          : std_logic_vector(63 downto 0) := (others => '0');
  signal actual_packet_header : std_logic_vector(63 downto 0) := (others => '0');

  signal mptl128_data_i       : std_logic_vector(127 downto 0);
  signal mptl128_valid_i      : std_logic := '0';
  signal mptl128_accept_i     : std_logic := '0';

  signal mptl128_hvalid_q : std_logic := '0';

  signal header_chksum : std_logic_vector(3 downto 0);

  signal burst_accepted : std_logic := '0';

  signal online_r1 : std_logic := '0';

  
begin  -- rtl
  -- Don't drive response dataflow path 
  ocp_s2m.Data       <= (others => 'Z');
  ocp_s2m.Resp       <= (others => 'Z');
  ocp_s2m.Tag        <= (others => 'Z');
  ocp_s2m.CmdAccept  <= cmd_accept;
  ocp_s2m.DataAccept <= data_accept;

  -- Keep track of remote FIFO level
  remote_fifo_level0 : process (ocp_clk)
  begin  -- process remote_fifo_level0
    if ocp_clk'event and ocp_clk = '1' then  -- rising clock edge
      -- Compute number of packets in flight or in remote FIFO     
      remote_fifo_level <= current_packet-remote_last_packet;    
      if limit_inflight_packets then
        -- Allow only 3 packets in flight
        can_accept_data  <=((not remote_fifo_level(4)) and (not remote_fifo_level(3)) and (not remote_fifo_level(2))) and online_r1;
      else
        -- Allow up to 16 packets in flight
        can_accept_data   <= (not remote_fifo_level(4)) and online_r1;
      end if;


      if online_r1 = '0' then
        current_packet <= (others => '0');
      elsif mptl128_hvalid = '1' and mptl128_accept_i = '1' then
        current_packet <= current_packet+1;
      end if;

     
    end if;
  end process remote_fifo_level0;

  

  -- FSM
  fsm0 : process (ocp_clk)
  begin  -- process fsm0
    if ocp_clk'event and ocp_clk = '1' then  -- rising clock edge
      -- Default Values (overwritten)
      mptl128_hvalid <= '0';
      mptl128_dvalid <= '0';
      online_r1 <= online;
      
      case state is
        when idle =>
          state_is_idle <= '1';
          if can_accept_data = '1' and online_r1 = '1' then
            if ocp_m2s.Cmd = OCP_CMD_WRITE then
              if ocp_m2s.BurstLength = CONV_STD_LOGIC_VECTOR(1, ADB3_OCP_BURST_WIDTH) then
                state         <= send_single_write;
                mptl128_hvalid <=  ocp_m2s.DataValid;
                state_is_idle <= '0';
              else
                state         <= send_burst_cmd;
                mptl128_hvalid <= ocp_m2s.DataValid;
                state_is_idle <= '0';
              end if;
            elsif ocp_m2s.Cmd = OCP_CMD_READ then
              state         <= send_read_request;
              mptl128_hvalid <= '1';
              state_is_idle <= '0';
            end if;
          end if;
          beat_counter  <= (others => '0');
          beat_counter4 <= (others => '0');
        when send_read_request =>
          if mptl128_accept_i = '1' then
            state <= idle;
            state_is_idle <= '1';
          else                    
            mptl128_hvalid <= '1';
          end if;
          if online_r1 = '0' then
            state <= idle;
          end if;
        when send_single_write =>
          if beat_counter4 = "00" then                                  
            if be_latched = '1' and mptl128_accept_i = '1' then
              beat_counter4 <= "01";
              mptl128_dvalid <= '1';
            else
              mptl128_hvalid <= be_latched or ocp_m2s.DataValid;
            end if;
          else           
            state         <= idle;
            state_is_idle <= '1';
          end if;
          if online_r1 = '0' then
            state <= idle;
          end if;
        when send_burst_cmd =>          
          if be_latched = '1' and mptl128_accept_i = '1' then
--             if not IMPROVE_TIMING_ON_MPTL128 then
--               state         <= send_burst_data;
--               beat_counter  <= (others => '0');
--               mptl128_hvalid <= '1';
--             else
              state         <= send_burst_data_wait;
              beat_counter  <= (others => '0');              
--            end if;
          else
            mptl128_hvalid <= be_latched or ocp_m2s.DataValid;
          end if;
          if online_r1 = '0' then
            state <= idle;
          end if;
        when send_burst_data_wait =>
          if can_accept_data = '1' then
            state         <= send_burst_data;
            beat_counter  <= (others => '0');
            mptl128_hvalid <= '1';
          end if;
          if online_r1 = '0' then
            state <= idle;
          end if;
        when send_burst_data =>
          if mptl128_hvalid = '1' then
            if mptl128_accept_i = '0' then
              mptl128_hvalid <= '1';
            else
              mptl128_dvalid <= '1';
              beat_counter <= beat_counter+1;
            end if;
          else
            mptl128_dvalid <= '1';
            if ocp_m2s.DataValid = '1' then            
              if beat_counter = beat_limit then
                if last_burst = '1' then
                  state         <= idle;
                  state_is_idle <= '1';
                  mptl128_dvalid <= '0';
                else
                  if not IMPROVE_TIMING_ON_MPTL128 then
                    if can_accept_data = '1' then
                      state         <= send_burst_data;
                      beat_counter  <= (others => '0');
                      mptl128_hvalid <= '1';
                    else
                      state         <= send_burst_data_wait;
                      beat_counter  <= (others => '0');
                      mptl128_hvalid <= '0';
                    end if;
                  else
                    state         <= send_burst_data_wait;
                    beat_counter  <= (others => '0');
                    mptl128_hvalid <= '0';
                  end if;
                  mptl128_dvalid <= '0';
                end if;
              else
                beat_counter <= beat_counter+1;
              end if;
            end if;          
          end if;
          if online_r1 = '0' then
            state <= idle;
          end if;
        when others =>
          state <= idle;
          state_is_idle <= '1';
      end case;
    end if;
  end process fsm0;

  fsm_outputs0 : process (state, state_is_idle, beat_counter, beat_counter4, can_accept_data, mptl128_hvalid)
  begin  -- process fsm_outputs0
    -- Control CmdAccept
    if state_is_idle = '1' then
      cmd_accept <= can_accept_data;
    else
      cmd_accept <= '0';
    end if;
    -- Control DataAccept
    if state = send_single_write and beat_counter4 = "01" then
      data_accept <= '1';
    elsif state = send_burst_data and mptl128_hvalid = '0' then
      data_accept <= '1';
    else
      data_accept <= '0';
    end if;
  end process fsm_outputs0;


  mptl128_valid_i <= mptl128_hvalid or (mptl128_dvalid and ocp_m2s.DataValid);

  mptl128_data_i <= packet_addr & actual_packet_header when mptl128_hvalid = '1' else               
                   ocp_m2s.Data;

  gen_better_timing : if IMPROVE_TIMING_ON_MPTL128 generate
    -- Register MPTL Output
    -- Improves timing, but at cost of 1 clock cycle 
    reg_mptl0 : process (ocp_clk)
    begin  -- process reg_mptl0
      if ocp_clk'event and ocp_clk = '1' then  -- rising clock edge
        if mptl128_valid_i = '1' then
          mptl128_data <= mptl128_data_i;         
        end if;
        mptl128_valid <= (mptl128_hvalid and not mptl128_accept_i) or (mptl128_dvalid and ocp_m2s.DataValid);     
      end if;
    end process reg_mptl0;
  end generate gen_better_timing;

  gen_better_throughput : if not IMPROVE_TIMING_ON_MPTL128 generate
    mptl128_data  <= mptl128_data_i;
    mptl128_valid <= mptl128_valid_i;
  end generate gen_better_throughput;

  mptl128_accept_i <= mptl128_accept;

  -- packet header only valid for 1 clock cycle, if not instantly accapted,
  -- use latched version
  actual_packet_header <= packet_header when prev_mptl128_hvalid = '0' else
                          prev_packet_header;

  -- Select Packet Header contents before registering
  -- Packet word (header or data) selected after registering
  next_packet_hdr : process (current_packet, next_burst_packet_size,
                            ocp_m2s, state)
  begin  -- process next_packet_hdr
    next_packet_header               <= (others => '0');  -- Assign default values if not
                                        -- overwritten in process
    -- Set Fixed Header values for all packets 
    next_packet_header(7 downto 4)   <= CONV_STD_LOGIC_VECTOR(channel, 4);  -- CHANNEL
    next_packet_header(23 downto 16) <= current_packet;   -- PACKET_ID

    if state = send_burst_data or state = send_burst_cmd or state = send_burst_data_wait then
      next_packet_header(3 downto 0)   <= MPTL_BURST_WRITE_DATA;     -- TYPE
      next_packet_header(12 downto 8)  <= next_burst_packet_size;
      next_packet_header(47 downto 40) <= ocp_m2s.Tag;
    elsif ocp_m2s.Cmd = OCP_CMD_READ then
      next_packet_header(3 downto 0)   <= MPTL_READ_REQUEST;         -- TYPE
      next_packet_header(12 downto 8)  <= CONV_STD_LOGIC_VECTOR(MPTL128_READ_REQUEST_PACKET_SIZE, 5);
      next_packet_header(35 downto 24) <= ocp_m2s.BurstLength;
      next_packet_header(47 downto 40) <= ocp_m2s.Tag;
      next_packet_header(63 downto 48) <= ocp_m2s.Addr(63 downto 48);
    elsif ocp_m2s.Cmd = OCP_CMD_WRITE then
      if ocp_m2s.BurstLength = CONV_STD_LOGIC_VECTOR(1, ADB3_OCP_BURST_WIDTH) then
        next_packet_header(3 downto 0)   <= MPTL_SINGLE_BEAT_WRITE;  -- TYPE
        next_packet_header(12 downto 8)  <= CONV_STD_LOGIC_VECTOR(MPTL128_SINGLE_BEAT_WRITE_PACKET_SIZE, 5);
        next_packet_header(35 downto 24) <= ocp_m2s.BurstLength;
        next_packet_header(47 downto 40) <= ocp_m2s.Tag;
        next_packet_header(63 downto 48) <= ocp_m2s.Addr(63 downto 48);
      else
        next_packet_header(3 downto 0)   <= MPTL_BURST_WRITE_CMD;    -- TYPE
        next_packet_header(12 downto 8)  <= CONV_STD_LOGIC_VECTOR(MPTL128_BURST_WRITE_CMD_PACKET_SIZE, 5);
        next_packet_header(35 downto 24) <= ocp_m2s.BurstLength;
        next_packet_header(47 downto 40) <= ocp_m2s.Tag;
        next_packet_header(63 downto 48) <= ocp_m2s.Addr(63 downto 48);
      end if;
    end if;
  end process next_packet_hdr;


  header_chksum <= next_packet_header(3 downto 0) xor next_packet_header(7 downto 4) xor next_packet_header(11 downto 8) xor next_packet_header(15 downto 12) xor
                   next_packet_header(19 downto 16) xor next_packet_header(23 downto 20) xor next_packet_header(27 downto 24) xor next_packet_header(31 downto 28) xor
                   next_packet_header(35 downto 32) xor next_packet_header(43 downto 40) xor next_packet_header(47 downto 44);
   
  
  -- Format Packet Header records
  phr0 : process (ocp_clk)
  begin  -- process phr0
    if ocp_clk'event and ocp_clk = '1' then  -- rising clock edge
      if IMPROVE_TIMING_ON_MPTL128 then
        if state = idle or state = send_burst_data_wait then
          packet_header <= next_packet_header(63 downto 40) & header_chksum & next_packet_header(35 downto 0);
        end if;   
      else
        if state = idle or state = send_burst_data_wait or (mptl128_hvalid = '0' and (state = send_read_request or state = send_burst_data)) then
          packet_header <= next_packet_header(63 downto 40) & header_chksum & next_packet_header(35 downto 0);
        end if;   
      end if;
      
      if mptl128_hvalid = '1' and prev_mptl128_hvalid = '0' then
        prev_packet_header <= packet_header;
      end if;
      prev_mptl128_hvalid <= mptl128_hvalid;

      if state = idle and ocp_m2s.Cmd /= OCP_CMD_IDLE then
        packet_addr(47 downto 0) <= ocp_m2s.Addr(47 downto 0);
        be_latched               <= ocp_m2s.DataValid;
      end if;
      if ocp_m2s.DataValid = '1' then
        packet_addr(63 downto 48) <= ocp_m2s.DataByteEn;
        be_latched                <= '1';
      end if;
      
    end if;
  end process phr0;

  -- Compute Burst Sizes
  burst_size0 : process (ocp_clk)
    variable nbps_var : std_logic_vector(11 downto 0);
    variable bl_var   : std_logic_vector(11 downto 0);
  begin  -- process burst_size0
    if ocp_clk'event and ocp_clk = '1' then  -- rising clock edge
      if state = send_burst_data and mptl128_accept_i = '1' and mptl128_hvalid = '1' then
        burst_accepted <= '1';
      else
        burst_accepted <= '0';
      end if;
      if state = idle then
        burst_length_left <= ocp_m2s.BurstLength;
      else
        if burst_accepted = '1' then
          if next_burst_last = '1' then
            burst_length_left <= (others => '0');
          else
            burst_length_left <= burst_length_left-MPTL128_MAX_BURST_PAYLOAD_SIZE;
          end if;
          last_burst <= next_burst_last;
          beat_limit <= beat_limit_calc;
        end if;
      end if;
      if burst_length_left > MPTL128_MAX_BURST_PAYLOAD_SIZE then
        next_burst_packet_size <= CONV_STD_LOGIC_VECTOR(MPTL128_MAX_BURST_PAYLOAD_PACKET_SIZE, 5);
        beat_limit_calc        <= CONV_STD_LOGIC_VECTOR(MPTL128_MAX_BURST_PAYLOAD_SIZE, 5);
        next_burst_last        <= '0';
      else
        nbps_var               := burst_length_left + 1;
        next_burst_packet_size <= nbps_var(4 downto 0);
        bl_var                 := burst_length_left;
        beat_limit_calc        <= bl_var(4 downto 0);
        next_burst_last        <= '1';
      end if;
      

    end if;
  end process burst_size0;

  
end rtl;
