--------------------------------------------------------------------------------
--
-- Module Name      : mptl128_gtp_buffer
--
--                    Copyright 2009 by
--                    Alpha Data Parallel Systems Ltd.
--                    All Rights Reserved.
--
-- Original Author  : Andrew McCormick
-- Created          : 
--
-- Modified By      : 
-- Date             :
-- Change Notes     :     
--
-- Description      :  
--
--
-- Dependencies     : fifo, mptl
--
-- Disclaimer       : THESE DESIGNS ARE PROVIDED "AS IS" WITH NO WARRANTY
--                    WHATSOEVER AND ALPHA DATA SPECIFICALLY DISCLAIMS ANY
--                    WARRANTIES IMPLIED OF MERCHANTABILITY, FITNESS FOR A
--                    PARTICULAR PURPOSE, OR AGAINST INFRINGEMENT.
--
--------------------------------------------------------------------------------



library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_misc.all;


Library UNISIM;
use UNISIM.vcomponents.all;
library UNIMACRO;  
use UNIMACRO.Vcomponents.all;


library work;
use work.mptl.all;
use work.fifo.all;

entity mptl128_gtp_buffer_macro is

  generic (
    four_gtps : boolean := true;
    DEVICE    : string  := "VIRTEX6");
  port (
    -- OCP Interface
    ocp_clk       : in  std_logic;
    -- MPTL Multi-port output interface
    mptl128_data  : out std_logic_vector(127 downto 0);
    mptl128_valid : out std_logic;
    -- GTP Outputs
    gtp_clk       : in  std_logic;
    gtp_data      : in  std_logic_vector(127 downto 0);
    gtp_charisk   : in  std_logic_vector(15 downto 0);
    gtp_online    : out std_logic := '0';
    tx_offline    : in  std_logic;
    error8b10b    : in  std_logic;
    debug         : out std_Logic_vector(139 downto 0);
    mptl_retry    : out std_logic);

end mptl128_gtp_buffer_macro;

architecture rtl of mptl128_gtp_buffer_macro is

  constant use_macro : boolean := false;
  constant K28_5 : std_logic_vector(7 downto 0) := X"BC";
  constant idleD : std_logic_vector(7 downto 0) := X"95";

  component small_sfifo32
    generic (
      width : integer := 64);
    port (
      clk    : in  std_logic;
      rst    : in  std_logic;
      wadv   : in  std_logic;
      wdata  : in  std_logic_vector(width-1 downto 0);
      nfull  : out std_logic;
      radv   : in  std_logic;
      rempty : out std_logic;
      rdata  : out std_logic_vector(width-1 downto 0));
  end component;

  signal beats_left : std_logic_vector(5 downto 0) := (others => '0');



  type fifo_data_type is array (0 to 7) of std_logic_vector(17 downto 0);
  signal wdata, rdata : fifo_data_type;
  signal wadv         : std_logic_vector(7 downto 0) := (others => '0');
  signal rempty       : std_logic_vector(7 downto 0);
  signal radv         : std_logic_vector(7 downto 0) := "00000000";

  signal w2adv   : std_logic := '0';
  signal w2adv_e : std_logic := '0';
  signal w2data  : std_logic_vector(127 downto 0);

  signal w3adv   : std_logic := '0';
  signal w3data  : std_logic_vector(127 downto 0);


  
  signal r2adv, r2empty : std_logic := '0';
  signal r2aempty : std_logic;
  signal r2data         : std_logic_vector(127 downto 0);

  signal r2adv1           : std_logic := '0';
  signal r2data1         : std_logic_vector(127 downto 0);

  signal online           : std_logic                     := '0';
  signal online_counters  : std_logic_vector(47 downto 0) := (others => '0');
  signal offline_counters : std_logic_vector(47 downto 0) := (others => '0');
  signal offline,offline1,offline2,offline3          : std_logic                     := '1';

  signal reverse_words : std_logic_vector(3 downto 0);
  

  signal gtp_data_r1      : std_logic_vector(127 downto 0);
  signal gtp_charisk_r1   : std_logic_vector(15 downto 0);

   signal gtp_data_reg      : std_logic_vector(127 downto 0);
  signal gtp_charisk_reg   : std_logic_vector(15 downto 0);

  signal gtp_phase : std_logic_vector(3 downto 0) := (others => '0');
  signal gtp_phase_changed : std_logic_vector(3 downto 0) := (others => '0');
  
  
  signal rst, rst_reg, rst_reg2 : std_logic := '0';

  attribute shreg_extract : string;
  attribute shreg_extract of r2data1 : signal is "no";
  attribute shreg_extract of r2adv1  : signal is "no";
  attribute shreg_extract of mptl128_data  : signal is "no";
  attribute shreg_extract of mptl128_valid  : signal is "no";
  attribute shreg_extract of gtp_data_r1 : signal is "no";
  attribute shreg_extract of gtp_charisk_r1 : signal is "no";

  signal channel_bond_rst : std_logic := '1';


  type retry_data_buffer_type is array (0 to 31) of std_logic_vector(127 downto 0);
  
  signal retry_data_buffer : retry_data_buffer_type := (others => (others => '0'));
  signal retry_adv_buffer : std_logic_vector(31 downto 0) := (others => '0');
  
  signal retrying : std_logic := '0';
  signal error8b10b_latch : std_logic := '0';
  
  signal error8b10b_latch_delayed : std_logic := '0';
  signal error8b10b_sr : std_logic_vector(31 downto 0) := (others => '0');
  signal packet_words_left : std_logic_vector(4 downto 0) := (others => '0');
  signal chksum : std_logic_vector(3 downto 0);
  
  signal last_good_packet_counter : std_logic_vector(7 downto 0):=(others => '0');
  signal last_good_packet : std_logic_vector(23 downto 0);
  
  signal mptl_retry_shift : std_logic_vector(7 downto 0);
  signal retry_flush_counter : std_logic_vector(6 downto 0) := (others => '0');

  signal chksum_err : std_logic := '0';

  signal EMPTY0,EMPTY1,ALMOSTFULL0,ALMOSTFULL1,ALMOSTEMPTY0,ALMOSTEMPTY1 : std_logic;
  signal RDCOUNT0,RDCOUNT1,WRCOUNT0,WRCOUNT1 : std_logic_vector(8 downto 0);

  signal cb_fail_counter : std_logic_vector(5 downto 0) := (others => '0');
  signal cb_fail : std_logic := '0';
  --signal cb_fail_multiple_counter : std_logic_vector(3 downto 0) := (others => '0');
  --signal cb_fail_multiple : std_logic := '0';

  signal error8b10bdet : std_logic_vector(7 downto 0);
  signal error8b10bdet_r1 : std_logic_vector(7 downto 0);
  signal error8b10bdet_r2 : std_logic_vector(7 downto 0);
  signal error8b10bdet_i : std_logic;

  signal bond_delay : std_logic_vector(6 downto 0);
  signal bond_delay_count_en : std_logic := '0';
  signal bond_delay_counted : std_logic := '0';

  signal fifo_reset : std_logic;

  signal dbg_counter : std_logic_vector(7 downto 0) := (others => '0');
  signal dbg_counter2 : std_logic_vector(7 downto 0) := (others => '0');

  signal online_delay_count : std_logic_vector(4 downto 0) := (others => '0');
  signal online_q : std_logic := '0';

  attribute ASYNC_REG : boolean;
  attribute ASYNC_REG of offline1,offline2: signal is TRUE;
  

  
begin  -- rtl

  reg_gtp_in : process (gtp_clk)
    --variable non_comma_detect : std_logic_vector(7 downto 0);
  begin  -- process reg_gtp_in
    if gtp_clk'event and gtp_clk = '1' then  -- rising clock edge
      
      error8b10bdet <= X"00";
      error8b10bdet_r1 <= error8b10bdet;
      error8b10bdet_r2 <= error8b10bdet_r1;

      gtp_data_reg <= gtp_data;
      gtp_charisk_reg <= gtp_charisk;

      gtp_phase_changed <= (others => '0');
      if online = '0' then
        
        for i in 0 to 3 loop      
          if gtp_phase_changed(i) = '0' and gtp_data_r1(32*i+15 downto 32*i) = K28_5 & idleD and gtp_charisk_r1(4*i+1 downto 4*i) = "10" then
            gtp_phase(i) <= not gtp_phase(i);
            gtp_phase_changed(i) <= '1';
          end if;
        end loop;  -- i       
      end if;

      for i in 0 to 3 loop
        if gtp_phase(i) = '0' then
          gtp_data_r1(32*i+31 downto 32*i) <= gtp_data(32*i+31 downto 32*i);
          gtp_charisk_r1(4*i+3 downto 4*i) <= gtp_charisk(4*i+3 downto 4*i);
        else
          gtp_data_r1(32*i+31 downto 32*i) <= gtp_data(32*i+23 downto 32*i) & gtp_data_reg(32*i+31 downto 32*i+24);
          gtp_charisk_r1(4*i+3 downto 4*i) <= gtp_charisk(4*i+2 downto 4*i) & gtp_charisk_reg(4*i+3 downto 4*i+3);
        end if;
      end loop;  -- i
      
      for i in 0 to 7 loop
        wdata(i) <= gtp_charisk_r1(2*i+1 downto 2*i) & gtp_data_r1(16*i+15 downto 16*i);
        if gtp_data_r1(16*i+15 downto 16*i) = idleD & K28_5 and gtp_charisk_r1(2*i+1 downto 2*i) = "01" then
          wadv(i) <= '0';
          if online_counters(6*i+5) = '0' then
            online_counters(6*i+5 downto 6*i) <= online_counters(6*i+5 downto 6*i)+1;
          end if;
          offline_counters <= (others => '0');
        else
          if offline = '1' then
            online_counters <= (others => '0');
          end if;
          if gtp_charisk_r1(2*i+1 downto 2*i) = "00" then
            wadv(i) <= online;
          else
            wadv(i) <= '0';
            if offline_counters(6*i+5) = '0' then
              offline_counters(6*i+5 downto 6*i) <= offline_counters(6*i+5 downto 6*i)+1;
            end if;
            --offline_counters(6*i+5 downto 6*i) <= (others => '1');
            error8b10bdet(i) <= '1';
          end if;
        end if;
      end loop;  -- i

      
      if tx_offline = '1' or cb_fail = '1' then
        online_counters  <= (others => '0');
        offline_counters <= (others => '0');
        online           <= '0';
        offline          <= '1';
      elsif (online_counters(5) and online_counters(11) and online_counters(17) and online_counters(23) and online_counters(29) and online_counters(35) and online_counters(41) and online_counters(47)) = '1' then
        online  <= '1';
        offline <= '0';
      elsif (offline_counters(5) or offline_counters(11) or offline_counters(17) or offline_counters(23) or offline_counters(29) or offline_counters(35) or offline_counters(41) or offline_counters(47)) = '1' or error8b10b = '1'then
        online  <= '0';
        offline <= '1';
      end if;
      gtp_online <= online;

      if online_q = '0' then 
        reverse_words <= (others => '0');
      else
        for i in 0 to 3 loop
          if gtp_charisk_r1(4*i+3 downto 4*i) = "0001" then
            reverse_words(i) <= '1';
          end if;
        end loop;  -- i
      end if;

      if online = '1' then
        if online_delay_count = "11111" then
          online_q <= '1';
        else
          if online_delay_count = "00000" then
            if gtp_charisk_r1 /= "0101010101010101" then
              online_delay_count <= "00001";
            end if;  
          else
            online_delay_count <= online_delay_count+1;
          end if;  
        end if;
      else
        online_q <= '0';
        online_delay_count <= (others => '0');
      end if;

      
      channel_bond_rst <= cb_fail or offline or (retry_flush_counter(6) and (retry_flush_counter(5) or retry_flush_counter(4) or retry_flush_counter(3)));-- or error8b10b_latch;

      

      if cb_fail = '0' and channel_bond_rst = '0' and rempty /= "11111111" and rempty /= "00000000" and gtp_data_r1 = (idleD & K28_5 & idleD & K28_5 & idleD & K28_5 & idleD & K28_5 & idleD & K28_5 & idleD & K28_5 & idleD & K28_5 & idleD & K28_5) then
        cb_fail_counter <= cb_fail_counter+1;
      else
        cb_fail_counter <= (others => '0');
      end if;
      cb_fail <= online and AND_reduce(cb_fail_counter);


      --if online = '1' then
      --  if cb_fail = '1' then
      --    cb_fail_multiple_counter <= cb_fail_multiple_counter+1;
      --  end if;
      --else
      --  cb_fail_multiple_counter <= (others => '0');
      --end if;
      --cb_fail_multiple <= online and AND_reduce(cb_fail_multiple_counter);

      
      --if channel_bond_rst = '1' then
      --  bond_delay_count_en <= '0';
      --  bond_delay_counted <= '0';
      --elsif bond_delay_counted = '0' then
      --  if bond_delay_count_en = '0' then
      --    if wadv /= "00000000" then
      --      bond_delay_count_en <= '1';
      --      bond_delay <= (others => '0');
      --    end if;
      --  else
      --    bond_delay <= bond_delay+1;
      --    if w2adv_e = '1' then
      --      bond_delay_counted <= '1';
      --    end if;
      --  end if;
      --     end if;
      
    end if;
  end process reg_gtp_in;

  error8b10bdet_i <= OR_reduce(error8b10bdet or error8b10bdet_r1 or error8b10bdet_r2);
  
  -- Channel Bonding FIFOs
  gen_sync_fifos : for i in 0 to 7 generate
    small_sfifo_i : small_sfifo32
      generic map (
        width => 18)
      port map (
        clk    => gtp_clk,
        rst    => channel_bond_rst,
        wadv   => wadv(i),
        wdata  => wdata(i),
        nfull  => open,
        radv   => radv(i),
        rempty => rempty(i),
        rdata  => rdata(i));
  end generate gen_sync_fifos;


  --radv    <= (not rempty) when offline = '1' else "11111111" when rempty = "00000000" else "00000000";
  radv    <= "11111111" when rempty = "00000000" else "00000000";
  w2adv_e <= '1'        when rempty = "00000000" else '0';

  transfer_data0 : process (gtp_clk)
  begin  -- process transfer_data0
    if gtp_clk'event and gtp_clk = '1' then  -- rising clock edge
      offline1 <= offline;
      w2adv  <= w2adv_e and not offline1;
      if four_gtps then
        for i in 0 to 3 loop
          if reverse_words(i) = '1' then
            w2data(32*i+31 downto 32*i) <= rdata(2*i)(15 downto 0) & rdata(2*i+1)(15 downto 0);
          else
            w2data(32*i+31 downto 32*i) <= rdata(2*i+1)(15 downto 0) & rdata(2*i)(15 downto 0);
          end if;
        end loop;  -- i
      else
        w2data <= rdata(7)(15 downto 0) & rdata(6)(15 downto 0) & rdata(5)(15 downto 0) & rdata(4)(15 downto 0) & rdata(3)(15 downto 0) & rdata(2)(15 downto 0) & rdata(1)(15 downto 0) & rdata(0)(15 downto 0);
      end if;
      
    end if;
  end process transfer_data0;

  --debug(127 downto 0) <= w2data;
  --debug(131 downto 128) <= reverse_words;
  --debug(139 downto 132) <= rempty;
-------------------------------------------------------------------------------
  -- Retry Control Logic
-------------------------------------------------------------------------------  

  --chksum <= retry_data_buffer(31)(3 downto 0) xor retry_data_buffer(31)(7 downto 4) xor retry_data_buffer(31)(11 downto 8) xor retry_data_buffer(31)(15 downto 12) xor
    --        retry_data_buffer(31)(19 downto 16) xor retry_data_buffer(31)(23 downto 20) xor retry_data_buffer(31)(27 downto 24) xor retry_data_buffer(31)(31 downto 28) xor
    --        retry_data_buffer(31)(35 downto 32) xor retry_data_buffer(31)(43 downto 40) xor retry_data_buffer(31)(47 downto 44);



  w3data <= w2data;
  w3adv <= w2adv;
  mptl_retry <= '0';



  
  -- Clock domain crossing FIFO
  -- Assumes ocp clock (250MHz) faster or equal to gtp clock
  local_rst : process (ocp_clk)
  begin  -- process local_rst
    if ocp_clk'event and ocp_clk = '1' then  -- rising clock edge
      rst <= not rst_reg2;
      rst_reg2 <= rst_reg;
      rst_reg <= '1';
    end if;
  end process local_rst;

gen_afifo: if not use_macro generate
  


fifo2 : afifo
    generic map (
      width  => 128,
      order  => 9,
      wpfl   => 8,
      wpfl2  => 32,
      rpel   => 2,
      family => family_virtex5,
      safety => false,
      fflush => false)
    port map (
      rst    => rst,
      wk     => gtp_clk,
      wd     => w3data,
      wadv   => w3adv,
      we     => open,
      wpf    => open,
      wpf2   => open,
      wovf   => open,
      wlevel => open,
      rk     => ocp_clk,
      rq     => r2data,
      radv   => r2adv,
      rflush => '0',
      re     => r2empty,
      rpe    => r2aempty,
      rf     => open,
      rudf   => open,
      rlevel => open);

end generate gen_afifo;


           gen_mfifo: if use_macro generate
             
           
FIFO_DUALCLOCK_MACRO_inst0 : FIFO_DUALCLOCK_MACRO
generic map (
DEVICE => DEVICE,
ALMOST_FULL_OFFSET => X"00C0",
ALMOST_EMPTY_OFFSET => X"0006",
DATA_WIDTH => 64,
FIFO_SIZE => "36Kb",
FIRST_WORD_FALL_THROUGH => TRUE)
port map (
ALMOSTEMPTY => ALMOSTEMPTY0,
ALMOSTFULL => ALMOSTFULL0,
DO => r2data(63 downto 0),
EMPTY => EMPTY0,
FULL => open,
RDCOUNT => RDCOUNT0,
RDERR => open,
WRCOUNT => WRCOUNT0,
WRERR => open,
DI => w3data(63 downto 0),
RDCLK => ocp_clk,
RDEN => r2adv,
RST => offline,
WRCLK => gtp_clk,
WREN => w3adv
);

FIFO_DUALCLOCK_MACRO_inst1 : FIFO_DUALCLOCK_MACRO
generic map (
DEVICE => DEVICE,
ALMOST_FULL_OFFSET => X"00C0",
ALMOST_EMPTY_OFFSET => X"0006",
DATA_WIDTH => 64,
FIFO_SIZE => "36Kb",
FIRST_WORD_FALL_THROUGH => TRUE)
port map (
ALMOSTEMPTY => ALMOSTEMPTY1,
ALMOSTFULL => ALMOSTFULL1,
DO => r2data(127 downto 64),
EMPTY => EMPTY1,
FULL => open,
RDCOUNT => RDCOUNT1,
RDERR => open,
WRCOUNT => WRCOUNT1,
WRERR => open,
DI => w3data(127 downto 64),
RDCLK => ocp_clk,
RDEN => r2adv,
RST => offline,
WRCLK => gtp_clk,
WREN => w3adv
);

r2empty <= EMPTY0 or EMPTY1;
r2aempty <= ALMOSTEMPTY0 or ALMOSTEMPTY1;

end generate gen_mfifo;

  

  process (ocp_clk) is
  begin  -- process
    if rising_edge(ocp_clk) then        -- rising clock edge
      r2data1 <= r2data;
      r2adv1 <= r2adv;      
      mptl128_data  <= r2data1;
      mptl128_valid <= r2adv1;
      offline2 <= offline;
      offline3 <= offline2;
      r2adv         <= ((not r2aempty) or ((not r2empty) and (not r2adv))) and (not offline3);
    end if;
  end process; 

  debug(15 downto 0) <= dbg_counter2 & dbg_counter;
  --debug(8 downto 0) <= "0" & rempty;
  --debug(17 downto 9) <= "000" &  cb_fail_counter;
  debug(18) <= ALMOSTFULL0;
  debug(19) <= error8b10bdet_i;
  debug(20) <= cb_fail;
  debug(21) <= channel_bond_rst;
  debug(22) <= retry_adv_buffer(31);
  debug(63 downto 32) <= retry_data_buffer(31)(31 downto 0);
  
end rtl;


-- Include an mptl128_gtp_buffer instance for backwards compatibility
-- of interface on 7K1 and 7V1 bridges 

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_misc.all;

library work;
use work.mptl.all;
use work.fifo.all;

entity mptl128_gtp_buffer is

  generic (
    four_gtps : boolean := true);
  port (
    -- OCP Interface
    ocp_clk       : in  std_logic;
    -- MPTL Multi-port output interface
    mptl128_data  : out std_logic_vector(127 downto 0);
    mptl128_valid : out std_logic;
    -- GTP Outputs
    gtp_clk       : in  std_logic;
    gtp_data      : in  std_logic_vector(127 downto 0);
    gtp_charisk   : in  std_logic_vector(15 downto 0);
    gtp_online    : out std_logic := '0';
    tx_offline    : in  std_logic;
    error8b10b    : in  std_logic;
    mptl_retry    : out std_logic);

end mptl128_gtp_buffer;

architecture rtl of mptl128_gtp_buffer is

begin
  
  mptl128_gtp_buffer_macro_1: entity work.mptl128_gtp_buffer_macro
    generic map (
      four_gtps => four_gtps,
      DEVICE    => "VIRTEX6")
    port map (
      ocp_clk       => ocp_clk,
      mptl128_data  => mptl128_data,
      mptl128_valid => mptl128_valid,
      gtp_clk       => gtp_clk,
      gtp_data      => gtp_data,
      gtp_charisk   => gtp_charisk,
      gtp_online    => gtp_online,
      tx_offline    => tx_offline,
      error8b10b    => error8b10b,
      debug         => open,
      mptl_retry    => mptl_retry);
  
end rtl;
    
