--
--

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_misc.all;

library work;
use work.adb3_ocp.all;
use work.mptl.all;
use work.mptl_pins.all;
use work.today.all;

library UNISIM;
use UNISIM.VCOMPONENTS.all;

entity mptl128_interface_target_ku is
  generic (
    enable_bridge_master : boolean := true;
    enable_target_master : boolean := false;
    enable_dma           : boolean := true);
  port (
    -- OCP Interface
    ocp_clk                  : in  std_logic;
    direct_slave_m2s         : out adb3_ocp_m2sT;
    direct_slave_s2m         : in  adb3_ocp_s2mT;
    dma_channels_m2s         : out adb3_ocp_m2sT_array(0 to 1);
    dma_channels_s2m         : in  adb3_ocp_s2mT_array(0 to 1);
    dma_abort                : in  std_logic_vector(1 downto 0);
    abmi_m2s                 : in  adb3_ocp_m2sT;
    abmi_s2m                 : out adb3_ocp_s2mT;
    -- MPTL Interface
    mptl_tx                  : out mptl_pins_type;
    mptl_rx                  : in  mptl_pins_type;
    mptl_clk                 : in  mptl_clk_type;
    mptl_clk_out             : out std_logic;
    -- Bridge to Target Sideband Signals
    mptl_target_configured_n : out std_logic;
    mptl_target_gtp_online_n : out std_logic;
    mptl_bridge_gtp_online_n : in  std_logic;
    -- MPTL64 Sideband Interface
    gpio_t2b                 : in  std_logic_vector(15 downto 0);
    gpio_b2t                 : out std_logic_vector(15 downto 0));

end mptl128_interface_target_ku;

architecture rtl of mptl128_interface_target_ku is

  constant NUMBER_OF_CHANNELS                   : natural := 9;
  constant DIRECT_SLAVE_CHANNEL                 : integer := 0;
  constant DMA_ENGINE0_CHANNEL                  : integer := 1;
  constant DMA_ENGINE1_CHANNEL                  : integer := 2;
  constant DMA_ENGINE2_CHANNEL                  : integer := 3;
  constant DMA_ENGINE3_CHANNEL                  : integer := 4;
  constant HOST_DIRECT_MASTER_CHANNEL           : integer := 5;
  constant BRIDGE_GENERIC_DIRECT_MASTER_CHANNEL : integer := 6;
  constant BRIDGE_MODEL_DIRECT_MASTER_CHANNEL   : integer := 7;
  constant MPTL_CONTROLLER_CHANNEL              : integer := 8;

  signal mclk, mclk_d2, mclk_bufr, ref_clk, gtp_clk, gtp_rclk, gtp_clk2x : std_logic;


  signal remote_last_packets   : std_logic_vector(127 downto 0) := (others => '0');
  signal local_last_packets    : std_logic_vector(127 downto 0) := (others => '0');
  signal local_current_packets : std_logic_vector(127 downto 0) := (others => '0');

  signal mptl128_data_send   : mptl128_data_array(0 to NUMBER_OF_CHANNELS-1);
  signal mptl128_valid_send  : std_logic_vector(NUMBER_OF_CHANNELS-1 downto 0);
  signal mptl128_accept_send : std_logic_vector(NUMBER_OF_CHANNELS-1 downto 0);

  signal mptl128_data_recv  : mptl128_data_array(0 to NUMBER_OF_CHANNELS-1);
  signal mptl128_valid_recv : std_logic_vector(NUMBER_OF_CHANNELS-1 downto 0);

  signal local_ocp_read0     : std_logic := '0';
  signal local_ocp_nfull0    : std_logic := '0';
  signal local_burst_length0 : std_logic_vector(11 downto 0);
  signal local_tag0          : std_logic_vector(7 downto 0);
  signal tag_mismatch0       : std_logic;


  signal local_ocp_read     : std_logic_vector(3 downto 0) := "0000";
  signal local_ocp_nfull    : std_logic_vector(3 downto 0) := "0000";
  type   local_burst_length_array is array (0 to 3) of std_logic_vector(11 downto 0);
  signal local_burst_length : local_burst_length_array;
  type   local_tag_array is array (0 to 3) of std_logic_vector(7 downto 0);
  signal local_tag          : local_tag_array;
  signal tag_mismatch       : std_logic_vector(3 downto 0);

  signal mptl128_data       : std_logic_vector(127 downto 0);
  signal mptl128_valid      : std_logic;
  signal mptl128_fifo_nfull : std_logic := '0';

  signal mptl128_data_g  : std_logic_vector(127 downto 0);
  signal mptl128_valid_g : std_logic;

  signal gtp_data_tx    : std_logic_vector(127 downto 0);
  signal gtp_charisk_tx : std_logic_vector(15 downto 0);
  signal gtp_data_rx    : std_logic_vector(127 downto 0);
  signal gtp_charisk_rx : std_logic_vector(15 downto 0);

  signal en_comma_align : std_logic := '0';
  signal powerdown      : std_logic := '0';
  signal pll_locked     : std_logic;

  signal dcm_rst,dcm_rst_r1           : std_logic := '0';
  signal locked_mclk       : std_logic := '0';
  signal mclk0, gtpclk2x_i : std_logic;
  signal fb_clk, gtp_clk_i : std_logic;

  signal local_gtp_online : std_logic;
  signal gtp_online       : std_logic;
  signal gtp_online1      : std_logic;
  signal gtp_online_r1    : std_logic;

  signal gtp_online_tck       : std_logic;
  signal gtp_online_tck1       : std_logic;
  signal gtp_online_ock       : std_logic;
  signal gtp_online_ock1       : std_logic;

  signal rx_reset  : std_logic_vector(7 downto 0) := (others => '0');
  signal rx_status : std_logic_vector(23 downto 0);
  signal rx_los    : std_logic_vector(15 downto 0);

  signal reset_count     : std_logic_vector(11 downto 0) := (others => '0');
  signal dcm_reset_count : std_logic_vector(27 downto 0) := X"0000000";
  signal dcm_reset_count2 : std_logic_vector(9 downto 0) := "1111000000";
  signal dcm_reset_count3 : std_logic_vector(2 downto 0) := (others => '0');

  signal rst, rst_reg, rst_rck, rst_rck1    : std_logic                     := '0';
  signal ref_clk_bufg    : std_logic;
  signal gtp_reset       : std_logic                     := '0';
  signal gtp_ready       : std_logic                     := '0';
  signal gtp_ready1       : std_logic                     := '0';
  signal gtp_ready2       : std_logic                     := '0';
  signal gtp_reset_count : std_logic_vector(23 downto 0) := INIT_GTP_RESET_7V1H;


  signal direct_slave_s2m_A   : adb3_ocp_s2mT;
  signal dma_channels_s2m_A   : adb3_ocp_s2mT_array(0 to 3);
  signal direct_masters_m2s_A : adb3_ocp_m2sT_array(0 to 2);
  signal direct_slave_s2m_B   : adb3_ocp_s2mT;
  signal dma_channels_s2m_B   : adb3_ocp_s2mT_array(0 to 3);
  signal direct_masters_m2s_B : adb3_ocp_m2sT_array(0 to 2);

  signal direct_slave_m2s_A : adb3_ocp_m2sT;
  signal dma_channels_m2s_A : adb3_ocp_m2sT_array(0 to 1);

  signal direct_slave_m2s_B : adb3_ocp_m2sT;
  signal dma_channels_m2s_B : adb3_ocp_m2sT_array(0 to 1);

  signal rlp,llp,rlp1,llp1 : std_logic_vector(7 downto 0);

  signal psent, prcvd : std_logic;

  signal gtp_status_i  : std_logic_vector(63 downto 0);
  signal gtp_status_ii : std_logic_vector(63 downto 0);

  signal dbg1, dbg2, dbg3, dbg4 : std_logic_vector(3 downto 0);

  signal los        : std_logic;
  signal los_r1     : std_logic;
  signal tx_offline : std_logic;

  signal los_count           : std_logic_vector(7 downto 0)  := (others => '0');
  signal los_offline         : std_logic                     := '0';
  signal en_commaalign_count : std_logic_vector(7 downto 0)  := (others => '0');
  signal ref_clk_locked      : std_logic;
  signal los_counter         : std_logic_vector(15 downto 0) := (others => '0');


  signal demux_err       : std_logic;
  signal demux_err_count : std_logic_vector(7 downto 0) := (others => '0');

  signal gtp_debug     : std_logic_vector(15 downto 0);
  signal big_los_count : std_logic_vector(23 downto 0) := (others => '0');

  signal clk_counter : std_logic_vector(31 downto 0) := (others => '0');
  signal packet_err  : std_logic;

  signal gtp_clk_p : std_logic;
  signal gtp_clk_n : std_logic;
  signal tx_p      : std_logic_vector(3 downto 0);
  signal tx_n      : std_logic_vector(3 downto 0);
  signal rx_p      : std_logic_vector(3 downto 0);
  signal rx_n      : std_logic_vector(3 downto 0);

  attribute shreg_extract                  : string;
  attribute shreg_extract of gtp_online    : signal is "no";
  attribute shreg_extract of gtp_online1   : signal is "no";
  attribute shreg_extract of gtp_online_r1 : signal is "no";

  attribute max_fanout               : string;
  attribute max_fanout of gtp_online : signal is "5";

  signal target_configured : std_logic;
  signal target_gtp_online : std_logic;
  signal bridge_gtp_online : std_logic;

  signal bridge_gtp_online_i  : std_logic;
  signal bridge_gtp_online_i1  : std_logic;
  signal bridge_gtp_online_i2  : std_logic;
  signal bridge_gtp_online_rck  : std_logic := '0';
  signal bridge_gtp_online_rck1  : std_logic := '0';
  signal bridge_gtp_online_sr : std_logic_vector(27 downto 0);

  signal retry_b2t  : std_logic;
  signal retry_t2b  : std_logic;
  signal error8b10b : std_logic := '0';


  signal gpio_in_i  : std_logic_vector(31 downto 0);
  signal gpio_out_i : std_logic_vector(31 downto 0);

  signal packet_words_left : std_logic_vector(4 downto 0);

  signal pkid1, pkid2, pkid3, pkid4 : std_logic_vector(7 downto 0);

  signal prcvd_counter   : std_logic_vector(15 downto 0) := (others => '0');
  signal pnrcvd          : std_logic                     := '0';
  signal prcvd_toggle    : std_logic                     := '0';
  signal prcvd_toggle_1    : std_logic                     := '0';
  signal prcvd_toggle_2    : std_logic                     := '0';
  signal prcvd_toggle_r1 : std_logic                     := '0';
  signal prcvd_toggle_r2 : std_logic                     := '0';
  signal prcvd_mclk      : std_logic                     := '0';

  signal short_offline_counter : std_logic_vector(6 downto 0) := (others => '0');
  signal short_offline_detect  : std_logic                    := '0';

  signal status_counter : std_logic_vector(19 downto 0) := (others => '0');




  signal bridge_online_counter : std_logic_vector(13 downto 0) := (others => '0');
  signal bridge_online_stable  : std_logic                     := '0';

  signal dma_abort_i : std_logic_vector(1 downto 0) := "00";

  signal abmi_s2m_B, abmi_s2m_A : adb3_ocp_s2mT;

  signal gtp_online_d2 : std_logic := '0';
  signal gtp_online_d3 : std_logic := '0'; 
  signal gtp_debug2 : std_logic_vector(139 downto 0);
  signal err_kchar : std_logic;


  attribute ASYNC_REG : boolean;
  attribute ASYNC_REG of bridge_gtp_online_rck1,rst_rck1,rlp1,llp1,gtp_ready1,prcvd_toggle_2,gtp_online_d3,bridge_gtp_online_i1,gtp_online_tck1,gtp_online_ock1: signal is TRUE;

  
begin  -- rtl

  
  
  bridge_gtp_online        <= not mptl_bridge_gtp_online_n;

  mptl_tx.p(tx_p'length-1 downto 0) <= tx_p;
  mptl_tx.n(tx_p'length-1 downto 0) <= tx_n;
  rx_p                              <= mptl_rx.p(tx_p'length-1 downto 0);
  rx_n                              <= mptl_rx.n(tx_p'length-1 downto 0);

  gtp_clk_p <= mptl_clk.p;
  gtp_clk_n <= mptl_clk.n;

  mptl_clk_out <= gtp_clk;

  --
  -- Connect Direct Slave OCP to MPTL MUX input #0
  --

  gen_bridge_master : if enable_bridge_master generate
    
    
    direct_slave_m2s.RespAccept  <= direct_slave_m2s_A.RespAccept;
    direct_slave_m2s.Cmd         <= direct_slave_m2s_B.Cmd;
    direct_slave_m2s.Addr        <= direct_slave_m2s_B.Addr;
    direct_slave_m2s.BurstLength <= direct_slave_m2s_B.BurstLength;
    direct_slave_m2s.Data        <= direct_slave_m2s_B.Data;
    direct_slave_m2s.DataByteEn  <= direct_slave_m2s_B.DataByteEn;
    direct_slave_m2s.DataValid   <= direct_slave_m2s_B.DataValid;
    direct_slave_m2s.Tag         <= direct_slave_m2s_B.Tag;



    mptl_resp_a2m_br0 : adb3_ocp_resp_to_mptl128
      generic map (
        channel => 0)
      port map (
        ocp_clk            => ocp_clk,
        ocp_m2s            => direct_slave_m2s_A,
        ocp_s2m            => direct_slave_s2m,
        mptl128_data       => mptl128_data_send(DIRECT_SLAVE_CHANNEL),
        mptl128_valid      => mptl128_valid_send(DIRECT_SLAVE_CHANNEL),
        mptl128_accept     => mptl128_accept_send(DIRECT_SLAVE_CHANNEL),
        online             => gtp_online_ock,
        remote_last_packet => remote_last_packets(8*DIRECT_SLAVE_CHANNEL+7 downto 8*DIRECT_SLAVE_CHANNEL),
--      current_packet_o => local_current_packets(8*DIRECT_SLAVE_CHANNEL+7 downto 8*DIRECT_SLAVE_CHANNEL),
        local_ocp_read     => local_ocp_read0,
        local_burst_length => local_burst_length0,
        local_tag          => local_tag0,
        local_ocp_nfull    => local_ocp_nfull0,
        tag_mismatch       => tag_mismatch0);

    mptl_m2a_br0 : mptl128_to_adb3_ocp
      port map (
        ocp_clk            => ocp_clk,
        ocp_m2s            => direct_slave_m2s_B,
        ocp_s2m            => direct_slave_s2m,
        mptl128_data       => mptl128_data_recv(DIRECT_SLAVE_CHANNEL),
        mptl128_valid      => mptl128_valid_recv(DIRECT_SLAVE_CHANNEL),
        online             => gtp_online_ock,
        remote_last_packet => local_last_packets(8*DIRECT_SLAVE_CHANNEL+7 downto 8*DIRECT_SLAVE_CHANNEL),
        local_ocp_read     => local_ocp_read0,
        local_burst_length => local_burst_length0,
        local_tag          => local_tag0,
        local_ocp_nfull    => local_ocp_nfull0);

    --
    -- Connect DMA Channels #0-3 OCP to MPTL MUX input #1-4
    --

    gen_dma1 : if enable_dma generate
      
      
      
      gen_dma : for i in 0 to 1 generate
        mptl_resp_a2m_br_i : adb3_ocp_resp_to_mptl128
          generic map (
            channel => DMA_ENGINE0_CHANNEL+i)
          port map (
            ocp_clk            => ocp_clk,
            ocp_m2s            => dma_channels_m2s_A(i),
            ocp_s2m            => dma_channels_s2m(i),
            mptl128_data       => mptl128_data_send(DMA_ENGINE0_CHANNEL+i),
            mptl128_valid      => mptl128_valid_send(DMA_ENGINE0_CHANNEL+i),
            mptl128_accept     => mptl128_accept_send(DMA_ENGINE0_CHANNEL+i),
            online             => gtp_online_ock,
            remote_last_packet => remote_last_packets(8*(DMA_ENGINE0_CHANNEL+i)+7 downto 8*(DMA_ENGINE0_CHANNEL+i)),
--        current_packet_o => local_current_packets(8*(DMA_ENGINE0_CHANNEL+i)+7 downto 8*(DMA_ENGINE0_CHANNEL+i)),
            local_ocp_read     => local_ocp_read(i),
            local_burst_length => local_burst_length(i),
            local_tag          => local_tag(i),
            local_ocp_nfull    => local_ocp_nfull(i),
            tag_mismatch       => tag_mismatch(i));

        mptl_m2a_br_i : mptl128_to_adb3_ocp
          port map (
            ocp_clk            => ocp_clk,
            ocp_m2s            => dma_channels_m2s_B(i),
            ocp_s2m            => dma_channels_s2m(i),
            mptl128_data       => mptl128_data_recv(DMA_ENGINE0_CHANNEL+i),
            mptl128_valid      => mptl128_valid_recv(DMA_ENGINE0_CHANNEL+i),
            online             => gtp_online_ock,
            remote_last_packet => local_last_packets(8*(DMA_ENGINE0_CHANNEL+i)+7 downto 8*(DMA_ENGINE0_CHANNEL+i)),
            local_ocp_read     => local_ocp_read(i),
            local_burst_length => local_burst_length(i),
            local_tag          => local_tag(i),
            local_ocp_nfull    => local_ocp_nfull(i)); 

        dma_channels_m2s(i).RespAccept  <= dma_channels_m2s_A(i).RespAccept;
        dma_channels_m2s(i).Cmd         <= dma_channels_m2s_B(i).Cmd;
        dma_channels_m2s(i).Addr        <= dma_channels_m2s_B(i).Addr;
        dma_channels_m2s(i).BurstLength <= dma_channels_m2s_B(i).BurstLength;
        dma_channels_m2s(i).Data        <= dma_channels_m2s_B(i).Data;
        dma_channels_m2s(i).DataByteEn  <= dma_channels_m2s_B(i).DataByteEn;
        dma_channels_m2s(i).DataValid   <= dma_channels_m2s_B(i).DataValid;
        dma_channels_m2s(i).Tag         <= dma_channels_m2s_B(i).Tag;

        dma_abort_i <= dma_abort;
      end generate gen_dma;
    end generate gen_dma1;
    
  end generate gen_bridge_master;


  --
  -- Connect MPTL MUX inputs #5-7 to Direct Master OCP Ports
  --

  gen_tgt_master : if enable_target_master generate


    mptl_a2m_br0 : adb3_ocp_to_mptl128
      generic map (
        channel                => BRIDGE_MODEL_DIRECT_MASTER_CHANNEL,
        limit_inflight_packets => true)
      port map (
        ocp_clk            => ocp_clk,
        ocp_m2s            => abmi_m2s,
        ocp_s2m            => abmi_s2m_A,
        mptl128_data       => mptl128_data_send(BRIDGE_MODEL_DIRECT_MASTER_CHANNEL),
        mptl128_valid      => mptl128_valid_send(BRIDGE_MODEL_DIRECT_MASTER_CHANNEL),
        mptl128_accept     => mptl128_accept_send(BRIDGE_MODEL_DIRECT_MASTER_CHANNEL),
        online             => gtp_online_ock,
        remote_last_packet => remote_last_packets(8*BRIDGE_MODEL_DIRECT_MASTER_CHANNEL+7 downto 8*BRIDGE_MODEL_DIRECT_MASTER_CHANNEL));

    mptl_resp_m2a_br0 : mptl128_to_adb3_ocp_resp
      port map (
        ocp_clk            => ocp_clk,
        ocp_m2s            => abmi_m2s,
        ocp_s2m            => abmi_s2m_B,
        mptl128_data       => mptl128_data_recv(BRIDGE_MODEL_DIRECT_MASTER_CHANNEL),
        mptl128_valid      => mptl128_valid_recv(BRIDGE_MODEL_DIRECT_MASTER_CHANNEL),
        online             => gtp_online_ock,
--      debug => open,
        remote_last_packet => local_last_packets(8*BRIDGE_MODEL_DIRECT_MASTER_CHANNEL+7 downto 8*BRIDGE_MODEL_DIRECT_MASTER_CHANNEL));

    abmi_s2m.Data       <= abmi_s2m_B.Data;
    abmi_s2m.Resp       <= abmi_s2m_B.Resp;
    abmi_s2m.Tag        <= abmi_s2m_B.Tag;
    abmi_s2m.CmdAccept  <= abmi_s2m_A.CmdAccept;
    abmi_s2m.DataAccept <= abmi_s2m_A.DataAccept;

    
  end generate gen_tgt_master;


  --
  -- MPTL Controller connected to Channel #4
  --

  ctlr0 : mptl128_controller
    generic map (
      channel => MPTL_CONTROLLER_CHANNEL)
    port map (
      clk                 => ocp_clk,
      mptl128_data_out    => mptl128_data_send(MPTL_CONTROLLER_CHANNEL),
      mptl128_valid_out   => mptl128_valid_send(MPTL_CONTROLLER_CHANNEL),
      mptl128_accept_out  => mptl128_accept_send(MPTL_CONTROLLER_CHANNEL),
      mptl128_data_in     => mptl128_data_recv(MPTL_CONTROLLER_CHANNEL),
      mptl128_valid_in    => mptl128_valid_recv(MPTL_CONTROLLER_CHANNEL),
      local_last_packets  => local_last_packets,
      remote_last_packets => remote_last_packets,
      gpio_in             => gpio_in_i,
      gpio_out            => gpio_out_i,
      gpio_status         => gtp_status_ii(31 downto 0),
      remote_status       => open,
      packet_sent         => psent,
      packet_received     => prcvd,
      online              => gtp_online_ock);

  gpio_in_i <= X"000" & "00" & dma_abort_i & gpio_t2b;
  gpio_b2t  <= gpio_out_i(15 downto 0);

  --
  -- MPTL MUX and DeMUX
  -- 

  mptl_mux0 : mptl128_packet_mux
    generic map (
      number_of_ports => NUMBER_OF_CHANNELS)
    port map (
      clk                => ocp_clk,
      online             => '1',
      mptl128_data_in    => mptl128_data_send,
      mptl128_valid_in   => mptl128_valid_send,
      mptl128_accept_in  => mptl128_accept_send,
      mptl128_data_out   => mptl128_data,
      mptl128_valid_out  => mptl128_valid,
      mptl128_fifo_nfull => mptl128_fifo_nfull);

  mptl_demux0 : mptl128_packet_demux
    generic map (
      number_of_ports => NUMBER_OF_CHANNELS)
    port map (
      clk               => ocp_clk,
      online            => '1',
      mptl128_data_out  => mptl128_data_recv,
      mptl128_valid_out => mptl128_valid_recv,
      mptl128_data_in   => mptl128_data_g,
      mptl128_valid_in  => mptl128_valid_g,
      error_detected    => demux_err);


  --
  -- Buffer Data between MUX and GTP TX
  --
  
  
  packet_buffer0 : mptl128_packet_buffer
    generic map (
      DEVICE => "7SERIES")
    port map (
      ocp_clk          => ocp_clk,
      mptl128_data_in  => mptl128_data,
      mptl128_valid_in => mptl128_valid,
      mptl128_nfull    => mptl128_fifo_nfull,
      mptl_retry       => retry_b2t,
      gtp_clk          => gtp_clk,
      gtp_data         => gtp_data_tx,
      gtp_charisk      => gtp_charisk_tx,
      gtp_online       => gtp_online_tck);


  --
  -- Buffer Data between GTP RX and DeMUX
  --
  
  gtp_buffer0 : mptl128_gtp_buffer_macro
    generic map (
      DEVICE => "7SERIES")
    port map (
      ocp_clk       => ocp_clk,
      mptl128_data  => mptl128_data_g,
      mptl128_valid => mptl128_valid_g,
      gtp_clk       => gtp_rclk,
      gtp_data      => gtp_data_rx,
      gtp_charisk   => gtp_charisk_rx,
      gtp_online    => local_gtp_online,
      tx_offline    => tx_offline,
      error8b10b    => los,
      debug         => gtp_debug2,
      mptl_retry    => retry_t2b);

err_kchar <= gtp_debug2(19);

  --
  -- GTP Interface
  --
  
  gtp_if0 : entity work.mptl128_gthu_interface
    port map (
      clkin           => mclk,
      pllrst          => dcm_rst,
      txrst           => gtp_reset,
      ref_clk         => ref_clk,
      rref_clk        => gtp_rclk,
      gtp_clk         => gtp_clk,
      drpclk          => mclk_bufr,
      gtp_data_in     => gtp_data_tx,
      gtp_charisk_in  => gtp_charisk_tx,
      gtp_data_out    => gtp_data_rx,
      gtp_charisk_out => gtp_charisk_rx,
      tx_p            => tx_p,
      tx_n            => tx_n,
      rx_p            => rx_p,
      rx_n            => rx_n,
      loopback        => "000",
      config          => "0000000000000000",
      rx_reset        => rx_reset,
      en_comma_align  => en_comma_align,
      powerdown       => powerdown,
      pll_locked      => pll_locked,
      rx_status       => rx_status,
      rx_loss_of_sync => rx_los);


  --target_configured <= '1';
  powerdown <= '0';


  process (ocp_clk)
  begin
    if rising_edge(ocp_clk) then
      gtp_online_ock1 <= gtp_online;
      gtp_online_ock <= gtp_online_ock1;
    end if;      
  end process;  
  
  gtp_control1 : process (gtp_clk)
  begin
    if rising_edge(gtp_clk) then
      bridge_gtp_online_sr <= bridge_gtp_online_sr(26 downto 0) & bridge_gtp_online;
      bridge_gtp_online_i  <= bridge_gtp_online_sr(27) or bridge_gtp_online_sr(0);
      retry_b2t            <= (not bridge_gtp_online_sr(0)) and bridge_gtp_online_sr(27);
      gtp_online_tck1 <= gtp_online;
      gtp_online_tck <= gtp_online_tck1;
      
    end if;
  end process;

  
  gtp_control0 : process (gtp_rclk)
  begin  -- process gtp_control0
    if gtp_rclk'event and gtp_rclk = '1' then  -- rising clock edge
      rst_rck1 <= rst;
      rst_rck <= rst_rck1;
      gtp_online     <= gtp_online1;
      bridge_gtp_online_rck1 <= bridge_gtp_online_i;
      bridge_gtp_online_rck <= bridge_gtp_online_rck1;
      gtp_online1    <= local_gtp_online and bridge_gtp_online_rck;
      gtp_online_r1  <= gtp_online;
      en_comma_align <= not gtp_online_r1;

      target_gtp_online <= local_gtp_online;-- and not retry_t2b;
      mptl_target_gtp_online_n <= not target_gtp_online;
 

      -- If Powered Up but Local GTP not online after
      -- ~4000 clock cycles, reset GTP RX
      if (los_count = "11111111" and rx_reset(0) = '0') or pnrcvd = '1' or short_offline_detect = '1' then
        reset_count <= X"FF8";
        --elsif short_offline_detect = '1' then
        --  reset_count <= reset_count+X"04000";
      elsif local_gtp_online = '0' then
        reset_count <= reset_count+1;
--      else
--        reset_count <= (others => '0');
      end if;
      if AND_reduce(reset_count(11 downto 3)) = '1' then
        rx_reset <= (others => '1');
      else
        rx_reset <= (others => '0');
      end if;

--       gtp_status_i(23 downto 12) <= rx_status(11 downto 0);
--       gtp_status_i(31 downto 24) <= rx_los(7 downto 0);
--       gtp_status_i(55 downto 44) <= rx_status(23 downto 12);
--       gtp_status_i(63 downto 56) <= rx_los(15 downto 8);
--       gtp_status_i(0)            <= '1';
--       gtp_status_i(1)            <= bridge_gtp_online;
--       gtp_status_i(2)            <= local_gtp_online;
--       gtp_status_i(3)            <= locked_mclk;
--       gtp_status_i(11 downto 4)  <= reset_count(11 downto 4);
      gtp_status_i(0) <= '1';
      gtp_status_i(1) <= bridge_gtp_online_rck;
      gtp_status_i(2) <= local_gtp_online;
      gtp_status_i(3) <= locked_mclk;

      status_counter <= status_counter+1;

      rlp1 <= remote_last_packets(7 downto 0);
      llp1 <= local_last_packets(7 downto 0);
      rlp <= rlp1;
      llp <= llp1;
      case status_counter(19 downto 18) is
        when "00" =>
          --gtp_status_i(31 downto 8) <= big_los_count;
          gtp_status_i(31 downto 8) <= rlp(7 downto 0) & llp(7 downto 0) & big_los_count(7 downto 0);
          gtp_status_i(7 downto 4)  <= "0000";
        when "11" =>
          gtp_status_i(31 downto 8) <= TODAYS_DATE(31 downto 16) & TODAYS_DATE(7 downto 0);
          gtp_status_i(7 downto 4)  <= "1111";
        when "10" =>
          gtp_status_i(31 downto 8) <= TODAYS_DATE(63 downto 40);
          gtp_status_i(7 downto 4)  <= "1110";
        when others =>
          gtp_status_i(31 downto 8) <= big_los_count;
          gtp_status_i(7 downto 4)  <= "0000";
      end case;


      --gtp_status_i(31 downto 8) <= rx_status(23 downto 0);
      --gtp_status_i(31 downto 8) <= big_los_count;
      --gtp_status_i(7 downto 4) <= "1110";
      --gtp_status_i(31 downto 8) <=pkid2 & pkid1 & remote_last_packets(23 downto 16);
      gtp_status_ii <= gtp_status_i;

      if rx_los(15) = '0' and rx_los(13) = '0' and rx_los(11) = '0' and rx_los(9) = '0' and rx_los(7) = '0' and rx_los(5) = '0' and rx_los(3) = '0' and rx_los(1) = '0' then
        los <= '0';
      else
        los <= '1';
      end if;

      los_r1 <= los;

      if local_gtp_online = '1' then
        short_offline_counter <= (others => '0');
        if short_offline_counter(5 downto 0) /= "000000" then
          short_offline_detect <= '1';
        else
          short_offline_detect <= '0';
        end if;
      else
        if short_offline_counter(6) = '0' then
          short_offline_counter <= short_offline_counter+1;
        end if;
        short_offline_detect <= '0';
      end if;

      if los = '1' and los_r1 = '0' and rx_reset(0) = '0' then  -- detect rising edge of los
        los_count <= los_count+1;
        if AND_reduce(big_los_count) = '0' and los_r1 = '0' then
          big_los_count <= big_los_count+1;
        end if;
        -- if los_count = "11111111" then
        --   rx_reset <= (others => '1');
        -- end if;
      elsif AND_reduce(los_counter) = '1' then
        -- If los infrequent don't count up and reset
        if OR_reduce(los_count) = '1' then
          los_count <= los_count-1;
        end if;
      elsif rx_reset(0) = '1' then
        los_count <= (others => '0');
      end if;

      los_counter <= los_counter+1;

      gtp_ready1 <= gtp_ready;
      gtp_ready2 <= gtp_ready1;
      tx_offline <= rx_reset(0) or not gtp_ready2;  -- or not bridge_gtp_online; --los and los_r1;

      -- Only take link offline every 16 errors 
      los_offline <= '0';  --los and OR_reduce(los_count(7 downto 4)) and AND_reduce(los_count(3 downto 0));   

      if demux_err = '1' then
        if AND_reduce(demux_err_count) = '0' then
          demux_err_count <= demux_err_count +1;
        end if;
      end if;

      if rst_rck = '1' then
        reset_count  <= (others => '0');
        rx_reset     <= (others => '0');
        gtp_status_i <= (others => '0');
      end if;

      if prcvd_mclk = '1' or gtp_online = '0' then
        prcvd_counter <= (others => '0');
      else
        prcvd_counter <= prcvd_counter +1;
      end if;
      pnrcvd <= AND_reduce(prcvd_counter(15 downto 4));

      prcvd_toggle_2 <= prcvd_toggle_1;
      prcvd_toggle_r1 <= prcvd_toggle_2;
      prcvd_toggle_r2 <= prcvd_toggle_r1;
      prcvd_mclk      <= prcvd_toggle_r2 xor prcvd_toggle_r1;
    end if;
  end process gtp_control0;

  process (ocp_clk) is
  begin  -- process
    if rising_edge(ocp_clk) then        -- rising clock edge
      if prcvd = '1' then
        prcvd_toggle <= not prcvd_toggle;
      end if;
      prcvd_toggle_1 <= prcvd_toggle;
    end if;
  end process;

  bufr_mclk : BUFG_GT
    port map(
      I       => mclk_d2,
      DIV     => "000",
      CLRMASK => '0',
      CLR     => '0',
      CEMASK  => '0',
      CE      => '1',
      O       => mclk_bufr);



  -- Local Reset on Power Up
  local_rst : process (mclk_bufr)
  begin  -- process local_rst
    if mclk_bufr'event and mclk_bufr = '1' then
      
      if dcm_reset_count2 /= "1111111111" then
        dcm_reset_count2 <= dcm_reset_count2+1;
        dcm_rst         <= '1';
        rst             <= '1';
      else
        dcm_rst <= '0';
        rst     <=  not pll_locked;
        if dcm_reset_count(23) = '1' or (err_kchar = '1' and gtp_online_d2 = '1') then
          dcm_reset_count2 <= (others => '0');
        end if;
      end if;
     
            
      if dcm_rst = '1' then
        dcm_reset_count <= (others => '0');
      else
        if gtp_online_d2 = '0' then
          dcm_reset_count <=  dcm_reset_count +1;
        end if;
      end if;
      gtp_online_d3 <= gtp_online_r1;
      gtp_online_d2 <= gtp_online_d3;
  
      target_configured <= not rst;     --pll_locked and gtp_ready;
      mptl_target_configured_n <= not target_configured;
      
    end if;
  end process local_rst;

  gtp_reset0 : process (mclk_bufr)
  begin  -- process gtp_control0
    if mclk_bufr'event and mclk_bufr = '1' then  -- rising clock edge
      if rst = '1' then
        gtp_reset_count       <= X"FF0000";
        gtp_reset             <= '1';
        gtp_ready             <= '0';
        bridge_online_counter <= (others => '0');      
      else
        if AND_reduce(gtp_reset_count(23 downto 16)) = '1' then
          gtp_reset_count <= gtp_reset_count+1;
        elsif bridge_gtp_online_i2 = '0' then
          if bridge_online_stable = '0' then
            gtp_reset_count <= gtp_reset_count+1;
          else
            gtp_reset_count <= X"FF0000";
          end if;
        end if;
        if AND_reduce(gtp_reset_count(23 downto 4)) = '1' then
          gtp_reset <= '1';
        
        else
          gtp_reset <= '0';
        end if;
        gtp_ready <= not AND_reduce(gtp_reset_count(23 downto 16));

        if gtp_reset = '1' then
          bridge_online_counter <= (others => '0');
        elsif bridge_gtp_online_i2 = '1' then
          if bridge_online_counter(10) = '0' then
            bridge_online_counter <= bridge_online_counter +1;
          end if;
        end if;
        bridge_online_stable <= bridge_online_counter(10);
      end if;
      bridge_gtp_online_i1 <= bridge_gtp_online_i;
      bridge_gtp_online_i2 <= bridge_gtp_online_i1;
    end if;
  end process gtp_reset0;



  -----------------------------------------------------------------------------
  -- Generate PLL and clocking for GTPs
  -----------------------------------------------------------------------------

  q3_clk0_refclk_ibufds_i : IBUFDS_GTE3
    generic map
    (
      -- CLKSWING_CFG => "11" -- see answer record 43339
      REFCLK_HROW_CK_SEL => "01"
      )
    port map
    (
      O     => mclk,
      ODIV2 => mclk_d2,
      CEB   => '0',
      I     => gtp_clk_p,
      IB    => gtp_clk_n
      );

  


  gtp_clk <= ref_clk;

  
end rtl;
