--------------------------------------------------------------------------------
--
-- Module Name      : mptl128_packet_buffer
--
--                    Copyright 2009 by
--                    Alpha Data Parallel Systems Ltd.
--                    All Rights Reserved.
--
-- Original Author  : Andrew McCormick
-- Created          : 
--
-- Modified By      : 
-- Date             :
-- Change Notes     :     
--
-- Description      : Buffer for storing packets before transmission across GTP link
--
--
-- Dependencies     : mptl,fifo
--
-- Disclaimer       : THESE DESIGNS ARE PROVIDED "AS IS" WITH NO WARRANTY
--                    WHATSOEVER AND ALPHA DATA SPECIFICALLY DISCLAIMS ANY
--                    WARRANTIES IMPLIED OF MERCHANTABILITY, FITNESS FOR A
--                    PARTICULAR PURPOSE, OR AGAINST INFRINGEMENT.
--
--------------------------------------------------------------------------------


library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_misc.all;

library work;
use work.mptl.all;
use work.fifo.all;


entity mptl128_packet_buffer is

  generic (
    DEVICE : string := "VIRTEX6");
  port (
    -- OCP Interface
    ocp_clk         : in  std_logic;
    -- MPTL Multi-port output interface
    mptl128_data_in  : in  std_logic_vector(127 downto 0);
    mptl128_valid_in : in  std_logic;
    mptl128_nfull    : out std_logic;
    mptl_retry       : in  std_logic;
    -- GTP Outputs
    gtp_clk         : in  std_logic;
    gtp_data        : out std_logic_vector(127 downto 0);
    gtp_charisk     : out std_logic_vector(15 downto 0);
    gtp_online      : in  std_logic);

end mptl128_packet_buffer;

architecture rtl of mptl128_packet_buffer is

  constant K28_5 : std_logic_vector(7 downto 0) := X"BC";
  constant idleD : std_logic_vector(7 downto 0) := X"95";

  signal radv    : std_logic := '0';
  signal rempty  : std_logic;
  signal raempty  : std_logic;
  signal rlevel  : std_logic_vector(8 downto 0);
  signal rlevel1 : std_logic_vector(15 downto 0);
  signal rdata   : std_logic_vector(127 downto 0);

  signal rst, rst_reg,rst_reg2 : std_logic := '0';


  signal rdata_reg   : std_logic_vector(127 downto 0) := (others => '0');
  signal radv_reg    : std_logic := '0';

  type retry_data_buffer_type is array (0 to 127) of std_logic_vector(127 downto 0);
  type retry_charisk_buffer_type is array (0 to 127) of std_logic_vector(15 downto 0);

  signal retry_data_buffer : retry_data_buffer_type := (others => (others => '0'));
  signal retry_charisk_buffer : retry_charisk_buffer_type := (others => (others => '0'));
  
  signal gtp_data_i        : std_logic_vector(127 downto 0);
  signal gtp_charisk_i     : std_logic_vector(15 downto 0);

  signal retry_count : std_logic_vector(6 downto 0) := (others => '0');
  signal retry,retry_reg,retry_reg2 : std_logic := '0';

  signal retry_pending : std_logic := '0';
  signal retry_start : std_logic;
  
  signal packet_words_left : std_logic_vector(4 downto 0) := (others => '0');

  attribute ASYNC_REG : boolean;
  attribute ASYNC_REG of rst_reg: signal is TRUE;
begin  -- rtl


    local_rst : process (gtp_clk)
    begin  -- process local_rst
      if gtp_clk'event and gtp_clk = '1' then  -- rising clock edge
        rst <= '0'; --not rst_reg;     
        rst_reg <= not gtp_online;       
      end if;
    end process local_rst;

    fifo0 : afifo
      generic map (
        width  => 128,
        order  => 9,
        wpfl   => 192,
        wpfl2  => 32,
        rpel   => 2,
        family => family_virtex5,
        safety => false,
        fflush => false)
      port map (
        rst    => rst,
        wk     => ocp_clk,
        wd     => mptl128_data_in,
        wadv   => mptl128_valid_in,
        we     => open,
        wpf    => mptl128_nfull,
        wpf2   => open,
        wovf   => open,
        wlevel => open,
        rk     => gtp_clk,
        rq     => rdata,
        radv   => radv,
        rflush => rst_reg,   -- '0'
        re     => rempty,
        rpe    => raempty,
        rf     => open,
        rudf   => open,
        rlevel => rlevel1); 
 
  rlevel <= rlevel1(8 downto 0);

  output_data : process (gtp_clk)
  begin  -- process output_data
    if gtp_clk'event and gtp_clk = '1' then  -- rising clock edge
      rdata_reg <= rdata;
      radv_reg <= radv;
      if retry_reg2 = '1' then
        gtp_data_i <= retry_data_buffer(126);
        gtp_charisk_i <= retry_charisk_buffer(126);
      elsif radv_reg = '0' then         
        gtp_data_i    <= idleD & K28_5 & idleD & K28_5 & idleD & K28_5 & idleD & K28_5 & idleD & K28_5 & idleD & K28_5 & idleD & K28_5 & idleD & K28_5;
        gtp_charisk_i <= "0101010101010101";  
      else
        gtp_data_i    <= rdata_reg;
        gtp_charisk_i <= (others => '0');
      end if;

      radv <= ((not raempty) or ((not rempty) and (not radv))) and not (retry or retry_start);

      for i in 1 to 127 loop
        retry_data_buffer(i) <= retry_data_buffer(i-1);
        retry_charisk_buffer(i) <= retry_charisk_buffer(i-1);        
      end loop;  -- i

      retry_data_buffer(0) <= gtp_data_i;
      retry_charisk_buffer(0) <= gtp_charisk_i;

      
      
      if OR_reduce(retry_count) = '0' then      
        if retry_start = '1' and gtp_online = '1' then
          retry_count <= (others => '1');
          retry <= '1';
          retry_pending <= '0';
        else
          retry <= '0';
          if mptl_retry = '1' then
            retry_pending <= '1';
          end if;
        end if;
      else
        retry_count <= retry_count-1; 
      end if;
      
      retry_reg <= retry and not retry_count(6);
      retry_reg2 <= retry_reg;

      if gtp_online = '0' then
        packet_words_left <= (others => '0');
      elsif packet_words_left = "00000" or packet_words_left = "00001" then
        if radv = '1' then           
          packet_words_left <= rdata(12 downto 8);           
        end if;
      else
        if radv  = '1' then
          packet_words_left <= packet_words_left-1;
        end if;    
      end if;
    end if;
  end process output_data;

  retry_start <= (mptl_retry or retry_pending) when packet_words_left(4 downto 1) = "0000" or (packet_words_left = "00010" and radv = '1')
                 else '0';

    
  gtp_data <= gtp_data_i;
  gtp_charisk <= gtp_charisk_i;
  
end rtl;
